
/* USB Driver Stack v$Revision: 1.65 $ 
 *
 * (C) 1997 by Iaky Prez Gonzlez <inaky@peloncho.fis.ucm.es>
 *
 * Distributed under the terms of the GNU Public License Version 2
 * See files COPYRIGHT and COPYING at the root distribution directory
 *
 * $Id: generic.c,v 1.65 1998/11/14 17:12:14 inaky Exp $
 *
 * Changelog at the bottom 
 *
 *  USB generic stuff
 *
 */

#include <uusbd/usb.h>
#include <uusbd/usbd.h>
#include <uusbd/usb-hub.h>
#include <uusbd/usb-hcd.h>

#define GDLF_IDENT          "USB"
#define GDLF_LOCAL          /* Define if want local settings */
#define GDLF_DEF_LOGLEVEL   9
#define GDLF_ASSERTIONS     /* want assertions (#undef if not) */
#undef GDLF_FNTRACE        /* want function traces */
#define GDLF_LOCATIONS      /* want locations in log messages */
#include <uusbd/gdlf.h>

    /* Methods to initialize/clear ts_usb_* structures, arrays of
     * these structures and arrays of pointers to these structures */

    /* ts_usb_xx_init(), ts_usb_xx_a_init()
     *
     * All these functions share a common behaviour (change xx with
     * the data type). ts_usb_xx_init() initializes a single
     * structure, ts_usb_xx_a_init() initializes an array of these
     * structures, ts_usb_xx_pa_init() initializes an array of
     * pointers to these structures.
     *
     * They are suppossed to leave the structure in a known, fully
     * resetter and resource free state, will all the items (where
     * posible) set to the invalid value.
     *
     * Parameters:  
     *  ts_usb_xx*  [ts_usb_xx_[a_]init()] Address of the structure to
     *              initialize. It will *overwrite* everything in the
     *              struct, without previously releasing anything. If
     *              NULL, space will be allocated with kmalloc() for
     *              it.
     *  ts_usb_xx** [ts_usb_xx_pa_init()] The same as ts_usb_xx* in
     *              the other set of functions, but in the form of an
     *              array of pointers :)
     *  unsigned    [ts_usb_xx_a_init()] size of the array of
     *              structures to initialize.
     *
     * Returns:
     *  ts_usb_xx*  Pointer to the initialized (and/or allocated)
     *              structure/array of structures. NULL on error
     *              (normally memory allocation error).
     */


    /* ts_usb_xx_release(), ts_usb_xx_a_release(), ts_usb_xx_pa_release()
     *
     * All these functions share a common behaviour (change xx with
     * the data type). ts_usb_xx_release() releases THE CONTENTS of a
     * single structure, ts_usb_xx_a_release() releases an array of these
     * structures. Beware that the space allocated for this structures
     * is not released, JUST THE CONTENTS are released.
     *
     * Parameters:  
     *  ts_usb_xx*  [ts_usb_xx_[a_]free()] Address of the
     *              structure/array to release. It will release
     *              everything in the struct and clear it to
     *              known-reset state values. If NULL, nothing will be
     *              done.
     *  ts_usb_xx** [ts_usb_xx_pa_release()] pointer to the array of
     *              pointers to release. Same behaviour as ts_usb_xx*
     *              above.
     *  unsigned    [ts_usb_xx_[p]a_release()] size of the array of
     *              structures/pointers to release.
     *
     * Returns:
     *  void
     */


#define FREE_PTR(p)           \
    if ((p) != NULL)          \
    {                         \
        usb_kfree ((p));      \
        (p) = NULL;           \
    }

    /* ts_usb_addr */

ts_usb_addr* ts_usb_addr_init (ts_usb_addr* addr)
{
    gdlf_fntrace ("ts_usb_addr_init (0x%p)", addr);

    if (addr == NULL)
        addr = usb_kmalloc (sizeof (ts_usb_addr), GFP_KERNEL);
    if (addr == NULL)
        return NULL;

    ts_usb_addr_clean (addr);
    return addr;
}

void ts_usb_addr_release (ts_usb_addr* addr)
{
    gdlf_fntrace ("ts_usb_addr_release (0x%p)", addr);
    gdlf_assert (addr);
    return;
}

    /* ts_usb_tr */

ts_usb_tr* ts_usb_tr_init (void)
{
    ts_usb_tr *addr;
    
    gdlf_fntrace ("ts_usb_tr_init()");

    addr = usb_kmalloc (sizeof (ts_usb_tr), GFP_KERNEL);
    if (addr == NULL)
        return NULL;

    ts_usb_tr_clean (addr);
    return addr;
}

void ts_usb_tr_release (ts_usb_tr* tr)
{
    gdlf_fntrace ("ts_usb_tr_release (0x%p)", tr);
    gdlf_aassert (tr, return);
    gdlf_assert (atomic_read (&tr->snode.usecnt) == 0);
    gdlf_assert (tr->hci == NULL);
    return;
}

    /* ts_usb_endp */

ts_usb_endp* ts_usb_endp_init (ts_usb_endp* endp)
{
    gdlf_fntrace ("ts_usb_endp_init (0x%p)", endp);

    if (endp == NULL)               /* Alloc memory if so requested */
        endp = usb_kmalloc (sizeof (ts_usb_endp), GFP_KERNEL);
    if (endp == NULL)
        return NULL;
    
    ts_usb_endp_clean (endp);
    return endp;
}

ts_usb_endp* ts_usb_endp_a_init (size_t size)
{
    ts_usb_endp *endps;
    ts_usb_endp *itr;
    
    gdlf_fntrace ("ts_usb_endp_a_init (%u)", size);

    endps = usb_kmalloc (sizeof (ts_usb_endp) * size, GFP_KERNEL);
    if (endps == NULL)
        return NULL;
    
    for (itr = endps + size - 1; itr >= endps; itr--)
    {
            /* Need braces, this is a macro! */
        ts_usb_endp_clean (itr);
    }

    return endps;
}

ts_usb_endp** ts_usb_endp_pa_init (ts_usb_endp** pendps, size_t size)
{
    ts_usb_endp **itr;
    
    gdlf_fntrace ("ts_usb_endp_pa_init (0x%p, %u)", pendps, size);

    if (pendps == NULL)              /* Alloc memory if so requested */
        pendps = usb_kmalloc (sizeof (ts_usb_endp*) * size, GFP_KERNEL);
    if (pendps == NULL)
        return NULL;
    
    for (itr = pendps + size - 1; itr >= pendps; itr--)
        *itr = NULL;

    return pendps;
}

void ts_usb_endp_release (ts_usb_endp* endp)
{
    gdlf_fntrace ("ts_usb_endp_release (0x%p)", endp);
    
    gdlf_assert (endp);

        /* ts_usb_endp->desc is not kfree()'ed, as it is only a
         * reference, we do not own it */

    return;
}

void ts_usb_endp_a_release (ts_usb_endp* endps, size_t size)
{
    ts_usb_endp *itr;
    
    gdlf_fntrace ("ts_usb_endp_a_release (0x%p, %u)", endps, size);

    gdlf_aassert (endps != NULL, return);
    gdlf_aassert (size != 0, return);

    for (itr = endps + size - 1; itr >= endps; itr--)
        ts_usb_endp_release (itr);

    return;
}

void ts_usb_endp_pa_release (ts_usb_endp** pendps, size_t size)
{
    gdlf_fntrace ("ts_usb_endp_pa_release (0x%p, %u)", pendps, size);
    
    gdlf_aassert (pendps != NULL, return);
    gdlf_aassert (size != 0, return);

        /* 
         * This is only an array of references to endpoints, we do not
         * own them, so we don't free'em 
         */

    FREE_PTR (pendps);
    return;
}

void ts_usb_endp_printk (const ts_usb_endp* endp)
{
    gdlf_fntrace ("ts_usb_endp_printk (0x%p)", endp);
    gdlf_aassert (endp != NULL, return);
    
    gdlf_debug ("ts_usb_endp 0x%p\n", endp);
    gdlf_debug ("  addr:       %d >> 1\n", endp->addr);
    gdlf_debug ("  type:       %d\n", endp->type);
    gdlf_debug ("  stat:       %d\n", endp->stat);
    gdlf_debug ("  maxpktsize: %d\n", endp->maxpktsize);
    gdlf_debug ("  interval:   %d\n", endp->interval);
    gdlf_debug ("  desc:       0x%p\n", endp->desc);
    return;
}


    /* ts_usb_cfel */    

ts_usb_cfel* ts_usb_cfel_init (void)
{
    ts_usb_cfel* cfel;
    
    gdlf_fntrace ("ts_usb_cfel_init()");
    cfel = usb_kmalloc (sizeof (ts_usb_cfel), GFP_KERNEL);
    if (cfel == NULL)
    {
        gdlf_err ("Cannot allocate space for a "
                  "ts_usb_cfel structure\n");
        return NULL;
    }
    ts_usb_cfel_clean (cfel);
    return cfel;              
}

void ts_usb_cfel_release (ts_usb_cfel* cfel)
{
    gdlf_fntrace ("ts_usb_cfel_release (0x%p)", cfel);
    gdlf_aassert (cfel != NULL, return);

        /* Device should be already unconfigured. See khubdd.c */

    gdlf_assert (cfel->data == NULL);
    return;
}

void ts_usb_cfel_printk (const ts_usb_cfel* cfel)
{
    gdlf_fntrace ("ts_usb_cfel_printk (0x%p)", cfel);
    gdlf_aassert (cfel != NULL, return);

    gdlf_debug ("ts_usb_cfel 0x%p\n", cfel);
    gdlf_debug ("  class:       0x%02x\n", cfel->class);
    gdlf_debug ("  subclass:    0x%02x\n", cfel->subclass);
    gdlf_debug ("  protocol:    0x%02x\n", cfel->protocol);
    gdlf_debug ("  configured:  0x%lx\n", cfel->configured);

    gdlf_debug ("  data:        0x%p\n", cfel->data);
    return;
}


    /* ts_usb_aset */

ts_usb_aset* ts_usb_aset_init (ts_usb_aset* aset)
{
    gdlf_fntrace ("ts_usb_aset_init (0x%p)", aset);

    if (aset == NULL)               /* Alloc memory if so requested */
        aset = usb_kmalloc (sizeof (ts_usb_aset), GFP_KERNEL);
    if (aset == NULL)
        return NULL;

    ts_usb_aset_clean (aset);
    return aset;
}

ts_usb_aset* ts_usb_aset_a_init (size_t size)
{
    ts_usb_aset *asets;
    ts_usb_aset *itr;
    
    gdlf_fntrace ("ts_usb_aset_a_init (%u)", size);

    asets = usb_kmalloc (sizeof (ts_usb_aset) * size, GFP_KERNEL);
    if (asets == NULL)
        return NULL;
    
    for (itr = asets + size - 1; itr >= asets; itr--)
    {
            /* Need braces, this is a macro! */
        ts_usb_aset_clean (itr);
    }

    return asets;
}

void ts_usb_aset_release (ts_usb_aset* aset)
{
    gdlf_fntrace ("ts_usb_aset_release (0x%p)", aset);

    gdlf_aassert (aset != NULL, return);

        /* 
         * Alternate setting should be already unconfigured. See
         * khubdd.c. 
         */

    ts_usb_cfel_release (&aset->cfel);

    FREE_PTR (aset->descr);
    if (aset->endp)
    {
        ts_usb_endp_a_release (aset->endp, aset->endps);
        FREE_PTR (aset->endp);
    }

    return;
}

void ts_usb_aset_a_release (ts_usb_aset* asets, size_t size)
{
    ts_usb_aset *itr;
    
    gdlf_fntrace ("ts_usb_aset_a_release (0x%p, %u)", asets, size);

    gdlf_aassert (asets != NULL, return);
    gdlf_aassert (size != 0, return);

    for (itr = asets + size - 1; itr >= asets; itr--)
        ts_usb_aset_release (itr);

    return;
}

void ts_usb_aset_printk (const ts_usb_aset* aset)
{
    int cnt;
    gdlf_fntrace ("ts_usb_aset_printk (0x%p)", aset);
    gdlf_aassert (aset, return);

    gdlf_debug ("ts_usb_aset 0x%p\n", aset);
    gdlf_debug ("  addr:   %u\n", aset->addr);
    ts_usb_cfel_printk (&aset->cfel);
    gdlf_debug ("  endp:   0x%p\n", aset->endp);
    gdlf_debug ("  endps:  %u\n", aset->endps);

    gdlf_debug ("  descr:  %s\n", aset->descr? aset->descr : "(null)");

    if (aset->endps)
    {
        gdlf_debug ("Endpoints dump for aset 0x%p\n", aset);
        for (cnt = 0; cnt < aset->endps; cnt++)
            ts_usb_endp_printk (&aset->endp[cnt]);
    }
    return;
}


    /* ts_usb_iface */

ts_usb_iface* ts_usb_iface_init (ts_usb_iface* iface)
{
    gdlf_fntrace ("ts_usb_iface_init (0x%p)", iface);

    if (iface == NULL)               /* Alloc memory if so requested */
        iface = usb_kmalloc (sizeof (ts_usb_iface), GFP_KERNEL);
    if (iface == NULL)
        return NULL;

    ts_usb_iface_clean (iface);
    return iface;
}

ts_usb_iface* ts_usb_iface_a_init (size_t size)
{
    ts_usb_iface* ifaces;
    ts_usb_iface *itr;
    
    gdlf_fntrace ("ts_usb_iface_a_init (%u)", size);

    ifaces = usb_kmalloc (sizeof (ts_usb_iface) * size, GFP_KERNEL);
    if (ifaces == NULL)
        return NULL;
    
    for (itr = ifaces + size - 1; itr >= ifaces; itr--)
    {
            /* Need braces, this is a macro! */
        ts_usb_iface_clean (itr);
    }

    return ifaces;
}

void ts_usb_iface_release (ts_usb_iface* iface)
{
    gdlf_fntrace ("ts_usb_iface_release (0x%p)", iface);
    gdlf_aassert (iface, return);

    if (iface->aset)
    {
        ts_usb_aset_a_release (iface->aset, iface->asets);
        FREE_PTR (iface->aset);
    }

    return;
}

void ts_usb_iface_a_release (ts_usb_iface* ifaces, size_t size)
{
    ts_usb_iface *itr;
    
    gdlf_fntrace ("ts_usb_iface_a_release (0x%p, %u)", ifaces, size);

    gdlf_aassert (ifaces != NULL, return);
    gdlf_aassert (size != 0, return);

    for (itr = ifaces + size - 1; itr >= ifaces; itr--)
        ts_usb_iface_release (itr);

    return;
}

void ts_usb_iface_printk (const ts_usb_iface* iface)
{
    int cnt;
    gdlf_fntrace ("ts_usb_iface_printk (0x%p)", iface);
    gdlf_aassert (iface, return);

    gdlf_debug ("ts_usb_iface 0x%p\n", iface);
    gdlf_debug ("  addr:   %u\n", iface->addr);
                           
    gdlf_debug ("  aset:   0x%p\n", iface->aset);
    gdlf_debug ("  asets:  %u\n", iface->asets);
    gdlf_debug ("  active: 0x%p\n", iface->active);

    gdlf_debug ("Alternate settings dump for iface 0x%p\n", iface);
    for (cnt = 0; cnt < iface->asets; cnt++)
        ts_usb_aset_printk (&iface->aset[cnt]);
    return;
}

    /* ts_usb_cfg */

ts_usb_cfg* ts_usb_cfg_init (void)
{
    ts_usb_cfg *cfg;
    
    gdlf_fntrace ("ts_usb_cfg_init (void)");

    cfg = usb_kmalloc (sizeof (ts_usb_cfg), GFP_KERNEL);
    if (cfg == NULL)
        return NULL;
    
    ts_usb_cfg_clean (cfg);
    return cfg;
}

#if 0
ts_usb_cfg* ts_usb_cfg_a_init (ts_usb_cfg* cfgs, size_t size)
{
    ts_usb_cfg *itr;
    
    gdlf_fntrace ("ts_usb_cfg_a_init (0x%p, %u)", cfgs, size);

    if (cfgs == NULL)              /* Alloc memory if so requested */
        cfgs = usb_kmalloc (sizeof (ts_usb_cfg) * size, GFP_KERNEL);
    if (cfgs == NULL)
        return NULL;
    
    for (itr = cfgs + size - 1; itr >= cfgs; itr--)
    {
            /* Need braces, this is a macro! */
        ts_usb_cfg_clean (itr);
    }

    return cfgs;
}
#endif

void ts_usb_cfg_release (ts_usb_cfg* cfg)
{
    gdlf_fntrace ("ts_usb_cfg_release (0x%p)", cfg);

    gdlf_aassert (cfg, return);
    gdlf_assert (atomic_read (&cfg->usecnt) == 0);

    FREE_PTR (cfg->descr);
    FREE_PTR (cfg->pendp);
    ts_usb_iface_a_release (cfg->iface, cfg->ifaces);
    FREE_PTR (cfg->iface);
    return;
}

void ts_usb_cfg_printk (const ts_usb_cfg* cfg)
{
    int cnt;
    
    gdlf_fntrace ("ts_usb_cfg_printk (0x%p)", cfg);
    gdlf_aassert (cfg != NULL, return);
    
    gdlf_debug ("ts_usb_cfg 0x%p\n", cfg);
    gdlf_debug ("  addr:       %u\n", cfg->addr);

    gdlf_debug ("  failed:     %u\n", cfg->failed);
    gdlf_debug ("  buspower:   %u\n", cfg->buspower);
    gdlf_debug ("  selfpower:  %u\n", cfg->selfpower);
    gdlf_debug ("  remwakeup:  %u\n", cfg->remwakeup);
    gdlf_debug ("  maxpower:   %u\n", cfg->maxpower);

    gdlf_debug ("  iface:      0x%p\n", cfg->iface);
    gdlf_debug ("  ifaces:     %u\n", cfg->ifaces);

    gdlf_debug ("  pendps:     %u\n", cfg->pendps);
    for (cnt = 0; cnt < cfg->pendps; cnt++)
        gdlf_debug ("  pendp[%02d]:  0x%p\n", cnt, cfg->pendp[cnt]);
    gdlf_debug ("  descr:      %s\n", cfg->descr? cfg->descr : "(null)");
    gdlf_debug ("  usecnt:     %d\n", atomic_read (&cfg->usecnt));

    gdlf_debug ("Interfaces dump for configuration 0x%p:\n", cfg);
    for (cnt = 0; cnt < cfg->ifaces; cnt++)
        ts_usb_iface_printk (&cfg->iface[cnt]);
    return;
}

#if 0
void ts_usb_cfg_a_release (ts_usb_cfg* cfgs, size_t size)
{
    ts_usb_cfg *itr;
    
    gdlf_fntrace ("ts_usb_cfg_a_release (0x%p, %u)", cfgs, size);

    gdlf_aassert (cfgs != NULL, return);
    gdlf_aassert (size != 0, return);
    
    for (itr = cfgs + size - 1; itr >= cfgs; itr--)
        ts_usb_cfg_release (itr);

    return;
}
#endif

    /* ts_usb_device */

ts_usb_device* ts_usb_device_init (void)
{
    ts_usb_device* device;
    
    gdlf_fntrace ("ts_usb_device_init ()");

    device = usb_kmalloc (sizeof (ts_usb_device), GFP_KERNEL);
    if (device == NULL)
    {
        gdlf_err ("Cannot allocate a ts_usb_device\n");
        return NULL;
    }

    ts_usb_device_clean (device);
    return device;
}

#if 0
ts_usb_device* ts_usb_device_a_init (size_t size)
{
    ts_usb_device *devices;
    ts_usb_device *itr;
    
    gdlf_fntrace ("ts_usb_device_a_init (%u)", size);

    devices = usb_kmalloc (sizeof (ts_usb_device) * size, GFP_KERNEL);
    if (devices == NULL)
        return NULL;
    
    for (itr = devices + size - 1; itr >= devices; itr--)
    {
            /* Need braces, this is a macro! */
        ts_usb_device_clean (itr);
    }

    return devices;
}
#endif

void ts_usb_device_release (ts_usb_device* device)
{
    t_usb_cfg cnt;
    
    gdlf_fntrace ("ts_usb_device_release (0x%p)", device);

    gdlf_assert (device);
    gdlf_passert (atomic_read (&device->msnode.usecnt) == 0,
                  "use count for device 0x%p is %u\n", 
                  device, atomic_read (&device->msnode.usecnt));
    gdlf_assert (device->msnode.next[0] == NULL);
    gdlf_assert (device->msnode.next[1] == NULL);
    gdlf_assert (device->msnode.next[2] == NULL);

        /* 
         * Device should be already unconfigured. See khubdd.c.
         */

    ts_usb_cfel_release (&device->cfel);


        /* 
         * Device should be already removed from the hub
         */

    gdlf_passert (device->hub == NULL,
                  "The hub is still 0x%p\n", device->hub);

        /* Remove the configurations and strings */
    
    for (cnt = 0; cnt < device->pcfgs; cnt++)
        if (device->pcfg[cnt])
        {
            ts_usb_cfg_release (device->pcfg[cnt]);
            FREE_PTR (device->pcfg[cnt]);
        }
    FREE_PTR (device->pcfg);

    FREE_PTR (device->serial);
    FREE_PTR (device->descr);
    FREE_PTR (device->mfrer);
    
    return;
}

#if 0
void ts_usb_device_a_release (ts_usb_device* devices, size_t size)
{
    ts_usb_device *itr;
    
    gdlf_fntrace ("ts_usb_device_a_release (0x%p, %u)", devices, size);

    gdlf_aassert (size != 0, return);
    gdlf_aassert (devics != NULL, return);
    
    for (itr = devices + size - 1; itr >= devices; itr--)
        ts_usb_device_release (itr);

    return;
}
#endif

void ts_usb_device_printk (const ts_usb_device* device)
{
    int cnt;

    gdlf_fntrace ("ts_usb_device_printk (0x%p)", device);
    gdlf_aassert (device != NULL, return);

    gdlf_debug ("ts_usb_device 0x%p\n", device);
    ts_3snode_printk (&device->msnode);
    gdlf_debug ("  addr:        %d\n", device->addr);
    gdlf_debug ("  hub:         0x%p\n", device->hub);
    gdlf_debug ("  port:        %u\n", device->port);

    gdlf_debug ("  remwakeup:   %d\n", device->remwakeup);
    gdlf_debug ("  selfpwrd:    %d\n", device->selfpwrd);
    gdlf_debug ("  cfgzero:     %d\n", device->cfgzero);

    ts_usb_endp_printk (&device->endp0);
    ts_usb_open_printk (&device->endp0d);

    gdlf_debug ("  pcfg:        0x%p\n", device->pcfg);
    gdlf_debug ("  pcfgs:       %d\n", device->pcfgs);
    gdlf_debug ("  active:      0x%p\n", device->active);

    gdlf_debug ("  cfels:       %d\n", device->cfels);
    gdlf_debug ("  cfelu:       %d\n", device->cfelu);
    gdlf_debug ("  acfels:      %d\n", device->acfels);
    gdlf_debug ("  acfelu:      %d\n", device->acfelu);

    ts_usb_cfel_printk (&device->cfel);

    gdlf_debug ("  usb_version: %x\n", device->usb_version);
    gdlf_debug ("  vendor:      %x\n", device->vendor);
    gdlf_debug ("  product:     %x\n", device->product);
    gdlf_debug ("  version:     %x\n", device->version);

    gdlf_debug ("  mfrer:       %s\n", 
                device->mfrer? device->mfrer : "(null)");
    gdlf_debug ("  descr:       %s\n", 
                device->descr? device->descr : "(null)");
    gdlf_debug ("  serial:      %s\n", 
                device->serial? device->serial : "(null)");
    
    gdlf_debug ("Configuration tree dump for device 0x%p\n", device);
    for (cnt = 0; cnt < device->pcfgs; cnt++)
        if (device->pcfg[cnt])
            ts_usb_cfg_printk (device->pcfg[cnt]);
    return;
}


    /* ts_usb_host */

ts_usb_host* ts_usb_host_init (void)
{
    ts_usb_host* host;
    
    gdlf_fntrace ("ts_usb_host_init (void)");

    host = usb_kmalloc (sizeof (ts_usb_host), GFP_KERNEL);
    if (host == NULL)
        return NULL;

    ts_usb_host_clean (host);
    return host;
}

void ts_usb_host_release (ts_usb_host* host)
{
    gdlf_fntrace ("ts_usb_host_release (0x%p)", host);

    gdlf_assert (host);
    gdlf_assert (atomic_read (&host->snode.usecnt) == 0);
    gdlf_assert (host->snode.next == NULL);

    if (host->master) /* ??? It should be clean */
    {
        gdlf_err ("Master device field of host 0x%p is 0x%p "
                  "instead of NULL\n", host, host->master);
        usb_device_unref (host->master, NULL);
        host->master = NULL;
    }

        /* No checks on HCI, as it must have been destroyed already */

    return;
}


    /* ts_usb_open */

ts_usb_open* ts_usb_open_init (ts_usb_open* open)
{
    gdlf_fntrace ("ts_usb_open_init (0x%p)", open);

    if (open == NULL)
        open = usb_kmalloc (sizeof (ts_usb_open), GFP_KERNEL);
    if (open == NULL)
        return NULL;

    ts_usb_open_clean (open);
    return open;
}

void ts_usb_open_release (ts_usb_open* open)
{
    gdlf_fntrace ("ts_usb_open_release (0x%p)", open);

    if (open == NULL)
        return;

    if (open->dev)
        usb_device_unref (open->dev, NULL);

    if (open->host != NULL)
        usb_host_unref (open->host, NULL);

    return;
}

void ts_usb_open_printk (const ts_usb_open* open)
{
    gdlf_fntrace ("ts_usb_open_printk (0x%p)", open);
    gdlf_aassert (open != NULL, return);
    
    gdlf_debug ("ts_usb_open 0x%p\n", open);
    gdlf_debug ("  host:     0x%p\n", open->host);
    gdlf_debug ("  dev:      0x%p\n", open->dev);
    gdlf_debug ("  endp:     0x%p\n", open->endp);
    ts_slist_printk (&open->list);
    gdlf_debug ("  running:  0x%p\n", open->running);
    gdlf_debug ("  finished: 0x%p\n", open->finished);
    return;
}


    /* ts_usb_driver */

ts_usb_driver* ts_usb_driver_init (void)
{
    ts_usb_driver* driver;

    gdlf_fntrace ("ts_usb_driver_init()");

    driver =
        usb_kmalloc (sizeof (ts_usb_driver), GFP_KERNEL);

    if (driver == NULL)
        return NULL;
    ts_usb_driver_clean (driver);
    return driver;
}


void ts_usb_driver_release (ts_usb_driver* driver)
{
    gdlf_fntrace ("ts_usb_driver_release (0x%p)", driver);
    gdlf_aassert (driver != NULL, return);
    gdlf_passert (atomic_read (&driver->snode.usecnt) == 0,
                  "aiee!! cleaning used ts_usb_driver %p "
                  "(usecount = %d)\n",
                  driver, atomic_read (&driver->snode.usecnt));
    return;
}

void ts_usb_driver_printk (const ts_usb_driver* driver)
{
    gdlf_fntrace ("ts_usb_driver_printk (0x%p)", driver);
    gdlf_aassert (driver != NULL, return);
    
    gdlf_debug ("ts_usb_driver 0x%p\n", driver);
    ts_snode_printk (&driver->snode);
    gdlf_debug ("  vendor:       0x%04x\n", driver->vendor);
    gdlf_debug ("  product:      0x%04x\n", driver->product);
    gdlf_debug ("  class:        0x%02x\n", driver->class);
    gdlf_debug ("  subclass:     0x%02x\n", driver->subclass);
    gdlf_debug ("  protocol:     0x%02x\n", driver->protocol);

    gdlf_debug ("  id:           0x%x\n", driver->id);

    gdlf_debug ("  name:         %s\n", 
                driver->name? driver->name : "(null)");
    gdlf_debug ("  copyright:    %s\n", 
                driver->copyright? driver->copyright : "(null)");
    gdlf_debug ("  vendorname:   %s\n", 
                driver->vendorname? driver->vendorname : "(null)");
    gdlf_debug ("  productname:  %s\n", 
                driver->productname? driver->productname : "(null)");
    gdlf_debug ("  classname:    %s\n", 
                driver->classname? driver->classname : "(null)");
    gdlf_debug ("  subclassname: %s\n", 
                driver->subclassname? driver->subclassname : "(null)");
    gdlf_debug ("  protocolname: %s\n", 
                driver->protocolname? driver->protocolname : "(null)");

    gdlf_debug ("  config:       0x%p\n", driver->config);
    gdlf_debug ("  unconfig:     0x%p\n", driver->unconfig);
    gdlf_debug ("  reference:    0x%p\n", driver->reference);

    ts_slist_printk (&driver->uselist);

    return;
}

    /* ts_usb_hci */

ts_usb_hci* ts_usb_hci_init (void)
{
    ts_usb_hci* hci;
    
    gdlf_fntrace ("ts_usb_hci_init()");
    hci = usb_kmalloc (sizeof (ts_usb_hci), GFP_KERNEL);
    if (hci == NULL)
    {
        gdlf_err ("Cannot allocate space for a "
                  "ts_usb_hci structure\n");
        return NULL;
    }
    ts_usb_hci_clean (hci);
    return hci;              
}


void ts_usb_hci_release (ts_usb_hci* hci)
{
    gdlf_fntrace ("ts_usb_hci_release (0x%p)", hci);
    gdlf_aassert (hci != NULL, return);
    gdlf_assert (hci->snode.dtor.fn == (tf_dtor)usb_hci_destroy);
    return;
}


int ts_usb_drv_use_destroy (ts_usb_drv_use* use)
{
    gdlf_fntrace ("ts_usb_drv_use_destroy (0x%p)", use);
    gdlf_aassert (use != NULL, return -EINVAL);
    
    ts_usb_drv_use_release (use);
    return 0;
}

/* Log:
 *
 * $Log: generic.c,v $
 * Revision 1.65  1998/11/14 17:12:14  inaky
 * Moved the include files to uusbd/include/uusbd, to ease integration
 * into the Linux kernel. Linkage into the kernel made possible.
 *
 * Revision 1.64  1998/11/09 21:20:33  inaky
 * Fixed prototypes for array generation/deletion, using size_t as size
 * type.
 *
 * Revision 1.63  1998/10/31 02:07:35  inaky
 * Removed all the d_* pointers from ts_usb_cfel and collapsed them into
 * a single pointer; only one was really used.
 *
 * Added the fix for devices using index 0 as first configuration
 * descriptor, with a bit which indicates it.
 *
 * Revision 1.62  1998/09/14 02:27:42  inaky
 * Added a comment on being only able to compile as module. Added the new
 * dynamic cfg->pendp scheme.
 *
 * Revision 1.61  1998/08/30 14:03:40  inaky
 * Fixed endpoint addres reporting.
 *
 * Revision 1.60  1998/08/30 12:38:55  inaky
 * Fixed reports of the endpoint pointer array numbers.
 *
 * Revision 1.59  1998/08/28 23:43:33  inaky
 * Added missing closing parentheses.
 *
 * Revision 1.58  1998/08/27 22:13:12  inaky
 * Removed traces.
 *
 * Revision 1.57  1998/08/27 10:52:16  inaky
 * Moved ts_*_free() to ts_*_release(), made them return void and severed
 * assertions.
 *
 * Revision 1.56  1998/08/25 03:58:45  inaky
 * Fixed the printing style for class, subclass and protocols, added a
 * driver printk function.
 *
 * Revision 1.55  1998/08/24 02:13:01  inaky
 * Added braces to `for' loops which call ts_*_clean()! it was only
 * executing the first command of the macro. Proper sollution would be to
 * define'em with do { } while (0).
 *
 * Fixed assertions.
 *
 * Added a set of functions to dump the state of devices, configurations,
 * ifaces and asets, printing the whole trees. Fixed some ts_*_free()
 * calls to return void. Fixed others to call ts_*_a_free() only if
 * there's array to free.
 *
 * Revision 1.54  1998/08/03 13:21:46  inaky
 * Fixed assertions to give clearer error messages on device's free
 * method.
 *
 * Revision 1.53  1998/07/30 22:57:15  inaky
 * Fixed all the for loops, as some could be broken.
 *
 * Revision 1.52  1998/07/23 14:18:20  inaky
 * Renamed the ts_usb_trbuf struct to ts_usb_tr
 *
 * Removed the host printk function, it wasn't used.
 *
 * Revision 1.51  1998/07/16 15:59:35  inaky
 * Fixed the mess with transactions on device-less hubs.
 *
 * Revision 1.50  1998/07/16 14:03:49  inaky
 * Fixed the ts_usb_host_free() function, as it was destroying the HCI
 * part, which it shouldn't, as it should be already destroyed and freed
 * when calling this function.
 *
 * Revision 1.49  1998/07/16 12:43:15  inaky
 * Added the transaction (ts_usb_trbuf) housekeeping functions.
 *
 * Moved from the old read/write interface to the new, request-based
 * interface.
 *
 * Revision 1.48  1998/07/12 22:38:30  inaky
 * Fixed the array cleaning code, as it was wroken.
 * Removed the old root hub code from host house-keeping; added the
 * reference method in hosts.
 *
 * Revision 1.47  1998/06/21 20:09:12  inaky
 * Removed the structure clean call from all ts_*_free() methods.
 *
 * Revision 1.46  1998/06/20 21:54:48  inaky
 * Massive changes to split the old monolithic hubd driver into a hub
 * manager (hub wrappers and daemon) and a dynamic driver, which can be
 * replaced and its different parts reused.
 *
 * Revision 1.45  1998/04/26 03:24:05  inaky
 * Fixed ts_usb_(iface|endp|aset)_a_init() routines not to need a
 * pointer. It should only allocate and clean. Disabled the
 * ts_usb_(cfg|device)_a_(init|free)() fn, which isn't used. Added the
 * fix to make the cfg's array on a device be a pointers array.
 *
 * Revision 1.44  1998/04/13 21:48:56  inaky
 * On ts_usb_device_init() and ts_usb_device_a_init() removed the
 * allocation stuff.
 *
 * Revision 1.43  1998/03/28 19:24:22  inaky
 * Removed inclusion of usb-methods.h
 *
 * Revision 1.42  1998/02/10 17:14:29  inaky
 * Moved to use ts_snode_[un]ref() to ref/unref snodes, fixed some
 * mistakes usint atomic_read(); Also changed the error handling when
 * any of the root, master or hci field of a host are set when freeing
 * it.
 *
 * Revision 1.41  1998/02/09 13:09:43  inaky
 * Moved all the use counts to atomic, using atomic_t and atomic_*()
 *
 * Revision 1.40  1998/01/11 15:44:15  inaky
 * Finished functions from the descriptor, inteface and endpoint modules,
 * fixed docs, pretty printed and removed craps.
 *
 * Revision 1.39  1998/01/08 12:19:06  inaky
 * Added copyright strings to all important files
 *
 * Revision 1.38  1998/01/02 19:39:36  inaky
 * Added the definition for endpoint zero to the device structure and
 * removed it from the list of endpoints kept at each configuration. Also
 * updated it on places where needed and removed the concept of basic
 * configuration, as is no longer needed.
 *
 * Revision 1.37  1998/01/02 00:12:15  inaky
 * Finished the new driver scheme. Documented at usbd-drv.c
 *
 * Revision 1.36  1998/01/01 16:11:03  inaky
 * Fixed bug, a couple _init routines where returning NULL instead of the
 * pointer :)
 *
 * Revision 1.35  1997/12/31 16:37:10  inaky
 * Finished the HCI interface
 *
 * Revision 1.34  1997/12/31 05:29:45  inaky
 * Added the handling functions for ts_usb_cfel
 *
 * Revision 1.33  1997/12/30 20:15:42  inaky
 * Moved everything to the new lists scheme. Untested yet.
 *
 * Revision 1.32  1997/12/30 00:02:46  inaky
 * Stupid change to fix code alignment pretty printing
 *
 * Revision 1.31  1997/12/29 00:25:35  inaky
 * Fixed unimplemented functions to return -ENOSYS and moving the khubdd
 * to use signals as commands.
 * Also added function traces on most files that had them disabled (donkey!)
 *
 * Revision 1.30  1997/12/27 03:00:13  inaky
 * Moved to GDLF and fixed some docs
 *
 * Revision 1.29  1997/12/27 02:24:03  inaky
 * Moving everything to use the GDLF for logging and debug checkings.
 * Removed the global array of USB hosts and moved it to a linked
 * list. Added an address field to each USB host controller.
 * Removed lots of cruft that were accumulated and useless. Still some
 * missing, but may be useful someday.
 *
 * Revision 1.28  1997/12/22 02:44:28  inaky
 * Trying to move all the documentation to the source files from the
 * header files, as it was duplicated.
 *
 * Revision 1.27  1997/12/08 02:40:57  inaky
 * Fixed some stuff of the #includes, modified the makefile targets to be
 * more flexible, tried to finish the device configuration section.
 *
 * Revision 1.26  1997/12/06 01:55:47  inaky
 * Now finishing the configuration system
 *
 * Revision 1.25  1997/12/06 00:19:39  inaky
 * Almost finished the usbd/host, usbd/cfg and a lot of meat from
 * usbd/device. Also added the driver definition. Got to polish it yet
 * ...
 *
 * Revision 1.24  1997/11/29 20:21:47  inaky
 * Documented and almost fixes the read/write/abort API, documented a
 * damn lot and implemented some stuff
 *
 * Revision 1.23  1997/11/29 00:46:29  inaky
 * Documented almost all the hub basic API, added spin locks to
 * usb_host_register() and some more stuff. Going for more.
 *
 * Revision 1.22  1997/10/31 05:44:29  inaky
 * *** empty log message ***
 *
 * Revision 1.21  1997/10/24 04:20:18  inaky
 * Security backup
 *
 * Revision 1.20  1997/10/02 22:29:36  inaky
 * Finally managed to understand the workings of the MODVERSIONS (ok, not
 * exactly the working, but how to make it work).
 *
 * Now I can go to sleep happy :) -- I know, I'm really sick --
 *
 * Revision 1.19  1997/10/02 21:28:59  inaky
 * Remade all the directory structure to make it cleaner when the USB
 * class drivers come.
 *
 * Revision 1.18  1997/09/30 13:04:42  inaky
 * Moved the name of the package to uusbd: Universal USB Driver
 *
 * Revision 1.17  1997/09/25 00:12:29  inaky
 * At the very least I hunted down the bug which was causing all that
 * damned exceptions, all memory related. Well, I'm a cock-fingered. Ix
 * don't know how to code; I'm stupid; I'm dreadful; I'm ....
 *
 * Revision 1.16  1997/09/08 10:54:48  inaky
 * Modified the generic functions to gain some performance, also some other checks
 *
 * Revision 1.15  1997/07/03 20:00:02  inaky
 * Moved all the hw dependencies to the HCD module, modified the
 * controller detections stuff to use just PCI means (okay, other buses
 * to come, but when they come). General fixes, but no really new deep
 * enhancements.
 *
 * Revision 1.14  1997/06/22 19:54:30  inaky
 * Seems that the moving of all the hw dependants stuff to the HCI
 * modules is done. Got to test it yet, but at least it compiles cleanly.
 *
 * Revision 1.13  1997/06/19 20:38:17  inaky
 * Moving al hardware dependant stuff to the HCD modules
 *
 * Revision 1.12  1997/05/04 14:38:34  inaky
 * Some fixes ...
 *
 * Revision 1.11  1997/05/03 21:22:48  inaky
 * Adding a lot of code, with the only objetive of adding the first
 * transfer type: control :) Still there's a lot of stuff to do, mostly
 * concerning data structures handling.
 *
 * Revision 1.10  1997/04/20 21:25:56  inaky
 * Finished the spec for the Hub Class (now gotta implement it)
 * :). Started coding the Frame Lists code, hope UHCI support will be
 * finished in some sessions.
 * This code is breaking, I'm doing something bad with kmallocs. kFree()ing
 * something not kmalloc()ed.
 *
 * Revision 1.9  1997/04/14 16:56:01  inaky
 * *** empty log message ***
 *
 * Revision 1.8  1997/04/12 16:47:49  inaky
 * Added the code for buffering reads/writes to/from userland
 *
 * Revision 1.7  1997/04/10 21:30:48  inaky
 * Fixed the file code, so it opens/closes ok; also reads/writes ok ... ok, just theory
 *
 * Revision 1.6  1997/04/06 22:56:44  inaky
 * Added the functions to set destinations of an USB open descriptor
 * (ts_usb_open); however, I haven't tested it yet ... bugs flowing!
 *
 * Revision 1.5  1997/04/06 21:14:31  inaky
 * Still with the definition of the access interface. More or less it is
 * taking an stable form.
 *
 * Revision 1.4  1997/04/02 20:02:03  inaky
 * Added init/free code for some struct more. Some still missing.
 *
 * Revision 1.3  1997/03/31 20:24:46  inaky
 * Pfff ... a lot. Mostly added descriptors stuff (HUB class still is
 * missing a lot) and I think I've got a definitive idea of how to
 * implement the storing of configurations and interfaces.
 *
 * Revision 1.2  1997/03/09 23:10:31  inaky
 * *** empty log message ***
 *
 * Revision 1.1  1997/03/02 22:11:02  inaky
 * *** empty log message ***
 *
 * Revision 1.1  1997/02/22 22:25:44  inaky
 * *** empty log message ***
 *
 *
 */

