
/* USB Driver Stack v$Revision: 1.17 $ 
 *
 * (C) 1997 by Iaky Prez Gonzlez <inaky@peloncho.fis.ucm.es>
 *
 * Distributed under the terms of the GNU Public License Version 2
 * See files COPYRIGHT and COPYING at the root distribution directory
 *
 * $Id: hcd.c,v 1.17 1998/11/14 17:12:28 inaky Exp $
 *
 * Changelog at the bottom 
 *
 *
 *     Host Controller Driver implementation
 *     -------------------------------------
 */


#include <uusbd/usb.h>
#include <uusbd/usb-hcd.h>
#include <uusbd/usbd.h>           /* usb_host_destroy() */

#define GDLF_IDENT          "USB-HCD"
#define GDLF_LOCAL          /* Define if want local settings */
#define GDLF_DEF_LOGLEVEL   9
#define GDLF_ASSERTIONS     /* want assertions (#undef if not) */
#undef GDLF_FNTRACE        /* want function traces */
#define GDLF_LOCATIONS      /* want locations in log messages */
#include <uusbd/gdlf.h>


ts_slist hcis;              /* list of registered HCIs */
extern ts_slist hosts;      /* list of registered hosts */


    /* Register an HCI into the UUSBD's HCD
     *
     *  This function will add a new HCD into the UUSBD
     * system. However, we can only expect IANF to see two, UHCI and
     * OHCI, but ...
     *
     *  The HCI structure (ts_usb_hci) must be properly initialized
     * with (ts_usb_hci_init()) and must have a name set (the 'name'
     * member). Note this member is a 'const char*', what means the
     * structure doesn't hold a local copy of the name, so do not
     * modify it :) nor free it while the name is in use.
     *
     * Parameters:  
     *
     *  ts_usb_hci*     Pointer to the host controller interface to
     *                  registrate. 
     *
     * Returns:
     *
     *  int             0 if ok, errno code < 0 on error.
     *
     * Global:
     *
     *  int usb_errno   On error, error code.  
     */

int usb_hci_register (ts_usb_hci* hci)
{
    gdlf_fntrace ("usb_hci_register (0x%p)", hci);
    gdlf_aassert (hci != NULL, return -EINVAL);
    gdlf_aassert (hci->name != NULL, return -EINVAL);
    gdlf_aassert (hci->type != 0, return -EINVAL);

        /* Fill the destructor and add to list */

    ts_dtor_fill (&hci->snode.dtor, (tf_dtor)usb_hci_destroy);
    return ts_slist_prep (&hcis, &hci->snode);    
}


    /* Unregister an HCI from the UUSBD's HCD
     *
     * Parameters:  
     *
     *  ts_usb_hci*     Pointer to the HCI structure.
     *
     * Returns:
     *
     *  int             0 if ok, errno code < 0 on error.
     *
     * Global:
     *
     *  int usb_errno   On error, error code.  
     */

int usb_hci_unregister (ts_usb_hci* hci)
{
    gdlf_fntrace ("usb_hci_unregister (0x%p)", hci);
    gdlf_aassert (hci != NULL, return -EINVAL);
    
    usb_hci_release (hci);
    ts_slist_del (&hcis, &hci->snode);
    return 0;
}


    /* Release an HCI
     *
     *  To accomplish it, we request the destruction of all the
     * ts_usb_hosts who use an hci registered within this HCI's list.
     *
     * Parameters:  
     *
     *  ts_usb_hci*     Pointer to HCI structure we want released
     *
     * Returns:
     *
     *  int             0 if ok, errno code < 0 on error.
     *
     * Global:
     *
     *  int usb_errno   On error, error code.  
     */

int usb_hci_release (ts_usb_hci* hci)
{
    int errcode;
    ts_usb_host *host, *next;
    
    gdlf_fntrace ("usb_hci_release (0x%p)", hci);
    gdlf_aassert (hci != NULL, return -EINVAL);

        /*
         * we must unregister every ts_usb_host which uses a
         * controller from this HCI.
         */

    read_lock (&hosts->lock);
    host = (ts_usb_host*)ts_slist_head (&hosts);
    while (host != NULL)
    {
        next = (ts_usb_host*)ts_snode_next (&host->snode);
        if (host->hcitype == hci->type)     /* HCI is this kind's? */
        {   
            errcode = usb_host_destroy (host);  /* kill it */
            gdlf_passert (errcode == 0,
                          "Couldn't destroy host 0x%p: "
                          "errno/usb %d/%d. "
                          "Killing it anyway.\n",
                          host, errcode, usb_errno);
            usb_kfree (host);
        }
        host = next;
    }
    read_unlock (&hosts->lock);
    return 0;
}


    /* Destructor for ts_usb_hci
     *
     * Parameters:  
     *
     *  ts_usb_hci*     Pointer to the HCI structure.
     *
     * Returns:
     *
     *  int             0 if ok, errno code < 0 on error.
     *
     * Global:
     *
     *  int usb_errno   On error, error code.  
     */

int usb_hci_destroy (ts_usb_hci* hci)
{    
    gdlf_fntrace ("usb_hci_destroy (0x%p)", hci);
    gdlf_aassert (hci != NULL, return -EINVAL);
    
    ts_snode_ref (&hci->snode);
    usb_hci_unregister (hci);
    ts_snode_unref_nd (&hci->snode);
    ts_usb_hci_release (hci);
    return 0;
}


    /* Register a host controller into an HCI
     *
     *  This function is used to register a host controller data
     * structure (ts_*hci_host) into it's correspoing HCI list
     * (ts_usb_hci).
     *
     * Parameters:  
     *
     *  ts_usb_hci*     Pointer to the HCI structure where to register
     *                  the host controller.
     *
     *  ts_snode*       Pointer to the host controller interface's
     *                  snode. 
     *
     * Returns:
     *
     *  int             0 if ok, errno code < 0 on error.
     *
     * Global:
     *
     *  int usb_errno   On error, error code.  
     */

int usb_hci_register_host (ts_usb_hci* hci, ts_snode* snode)
{
    int errcode;
    gdlf_fntrace ("usb_hci_register_host (0x%p, 0x%p)",
                  hci, snode);
    gdlf_aassert (hci != NULL, return -EINVAL);
    gdlf_aassert (snode != NULL, return -EINVAL);
                   
    errcode = ts_slist_prep (&hci->hosts, snode);
    gdlf_passert (errcode == 0,
                  "Cannot add host controller 0x%p to "
                  "HCI 0x%p: errno/usb %d/%d\n",
                  hci, snode, errcode, usb_errno);
    
    return errcode;
}


    /* Unregister a host controller from an HCI
     *
     *  This function is used to unregister a host controller data
     * structure (ts_*hci_host) from it's correspoing HCI list
     * (ts_usb_hci).
     *
     * Parameters:  
     *
     *  ts_usb_hci*     Pointer to the HCI structure from where to
     *                  unregister the host controller.
     *
     *  ts_snode*       Pointer to the host controller interface's
     *                  snode. 
     *
     * Returns:
     *
     *  int             0 if ok, errno code < 0 on error.
     *
     * Global:
     *
     *  int usb_errno   On error, error code.  
     */

int usb_hci_unregister_host (ts_usb_hci* hci, ts_snode* snode)
{
    gdlf_fntrace ("usb_hci_unregister_host (0x%p, 0x%p)",
                  hci, snode);
    gdlf_aassert (hci != NULL, return -EINVAL);
    gdlf_aassert (snode != NULL, return -EINVAL);
                   
    ts_slist_del (&hci->hosts, snode);
    return 0;
}


/* Log:
 *
 * $Log: hcd.c,v $
 * Revision 1.17  1998/11/14 17:12:28  inaky
 * Moved the include files to uusbd/include/uusbd, to ease integration
 * into the Linux kernel.
 *
 * Revision 1.16  1998/09/14 19:44:07  inaky
 * Severed assertions.
 *
 * Revision 1.15  1998/08/28 23:44:03  inaky
 * Severed assertions and the function, in general.
 *
 * Revision 1.14  1998/08/27 11:14:57  inaky
 * Moved ts_*_free() to ts_*_release(), made them return void and severed
 * assertions.
 *
 * Revision 1.13  1998/08/24 02:13:14  inaky
 * Removed function traces
 *
 * Revision 1.12  1998/07/16 12:43:44  inaky
 * Removed lockings when adding to lists, the list code already does it.
 *
 * Revision 1.11  1998/07/12 22:39:12  inaky
 * Fixed destructor to use reference and non-destructive unreference to
 * keep the object alive during destruction.
 *
 * Revision 1.10  1998/06/20 21:54:48  inaky
 * Massive changes to split the old monolithic hubd driver into a hub
 * manager (hub wrappers and daemon) and a dynamic driver, which can be
 * replaced and its different parts reused.
 *
 * Revision 1.9  1998/02/12 14:28:51  inaky
 * Replaced the old check-and-delete code with the new one, simpler and
 * more effective.
 *
 * Revision 1.8  1998/02/10 19:39:56  inaky
 * Fixed the destructor and unregistration methods, to do a check on
 * presence before deleting from their respective lists.
 *
 * Revision 1.7  1998/01/08 12:19:06  inaky
 * Added copyright strings to all important files
 *
 * Revision 1.6  1998/01/02 00:12:15  inaky
 * Finished the new driver scheme. Documented at usbd-drv.c
 *
 * Revision 1.5  1997/12/31 18:30:01  inaky
 * Finished adapting the UHCI module to the new HCD driver
 *
 * Revision 1.4  1997/12/31 17:06:35  inaky
 * Removed some unneeded locks and fixed some others.
 *
 * Revision 1.3  1997/12/31 16:25:10  inaky
 * Added the functions to register host controllers into their HCIs
 *
 * Revision 1.2  1997/12/31 05:37:00  inaky
 * Almost finished all the lists and life cycle for hosts, drivers and
 * devices. Some craps to be fixed yet.
 *
 * Revision 1.1  1997/12/30 22:30:17  inaky
 * Added the foundations of the HCD and fixed some bugs.
 *
 * Revision 1.2  1997/12/27 19:54:46  inaky
 * Removed .txt file who got in by mistake and added some stuff to the
 * templates.
 *
 * Revision 1.1  1997/10/02 21:29:17  inaky
 * Remade all the directory structure to make it cleaner when the USB
 * class drivers come.
 *
 * Revision 1.1  1997/02/22 22:25:44  inaky
 * *** empty log message ***
 *
 *
 */
















