
/*
 * Keyboard Boot Protocol Test
 * ---------------------------
 *
 *  The HID Boot Protocol for keyboards specifies the keyboard will
 * return, when polled, eight bytes of data. The disposition and
 * meaning of the data is specified on struct ts_bkbd_report in
 * ../hidd/usb-hidd.h.
 *
 *  What we do is just open the first registered keyboard and read
 * from it. Note what happens if you suddenly unplug the keyboard
 * [wait some time for khubdd to notice].
 */

#include <stdio.h>
#include <unistd.h>
#include <fcntl.h>
#include <errno.h>
#include <getopt.h>

#define DEFAULT_POLL_INTERVAL 100000

void version (FILE* file)
{
    fprintf (file,
             "UUSBD Utils - USB HID Boot Protocol Keyboard Test v$Version$\n");
    return;
}

void help (FILE* file)
{
    fprintf (file,
"(C) 1998 Inaky Perez Gonzalez <inaky@peloncho.fis.ucm.es>\n"
"\n"
"bkbd-test [OPTIONS] DEVICE\n"
"\n"
"Options:\n"
"\n"
"  -V, --version               Show version info\n"
"  -h, --help                  Show this help\n"
"  -i, --interval MICROSECONDS Set delay between reads (polls) to \n"
"                              MICROSECONDS (default is %d us). 0\n"
"                              don't wait\n"
"  -p, --dont-print            Don't print any data, just read\n"
"\n",
             DEFAULT_POLL_INTERVAL);
    return;
}


int main (int argc, char **argv)
{
    int print = 1;
    char option;
    int idxptr;
    ssize_t res;
    int fd;
    char* dev;
    char data[8];
    
    unsigned long interval = DEFAULT_POLL_INTERVAL;
    
    struct option longopts[] =
    {
        { "interval", required_argument, NULL, 'i' },
        { "help", no_argument, NULL, 'h' },
        { "version", no_argument, NULL, 'V' },
        { "dont-print", no_argument, NULL, 'p' },
        { NULL, 0, NULL, 0 }
    };
    const char shortopts[] = "i:phV";

    while ((option = getopt_long (argc, argv, shortopts, 
                                  longopts, &idxptr)) 
           >= 0)
    {
        switch (option)
        {
          case 'i':
            if (sscanf (optarg, "%lu", &interval) == 0)
                perror ("rate");
            break;
            
          case 'h':
            version (stdout);
            help (stdout);
            exit (0);
            break;
            
          case 'V':
            version (stdout);
            exit (0);
            break;

          case 'p':
            print = 0;
            break;
            
          default:
            version (stderr);
            help (stderr);
            exit (1);
            break;
        }
    }

    if (optind >= argc)
    {
        version (stderr);
        help (stderr);
        exit (1);
    }
    else
        dev = argv[optind];

    fd = open (dev, O_RDONLY);
    if (fd == -1)
    {
        perror ("open");
        exit (1);
    }
    
    do
    {
        res = read (fd, &data, 8);
        if (res == 0)
            break;
        if (res < 0)
        {
            if (errno == EBADFD)
            {
                fprintf (stderr, "Device %s unplugged/turned off!!\n", dev);
                break;
            }
            perror ("read");
        }
        else if (print)
            printf ("Modifiers: 0x%02x "
                    "cte  0x%02x "
                    "key1 0x%02x "
                    "key2 0x%02x "
                    "key3 0x%02x "
                    "key4 0x%02x "
                    "key5 0x%02x "
                    "key6 0x%02x\n",
                    (unsigned)data[0], 
                    (unsigned)data[1], 
                    (unsigned)data[2], 
                    (unsigned)data[3],
                    (unsigned)data[4], 
                    (unsigned)data[5], 
                    (unsigned)data[6], 
                    (unsigned)data[7]);
        usleep (interval);
    } while (1);
    close (fd);
    return 0;
}

/*
 * Local Variables:
 *  compile-command: "make bkbd-test"
 * End:
 */
