
/* USB Human Interface Devices Driver v$Revision: 1.8 $ 
 *
 * (C) 1997 by Iaky Prez Gonzlez <inaky@peloncho.fis.ucm.es>
 *
 * Distributed under the terms of the GNU Public License Version 2
 * See files COPYRIGHT and COPYING at the root distribution directory
 *
 * $Id: cfg.c,v 1.8 1998/11/14 17:18:54 inaky Exp $
 *
 * Changelog at the bottom 
 *
 *     USB HID Boot Protocol
 *     ---------------------
 *
 *  This is an smaller driver then HID, but is capable of reading from
 * keyboards and mice which understand the boot protocol. This makes
 * it nice to have it compiled for boot disks and that stuff with size
 * constraints.
 */

#define __NO_VERSION__
#include <linux/module.h>
#include <asm/uaccess.h>
#include <asm/spinlock.h>   /* locks */
#include <linux/miscdevice.h>

#include <uusbd/usbd.h>

#define GDLF_IDENT          "USB-HIDBP"
#define GDLF_LOCAL          /* Define if want local settings */
#define GDLF_DEF_LOGLEVEL   9
#define GDLF_ASSERTIONS     /* want assertions (#undef if not) */
#define GDLF_FNTRACE        /* want function traces */
#define GDLF_LOCATIONS      /* want locations in log messages */
#include <uusbd/gdlf.h>

#include <uusbd/usb-hidbp.h>



    /* Configure a HID BP interface
     *
     *  This function may be used to be called from hidd_config() to
     * configure the class part of a configurable element.
     *
     * Returns:
     *
     *  int             Number of cfels configured, errno code < 0 on
     *                  error. 
     */
           
int hidbp_config (ts_usb_cfel* cfel, ts_usb_device* device, 
                 ts_usb_cfg* cfg, t_usb_cfg cfg_idx, 
                 ts_usb_iface* iface, t_usb_iface iface_idx, 
                 ts_usb_aset* aset, t_usb_aset aset_idx, __u32 flags)
{
    int errcode;
    int base;
    const char *name;
    ts_hidbp *hidbp;
    struct file_operations *fops;
    ts_usb_request req;
    
    gdlf_fntrace ("hidbp_config (0x%p, 0x%p, 0x%p, "
                  "0x%p, 0x%p, %u)", cfel, device, cfg, 
                  iface, aset, flags);
    gdlf_aassert (cfel != NULL, return -EINVAL);
    gdlf_aassert (cfel->class == HIDD_CLASS, return -EBADFD);
    gdlf_aassert (device != NULL, return -EINVAL); 
    gdlf_aassert (iface != NULL, return -EINVAL);
    gdlf_aassert (aset != NULL, return -EINVAL);
    
    if (cfel->class != HIDD_CLASS)
    {
        gdlf_err ("Pretended to configure as HID non-HID "
                  "class interface %d of device (%d %d)\n",
                  iface->addr, 
                  usb_host_addr (device), usb_dev_addr (device));
        return 0;
    }
    
    if (cfel->subclass != HIDD_SUBCLASS_BOOT)
    {
        gdlf_err ("Pretended to configure as HID Boot Protocol "
                  "non-HIDBP class interface %d of device (%d %d)\n",
                  iface->addr, 
                  usb_host_addr (device), usb_dev_addr (device));
        return 0;
    }
    
        /* make sure some theorical facts */

    gdlf_assert (aset->endps == 1);
    gdlf_assert (aset->endp[0].type == e_usb_tr_type_intr);

        /* Ok, configure this cfel */

    switch (cfel->protocol)
    {
      case HIDBP_PROTOCOL_KBD:
#ifdef CONFIG_USB_HIDBP_KBD
        gdlf_info ("Interface %d of device (%d %d) supports "
                   "HID Keyboard Boot Protocol\n", iface->addr, 
                   usb_host_addr (device), usb_dev_addr (device));
        base = HIDBP_BKBD_MINOR;
        name = "keyboard";
        fops = &bkbd_fops;
#else
        gdlf_err ("Interface %d of device (%d %d) supports "
                  "HID Keyboard Boot Protocol, not compiled in\n", 
                  iface->addr, usb_host_addr (device), 
                  usb_dev_addr (device));
        return -ENOSYS;
#endif
        break;  
            
      case HIDBP_PROTOCOL_MSE:
#ifdef CONFIG_USB_HIDBP_KBD
        gdlf_info ("Interface %d of device (%d %d) supports "
                   "HID Boot Protocol Mouse\n", iface->addr, 
                   usb_host_addr (device), usb_dev_addr (device));
        base = HIDBP_BMSE_MINOR;
        name = "mouse";
        fops = &bmse_fops;
#else
        gdlf_err ("Interface %d of device (%d %d) supports "
                  "HID Boot Protocol Mouse, not compiled in\n", 
                  iface->addr, usb_host_addr (device), 
                  usb_dev_addr (device));
        return -ENOSYS;
#endif
        break;
            
      default:
        gdlf_err ("Interface %d of device (%d %d) supports "
                  "unknown HID Boot Protocol %d\n", iface->addr, 
                  usb_host_addr (device), usb_dev_addr (device),
                  cfel->protocol);
        return -ENOSYS;
        break;
    }

        /* 
         * Ok, we can get on with this guy.
         *
         * set protocol to boot protocol 
         */

    req.bmreqtype = 0x21;
    req.request = e_hidd_req_set_protocol;
    req.value = 0;
    req.index = iface->addr;
    req.length = 0;
    
    errcode = usb_request_dataless (device, &req, "set protocol");
    if (errcode < 0)
    {
        gdlf_err ("Cannot set interface %d of device (%d %d) "
                  "to HID Boot Protocol: errno/usb %d/%d\n",
                  iface->addr, usb_host_addr (device), 
                  usb_dev_addr (device), errcode, usb_errno);
        return errcode;
    }

        /* set report mode to idle until change */

    req.bmreqtype = 0x21;
    req.request = e_hidd_req_set_idle;
    req.value = (32)<<8;      /* report only if change detected */
    req.index = iface->addr;
    req.length = 0;
    
    errcode = usb_request_dataless (device, &req, "set idle");
    if (errcode < 0)
        gdlf_err ("Cannot set idle mode on HID Boot %s, "
                  "interface %d of device (%d %d) errno/usb %d/%d\n",
                  name, iface->addr, usb_host_addr (device), 
                  usb_dev_addr (device), errcode, usb_errno);

        /* allocate a ts_hidbp structure and register it */

    hidbp = ts_hidbp_init();
    if (hidbp == NULL)
    {
        gdlf_err ("Cannot allocate memory for using interface "
                  "%d of device (%d %d) as HID Boot Protocol %s\n",
                  iface->addr, usb_host_addr (device), 
                  usb_dev_addr (device), name);
        return -ENOMEM;
    }
    ts_dtor_fill (&hidbp->snode.dtor, (tf_dtor)hidbp_destroy);
    errcode = hidbp_register (hidbp);
    gdlf_aassert (errcode >= 0, return errcode);
    usb_dst_host_set (&hidbp->open, usb_host_addr (device));
    usb_dst_dev_set (&hidbp->open, usb_dev_addr (device));
    cfel->configured = (unsigned long)hidbp;
    cfel->data = ts_snode_ref (&hidbp->snode);

    hidbp->aset = aset;
    hidbp->tr.open = &hidbp->open;        
    hidbp->tr.size =
        hidbp->aset->cfel.protocol == HIDBP_PROTOCOL_MSE? 
        sizeof (ts_bmse_report) : sizeof (ts_bkbd_report);
    hidbp->tr.status = e_usb_tr_status_filled;

        /* alloc a misc char device */

    hidbp->miscdev.minor = hidbp_free_address (base);
    if (hidbp->miscdev.minor < 0)
    {
        gdlf_err ("No minor number space to allocate HIDBP %s at "
                  "device (%d %d): errno %d\n", 
                  name, usb_host_addr (device), usb_dev_addr (device),
                  hidbp->miscdev.minor);
        hidbp_destroy (hidbp);
        usb_kfree (hidbp);
    }
    hidbp->miscdev.fops = fops;

        /*
         * basic name is 64 chars at most, also we add lengths for
         * manufacturer, device name and serial. +3 is for the
         * separators.
         */  

    hidbp->miscdev.name =   
        usb_kmalloc (
            (64 + (device->mfrer? strlen (device->mfrer) : 0)
             + (device->descr? strlen (device->descr) : 0)
             + (device->serial? strlen (device->serial) : 0)
             + 3) * sizeof (char), GFP_KERNEL);
    if (hidbp->miscdev.name == NULL)
    {
        gdlf_err ("No memory for device 0x%p (%d %d)'s name! "
                  "Aborting!!\n", device, usb_host_addr (device),
                  usb_dev_addr (device));
        hidbp_destroy (hidbp);
        usb_kfree (hidbp);
        return -ENOMEM;
    }
    sprintf ((char*)hidbp->miscdev.name, 
             "HIDBP %s/USB dev (%d %d)%s %s %s %s",
             name, usb_host_addr (device), usb_dev_addr (device),
             (device->mfrer || device->descr || device->serial)? ":" : "",
             device->mfrer? device->mfrer : "",
             device->descr? device->descr : "", 
             device->serial? device->serial : "");
    errcode = misc_register (&hidbp->miscdev);
    if (errcode < 0)
    {
        gdlf_err ("Unable to register misc device with minor "
                  "%d! Aborting!!\n", hidbp->miscdev.minor);
        hidbp_destroy (hidbp);
        usb_kfree (hidbp);
    }
    return errcode;
}


    /* Unconfigure all cfels configured by this driver
     *
     *  This function may be used to unconfigure a driver's
     * class-configured configurable elements.
     *
     * Parameters:  
     *
     *  ts_usb_device*  Pointer to the device structure to
     *                  unconfigure. 
     *
     *  unsigned        USB_CFG_* flags indicating why is this
     *                  function called.  
     *
     * Returns:
     *
     *  int             0 if ok, errno code < 0 on error.
     */

int hidbp_unconfig (ts_usb_device* device, unsigned flags)
{
    t_usb_cfg cfg_cnt;
    ts_usb_cfg *cfg;
    ts_usb_iface *iface, *iface_top;
    ts_usb_aset *aset, *aset_top;
    
    gdlf_fntrace ("hidbp_unconfig (0x%p, %u)", device, flags);
    gdlf_aassert (device != NULL, return -EINVAL);

    for (cfg_cnt = 0; cfg_cnt < device->pcfgs; cfg_cnt++)
    {
        cfg = device->pcfg[cfg_cnt];
        if (cfg == NULL)        /* unloaded? skip */
            continue;
        
            /* Walk cfg's interfaces */

        for (iface = cfg->iface, iface_top = iface + cfg->ifaces; 
             iface < iface_top; iface++)
        {
                /* Walk the iface's asets */

            for (aset = iface->aset, aset_top = aset + iface->asets; 
                 aset < aset_top; aset++)

                      /* cfged by us? */

                if (aset->cfel.configured
                    && aset->cfel.data
                    && (aset->cfel.data->dtor.fn
                        == (tf_dtor)hidbp_destroy)
                        )
                {
                    ts_dtor_exec (&aset->cfel.data->dtor);
                    usb_kfree (aset->cfel.data = NULL);
                    aset->cfel.data = NULL;
                    aset->cfel.configured = 0;      /* uncfged */
                }
        }         
    }
    return 0;
}


/* Log:
 *
 * $Log: cfg.c,v $
 * Revision 1.8  1998/11/14 17:18:54  inaky
 * Moved the include files to uusbd/include/uusbd, to ease integration
 * into the Linux kernel.
 *
 * Revision 1.7  1998/11/09 21:33:32  inaky
 * Changed the name of the configuration pointer array to pcfg.
 *
 * Revision 1.6  1998/10/31 02:16:26  inaky
 * Removed all the d_* pointers from ts_usb_cfel and collapsed them into
 * a single pointer; only one was really used.
 *
 * Revision 1.5  1998/10/22 00:05:43  inaky
 * Moved to hidbp_config() the selection of the size to be used as data
 * buffer.
 *
 * Revision 1.4  1998/10/18 15:11:11  inaky
 * Fixed with parentheses as it seems I still don't know as well as I
 * should the operators precedence. Somewhere it was causing a NULL
 * access.
 *
 * Revision 1.3  1998/09/24 15:59:48  inaky
 * Made the registered name on miscdevices be a descriptive name plus the
 * device's name, manufacturer and serial number (if available).
 *
 * Revision 1.2  1998/09/14 02:03:27  inaky
 * Moved to a HIDBP header, recoded the routines to have only specifics
 * for reading and ioctls.
 *
 * Revision 1.1  1998/09/11 00:50:57  inaky
 * Moved the HID Boot Protocol to a single driver, so it can be isolated
 * more easily. It supports keyboards and mice, settable at compile time.
 *
 * Revision 1.1  1998/08/30 13:55:46  inaky
 * Added boot protocol support and a simple test program.
 *
 * Revision 1.1  1998/08/28 23:53:14  inaky
 * New driver
 *
 * Revision 1.3  1998/08/28 00:14:22  inaky
 * New driver ...
 *
 */
