
/* USB Human Interface Devices Driver v$Revision: 1.3 $ 
 *
 * (C) 1997 by Iaky Prez Gonzlez <inaky@peloncho.fis.ucm.es>
 *
 * Distributed under the terms of the GNU Public License Version 2
 * See files COPYRIGHT and COPYING at the root distribution directory
 *
 * $Id: driver.c,v 1.3 1998/11/14 17:18:55 inaky Exp $
 *
 * Changelog at the bottom 
 *
 *     USB HID Boot Protocol
 *     ---------------------
 *
 *  This is an smaller driver then HID, but is capable of reading from
 * keyboards and mice which understand the boot protocol. This makes
 * it nice to have it compiled for boot disks and that stuff with size
 * constraints.
 */

#include <linux/module.h>
#include <linux/init.h>     /* __init */
#include <asm/uaccess.h>
#include <asm/spinlock.h>   /* locks */
#include <linux/miscdevice.h>

#include <uusbd/usbd.h>

#define GDLF_IDENT          "USB-HIDBP"
#define GDLF_LOCAL          /* Define if want local settings */
#define GDLF_DEF_LOGLEVEL   9
#define GDLF_ASSERTIONS     /* want assertions (#undef if not) */
#define GDLF_FNTRACE        /* want function traces */
#define GDLF_LOCATIONS      /* want locations in log messages */
#include <uusbd/gdlf.h>

#include <uusbd/usb-hidbp.h>



#define NAME   /* short!! */                    \
    "USB HID Boot Protocol Driver v$Revision: 1.3 $"

#define COPYRIGHT                               \
    "(c) 1998 Inaky Perez Gonzalez "            \
    "<inaky@peloncho.fis.ucm.es>"

#define VENDOR 0
#define VENDORNAME                              \
    NULL

#define PRODUCT 0
#define DEVICENAME                              \
    NULL

#define CLASS HIDD_CLASS
#define CLASSNAME                               \
    "USB Human Interface Devices"

#define SUBCLASS 1
#define SUBCLASSNAME                            \
    "Boot Protocol"

#define PROTOCOL 0
#define PROTOCOLNAME                            \
    NULL

#define SNODE_NULL { {NULL}, NULL, 0}

    /* Driver stuff */

atomic_t usecnt;
ts_usb_driver* hidbp_driver = NULL;
ts_slist hidbp_list;


    /* Initialize the hidbp driver
     *
     * Returns:
     *
     *  int             0 if ok, errno code < 0 on error.
     */

int __init hidbp_init (void)
{
    int errcode;
    
    gdlf_fntrace ("hidbp_init()");
    gdlf_info (NAME "\n");
#ifdef MODULE
    gdlf_info (COPYRIGHT "\n");
#endif

    hidbp_driver = ts_usb_driver_init();
    if (hidbp_driver == NULL)
    {
        usb_errno = -USBE_NO_MEMORY;
        return -ENOMEM;
    }

        /* Set driver's attributes */

    hidbp_driver->vendor = VENDOR;
    hidbp_driver->product = PRODUCT;
    hidbp_driver->class = CLASS;
    hidbp_driver->subclass = SUBCLASS;
    hidbp_driver->protocol = PROTOCOL;

        /* Set info */

    hidbp_driver->name = NAME;                   /* driver name */
    hidbp_driver->copyright = COPYRIGHT;         /* driver copyright */
    hidbp_driver->vendorname = VENDORNAME;       /* device's vendor name */
    hidbp_driver->productname = DEVICENAME;      /* device's name */
    hidbp_driver->classname = CLASSNAME;         /* class' name */
    hidbp_driver->subclassname = SUBCLASSNAME;   /* subclass' name */
    hidbp_driver->protocolname = PROTOCOLNAME;   /* protocol name */

    hidbp_driver->config = hidbp_config;         /* config method */
    hidbp_driver->unconfig = hidbp_unconfig;     /* unconfig method */
    hidbp_driver->reference = hidbp_reference;   /* reference method */
    ts_snode_ref (&hidbp_driver->snode);         /* we hold a reference to it */
    
    ts_slist_clean (&hidbp_list);                /* initialize the list */
    errcode = usb_drv_register (hidbp_driver);   /* register driver */
    if (errcode)
    {
        gdlf_err ("Cannot register driver: "
                  "errno/usb %d/%d\n",
                  errcode, usb_errno);
        usb_drv_destroy (hidbp_driver);
        usb_kfree (hidbp_driver);
        return errcode;
    }        
    return 0;
}


    /* Release the hidbp driver
     *
     *  Do what is needed to ensure the use count of the driver is
     * zero and can be unloaded from the kernel.
     *
     * Returns:
     *
     *  int             0 if ok, errno code < 0 on error.
     *
     * Global:
     *
     *  int hidbp_errno On error, error code.  
     *  int usb_errno   On error, error code.  
     */

int hidbp_release (void)
{
    gdlf_fntrace ("hidbp_release()");

#warning FIXME: hidbp_release() not implemented!

    return -ENOSYS;
}


    /* Shutdown the skel driver
     *
     * Returns:
     *
     *  int             0 if ok, errno code < 0 on error.
     *
     * Global:
     *
     *  int hidbp_errno On error, error code.  
     *  int usb_errno   On error, error code.  
     */

int hidbp_shutdown (void)
{
    gdlf_fntrace ("hidbp_shutdown()");
    ts_snode_unref (&hidbp_driver->snode, NULL);
    return usb_drv_unregister (hidbp_driver);
}


    /* Driver reference
     *
     *  The 'reference' method is used when we take any of the driver's
     * code or data structures to use it. If so, we call it with a
     * positive argument (one reference, regardless the value of the
     * _positive_ number). When we are done, we release the reference
     * calling it with a negative number. It always returns the number
     * of references or < 0 on error. Calling it with an argument of
     * zero will just return the number of references. 
     *
     * Parameters:  
     *
     *  type            descr
     *
     * Global:
     *
     *  type name       descr
     *
     * Returns:
     *
     *  type            descr
     *
     * Global:
     *
     *  type name       descr
     */

int hidbp_reference (ts_usb_device* device, int n)
{
    gdlf_fntrace ("hidbp_reference (0x%p, %d)", device, n);
    gdlf_aassert (device != NULL, return -EINVAL);

    if (n > 0)
    {
        atomic_inc (&usecnt);
        MOD_INC_USE_COUNT;
    }
    else if (n < 0)
    {
        atomic_dec (&usecnt);
        MOD_DEC_USE_COUNT;
    }
    return atomic_read (&usecnt);
}


    /* Linux module support */

#ifdef MODULE

int init_module (void)
{
    return hidbp_init();
}

void cleanup_module (void)
{
    hidbp_release();
    hidbp_shutdown();
    return;
}

#endif


/* Log:
 *
 * $Log: driver.c,v $
 * Revision 1.3  1998/11/14 17:18:55  inaky
 * Moved the include files to uusbd/include/uusbd, to ease integration
 * into the Linux kernel.
 *
 * Revision 1.2  1998/09/14 02:03:58  inaky
 * Reworked the driver and config system to register the misc device only
 * on device configuration time.
 *
 * Revision 1.1  1998/09/11 00:50:57  inaky
 * Moved the HID Boot Protocol to a single driver, so it can be isolated
 * more easily. It supports keyboards and mice, settable at compile time.
 *
 * Revision 1.1  1998/08/30 13:55:46  inaky
 * Added boot protocol support and a simple test program.
 *
 * Revision 1.1  1998/08/28 23:53:14  inaky
 * New driver
 *
 * Revision 1.3  1998/08/28 00:14:22  inaky
 * New driver ...
 *
 */



