
/* USB Human Interface Devices Driver v$Revision: 1.5 $ 
 *
 * (C) 1997 by Iaky Prez Gonzlez <inaky@peloncho.fis.ucm.es>
 *
 * Distributed under the terms of the GNU Public License Version 2
 * See files COPYRIGHT and COPYING at the root distribution directory
 *
 * $Id: hidbp.c,v 1.5 1998/11/14 17:18:56 inaky Exp $
 *
 * Changelog at the bottom 
 *
 *     USB HID Boot Protocol
 *     ---------------------
 *
 *  Stuff common to both mice and kbds.
 */

#define __NO_VERSION__
#include <linux/module.h>
#include <asm/uaccess.h>
#include <asm/spinlock.h>   /* locks */
#include <linux/miscdevice.h>

#include <uusbd/usbd.h>

#define GDLF_IDENT          "USB-HIDBP"
#define GDLF_LOCAL          /* Define if want local settings */
#define GDLF_DEF_LOGLEVEL   9
#define GDLF_ASSERTIONS     /* want assertions (#undef if not) */
#define GDLF_FNTRACE        /* want function traces */
#define GDLF_LOCATIONS      /* want locations in log messages */
#include <uusbd/gdlf.h>

#include <uusbd/usb-hidbp.h>


    /* Seek a free minor address and return it.
     *
     * NOTE: The method for doing this search could be enhanced, as it
     *       is no secret I haven't squeezed my brains finding one :)
     *
     *  We'll just start with address number 'base' and count up. With
     * each one, we'll test all the elements in the list and see if
     * someone owns it. If it does, we'll use the next one, and so
     * on.
     *
     * Returns:
     *
     *  int             Free address if ok, errno code < 0 on
     *                  error. Address is guaranteed to be between
     *                  'base' and 'base + HIDBP_DEVICES_MAX'. If no
     *                  address is available, -ENOSPC (no space left
     *                  on device) will be returned.
     */

#define HIDBP_DEVICES_MAX 16

int hidbp_free_address (unsigned base)
{
    ts_hidbp* itr;
    int addr = base;       /* default initial address */
    int top = addr + HIDBP_DEVICES_MAX;
    
    gdlf_fntrace ("hidbp_free_address (%u)", base);

    itr = (ts_hidbp*)                   /* get last registered device */
        ts_slist_head (&hidbp_list);
    for (addr = base; addr < top; addr++)   /* cycle all addresses */
    {
        while (itr)                         /* on each cycle hidbps */
        {
            if (itr->miscdev.minor >= base
                || itr->miscdev.minor < base + HIDBP_DEVICES_MAX)
            {
                if (itr->miscdev.minor == addr) /* Used? */
                    break;                      /* ok, slip */
            }
            itr = (ts_hidbp*) ts_slist_next (&itr->snode);
        }
        if (itr == NULL)        /* unused addr */
            return addr;    
    }

        /* Bad, none found */

    return -ENOSPC; 
}


    /* Open a boot protocol device
     *
     * Major number is HIDBP_MAJOR, minor number is at the ts_hidbp. They
     * are numbered after the configuration order. devfs will rock
     * here.
     *
     * Parameters:  
     *
     *  struct inode*   i-node description structure.
     *
     *  struct file*    file description structure.
     *
     * Returns:
     *
     *  int             0 if ok, errno code < 0 on error.
     */

int hidbp_open (struct inode *inode, struct file *file)
{
    unsigned minor;     /* hidbp struct index number */
    int errcode;
    ts_hidbp *hidbp;
    
    gdlf_fntrace ("hidbp_open (0x%p, 0x%p)", inode, file);
    gdlf_aassert (inode != NULL, return -EINVAL);
    gdlf_aassert (file != NULL, return -EINVAL);
    
        /* Find which _hidbp_ are we referring to */

    minor = MINOR (inode->i_rdev);      /* get the minor number */
    hidbp = (ts_hidbp*)ts_slist_head (&hidbp_list);
    while (hidbp)
    {
        if (gdlf_c1assert (hidbp->aset)
            && gdlf_c1assert (hidbp->aset->cfel.class  == HIDD_CLASS)
            && gdlf_c1assert (hidbp->aset->cfel.subclass == HIDD_SUBCLASS_BOOT)
            && hidbp->miscdev.minor == minor)
            break;
        hidbp = (ts_hidbp*)ts_slist_next (&hidbp->snode);
    }

    if (hidbp == NULL)                   /* not found */
        return -ENODEV;

    if (hidbp->open.endp != NULL)
        return -EBUSY;
    
        /* Open the interrupt pipe */

    gdlf_aassert (hidbp->open.dev != NULL, return -EINVAL);
    gdlf_aassert (hidbp->open.dev->active != NULL, return -EINVAL);
    gdlf_aassert (hidbp->aset != NULL, return -EINVAL);

    errcode = usb_dst_endp_set (&hidbp->open, hidbp->aset->endp[0].addr);
    if (errcode < 0)
    {
        gdlf_err ("Cannot open endpoint %d, interface %d of "
                  "device (%d %d) to read keyboard input: "
                  "errno/usb %d/%d\n",
                  hidbp->aset->endp[0].addr, hidbp->aset->addr, 
                  usb_host_addr (hidbp->open.dev), 
                  usb_dev_addr (hidbp->open.dev), errcode, usb_errno);
        usb_dst_endp_unset (&hidbp->open);
        return -ENXIO;
    }

        /* Insert the tr */
    
    hidbp->tr.data = usb_kmalloc (hidbp->tr.size, GFP_KERNEL);
    if (hidbp->tr.data == NULL)
    {
        gdlf_err ("No memory to set up a buffer for input on "
                  "%s\n", hidbp->miscdev.name);
        usb_dst_endp_unset (&hidbp->open);
        return -ENOMEM;
    }
    
    errcode = usb_tr_setup (&hidbp->tr);
    if (errcode < 0)
    {
        usb_kfree (hidbp->tr.data);
        hidbp->tr.data = NULL;
        usb_tr_setdn (&hidbp->tr);
        usb_dst_endp_unset (&hidbp->open);
        return errcode;
    }

    errcode = usb_tr_insert (&hidbp->tr);
    if (errcode < 0)
    {
        usb_kfree (hidbp->tr.data);
        hidbp->tr.data = NULL;
        usb_tr_setdn (&hidbp->tr);
        usb_dst_endp_unset (&hidbp->open);
        return errcode;
    }
    
    ts_snode_ref (&hidbp->snode);
    file->private_data = hidbp;          /* set the private data */
    hidbp->file = file;
    MOD_INC_USE_COUNT;
    return 0;
}


    /* Close a boot protocol device
     *
     * Parameters:  
     *  struct inode*   i-node description structure.
     *  struct file*    file description structure.
     *
     * Returns:
     *
     *  int             0 no error detected, != errno error code.
     */

int hidbp_close (struct inode *inode, struct file *file)
{
    int errcode;
    ts_hidbp *hidbp;
    
    gdlf_fntrace ("hidbp_close (0x%p, 0x%p)", inode, file);   
    gdlf_aassert (inode != NULL, return -EINVAL);
    gdlf_aassert (file != NULL, return -EINVAL);
    
    hidbp = file->private_data;          /* get the private data */
    if (hidbp == NULL)                   /* device removed? */
        return 0;
    gdlf_aassert (hidbp->snode.dtor.fn == (tf_dtor)hidbp_destroy, 
                  return -EINVAL);
    ts_snode_unref (&hidbp->snode, NULL);    
    usb_tr_setdn (&hidbp->tr);
    usb_kfree (hidbp->tr.data);
    hidbp->tr.data = NULL;
    errcode = usb_dst_endp_unset (&hidbp->open);  /* release the data ... */
    hidbp->file = NULL;
    MOD_DEC_USE_COUNT;
    return errcode;
}



/* Log:
 *
 * $Log: hidbp.c,v $
 * Revision 1.5  1998/11/14 17:18:56  inaky
 * Moved the include files to uusbd/include/uusbd, to ease integration
 * into the Linux kernel.
 *
 * Revision 1.4  1998/11/09 21:33:54  inaky
 * You should get an error code before checking it ...
 *
 * Revision 1.3  1998/10/22 00:06:45  inaky
 * Moved to hidbp_config() the selection of the size to use for the
 * data. Also fixed bits when freeing.
 *
 * Revision 1.2  1998/09/14 02:04:23  inaky
 * Added all the common routines, such as config, open, close and search
 * for a free address.
 *
 * Revision 1.1  1998/09/11 00:50:57  inaky
 * Moved the HID Boot Protocol to a single driver, so it can be isolated
 * more easily. It supports keyboards and mice, settable at compile time.
 *
 * Revision 1.1  1998/08/30 13:55:46  inaky
 * Added boot protocol support and a simple test program.
 *
 * Revision 1.1  1998/08/28 23:53:14  inaky
 * New driver
 *
 * Revision 1.3  1998/08/28 00:14:22  inaky
 * New driver ...
 *
 */



