
/* USB Human Interface Devices Driver v$Revision: 1.5 $ 
 *
 * (C) 1997 by Iaky Prez Gonzlez <inaky@peloncho.fis.ucm.es>
 *
 * Distributed under the terms of the GNU Public License Version 2
 * See files COPYRIGHT and COPYING at the root distribution directory
 *
 * $Id: cfg.c,v 1.5 1998/11/14 17:19:46 inaky Exp $
 *
 * Changelog at the bottom 
 *
 *     USB Human Interface Devices - Configuration
 *     -------------------------------------------
 *
 *  This is the embrion of the USB HID Driver.
 *
 */

#include <uusbd/usbd.h>

#define GDLF_IDENT          "USB-HIDD"
#define GDLF_LOCAL          /* Define if want local settings */
#define GDLF_DEF_LOGLEVEL   9
#define GDLF_ASSERTIONS     /* want assertions (#undef if not) */
#define GDLF_FNTRACE        /* want function traces */
#define GDLF_LOCATIONS      /* want locations in log messages */
#include <uusbd/gdlf.h>

#include <uusbd/usb-hidbp.h>


    /* Configure a device class interface
     *
     *  This function may be used to be called from hidd_config() to
     * configure the class part of a configurable element.
     *
     * Returns:
     *
     *  int             Number of cfels configured, errno code < 0 on
     *                  error. 
     */
           
int hidd_config (ts_usb_cfel* cfel, ts_usb_device* device, 
                 ts_usb_cfg* cfg, t_usb_cfg cfg_idx, 
                 ts_usb_iface* iface, t_usb_iface iface_idx, 
                 ts_usb_aset* aset, t_usb_aset aset_idx, __u32 flags)
{
    int errcode;
    
    gdlf_fntrace ("hidd_config (0x%p, 0x%p, 0x%p, "
                  "0x%p, 0x%p, %u)", cfel, device, cfg, 
                  iface, aset, flags);
    gdlf_aassert (cfel != NULL, return -EINVAL);
    gdlf_aassert (device != NULL, return -EINVAL);
    
    if (cfel->class != HIDD_CLASS)
    {
        gdlf_err ("Pretended to configure as HID non-HID "
                  "class interface %d of device (%d %d)\n",
                  iface->addr, 
                  usb_host_addr (device), usb_dev_addr (device));
        return 0;
    }

        /* Ok, configure this cfel */

    switch (cfel->subclass)
    {
      case HIDD_SUBCLASS_BOOT:  /* we don't care about this one */
        gdlf_err ("Interface %d of device (%d %d) supports "
                  "HID Boot Protocol, will use Report Protocol\n",
                  iface->addr, usb_host_addr (device), 
                  usb_dev_addr (device));

      case HIDD_SUBCLASS_REPORT:
        gdlf_err ("Interface %d of device (%d %d) supports "
                  "HID Report Protocol, unimplemented\n", iface->addr, 
                  usb_host_addr (device), usb_dev_addr (device));
        errcode = -ENOSYS;
        break;
        

      default:
        gdlf_err ("Interface %d of device (%d %d): unknown "
                  "HID subclass code %d\n", iface->addr, 
                  usb_host_addr (device), usb_dev_addr (device),
                  cfel->subclass);
        errcode = -EINVAL;
        break;
    }        
    return errcode;
}


    /* Unconfigure all cfels configured by this driver
     *
     *  This function may be used to unconfigure a driver's
     * class-configured configurable elements.
     *
     * Parameters:  
     *
     *  ts_usb_device*  Pointer to the device structure to
     *                  unconfigure. 
     *
     *  unsigned        USB_CFG_* flags indicating why is this
     *                  function called.  
     *
     * Returns:
     *
     *  int             0 if ok, errno code < 0 on error.
     */

int hidd_unconfig (ts_usb_device* device, unsigned flags)
{
    t_usb_cfg cfg_cnt;
    ts_usb_cfg *cfg;
    ts_usb_iface *iface, *iface_top;
    ts_usb_aset *aset, *aset_top;
    
    gdlf_fntrace ("hidd_unconfig (0x%p, %u)", device, flags);
    gdlf_aassert (device != NULL, return -EINVAL);

    for (cfg_cnt = 0; cfg_cnt < device->pcfgs; cfg_cnt++)
    {
        cfg = device->pcfg[cfg_cnt];
        if (cfg == NULL)        /* unloaded? skip */
            continue;
        
            /* Walk cfg's interfaces */

        for (iface = cfg->iface, iface_top = iface + cfg->ifaces; 
             iface < iface_top; iface++)
        {
                /* Walk the iface's asets */

            for (aset = iface->aset, aset_top = aset + iface->asets; 
                 aset < aset_top; aset++)

                      /* cfged by us? */

                if (aset->cfel.configured
                    && aset->cfel.data
                    && (aset->cfel.data->dtor.fn
                         == (tf_dtor)hidd_destroy)
                        )
                {
                    ts_dtor_exec (&aset->cfel.data->dtor);
                    usb_kfree (aset->cfel.data = NULL);
                    aset->cfel.data = NULL;
                    aset->cfel.configured = 0;      /* uncfged */
                }
        }         
    }
    return 0;
}

/* Log:
 *
 * $Log: cfg.c,v $
 * Revision 1.5  1998/11/14 17:19:46  inaky
 * Moved the include files to uusbd/include/uusbd, to ease integration
 * into the Linux kernel.
 *
 * Revision 1.4  1998/10/31 02:16:55  inaky
 * Removed all the d_* pointers from ts_usb_cfel and collapsed them into
 * a single pointer; only one was really used.
 *
 * Revision 1.3  1998/09/11 00:51:50  inaky
 * Moved the boot protocol to a single driver, to isolate it more easily.
 *
 * Revision 1.2  1998/08/30 13:56:11  inaky
 * Fixed comments. Added support to multiplex to boot protocols.
 *
 * Revision 1.1  1998/08/28 23:53:14  inaky
 * New driver
 *
 * Revision 1.3  1998/08/28 00:14:22  inaky
 * New driver ...
 *
 */



