
/* USB Driver Stack v$Revision: 1.2 $ 
 *
 * (C) 1997 by Iaky Prez Gonzlez <inaky@peloncho.fis.ucm.es>
 *
 * Distributed under the terms of the GNU Public License Version 2
 * See files COPYRIGHT and COPYING at the root distribution directory
 *
 * $Id: cfg.c,v 1.2 1998/11/14 17:20:29 inaky Exp $
 *
 * Changelog at the bottom 
 *
 *
 *     Hub Driver - Configuration and friends
 *     --------------------------------------
 *
 */

#include <asm/spinlock.h>   /* spinlock_init() */
#define __NO_VERSION__
#include <linux/module.h>   /* MOD_*_USE_COUNT */

#include <uusbd/usbd.h>
#include <uusbd/usb-hubd.h>

#define GDLF_IDENT          "USB-HUBD"
#define GDLF_LOCAL          /* Define if want local settings */
#define GDLF_DEF_LOGLEVEL   gdlf_level_top
#define GDLF_ASSERTIONS     /* want assertions (#undef if not) */
#define GDLF_FNTRACE        /* want function traces */
#define GDLF_LOCATIONS      /* want locations in log messages */
#include <uusbd/gdlf.h>


    /* Configure a hub
     *
     *  This function is called by the UUSBD when it needs to
     * configure a hub.
     *
     *  It sets up all the data structures and registrations on all
     * needed places. Also initializes the 'tr' member to do the
     * status change transactions with the default callback. Use your
     * callback there if needed, but follow the working guidelines
     * (see the source code for it).
     *
     *  This stuff:
     *
     * usb_configuration_ref (cfg);
     * usb_device_ref (&device->msnode);
     * usb_host_ref (&hub->host->snode);
     *
     * is needed for the counts to be ok. The hub destructor,
     * usb_hub_destroy() will unreference on hub structure
     * destruction.
     *
     * Parameters:  
     *
     *  ts_usb_cfel*    Configurable element to configure.
     *
     *  ts_usb_device*  Pointer to the device structure to configure.
     *
     *  ts_usb_cfg*     USB configuration where the configurable
     *                  element is (NULL if it is the Device
     *                  Configurable Element).
     *
     *  ts_usb_iface*   Interface where the configurable element is
     *                  (NULL if it is the Device Configurable
     *                  Element).
     *
     *  ts_usb_aset*    Alternate setting where the configurable
     *                  element is (NULL if it is the Device
     *                  Configurable Element).
     *
     *  __u32           USB_CFG_* flags indicating why is this
     *                  function called.  
     *
     * Returns:
     *
     *  int             Number of cfels configured, errno code < 0 on
     *                  error. 
     */

int hubd_config (ts_usb_cfel* cfel, ts_usb_device* device, 
                 ts_usb_cfg* cfg, t_usb_cfg cfg_idx, 
                 ts_usb_iface* iface, t_usb_iface iface_idx, 
                 ts_usb_aset* aset, t_usb_aset aset_idx, __u32 flags)
{
    int errcode, cnt;
    ts_usb_hub *hub;
    ts_usb_addr addr;
    
    gdlf_fntrace ("hubd_config (0x%p, 0x%p, 0x%p, %u, "
                  "0x%p, %u, 0x%p, %u, %u)", cfel, device, 
                  cfg, cfg_idx, iface, iface_idx, 
                  aset, aset_idx, flags);
    
        /* Check class */

    if (cfel->class != HUBD_CLASS)
    {
        gdlf_err ("Tried to configure non-hub device "
                  "0x%p (%d %d) as hub\n",
                  device, usb_host_addr (device), 
                  usb_dev_addr (device));
        return 0;               /* nothing configured */
    }        

        /* check this is the device cfel */

    if (cfel != &device->cfel)
    {
        gdlf_err ("Hub configuration not called on device "
                  "wide basis for device 0x%p (%d %d)\n",
                  device, usb_host_addr (device), 
                  usb_dev_addr (device));
        return -EBADFD;
    }
    
        /* 
         * Set the active configuration to the self powered one if
         * possible
         */  

#warning FIXME: get the configuration from the persistence system, choose the best one [self powered]

    cnt = 0;
    errcode = usb_configuration_set (device, device->pcfg[cnt]->addr, 0);
    if (errcode < 0)
    {
        gdlf_err ("Cannot switch to configuration index "
                  "%d, address %d (%s) of device 0x%p (%d %d): "
                  "errno/usb %d/%d\n",
                  cnt, device->pcfg[cnt]->addr, 
                  device->pcfg[cnt]->descr?
                  device->pcfg[cnt]->descr : "name n/a",
                  device, usb_host_addr (device), 
                  usb_dev_addr (device),
                  errcode, usb_errno);
        return errcode;
    }

        /*
         *   Hub spec (USB 1.0 Spec, Chapter 11th) mandates just an
         * interface and one endpoint.
         */

    gdlf_aassert (device->active != NULL, return -EINVAL);
    gdlf_assert (device->active->ifaces == 1);
    gdlf_assert (device->active->iface[0].active->endps == 1);

        /* 
         * Now set all given pointers to what we need 
         *
         *  This is left so in order to prepare it to the time when a
         * device can have a hub in a single interface. This may
         * happen in a future. Now it looks like extra work? yep, but
         * it should be optimized out. If not, we can #ifdef it.
         */

    cfg = device->active;
    iface = &cfg->iface[0];
    aset = iface->active;
    cfel = &aset->cfel;
    
    gdlf_aassert ((cfg->iface - iface) / sizeof (ts_usb_iface) 
                  < cfg->ifaces,
                  return -EINVAL);
    gdlf_aassert ((iface->aset - aset) / sizeof (ts_usb_aset) 
                  < iface->asets,
                  return -EINVAL);
        
        /* Create a hub structure and fill basics */

    hub = ts_usb_hub_init();    /* alloc */
    if (hub == NULL)            /* on error ... */
        return -ENOMEM;         /* ... ts_usb_hub_init() complains itself */

        /* (hub destructor set when registering) */
    
    hub->configuring = 1;           /* we are configuring it */
    hub->host = device->hub->host;  /* set hub's host */
    usb_host_ref (hub->host);
    hub->device = device;           /* set hub's device */
    usb_device_ref (device);
    hub->cfg = cnt;
    usb_configuration_ref (cfg);
    hub->iface = (cfg->iface - iface) / sizeof (ts_usb_iface);
    hub->aset = (iface->aset - aset) / sizeof (ts_usb_aset);
    hub->operations = &hubd_operations;
    
        /* Fill in data from descriptor */

    errcode = hubd_descriptor_fill (hub);
    if (errcode < 0)
    {
        gdlf_err ("Couldn't fill hub 0x%p (%d %d) "
                  "descriptor data: "
                  "errno/usb %d/%d\n",
                  hub, hub->host->addr, hub->device->addr,
                  errcode, usb_errno);
        usb_configuration_unref (cfg);
        usb_device_unref (device, NULL);
        usb_host_unref (hub->host, NULL);
        ts_usb_hub_release (hub);
        usb_kfree (hub);
        return errcode;
    }

        /* Set up the status change descriptor */
    
    addr.host = hub->host->addr;
    addr.device = hub->device->addr;
    addr.endp = aset->endp[USB_HUB_ENDP_STAT_CH].addr;
    errcode = usb_open (&addr, &hub->statch);
    if (errcode < 0)
    {
        gdlf_err ("Couldn't open hub 0x%p (%d %d)'s "
                  "status change endpoint: "
                  "errno/usb %d/%d\n",
                  hub, hub->host->addr, hub->device->addr,
                  errcode, usb_errno);
        usb_configuration_unref (cfg);
        usb_device_unref (device, NULL);
        usb_host_unref (hub->host, NULL);
        ts_usb_hub_release (hub);
        usb_kfree (hub);
        return errcode;
    }

        /* set the transfers data */

    hub->statch_tr.size = (1 + hub->ports)/8  
        + ((1 + hub->ports) % 8? 1 : 0);
    hub->statch_tr.data = 
        usb_kmalloc (hub->statch_tr.size, GFP_KERNEL);
    if (hub->statch_tr.data == NULL)
    {
        gdlf_err ("No memory to set up a %u bytes data "
                  "buffer for hub 0x%p (%d %d)'s status "
                  "change endpoint\n",
                  hub->statch_tr.size, hub, hub->host->addr, 
                  hub->device->addr);
        hub->statch_tr.size = 0;
        usb_configuration_unref (cfg);
        usb_device_unref (device, NULL);
        usb_host_unref (hub->host, NULL);
        usb_kfree (hub);
        return errcode;
    }
    hub->statch_tr.warn.callback = usb_hub_status_callback;
    hub->statch_tr.method = 1;
#if 0
    hub->statch_tr.user = 0;    
#endif
    hub->statch_tr.status = e_usb_tr_status_filled;
    hub->statch_tr.open = &hub->statch;

    errcode = usb_tr_setup (&hub->statch_tr);
    if (errcode < 0)
    {
        gdlf_err ("Cannot setup a read from hub 0x%p "
                  "(%d %d)'s status change endpoint: "
                  "errno/usb %d/%d\n",
                  hub, hub->host->addr, hub->device->addr,
                  errcode, usb_errno);
        usb_tr_destroy (&hub->statch_tr);
        usb_configuration_unref (cfg);
        usb_device_unref (device, NULL);
        usb_host_unref (hub->host, NULL);
        ts_usb_hub_release (hub);
        usb_kfree (hub);
        return errcode;
    }

    errcode = usb_tr_insert (&hub->statch_tr);
    if (errcode < 0)
    {
        gdlf_err ("Cannot insert a read from hub 0x%p "
                  "(%d %d)'s status change endpoint: "
                  "errno/usb %d/%d\n",
                  hub, hub->host->addr, hub->device->addr,
                  errcode, usb_errno);
        usb_tr_destroy (&hub->statch_tr);
        usb_configuration_unref (cfg);
        usb_device_unref (device, NULL);
        usb_host_unref (hub->host, NULL);
        ts_usb_hub_release (hub);
        usb_kfree (hub);
        return errcode;
    }
    
    cfel->data = ts_snode_ref (&hub->snode);    /* attach to device */
    device->cfel.configured = 
        cfel->configured = (__u32)hub;          /* it's configured! */    
    hub->configuring = 0;                       /* done configuring the hub */
    hubd_reference (hub, 1);                    /* we use the driver */


        /* Register the hub structure (it starts scheduling) */

    errcode = usb_hub_register (hub);
    if (errcode < 0)
    {
        gdlf_err ("Couldn't register hub 0x%p (%d %d): "
                  "errno/usb %d/%d\n",
                  hub, hub->host->addr, hub->device->addr,
                  errcode, usb_errno);
        hubd_reference (hub, -1);
        cfel->data = NULL;
        cfel->configured = 0;           /* it's configured! */
        usb_configuration_unref (cfg);
        usb_device_unref (device, NULL);
        usb_host_unref (hub->host, NULL);
        ts_snode_unref (&hub->snode, NULL); /* this will kfree() */
        ts_usb_hub_release (hub);
        return errcode;
    }

        /* Power on all the ports */

    if (hub->power <= 2)            /* can we turn ports' power on? */
    {
        for (cnt = 0; cnt < hub->ports; cnt++)
        {
            errcode = usb_port_feature_set (hub, cnt, e_usb_port_fes_power);
            if (errcode < 0)    
                gdlf_err ("Error powering on port %u "
                          "of hub 0x%p, device (%d %d): "
                          "errno/usb %d/%d\n",
                          cnt, hub, hub->host->addr, 
                          hub->device->addr, 
                          errcode, usb_errno);
        }
        usb_waitms (2 * hub->pwrontime);
    }
    return 1;                           /* we've configured one cfel */
}


    /* Unconfigure a device if hub
     *
     *  This routine is called by when it is required to unconfigure
     * all devices which use a particular driver. It has to make sure
     * that any device which have any cfel configured by this driver
     * is unconfigured.
     *
     *  So, the easiest way is to find this device into the driver use
     * list. The use list keeps pointers to hub structures. If we find
     * the pointer to this device's hub structure at our use list,
     * then we must unconfigure the device.
     *
     *  To accomplish that task, instead of calling directly the
     * usb_hub_destroy() destructor, we call the structure's
     * destructor and then free the data.
     *
     * Parameters:  
     *
     *  ts_usb_device*  Pointer to the device structure to
     *                  unconfigure. 
     *
     *  __u32           USB_CFG_* flags indicating why is this
     *                  function called.  
     *
     * Returns:
     *
     *  int             0 if ok, errno code < 0 on error.
     */

int hubd_unconfig (ts_usb_device* device, __u32 flags)
{
    int count;                      /* unconfigured cfels */
    ts_snode *itr, *next, *this;

    gdlf_fntrace ("hubd_unconfig (0x%p, 0x%x)", device, flags);

    this = device->cfel.data;       /* hub drivers hook there */
    count = 0;
    itr = ts_slist_head (&hubd_driver->uselist);
    while (itr)
    {
        next = ts_slist_next (itr);
        if (((ts_usb_drv_use*)itr)->user == this)
        {
            ts_dtor_exec (&itr->dtor);  /* destroy ~ unconfigure */  
            count++;
        }
        itr = next;
    }
    return count; 
}

/* Log:
 *
 * $Log: cfg.c,v $
 * Revision 1.2  1998/11/14 17:20:29  inaky
 * Moved the include files to uusbd/include/uusbd, to ease integration
 * into the Linux kernel.
 *
 * Revision 1.1  1998/11/09 21:34:47  inaky
 * Split the driver.c file into driver-only and configuration issues
 * stuff.
 *
 *
 */

