
/* USB Driver Stack v$Revision: 1.32 $ 
 *
 * (C) 1997 by Iaky Prez Gonzlez <inaky@peloncho.fis.ucm.es>
 *
 * Distributed under the terms of the GNU Public License Version 2
 * See files COPYRIGHT and COPYING at the root distribution directory
 *
 * $Id: driver.c,v 1.32 1998/11/14 17:20:30 inaky Exp $
 *
 * Changelog at the bottom 
 *
 *
 *     Hub Driver - Driver housekeeping
 *     --------------------------------
 *
 *  Driver definition and basic code for the Hubs
 *
 */

#include <asm/spinlock.h>   /* spinlock_init() */
#include <linux/init.h>     /* MOD_*_USE_COUNT */
#include <linux/module.h>   /* MOD_*_USE_COUNT */

#include <uusbd/usbd.h>
#include <uusbd/usb-hubd.h>

#if defined (CONFIG_USB_HUBD_MODULE)
#include <linux/kmod.h>
#endif

#define GDLF_IDENT          "USB-HUBD"
#define GDLF_LOCAL          /* Define if want local settings */
#define GDLF_DEF_LOGLEVEL   gdlf_level_top
#define GDLF_ASSERTIONS     /* want assertions (#undef if not) */
#define GDLF_FNTRACE        /* want function traces */
#define GDLF_LOCATIONS      /* want locations in log messages */
#include <uusbd/gdlf.h>


    /* Driver information */

#define NAME                /* short!! */       \
    "UUSBD Hub Driver v$Revision: 1.32 $"

#define COPYRIGHT                               \
    "(c) 1997-98 Inaky Perez Gonzalez "         \
    "<inaky@peloncho.fis.ucm.es>"

#define VENDOR 0
#define VENDORNAME                              \
    NULL

#define PRODUCT 0
#define DEVICENAME                              \
    NULL

#define CLASS HUBD_CLASS
#define CLASSNAME                               \
    "Hub"

#define SUBCLASS 0
#define SUBCLASSNAME                            \
    NULL

#define PROTOCOL 0
#define PROTOCOLNAME                            \
    NULL


ts_usb_driver *hubd_driver = NULL;      /* Driver structure */

const ts_usb_hub_ops hubd_operations =  /* Operations this driver does */
{
    hubd_schedule_status_change,
    hubd_descriptor_get,
    hubd_descriptor_set,
    hubd_descriptor_fill,
    hubd_status_get,
    hubd_bus_status_get,
    hubd_feature,
    hubd_port_feature,
    hubd_port_status_get,
    hubd_port_reset,
    hubd_free
};


    /* Initialize the hubd driver
     *
     * Returns:
     *
     *  int             0 if ok, errno code < 0 on error.
     *
     * Global:
     *
     *  int usb_errno      On error, error code.  
     */

int __init hubd_init (void)
{
    int errcode;
    
    gdlf_fntrace ("hubd_init (void)");
    gdlf_info (NAME "\n");
#ifdef MODULE
    gdlf_info (COPYRIGHT "\n");
#endif

    hubd_driver = ts_usb_driver_init();
    if (hubd_driver == NULL)
    {
        usb_errno = -USBE_NO_MEMORY;
        return -ENOMEM;
    }

        /* Set driver's attributes */

    hubd_driver->vendor = VENDOR;
    hubd_driver->product = PRODUCT;
    hubd_driver->class = CLASS;
    hubd_driver->subclass = SUBCLASS;
    hubd_driver->protocol = PROTOCOL;

        /* Set info */

    hubd_driver->name = NAME;                   /* driver name */
    hubd_driver->copyright = COPYRIGHT;         /* driver copyright */
    hubd_driver->vendorname = VENDORNAME;       /* device's vendor name */
    hubd_driver->productname = DEVICENAME;      /* device's name */
    hubd_driver->classname = CLASSNAME;         /* class' name */
    hubd_driver->subclassname = SUBCLASSNAME;   /* subclass' name */
    hubd_driver->protocolname = PROTOCOLNAME;   /* protocol name */

    hubd_driver->config = hubd_config;          /* config method */
    hubd_driver->unconfig = hubd_unconfig;      /* unconfig method */
    ts_snode_ref (&hubd_driver->snode);         /* we hold a reference to it */    

        /* Register the driver */

    errcode = usb_drv_register (hubd_driver);   /* register driver */
    if (errcode)
    {
        gdlf_err ("Cannot register UUSBD Hub Driver: "
                  "errno/usb %d/%d\n",
                  errcode, usb_errno);
        if (hubd_driver->snode.next == NULL)
        {
            ts_usb_driver_clean (hubd_driver);
            usb_kfree (hubd_driver);
        }
        return errcode;
    }

    return 0;
}


    /* Release the hubd driver
     *
     *  Do what is needed to ensure the use count of the driver is
     * zero so it can be unloaded from the kernel. What we got to do
     * is to call the driver use release function and then unconfigure
     * all hubs which still use this driver. 
     *
     * Returns:
     *
     *  int             0 if ok, errno code < 0 on error.
     */


int hubd_release (void)
{    
    ts_snode *itr, *next;

    gdlf_fntrace ("hubd_release (void)");

    usb_drv_release (hubd_driver);  /* release devices using us */

        /* 
         * Now try to release all devices which might be using us
         * indirectly by unreferencing them 
         */

    itr = ts_slist_head (&hubd_driver->uselist);
    while (itr)
    {
        next = ts_slist_next (itr);
        ts_snode_unref (itr, NULL);
        itr = next;
    }

        /*
         * Ok, if there still are people on the list, it means some
         * failed, so we'll do it in the rude way.
         */

    itr = ts_slist_head (&hubd_driver->uselist);
    if (itr)
        gdlf_warning ("Some devices neglected to release their "
                      "use of the hub driver. Will have to use "
                      "the rude way. Expect inconsistencies\n");
    while (itr)
    {
        next = ts_slist_next (itr);
        usb_drv_release_use (hubd_driver, (ts_usb_drv_use*)itr, NULL);
        itr = next;
    }

    return 0;
}


    /* Shutdown the hub driver
     *
     * Returns:
     *
     *  int             0 if ok, errno code < 0 on error.
     */

int hubd_shutdown (void)
{
    int errcode;

    gdlf_fntrace ("hubd_shutdown()");

    hubd_release();
    ts_snode_unref (&hubd_driver->snode, NULL);

        /* 
         *  We'll ignore the error if it cannot be found on the list,
         * as it may happend that when the initialization fails, the
         * hub driver is not registered, and thus there's no need to
         * unregister it :) 
         */

    errcode = usb_drv_unregister (hubd_driver);
    hubd_driver = NULL;
    if (errcode && usb_errno != -USBE_DRIVER_UNREFERENCED)
        gdlf_err ("Cannot unregister hub driver: "
                  "errno/usb %d/%d\n",
                  errcode, usb_errno);
    return errcode;
}


    /* Add/remove a reference to the driver
     *
     *  This function is used to tell the hub driver if we are or not
     * using any of it's data structures or functions. When we take a
     * pointer to any of the driver's function in the operations
     * structure, we must reference (use) it, so we call this with a
     * positive argument. When we are done using them, we call it with
     * a negative argument.
     *
     * Parameters:  
     *
     *  ts_usb_hub*     Pointer to the hub with which we are using
     *                  the driver.  
     *
     *  __u8            Positive for ONE reference, negative for ONE
     *                  unreference. If zero, return use count.
     *
     * Returns:
     *
     *  int             0 if ok, errno code < 0 on error.
     */
    
int hubd_reference (ts_usb_hub* hub, int no)
{
    gdlf_fntrace ("hubd_reference (0x%p, %d)", hub, no);
    
    if (no > 0)             /* reference */
        MOD_INC_USE_COUNT;  /* module's too */
    else if (no < 0)        /* unreference */
        MOD_DEC_USE_COUNT;  /* module's too */

    return usb_drv_reference (hubd_driver, &hub->snode, no);
}


    /* Modular stuff */

#ifdef MODULE

int init_module (void)
{
    return hubd_init();
}

void cleanup_module (void)
{
    hubd_shutdown();
    return;
}

#endif

/* Log:
 *
 * $Log: driver.c,v $
 * Revision 1.32  1998/11/14 17:20:30  inaky
 * Moved the include files to uusbd/include/uusbd, to ease integration
 * into the Linux kernel.
 *
 * Revision 1.31  1998/11/09 21:35:28  inaky
 * Moved the configuration stuff to a single file, and fixed the
 * reference function to decrement mod use count when unreferencing ...
 *
 * Revision 1.30  1998/10/31 02:17:13  inaky
 * Removed all the d_* pointers from ts_usb_cfel and collapsed them into
 * a single pointer; only one was really used.
 *
 * Revision 1.29  1998/09/24 16:08:48  inaky
 * Moved 'hub_port' to 'port'.
 *
 * Revision 1.28  1998/09/14 20:28:54  inaky
 * Removed the usb_hub_errno variable.
 *
 * Revision 1.27  1998/08/27 22:19:00  inaky
 * As this will crash and has to be tested step by step, I -ENOSYS it, so
 * I won't have to reboot ... :)
 *
 * Revision 1.26  1998/08/24 02:28:58  inaky
 * Changed name of all enumerations to be prefixed with the enumeration
 * type name.
 *
 * Revision 1.25  1998/07/23 14:38:50  inaky
 *  Removed the direct user memory access (temporarily).
 *
 * Revision 1.24  1998/07/16 14:14:42  inaky
 * Moved from the old read/write interface to the new, request-based
 * interface.
 *
 * Revision 1.23  1998/07/10 10:38:30  inaky
 * Now the config routine delegates on usb_hub_register() to do the
 * initial schedulement of status change read.
 *
 * Revision 1.22  1998/06/21 20:07:52  inaky
 * Fixed the operations structure; moved the use count and use list to
 * the driver structure; fixed some docs and the hubd_release() function
 * to release in three stages (to try to release _everything_).
 *
 * Added the calculation of the count of unconfigured cfels to
 * hubd_unconfig().
 *
 * Fixed hubd_reference() to use the global driver reference function
 * after doing the module use count update.
 *
 * Fixed the module initialization calls.
 *
 * Revision 1.21  1998/06/20 21:55:16  inaky
 * Massive changes to split the old monolithic hubd driver into a hub
 * manager (hub wrappers and daemon) and a dynamic driver, which can be
 * replaced and its different parts reused.
 *
 * Revision 1.20  1998/05/02 00:52:04  inaky
 * Added the khubdd initialization code to hubd_init() and shutdown to
 * hubd_shutdown()
 *
 * Revision 1.19  1998/05/01 19:15:38  inaky
 * Fixed device address reporting on error strings.
 *
 * Added power on of all ports to the configuration routine [as it's
 * normal that powered off ports don't work ... :)]
 *
 * Revision 1.18  1998/04/26 03:39:55  inaky
 * Added the spinlock initialization.
 *
 * Revision 1.17  1998/04/13 22:18:39  inaky
 * Finished most of the hub driver; some updates still needed.
 *
 * Revision 1.16  1998/02/12 14:39:54  inaky
 * Drivers config method should inc the module use count when entering
 * and decrement it upon return.
 *
 * Revision 1.15  1998/02/10 17:40:46  inaky
 * Moved to use ts_[N]snode_[un]ref() and clean the driver structure
 * pointer when releasing it.
 *
 * Revision 1.14  1998/02/09 13:46:49  inaky
 * Moved all use counts to atomic, added #defines for the vicsp codes,
 * fixed the releasing of the driver structure when there's an error on
 * registration, to evitate freeing it if it's already on the drivers
 * list. This should be checked. FIXME.
 *
 * Revision 1.13  1998/02/08 22:01:08  inaky
 * The hubd_init() function wasn't registering properly the driver, as
 * the class code wans't being set.
 *
 * Revision 1.12  1998/01/11 15:44:36  inaky
 * Added docs?
 *
 * Revision 1.11  1998/01/08 12:19:18  inaky
 * Added copyright strings to all important files
 *
 * Revision 1.10  1998/01/08 11:50:02  inaky
 * Adapted to the new configuration scheme
 *
 * Revision 1.9  1998/01/02 00:07:29  inaky
 * Moved to the new driver registration scheme
 *
 * Revision 1.8  1997/12/30 22:30:29  inaky
 * Fixed some bugs.
 *
 * Revision 1.7  1997/12/30 20:14:58  inaky
 * Moved to the new lists code. Untested yet.
 *
 * Revision 1.6  1997/12/29 03:23:44  inaky
 * Removed some stale labels
 *
 * Revision 1.5  1997/12/29 00:26:14  inaky
 * Moving the khubdd to use signals as commands.
 *
 * Revision 1.4  1997/12/27 16:44:31  inaky
 * Moving to GDLF and fixinf docs. Removing more cruft.
 *
 * Revision 1.3  1997/12/27 02:06:13  inaky
 * Moving everything to use the GDLF for logging and debug checkings.
 * Removed the global array of USB hosts and moved it to a linked
 * list. Added an address field to each USB host controller.
 * Removed lots of cruft that were accumulated and useless. Still some
 * missing, but may be useful someday.
 *
 * Revision 1.2  1997/12/13 20:56:37  inaky
 * Moving all the documentation to the .c files except the ones for data
 * structures and inline functions. Helps making the header files cleaner.
 *
 * Revision 1.1  1997/12/08 02:41:52  inaky
 * Moving the hubd to the new driver style
 *
 */

