
/* USB Driver Stack v$Revision: 1.15 $ 
 *
 * (C) 1997 by Iaky Prez Gonzlez <inaky@peloncho.fis.ucm.es>
 *
 * Distributed under the terms of the GNU Public License Version 2
 * See files COPYRIGHT and COPYING at the root distribution directory
 *
 * $Id: ports.c,v 1.15 1998/11/14 17:20:35 inaky Exp $
 *
 * Changelog at the bottom 
 *
 *
 *     HUB Driver
 *     ----------
 *
 *     Code for making standard requests to the hubs' ports
 */

#include <uusbd/usbd.h>
#include <uusbd/usb-hubd.h>

#define GDLF_IDENT          "USB-HUBD"
#define GDLF_LOCAL          /* Define if want local settings */
#define GDLF_DEF_LOGLEVEL   9
#define GDLF_ASSERTIONS     /* want assertions (#undef if not) */
#define GDLF_FNTRACE        /* want function traces */
#define GDLF_LOCATIONS      /* want locations in log messages */
#include <uusbd/gdlf.h>


    /* Dirty-work function for hubd_port_feature_*()
     *
     *  THIS FUNCTION IS FOR INTERNAL USE, DO NOT CALL IT DIRECTLY
     *
     * Parameters:  
     *
     *  ts_usb_hub*     Pointer to hub structure
     *
     *  __u8            Port number
     *
     *  e_usb_featsel   Feature
     *
     *  e_usb_req       USB request code
     *
     * Returns:
     *
     *  int             0 if ok, errno code < 0 on error.
     */

int hubd_port_feature (ts_usb_hub* hub, __u8 port,
                       e_usb_featsel feature, e_usb_req request)
{
    ssize_t errcode;
    ts_usb_request req;
    
    gdlf_fntrace ("hubd_port_feature (0x%p, %u, %u, %u)",
                  hub, port, feature, request);
    gdlf_aassert (hub != NULL, return -EINVAL);
    gdlf_aassert (port < hub->ports, return -EINVAL);

    req.bmreqtype = 0x23;
    req.request = request;
    req.value = feature;
    req.index = port;

    read_lock (&hub->device->msnode.lock);    
    errcode = usb_request_dataless (hub->device, &req,
                                    "hub port feature change");
    if (errcode < 0)
        gdlf_err ("Cannot change hub port feature; request failed: "
                  "errno/usb %d/%d\n", errcode, usb_errno);
    read_unlock (&hub->device->msnode.lock);
    return errcode;
}


    /* Get the port status, as well as the status bits which have
     * experienced a change. On the lower word are stored the status
     * bits and in the higher word the status change bits. You can use
     * the #defined masks USB_PORT_STAT_* for the status bits and
     * USB_PORT_STAT_C_* for the status change bits.
     *
     *  To acknowledge the changes reported with this function [most
     * of them], use the usb_port_feature_clear() function.
     *
     * NOTE: all the values are in host order, though USB uses little
     *       endian.
     *
     * Compliance:      USB Spec 1.0 - 11.12.2.6
     *
     * Parameters:  
     *
     *  ts_usb_hub*     Pointer to the hub structure. If NULL, -EINVAL
     *                  will be returned.  
     *
     *  __u8            Port number.
     *
     * Returns:
     *
     *  __s32           If > 0, status code (and with HUBD_PORT_STAT_* to
     *                  get the bits) [FIXME: things may get hairy if
     *                  sometime someone decides to use bit 31 for
     *                  status ...]. On error, <0 errno code.
     */

__s32 hubd_port_status_get (ts_usb_hub* hub, __u8 port)
{
    int errcode;
    ts_usb_request req;
    __u32 status;
    
    gdlf_fntrace ("hubd_port_status_get (0x%p, %u)", hub, port);
    gdlf_aassert (hub != NULL, return -EINVAL);
    gdlf_aassert (port < hub->ports, return -EINVAL);

    req.bmreqtype = 0x80;
    req.request = e_usb_req_get_status;
    req.value = 0;
    req.index = port+1;
    req.length = sizeof (status);
    
    read_lock (&hub->device->msnode.lock);
    errcode = usb_request_data (hub->device, &req, &status, 
                                "get port status");
    if (errcode < 0)
        gdlf_err ("Cannot get port status; request failed: "
                  "errno/usb %d/%d\n", errcode, usb_errno);
    read_unlock (&hub->device->msnode.lock);

        /* The hub manager updates the port structure */

    return errcode < 0? errcode : le32_to_cpu (status);
}


    /* Reset a hub's port.
     *
     *  This function will reset a USB port, blocking [interruptibly]
     * until the reset is over. FIXME: Must use some sort of wait
     * queuing here; adding a pointer field to the port structure to
     * the wait queue.
     *
     * Parameters:  
     *
     *  ts_usb_hub*     Pointer to the hub structure. If NULL, -EINVAL
     *                  will be returned.  
     *
     *  __u8            Number of the port to reset. -ENODEV if the
     *                  port doesn't exist.
     *
     * Returns:
     *
     *  int             0 if ok, errno code < 0 on error.
     */

#define USB_RESET_WAIT_RETRIES 10

int hubd_port_reset (ts_usb_hub* hub, __u8 port)
{
    int errcode, cnt;
    ts_usb_port *pport;
    
    gdlf_fntrace ("hubd_port_reset (0x%p, %u)", hub, port);

    pport = &hub->port[port];
    
        /* Request the reset */

    pport->reset = 1;
    errcode = usb_port_feature_set (hub, port, e_usb_port_fes_reset);
    if (errcode < 0)
    {
        gdlf_err ("Error resetting port %u of hub 0x%p, "
                  "device (%d %d): "
                  "errno/usb %d/%d\n",
                  port, hub, hub->host->addr, 
                  hub->device->addr, 
                  errcode, usb_errno);
        return errcode;
    }

        /* Wait 10ms, as stated by USB 1.0 Spec 11.6.2 */

    cnt = 0;
    while (cnt < USB_RESET_WAIT_RETRIES)
    {
        usb_waitms (USB_PORT_RESET_TIME);
        if (usb_port_status_get (hub, port) 
            & USB_PORT_STAT_C_RESET)
            break;
        cnt++;
    }

    if (cnt >= USB_RESET_WAIT_RETRIES) /* not resetted? */
    {
        gdlf_err ("Error resetting port %u of hub 0x%p (%d -1): "
                  "reset process was interrupted or not completed\n",
                  port, hub, hub->host->addr);
        usb_errno = -USBE_PORT_RESET_INT;
        write_unlock (&hci->snode.lock);
        return -ENOMEDIUM;
    }

    return 0;
}


/* Log:
 *
 * $Log: ports.c,v $
 * Revision 1.15  1998/11/14 17:20:35  inaky
 * Moved the include files to uusbd/include/uusbd, to ease integration
 * into the Linux kernel.
 *
 * Revision 1.14  1998/10/14 21:48:18  inaky
 * Fixed request to clear feature. I don't know where did I get the idea
 * that was needed.
 *
 * Revision 1.13  1998/09/24 16:08:58  inaky
 * Moved 'hub_port' to 'port'.
 *
 * Revision 1.12  1998/09/14 20:29:57  inaky
 * Severed assertions, removed the usb_hub_errno variable, changed the
 * reset behaviour to wait for reset to complete by looping instead of
 * using a wait queue ... it made no sense, anyway.
 *
 * Revision 1.11  1998/08/24 02:29:08  inaky
 * Changed name of all enumerations to be prefixed with the enumeration
 * type name.
 *
 * Revision 1.10  1998/07/23 14:39:03  inaky
 * Added the length parameter to the request structure.
 *
 * Revision 1.9  1998/07/16 14:15:00  inaky
 * Moved from the old read/write interface to the new, request-based
 * interface.
 *
 * Revision 1.8  1998/07/14 15:44:05  inaky
 * Moved the status bit update code from the hub driver to the hub
 * manager wrapping functions.
 *
 * Revision 1.7  1998/06/20 21:55:17  inaky
 * Massive changes to split the old monolithic hubd driver into a hub
 * manager (hub wrappers and daemon) and a dynamic driver, which can be
 * replaced and its different parts reused.
 *
 * Revision 1.6  1998/05/01 19:18:06  inaky
 * Fixed device address reporting on messages.
 *
 * Revision 1.5  1998/04/26 03:41:26  inaky
 * Addded the delays where needed
 *
 * Revision 1.4  1998/04/13 22:18:40  inaky
 * Finished most of the hub driver; some updates still needed.
 *
 * Revision 1.3  1998/01/08 12:19:19  inaky
 * Added copyright strings to all important files
 *
 * Revision 1.2  1997/12/29 00:26:15  inaky
 * Moving the khubdd to use signals as commands.
 *
 * Revision 1.1  1997/12/13 20:56:38  inaky
 * Moving all the documentation to the .c files except the ones for data
 * structures and inline functions. Helps making the header files cleaner.
 *
 *
 */

