
/* USB Driver Stack v$Revision: 1.15 $ 
 *
 * (C) 1997 by Iaky Prez Gonzlez <inaky@peloncho.fis.ucm.es>
 *
 * Distributed under the terms of the GNU Public License Version 2
 * See files COPYRIGHT and COPYING at the root distribution directory
 *
 * $Id: requests.c,v 1.15 1998/11/14 17:20:36 inaky Exp $
 *
 * Changelog at the bottom 
 *
 *
 *     HUB Driver
 *     ----------
 *
 *     Code for making standard requests to the hubs
 */

#include <uusbd/usbd.h>
#include <uusbd/usb-hubd.h>

#define GDLF_IDENT          "USB-HUBD"
#define GDLF_LOCAL          /* Define if want local settings */
#define GDLF_DEF_LOGLEVEL   9
#define GDLF_ASSERTIONS     /* want assertions (#undef if not) */
#define GDLF_FNTRACE        /* want function traces */
#define GDLF_LOCATIONS      /* want locations in log messages */
#include <uusbd/gdlf.h>


    /* Get the hub descriptor.
     *
     * Compliance:      USB Spec 1.0 11.12.2.4
     *
     * Parameters:  
     *
     *  ts_usb_hub*     Pointer to hub structure.
     *
     *  __u8            Descriptor index. Normally zero.
     *
     *  void*           Pointer to the buffer.
     *
     *  size_t          Size of the buffer for the descriptor
     *
     * Returns:
     *
     *  ssize_t         Bytes readen if ok [excluding request
     *                  phase's], errno code < 0 on error. Note that
     *                  no more than size_t bytes will be readen. Is
     *                  up to the caller to determine if the
     *                  descriptor has been fully read. 
     *
     * Global:
     *
     *  int usb_errno   On error, error code.  
     */

ssize_t hubd_descriptor_get (ts_usb_hub* hub, __u8 index, 
                             void* data, size_t size)
{
    int errcode;
    ts_usb_request req;
        
    gdlf_fntrace ("hubd_descriptor_get (0x%p, %u, 0x%p, %u)",
                  hub, index, data, size);

    req.bmreqtype = 0xa0;
    req.request = e_usb_req_get_descriptor;
    req.value = ((e_usb_desc_hub & 0xff) << 8) | index;
    req.index = 0;
    req.length = size;
    
    read_lock (&hub->device->msnode.lock);
    errcode = usb_request_data (hub->device, &req, data,
                                "get hub descriptor");
    if (errcode < 0)
        gdlf_err ("Cannot get hub descriptor; request failed: "
                  "errno/usb %d/%d\n", errcode, usb_errno);
    read_unlock (&hub->device->msnode.lock);
    return errcode;
}


    /* Set a hub's descriptor
     *
     *  Overwrites the previous upon succesful reception of the new
     * one. Any of the hub's available descriptors may be overwritten,
     * specifying it's type and it's index number. However, not all
     * hubs support this operation, as this request is optional and
     * writes data to a class-specific descriptor.
     *
     * Compliance:      USB Spec 1.0 11.12.2.7
     *
     * Parameters:  
     *
     *  ts_usb_hub*     Pointer to the hub structure. If NULL, -EINVAL
     *                  will be returned.  
     *
     *  __u8            Descriptor index.
     *
     *  const void*     Pointer to the descriptor data in little
     *                  endian format. If NULL, -EINVAL will be
     *                  returned. 
     *
     *  size_t          Length of the descriptor data. If 0, -EINVAL
     *                  will be returned. 
     *
     * Returns:
     *
     *  ssize_t         Bytes transmited [excluding request phase's]
     *                  or errno code < 0 on error. 
     *
     * Global:
     *
     *  int usb_errno   On error, error code.  
     */

ssize_t hubd_descriptor_set (ts_usb_hub* hub, __u8 index, 
                             const void* data, size_t size)
{
    int errcode;
    ts_usb_request req;
        
    gdlf_fntrace ("hubd_descriptor_set (0x%p, %u, 0x%p, %u)",
                  hub, index, data, size);

    req.bmreqtype = 0x20;
    req.request = e_usb_req_set_descriptor;
    req.value = ((e_usb_desc_hub & 0xff) << 8) | index;
    req.index = 0;
    req.length = size;
    
    write_lock (&hub->device->msnode.lock);
    errcode = usb_request_data (hub->device, &req, (void*)data, 
                                "set hub descriptor");
    if (errcode < 0)
        gdlf_err ("Cannot set hub descriptor; request failed: "
                  "errno/usb %d/%d\n", errcode, usb_errno);
    write_unlock (&hub->device->msnode.lock);
    return errcode;
}


    /* Dirty-work function for hubd_feature_*()
     *
     *  THIS FUNCTION IS FOR INTERNAL USE, DO NOT CALL IT DIRECTLY
     *
     * Parameters:  
     *
     *  ts_usb_hub*     Pointer to hub structure
     *
     *  e_usb_featsel   Feature
     *
     *  e_usb_req       USB request code
     *
     * Returns:
     *
     *  int             0 if ok, errno code < 0 on error.
     *
     * Global:
     *
     *  int usb_errno   On error, error code.  
     */

int hubd_feature (ts_usb_hub* hub, e_usb_featsel feature, e_usb_req request)
{
    ssize_t errcode;
    ts_usb_request req;
    
    gdlf_fntrace ("hubd_feature (0x%p, %u, %u)",
                  hub, feature, request);
    gdlf_aassert (hub != NULL, return -EINVAL);

    req.bmreqtype = 0x20;
    req.request = request;
    req.value = feature;
    req.index = 0;

    write_lock (&hub->device->msnode.lock);
    errcode = usb_request_dataless (hub->device, &req,
                                    "hub feature change");
    if (errcode < 0)
        gdlf_err ("Cannot change hub feature; request failed: "
                  "errno/usb %d/%d\n", errcode, usb_errno);
    write_unlock (&hub->device->msnode.lock);
    return errcode;
}


    /* Get hub status and change bits
     *
     *  The data is returned in 32 bits, being testable with the
     * #defined masks HUBD_STATUS_* for the status bits and
     * HUBD_STATUS_C_* for the status change bits.
     *
     * Compliance:      USB Spec 1.0 11.12.2.5
     *
     * Parameters:  
     *
     *  ts_usb_hub*     Pointer to the hub structure. If NULL, -EINVAL
     *                  will be returned.  
     *
     * Returns:
     *
     *  __s32           If > 0, status code (and with HUBD_STAT_* to
     *                  get the bits) [FIXME: things may get hairy if
     *                  sometime someone decides to use bit 31 for
     *                  status ...]. On error, <0 errno code.
     */

__s32 hubd_status_get (ts_usb_hub* hub)
{
    int errcode;
    ts_usb_request req;
    __u32 status;
    
    gdlf_fntrace ("hubd_status_get (0x%p)", hub);
    gdlf_aassert (hub != NULL, return -EINVAL);

    req.bmreqtype = 0xa0;
    req.request = e_usb_req_get_status;
    req.value = 0;
    req.index = 0;
    req.length = sizeof (status);
    
    read_lock (&hub->device->msnode.lock);
    errcode = usb_request_data (hub->device, &req, &status, 
                                "get hub status");
    if (errcode < 0)
        gdlf_err ("Cannot get hub status; request failed: "
                  "errno/usb %d/%d\n", errcode, usb_errno);
    read_unlock (&hub->device->msnode.lock);
    return errcode < 0? errcode : le32_to_cpu (status);    
}


    /* Read the status of the bus
     *
     *  This function will read the bus status on fiven port as
     * sampled at the last EOF2 hub state. It will be returned on bits
     * 0 and 1 of the return value if > 0. You can then use the mask
     * values USB_HUB_BUS_STAT_DP and USB_HUB_BUS_STAT_DN to sample the
     * values of the D- and D+ lines.
     *
     * NOTE: This is a very lowlevel function, intended as a debugging
     *       aid.
     *
     * Compliance:      USB Spec 1.0 - 11.12.2.3
     *
     * Parameters:  
     *
     *  ts_usb_hub*     Pointer to the hub structure. If NULL, -EINVAL
     *                  will be returned.  
     *
     *  __u8            Port number.
     *
     * Returns:
     *
     * __s8             > 0 if ok, encoding the bus state (and with
     *                  USB_HUB_BUS_STAT_* to get the bits). <0 errno
     *                  code on error. If the hub doesn't support this
     *                  operation, it will answer with an stall and
     *                  return. The code returned will be -EOPNOTSUPP.
     */

__s8 hubd_bus_status_get (ts_usb_hub* hub, __u8 port)
{
    int errcode;
    ts_usb_request req;
    __u8 status;
    
    gdlf_fntrace ("hubd_bus_status_get (0x%p, %u)", hub, status);
    gdlf_aassert (hub != NULL, return -EINVAL);
    gdlf_aassert (port < hub->ports, return -EINVAL);

    req.bmreqtype = 0xa3;
    req.request = e_usb_req_get_status;
    req.value = 0;
    req.index = port+1;
    req.length = sizeof (status);
    
    read_lock (&hub->device->msnode.lock);
    errcode = usb_request_data (hub->device, &req, &status, 
                                "get bus status");
    if (errcode < 0)
        gdlf_err ("Cannot get bus status; request failed: "
                  "errno/usb %d/%d\n", errcode, usb_errno);
    read_unlock (&hub->device->msnode.lock);
    return errcode < 0? errcode : status;    
}


    /* Initializes a hub control structure from a Hub Descriptor
     *
     *  It will read in the descriptor, parse it, and fill in the
     * structure, allocating space for it as needed. The structure
     * will hold data which must be freed with ts_usb_hub_free() before
     * usb_kfree()ing the structure.
     *
     * Parameters:  
     *
     *  ts_usb_hub*     Pointer to hub structure to fill in.
     *
     * Returns:
     *
     *  int             0 if ok, errno code < 0 on error.
     */

int hubd_descriptor_fill (ts_usb_hub* hub)
{
    int errcode;
    ts_usb_hub_desc *desc;
    ts_usb_desc_head head;
    
    gdlf_fntrace ("ts_usb_hub_init_desc (0x%p)", hub);
    gdlf_aassert (hub != NULL, return -EINVAL);

        /* get the header, enough to get the full length */

    errcode = hubd_descriptor_get (hub, 0, &head, sizeof (head));
    if (errcode < 0)
    {
        gdlf_err ("Error reading "
                  "hub 0x%p (%u %u)'s descriptor head: "
                  "errno/usb %d/%d\n",
                  hub, hub->host->addr, hub->device->addr,
                  errcode, usb_errno);
        return errcode;
    }

        /* check for consistency on descriptor */

    if (head.bDescriptorType != e_usb_desc_hub)
    {
        gdlf_err ("Incorrect descriptor passed as hub descriptor\n");
        return -ENODATA;    /* FIXME: errno code? */
    }

        /* alloc mem for the descriptor */

    desc = usb_kmalloc (head.bLength * sizeof (__u8), 
                        GFP_KERNEL);
    if (desc == NULL)
    {
        gdlf_err ("Failed to allocated %d bytes to store "
                  "hub 0x%p (%u %u)'s descriptor\n",
                  head.bLength * sizeof (__u8), hub, 
                  hub->host->addr, hub->device->addr);
        return -ENOMEM;
    }

        /* get the descriptor */
    
    errcode = hubd_descriptor_get (hub, 0, desc, 
                                   head.bLength * sizeof (__u8));
    if (errcode < 0)
    {
        gdlf_err ("Error reading "
                  "hub 0x%p (%u %u)'s descriptor: "
                  "errno/usb %d/%d\n",
                  hub, hub->host->addr, hub->device->addr,
                  errcode, usb_errno);
        usb_kfree (desc);
        return errcode;
    }

        /* Ok, fill in the data */

    hub->ports = desc->bNbrPorts;
    hub->port = ts_usb_port_a_init (hub->ports);
    if (hub->port == NULL)
    {
        ts_usb_hub_release (hub);
        usb_kfree (desc);
        return -ENOMEM;
    }        
    hub->pwrontime = desc->bPwrOn2PwrGood;
    hub->maxcurrent = desc->bHubContrCurrent;

    hub->power = (le16_to_cpu (desc->wHubCharacteristics)
                  | USB_HUB_CRTS_PWR_CTL_MASK);
    hub->compound = (le16_to_cpu (desc->wHubCharacteristics)
                     | USB_HUB_CRTS_COMPD_MASK)? 1 : 0;
    hub->overc_prot = (le16_to_cpu (desc->wHubCharacteristics)
                       | USB_HUB_CRTS_CURRENT_MASK) 
                           >> USB_HUB_CRTS_CURRENT_MASK_S;

        /* Initialize the ports' data */
    
#warning FIXME: test this block, guess it wont work the first time 
    
    {
        int             byte;           /* byte counter */
        int             port, top;      /* port counter, top */
        int             bit;            /* bit counter */
        unsigned char   bits1, bits2;   /* Data for each byte of the bit arrays */

        for (port = byte = 0, top = hub->ports; port < top; byte++)
        {
            bits1 = desc->PortData[byte];          /* Device removable array */
            bits2 = desc->PortData[byte + top];    /* Power Ctl Mask array */
            for (bit = 0; 
                 bit < sizeof (char) && port < top;
                 port++, bit++, bits1 >>= 1, bits2 >>= 1)
            {
                if (byte == 0 && port == 0)  /* This one is reserved */
                    continue;
                hub->port[port].removable = bits1 & 0x1;
                hub->port[port].pwrctl = bits2 & 0x1;
            }
        }
    }
    
    usb_kfree (desc);
    return 0;
}


    /* Schedule a read from the hub's status change endpoint
     *
     *  If error re-scheduling, we take the most radical option, to
     * reset the hub directly. However, FIXME, this is quite
     * radical. Should think of a possible better way.
     *
     *  Note this function _will_ complain itself, so the caller has
     * to do nothing in case of error but raise it on the call chain
     * [if needed].
     *
     * Parameters:  
     *
     *  ts_hubd*        Pointer to the hub structure we want to
     *                  schedule with.
     *
     * Returns:
     *
     *  int             0 if ok, errno code < 0 on error.
     *
     * Global:
     *
     *  int usb_errno   On error, error code.  
     *
     *  int hubd_errno  On error, error code.  
     */

int hubd_schedule_status_change (ts_usb_hub* hub)
{
    int errcode;

    gdlf_fntrace ("hubd_schedule_status_change (0x%p)", hub);

    errcode = usb_tr_run (&hub->statch_tr);
    if (errcode < 0)
    {
        gdlf_err ("Error runing status change read for "
                  "hub 0x%p (%d %d): errno/usb_errno %d/%d\n",
                  hub, hub->host->addr, hub->device->addr,
                  hub->statch_tr.errno, hub->statch_tr.usb_errno);

#warning FIXME: do some more retries, there may be some timing issues related to hubs

            /*
             * FIXME: see msg ... for the timing issues stuff
             *
             *        From: Reid Augustin <reid@NetChip.com>
             *        To: inaky@peloncho.fis.ucm.es
             *        Subject: Re: [linux-usb] Self powered hub behaviours on local power loose?
             *        Date: Tue, 07 Apr 1998 12:32:37 -0700 (PDT)
             *
             *  We can only reset if we aren't configuring the hub, as
             * we would  start a reentrancy. If configuring [aka,
             * called from hubd_config()] and fails, the configuration
             * process will fail, and will be more or less the same.
             */

        if (!hub->configuring)
            usb_device_reset (hub->device);
        return errcode;
    }
    return 0;
}


/* Log:
 *
 * $Log: requests.c,v $
 * Revision 1.15  1998/11/14 17:20:36  inaky
 * Moved the include files to uusbd/include/uusbd, to ease integration
 * into the Linux kernel.
 *
 * Revision 1.14  1998/11/09 21:36:31  inaky
 * Fixed getting the hub descriptor, as it was trying to get too much
 * data and it was stalling the xfer. Now it gets the header and then the
 * exact data.
 *
 * Moved hubd_schedule_status_change() to requests.c
 *
 * Revision 1.13  1998/09/24 16:09:07  inaky
 * Moved 'hub_port' to 'port'.
 *
 * Revision 1.12  1998/09/14 20:30:25  inaky
 * Severed assertions, removed the usb_hub_errno variable.
 *
 * Revision 1.11  1998/08/27 22:19:10  inaky
 * Moved _free to _release.
 *
 * Revision 1.10  1998/08/24 02:29:23  inaky
 * Changed name of all enumerations to be prefixed with the enumeration
 * type name.
 *
 * Revision 1.9  1998/07/23 14:39:16  inaky
 * Added the length parameter to the request structure.
 *
 * Revision 1.8  1998/07/16 14:15:17  inaky
 * Moved from the old read/write interface to the new, request-based
 * interface.
 *
 * Revision 1.7  1998/07/14 15:44:05  inaky
 * Moved the status bit update code from the hub driver to the hub
 * manager wrapping functions.
 *
 * Revision 1.6  1998/06/20 21:55:17  inaky
 * Massive changes to split the old monolithic hubd driver into a hub
 * manager (hub wrappers and daemon) and a dynamic driver, which can be
 * replaced and its different parts reused.
 *
 * Revision 1.5  1998/05/01 19:18:36  inaky
 * Fixed device address reporting on messages.
 *
 * Revision 1.4  1998/04/13 22:18:40  inaky
 * Finished most of the hub driver; some updates still needed.
 *
 * Revision 1.3  1998/01/08 12:19:19  inaky
 * Added copyright strings to all important files
 *
 * Revision 1.2  1997/12/29 00:26:16  inaky
 * Moving the khubdd to use signals as commands.
 *
 * Revision 1.1  1997/12/13 20:56:38  inaky
 * Moving all the documentation to the .c files except the ones for data
 * structures and inline functions. Helps making the header files cleaner.
 *
 *
 */

