
/* USB Driver Stack v$Revision: 1.13 $ 
 *
 * (C) 1997 by Iaky Prez Gonzlez <inaky@peloncho.fis.ucm.es>
 *
 * $Id: hubm-generic.c,v 1.13 1998/11/14 17:12:39 inaky Exp $
 *
 * Distributed under the terms of the GNU Public License Version 2
 * See files COPYRIGHT and COPYING at the root distribution directory
 *
 * Changelog at the bottom 
 *
 *
 *     Hub manager generic code
 *     ------------------------
 *
 */

#include <uusbd/usb.h>
#include <uusbd/usbd.h>
#include <uusbd/usb-stddesc.h>
#include <uusbd/usb-hub.h>
#include "hubm.h"

#define GDLF_IDENT          "USB-HUBM"
#define GDLF_LOCAL          /* Define if want local settings */
#define GDLF_DEF_LOGLEVEL   gdlf_level_top
#define GDLF_ASSERTIONS     /* want assertions (#undef if not) */
#undef GDLF_FNTRACE        /* want function traces */
#define GDLF_LOCATIONS      /* want locations in log messages */
#include <uusbd/gdlf.h>


    /* Allocates and initializes to a known state a ts_usb_hub*
     * structure.
     *
     * Returns:
     *  ts_usb_hub* Pointer to newly allocated hub structure. Must be
     *              freed with usb_kfree().
     */

ts_usb_hub* ts_usb_hub_init (void)
{
    ts_usb_hub* class;
    
    gdlf_fntrace ("ts_usb_hub_init()");
    class = usb_kmalloc (sizeof (ts_usb_hub), GFP_KERNEL);
    if (class == NULL)
    {
        gdlf_err ("Cannot allocate space for a "
                  "ts_usb_hub structure\n");
        return NULL;
    }
    ts_usb_hub_clean (class);
    return class;              
}


    /* Free a ts_usb_hub structure's contents. Checks for the magic
     * number being ok. Remember that the structure iself is not
     * freed, just it's contents.
     *
     * Parameters:  
     *
     *  ts_usb_hub*        Pointer to the hub structure to free.
     *
     * Returns:
     *
     *  int             0 if ok, errno code < 0 on error.
     *
     */

void ts_usb_hub_release (ts_usb_hub* hub)
{
    gdlf_fntrace ("ts_usb_hub_release (0x%p)", hub);
    gdlf_aassert (hub != NULL, return);

    if (gdlf_c0assert (hub->snode.dtor.fn == (tf_dtor)usb_hub_destroy))
    {
        gdlf_assert (hub->host == NULL);
        gdlf_assert (hub->device == NULL);
    }

    if (hub->port)
    {
        ts_usb_port_a_release (hub->port, hub->ports);
        usb_kfree (hub->port);
    }

    if (hub->statch_tr.data)
        usb_kfree (hub->statch_tr.data);

    ts_usb_open_release (&hub->statch);
    ts_usb_tr_release (&hub->statch_tr);

    return;
}


    /*  Creates and initializes a port structure.
     *
     * Returns:
     *
     *  ts_usb_port* Pointer to port structure.
     *
     * Global:
     *
     *  int usb_errno   On error, error code.  
     */

ts_usb_port* ts_usb_port_init (void)
{
    ts_usb_port* port;
    
    gdlf_fntrace ("ts_usb_port_init()");

    port = usb_kmalloc (sizeof (ts_usb_port), GFP_KERNEL);
    if (port == NULL)
    {
        usb_errno = -USBE_NO_MEMORY;
        gdlf_err ("No memory to create a port"
                  "structure\n");
        return NULL;
    }    

    ts_usb_port_clean (port);
    return port;
}

void ts_usb_port_release (ts_usb_port* port)
{
    gdlf_fntrace ("ts_usb_port_release (0x%p)", port);
    gdlf_aassert (port != NULL, return);

    if (port->device)
        usb_device_unref (port->device, NULL);
    
    ts_usb_port_clean (port);
    return;
}


ts_usb_port* ts_usb_port_a_init (size_t size)
{
    ts_usb_port *top, *ports;
    
    gdlf_fntrace ("ts_usb_port_a_init (%u)", size);
    gdlf_aassert (size != 0, return NULL);
    
    ports = usb_kmalloc (sizeof (ts_usb_port) * size, GFP_KERNEL);
    if (ports == NULL)
    {
        usb_errno = -USBE_NO_MEMORY;
        gdlf_err ("No memory to create a ts_usb_port array\n");
        return NULL;
    }
   
    for (top = ports + size - 1; top >= ports; top--)
    {
            /* Need braces, this is a macro! */
        ts_usb_port_clean (top);
    }
    return ports;
}


void ts_usb_port_a_release (ts_usb_port* ports, size_t size)
{
    ts_usb_port *top;
    
    gdlf_fntrace ("ts_usb_port_a_release (0x%p, %u)", ports, size);
    gdlf_aassert (ports != NULL, return);
    
    for (top = ports + size; ports < top; ports++)
        ts_usb_port_release (ports);

    return;
}


/* Log:
 *
 * $Log: hubm-generic.c,v $
 * Revision 1.13  1998/11/14 17:12:39  inaky
 * Moved the include files to uusbd/include/uusbd, to ease integration
 * into the Linux kernel.
 *
 * Revision 1.12  1998/11/09 21:21:53  inaky
 * Added assertions in function of the degree of initialization of an
 * structure. Fixed the hub_release function to release the open and tr
 * structures. Moved all device ref/unref to usb_device_[un]ref(). Moved
 * all array size parameters to size_t.
 *
 * Revision 1.11  1998/09/24 15:50:05  inaky
 * Moved 'hub_port' to 'port'.
 *
 * Revision 1.10  1998/09/14 02:28:02  inaky
 * Removed the usb_hub_errno variable.
 *
 * Revision 1.9  1998/08/27 22:13:24  inaky
 * Removed traces.
 *
 * Revision 1.8  1998/08/27 10:48:25  inaky
 * Moved ts_*_free() to ts_*_release(), made them return void and severed
 * assertions.
 *
 * Revision 1.7  1998/08/24 02:14:08  inaky
 * Removed debug messages.
 *
 * Revision 1.6  1998/08/19 23:33:02  inaky
 * Added debug messages to track memory corruption problems. Still here,
 * just in case.
 *
 * Revision 1.5  1998/08/03 13:22:04  inaky
 * Removed debug messages at port clean.
 *
 * Revision 1.4  1998/07/16 15:59:36  inaky
 * Fixed the mess with transactions on device-less hubs.
 *
 * Revision 1.3  1998/07/14 21:43:49  inaky
 * Removed spurios debug info
 *
 * Revision 1.2  1998/07/12 22:39:53  inaky
 * When freeing the array of ports, we got to _free_ it :) Some of the
 * debug code _should_ go away someday.
 *
 * Revision 1.1  1998/06/20 21:54:48  inaky
 * Massive changes to split the old monolithic hubd driver into a hub
 * manager (hub wrappers and daemon) and a dynamic driver, which can be
 * replaced and its different parts reused.
 *
 * Revision 1.15  1998/04/13 22:18:39  inaky
 * Finished most of the hub driver; some updates still needed.
 *
 * Revision 1.14  1998/02/10 17:41:07  inaky
 * Fixed to use ts_[S]snode_[un]ref()
 *
 * Revision 1.13  1998/02/09 13:42:56  inaky
 * Moved all use counts to atomic
 *
 * Revision 1.12  1998/01/08 12:19:18  inaky
 * Added copyright strings to all important files
 *
 * Revision 1.11  1997/12/31 05:37:40  inaky
 * Fixed some craps.
 *
 * Revision 1.10  1997/12/30 22:30:29  inaky
 * Fixed some bugs.
 *
 * Revision 1.9  1997/12/30 20:14:58  inaky
 * Moved to the new lists code. Untested yet.
 *
 * Revision 1.8  1997/12/29 00:26:14  inaky
 * Moving the khubdd to use signals as commands.
 *
 * Revision 1.7  1997/12/27 17:25:42  inaky
 * Finished moving to GDLF. Deprecated usb_printk and __LOCATION__
 *
 * Revision 1.6  1997/12/27 02:06:13  inaky
 * Moving everything to use the GDLF for logging and debug checkings.
 * Removed the global array of USB hosts and moved it to a linked
 * list. Added an address field to each USB host controller.
 * Removed lots of cruft that were accumulated and useless. Still some
 * missing, but may be useful someday.
 *
 * Revision 1.5  1997/12/13 20:56:37  inaky
 * Moving all the documentation to the .c files except the ones for data
 * structures and inline functions. Helps making the header files cleaner.
 *
 * Revision 1.4  1997/12/08 02:41:53  inaky
 * Moving the hubd to the new driver style
 *
 * Revision 1.3  1997/12/06 00:19:51  inaky
 * Almost finished the usbd/host, usbd/cfg and a lot of meat from
 * usbd/device. Also added the driver definition. Got to polish it yet
 * ...
 *
 * Revision 1.2  1997/11/29 20:21:58  inaky
 * Documented and almost fixes the read/write/abort API, documented a
 * damn lot and implemented some stuff
 *
 * Revision 1.1  1997/10/31 05:44:31  inaky
 * *** empty log message ***
 *
 */

