
/* USB Driver Stack v$Revision: 1.23 $ 
 *
 * (C) 1997 by Iaky Prez Gonzlez <inaky@peloncho.fis.ucm.es>
 *
 * Distributed under the terms of the GNU Public License Version 2
 * See files COPYRIGHT and COPYING at the root distribution directory
 *
 * $Id: hubm-hkeep.c,v 1.23 1998/11/14 17:12:53 inaky Exp $
 *
 * Changelog at the bottom 
 *
 *
 *     Hub Manager - Driver housekeeping
 *     --------------------------------
 *
 *  Driver definition and basic code for the Hubs.
 */

#include <asm/spinlock.h>   /* spinlock_init() */

#include <uusbd/usb.h>
#include <uusbd/usb-hub.h>
#include <uusbd/usbd.h>
#include "hubm.h"

#define GDLF_IDENT          "USB-HUBM"
#define GDLF_LOCAL          /* Define if want local settings */
#define GDLF_DEF_LOGLEVEL   gdlf_level_top
#define GDLF_ASSERTIONS     /* want assertions (#undef if not) */
#undef GDLF_FNTRACE        /* want function traces */
#define GDLF_LOCATIONS      /* want locations in log messages */
#include <uusbd/gdlf.h>

ts_slist usb_hub_list;


    /* Detach all devices from a hub
     *
     * Compliance:      USB Spec 1.0 xxxx
     *
     * Parameters:  
     *
     *  type            descr
     *
     * Global:
     *
     *  type name       descr
     *
     * Returns:
     *
     *  type            descr
     *
     * Global:
     *
     *  type name       descr
     */

int usb_hub_detach_devices (ts_usb_hub* hub, __u32 flags)
{
    int errcode;
    int cnt;
    ts_usb_device *device;
    
    gdlf_fntrace ("usb_hub_detach_devices (0x%p)", hub);

    for (cnt = 0; cnt < hub->ports; cnt++)
    {
        device = hub->port[cnt].device;
        if (device == NULL)
            continue;

        errcode = usb_device_detach (device, flags);
        if (errcode < 0)
            gdlf_err ("Error detaching device 0x%p (%d %d) "
                      "from port %u of hub 0x%p (%d %d): "
                      "errno/usb %d/%d; "
                      "expect inconsistencies\n",
                      device,
                      device->hub->host->addr, 
                      device->addr, 
                      cnt, hub, 
                      hub->host->addr, 
                      hub->device? hub->device->addr : -1,
                      errcode, usb_errno);
        usb_port_reset (hub, cnt);
    }
    return 0;
}



    /* Initialize the hub manager
     *
     * Returns:
     *
     *  int             0 if ok, errno code < 0 on error.
     *
     * Global:
     *
     *  int usb_errno       On error, error code.  
     */

int usb_hub_init (void)
{
    int errcode;
    
    gdlf_fntrace ("usb_hub_init (void)");

        /* Initialize */

    ts_slist_clean (&usb_hub_list);
    spin_lock_init (&dqueue_lock);              /* lock init'on */
    memset ((void*)dqueue, 0,                   /* clean the queue */
            sizeof (struct s_dqueue) * USB_HUB_DQUEUE_SLOTS);

    errcode = usb_hub_startd();                 /* start daemon */
    if (errcode)
        return errcode;

    return 0;
}


    /* Release the hub manager
     *
     *  Free and clean all data structs managed by the hub manager.
     *
     * Returns:
     *
     *  int             0 if ok, errno code < 0 on error.
     */

int usb_hub_release (void)
{    
    ts_usb_hub *itr, *next;

    gdlf_fntrace ("usb_hub_release (void)");

    itr = (ts_usb_hub*) ts_slist_head (&usb_hub_list);
    while (itr)
    {
        next = (ts_usb_hub*) ts_slist_next (&itr->snode);
        usb_hub_destroy (itr);
        itr = next;
    }
    return 0;
}


    /* Shutdown the hub manager
     *
     * Returns:
     *
     *  int             0 if ok, errno code < 0 on error.
     */

int usb_hub_shutdown (void)
{
    gdlf_fntrace ("usb_hub_shutdown()");

        /* we'll go over errors ... */

    usb_hub_release();
    return usb_hub_stopd();    
}


    /* Unregister an hub
     *
     *  This function will unregister a hub structure from the system,
     * following the inverse steps than when registering (this means
     * remove the hub from the linked list of hubs), previosuly
     * detaching all the pending devices.
     *
     * Parameters:  
     *
     *  ts_usb_hub*     Class element to unregister
     *
     * Returns:
     *
     *  int             0 if ok, errno code < 0 on error.
     */

int usb_hub_unregister (ts_usb_hub* hub)
{
    gdlf_fntrace ("usb_hub_unregister (0x%p)", hub);

    gdlf_aassert (hub != NULL, return -EINVAL);
    gdlf_passert (atomic_read (&hub->snode.usecnt) != 0,
                  "ts_usb_hub 0x%p's reference count is zero!\n",
                  hub);

        /* Stop scheduling reads (if doing it) */

    if (hub->operations
        && hub->operations->schedule_stch != NULL
        && hub->statch_tr.open != NULL)
        usb_close (&hub->statch);

        /* Detach everything */

    usb_hub_detach_devices (hub, USB_CFG_HUB_DESTROY);

        /* Remove the hub from the hub's list */

    ts_slist_del (&usb_hub_list, &hub->snode);
    return 0;                   /* bye */
}


    /* Register a hub
     *
     *  This function expects a ts_usb_hub structure will all the fields
     * filled in but `dtor'. This implies having the statch member
     * open correctly, the statch_tr member set up ready for
     * transfers (it's up to the specific hub driver in which state
     * must the transaction be; normally you'd have it INSERTED and
     * the drivers schedule status change function would just EXECUTE
     * it and reset it down from ABORTED, DONE or ERROR to INSERTED
     * again to repeat the execution).
     *
     *  Note that filling the `device' pointer means to also increment
     * the use count `usecount' of the device structure pointed to by
     * that pointer. Also note that an structure with any parameter
     * unspecified will be considered invalid, and thus -EINVAL
     * returned.
     *
     *  If everything is valid, it then will:
     *
     * 1. Fill in the snode.dtor member with the usb_hub_unregister()
     *    function. 
     *
     * 2. append the hub structure to the list of hubs (and increment
     *    the hub use count),
     *
     * 3. start scheduling polls on the hub by calling
     *    usb_hub_schedule_status_chage(). If not supported (returns
     *    -ENOSYS), then the hub is taken to be polling [root hubs],
     *    and the responsibility for it is left to the HCD.
     *
     * Parameters:  
     *
     *  ts_usb_hub*     Class element to register
     *
     * Returns:
     *
     *  int             0 if ok, errno code < 0 on error.
     */

int usb_hub_register (ts_usb_hub* hub)
{
    int errcode;
    int cnt;
    ts_usb_hub *root;
    ts_usb_hub *itr;
    
    gdlf_fntrace ("usb_hub_register (0x%p)", hub);
    gdlf_aassert (hub != NULL, return -EINVAL);
    
        /* 
         *  First of all, we check the depth in the USB tree and beep
         * if greater then 5 (See USB Spec 1.0, 7.1.16). We must allow
         * a maximum of 5 hub levels, counting the root hub. Note we
         * start counting on zero, so five must not be reached.
         *
         *  Instead of keeping a depth count on each hub, each time we
         * plug a hub in, we walk back the list until we reach the
         * root hub, counting each hub. If we count five, then this
         * one is not allowed to be plugged, as we'd violate the USB
         * topology as described on USB Spec 1.0 7.1.16.
         *
         *  Note we `blindly' assign itr to itr->device->hub. We know
         * that at least, UHCI root hubs have device set as NULL. Ok,
         * we can do that, as if itr is the root hub, it'll be catched
         * before the assignment and it will not happen.
         */

#warning FIXME: cli() == MORTAL SIN!!! move to lock the host->lock.

    cli();                                  /* FIXME: hmm */
    root = &hub->host->root;                /* huuuh ... loooong :) */
    for (itr = hub, cnt = 0;                /* Walk the list back to the root */
         itr != root && cnt < 5;
         itr = itr->device->hub, cnt++);
    sti();
  
    if (cnt >= 5)                           /* out of luck, too many hubs */ 
    {
            /*
             *  No check for hub->device != NULL, as it is suppossed
             * to happen _only_ on root hubs, and they cannot raise
             * this error [unless mistake in the code :)].
             */
        gdlf_err ("Cannot add new hub to port "
                  "%u of hub 0x%p (%d %d): "
                  "no more than four hubs may "
                  "be chained (five counting root hub)\n",
                  hub->device->port, hub->device->hub,
                  hub->host->addr, 
                  hub->device->hub->device->addr);
        return -EMLINK;
    }    
  
        /* Set the destructor */

    write_lock (&hub->snode.lock);
    ts_dtor_fill (&hub->snode.dtor, (tf_dtor)usb_hub_destroy);

        /* add it to the list */

    errcode = ts_slist_prep (&usb_hub_list, &hub->snode);
    if (errcode)
    {
        gdlf_err ("Cannot register hub 0x%p: couldn't "
                  "add it to the hub's list, errno %d\n",
                  hub, errcode);
        write_unlock (&hub->snode.lock);
        return errcode;
    }
    write_unlock (&hub->snode.lock);

#warning FIXME: power on all ports [move them to disconnected]

        /* Start scheduling */

    errcode = usb_hub_schedule_status_change (hub); /* start it */
    if (errcode < 0 && errcode != -ENOSYS)          /* any error? */
    {
        gdlf_err ("Cannot schedule status change read "
                  "on hub 0x%p (%d %d): errno/usb %d/%d. "
                  "Registration failed\n",
                  hub, hub->host->addr, 
                  hub->device? hub->device->addr : -1,
                  errcode, usb_errno);
        write_lock (&hub->snode.lock);
        ts_slist_del (&usb_hub_list, &hub->snode);
        write_unlock (&hub->snode.lock);
        return errcode;
    }
    return 0;
}


    /* Register a device connected to one of this hub's ports to this
     * hub's structures. 
     *
     *  This function will be called by usb_device_attach(). It will
     * get the hub structure to which we want to attach the device,
     * the port number and the device structure. 
     *
     *  Then it will point the device->hub to the hub, increment the
     * hubs usecount and set device->port. It will also set the
     * hub->port[port]->device to device and increment the device use
     * count. Now the device has a pointer to the hub and port, and
     * the hub knows which device is plugged in that port.
     *
     *  Actually, it should not talk to the device, as the control
     * endpoint 0 may not initialized yet.
     *
     * Parameters:  
     *
     *  ts_usb_device*  Device structure to attach to the hub. Note
     *                  that the hub is taken from the hub field of
     *                  the ts_usb_device structure.
     *
     *  ts_usb_hub*     Hub structure into which we are registering.
     *
     *  __u8            Port number.
     *
     * Returns:
     *
     *  int             0 if ok, errno code < 0 on error.
     */

int usb_hub_register_device (ts_usb_hub* hub, ts_usb_device* device, 
                             __u8 port)
{
    gdlf_fntrace ("usb_hub_register_device (0x%p, 0x%p, %u)", 
                  device, hub, port);
    gdlf_aassert (device != NULL, return -EINVAL);
    gdlf_aassert (hub != NULL, return -EINVAL);
    gdlf_aassert (port < hub->ports, return -EINVAL);

    write_lock (&device->msnode.lock);
    write_lock (&hub->snode.lock);

    device->hub = hub;
    ts_snode_ref (&hub->snode);
    device->port = port;

    hub->port[port].device = device;
    usb_device_ref (device);

    write_unlock (&hub->snode.lock);
    write_unlock (&device->msnode.lock);
    return 0;
}


    /* Unregisters a device from a hub
     *
     *  This function undoes what usb_hub_register_device() does. It will
     * just unset all the data fields the previous one set, taking
     * care of resource counts and cleaning fields. Take into account
     * it does NO operation on the USB bus.
     *
     * Parameters:  
     *
     *  ts_usb_device*  Device structure to attach to the hub. Note
     *                  that the hub is taken from the hub field of
     *                  the ts_usb_device structure.
     *
     * Returns:
     *
     *  int             0 if ok, errno code < 0 on error.
     */

int usb_hub_unregister_device (ts_usb_device* device)
{
    ts_usb_hub* hub;
    
    gdlf_fntrace ("usb_hub_unregister_device (0x%p)", device);
    gdlf_aassert (device != NULL, return -EINVAL);

    hub = device->hub;
    if (hub == NULL)
    {
        gdlf_warning ("Cannot unregister device 0x%p, "
                      "it has no hub!\n",
                      device);
        return -EBADF; 
    }

    write_lock (&device->msnode.lock);
    write_lock (&hub->snode.lock);

    usb_device_unref (device, NULL);
    hub->port[device->port].device = NULL;    

    ts_snode_unref (&hub->snode, NULL);
    device->hub = NULL;
    device->port = 0;

    write_unlock (&hub->snode.lock);
    write_unlock (&device->msnode.lock);
    return 0;
}


    /* Destroy a hub
     *
     * Parameters:  
     *
     *  ts_usb_hub*     Class element to unregister
     *
     * Returns:
     *
     *  int             0 if ok, errno code < 0 on error.
     */

int usb_hub_destroy (ts_usb_hub* hub)
{
    int cnt;
    int errcode;
    ts_usb_device *device;
    ts_usb_cfel *cfel;
    
    gdlf_fntrace ("usb_hub_destroy (0x%p)", hub);
    gdlf_aassert (hub != NULL, return -EINVAL);

    ts_snode_ref (&hub->snode); /* evitate it's destruction */
    atomic_inc (&hub->in_progress);
    if (atomic_read (&hub->in_progress) > 1)
        gdlf_info ("A change is still being processed "
                   "in hub 0x%p (%d %d); destroying anyway\n",
                   hub, hub->host->addr, 
                   hub->device? hub->device->addr : -1);
    
        /* Power off all ports */

    if (hub->power <= 2)            /* can we turn ports' power off? */
    {
        for (cnt = 0; cnt < hub->ports; cnt++)
        {
            errcode = usb_port_feature_clear (hub, cnt, e_usb_port_fes_power);
            if (errcode < 0)    
                gdlf_err ("Error powering off port %u "
                          "of hub 0x%p, device (%d %d): "
                          "errno/usb %d/%d\n",
                          cnt, hub, hub->host->addr, 
                          hub->device? hub->device->addr : -1, 
                          errcode, usb_errno);
        }
    }

            /*
             * Leave the cfel unconfigured 
             *
             *  Note we'll only do it if there's a device and we have
             * to search which cfel it is [thus, using only the
             * configuration, interface and alternate setting info if
             * configured independently, what I'm not sure if it is
             * possible].
             *
             *  Setting cfel->data to NULL and kfree()ing should be
             * done by the caller.
             */
    
    device = hub->device;   /* NOTE THIS ONE!! */
    if (device)
    {
#if 0
        if (USB_DEVICE_AG_INDEP (device))
            cfel = &device->pcfg[hub->cfg]->iface[hub->iface]
                .aset[hub->aset].cfel;
        else
            cfel = &device->cfel;
        cfel->configured = 0;
#endif
        device->cfel.configured = 0;
        cfel = &device->pcfg[hub->cfg]->iface[hub->iface]
            .aset[hub->aset].cfel;
    }
    
        /* Unregister the hub structure */

    usb_hub_unregister (hub);                   /* rm from hub's list */

    if (hub->host)  /* when errors at the very initial stages ... */
    {               /* ... things ain't set up */
        usb_host_unref (hub->host, NULL);       /* no more refs to host */
        hub->host = NULL;
    }
    if (device)                                 /* no more refs to dev */
    {
        usb_configuration_unref (device->pcfg[hub->cfg]);
        usb_device_unref (device, NULL);/* (it may be root!) */
        hub->device = NULL;
    }

    if (hub->operations)
        usb_hub_free (hub);                     /* release drivers */
    ts_snode_unref_nd (&hub->snode);
    atomic_dec (&hub->in_progress);
        /* This is safe, we keep a reference from the structure */
    gdlf_assert_act (ts_snode_unref_nd (&hub->statch_tr.snode));
    ts_usb_hub_release (hub);
        /* a destructor must not kfree() it's struct! */
    return 0;
}

/* Log:
 *
 * $Log: hubm-hkeep.c,v $
 * Revision 1.23  1998/11/14 17:12:53  inaky
 * Moved the include files to uusbd/include/uusbd, to ease integration
 * into the Linux kernel.
 *
 * Revision 1.22  1998/11/09 21:23:30  inaky
 * Removed abortion of the pipe when destroying a host. Closing it does
 * the trick. Moved all device ref/unref to usb_device_[un]ref(). Changed
 * the name of the configuration pointer array to pcfg. Moved all host
 * ref/unref to usb_host_[un]ref().
 *
 * Revision 1.21  1998/10/14 21:37:19  inaky
 * Added a cast, perhaps unneeded.
 *
 * Revision 1.20  1998/09/24 15:50:18  inaky
 * Moved 'hub_port' to 'port'.
 *
 * Revision 1.19  1998/09/14 02:28:21  inaky
 * Added a comment on things which should be done.
 *
 * Revision 1.18  1998/09/02 21:10:15  inaky
 * Removed the usb_hub_errno variable and the polled hubs array (in
 * favour of per HCD polling).
 *
 * Revision 1.17  1998/08/30 13:43:15  inaky
 * Severed assertions. Added check of semaphore for hub modifications
 * (in_progress).
 *
 * Revision 1.16  1998/08/27 22:13:41  inaky
 * Removed traces. Severed assertions.
 *
 * Revision 1.15  1998/08/27 11:14:51  inaky
 * Moved ts_*_free() to ts_*_release(), made them return void and severed
 * assertions.
 *
 * Revision 1.14  1998/08/24 02:15:31  inaky
 * Changed name of all enumerations to be prefixed with the enumeration
 * type name.
 *
 * Revision 1.13  1998/08/19 23:42:20  inaky
 * Removed the debug code which prevented usb_hub_detach_devices() to
 * complete.
 *
 * Revision 1.12  1998/08/19 23:34:27  inaky
 * Removed the debug code to track the memory corruption problems. I can
 * only blame it to the failure to initialize the usb_hub_list, as it
 * doesn't appear any more, and the ts_slist_(prep|del) code seems
 * completely correct.
 *
 * Revision 1.11  1998/08/19 23:26:45  inaky
 * Added debug code to hunt the memory corruption stuff.
 *
 * Fixed usb_hub_init() to initialize the usb hubs list.
 *
 * Revision 1.10  1998/08/03 13:22:51  inaky
 * Severed assertions for checking arguments.
 *
 * Revision 1.9  1998/07/23 14:18:52  inaky
 * Added volatile to errno variables, renamed the ts_usb_trbuf struct to
 * ts_usb_tr.
 *
 * Revision 1.8  1998/07/16 15:59:36  inaky
 * Fixed the mess with transactions on device-less hubs.
 *
 * Revision 1.7  1998/07/16 14:04:12  inaky
 * Moved from the old read/write interface to the new, request-based
 * interface.
 *
 * Revision 1.6  1998/07/14 21:55:19  inaky
 * Removed spurious debug info.
 *
 * Revision 1.5  1998/07/12 22:42:54  inaky
 * Added a function to detach all devices from a hub (and reset their
 * ports). Fixed the depth detection code (to take care on root hubs
 * without device).
 *
 * Fixed hub destructor to keep object alive during destruction using
 * non-destructive unreferencing. Fixed the device unconfiguration code
 * on hubs.
 *
 * Revision 1.4  1998/07/10 12:30:23  inaky
 * Removed all the old root hub stuff and fixed for the new one.
 *
 * Revision 1.3  1998/07/10 10:37:17  inaky
 * Moved the initial scheduling of status change reads to
 * usb_hub_register(). This way the hub manager knows automagically how
 * to treat the hub [scheduled or polled] and we take that burden out of
 * the driver.
 *
 * Revision 1.2  1998/06/21 20:09:47  inaky
 * Fixed some doc and added the declaration of the usb_hub_list (which
 * was missing).
 *
 * Revision 1.1  1998/06/20 21:54:49  inaky
 * Massive changes to split the old monolithic hubd driver into a hub
 * manager (hub wrappers and daemon) and a dynamic driver, which can be
 * replaced and its different parts reused.
 *
 * Revision 1.20  1998/05/02 00:52:04  inaky
 * Added the khubdd initialization code to hubd_init() and shutdown to
 * hubd_shutdown()
 *
 * Revision 1.19  1998/05/01 19:15:38  inaky
 * Fixed device address reporting on error strings.
 *
 * Added power on of all ports to the configuration routine [as it's
 * normal that powered off ports don't work ... :)]
 *
 * Revision 1.18  1998/04/26 03:39:55  inaky
 * Added the spinlock initialization.
 *
 * Revision 1.17  1998/04/13 22:18:39  inaky
 * Finished most of the hub driver; some updates still needed.
 *
 * Revision 1.16  1998/02/12 14:39:54  inaky
 * Drivers config method should inc the module use count when entering
 * and decrement it upon return.
 *
 * Revision 1.15  1998/02/10 17:40:46  inaky
 * Moved to use ts_[N]snode_[un]ref() and clean the driver structure
 * pointer when releasing it.
 *
 * Revision 1.14  1998/02/09 13:46:49  inaky
 * Moved all use counts to atomic, added #defines for the vicsp codes,
 * fixed the releasing of the driver structure when there's an error on
 * registration, to evitate freeing it if it's already on the drivers
 * list. This should be checked. FIXME.
 *
 * Revision 1.13  1998/02/08 22:01:08  inaky
 * The hubd_init() function wasn't registering properly the driver, as
 * the class code wans't being set.
 *
 * Revision 1.12  1998/01/11 15:44:36  inaky
 * Added docs?
 *
 * Revision 1.11  1998/01/08 12:19:18  inaky
 * Added copyright strings to all important files
 *
 * Revision 1.10  1998/01/08 11:50:02  inaky
 * Adapted to the new configuration scheme
 *
 * Revision 1.9  1998/01/02 00:07:29  inaky
 * Moved to the new driver registration scheme
 *
 * Revision 1.8  1997/12/30 22:30:29  inaky
 * Fixed some bugs.
 *
 * Revision 1.7  1997/12/30 20:14:58  inaky
 * Moved to the new lists code. Untested yet.
 *
 * Revision 1.6  1997/12/29 03:23:44  inaky
 * Removed some stale labels
 *
 * Revision 1.5  1997/12/29 00:26:14  inaky
 * Moving the khubdd to use signals as commands.
 *
 * Revision 1.4  1997/12/27 16:44:31  inaky
 * Moving to GDLF and fixinf docs. Removing more cruft.
 *
 * Revision 1.3  1997/12/27 02:06:13  inaky
 * Moving everything to use the GDLF for logging and debug checkings.
 * Removed the global array of USB hosts and moved it to a linked
 * list. Added an address field to each USB host controller.
 * Removed lots of cruft that were accumulated and useless. Still some
 * missing, but may be useful someday.
 *
 * Revision 1.2  1997/12/13 20:56:37  inaky
 * Moving all the documentation to the .c files except the ones for data
 * structures and inline functions. Helps making the header files cleaner.
 *
 * Revision 1.1  1997/12/08 02:41:52  inaky
 * Moving the hubd to the new driver style
 *
 */

