
/* USB Driver Stack v$Revision: 1.16 $ 
 *
 * (C) 1997 by Iaky Prez Gonzlez <inaky@peloncho.fis.ucm.es>
 *
 * Distributed under the terms of the GNU Public License Version 2
 * See files COPYRIGHT and COPYING at the root distribution directory
 *
 * $Id: hubm-stch.c,v 1.16 1998/11/14 17:14:56 inaky Exp $
 *
 * Changelog at the bottom 
 *
 *
 *     Hub Manager - Response to status changes
 *     ----------------------------------------
 *
 *  Functions to deal with the response to status changes from
 * hubs. Note that per hub driver / per hub function hooks are
 * provided in order to fix quirks.
 * 
 */

#warning FIXME: add driver hooks 

#include <uusbd/usbd.h>
#include "hubm.h"           /* Internal hubm stuff */

#define GDLF_IDENT          "USB-HUBM"
#define GDLF_LOCAL          /* Define if want local settings */
#define GDLF_DEF_LOGLEVEL   9
#define GDLF_ASSERTIONS     /* want assertions (#undef if not) */
#undef GDLF_FNTRACE        /* want function traces */
#define GDLF_LOCATIONS      /* want locations in log messages */
#include <uusbd/gdlf.h>


    /* Handle a hub status change
     *
     *  This function will act in consequence of a hub status
     * change. It will obtain the new hub status, and based upon the
     * changed bits, act.
     *
     * Parameters:  
     *
     *  ts_hubd*        Hub to operate on.
     *
     * Returns:
     *
     *  int             0 if ok, errno code < 0 on error.
     *
     * Global:
     *
     *  int hubd_errno  On error, error code.  
     */

int usb_hub_status_change (ts_usb_hub* hub)
{
    int errcode;
    __u32 status;    
    
        /* Get new hub status */

    status = usb_hub_status_get (hub);
    if (status < 0)
    {
        gdlf_err ("There was an error trying to process a hub "
                  "status change on hub 0x%p, device (%d %d): "
                  "cannot get status\n",
                  hub, hub->host->addr, 
                  hub->device? hub->device->addr : -1);
            /* hubd_errno set by hubd_status_get() */
        return status;
    }
    
        /* Process the status change [if any] */

    errcode = 0;

    if (status & USB_HUB_STATUS_C_LOCAL_PWR)
        errcode = usb_hub_local_power_change (hub);

    if (status & USB_HUB_STATUS_C_OVER_CURRENT)
        errcode = usb_hub_overcurrent_change (hub);

        /* 
         * yeah, we loose results, but each function is suppossed to
         * to complain adequately.  
         */

    return errcode; 
}


    /* Process a hub over current status change
     *
     *  An overcurrent condition has raised/gone if the overcurrent
     * change bit is set. 
     *
     *  If we have an overcurrent here, it means a global overcurrent.
     * Power has been shut off to all ports, so according to the port
     * state diagram (USB 1.0 Spec, Figure 11-5, Pag 221), it is like
     * having the ports powered off (or disconnected if hub doesn't
     * allow power switching). We have to detach the devices, warn of
     * the fact and then retry to power them on. This will re-start
     * the connection sequence for the devices, as well as it's
     * configuration. If any device again causes an overcurrent, this
     * will keep on failing until the user notices and removes the
     * offending device from the hub. Perhaps we should add a delay
     * here, but as the condition won't raise again until the device
     * is configured and working, it should be enough.
     * 
     *  If 0, the overcurrent condition is gone. This should
     * happen as soon as the ports are powered off.     
     *
     * NOTE: here we relay on usb_hub_status_get() having actualized the
     *       hub status fields (`local_power' and `overcurrent').
     *
     * Parameters:  
     *
     *  ts_usb_hub*     Hub we are handling
     *
     * Returns:
     *
     *  int             0 if ok, errno code < 0 on error.
     */

int usb_hub_overcurrent_change (ts_usb_hub* hub)
{
    int cnt;                        /* port counter */
    int errcode;
    
    gdlf_fntrace ("usb_hub_overcurrent_change (0x%p)", hub);

    if (hub->overcurrent)
    {
        ts_usb_device *device;
        
            /* Ok, we have a global overcurrent */

        gdlf_emerg ("Hub 0x%p, device (%d %d) is "
                    "in overcurrent; it cannot deliver "
                    "power to all it's attached devices; "
                    "detaching them\n",
                    hub, hub->host->addr, 
                    hub->device? hub->device->addr : -1);
        
        for (cnt = 0; cnt < hub->ports; cnt++)
        {
            device = hub->port[cnt].device;

            if (device == NULL)
                continue;

            errcode = usb_device_detach (device, 
                                         USB_CFG_HUB_OVR_CUR);
            if (errcode < 0)
                gdlf_err ("Error detaching device 0x%p (%d %d) "
                          "(because of global overcurrent) from "
                          "port %u of hub 0x%p, device (%d %d): "
                          "errno/usb %d/%d; "
                          "expect inconsistencies\n",
                          device, device->hub->host->addr, 
                          device->addr, cnt, hub, 
                          hub->host->addr, 
                          hub->device? hub->device->addr : -1, 
                          errcode, usb_errno);
            usb_kfree (device);
            hub->port[cnt].try = 1;
        }
    }
    else
    {
            /* 
             * out of overcurrent condition
             *
             *  We should make here sure that all the ports
             * are powered on. We do it on all as it is suppossed it
             * was a global overcurrent condition, and this, power was
             * shut down to all ports (as indicated on USB Spec 1.0
             * Table 11-14, bit 1).
             */

        gdlf_info ("Hub 0x%p, device (%d %d) has gone out of "
                   "global overcurrent state\n",
                   hub, hub->host->addr, 
                   hub->device? hub->device->addr : -1);

        if (hub->power <= 2)    /* if we can turn power on ... */
        {            
            for (cnt = 0; cnt < hub->ports; cnt++)
            {
                errcode = usb_port_feature_set (hub, cnt,
                                                e_usb_port_fes_power);
                if (errcode < 0)
                    gdlf_err ("Error powering on port %u "
                              "of hub 0x%p, device (%d %d): "
                              "errno/usb %d/%d\n",
                              cnt, hub, hub->host->addr, 
                              hub->device? hub->device->addr : -1, 
                              errcode, usb_errno);
            }
            usb_waitms (2 * hub->pwrontime);
        }
    }    

        /* Acknowledge the status change */

    errcode = usb_hub_feature_clear (hub, e_usb_hub_fes_c_over_current);
    if (errcode < 0)
    {
        gdlf_err ("Error acknowledging hub's overcurrent status "
                  "change to hub 0x%p, device (%d %d): "
                  "errno/usb %d/%d\n", 
                  hub, hub->host->addr, 
                  hub->device? hub->device->addr : -1,
                  errcode, usb_errno);
        return errcode;
    }
    return 0;
}


    /* Handle a local power change on a hub
     *
     *  Here we are faced with a problem. When this function is
     * invoked is because the local power supply in a hub has
     * transitioned from the lost to the ok state [the good news] or
     * from the ok to the lost state [the bad news].
     *
     *  If the hub is not self powered [hub->device->active->selfpower
     * != 0], then this will not happen. This comprobation is
     * asserted, as it should not happen, though it'd be a hardware
     * problem ... perhaps it should be always done. FIXME.
     *
     *  Ok, then we have this self-powered hub, who delivers at least
     * 5 unit loads/port. If local power is lost, I supposse it closes
     * current to all ports, as I cannot find on the spec what happens
     * exactly, but it could depend on the hub, as some ports may be
     * drawing a current amount which is supplied by the bus. 
     *
     *  Thus, the approach I'm using is all devices are powered off
     * until I find something better [FIXME].
     *
     *  If it's only selfpowered, then we are lost ... to the
     * river. It should signal a disconnect on the port and the port
     * change functions would detach it.
     *
     * NOTE: here we relay on usb_hub_status_get() having actualized the
     *       hub status fields (`local_power' and `overcurrent').
     *
     * Parameters:  
     *
     *  ts_usb_hub*     Hub we are handling
     *
     * Returns:
     *
     *  int             0 if ok, errno code < 0 on error.
     */

int usb_hub_local_power_change (ts_usb_hub* hub)
{
    int cnt;                        /* port counter */
    int errcode;

    gdlf_fntrace ("usb_hub_local_power_change (0x%p)", hub);

    gdlf_apassert (hub->device? hub->device->active->selfpower : 1,
                   return -EBADF,
                   "Ein?? Local power change on non self-powered "
                   "hub 0x%p, device (%d %d)\n", 
                   hub, hub->host->addr, 
                   hub->device? hub->device->addr : -1);
    
#warning FIXME: brute force method, should refine

    if (hub->local_power) /* Lost local power? */
    {
            /* 
             *  All ports to `power off' state if supported, else to
             * `disabled' state. Also we detach the device plugged
             * into it, giving a reason.
             *
             *  Note than for gang-power swiching mode, for all ports
             * to be powered off, it will happen when we have
             * requested all ports to be powered off [after the for
             * loop] (USB Spec 1.0 11.12.2.2).
             */

        int feature;                    /* feature to clear */
        ts_usb_device* device;
        
        gdlf_emerg ("Hub 0x%p, device (%d %d) has "
                    "lost it's local power supply\n",
                    hub, hub->host->addr, 
                    hub->device? hub->device->addr : -1);

        if (hub->power <= 2)            /* can we turn ports' power off? */
            feature = e_usb_port_fes_power;
        else
            feature = e_usb_port_fes_enable;
        
        for (cnt = 0; cnt < hub->ports; cnt++)
        {
            device = hub->port[cnt].device;

            if (device == NULL)
                continue;

            errcode = usb_device_detach (device, 
                                         USB_CFG_HUB_PWR_DN);
            if (errcode < 0)
                gdlf_err ("Error detaching device 0x%p (%d %d) "
                          "(because of local power down) from "
                          "port %u of hub 0x%p, device (%d %d): "
                          "errno/usb %d/%d; "
                          "expect inconsistencies\n",
                          device, device->hub->host->addr, 
                          device->addr, cnt, hub, 
                          hub->host->addr, 
                          hub->device? hub->device->addr : -1, 
                          errcode, usb_errno);
            usb_kfree (device);
            hub->port[cnt].try = 1;

            errcode = usb_port_feature_clear (hub, cnt, feature);
            if (errcode < 0)
                gdlf_err ("Error %s on port %u "
                          "of hub 0x%p, device (%d %d): "
                          "errno/usb %d/%d\n",
                          hub->power <= 2? "turning power off " : "disabling",
                          cnt, hub, hub->host->addr, 
                          hub->device? hub->device->addr : -1, 
                          errcode, usb_errno);
        }
    }
    else                                /* power is ok */
    {
            /* 
             * Power on all the ports if supported, else leave them at
             * their current state [should be disconnected].
             */

        gdlf_alert ("Hub 0x%p, device (%d %d) has "
                    "recovered it's local power supply\n",
                    hub, hub->host->addr, 
                    hub->device? hub->device->addr : -1);

        if (hub->power <= 2)            /* can we turn ports' power on? */
        {
            for (cnt = 0; cnt < hub->ports; cnt++)
            {
                errcode = usb_port_feature_set (hub, cnt, 
                                                e_usb_port_fes_power);
                if (errcode < 0)
                    gdlf_err ("Error powering on port %u "
                              "of hub 0x%p, device (%d %d): "
                              "errno/usb %d/%d\n",
                              cnt, hub, hub->host->addr, 
                              hub->device? hub->device->addr : -1, 
                              errcode, usb_errno);
            }
            usb_waitms (2 * hub->pwrontime);
        }
    }

        /* Acknowledge the status change */

    errcode = usb_hub_feature_clear (hub, e_usb_hub_fes_c_local_pwr);
    if (errcode < 0)
    {
        gdlf_err ("Error acknowledging hub's local power status "
                  "change to hub 0x%p, device (%d %d): "
                  "errno/usb %d/%d\n", 
                  hub, hub->host->addr, 
                  hub->device? hub->device->addr : -1,
                  errcode, usb_errno);
        return errcode;
    }
    return 0;
}





    /* The overcurrent status has changed on a port
     *
     *  If the overcurrent bit is set, we have an overcurrent on the
     * port; it's power has been shut off. We have to detach the
     * device and acknowledge.
     *
     *  I don't know if a change from overcurrent to no overcurrent
     * will be reported, as once the port is powered off, the
     * overcurrent is over. Thus, when acknowledging, it should clear
     * the overcurrent condition and warn again (change) that we have
     * no overcurrent, then we'd turn the power back on. 
     *
     *  If this doesn't work ... we'll have to change the method.
     *
     * Parameters:  
     *
     *  ts_usb_hub*     Hub
     *
     *  __u8            Port number
     *
     * Returns:
     *
     *  int             0 if ok, errno code < 0 on error.
     */

int port_over_current_change (ts_usb_hub* hub, __u8 port)
{
    int errcode;

    gdlf_fntrace ("port_over_current_change (0x%p, %u)", 
                  hub, port);

    if (hub->port[port].overcurrent)    /* we've got overcurrent */
    {
            /* check there's a device, jic */

        if (hub->port[port].device == NULL)
            gdlf_info ("Ein?? Overcurrent reported at "
                       "port %u of hub 0x%p (%d %d) "
                       "and no device attached\n",
                       port, hub, 
                       hub->host->addr, 
                       hub->device? hub->device->addr : -1);
        else
        {
            ts_usb_device *device = hub->port[port].device;

            gdlf_emerg ("Device (%d %d) at port %u of "
                        "hub 0x%p (%d %d) is in "
                        "overcurrent; detaching it\n",
                        device->hub->host->addr,
                        device->addr, port, hub, 
                        hub->host->addr, 
                        hub->device? hub->device->addr : -1);      

                /* detach it */

            errcode = usb_device_detach (device, 
                                         USB_CFG_HUB_OVR_CUR);
            if (errcode < 0)
                gdlf_err ("Error detaching device 0x%p (%d %d) "
                          "(because of global overcurrent) from "
                          "port %u of hub 0x%p, device (%d %d): "
                          "errno/usb %d/%d\n",
                          device, device->hub->host->addr, 
                          device->addr, port, hub, 
                          hub->host->addr, 
                          hub->device? hub->device->addr : -1, 
                          errcode, usb_errno);
            usb_kfree (device);
            hub->port[port].try = 1;
        }        
    }
    else    /* out of overcurrent condition, power it on */
    {
        gdlf_info ("Port %u of hub 0x%p (%d %d) got out of "
                   "overcurrent; powering it on\n",
                   port, hub, hub->host->addr, 
                   hub->device? hub->device->addr : -1);      

        if (hub->power <= 2)    /* if we can turn power on ... */
        {            
            errcode = usb_port_feature_set (hub, port, 
                                            e_usb_port_fes_power);
            if (errcode < 0)
                gdlf_err ("Error powering on port %u "
                          "of hub 0x%p (%d %d): "
                          "errno/usb %d/%d\n",
                          port, hub, hub->host->addr, 
                          hub->device? hub->device->addr : -1, 
                          errcode, usb_errno);
            usb_waitms (2 * hub->pwrontime);
        }
    }

        /* Acknowledge the status change */

    errcode = usb_port_feature_clear (hub, port,   
                                      e_usb_port_fes_c_over_current);
    if (errcode < 0)
    {
        gdlf_err ("Error acknowledging port's overcurrent status "
                  "change to port %u of hub 0x%p (%d %d): "
                  "errno/usb %d/%d\n", 
                  port, hub, hub->host->addr, 
                  hub->device? hub->device->addr : -1,
                  errcode, usb_errno);
        return errcode;
    }
    return 0;
}


    /* There was a hardware caused change in the enable status
     *
     *  A change caused by hardware has put the hub out or into
     * enabled state. One of the reasons may be an overcurrent [and
     * this is a reason for checking this overcurrent before
     * overcurrent]. 
     *
     *  If we have moved to the disabled state, we detach the device
     * and reset it; this will not cause another status change, from
     * disabled to enabled, as it'd be software generated and this bit
     * only is for hardware detection. Anyhow, we'll keep some code to
     * attach it in the case [again, should not happen] it happens.
     *
     * Parameters:  
     *
     *  ts_usb_hub*     Hub
     *
     *  __u8            Port number
     *
     * Returns:
     *
     *  int             0 if ok, errno code < 0 on error.
     *
     * Global:
     *
     *  int usb_errno   On error, error code.  
     */

int port_enable_change (ts_usb_hub* hub, __u8 port)
{
    int errcode = 0;
    ts_usb_device *device;
    
    gdlf_fntrace ("port_enable_change (0x%p, %u)", hub, port);

        /*
         * NOTE: This code, initially, is almost the same as for the
         *       port_connection_change(), function, so keep'em
         *       synchronized, right?
         */

    if (hub->port[port].enable)
    {
            /* 
             * enabled port, let's attach the device 
             *
             * I know, this would be quite strange, but just in case.
             */

        gdlf_warning ("Ein? Hardware caused enablement "
                      "on port %u of hub 0x%p (%d %d)?\n", 
                      port, hub, hub->host->addr, 
                      hub->device? hub->device->addr : -1);

        device = ts_usb_device_init();
        if (device == NULL)
            gdlf_err ("Cannot attach device to port %u "
                      "of hub 0x%p (%d %d): no memory\n", 
                      port, hub, hub->host->addr, 
                      hub->device? hub->device->addr : -1);
        else
        {
            errcode = usb_device_attach (device, hub, port, 
                                         USB_CFG_ATTACH);
            if (errcode < 0)
                gdlf_err ("Cannot attach device to port %u "
                          "of hub 0x%p (%d %d): "
                          "errno/usb %d/%d\n", 
                          port, hub, hub->host->addr, 
                          hub->device? hub->device->addr : -1,
                          errcode, usb_errno);
            hub->port[port].try++;  /* inc try count */
            if (hub->port[port].try > 10)
                hub->port[port].try = 1;
        }
    }   
    else    /* disabled port, let's detach the device */
    {
        hub->port[port].try = 1;
        device = hub->port[port].device;
        if (device == NULL)
        {
            gdlf_info ("Port %u of hub 0x%p (%d %d) "
                       "changed to disabled state and "
                       "no device was attached\n", 
                       port, hub, hub->host->addr, 
                       hub->device? hub->device->addr : -1);
        }
        else
        {
            errcode = usb_device_detach (device, USB_CFG_DETACH);
            if (errcode < 0)
                gdlf_err ("Cannot detach device from port %u "
                          "of hub 0x%p (%d %d): "
                          "errno/usb %d/%d; "
                          "going on anyway, expect memory leaks\n", 
                          port, hub, hub->host->addr, 
                          hub->device? hub->device->addr : -1,
                          errcode, usb_errno);
            gdlf_passert (atomic_read (&device->msnode.usecnt) == 0,
                          "Device (%d %d) use count when "
                          "detaching is %u, should be zero; "
                          "expect memory leaks\n",
                          device->addr, hub->host->addr,
                          atomic_read (&device->msnode.usecnt));
            usb_kfree (device);
        }        
    }
    
        /* Acknowledge status change */

    errcode = usb_port_feature_clear (hub, port,   
                                      e_usb_port_fes_c_enable);
    if (errcode < 0)
    {
        gdlf_err ("Error acknowledging port's enable status "
                  "change to port %u of hub 0x%p (%d %d): "
                  "errno/usb %d/%d\n", 
                  port, hub, hub->host->addr, 
                  hub->device? hub->device->addr : -1,
                  errcode, usb_errno);
        return errcode;
    }
    return errcode;
}


    /* A change has undergone in the port's connection status
     *
     *  If the change is a connection, it means we are in the
     * disabled state, and thus we must reset the port. After that we
     * must attach the device. Note we must *reset* the port. That
     * implies we are going to sleep, and thus, schedule().
     *
     *  If what we have is a disconnection, then *it* should have
     * signaled first an enable->disable transition, as it is a
     * hardware change. I _don't_ know. FIXME. Anyhow, we'll take the
     * same steps as on disable, to detach the device.
     *
     * Parameters:  
     *
     *  ts_usb_hub*     Hub
     *
     *  __u8            Port number
     *
     * Returns:
     *
     *  int             0 if ok, errno code < 0 on error.
     *
     * Global:
     *
     *  int usb_errno   On error, error code.  
     */

int port_connection_change (ts_usb_hub* hub, __u8 port)
{
    int errcode;
    ts_usb_device *device;
    
        /*
         * NOTE: This code, initially, is almost the same as for the
         *       port_enable_change(), function, so keep'em
         *       synchronized, right?
         */

    gdlf_fntrace ("port_connection_change (0x%p, %u)", hub, port);

    if (hub->port[port].connect)
    {
            /* 
             * we've got it connected, let's reset and attach the
             * device.
             *
             * Note on `enable_change' we do not reset
             */

        gdlf_info ("Device attachment detected "
                   "on port %u of hub 0x%p (%d %d)\n", 
                   port, hub, hub->host->addr, 
                   hub->device? hub->device->addr : -1);

        while (1)               /* dirty trick */
        {
            usb_waitms (hub->port[port].try * 100); /* device pwr up complete (USB 1.0 11.6.3) */
            errcode = usb_port_reset (hub, port);   /* Reset the port */
            if (errcode < 0)
            {
                gdlf_err ("Cannot reset for attachment port %u "
                          "of hub 0x%p (%d %d): "
                          "errno/usb %d/%d\n", 
                          port, hub, hub->host->addr, 
                          hub->device? hub->device->addr : -1,
                          errcode, usb_errno);
                break;
            }

                /* Reset+enable went ok */

            device = ts_usb_device_init();
            if (device == NULL)
            {
                gdlf_err ("Cannot attach device to port %u "
                          "of hub 0x%p (%d %d): no memory\n", 
                          port, hub, hub->host->addr, 
                          hub->device? hub->device->addr : -1);
                    /* Disable the port! */
                usb_port_feature_clear (hub, port, 
                                        e_usb_port_fes_enable);
                break;
            }            
                /* Got a device structure, we can attach to it */

            errcode = usb_device_attach (device, hub, port, 
                                         USB_CFG_ATTACH);
            if (errcode < 0)
            {
                gdlf_err ("Cannot attach device to port %u "
                          "of hub 0x%p (%d %d): "
                          "errno/usb %d/%d\n", 
                          port, hub, hub->host->addr, 
                          hub->device? hub->device->addr : -1,
                          errcode, usb_errno);
                    /* Disable the port! */
                usb_port_feature_clear (hub, port, 
                                        e_usb_port_fes_enable);
                usb_kfree (device);
                hub->port[port].try++;  /* inc try count */
                if (hub->port[port].try > USB_CONNECT_TRIES)
                    hub->port[port].try = 1;
            }
            break;
        }
    }   
    else    /* disconnected port, let's detach the device */
    {
        hub->port[port].try = 1;  /* inc try count */
        device = hub->port[port].device;
        if (device == NULL)
        {
            gdlf_info ("Port %u of hub 0x%p (%d %d) "
                       "changed to disconnected state and "
                       "no device was attached\n", 
                       port, hub, hub->host->addr, 
                       hub->device? hub->device->addr : -1);
        }
        else
        {
            errcode = usb_device_detach (device, USB_CFG_DETACH);
            if (errcode < 0)
                gdlf_err ("Cannot detach device from port %u "
                          "of hub 0x%p (%d %d): "
                          "errno/usb %d/%d; "
                          "going on anyway, expect memory leaks\n", 
                          port, hub, hub->host->addr, 
                          hub->device? hub->device->addr : -1,
                          errcode, usb_errno);
            gdlf_passert (atomic_read (&device->msnode.usecnt) == 0,
                          "Device (%d %d) use count when "
                          "detaching is %u, should be zero; "
                          "expect memory leaks\n",
                          device->addr, hub->host->addr,
                          atomic_read (&device->msnode.usecnt));
            usb_kfree (device);
            hub->port[port].device = NULL;
        }        
    }
    
        /* Acknowledge status change */

    errcode = usb_port_feature_clear (hub, port,   
                                      e_usb_port_fes_c_connection);
    if (errcode < 0)
    {
        gdlf_err ("Error acknowledging port's connection status "
                  "change to port %u of hub 0x%p (%d %d): "
                  "errno/usb %d/%d\n", 
                  port, hub, hub->host->addr, 
                  hub->device? hub->device->addr : -1,
                  errcode, usb_errno);
        return errcode;
    }
    return errcode;
}


    /* The suspend process has finished, device has resumed
     *
     *  This function gets called when a device has got out
     * of the suspend state and has completely resumed (See USB Spec
     * 1.0 Table 11-17, bit 2 C_PORT_SUSPEND).
     *
     *  Currently this function does nothing but to warn of the
     * fact. It could be implemented something to make drivers
     * unload on device suspend and reload on resume. However, this is
     * quite difficult. FIXME.
     *
     * FIXME: APM interactions here? if this is a modem which gets a
     *        RING and wants to wake the system up, it should send a
     *        resume event somehow and somewhere. Devices send remote
     *        wake ups if they are allowed to, via the feature
     *        USB_FES_RMT_WAKE_UP. See USB Spec 1.0 9.4.5, 5th
     *        paragraph.
     *
     * Parameters:  
     *
     *  ts_usb_hub*     Hub
     *
     *  __u8            Port number
     *
     * Returns:
     *
     *  int             0 if ok, errno code < 0 on error.
     *
     * Global:
     *
     *  int usb_errno   On error, error code.  
     */

int port_resumed (ts_usb_hub* hub, __u8 port)
{
    int errcode;
    
    gdlf_fntrace ("port_resumed (0x%p, %u)", hub, port);
    
    gdlf_info ("Device 0x%p (%d %d), on port %u of "
               "hub 0x%p (%d %d) has resumed\n",
               hub->port[port].device, 
               hub->host->addr, 
               hub->port[port].device? hub->port[port].device->addr : -1, 
               port,
               hub, hub->host->addr, 
               hub->device? hub->device->addr : -1);

        /* Acknowledge the status change */

    errcode = usb_port_feature_clear (hub, port,   
                                      e_usb_port_fes_c_suspend);
    if (errcode < 0)
    {
        gdlf_err ("Error acknowledging port's resume status "
                  "change to port %u of hub 0x%p, "
                  "device (%d %d): errno/usb "
                  "%d/%d\n", 
                  port, hub, hub->host->addr, 
                  hub->device? hub->device->addr : -1, 
                  errcode, usb_errno);
        return errcode;
    }
        
    return 0;
}


    /* Process an status change on a given port
     *
     *  Well get the status change bitmap and process each changed
     * bit, in importance order.
     *
     * NOTE: we depend on usb_port_status_get() actualizing the
     *       status information for the given port.
     *
     * Parameters:  
     *
     *  ts_ubd*         Hub structure to process.
     *
     *  __u8            Port number
     *
     * Returns:
     *
     *  int             0 if ok, errno code < 0 on error.
     *
     * Global:
     *
     *  int usb_errno   On error, error code.  
     */

int usb_port_status_change (ts_usb_hub* hub, __u8 port)
{
    __u32 status;
    int errcode;
    
    gdlf_fntrace ("usb_port_status_change (0x%p, %u)", hub, port);

        /* Let's get port's status */

    status = usb_port_status_get (hub, port);
    if (status < 0)
    {
        gdlf_err ("There was an error trying to process a "
                  "port status change on port %u of "
                  "hub 0x%p, device (%d %d)\n",
                  port, hub, hub->host->addr, 
                  hub->device? hub->device->addr : -1);
        return errcode;
    }
#if 0
    gdlf_debug ("Got port status 0x%x\n", status);
#endif
    
        /* 
         *  Ok, now this ugly if sequence to determine which bit's
         * changed. Note the order, as there are some events more
         * important than others.
         *
         * FIXME: is this order right?
         */

    errcode = 0;
    
    if (status & USB_PORT_STAT_C_ENABLE)
        errcode = port_enable_change (hub, port);

    if (status & USB_PORT_STAT_C_OVER_CURRENT)
        errcode = port_over_current_change (hub, port);

        /* We ignore USB_PORT_STAT_C_RESET */

    if (status & USB_PORT_STAT_C_CONNECTION)
        errcode = port_connection_change (hub, port);

    if (status & USB_PORT_STAT_C_SUSPEND)
        errcode = port_resumed (hub, port);

        /*
         *  These are safety mesures, in case something went wrong, to
         * try to recover from it.
         */

        /* 
         *  Device connected, but no device registered on it? This
         * means the connect process failed. We probably want to keep
         * retrying, so we try again.
         *
         * FIXME: Add some sort of protecion against DOS, like a timer
         *        ...
         */

    if (status & USB_PORT_STAT_CONNECTION
        && hub->port[port].device == NULL)
        errcode = port_connection_change (hub, port);

        /* 
         * Yeah, we loose result codes, but each function is supposed
         * to complain adequately.
         */

    return errcode;
}


/* Log:
 *
 * $Log: hubm-stch.c,v $
 * Revision 1.16  1998/11/14 17:14:56  inaky
 * Moved the include files to uusbd/include/uusbd, to ease integration
 * into the Linux kernel.
 *
 * Revision 1.15  1998/10/18 15:10:24  inaky
 * Added the definition of maximum number of retries on a port to connect
 * a device. After that tries, the count wraps to zero. This affects the
 * times waited on the reset and after-reset stages.
 *
 * Revision 1.14  1998/10/14 21:37:40  inaky
 * Remove function traces.
 *
 * Revision 1.13  1998/09/24 15:50:45  inaky
 * Moved 'hub_port' to 'port'.
 *
 * Revision 1.12  1998/09/14 02:29:00  inaky
 * Fixed comment priorities, removed the reset change callback [now we
 * wait for reset to complete].
 *
 * Revision 1.11  1998/09/02 21:11:00  inaky
 * Removed the usb_hub_errno variable.
 *
 * Revision 1.10  1998/08/30 13:49:14  inaky
 * Added the support for the port attachment try count.
 *
 * Revision 1.9  1998/08/28 23:44:19  inaky
 * Added comment.
 *
 * Revision 1.8  1998/08/24 02:15:20  inaky
 * Changed name of all enumerations to be prefixed with the enumeration
 * type name.
 *
 * Changed category of disconnected-port-with-device message.
 *
 * Revision 1.7  1998/08/19 23:34:53  inaky
 * Added a dirty trick, not permanent.
 *
 * Revision 1.6  1998/08/03 13:23:32  inaky
 * Fixed the port_connection_change() function to disable ports if
 * the attachment of a device fails. This way hubs aren't confused.
 *
 * Revision 1.5  1998/07/23 14:19:24  inaky
 * When a failed device attach happens, we have to FREE the device
 * structure. usb_device_attach() won't do it for us.
 *
 * Revision 1.4  1998/07/14 21:45:15  inaky
 * Fixed some texts and added the first of the safety measures to
 * usb_port_status_change() to try to recover from some failure.
 *
 * Revision 1.3  1998/07/14 14:51:30  inaky
 * Fixed messages to take into account the case of root hubs without
 * devices.
 *
 * Revision 1.2  1998/07/01 22:01:48  inaky
 * *** empty log message ***
 *
 * Revision 1.1  1998/06/20 21:54:49  inaky
 * Massive changes to split the old monolithic hubd driver into a hub
 * manager (hub wrappers and daemon) and a dynamic driver, which can be
 * replaced and its different parts reused.
 *
 * Revision 1.2  1997/12/27 19:54:46  inaky
 * Removed .txt file who got in by mistake and added some stuff to the
 * templates.
 *
 * Revision 1.1  1997/10/02 21:29:17  inaky
 * Remade all the directory structure to make it cleaner when the USB
 * class drivers come.
 *
 * Revision 1.1  1997/02/22 22:25:44  inaky
 * *** empty log message ***
 *
 *
 */
