
/* USB Driver Stack v$Revision: 1.1 $ 
 *
 * (C) 1997 by Iaky Prez Gonzlez <inaky@peloncho.fis.ucm.es>
 *
 * Distributed under the terms of the GNU Public License Version 2
 * See files COPYRIGHT and COPYING at the root distribution directory
 *
 * $Id: usb-dtor.h,v 1.1 1998/11/14 00:38:26 inaky Exp $
 *
 * Changelog at the bottom 
 *
 *  Structure destructors
 *
 *  Define TS_DTOR_MAGIC to get magic number checks on ts_dtor
 * structures, who by default don't wear one. Define TS_DTOR_DEBUG to
 * get debug checks on all operations.
 */

#ifndef _usb_dtor_h_
#define _usb_dtor_h_


    /* 
     * Environment definition
     * ----------------------
     *
     *  This is to ease reusability, by defining an environment, we can
     * use this code to run at many places almost without tweaking.
     *
     *  Currently only defined for Linux Kernel and UUSBD usage.
     */

#ifdef __linux__
#  ifdef __KERNEL__
#    define TS_DTOR_EXPORT
#    ifdef __UUSBD__
#      include <uusbd/usb-mm.h>       /* usb_kmalloc(), usb_kfree(), GFP_KERNEL */
#      define ts_dtor_malloc(a) usb_kmalloc ((a), GFP_KERNEL)
#      define ts_dtor_mfree(a) usb_kfree ((a))
#    else
#      include <linux/mm.h>     /* kmalloc(), kfree(), GFP_KERNEL */
#      define ts_dtor_malloc(a) kmalloc ((a), GFP_KERNEL)
#      define ts_dtor_mfree(a) kfree ((a))
#    endif /* __UUSBD__ */
#  endif /* __KERNEL__ */
#else

#error DTOR: Undefined environment!

#endif


    /* Structure destructors
     * ---------------------
     * 
     *  This is used for structures which hold complex data and
     * thus are harder to free. The thing is to have one of the
     * data members who's a "destructor" (ala C++), who will take
     * care when executed, given the structure pointer, of freeing
     * the data associated to the struct, but not the storage for
     * the structure itself (thus, after a succesful call to the
     * destructor, you should kfree() the structure.
     *
     *  The idea is to make an structure for describing each
     * struct, containing a magic number (#ifdef'ef, to allow
     * removing it when it's known to work ok) and a destructor
     * pointer. Then every complex data structure should be an
     * union of this struct as well as it's own data. This should
     * be like this due to the fact that lot of pointers will be
     * void*, so it's ok to know we have, at least, a secure type
     * to relay on.
     *
     *  The destructor should first check it understands the magic
     * number and then proceed. If it doesn't understand the magic
     * number, should print an error message, critical level,
     * saying the fact and advising of possible corruption. Then
     * it should return a value < 0. If ok, it should free the
     * structure data and return 0. If any error when freeing, it
     * should return a value > 0. If NULL pointer it should return
     * -1.
     *
     *  Note that the function pointer is always first in the two
     * cases (with or without magic number). This is done so that when
     * you declare an struct, it should cointain a ts_[m]dtor as the
     * first element. Then, using the first pointer in the structure,
     * no matter if referenced via a void*, we can always access the
     * destructor function.
     */

typedef int (*tf_dtor)(void*);

typedef struct      /* Destructor without magic number */
{
    tf_dtor     fn; /* Destructor */
} ts_dtor;

typedef struct      /* Destructor with magic number */
{
    tf_dtor     fn;     /* Destructor */
    unsigned    magic;  /* Magic number */
} ts_mdtor;


    /* Execute the handler to free the data structure associated to
     * ptr. The handler is a tf_dtor type.
     *
     *  A tf_dtor (int (*) (tf_mdtor*, void*)), is passed the address
     * of the dtor object, which in turn will be the address of the
     * object to destroy, AS THE ts_[m]dtor OBJECT MUST BE THE FIRST
     * ONE DEFINED FOR ANY STRUCTURE USING THIS SCHEME. Thus, with
     * this single data, we can get the pointer to the structure whose
     * contents to free, the magic number (if any) or whatever is
     * needed. THE DESTRUCTOR MUST NOT FREE THE STORAGE FOR THE
     * STRUCTURE, JUST IT'S CONTENTS. It must return < 0 on any error,
     * being free to log any messages.
     *
     * NOTE: Don't want'em #defined as I want type checking
     *
     * Parameters:  
     *
     *  tf_dtor*    Pointer to the destructor object. WARNING! This
     *              function doesn't check for the pointer to be or
     *              not NULL, you must perform this check before
     *              calling it!
     *
     * Returns:
     *  
     *  int         < 0 on error, else success
     *
     * Global:
     *
     *  int usb_errno   On error, error code.  
     */

static inline 
int ts_dtor_exec (ts_dtor* _dtor)
{
    return _dtor->fn (_dtor);
}
    
static inline 
int ts_mdtor_exec (ts_mdtor* _dtor)
{
    return _dtor->fn (_dtor);
}    


    /* [ts_[m]dtor_fill] Fill a ts_[m]dtor structure
     *
     *  Given a pointer to a desctructor function (tf_dtor = int
     * (*)(void*)) and a magic number (give it always, as may be or
     * not be used; if not, will be discarded), initialize a
     * ts_[m]dtor structure.
     *
     * Parameters:  
     *
     *  ts_[m]dtor*     Pointer to destructor structure to fill.
     *
     *  int(*)(void*)   Pointer to destructor function.
     *
     *  unsigned int    [ts_mdtor_fill] Magic number
     *
     *
     * [ts_[m]dtor_clean] Initialize a ts_[m]dtor structure.
     *
     *  Given a pointer to a ts_[m]dtor structure, cleans it 
     * uninitalized values (NULL destructor and zero magic number).
     *
     * Parameters:  
     *
     *  ts_[m]dtor*     Pointer to destructor structure to clean.
     */

#define ts_dtor_fill(_dtor, _fn)                \
do                                              \
{                                               \
    (_dtor)->fn = _fn;                          \
} while (0)

#define ts_dtor_nozero(_dtor)                   \
do                                              \
{                                               \
} while (0);

#define ts_dtor_clean(_dtor)                    \
do                                              \
{                                               \
    memset (_dtor, 0, sizeof (*(_dtor)));       \
    ts_dtor_nozero (_dtor);                     \
} while (0)

#define ts_mdtor_fill(_dtor, _fn, _magic)       \
do                                              \
{                                               \
    (_dtor)->magic = _magic;                    \
    (_dtor)->fn = _fn;                          \
} while (0)

#define ts_mdtor_nozero(_dtor)                  \
do                                              \
{                                               \
} while (0);

#define ts_mdtor_clean(_dtor)   \
do                                              \
{                                               \
    memset (_dtor, 0, sizeof (*(_dtor)));       \
    ts_mdtor_nozero (_dtor);                    \
} while (0)


    /* Allocate and clean a ts_[m]dtor structure
     *
     * Returns:
     *
     *  ts_[m]dtor*     Pointer to the new ts_[m]dtor allocated and
     *                  initialized, NULL on allocation error.
     */

#if 0
static inline 
ts_dtor* ts_dtor_init (void)
{
    ts_dtor* dtor;

    dtor = ts_dtor_malloc (sizeof (ts_dtor));

    if (dtor == NULL)
    {
        gdlf_err ("No memory to allocate a ts_dtor structure\n");
        return NULL;
    }
    ts_dtor_clean (dtor);
    return dtor;
}

static inline 
ts_mdtor* ts_mdtor_init (void)
{
    ts_mdtor* dtor;

    dtor = ts_dtor_malloc (sizeof (ts_mdtor));

    if (dtor == NULL)
    {
        gdlf_err ("No memory to allocate a ts_mdtor structure\n");
        return NULL;
    }
    ts_mdtor_clean (dtor);
    return dtor;
}
#endif


    /* Free an allocated ts_[m]dtor structure
     *
     *  Free's it's contents, not the structure itself.
     *
     * Parameters:
     *
     *  ts_[m]dtor*     Pointer to the ts_[m]dtor to free. It must
     *                  have been allocated previously.
     *
     * Returns:
     *
     *  int             0 if ok, errno code < 0 on error.
     */

#if 0
static inline 
int ts_dtor_release (ts_dtor* _dtor)
{
    gdlf_apassert (_dtor != NULL,
                   return -EINVAL;,
                   "Tried to free a NULL ts_dtor\n");

    ts_dtor_clean (_dtor);
    return 0;
}

static inline 
int ts_mdtor_release (ts_mdtor* _dtor)
{
    gdlf_apassert (_dtor != NULL,
                   return -EINVAL;,
                   "Tried to free a NULL ts_mdtor\n");

    ts_mdtor_clean (_dtor);
    return 0;
}
#endif

#include <uusbd/gdlf-undef.h> /* clean */

#endif /* _usb_dtor_h_ */

/* Log:
 *
 * $Log: usb-dtor.h,v $
 * Revision 1.1  1998/11/14 00:38:26  inaky
 * Isolated all the header files
 *
 * Revision 1.10  1998/09/24 15:51:41  inaky
 * Finished the split of _clean into _nozero and _clean.
 *
 * Revision 1.9  1998/08/27 22:14:19  inaky
 * Moved _free to _release.
 *
 * Revision 1.8  1998/08/24 02:20:13  inaky
 * Removed GDLF dependency and unused functions.
 *
 * Revision 1.7  1998/01/08 12:19:08  inaky
 * Added copyright strings to all important files
 *
 * Revision 1.6  1998/01/02 00:12:16  inaky
 * Finished the new driver scheme. Documented at usbd-drv.c
 *
 * Revision 1.5  1997/12/30 20:15:43  inaky
 * Moved everything to the new lists scheme. Untested yet.
 *
 * Revision 1.4  1997/12/30 00:03:45  inaky
 * Adding a generic system to handle singly linked lists and the HCI
 * registration control system
 *
 * Revision 1.3  1997/12/29 23:55:12  inaky
 * Removed the posibility of using a magic number into ts_dtor, as it was
 * useless.
 *
 * Revision 1.2  1997/12/27 20:06:27  inaky
 * Finished fixinf the __KERNEL__ #ifdef wrapping and moved the ts_dtor.h
 * to usb-dtor.h
 *
 * Revision 1.1  1997/12/27 16:44:24  inaky
 * Moving to GDLF and fixinf docs. Removing more cruft.
 *
 */

