
/* USB Human Interface Devices Driver Boot Protocol v$Revision: 1.1 $ 
 *
 * (C) 1997 by Iaky Prez Gonzlez <inaky@peloncho.fis.ucm.es>
 *
 * Distributed under the terms of the GNU Public License Version 2
 * See files COPYRIGHT and COPYING at the root distribution directory
 *
 * $Id: usb-hidbp.h,v 1.1 1998/11/14 00:38:26 inaky Exp $
 *
 * Changelog at the bottom 
 *
 *     USB Human Interface Devices - Boot Protocol
 *     -------------------------------------------
 */

#ifndef _usb_hidbp_h_
#define _usb_hidbp_h_

#include <linux/fs.h>       /* file_operations */
#include <linux/miscdevice.h>   /* struct miscdevice */

#include <uusbd/usbd.h>
#include <uusbd/usb-hidd.h>

    /* Protocol codes */

#define HIDBP_PROTOCOL_KBD  0x1 /* Boot keyboard */
#define HIDBP_PROTOCOL_MSE  0x2 /* Boot mouse */

    /* Allocating 16 minors for boot protocol keyboards and 16 for mice */

#define HIDBP_BKBD_MINOR 16         /* minor base for boot kbds */
#define HIDBP_BMSE_MINOR 32         /* minor base for boot mouses */
#define HIDBP_CTL_STR USB_CTL_STR "-hidbp"

    /*
     * HID BP structure 
     * ----------------
     *
     *  We always keep the 'open' open at the host and device level,
     * and only open the endp when calling hidb_open(). If
     * miscdev->name is NULL, then the miscdev hasn't been
     * registered. Trough ->name is a const char*, it is a dynamically
     * allocated pointer which must be freed _when_
     * unregistered. ->file is NULL when not opened.
     */

typedef struct
{
    ts_snode    snode;
    ts_usb_aset *aset;
    ts_usb_open open;
    ts_usb_tr   tr;
    struct miscdevice miscdev;
    struct file *file;  /* need it to close when destroying */
} ts_hidbp;

#ifdef __KERNEL__
#define ts_hidbp_nozero(boot)                   \
do                                              \
{                                               \
    ts_snode_clean (&(boot)->snode);            \
    ts_usb_tr_clean (&(boot)->tr);              \
    ts_usb_open_clean (&(boot)->open);          \
    ts_snode_ref (&(boot)->tr.snode);           \
} while (0)

#define ts_hidbp_clean(boot)                    \
do                                              \
{                                               \
    memset ((boot), 0, sizeof (*(boot)));       \
    ts_hidbp_nozero (boot);                     \
} while (0)
    
static inline
ts_hidbp* ts_hidbp_init (void)
{
    ts_hidbp* boot;
    gdlf_fntrace ("ts_hidbp_init (void)");

    boot = usb_kmalloc (sizeof (*boot), GFP_KERNEL);
    if (boot == NULL)
        return NULL;
    ts_hidbp_clean (boot);
    return boot;
}

static inline
void ts_hidbp_release (ts_hidbp* boot)
{
    gdlf_fntrace ("ts_hidbp_release (0x%p)", boot);
    gdlf_aassert (boot != NULL, return);
    gdlf_assert (boot);
    ts_snode_release (&boot->snode);
    ts_usb_open_release (&boot->open);
    if (boot->miscdev.name != NULL)
    {
        misc_deregister (&boot->miscdev);
        usb_kfree ((char*)boot->miscdev.name);
    }    
    return;
}

extern ts_slist hidbp_list;
extern int hidbp_register (ts_hidbp*);
extern int hidbp_unregister (ts_hidbp*);
extern int hidbp_destroy (ts_hidbp*);
extern int hidbp_reference (ts_usb_device*, int);
extern int hidbp_open (struct inode*, struct file*);
extern int hidbp_close (struct inode*, struct file*);
extern int hidbp_free_address (unsigned base);
extern int hidbp_config (ts_usb_cfel*, ts_usb_device*, 
                         ts_usb_cfg*, t_usb_cfg, 
                         ts_usb_iface*, t_usb_iface, 
                         ts_usb_aset*, t_usb_aset, __u32);
extern int hidbp_unconfig (ts_usb_device*, unsigned);
#endif /* __KERNEL__ */

    /* HID BP Mouse */

typedef struct s_bmse_report ts_bmse_report;
struct s_bmse_report
{
    union 
    {
        __u8    b1:1;   /* button 1 */
        __u8    b2:1;   /* button 2 */
        __u8    b3:1;   /* button 3 */
        __u8    spec:5; /* device specific */
    } buttons;
    __u8        xd;     /* X displacement */
    __u8        yd;     /* Y displacement */
};

#ifdef CONFIG_USB_HIDBP_MSE
extern struct file_operations bmse_fops;
#ifdef __KERNEL__
extern ssize_t bmse_read (struct file*, char*, size_t, loff_t*);
extern int bmse_ioctl (struct inode*, struct file *, 
                       unsigned int, unsigned long);
#endif /* __KERNEL__ */
#endif /* CONFIG_USB_HIDBP_MSE */


    /* HID BP Keyboard */

typedef struct s_bkbd_report ts_bkbd_report;
struct s_bkbd_report
{
    union
    {
        __u8    numlock:1;
        __u8    capslock:1;
        __u8    scrolllock:1;
        __u8    compose:1;
        __u8    kana:1;
        __u8    constant:3;
        
    } mod;
    __u8 reserved;
    __u8 k1;
    __u8 k2;
    __u8 k3;
    __u8 k4;
    __u8 k5;
    __u8 k6;
};

#ifdef CONFIG_USB_HIDBP_KBD
#ifdef __KERNEL__
extern struct file_operations bkbd_fops;
extern ssize_t bkbd_read (struct file*, char*, size_t, loff_t*);
extern int bkbd_ioctl (struct inode*, struct file *, 
                       unsigned int, unsigned long);
#endif /* CONFIG_USB_HIDBP_KBD */
#endif /* __KERNEL__ */

#endif /* _usb_hidbp_h_ */


/* Log:
 *
 * $Log: usb-hidbp.h,v $
 * Revision 1.1  1998/11/14 00:38:26  inaky
 * Isolated all the header files
 *
 * Revision 1.2  1998/10/22 00:07:43  inaky
 * Added a reference to hidbp->tr, as we _use_ it.
 *
 * Revision 1.1  1998/09/14 02:06:03  inaky
 * Removed the old boot-protocol.c, which had _everything_ in the same
 * file. Added a header file with boot protocol basis. It still depends
 * deeply on ../hidd/usb-hidd.h.
 *
 * Revision 1.3  1998/09/11 00:51:52  inaky
 * Moved the boot protocol to a single driver, to isolate it more easily.
 *
 * Revision 1.2  1998/08/30 13:58:25  inaky
 * Added boot protocol and basic user-mode read support.
 *
 * Revision 1.1  1998/08/28 23:53:16  inaky
 * New driver
 *
 * Revision 1.1  1998/08/28 00:14:23  inaky
 * New driver ...
 *
 */

