
/* USB Driver Stack v$Revision: 1.1 $ 
 *
 * (C) 1997 by Iaky Prez Gonzlez <inaky@peloncho.fis.ucm.es>
 *
 * Distributed under the terms of the GNU Public License Version 2
 * See files COPYRIGHT and COPYING at the root distribution directory
 *
 * $Id: usb-hub.h,v 1.1 1998/11/14 00:38:27 inaky Exp $
 *
 * Changelog at the bottom 
 *
 *
 *     Hub Manager
 *     -----------
 *
 *  Note some typedefs are defined at usb.h, though their underlying
 * structs are defined here.
 */

#ifndef _usb_hub_h_
#define _usb_hub_h_

#include <asm/spinlock.h>       /* spinlock_t */

#include <uusbd/usb.h>          /* UUSBD data structures */
#include <uusbd/usb-slist.h>    /* Sigle Linked Lists */
#include <uusbd/usb-stddesc.h>  /* Standard USB descriptors */


    /* 
     * Some definitions
     * ----------------
     */

#define USB_PORTS_PER_HUB_MAX   127
#define USB_HUB_ENDP_STAT_CH    0   /* status change endpoint index */
#define USB_CONNECT_TRIES       10  /* Max no. of tries to connect */

    /* See USB 1.0 11.6.2 */

#define USB_PORT_RESET_TIME        12  /* ms */
#define USB_PORT_AFTER_RESET_TIME  12  /* ms */

    /* 
     * 7 bytes the fixed part, and two which are bitfields of a
     * maximum of 256 elements each, what yields a maximum of 32
     * bytes each.
     */

#define USB_HUB_DESC_MAX_LEN       (7 + 2*32)

    /* Commands to khubdd */

#define USB_HUB_COM_NONE       0x00    /* No command */
#define USB_HUB_COM_STOP       0x01    /* Stop the hubd */
#define USB_HUB_COM_HUB_PROCESS 0x02   /* Process the hub */
#define USB_HUB_COM_NO_SLOTS   0x03    /* Report lack of slots */
#define USB_HUB_COM_HOST_ENUM  0x04    /* Enumerate hosts */
#define USB_HUB_COM_HOST_RM    0x05    /* Remove a host */


    /* khubdd control stuff */

#define USB_HUB_DQUEUE_SLOTS   5   /* Slots for the data queue */
#define USB_HUB_WFHS_LOOPS     10  /* Loops to wait for the khubdd to stop */
#define USB_HUB_WHFS_MSECS     10  /* msecs to wait per loop */



    /* 
     * Hub descriptor (USB Spec 1.0 11.11.2.1) 
     * ---------------------------------------
     *
     * FIXME: fucking GCC won't align correctly the bitfields, compile
     *        time option?  
     */

typedef struct
{
    ts_usb_desc_head head;
    __u8    bNbrPorts;                  /* Number of downstream ports this 
                                         * HUB has */
    __u16   wHubCharacteristics;        /* Port characteristics. See spec 
                                         * description */
    __u8    bPwrOn2PwrGood;             /* Time (in 2ms units) it takes 
                                         * to complete the power-on sequence 
                                         * on each port */
    __u8    bHubContrCurrent;           /* Maximum current (1mA units) 
                                         * requirement of the hub controller */
    __u8    PortData[64];               /* Comprises the two arrays below */
#if 0
    bit     DeviceRemovable[bNbrPorts]; /* Bitmap (byte-granular) of attach
                                         * type for each port. Warning! This
                                         * is a 1 based array, element 0 is
                                         * reserved, with BYTE granularity */

    bit     PortPwrCtrlMask[bNbrPorts];  /* Bitmap (byte-granular) of port
                                          * behaviour under gang-mode power
                                          * ctrl request. Warning! This is a
                                          * 1 based array, element 0 is
                                          * reserved, with BYTE granularity
                                          * (See the Spec Docs) */
#endif
} ts_usb_hub_desc;


    /*
     * Hub characteristics (USB 1.0 11.11.2.1) 
     * ---------------------------------------
     * 
     *  Masks to apply to the wHubCharacteristics field of a hub
     * descriptor to get the hub characteristics. Remember
     * wHubCharacteristics is little endian, so le16_to_cpu is done on
     * it.
     */

#define USB_HUB_CRTS_PWR_CTL_MASK   0x0003  /* Power control mode */
#define USB_HUB_CRTS_COMPD_MASK     0x0004  /* If compound device or not */
#define USB_HUB_CRTS_CURRENT_MASK   0x0018  /* Current protection mode */
#define USB_HUB_CRTS_CURRENT_MASK_S 3       /* Left shift count for ^ */


    /* 
     * Hub Class Requests (USB Spec 1.0 11.12.1) 
     * -----------------------------------------
     */

typedef enum
{
    e_usb_hub_req_get_status = 0,
    e_usb_hub_req_clear_feature,
    e_usb_hub_req_get_state,
    e_usb_hub_req_set_feature,
    e_usb_hub_req_get_descriptor = 6,
    e_usb_hub_req_set_descriptor
} e_usb_hub_req;


    /* 
     * Hub Class Hub Feature Selectors (USB Spec 1.0 Table 11-13) 
     * ----------------------------------------------------------
     */

typedef enum
{
    e_usb_hub_fes_c_local_pwr = 0,
    e_usb_hub_fes_c_over_current
} e_usb_hub_featsel;


    /* 
     * Hub Class Port Feature Selectors (USB Spec 1.0 Table 11-13) 
     * -----------------------------------------------------------
     *
     * FIXME: document each one
     */

typedef enum
{
    e_usb_port_fes_connection = 0,
    e_usb_port_fes_enable,
    e_usb_port_fes_suspend,
    e_usb_port_fes_over_current,
    e_usb_port_fes_reset,
    e_usb_port_fes_power = 8,
    e_usb_port_fes_low_speed,
    e_usb_port_fes_c_connection = 16,
    e_usb_port_fes_c_enable,
    e_usb_port_fes_c_suspend,
    e_usb_port_fes_c_over_current, 
    e_usb_port_fes_c_reset 
} e_usb_port_featsel;


    /* Bus status field (USB Spec 1.0 11.12.2.3)
     * -----------------------------------------
     * 
     * bmRequestType    0xA3
     * bRequest         USB_HUB_REQ_GET_STATE
     * wValue           0
     * wIndex           Port number, 1 based.
     * wLength          1
     * Data Returned    1 byte
     * 
     * If stall, it means the hub doesn't support it.
     */

#define USB_HUB_BUS_STAT_DP    0x01    /* D+ Line signal state */
#define USB_HUB_BUS_STAT_DN    0x02    /* D- Line signal state */


    /* Hub status and change (USB Spec 1.0 Tables 11-14 11-15) 
     * -------------------------------------------------------
     * 
     * bmRequestType    0xA0
     * bRequest         USB_HUB_REQ_GET_STATUS
     * wValue           0
     * wIndex           0
     * wLength          4
     * Data Returned    2 words, wHubStatus and wHubChange.
     *
     * NOTE: Use the macro le32_to_cpu() to transform the obtained data to
     *       CPU's byte order, then proceed. Obtained with request
     *       USB_HUB_REQ_GET_STATUS. Gives two words.  Acknowledge with
     *       USB_HUB_REQ_CLEAR_FEATURE.
     *
     *  The usb_hub_status_get() function will provide with a 32 bits
     * word containing the fourth values as specified by the masks. 
     */

#define USB_HUB_STATUS_LOCAL_PWR      0x00000001 /* & = 0 - Local power is ok, 1 - local power lost */
#define USB_HUB_STATUS_OVER_CURRENT   0x00000002 /* & = 0 - power ops ok, 1 - hub-wide over-current */
                                      
#define USB_HUB_STATUS_C_LOCAL_PWR    0x00010000  /* & = 1 - Change in local power status */
#define USB_HUB_STATUS_C_OVER_CURRENT 0x00020000  /* & = 1 - Change in over-current status */


    /* Port status and change (USB Spec 1.0 Tables 11-16 11-17)
     * --------------------------------------------------------
     * 
     * bmRequestType    0xA3
     * bRequest         USB_HUB_REQ_GET_STATUS
     * wValue           0
     * wIndex           Port number (1 based)
     * wLength          4
     * Data Returned    2 words, wPortStatus and wPortChange.
     *
     * NOTE: Acknowledge with USB_HUB_REQ_CLEAR_FEATURE. Use the macro
     *       le32_to_cpu() to transform the obtained data to CPU's byte
     *       order, then proceed.  
     */

#define USB_PORT_STAT_LOW_SPEED         0x00000100
#define USB_PORT_STAT_POWER             0x00000200
#define USB_PORT_STAT_RESET             0x00000010
#define USB_PORT_STAT_OVER_CURRENT      0x00000008
#define USB_PORT_STAT_SUSPEND           0x00000004
#define USB_PORT_STAT_ENABLE            0x00000002
#define USB_PORT_STAT_CONNECTION        0x00000001
                     
#define USB_PORT_STAT_C_RESET           0x00100000
#define USB_PORT_STAT_C_OVER_CURRENT    0x00080000
#define USB_PORT_STAT_C_SUSPEND         0x00040000
#define USB_PORT_STAT_C_ENABLE          0x00020000
#define USB_PORT_STAT_C_CONNECTION      0x00010000



    /* 
     * Port control structure 
     * ----------------------
     *
     *  A port is where we plug in devices, thus, we keep here info on
     * the kind of port, it's power control capabilities, the speed of
     * the connected device (if any), it's status, if there's any
     * device currently plugged and how is it. In case of having a
     * connected device, we'll keep a pointer to it's structure (which
     * incs it's use count).
     *
     * Note: to update this structure, you must use
     *       usb_port_status_get(). That will read in port's status and
     *       store it in this structure.  
     */

struct s_usb_port
{
    unsigned    removable:1;    /* Attached device is: 0 - removable, 1 - not removable */
    unsigned    pwrctl:1;       /* Port masks gang-mode pwr control? 0 - no, 1 - yes */
    unsigned    lowspeed:1;     /* Attached device's speed is: 0 - full, 1 - low */
    unsigned    reset:1;        /* Port is: 0 - resetted, 1 - still resetting */
    unsigned    overcurrent:1;  /* Power status is: 0 - ok, 1 - overcurrent */

    unsigned    power:1;        /* Port is: 0 - powered off, 1 - powered on */
    unsigned    connect:1;      /* Device is: 0 - disconnected, 1 - connected */
    unsigned    enable:1;       /* Port is: 0 - disabled, 1 - enabled */
    unsigned    suspend:1;      /* Port is: 0 - not suspended, 1 - suspended */

    __u8        try;            /* connection retry number */
    
    ts_usb_device   *device;    /* Pointer to the device hanging from this port. 
                                 * NULL if neither enabled nor suspended */    
};
 
#ifdef __KERNEL__
#define ts_usb_port_nozero(_port)               \
do                                              \
{                                               \
    (_port)->try = 1;                           \
} while (0)

#define ts_usb_port_clean(_port)                \
do                                              \
{                                               \
    memset ((_port), 0, sizeof (*(_port)));     \
    ts_usb_port_nozero (_port);                 \
} while (0)
extern ts_usb_port* ts_usb_port_init (void);
extern ts_usb_port* ts_usb_port_a_init (unsigned size);
extern void ts_usb_port_release (ts_usb_port*);
extern void ts_usb_port_a_release (ts_usb_port*, unsigned size);
#endif /* __KERNEL__ */


    /* 
     * Hub operations
     * --------------
     *
     *  The mandatory behaviour and interface for these functions is
     * the one given for their wrappers (below).
     *
     * FIXME: doc
     */

struct s_usb_hub_ops
{
    int (*schedule_stch) (ts_usb_hub*);

    ssize_t (*descriptor_get) (ts_usb_hub*, __u8, void*, size_t);
    ssize_t (*descriptor_set) (ts_usb_hub*, __u8, const void*, size_t);
    int (*descriptor_fill) (ts_usb_hub*);

    __s32 (*status_get) (ts_usb_hub*);
    __s8 (*bus_status_get) (ts_usb_hub*, __u8);

    int (*feature) (ts_usb_hub*, e_usb_hub_featsel, e_usb_req);
    int (*port_feature) (ts_usb_hub*,__u8,  e_usb_hub_featsel, e_usb_req);

    __s32 (*port_status_get) (ts_usb_hub*, __u8);
    int (*port_reset) (ts_usb_hub*, __u8);

    int (*free) (ts_usb_hub*);
};


    /* 
     * Hub control structure 
     * ---------------------
     *
     *  This is defined at usb.h
     */


    /* variables && housekeeping (hubm-hkeep.c) */

#ifdef __KERNEL__
extern int usb_hub_init (void);
extern int usb_hub_release (void);
extern int usb_hub_shutdown (void);

extern int usb_hub_destroy (ts_usb_hub*);
extern int usb_hub_register (ts_usb_hub*);
extern int usb_hub_unregister (ts_usb_hub*);
extern int usb_hub_register_device (ts_usb_hub*, ts_usb_device*, __u8);
extern int usb_hub_unregister_device (ts_usb_device*);
extern int usb_hub_detach_devices (ts_usb_hub*, __u32);


    /* Kernel hub driver daemon (khubdd.c) */

extern int usb_hub_startd (void);
extern int usb_hub_stopd (void);
extern void usb_hub_status_callback (ts_usb_tr*, int);
extern void usb_hub_status_poll (ts_usb_hub*);
extern void usb_hub_host_rm (ts_usb_host*);
extern void usb_hub_host_enum (void);


    /* Status changes (hubm-stch.c) */

extern int usb_port_status_change (ts_usb_hub* hub, __u8 port);


    /* 
     * Standard USB Hub class requests 
     * -------------------------------
     *
     * The functions which lack a NULLness check mean they are
     * mandatory, so it is not suppossed to be NULL.
     */


    /* Get the hub descriptor.
     *
     * Compliance:      USB Spec 1.0 11.12.2.4
     *
     * Parameters:  
     *
     *  ts_usb_hub*     Pointer to hub structure.
     *
     *  __u8            Descriptor index. Normally zero.
     *
     *  void*           Pointer to the buffer.
     *
     *  size_t          Size of the buffer for the descriptor
     *
     * Returns:
     *
     *  ssize_t         Bytes readen if ok [excluding request
     *                  phase's], errno code < 0 on error. Note that
     *                  no more than size_t bytes will be readen. Is
     *                  up to the caller to determine if the
     *                  descriptor has been fully read. 
     *
     * Global:
     *
     *  int usb_errno   On error, error code.  
     */

static inline
ssize_t usb_hub_descriptor_get (ts_usb_hub* hub, __u8 index, 
                                void* data, ssize_t size)
{
    return (*hub->operations->descriptor_get) (hub, index, data, size);
}

    /* Set a hub's descriptor
     *
     *  Overwrites the previous upon succesful reception of the new
     * one. Any of the hub's available descriptors may be overwritten,
     * specifying it's type and it's index number. However, not all
     * hubs support this operation, as this request is optional and
     * writes data to a class-specific descriptor.
     *
     * Compliance:      USB Spec 1.0 11.12.2.7
     *
     * Parameters:  
     *
     *  ts_usb_hub*     Pointer to the hub structure. If NULL, -EINVAL
     *                  will be returned.  
     *
     *  __u8            Descriptor index.
     *
     *  const void*     Pointer to the descriptor data in little
     *                  endian format. If NULL, -EINVAL will be
     *                  returned. 
     *
     *  size_t          Length of the descriptor data. If 0, -EINVAL
     *                  will be returned. 
     *
     * Returns:
     *
     *  ssize_t         Bytes transmited [excluding request phase's]
     *                  or errno code < 0 on error. 
     *
     * Global:
     *
     *  int usb_errno   On error, error code.  
     */

static inline
ssize_t usb_hub_descriptor_set (ts_usb_hub* hub, __u8 index, 
                                const void* data, ssize_t size)
{
    return 
        (hub->operations == NULL || hub->operations->descriptor_set == NULL)?
        -ENOSYS
        : (*hub->operations->descriptor_set) (hub, index, data, size);
}


    /* Initializes a hub control structure from a Hub Descriptor
     *
     *  It will read in the descriptor, parse it, and fill in the
     * structure, allocating space for it as needed. The structure
     * will hold data which must be freed with ts_usb_hub_release() before
     * usb_kfree()ing the structure.
     *
     * Parameters:  
     *
     *  ts_hub*         Pointer to hub structure to fill in.
     *
     * Returns:
     *
     *  int             0 if ok, errno code < 0 on error.
     */

static inline
int usb_hub_descriptor_fill (ts_usb_hub* hub)
{
    return (*hub->operations->descriptor_fill) (hub);
}


    /* Get hub status and change bits
     *
     *  The data is returned in 32 bits, being testable with the
     * #defined masks USB_HUB_STATUS_* for the status bits and
     * USB_HUB_STATUS_C_* for the status change bits. 
     *
     *  The wrapping function (this one) expects the stub function to
     * return the data in host order, so it must convert it from
     * little endian (as used by USB).
     *
     * NOTE: this funcion will update the hub structure's fields
     *       `local_power' and `overcurrent'.
     *
     * Compliance:      USB Spec 1.0 11.12.2.5
     *
     * Parameters:  
     *
     *  ts_usb_hub*     Pointer to the hub structure. If NULL, -EINVAL
     *                  will be returned.  
     *
     * Returns:
     *
     *  __s32           If > 0, status code (and with HUBD_STAT_* to
     *                  get the bits) [FIXME: things may get hairy if
     *                  sometime someone decides to use bit 31 for
     *                  status ...]. On error, <0 errno code.
     */

static inline
__s32 usb_hub_status_get (ts_usb_hub* hub)
{
    __s32 errcode;

    
    errcode = (*hub->operations->status_get) (hub);
    if (errcode < 0)
        return errcode;

        /* actualize hub data */

    write_lock (&hub->snode.lock);
    hub->local_power = errcode & USB_HUB_STATUS_LOCAL_PWR? 1 : 0;
    hub->overcurrent = errcode & USB_HUB_STATUS_OVER_CURRENT? 1 : 0;
    write_unlock (&hub->snode.lock);

    return errcode;
}


    /* Read the status of the bus
     *
     *  This function will read the bus status on fiven port as
     * sampled at the last EOF2 hub state. It will be returned on bits
     * 0 and 1 of the return value if > 0. You can then use the mask
     * values USB_HUB_BUS_STAT_DP and HUBD_BUS_STAT_DN to sample the
     * values of the D- and D+ lines.
     *
     * NOTE: This is a very lowlevel function, intended as a debugging
     *       aid.
     *
     * Compliance:      USB Spec 1.0 - 11.12.2.3
     *
     * Parameters:  
     *
     *  ts_usb_hub*     Pointer to the hub structure. If NULL, -EINVAL
     *                  will be returned.  
     *
     *  __u8            Port number.
     *
     * Returns:
     *
     * __s8             > 0 if ok, encoding the bus state (and with
     *                  HUBD_BUS_STAT_* to get the bits). <0 errno
     *                  code on error. If the hub doesn't support this
     *                  operation, it will answer with an stall and
     *                  return. The code returned will be -EOPNOTSUPP.
     */

static inline
__s8 usb_hub_bus_status_get (ts_usb_hub* hub, __u8 port)
{
    return 
        (hub->operations == NULL || hub->operations->bus_status_get == NULL)?
        -ENOSYS
        : (*hub->operations->bus_status_get) (hub, port);
}


    /* Reset/clear or acknowledge a hub feature. 
     *
     *  Acknowledge is performed if the feature is associated with an
     * status change as reported by usb_hub_status_get(); that's the case
     * of USB_HUB_FES_C_LOCAL_PWR and USB_HUB_FES_C_OVER_CURRENT (currently
     * the only ones implemented).
     *
     * Compliance:      USB Spec 1.0 11.12.2.1
     *
     * Parameters:  
     *
     *  ts_usb_hub*        Pointer to the hub structure. If NULL, -EINVAL
     *                  will be returned.  
     *
     *  e_usb_hub_featsel
     *
     *                  Feature to clear. If an unsupported feature is
     *                  selected, the endpoint stalls, it is
     *                  unstalled, and will return -EOPTNOTSUPP.
     *
     * Returns:
     *
     *  int             0 if ok, errno code < 0 on error.
     *
     * Global:
     *
     *  int usb_errno   On error, error code.  
     */

static inline
int usb_hub_feature_clear (ts_usb_hub* hub, e_usb_hub_featsel feature)
{
    return (*hub->operations->feature) (hub, feature, 
                                        e_usb_hub_req_clear_feature);
}


    /* Request a feature to be enabled
     *
     *  Will affect hub status, though change indicators aren't
     * affected by this request. 
     *
     * Compliance:      USB Spec 1.0 11.12.2.8
     *
     * Parameters:  
     *
     *  ts_usb_hub*        Pointer to the hub structure. If NULL, -EINVAL
     *                  will be returned.  
     *
     *  e_usb_hub_featsel
     *
     *                  Feature to set. If an unsupported feature is
     *                  selected, the endpoint stalls, it is
     *                  unstalled, and will return -EOPTNOTSUPP.
     *
     * Returns:
     *
     *  int             0 if ok, errno code < 0 on error.
     *
     * Global:
     *
     *  int usb_errno   On error, error code.  
     */

static inline
int usb_hub_feature_set (ts_usb_hub* hub, e_usb_hub_featsel feature)
{
    return (*hub->operations->feature) (hub, feature, 
                                        e_usb_req_set_feature);
}


    /* Reset a value reported in the port's status/clear a feature,
     * disabling it. 
     *
     *  Used to acknowledge status changes when the features are
     * associated with them [it is the case of USB_PORT_STAT_C_{RESET,
     * OVER_CURRENT, SUSPEND, ENABLE, CONNECTION}] bits being active.
     *
     *  Clearing USB_PORT_FES_SUSPEND will force a resume on the port
     * by the host.
     *
     *  Clearing USB_PORT_FES_ENABLE will disable the port. 
     *
     *  Clearing USB_PORT_FES_POWER will be switch off port's power if
     * the hub supports switching of individual port's
     * power. Otherwise, you need to turn off all the ports for power
     * off to work. 
     *
     * Compliance:      USB Spec 1.0 - 11.12.2.2
     *
     * Parameters:  
     *
     *  ts_usb_hub*        Pointer to the hub structure. If NULL, -EINVAL
     *                  will be returned.  
     *
     *  __u8            Port number.
     *
     *  e_usb_port_featsel
     *
     *                  Feature to clear/acknowledge. If an unsupported
     *                  feature is selected, the endpoint stalls, it's
     *                  unstalled and will return -EOPTNOTSUPP.
     *
     * Returns:
     *
     *  int             0 if ok, on error, errno code < 0.
     *
     * Global:
     *
     *  int usb_errno   On error, error code.  
     */

static inline
int usb_port_feature_clear (ts_usb_hub* hub, __u8 port, 
                             e_usb_port_featsel feature)
{
    return (*hub->operations->port_feature) (hub, port, feature, 
                                             e_usb_req_clear_feature);
}


    /*  Set a feature on the port reported in the port status. Change
     * indicators cannot be acknowledged using these requests.
     *
     *  Setting USB_PORT_FES_SUSPEND suspends the device hanged to the
     * selected port.
     *
     *  Setting USB_PORT_FES_RESET initiaties a reset on the selected
     * port, and when it is complete, the USB_PORT_STAT_C_RESET bit of
     * the port status change bits is set.
     * 
     * Compliance:      USB Spec 1.0 - 11.12.2.9
     *
     * Parameters:  
     *
     *  ts_usb_hub*        Pointer to the hub structure. If NULL, -EINVAL
     *                  will be returned.  
     *
     *  __u8            Port number
     *
     *  e_usb_port_featsel
     *
     *                  Feature to set
     *
     * Returns:
     *
     *  int             0 if ok, errno code < 0 on error.
     *
     * Global:
     *
     *  int usb_errno   On error, error code.  
     */

static inline
int usb_port_feature_set (ts_usb_hub* hub, __u8 port, 
                              e_usb_port_featsel feature)
{
    return (*hub->operations->port_feature) (hub, port, feature, 
                                             e_usb_hub_req_set_feature);
}


    /* Get the port status, as well as the status bits which have
     * experienced a change. On the lower word are stored the status
     * bits and in the higher word the status change bits. You can use
     * the #defined masks USB_PORT_STAT_* for the status bits and
     * USB_PORT_STAT_C_* for the status change bits.
     *
     *  To acknowledge the changes reported with this function [most
     * of them], use the port_feature_clear() function.
     *
     * NOTE: all the values are in host order, though USB uses little
     *       endian. The wrapper expect like this the values. Your
     *       function must convert them.
     *
     * NOTE: all the statys fields in the port structure will be
     *       updated to match it's new value, so you can rely on them
     *       instead of testing some bits.
     *
     * Compliance:      USB Spec 1.0 - 11.12.2.6
     *
     * Parameters:  
     *
     *  ts_usb_hub*     Pointer to the hub structure. If NULL, -EINVAL
     *                  will be returned.  
     *
     *  __u8            Port number.
     *
     * Returns:
     *
     *  __s32           If > 0, status code (and with
     *                  USB_PORT_STAT_* to get the bits) [FIXME:
     *                  things may get hairy if sometime someone
     *                  decides to use bit 31 for status ...]. On
     *                  error, <0 errno code. 
     */

static inline
__s32 usb_port_status_get (ts_usb_hub* hub, __u8 port)
{
    __s32 errcode;
    
    errcode = (*hub->operations->port_status_get) (hub, port);
    if (errcode < 0)
        return errcode;

        /* Update port status bits */

    write_lock (&hub->device->lock);
    hub->port[port].connect = errcode & USB_PORT_STAT_CONNECTION? 1 : 0;
    hub->port[port].enable = errcode & USB_PORT_STAT_ENABLE? 1 : 0;
    hub->port[port].suspend = errcode & USB_PORT_STAT_SUSPEND? 1 : 0;
    hub->port[port].overcurrent = 
        errcode & USB_PORT_STAT_OVER_CURRENT? 1 : 0;
    hub->port[port].reset = errcode & USB_PORT_STAT_RESET? 1 : 0;
    hub->port[port].power = errcode & USB_PORT_STAT_POWER? 1 : 0;
    hub->port[port].lowspeed = 
        errcode & USB_PORT_STAT_LOW_SPEED? 1 : 0;
    write_unlock (&hub->device->lock);

    return errcode;
}


    /* Reset a hub's port.
     *
     *  This function will reset a USB port, blocking [interruptibly]
     * until the reset is over. FIXME: Must use some sort of wait
     * queuing here; adding a pointer field to the port structure to
     * the wait queue.
     *
     * Parameters:  
     *
     *  ts_usb_hub*     Pointer to the hub structure. If NULL, -EINVAL
     *                  will be returned.  
     *
     *  __u8            Number of the port to reset. -ENODEV if the
     *                  port doesn't exist.
     *
     * Returns:
     *
     *  int             0 if ok, errno code < 0 on error.
     */

static inline
int usb_port_reset (ts_usb_hub* hub, __u8 port)
{
    return (*hub->operations->port_reset) (hub, port);
}


    /* Release the operations structure
     *
     *  This function is called when the hub structure is going to be
     * released/freed. You _MUST_ implement here a 'destructor' who
     * will unreference all referenced drivers, so to leave everything
     * as before configuring the hub.
     *
     * Parameters:  
     *
     *  ts_usb_hub*     Pointer to the hub structure. If NULL, -EINVAL
     *                  will be returned.  
     *
     * Returns:
     *
     *  int             0 if ok, errno code < 0 on error.
     *
     * Global:
     *
     *  int usb_errno   On error, error code.  
     */

static inline
int usb_hub_free (ts_usb_hub* hub)
{
    return (*hub->operations->free) (hub);
}


    /* Schedule a read from the hub's status change endpoint
     *
     *  Using the tr data in the ts_usb_hub structure (see sample in
     * hubd). Initially, if error re-scheduling, we take the most
     * radical option, to reset the hub directly. However, FIXME, this
     * is quite radical. Should think of a possible better way.
     *
     *  Note this function _will_ complain itself, so the caller
     * should have to do nothing in case of error but raise it on the
     * call chain [if needed].
     *
     *  If this function is not supported, the hub is taken to need
     * polling and automagically added to the poll list when
     * registered.
     *
     * Parameters:  
     *
     *  ts_usb_hub*     Pointer to the hub structure we want to
     *                  schedule with.
     *
     * Returns:
     *
     *  int             0 if ok, errno code < 0 on error.
     *
     * Global:
     *
     *  int usb_errno   On error, error code.  
     *
     *  int hubd_errno  On error, error code.  
     */

static inline
int usb_hub_schedule_status_change (ts_usb_hub* hub)
{
        /* FIXME: handle non -ENOSYS errors */
    return 
        (hub->operations == NULL || hub->operations->schedule_stch == NULL)?
        -ENOSYS
        : (*hub->operations->schedule_stch) (hub);
}

static inline
void usb_hub_dump_devices (ts_usb_hub* hub)
{
    int cnt;
    printk (KERN_DEBUG "hub 0x%p has %d ports at 0x%p\n", 
            hub, hub->ports, hub->port);
    for (cnt = 0; cnt < hub->ports; cnt++)
        printk (KERN_DEBUG "hub 0x%p, port %d, device 0x%p\n",
                hub, cnt, hub->port[cnt].device);
}

#endif /* __KERNEL__ */

#endif /* _usb_hub_h_ */


/* Log:
 *
 * $Log: usb-hub.h,v $
 * Revision 1.1  1998/11/14 00:38:27  inaky
 * Isolated all the header files
 *
 * Revision 1.48  1998/11/09 21:24:32  inaky
 * Fixed the hub descriptor to have the ts_usb_desc_head as first
 * element.
 *
 * Revision 1.47  1998/10/18 15:10:02  inaky
 * Added the definition of maximum number of retries on a port to connect
 * a device. After that tries, the count wraps to zero. This affects the
 * times waited on the reset and after-reset stages.
 *
 * Revision 1.46  1998/10/14 21:42:28  inaky
 * Added methods for telling khubdd to remove a host or re-enumerate the
 * hosts.
 *
 * Revision 1.45  1998/09/24 15:52:16  inaky
 * Moved 'hub_port' to 'port'. Finished the split of _clean into _nozero
 * and _clean.
 *
 * Revision 1.44  1998/09/14 02:32:35  inaky
 * Added some definitions for times, removed the usb_hub_errno variable.
 *
 * Revision 1.43  1998/09/02 21:09:10  inaky
 * Removed the polled hubs array in favour of doing it by HCD
 * means. Fixed parameters to descriptor hub functions.
 *
 * Revision 1.42  1998/08/30 13:45:30  inaky
 * Reduced to 5 seconds the poll interval. Added a try count to the port
 * structure to keep the count of failed retries (modulo 10) to attach a
 * device to a port.
 *
 * Revision 1.41  1998/08/27 22:14:44  inaky
 * Moved _clean() to use memset(0).
 *
 * Revision 1.40  1998/08/27 10:48:44  inaky
 * Moved ts_*_free() to ts_*_release(), made them return void and severed
 * assertions.
 *
 * Revision 1.39  1998/08/24 02:20:26  inaky
 * Changed name of all enumerations to be prefixed with the enumeration
 * type name.
 *
 * Revision 1.38  1998/08/19 23:47:08  inaky
 * Added a debug function to dump the device pointers of hub's ports.
 *
 * Revision 1.37  1998/07/30 17:50:58  inaky
 * Added reason argument to hub callback
 *
 * Revision 1.36  1998/07/23 14:21:17  inaky
 * Made all errnos volatile.
 *
 * Renamed the ts_usb_trbuf struct to ts_usb_tr.
 *
 * Revision 1.35  1998/07/16 14:06:15  inaky
 * Moved from the old read/write interface to the new, request-based
 * interface.
 *
 * Revision 1.34  1998/07/14 21:46:30  inaky
 * Modified the poll interval to ten seconds to give klogd a break
 * (during tests) and added some fixmes I got to check someday.
 *
 * Revision 1.33  1998/07/14 15:43:57  inaky
 * Moved the status bit update code from the hub driver to the hub
 * manager wrapping functions.
 *
 * Revision 1.32  1998/07/14 14:53:31  inaky
 * Added a comment on future directions :)
 *
 * Revision 1.31  1998/07/12 22:47:41  inaky
 * Added the schedulement of status change read to be one of the
 * hub-specific operations. Moved the definition of the hub structure to
 * usb.h so that the root hub can be included in the host structure (and
 * not as a pointer, as before).
 *
 * Added protos for new functions, fixed other protos and added the
 * inline for the status change read schedulement.
 *
 * Revision 1.30  1998/06/21 20:12:06  inaky
 * Removed the export declaration for the hub list and added a comment on
 * possible deprecation of the cfg, iface and aset fields of the hub
 * structure, as a device cannot have a hub at an interface, but at the
 * whole device.
 *
 * Revision 1.29  1998/06/20 21:54:51  inaky
 * Massive changes to split the old monolithic hubd driver into a hub
 * manager (hub wrappers and daemon) and a dynamic driver, which can be
 * replaced and its different parts reused.
 *
 * Revision 1.28  1998/05/24 12:01:18  inaky
 * Added inclusion of spinlock.h for hubd data queue locking.
 *
 * Revision 1.27  1998/05/02 00:47:52  inaky
 * Moved in the definitions, some declarations and prototypes from the
 * khubdd.
 *
 * Revision 1.26  1998/04/13 21:56:14  inaky
 * Added some standard definitions, fixed 'current' mask definition,
 * changed some stuff for stetical purposes, added comments on fields
 * which are automagically updated by hub*_status_get(). New fields to
 * ts_hubd [cfel position, status and trbuffer]; fixed some declarations,
 * adapted the hubd_config() to the new scheme; fixed flags for unconfig
 * and config methods; added destructor; added callback and callback
 * handling methods; fixed more declarations and finished rest of hub
 * functions.
 *
 * Revision 1.25  1998/03/28 19:26:10  inaky
 * Fixed declaration of hubd_search()
 *
 * Revision 1.24  1998/01/11 15:44:16  inaky
 * Finished functions from the descriptor, inteface and endpoint modules,
 * fixed docs, pretty printed and removed craps.
 *
 * Revision 1.23  1998/01/08 12:19:09  inaky
 * Added copyright strings to all important files
 *
 * Revision 1.22  1998/01/08 11:49:11  inaky
 * Finishing the configuration system. Only the final cfel-stuff is
 * missing to implement; however, nothing is tested.
 *
 * Revision 1.21  1998/01/02 00:12:16  inaky
 * Finished the new driver scheme. Documented at usbd-drv.c
 *
 * Revision 1.20  1997/12/31 05:37:00  inaky
 * Almost finished all the lists and life cycle for hosts, drivers and
 * devices. Some craps to be fixed yet.
 *
 * Revision 1.19  1997/12/30 22:30:17  inaky
 * Added the foundations of the HCD and fixed some bugs.
 *
 * Revision 1.18  1997/12/30 20:15:43  inaky
 * Moved everything to the new lists scheme. Untested yet.
 *
 * Revision 1.17  1997/12/29 00:25:36  inaky
 * Fixed unimplemented functions to return -ENOSYS and moving the khubdd
 * to use signals as commands.
 * Also added function traces on most files that had them disabled (donkey!)
 *
 * Revision 1.16  1997/12/27 20:06:27  inaky
 * Finished fixinf the __KERNEL__ #ifdef wrapping and moved the ts_dtor.h
 * to usb-dtor.h
 *
 * Revision 1.15  1997/12/27 16:44:24  inaky
 * Moving to GDLF and fixinf docs. Removing more cruft.
 *
 * Revision 1.14  1997/12/22 02:44:29  inaky
 * Trying to move all the documentation to the source files from the
 * header files, as it was duplicated.
 *
 * Revision 1.13  1997/12/13 20:56:31  inaky
 * Moving all the documentation to the .c files except the ones for data
 * structures and inline functions. Helps making the header files cleaner.
 *
 * Revision 1.12  1997/12/08 02:40:59  inaky
 * Fixed some stuff of the #includes, modified the makefile targets to be
 * more flexible, tried to finish the device configuration section.
 *
 * Revision 1.11  1997/12/06 00:19:40  inaky
 * Almost finished the usbd/host, usbd/cfg and a lot of meat from
 * usbd/device. Also added the driver definition. Got to polish it yet
 * ...
 *
 * Revision 1.10  1997/11/29 20:21:48  inaky
 * Documented and almost fixes the read/write/abort API, documented a
 * damn lot and implemented some stuff
 *
 * Revision 1.9  1997/11/29 00:46:30  inaky
 * Documented almost all the hub basic API, added spin locks to
 * usb_host_register() and some more stuff. Going for more.
 *
 * Revision 1.8  1997/10/31 05:44:30  inaky
 * *** empty log message ***
 *
 * Revision 1.7  1997/10/17 00:16:00  inaky
 * *** empty log message ***
 *
 * Revision 1.6  1997/10/02 21:29:01  inaky
 * Remade all the directory structure to make it cleaner when the USB
 * class drivers come.
 *
 * Revision 1.5  1997/10/01 21:28:11  inaky
 * Re-making a lot of stuff in the hadnling of host, hub and device
 * initialization and handling, following a more correct approach of
 * registering/unregistering. Also moved the usb_host array to be an
 * array of pointers, so dynamic reallocation of it is easier, and the
 * host structure needs not to be moved in memory. The same applies to
 * the device arrays for each host.
 *
 * Revision 1.4  1997/09/30 13:04:42  inaky
 * Moved the name of the package to uusbd: Universal USB Driver
 *
 * Revision 1.3  1997/04/20 21:25:56  inaky
 * Finished the spec for the Hub Class (now gotta implement it)
 * :). Started coding the Frame Lists code, hope UHCI support will be
 * finished in some sessions.
 * This code is breaking, I'm doing something bad with kmallocs. kFree()ing
 * something not kmalloc()ed.
 *
 * Revision 1.2  1997/04/06 21:14:31  inaky
 * Still with the definition of the access interface. More or less it is
 * taking an stable form.
 *
 * Revision 1.1  1997/03/31 20:24:46  inaky
 * Pfff ... a lot. Mostly added descriptors stuff (HUB class still is
 * missing a lot) and I think I've got a definitive idea of how to
 * implement the storing of configurations and interfaces.
 *
 * Revision 1.1  1997/02/22 22:25:44  inaky
 * *** empty log message ***
 */

