
/* USB Driver Stack v$Revision: 1.2 $ 
 *
 * (C) 1997 by Iaky Prez Gonzlez <inaky@peloncho.fis.ucm.es>
 *
 * Distributed under the terms of the GNU Public License Version 2
 * See files COPYRIGHT and COPYING at the root distribution directory
 *
 * $Id: usb-slist.h,v 1.2 1998/11/16 18:45:07 inaky Exp $
 *
 * Changelog at the bottom 
 *
 *
 *     Simply linked Lists
 *     -------------------
 *
 *  This file defines means for using (together with slist.c) singly
 * linked lists.
 *
 *  We define two kinds of singly linked lists: simple and
 * multiple. The simple lists are just one list. The multiple are
 * multiple lists which share the list's nodes. See below.
 *
 *  Mind that *all* the elements are managed by the list, and you
 * should only use the methods in this header file to manipulate
 * them. In special, when referencing/unreferencing them, as the code
 * will automatically call destructors and free memory (garbage
 * collection) whenever a use count drops to zero. This also implies
 * that every element plugged into a list must be dynamically
 * allocated or if static, never reach zero in it's use count.
 *
 *  If the symbol TS_SLIST_EXPORT is #defined, then this header file
 * will export macros and function declarations, as well as type
 * declarations. This is intended to be used as for the __KERNEL__
 * define in the Linux Kernel, to avoid user space code seeing things
 * intended only for kernel code. See the environments stuff.
 */


#ifndef _usb_slist_h_
#define _usb_slist_h_


    /* 
     * Environment definition
     * ----------------------
     *
     *  This is to ease reusability, by defining an environment, we can
     * use this code to run at many places almost without tweaking.
     *
     *  Currently only defined for Linux Kernel and UUSBD usage.
     */

#ifdef __linux__
#  include <asm/spinlock.h>     /* rwlock_t, write_[un]lock() */
#  include <asm/atomic.h>       /* rwlock_t, write_[un]lock() */
#  define ts_lock rwlock_t
#  ifdef __KERNEL__
#    include <linux/string.h>   /* memset() */
#    define ts_slist_write_lock(a) write_lock((a))
#    define ts_slist_write_unlock(a) write_unlock((a))
#    define TS_SLIST_EXPORT
#    ifdef __UUSBD__
#      include <uusbd/usb-dtor.h>     /* ts_dtor */
#      include <uusbd/usb-mm.h>       /* usb_kmalloc(), usb_kfree(), GFP_KERNEL */
#      define ts_slist_malloc(a) usb_kmalloc ((a), GFP_KERNEL)
#      define ts_slist_mfree(a) usb_kfree ((a))
#    endif /* __UUSBD__ */
#  endif /* __KERNEL__ */
#else

#error SLIST: Undefined environment!

#endif

    /*
     * Define to get, whenever a reference or unreference is done, a
     * dump on which object did it, and how many references it has
     */

#undef SLIST_REF_DUMP


    /* Use counts
     * ----------
     *
     *  Use counts are kept for tracking the number of references made
     * to each resource. Whenever a pointer is held to an structure
     * which keeps use counts, it must be incremented. Thus, when that
     * pointer is changed (reference released), the use count is
     * decremented.
     *
     *  Use counts are atomic_t. 0 means unused, any other means
     * used. Always use methods atomic_* to manipulate them, as they
     * guarantee the atomicity of the operation.
     *
     *  However, for referencing and unreferencing snodes and Nsnodes,
     * do not use the atomic operations. Use the already-for-that
     * defined functions ts_snode_ref() and ts_snode_unref().
     *
     *  IMPORTANT! To keep simplicity, _both_ ts_snode and ts_Nsnode
     * structures have the fields in the same order, starting with
     * destructors, use count and next fields. This way they can be
     * handled with the same set of functions to reference,
     * unreference and handle within lists (see ts_usb_drv_use).
     */


    /* 
     * Single Singly-Linked List node
     * ------------------------------
     *
     *  This is a node of a generic list. As the type of the
     * underlying structure which forms the node of the list is not
     * our problem, we just to use as first element of the
     * structure an special struct which comprises a ts_dtor and a
     * pointer to the next element in the list. 
     *
     *  This way, though the list may be on any type, we can walk it
     * by assuming all pointers to elements begin with a ts_snode. We
     * also can destroy each node, as the ts_dtor gives us a
     * destructor function to free the contents of the underlying
     * structure. Later we only have to free the structure itself
     * using the pointer. Note that the destructor has to be present
     * *always* if we plan to let the list destroy each node (what
     * will be done if we destroy the list and the node's use count
     * reaches zero when removing them from it).
     *
     *  Also note we have available a use count. Each reference into
     * any list will increment the use count.
     */

typedef struct s_snode ts_snode;

struct s_snode
{
    ts_dtor  dtor;
    atomic_t usecnt;
    ts_snode *next;
};

#ifdef TS_SLIST_EXPORT
static inline
ts_snode* ts_snode_next (ts_snode* snode)
{
    return snode->next;
}

#define ts_snode_nozero(snode)                  \
do                                              \
{                                               \
    ts_dtor_nozero (&(snode)->dtor);            \
    atomic_set (&(snode)->usecnt, 0);           \
} while (0)

#define ts_snode_clean(snode)                   \
do                                              \
{                                               \
    memset (snode, 0, sizeof (*(snode)));       \
    ts_snode_nozero (snode);                    \
} while (0)

extern ts_snode* ts_snode_init (tf_dtor);
extern void ts_snode_release (ts_snode*);
extern void ts_snode_printk (const ts_snode*);


    /* Reference an snode
     *
     *  Add a reference to a snode, incrementing its use count. You
     * should use it like this, as it [should :)] enhances
     * readability:
     *
     *  pointer_to_snode = ts_snode_ref (new_snode_pointer);
     *
     *  See ts_snode_unref() for more usage guides.
     *
     * Parameters:  
     *
     *  ts_snode*       Pointer to the snode to reference.
     *
     * Returns:
     *
     *  ts_snode*       Pointer to the snode referenced.
     */

static inline
ts_snode* ts_snode_ref (ts_snode* snode)
{
    atomic_inc (&snode->usecnt);
#ifdef SLIST_REF_DUMP
    printk (KERN_DEBUG "snode 0x%p usecount up to %d\n", 
            snode, atomic_read (&snode->usecnt));
#endif
    return snode;
}


    /* Unreference an snode
     *
     *  This function is called when a reference is dropped to an
     * snode. This will cause the reference count to be decremented
     * and if reaches zero, it's destructor will be called (using
     * ts_snode_release()). After that, it's memory will be freed.
     *
     *  Also note this function won't take care of removing the
     * snode from any list it could be in. That's a destructor's
     * mission. 
     *
     *  After doing this, you should set the pointer to a NULL value
     * or a new pointer. Note that substituting for a new value should
     * be done like this [this is orientative :)]:
     *
     *  .
     *  .
     *  pointer = ts_snode_ref (&snode_1);
     *  .
     *  .
     *  ts_snode_unref (pointer, NULL);
     *  pointer = ts_snode_ref (&snode_2);
     *  .
     *  .
     *  ts_snode_unref (pointer, NULL);
     *  pointer = NULL;
     *  .
     *  .
     *
     *  The _nd unref function works the same, but it _won't_ call the
     * destructor nor free the memory [it just dereferences]. This is
     * used by destructor functions to unreference without
     * destroying. The scenario is: the destructor gets called, it
     * references the struct to evitate its (again) destruction when
     * last 'administrative' reference is released, does all the
     * release opearations and then unreferences, dropping the
     * reference count to zero [should ensure this]. Then it cleans
     * and let's the caller free.
     *
     * Parameters:  
     *
     *  ts_snode*       Pointer to the snode to unreference.
     *
     *  int*            Pointer to a variable where to store the
     *                  result code of the call to the destructor. If
     *                  NULL, won't be used.
     *                  
     *
     * Returns:
     *
     *  int             Use count after decreasing (if zero, destroyed).
     */

static inline
int ts_snode_unref (ts_snode* snode, int* perrcode)
{
#ifdef SLIST_REF_DUMP
    printk (KERN_DEBUG "snode 0x%p usecount down to %d\n", 
            snode, atomic_read (&snode->usecnt) - 1);
#endif
    if (atomic_dec_and_test (&snode->usecnt))
    {
        int errcode;
        
        errcode = ts_dtor_exec (&snode->dtor);
        if (perrcode)
            *perrcode = errcode;
        ts_slist_mfree (snode);
        return 0;
    }
    return atomic_read (&snode->usecnt);
}

static inline
int ts_snode_unref_nd (ts_snode* snode)
{
    atomic_dec (&snode->usecnt);
#ifdef SLIST_REF_DUMP
    printk (KERN_DEBUG "snode 0x%p usecount down to %d\n", 
            snode, atomic_read (&snode->usecnt));
#endif
    return atomic_read (&snode->usecnt);
}
#endif /* TS_SLIST_EXPORT */


    /* 
     * Single Linked List
     * ------------------
     *
     *  Structure which keeps a singly linked list of ts_snode's or
     * structures with it as head.
     *
     *  Never access directly to the list, as the interface may change
     * in the future. Instead, use the helper macros or functions
     * except for.
     */

typedef struct s_slist ts_slist;

struct s_slist
{
    atomic_t    usecnt;     /* use count of this structure */
    size_t      elements;   /* number of elements in the list */
    ts_lock     lock;       /* lock for twikinling with the list */
    ts_snode    *head;      /* first element of the list */
};

#ifdef TS_SLIST_EXPORT
#define ts_slist_nozero(slist)                          \
do                                                      \
{                                                       \
 /* __USB_SMP ((slist)->lock = RW_LOCK_UNLOCKED); */    \
    atomic_set (&(slist)->usecnt, 0);                   \
} while (0)

#define ts_slist_clean(slist)                   \
do                                              \
{                                               \
    memset ((slist), 0, sizeof (*(slist)));     \
    ts_slist_nozero (slist);                    \
} while (0)

extern ts_slist* ts_slist_init (void);
extern void ts_slist_release (ts_slist*);
extern void ts_slist_printk (const ts_slist*);

static inline
ts_snode* ts_slist_head (ts_slist* slist)
{
    return slist->head;
}
#define ts_slist_next(psnode) ts_snode_next ((psnode))
extern ts_snode* ts_slist_prev (ts_slist*, ts_snode*);
extern ts_snode* ts_slist_last (ts_slist*);
extern int ts_slist_prep (ts_slist*, ts_snode*);
extern int ts_slist_append (ts_slist*, ts_snode*);
extern int ts_slist_move (ts_slist*, ts_snode*, ts_snode*);
extern int ts_slist_del (ts_slist*, ts_snode*);
extern int ts_slist_del_fast (ts_slist*, ts_snode*, ts_snode*);
#endif /* TS_SLIST_EXPORT */


    /*
     * Multiple Singly-linked list node
     * --------------------------------
     *
     *  Yes, this looks akward.
     *
     *  The idea behind this thingie is to allow a node to be in
     * multiple lists, sharing the use counts and destructors. This is
     * used in the UUSBD for the devices, which must be in three
     * lists. However, nicer then making different interfaces is to
     * make a common one, as it may be helpful for more things.
     *
     *  For a node who is wanted to be in N lists, you first have got
     * to declare the node type (ts_Nsnode). You do this by #defining
     * NLISTS to N; then you define a block of aliases (needed, should
     * change the way of doing someday) and then do the actual
     * declaration with 'ts_msnode_declare'. This defines a
     * 'ts_Nsnode' type which is an alias for 'struct ts_Nsnode'. N
     * must be an integer greater than zero.
     *
     *  It is VERY important that a ts_Nsnode looks like a ts_snode;
     * they must begin with the same set of elements, so that the same
     * set of functions can operate identically on the usecounts and
     * destructors. See the comment at the 'Use Counts' documentation
     * above.
     *
     *  We define 'NLIST' as the number of the list to which we want
     * to refer. There are N lists, from 0 to N-1.
     *
     *  Thus, in all the macros below, 'N' refers to the number of lists
     * in the ts_msnode, and 'NLIST' to the number of a list, and in
     * both cases it must be an expression that evaluates to a
     * positive or zero integer (with the implicit ranges, of course,
     * as there are NO RANGE CHECKS).  
     *
     *  For doing the declaration of the data types:
     *
     * #define NLISTS N
     * 
     * #define s_Nslist        slist_stk(s_,NLISTS,slist)
     * #define s_Nsnode        slist_stk(s_,NLISTS,snode)
     * #define ts_Nslist       slist_stk(ts_,NLISTS,slist)
     * #define ts_Nsnode       slist_stk(ts_,NLISTS,snode)
     *
     * ts_msnode_declare;
     * ts_mslist_declare;
     *
     *
     *  For doing the declaration of the functions which operate on
     * them: 
     *     
     * #define NLISTS N
     * 
     * #define s_Nslist        slist_stk(s_,NLISTS,slist)
     * #define s_Nsnode        slist_stk(s_,NLISTS,snode)
     * #define ts_Nslist       slist_stk(ts_,NLISTS,slist)
     * #define ts_Nsnode       slist_stk(ts_,NLISTS,snode)
     *
     * #define ts_Nslist_append   slist_stk(ts_,NLISTS,slist_append)
     * #define ts_Nslist_nozero   slist_stk(ts_,NLISTS,slist_nozero)
     * #define ts_Nslist_clean    slist_stk(ts_,NLISTS,slist_clean)
     * #define ts_Nslist_del      slist_stk(ts_,NLISTS,slist_del)
     * #define ts_Nslist_del_fast slist_stk(ts_,NLISTS,slist_del_fast)
     * #define ts_Nslist_release  slist_stk(ts_,NLISTS,slist_release)
     * #define ts_Nslist_printk   slist_stk(ts_,NLISTS,slist_printk)
     * #define ts_Nslist_head     slist_stk(ts_,NLISTS,slist_head)
     * #define ts_Nslist_init     slist_stk(ts_,NLISTS,slist_init)
     * #define ts_Nslist_last     slist_stk(ts_,NLISTS,slist_last)
     * #define ts_Nslist_move     slist_stk(ts_,NLISTS,slist_move)
     * #define ts_Nslist_next     slist_stk(ts_,NLISTS,slist_next)
     * #define ts_Nslist_prep     slist_stk(ts_,NLISTS,slist_prep)
     * #define ts_Nslist_prev     slist_stk(ts_,NLISTS,slist_prev)
     * #define ts_Nsnode_nozero   slist_stk(ts_,NLISTS,snode_nozero)
     * #define ts_Nsnode_clean    slist_stk(ts_,NLISTS,snode_clean)
     * #define ts_Nsnode_release  slist_stk(ts_,NLISTS,snode_release)
     * #define ts_Nsnode_printk   slist_stk(ts_,NLISTS,snode_printk)
     * #define ts_Nsnode_init     slist_stk(ts_,NLISTS,snode_init)
     * #define ts_Nsnode_next     slist_stk(ts_,NLISTS,snode_next)
     * #define ts_Nsnode_ref      slist_stk(ts_,NLISTS,snode_ref)
     * #define ts_Nsnode_unref    slist_stk(ts_,NLISTS,snode_unref)
     * #define ts_Nsnode_unref_nd slist_stk(ts_,NLISTS,snode_unref_nd)
     * 
     * ts_msnode_declare_functions;
     * ts_mslist_declare_functions;
     *
     *
     *  And for implementation (for .c files):
     * 
     * #define NLISTS N
     * 
     * #define s_Nslist           slist_stk(s_,NLISTS,slist)
     * #define s_Nsnode           slist_stk(s_,NLISTS,snode)
     * #define ts_Nslist          slist_stk(ts_,NLISTS,slist)
     * #define ts_Nslist_append   slist_stk(ts_,NLISTS,slist_append)
     * #define ts_Nslist_nozero   slist_stk(ts_,NLISTS,slist_nozero)
     * #define ts_Nslist_clean    slist_stk(ts_,NLISTS,slist_clean)
     * #define ts_Nslist_del      slist_stk(ts_,NLISTS,slist_del)
     * #define ts_Nslist_del_fast slist_stk(ts_,NLISTS,slist_del_fast)
     * #define ts_Nslist_printk   slist_stk(ts_,NLISTS,slist_printk)
     * #define ts_Nslist_head     slist_stk(ts_,NLISTS,slist_head)
     * #define ts_Nslist_init     slist_stk(ts_,NLISTS,slist_init)
     * #define ts_Nslist_last     slist_stk(ts_,NLISTS,slist_last)
     * #define ts_Nslist_move     slist_stk(ts_,NLISTS,slist_move)
     * #define ts_Nslist_next     slist_stk(ts_,NLISTS,slist_next)
     * #define ts_Nslist_prep     slist_stk(ts_,NLISTS,slist_prep)
     * #define ts_Nslist_prev     slist_stk(ts_,NLISTS,slist_prev)
     * #define ts_Nsnode          slist_stk(ts_,NLISTS,snode)
     * #define ts_Nsnode_clean    slist_stk(ts_,NLISTS,snode_clean)
     * #define ts_Nsnode_nozero   slist_stk(ts_,NLISTS,snode_nozero)
     * #define ts_Nsnode_release  slist_stk(ts_,NLISTS,snode_release)
     * #define ts_Nsnode_printk   slist_stk(ts_,NLISTS,snode_printk)
     * #define ts_Nsnode_init     slist_stk(ts_,NLISTS,snode_init)
     * #define ts_Nsnode_next     slist_stk(ts_,NLISTS,snode_next)
     * #define ts_Nsnode_ref      slist_stk(ts_,NLISTS,snode_ref)
     * #define ts_Nsnode_unref    slist_stk(ts_,NLISTS,snode_unref)
     * #define ts_Nsnode_unref_nd slist_stk(ts_,NLISTS,snode_unref_nd)
     *
     * ts_msnode_implement;
     * ts_mslist_implement;
     *
     *
     *  For cleanup of the #defined name space:
     *
     * #undef NLISTS
     * #undef s_Nslist
     * #undef s_Nsnode
     * #undef ts_Nslist
     * #undef ts_Nslist_append
     * #undef ts_Nslist_nozero
     * #undef ts_Nslist_clean
     * #undef ts_Nslist_del
     * #undef ts_Nslist_del_fast
     * #undef ts_Nslist_release
     * #undef ts_Nslist_printk
     * #undef ts_Nslist_init
     * #undef ts_Nslist_last
     * #undef ts_Nslist_move
     * #undef ts_Nslist_next
     * #undef ts_Nslist_prep
     * #undef ts_Nslist_prev
     * #undef ts_Nsnode
     * #undef ts_Nsnode_clean
     * #undef ts_Nsnode_nozero
     * #undef ts_Nsnode_release
     * #undef ts_Nsnode_printk
     * #undef ts_Nsnode_init
     * #undef ts_Nsnode_next
     * #undef ts_Nsnode_ref
     * #undef ts_Nsnode_unref
     * #undef ts_Nsnode_unref_nd
     *
     *
     *  See usb.h and usbd-device-list.c for an example implementation. 
     *
     *  The generic prototypes for the functions are:
     *
     * ts_Nsnode* ts_Nsnode_init (tf_dtor)
     * void ts_Nsnode_release (ts_Nsnode*)
     * void ts_Nsnode_printk (const ts_Nsnode*)
     * ts_Nsnode* ts_Nsnode_next (ts_Nsnode*, size_t)
     * void ts_Nsnode_nozero (ts_Nsnode*)
     * void ts_Nsnode_clean (ts_Nsnode*)
     * ts_Nsnode* ts_Nsnode_ref (ts_Nsnode*)
     * int ts_Nsnode_unref (ts_Nsnode*, int*)
     * int ts_Nsnode_unref_nd (ts_Nsnode*)
     *
     * void ts_Nslist_nozero (ts_Nslist* Nslist)
     * void ts_Nslist_clean (ts_Nslist* Nslist)
     * ts_Nsnode* ts_Nslist_head (ts_Nslist*)
     * ts_Nsnode* ts_Nsnode_next (ts_Nslist*, ts_Nsnode*)
     * ts_Nslist* ts_Nslist_init (void)
     * void ts_Nslist_release (ts_Nslist*)
     * void ts_Nslist_printk (const ts_Nslist*)
     * int ts_Nslist_del (ts_Nslist*, ts_Nsnode*)
     * int ts_Nslist_del_fast (ts_Nslist*, ts_Nsnode*, ts_Nsnode*)
     * int ts_Nslist_prep (ts_Nslist*, ts_Nsnode*)
     * int ts_Nslist_last (ts_Nslist*)
     * int ts_Nslist_append (ts_Nslist*, ts_Nsnode*)
     * int ts_Nslist_prev (ts_Nslist*, ts_Nsnode*)
     *
     *  The last parameter, unsigned, is the list number. Remember you
     * have to use always the same list number or you'll corrupt it.
     */

#define __slist_stk(a,b,c)  a##b##c
#define slist_stk(a,b,c) __slist_stk(a,b,c)

#define ts_msnode_declare                                       \
typedef struct s_Nsnode ts_Nsnode;                              \
struct s_Nsnode                                                 \
{                                                               \
    ts_dtor   dtor;                                             \
    atomic_t  usecnt;                                           \
    ts_Nsnode *next[NLISTS];                                    \
};

#define ts_msnode_declare_functions                                 \
extern ts_Nsnode* ts_Nsnode_init (tf_dtor);                         \
extern void ts_Nsnode_release (ts_Nsnode*);                         \
extern void ts_Nsnode_printk (const ts_Nsnode*);                    \
                                                                    \
static inline                                                       \
ts_Nsnode* ts_Nsnode_next (ts_Nsnode* msnode, size_t nlist)         \
{                                                                   \
    return msnode->next[nlist];                                     \
}                                                                   \
                                                                    \
static inline                                                       \
void ts_Nsnode_nozero (ts_Nsnode* Nsnode)                           \
{                                                                   \
    ts_dtor_nozero (&(Nsnode)->dtor);                               \
    return;                                                         \
}                                                                   \
                                                                    \
static inline                                                       \
void ts_Nsnode_clean (ts_Nsnode* Nsnode)                            \
{                                                                   \
    memset (Nsnode, 0,  sizeof (*(Nsnode)));                        \
    ts_Nsnode_nozero (Nsnode);                                      \
    return;                                                         \
}                                                                   \
                                                                    \
static inline                                                       \
ts_Nsnode* ts_Nsnode_ref (ts_Nsnode* Nsnode)                        \
{                                                                   \
    atomic_inc (&Nsnode->usecnt);                                   \
    return Nsnode;                                                  \
}                                                                   \
                                                                    \
static inline                                                       \
int ts_Nsnode_unref (ts_Nsnode* Nsnode, int* perrcode)              \
{                                                                   \
    if (atomic_dec_and_test (&Nsnode->usecnt))                      \
    {                                                               \
        int errcode;                                                \
                                                                    \
        errcode = ts_dtor_exec (&Nsnode->dtor);                     \
        if (perrcode)                                               \
            *perrcode = errcode;                                    \
        ts_slist_mfree (Nsnode);                                    \
        return 0;                                                   \
    }                                                               \
    return atomic_read (&Nsnode->usecnt);                           \
}                                                                   \
                                                                    \
static inline                                                       \
int ts_Nsnode_unref_nd (ts_Nsnode* Nsnode)                          \
{                                                                   \
    atomic_dec (&Nsnode->usecnt);                                   \
    return atomic_read (&Nsnode->usecnt);                           \
}


#define ts_msnode_implement                                             \
ts_Nsnode* ts_Nsnode_init (tf_dtor dtor)                                \
{                                                                       \
    ts_Nsnode* Nsnode;                                                  \
                                                                        \
    gdlf_fntrace ("ts_" gdlf_stringfy (NLISTS)                          \
                  "snode_init (0x%p)", dtor);                           \
                                                                        \
    Nsnode = ts_slist_malloc (sizeof (ts_Nsnode));                      \
    if (Nsnode == NULL)                                                 \
    {                                                                   \
        gdlf_err ("Couldn't allocate a ts_" gdlf_stringfy (NLISTS)      \
                  "snode structure\n");                                 \
        return NULL;                                                    \
    }                                                                   \
    ts_Nsnode_clean (Nsnode);                                           \
    ts_dtor_fill (&Nsnode->dtor, dtor);                                 \
    return Nsnode;                                                      \
}                                                                       \
                                                                        \
void ts_Nsnode_release (ts_Nsnode* Nsnode)                              \
{                                                                       \
    gdlf_fntrace ("ts_" gdlf_stringfy (NLISTS)                          \
                  "snode_release (0x%p)", Nsnode);                      \
    gdlf_aassert (Nsnode != NULL, return);                              \
    gdlf_passert (atomic_read (&Nsnode->usecnt) == 0,                   \
                  "Aieee! Freeing an " gdlf_stringfy (NLISTS)           \
                  "snode with use count %d!\n",                         \
                  atomic_read (&Nsnode->usecnt));                       \
                                                                        \
    gdlf_assert_act (                                                   \
        int cnt;                                                        \
        for (cnt = 0; cnt < NLISTS; cnt++)                              \
            gdlf_passert (Nsnode->next[cnt] == NULL,                    \
                          "Aieee! Freeing an " gdlf_stringfy(NLISTS)    \
                          "snode with `next[%u]' = 0x%p!\n",            \
                          cnt, Nsnode->next[cnt]);                      \
        );                                                              \
    return;                                                             \
}                                                                       \
                                                                        \
void ts_Nsnode_printk (const ts_Nsnode* Nsnode)                         \
{                                                                       \
    int cnt;                                                            \
    gdlf_fntrace ("ts_" gdlf_stringfy (NLISTS)                          \
                  "snode_printk (0x%p)", Nsnode);                       \
    gdlf_aassert (Nsnode != NULL, return);                              \
                                                                        \
    gdlf_debug ("ts_Nsnode: 0x%p\n", Nsnode);                           \
    gdlf_debug ("  dtor:    0x%p\n", Nsnode->dtor.fn);                  \
    gdlf_debug ("  usecnt:  %d\n", atomic_read (&Nsnode->usecnt));      \
    for (cnt = 0; cnt < NLISTS; cnt++)                                  \
         gdlf_debug ("  next[%d]: 0x%p\n", cnt, Nsnode->next[cnt]);     \
    return;                                                             \
}

    /* 
     * Multiple singly-linked lists
     * ----------------------------
     *
     *  Note that the list structure (ts_Nslist) has just one header,
     * not multiple as it could be done. We want MULTIPLE lists that
     * share SAME nodes. If the list had the same number of heads,
     * we'd be doing nothing useful :)
     *
     *  Also, a list must use always the same list number for it to
     * work properly.
     */

#define ts_mslist_declare                                       \
typedef struct s_Nslist ts_Nslist;                              \
struct s_Nslist                                                 \
{                                                               \
    atomic_t    usecnt;                                         \
    size_t      nlist;                                          \
    size_t      elements;                                       \
    ts_lock     lock;                                           \
    ts_Nsnode   *head;                                          \
};

#define ts_mslist_declare_functions                             \
static inline                                                   \
void ts_Nslist_nozero (ts_Nslist* Nslist)                       \
{                                                               \
}                                                               \
static inline                                                   \
void ts_Nslist_clean (ts_Nslist* Nslist)                        \
{                                                               \
    memset (Nslist, 0, sizeof (*(Nslist)));                     \
    ts_Nslist_nozero (Nslist);                                  \
}                                                               \
static inline                                                   \
ts_Nsnode* ts_Nslist_head (ts_Nslist* Nslist)                   \
{                                                               \
    return Nslist->head;                                        \
}                                                               \
static inline                                                   \
ts_Nsnode* ts_Nslist_next (ts_Nslist* Nslist, ts_Nsnode* msnode)\
{                                                               \
    return msnode->next[Nslist->nlist];                         \
}                                                               \
extern ts_Nsnode* ts_Nslist_prev (ts_Nslist*, ts_Nsnode*);      \
extern ts_Nsnode* ts_Nslist_last (ts_Nslist*);                  \
extern ts_Nslist* ts_Nslist_init (size_t);                      \
extern void ts_Nslist_release (ts_Nslist*);                     \
extern void ts_Nslist_printk (const ts_Nslist*);                \
extern int ts_Nslist_prep (ts_Nslist*, ts_Nsnode*);             \
extern int ts_Nslist_append (ts_Nslist*, ts_Nsnode*);           \
extern int ts_Nslist_append (ts_Nslist*, ts_Nsnode*);           \
extern int ts_Nslist_del (ts_Nslist*, ts_Nsnode*);              \
extern int ts_Nslist_del_fast (ts_Nslist*, ts_Nsnode*, ts_Nsnode*);


#define ts_mslist_implement                                                               \
ts_Nslist* ts_Nslist_init (size_t nlist)                                                  \
{                                                                                         \
    ts_Nslist* Nslist;                                                                    \
                                                                                          \
    gdlf_fntrace ("ts_" gdlf_stringfy (NLISTS)                                            \
                  "slist_init (%u)", nlist);                                              \
                                                                                          \
    gdlf_apassert (nlist < NLISTS,                                                        \
                   return NULL,                                                           \
                   "Index %d out of range for "                                           \
                   "ts_" gdlf_stringfy (NLISTS) "slist!\n", nlist);                       \
                                                                                          \
    Nslist = ts_slist_malloc (sizeof (ts_Nslist));                                        \
    if (Nslist == NULL)                                                                   \
    {                                                                                     \
        gdlf_err ("Couldn't allocate a ts_" gdlf_stringfy (NLISTS)                        \
                  "slist structure\n");                                                   \
        return NULL;                                                                      \
    }                                                                                     \
    ts_Nslist_clean (Nslist);                                                             \
    Nslist->nlist = nlist;                                                                \
    return Nslist;                                                                        \
}                                                                                         \
                                                                                          \
void ts_Nslist_release (ts_Nslist* Nslist)                                                \
{                                                                                         \
    int count;                                                                            \
    ts_Nsnode *itr, *next;                                                                \
                                                                                          \
    gdlf_fntrace (gdlf_stringfy (ts_Nslist_release) " (0x%p)",                            \
                  Nslist);                                                                \
    gdlf_aassert (Nslist != NULL, return);                                                \
                                                                                          \
    ts_slist_write_lock (&Nslist->lock);                                                  \
    itr = Nslist->head;                                                                   \
    Nslist->head = NULL;                                                                  \
    while (itr)                                                                           \
    {                                                                                     \
        next = itr->next[Nslist->nlist];                                                  \
        itr->next[Nslist->nlist] = NULL;                                                  \
        ts_slist_write_lock (&itr->lock);                                                 \
        count = atomic_read (&itr->usecnt);                                               \
        ts_Nsnode_unref (itr, NULL);                                                      \
        if (count != 1)                                                                   \
            ts_slist_write_unlock (&itr->lock);                                           \
        itr = next;                                                                       \
    }                                                                                     \
    ts_slist_write_unlock (&Nslist->lock);                                                \
    ts_slist_clean (Nslist);                                                              \
    return;                                                                               \
}                                                                                         \
                                                                                          \
void ts_Nslist_printk (const ts_Nslist* Nslist)                                           \
{                                                                                         \
    gdlf_fntrace (gdlf_stringfy (ts_Nslist_printk) " (0x%p)", Nslist);                    \
    gdlf_aassert (Nslist != NULL, return);                                                \
                                                                                          \
    gdlf_debug (gdlf_stringfy (ts_Nslist_printk) " 0x%p\n", Nslist);                      \
    gdlf_debug ("  usecnt:    %d\n", atomic_read (&Nslist->usecnt));                      \
    gdlf_debug ("  nlist:     %d\n", Nslist->nlist);                                      \
    gdlf_debug ("  elements:  %u\n", Nslist->elements);                                   \
    gdlf_debug ("  head:      0x%p\n", Nslist->head);                                     \
    return;                                                                               \
}                                                                                         \
                                                                                          \
int ts_Nslist_del (ts_Nslist* Nslist, ts_Nsnode* Nsnode)                                  \
{                                                                                         \
    int errcode;                                                                          \
    ts_Nsnode *prev;                                                                      \
                                                                                          \
    gdlf_fntrace (gdlf_stringfy (ts_Nslist_del)                                           \
                  " (0x%p, 0x%p)", Nslist, Nsnode);                                       \
                                                                                          \
    gdlf_apassert (Nslist != NULL,                                                        \
                   return -EINVAL,                                                        \
                   "Tried to remove from a NULL "                                         \
                   gdlf_stringfy (ts_Nslist) "\n");                                       \
                                                                                          \
    gdlf_apassert (Nsnode != NULL,                                                        \
                   return -EINVAL,                                                        \
                   "Tried to remove a NULL element from "                                 \
                   gdlf_stringfy (ts_Nslist) " 0x%p\n",                                   \
                   Nslist);                                                               \
                                                                                          \
    ts_slist_write_lock (&Nslist->lock);                                                  \
    prev = ts_Nslist_prev (Nslist, Nsnode);                                               \
    if (prev == NULL)                                                                     \
        errcode = -ENOENT;                                                                \
    else                                                                                  \
        errcode = ts_Nslist_del_fast (Nslist, Nsnode, prev);                              \
    ts_slist_write_unlock (&slist->lock);                                                 \
    return errcode;                                                                       \
}                                                                                         \
                                                                                          \
int ts_Nslist_del_fast (ts_Nslist* Nslist, ts_Nsnode* Nsnode,                             \
                        ts_Nsnode* prev)                                                  \
{                                                                                         \
    int count;                                                                            \
                                                                                          \
    gdlf_fntrace ("ts_" gdlf_stringfy (NLISTS)                                            \
                  "slist_del_fast (0x%p, 0x%p)", Nslist, Nsnode);                         \
                                                                                          \
    gdlf_apassert (Nslist != NULL,                                                        \
                   return -EINVAL,                                                        \
                   "Tried to remove from a NULL ts_"                                      \
                   gdlf_stringfy (NLISTS) "slist\n");                                     \
                                                                                          \
    gdlf_apassert (Nsnode != NULL,                                                        \
                   return -EINVAL,                                                        \
                   "Tried to remove a NULL element from "                                 \
                   "ts_" gdlf_stringfy (NLISTS) "slist 0x%p\n",                           \
                   Nslist);                                                               \
                                                                                          \
    gdlf_apassert (prev != NULL,                                                          \
                   return -EINVAL,                                                        \
                   "Tried to remove an element with NULL previous from "                  \
                   "ts_" gdlf_stringfy (NLISTS) "slist 0x%p\n",                           \
                   Nslist);                                                               \
                                                                                          \
    ts_slist_write_lock (&Nsnode->lock);                                                  \
    ts_slist_write_lock (&prev->lock);                                                    \
    if (prev == Nsnode)                                                                   \
        Nslist->head = Nsnode->next[Nslist->nlist];                                       \
    else                                                                                  \
    {                                                                                     \
        gdlf_apassert (prev->next[Nslist->nlist] == Nsnode,                               \
                       {                                                                  \
                           ts_slist_write_unlock (&prev->lock);                           \
                           ts_slist_write_unlock (&Nsnode->lock);                         \
                           return -EINVAL;                                                \
                       },                                                                 \
                       "Element 0x%p is not the previous element "                        \
                       "to 0x%p on ts_Nslist 0x%p\n",                                     \
                       prev, Nsnode, Nslist);                                             \
        prev->next[Nslist->nlist] = Nsnode->next[Nslist->nlist];                          \
    }                                                                                     \
    Nslist->elements--;                                                                   \
    Nsnode->next[Nslist->nlist] = NULL;                                                   \
    count = atomic_read (&Nsnode->usecnt);                                                \
    ts_Nsnode_unref (Nsnode, NULL);                                                       \
    ts_slist_write_unlock (&prev->lock);                                                  \
    if (count != 1)                                                                       \
        ts_slist_write_unlock (&Nsnode->lock);                                            \
    return 0;                                                                             \
}                                                                                         \
                                                                                          \
int ts_Nslist_prep (ts_Nslist* Nslist, ts_Nsnode* Nsnode)                                 \
{                                                                                         \
    gdlf_fntrace (gdlf_stringfy (ts_Nslist_prep)                                          \
                  " (0x%p, 0x%p)", Nslist, Nsnode);                                       \
                                                                                          \
    gdlf_apassert (Nslist != NULL,                                                        \
                   return -EINVAL,                                                        \
                   "Tried to prepend to a NULL "                                          \
                   gdlf_stringfy (ts_Nslist) "\n");                                       \
                                                                                          \
    gdlf_apassert (Nsnode != NULL,                                                        \
                   return -EINVAL,                                                        \
                   "Tried to prepend a NULL element to "                                  \
                   gdlf_stringfy (ts_Nslist) " 0x%p\n",                                   \
                   Nslist);                                                               \
                                                                                          \
    gdlf_apassert (Nsnode->next[Nslist->nlist] == NULL,                                   \
                   return -EBUSY,                                                         \
                   "Tried to prepend to " gdlf_stringfy (ts_Nslist)                       \
                   " 0x%p the element 0x%p whose `next[%u]' is already "                  \
                   "set to 0x%p\n",                                                       \
                   Nslist, Nsnode, Nslist->nlist,                                         \
                   Nsnode->next[Nslist->nlist]);                                          \
                                                                                          \
    ts_slist_write_lock (&Nslist->lock);                                                  \
    ts_slist_write_lock (&Nsnode->lock);                                                  \
    Nsnode->next[Nslist->nlist] = Nslist->head;                                           \
    ts_Nsnode_ref (Nsnode);                                                               \
    Nslist->head = Nsnode;                                                                \
    Nslist->elements++;                                                                   \
    ts_slist_write_unlock (&Nsnode->lock);                                                \
    ts_slist_write_unlock (&Nslist->lock);                                                \
    return 0;                                                                             \
}                                                                                         \
                                                                                          \
ts_Nsnode* ts_Nslist_prev (ts_Nslist* Nslist, ts_Nsnode* Nsnode)                          \
{                                                                                         \
    ts_Nsnode *itr;                                                                       \
    size_t nlist;                                                                         \
                                                                                          \
    gdlf_fntrace (gdlf_stringfy (ts_Nslist_prev)                                          \
                  " (0x%p, 0x%p)", Nslist, Nsnode);                                       \
                                                                                          \
    gdlf_apassert (Nslist != NULL,                                                        \
                   return NULL,                                                           \
                   "Tried to search previous in a NULL "                                  \
                   gdlf_stringfy (ts_Nslist) "\n");                                       \
                                                                                          \
    read_lock (&Nslist->lock);                                                            \
    itr = Nslist->head;                                                                   \
    if (itr == Nsnode)                                                                    \
    {                                                                                     \
        read_unlock (&Nslist->lock);                                                      \
        return itr;                                                                       \
    }                                                                                     \
    nlist = Nslist->nlist;                                                                \
    while (itr)     /* while there are elements to search */                              \
    {                                                                                     \
        if (itr->next[nlist] == Nsnode) /* if it is the next ... */                       \
            break;                      /* ... we're done */                              \
        itr = itr->next[nlist];         /* it isn't, get followin'*/                      \
    }                                                                                     \
    read_unlock (&Nslist->lock);                                                          \
                                                                                          \
    return itr;                                                                           \
}                                                                                         \
                                                                                          \
ts_Nsnode* ts_Nslist_last (ts_Nslist* Nslist)                                             \
{                                                                                         \
    ts_Nsnode *itr;                                                                       \
                                                                                          \
    gdlf_fntrace (gdlf_stringfy (ts_Nslist_last) " (0x%p)",                               \
                  Nslist);                                                                \
                                                                                          \
    read_lock (&Nslist->lock);                                                            \
    itr = Nslist->head;                                                                   \
    if (itr == NULL)        /* empty list? */                                             \
    {                                                                                     \
        read_unlock (&Nslist->lock);                                                      \
        return NULL;                                                                      \
    }                                                                                     \
    while (itr->next[Nslist->nlist])                                                      \
        itr = itr->next[Nslist->nlist];                                                   \
    read_unlock (&Nslist->lock);                                                          \
    return itr;                                                                           \
}                                                                                         \
                                                                                          \
int ts_Nslist_append (ts_Nslist* Nslist, ts_Nsnode* Nsnode)                               \
{                                                                                         \
    ts_Nsnode *last;                                                                      \
                                                                                          \
    gdlf_fntrace (gdlf_stringfy (ts_Nslist_append) " (0x%p, 0x%p)",                       \
                  Nslist, Nsnode);                                                        \
                                                                                          \
    gdlf_apassert (Nslist != NULL,                                                        \
                   return -EINVAL,                                                        \
                   "Cannot append to a NULL "                                             \
                   gdlf_stringfy (ts_Nslist) "\n");                                       \
                                                                                          \
    gdlf_apassert (Nsnode != NULL,                                                        \
                   return -EINVAL,                                                        \
                   "Cannot append a NULL " gdlf_stringfy (ts_Nsnode)                      \
                   " to " gdlf_stringfy (ts_Nslist) " 0x%p\n", Nslist);                   \
                                                                                          \
    ts_slist_write_lock (&Nslist->lock);                                                  \
    ts_slist_write_lock (&Nsnode->lock);                                                  \
                                                                                          \
    last = ts_Nslist_last (Nslist);                                                       \
    if (last == NULL)                                                                     \
        Nslist->head = Nsnode;                                                            \
    else                                                                                  \
        last->next[Nslist->nlist] = Nsnode;                                               \
    ts_Nsnode_ref (Nsnode);                                                               \
    Nsnode->next[Nslist->nlist] = NULL;                                                   \
    Nslist->elements++;                                                                   \
                                                                                          \
    ts_slist_write_unlock (&Nsnode->lock);                                                \
    ts_slist_write_unlock (&Nslist->lock);                                                \
    return 0;                                                                             \
}                                                                                         \
                                                                                          \
int ts_Nslist_move (ts_Nslist* Nslist, ts_Nsnode* Nsnode, ts_Nsnode* prev)                \
{                                                                                         \
    ts_Nsnode* Nsnode_prev;                                                               \
                                                                                          \
    gdlf_fntrace (gdlf_stringfy (ts_Nslist_move) " (0x%p, 0x%p, 0x%p)",                   \
                  Nslist, Nsnode, prev);                                                  \
                                                                                          \
    gdlf_aassert (Nslist != NULL, return -EINVAL);                                        \
    gdlf_aassert (Nsnode != NULL, return -EINVAL);                                        \
    gdlf_aassert (prev != Nsnode, return -EINVAL);                                        \
                                                                                          \
        /* moving to the same place? */                                                   \
                                                                                          \
    if (Nsnode == prev)                                                                   \
        return -EMLINK; /* would cause circular link */                                   \
                                                                                          \
        /* both in list? */                                                               \
                                                                                          \
    Nsnode_prev = ts_Nslist_prev (Nslist, Nsnode);                                        \
    gdlf_aassert (Nsnode_prev != NULL, return -EINVAL);                                   \
    gdlf_aassert (prev == NULL || ts_Nslist_prev (Nslist, prev) != NULL, return -EINVAL); \
                                                                                          \
        /* no change in position? */                                                      \
                                                                                          \
    if ((Nsnode_prev == prev)    /* in the middle of the list */                          \
        || (prev == NULL        /* or at the beginning */                                 \
            && Nslist->head == Nsnode))                                                   \
        return 0;                                                                         \
                                                                                          \
        /* Ok, we can proceed, let's go */                                                \
                                                                                          \
    ts_slist_write_lock (&Nslist->lock);                                                  \
    ts_slist_write_lock (&Nsnode->lock);                                                  \
                                                                                          \
        /* Unlink Nsnode */                                                               \
                                                                                          \
    if (Nsnode_prev == Nsnode)    /* First one? */                                        \
        Nslist->head = Nsnode->next[Nslist->nlist];                                       \
    else                                                                                  \
        Nsnode_prev->next[Nslist->nlist] = Nsnode->next[Nslist->nlist];                     \
                                                                                          \
        /* Relink it */                                                                   \
                                                                                          \
    if (prev == NULL)       /* At beginning of list */                                    \
    {                                                                                     \
        Nsnode->next[Nslist->nlist] = Nslist->head;                                        \
        Nslist->head = Nsnode;                                                            \
    }                                                                                     \
    else                    /* in the middle/end of the list */                           \
    {                                                                                     \
        Nsnode->next[Nslist->nlist] = prev->next[Nslist->nlist];                            \
        prev->next[Nslist->nlist] = Nsnode;                                                \
    }                                                                                     \
                                                                                          \
    ts_slist_write_unlock (&Nsnode->lock);                                                \
    ts_slist_write_unlock (&Nslist->lock);                                                \
    return 0;                                                                             \
}

#include <uusbd/gdlf-undef.h> /* clean */

#endif /* _usb_slist_h_ */

/* Log:
 *
 * $Log: usb-slist.h,v $
 * Revision 1.2  1998/11/16 18:45:07  inaky
 * Fixed ts_Nslist_move(), as it was, like ts_slist_move(), blatantly
 * broken.
 *
 * Revision 1.1  1998/11/14 00:38:27  inaky
 * Isolated all the header files
 *
 * Revision 1.27  1998/09/24 15:52:35  inaky
 * Finished the split of _clean into _nozero and _clean.
 *
 * Revision 1.26  1998/08/30 13:46:12  inaky
 * Made all ts_*_release() return void. Severed assertions.
 *
 * Revision 1.25  1998/08/27 11:08:42  inaky
 * Moved ts_*_free() to ts_*_release(), made them return void and severed
 * assertions.
 *
 * Revision 1.24  1998/08/25 04:00:26  inaky
 * Fixed ts_snode_printk() to take a const pointer as argument.
 *
 * Revision 1.23  1998/08/24 02:21:46  inaky
 * Added print functions, moved to return void some functions which
 * didn't need to return anything.
 *
 * Revision 1.22  1998/08/20 23:40:53  inaky
 * Synched both codes (Nslist was getting quite obsolete).
 *
 * Revision 1.21  1998/08/03 13:27:30  inaky
 * Fixed the order of cleaning in the ts_Nsnode_clean function to firstly
 * initialize to zero and then call specific clean methods.
 *
 * Revision 1.20  1998/07/30 17:51:59  inaky
 * Fixed function traces on ts_Nslist_del_fast() and toggled assertion.
 *
 * Revision 1.19  1998/07/23 14:21:52  inaky
 * Added a functio to move elements from one position to another into a
 * list.
 *
 * Revision 1.18  1998/07/14 21:55:58  inaky
 * Fixed usecount messages to say if they're incrementing or
 * decrementing.
 *
 * Revision 1.17  1998/07/12 22:48:36  inaky
 * Added the non-destructing unreference method, to allow keeping objects
 * alive during destruction.
 *
 * Revision 1.16  1998/06/21 17:19:59  inaky
 * Reverted the ts_Nsnode_[un]ref() function removal. Now these are still
 * available and must be used. They enhance type checking.
 *
 * Revision 1.15  1998/06/21 15:10:26  inaky
 * Made ts_snode and ts_Nsnode identical at the top, so we can operate on
 * usecount and destructor means identically for both. Useful for the
 * ts_usb_drv_use list, so it can store anything that has a destructor
 * and a usecount.
 *
 * Revision 1.14  1998/06/20 21:54:51  inaky
 * Massive changes to split the old monolithic hubd driver into a hub
 * manager (hub wrappers and daemon) and a dynamic driver, which can be
 * replaced and its different parts reused.
 *
 * Revision 1.13  1998/04/13 21:58:14  inaky
 * Added a debug #define to get dumps on reference counts; added some
 * advice to use ts_*snode_[un]ref(). Added declarations for *_last() and
 * *_append() functions. Fixed *_ref() functions to return a pointer to
 * the snode being referenced. Added the new _last() and _append()
 * functions to the Nslist code.
 *
 * Revision 1.12  1998/02/12 14:31:19  inaky
 * Fixed the delete functions to return error if item to delete isn't on
 * the list instead of failing miserably.
 *
 * Revision 1.11  1998/02/10 17:29:37  inaky
 * Removed GDLF, it was cluttering usb.h and usbd.h. There's no gdlf
 * tracing on inlined or declared functions. Just on implemented
 * functions, as they're suppossed to be on a file. Added some printk's
 * to help that will be removed soon. Added ts_Nslist_prev() declaration,
 * which was missing, fixed ts_Nslist_free() to NULLify the head pointer
 * before starting to free. Removed a #warning which was illegal on
 * macros. Fixed craps.
 *
 * Revision 1.10  1998/02/10 15:08:51  inaky
 * Added doc, removed environment check for compiling outta UUSBD, added
 * GDLF for inlined functions. Added ts_[N]snode_[un]ref(); modified rest
 * of functions to behave as on slist.c.
 *
 * Revision 1.9  1998/02/09 13:23:30  inaky
 * Moved to atomic use counts
 * 468: fixed missing parentheses on function trace messages
 * 537: fixed the order of the warning message parameters, which was
 *      swapped.
 *
 * Revision 1.8  1998/01/08 12:19:09  inaky
 * Added copyright strings to all important files
 *
 * Revision 1.7  1998/01/02 00:12:16  inaky
 * Finished the new driver scheme. Documented at usbd-drv.c
 *
 * Revision 1.6  1997/12/31 16:37:11  inaky
 * Finished the HCI interface
 *
 * Revision 1.5  1997/12/31 15:42:06  inaky
 * Added search of previous elements in lists
 *
 * Revision 1.4  1997/12/30 22:30:17  inaky
 * Added the foundations of the HCD and fixed some bugs.
 *
 * Revision 1.3  1997/12/30 20:15:43  inaky
 * Moved everything to the new lists scheme. Untested yet.
 *
 * Revision 1.2  1997/12/30 13:50:43  inaky
 * Moved the Nslists to not need to accept a list number, as they can
 * keep their list number into the ts_Nlist structure. Thus, the list
 * number is only needed in ts_Nsnode_next(), but this one is for
 * optimization.
 *
 * Revision 1.1  1997/12/30 00:03:45  inaky
 * Adding a generic system to handle singly linked lists and the HCI
 * registration control system
 *
 */

