
/* USB Driver Stack v$Revision: 1.1 $ 
 *
 * (C) 1997 by Iaky Prez Gonzlez <inaky@peloncho.fis.ucm.es>
 *
 * Distributed under the terms of the GNU Public License Version 2
 * See files COPYRIGHT and COPYING at the root distribution directory
 *
 * $Id: usb-stddesc.h,v 1.1 1998/11/14 00:38:27 inaky Exp $
 *
 * Changelog at the bottom 
 *
 *
 *     Standard USB descriptors
 *     ------------------------
 *
 *  The idea here is to stick this into a memory area and be able of
 * accesing it directly to move it to C structs. FIXME: trouble with
 * memory alignments.
 */

#ifndef _usb_stddesc_h_
#define _usb_stddesc_h_

#include <uusbd/usb.h>


    /* 
     * Structure which is the head of every descriptor
     * -----------------------------------------------
     */

typedef struct
{
    __u8    bLength;            /* Size of this descriptor (in bytes) */
    __u8    bDescriptorType;    /* DEVICE descriptor type */
} ts_usb_desc_head;


    /* 
     * Device descriptor (USB Spec 1.0 9.6.1) 
     * --------------------------------------
     */

typedef struct
{
    ts_usb_desc_head head;
    __u16   bcdUSB;             /* USB spec release (BCD coded) */
    __u8    bDeviceClass;       /* USB Class code (note the Spec Description) */
    __u8    bDeviceSubClass;    /* USB Subclass code (note the Spec Description) */
    __u8    bDeviceProtocol;    /* USB Protocol code (note the Spec Description) */
    __u8    bMaxPacketSize;     /* Maximum packet size for endpoint 0 */
    __u16   idVendor;           /* USB vendor code */
    __u16   idProduct;          /* Mfrer. Product code */
    __u16   bcdDevice;          /* BCD coded device release number */
    __u8    iManufacturer;      /* Index of string descriptor describing mfrer. */
    __u8    iProduct;           /* Index of string descriptor describing product */
    __u8    iSerialNumber;      /* Index of string descriptor describing serial no. */
    __u8    bNumConfigurations; /* Number of posible configurations */
} ts_usb_dev_descr;

    /* Length in bytes of a device descriptor following the 1.0 spec */

#define USB_DEV_DESCR_LEN_1_0 18


    /* 
     * Configuration descriptor (USB Spec 1.0 9.6.2) 
     * ---------------------------------------------
     */

typedef struct
{
    ts_usb_desc_head head;
    __u16   wTotalLength;       /* Total length of data returned for this configuration (cfg+iface+enpd+class/vendor descriptors) */
    __u8    bNumInterfaces;     /* Number of interfaces supported by this configuration */
    __u8    bConfigurationValue;/* Value to use as an argument to "Set Configuration" to select this configuration */
    __u8    iConfiguration;     /* Index of string descriptor describing this configuration */
    __u8    bmAttributes;       /* Configuration characteristics (see Spec Description) */
    __u8    MaxPower;           /* Max power consumption for this device in 2mA units (see Spec Description) */
} ts_usb_cfg_descr;

#define USB_CFG_DESCR_LEN_1_0 9

    /* Masks for descriptor decoding */

#define USB_CFG_DESCR_ATTR_BUS_PWR      0x80    /* bus powered */
#define USB_CFG_DESCR_ATTR_SELF_PWR     0x40    /* self powered */
#define USB_CFG_DESCR_ATTR_REM_WAKEUP   0x20    /* supports rem wake up */


    /* 
     * Interface descriptor (USB Spec 1.0 9.6.3) 
     * -----------------------------------------
     */

typedef struct
{
    ts_usb_desc_head head;
    __u8    bInterfaceNumber;   /* Zero-based number of interface in this configuration */
    __u8    bAlternateSetting;  /* Alternate setting for this interface */
    __u8    bNumEndpoints;      /* Number of endpoins used, excluding endpoint zero */
    __u8    bInterfaceClass;    /* USB class code (note the Spec Description) */
    __u8    bInterfaceSubClass; /* USB Subclass code (note the Spec Description) */
    __u8    bInterfaceProtocol; /* USB Protocol code (note the Spec Description) */
    __u8    iInterface;         /* Index of string descriptor describing this interface */
} ts_usb_iface_descr;

#define USB_IFACE_DESCR_LEN_1_0 9


    /* 
     * Endpoint descriptor (USB Spec 1.0 9.6.4) 
     * ----------------------------------------
     */

typedef struct
{
    ts_usb_desc_head head;
    __u8    bEndpointAddress;   /* Address of the endpoint in the USB device (see Spec Description) */
    __u8    bmAttributes;       /* Endpoint attributes (when selected/configured; see Spec Description) */
    __u16   wMaxPacketSize;     /* Max packet size this endp will send/receive (see Spec Description) */
    __u8    bInterval;          /* Interval for polling the endp for data transfes (in milliseconds) (see Spec Description) */
} ts_usb_endp_descr;

#define USB_ENDP_DESCR_LEN_1_0 7


    /* 
     * String descriptor (USB Spec 1.0 9.6.5)
     * --------------------------------------
     *
     * Though the bString member is an static string of the maximum
     * size, it doesn't have why to be that big. That's given by the
     * bLength member. bString is only the space for it to be stored.
     */

#define USB_STR_MAX 256         /* max size of a string in a descriptor */

typedef struct
{
    ts_usb_desc_head head;
    __u8    bString[USB_STR_MAX];/* UNICODE (UTF-8?) encoded string */
} ts_usb_str_descr;


    /* 
     * Standard request parameters  (USB Spec 1.0 9.3)
     * -----------------------------------------------
     */

typedef enum
{   
    e_usb_req_dir_h2d = 0,    /* Direction: host to device */
    e_usb_req_dir_d2h = 1     /* Direction: device to host */
} e_usb_req_dir;

typedef enum
{
    e_usb_req_type_std = 0,   /* Type: Standard */
    e_usb_req_type_class = 1, /*       Class specific */
    e_usb_req_type_vendor = 2 /*       Vendor specific */
} e_usb_req_type;

typedef enum
{
    e_usb_req_rec_dev = 0,    /* Destination: Device */
    e_usb_req_rec_iface = 1,  /*              Interface */
    e_usb_req_rec_endp = 2,   /*              Endpoint */
    e_usb_req_rec_other = 3   /*              Other */
} e_usb_req_rec;


    /*
     * Standard request types (USB Spec 1.0 9.4) 
     * -----------------------------------------
     */

typedef enum
{
    e_usb_req_get_status = 0,     /* Return status for the specified recipient (see USB Spec 1.0 9.4.5) */
    e_usb_req_clear_feature,      /* Clear/disable an specific feature */
    e_usb_req_reserved_1,
    e_usb_req_set_feature,        /* Set/enable an specific feature */
    e_usb_req_reserved_2,
    e_usb_req_set_address,        /* Set the device address */
    e_usb_req_get_descriptor,     /* Return the specified descriptor if it exists (note 3rd paragraph at page 177 USB Spec 1.0) */
    e_usb_req_set_descriptor,     /* Add/update descriptor (optional) */
    e_usb_req_get_configuration,  /* Return in the data phase the number of actual active configuration */
    e_usb_req_set_configuration,  /* Set the device configuration */
    e_usb_req_get_interface,      /* Return the currently selected alternate setting for the given interface */
    e_usb_req_set_interface,      /* Select an alternate setting for an interface */
    e_usb_req_synch_frame         /* Set'n'report and endpoint synchronization's frame */
} e_usb_req;


    /*
     * Standard descriptor types (USB Spec 1.0 9.4)
     * --------------------------------------------
     */

typedef enum
{
    e_usb_desc_dev = 1,   /* Device descriptor */
    e_usb_desc_cfg,       /* Configuration descriptor */
    e_usb_desc_str,       /* String descriptor */
    e_usb_desc_iface,     /* Interface descriptor */
    e_usb_desc_endp,      /* Endpoint descriptor */
    e_usb_desc_hub = 0x29 /* Hub descriptor (USB Spec 1.0 11.11.2.1 */
} e_usb_desc;


    /* 
     * Feature selectors 
     * -----------------
     */

typedef enum
{
    e_usb_fes_endp_stall = 0, /* endpoint stall */
    e_usb_fes_rmt_wake_up     /* Remote wake-up */
} e_usb_featsel;


    /* 
     * Status bitmaps
     * --------------
     *
     *  Anded to the status obtained from the usb_*_status_get()
     * calls, if 1, it indicates the item is in the state noted by the
     * comment by the right.
     */

#define USB_ENDP_STAT_STALL     0x0001    /* Endpoint stalled */
#define USB_DEV_STAT_SELF_PWR   0x0001    /* Device is self powered */
#define USB_DEV_STAT_REM_WAKEUP 0x0002    /* Dev may do remote wake ups */

    /*
     * Functions
     * ---------
     */

#ifdef __KERNEL__
#endif /* __KERNEL__ */

#endif /* _usb_stddesc_h_ */

/* Log:
 *
 * $Log: usb-stddesc.h,v $
 * Revision 1.1  1998/11/14 00:38:27  inaky
 * Isolated all the header files
 *
 * Revision 1.20  1998/08/24 15:47:22  inaky
 * Fixed little cosmetic stuff, removed old files.
 *
 * Revision 1.19  1998/08/24 02:22:00  inaky
 * Changed name of all enumerations to be prefixed with the enumeration
 * type name.
 *
 * Revision 1.18  1998/05/01 19:06:17  inaky
 * Fixed a typo error on interface descriptor declaration.
 *
 * Revision 1.17  1998/04/26 03:25:53  inaky
 * Added some masks for configuration, and some constants for stantard
 * USB length.
 *
 * Revision 1.16  1998/04/13 21:59:32  inaky
 * Fixed HUB descriptor type [copied to usb-hub.h]; added status bitmaps
 * [removed bitfields for status];
 *
 * Revision 1.15  1998/03/28 19:27:21  inaky
 * Added new type for common descriptor fields and reformatted some
 * comments to make them more noticeable.
 *
 * Revision 1.14  1998/01/11 15:44:16  inaky
 * Finished functions from the descriptor, inteface and endpoint modules,
 * fixed docs, pretty printed and removed craps.
 *
 * Revision 1.13  1998/01/08 12:19:10  inaky
 * Added copyright strings to all important files
 *
 * Revision 1.12  1997/12/27 15:07:42  inaky
 * Added wrappers for __KERNEL__ #define
 *
 * Revision 1.11  1997/12/22 02:44:30  inaky
 * Trying to move all the documentation to the source files from the
 * header files, as it was duplicated.
 *
 * Revision 1.10  1997/12/08 02:40:59  inaky
 * Fixed some stuff of the #includes, modified the makefile targets to be
 * more flexible, tried to finish the device configuration section.
 *
 * Revision 1.9  1997/11/29 20:21:48  inaky
 * Documented and almost fixes the read/write/abort API, documented a
 * damn lot and implemented some stuff
 *
 * Revision 1.8  1997/11/29 00:46:30  inaky
 * Documented almost all the hub basic API, added spin locks to
 * usb_host_register() and some more stuff. Going for more.
 *
 * Revision 1.7  1997/10/31 05:44:30  inaky
 * *** empty log message ***
 *
 * Revision 1.6  1997/06/19 20:38:17  inaky
 * Moving al hardware dependant stuff to the HCD modules
 *
 * Revision 1.5  1997/05/03 21:22:49  inaky
 * Adding a lot of code, with the only objetive of adding the first
 * transfer type: control :) Still there's a lot of stuff to do, mostly
 * concerning data structures handling.
 *
 * Revision 1.4  1997/04/20 21:25:57  inaky
 * Finished the spec for the Hub Class (now gotta implement it)
 * :). Started coding the Frame Lists code, hope UHCI support will be
 * finished in some sessions.
 * This code is breaking, I'm doing something bad with kmallocs. kFree()ing
 * something not kmalloc()ed.
 *
 * Revision 1.3  1997/04/06 21:14:32  inaky
 * Still with the definition of the access interface. More or less it is
 * taking an stable form.
 *
 * Revision 1.2  1997/04/02 20:01:27  inaky
 * Moved all the descriptor abreviature `desc' to `descr' (will use desc
 * for `file' descriptors).
 *
 * Revision 1.1  1997/03/31 20:24:46  inaky
 * Pfff ... a lot. Mostly added descriptors stuff (HUB class still is
 * missing a lot) and I think I've got a definitive idea of how to
 * implement the storing of configurations and interfaces.
 *
 * Revision 1.1  1997/02/22 22:25:44  inaky
 * *** empty log message ***
 *
 *
 */

