
/* USB Driver Stack v$Revision: 1.3 $
 *
 * (C) 1997 by Iaky Prez Gonzlez <inaky@peloncho.fis.ucm.es>
 *
 * Distributed under the terms of the GNU Public License Version 2
 * See files COPYRIGHT and COPYING at the root distribution directory
 *
 * $Id: usb.h,v 1.3 1998/11/16 00:01:02 inaky Exp $
 *
 * Changelog at the bottom
 *
 * NOTES:
 *
 *  1. All USB data reported from devices is reported in LITTLE
 *     ENDIAN format. However, you should forget this except when
 *     dealing with descriptors not handled by the UUSBD, as all the
 *     functions will translate for you if needed.
 *
 *  2. To control the amount of debugging information and checks, go
 *     to the gdlf.h header file, as it is the one who controls the
 *     assertions and debug/log printings.
 */

#ifndef _usb_h_
#define _usb_h_

#include <linux/sched.h>
#include <linux/mm.h>
#include <linux/kernel.h>       /* printk */
#include <linux/config.h>       /* Linux kernel configuration */
#include <asm/spinlock.h>       /* spinlock_t */
#include <asm/ptrace.h>         /* struct pt_regs */
#include <asm/atomic.h>         /* atomic_t, atomic_*() */
#include <stdarg.h>             /* Variadic arguments ... */

#include <uusbd/usb-dtor.h>     /* structure destructors */
#include <uusbd/usb-slist.h>    /* generic simple and multiple single linked lists */


    /* USB defines */

#define USB_MAINTAINER  "Inaky Perez Gonzalez <inaky@peloncho.fis.ucm.es>"
#define USB_CTL_STR     "usb"


    /* Known USB Class codes */

#define USB_CLASS_IFACE     0x00    /* Each interface has it's own class */
#define USB_CLASS_HUB       0x09    /* Hub class */
#define USB_CLASS_VENDOR    0xFF    /* Vendor specific class */

    /* Known USB Protocol codes */

#define USB_PROTOCOL_IFACE  0x00    /* Each interface has it's own protocol */
#define USB_PROTOCOL_VENDOR 0xFF    /* Vendor specific protocol */

    /* Known USB Protocol Revision */

#define USB_MAX_RELEASE     0x0100  /* Max release we know about, BCD coded */

    /* USB Packet IDentifications */

#define USB_PID_IN      0x69
#define USB_PID_OUT     0xe1
#define USB_PID_SETUP   0x2d


    /*
     * Miscellanea
     * -----------
     *
     * mostly configuration/compile time options.
     */

    /*
     * Max number of in and out endpoints (15*2), excluding endpoint 0
     */

#define USB_ENDPS               30
#define USB_DEVICES_MAX         126 /* Maximum number of devices/bus */
#define USB_CFEL_CHANGES_MAX    15  /* max cfel vicsp code changes */
#define USB_PORT_RESET_RETRIES  3   /* Maximum number of retries to reset a port */

    /* Defaults */

    /* FIXME: should adjust this ... */

#define USB_DEFAULT_TIMEOUT     3000    /* Timeout [ms] */
#define USB_TIMEOUT_HZ          1000    /* Timeout units/second [ms] */
#define USB_DEFAULT_MAX_RATIO   512     /* [transaction_errors/bytes*1024] */
#define USB_DEFAULT_MAX_ERRORS  4096    /* [transaction_errors] */


    /*
     * Hacks and some stuff
     * --------------------
     *
     * AFAIK, ssize_t is an int; this'll make sure on axp's
     */

#define SSIZE_MAX   LONG_MAX


    /*
     * USB errno codes
     * ---------------
     *
     * Like the ERRNO codes, when retuned to indicate an error,
     * they're negative.
     *
     * FIXME: add a usb_strerror() function to translate to `const char*'
     */

#define USBE_OK                   0
#define USBE_ADDR_NULL            1     /* NULL ts_usb_addr struct passed to usb_set_dst() */
#define USBE_ADDR_UNKNOWN         2     /* Unknown USB address */
#define USBE_ASET_BAD_INDEX       3     /* Invalid or out of range alternate setting number */
#define USBE_ASET_NULL            4     /* NULL pointer to alternate setting structure */
#define USBE_BAD_DEV_DESCR        5     /* Invalid device descriptor */
#define USBE_BAD_HOST_STRUCT      6     /* Bad host description structure */
#define USBE_BAD_MAGIC_NUMBER     7     /* Found an incorrect magic number */
#define USBE_BAD_PARAMETERS       8     /* Bad parameter passed to function */
#define USBE_BUFFER_BAD_PARAMS    9     /* Bad data buffer specification */
#define USBE_BUG                  10    /* Bug in code */
#define USBE_BUS_BABBLE           11    /* Babble on USB bus */
#define USBE_BUS_BITSTUFF         12    /* Bit Stuff error on USB bus */
#define USBE_BUS_CRCTO            13    /* CRC/Time out error on USB bus*/
#define USBE_BUS_NAK_RECEIVED     14    /* NAK received */
#define USBE_BUS_TIMEOUT          15    /* Time out error on USB bus */
#define USBE_CFEL_NULL            16    /* NULL pointer to configurable element structure */
#define USBE_CFG_BAD_INDEX        17    /* Invalid or out of range configuration number */
#define USBE_CFG_DESCR_BAD_TYPE   18    /* Bad descriptor type as confiruration descriptor */
#define USBE_CFG_DESCR_COMP_LEN   19    /* Could not compute lengths from configuration descriptor */
#define USBE_CFG_DESCR_SIZE       20    /* Bad length on configuration descriptor (or error reading it) */
#define USBE_CFG_NULL             21    /* NULL ts_usb_cfg struct passed to ts_usb_cfg_set_aset() */
    /* 22 unused */
#define USBE_DATA_BUFFER_ERROR    23    /* Error accessing data buffer from host controller [(over|under)run] */
#define USBE_DESC_NULL            24    /* NULL descriptor passed to function */
#define USBE_DEVICE_MASTER_BUSY   25    /* Bus mastering denied, as it is being used by other */
#define USBE_DEVICE_NOT_CFGED     26    /* Device isn't configured */
#define USBE_DEVICE_NOT_FOUND     27    /* Device isn't in the configured devices list */
#define USBE_DEVICE_NOT_MASTER    28    /* Device is not the bus master */
#define USBE_DEVICE_NULL          29    /* Pointer to device structure is NULL */
#define USBE_DEVICE_STAT_NULL     30    /* NULL ts_usb_dev_stat pointer */
#define USBE_DEVICE_UN_NOT_FOUND  31    /* Device isn't in the unconfigured devices list */
#define USBE_DEV_DESCR_BAD_TYPE   32    /* Bad descriptor type as device descriptor */
#define USBE_DEV_DESCR_SIZE       33    /* Bad length on device descriptor (or error reading it) */
#define USBE_DRIVER_BUSY          34    /* Driver is still referenced by someone but the drivers list */
#define USBE_DRIVER_CANT_REG      35    /* Cannot register driver */
#define USBE_DRIVER_LIST_ADD      36    /* Cannot add to driver's list */
#define USBE_DRIVER_NULL          37    /* NULL pointer to driver structure */
#define USBE_DRIVER_NULL_CONFIG   38    /* Driver doesn't have a configuration method */
#define USBE_DRIVER_NULL_UNCONFIG 39    /* Driver doesn't have an unconfiguration method */
#define USBE_DRIVER_UNREFERENCED  40    /* Driver isn't referenced by the drivers list */
#define USBE_DRIVER_USECOUNT      41    /* Driver structure is already in use! */
#define USBE_DRV_NOT_UNLOADED     42    /* Tried to register a driver which is already registered and loaded */
#define USBE_ENDP_BAD_INDEX       43    /* Invalid or out of range endpoint number */
#define USBE_ENDP_DESCR_BAD_TYPE  44    /* Bad descriptor type as endpoint descriptor */
#define USBE_ENDP_NOT_CNTL        45    /* Enpdpoint doesn't support control trasnfers */
#define USBE_ENDP_NO_SYNCH        46    /* Cannot get synch frame on non-isochronous endpoint */
#define USBE_ENDP_RDONLY          47    /* Endpoint is only readable */
#define USBE_ENDP_STALLED         48    /* Endpoint is stalled */
#define USBE_ENDP_STAT_NULL       49    /* NULL ts_usb_endp_stat pointer */
#define USBE_ENDP_WRONLY          50    /* Endpoint is only writable */
#define USBE_ERROR_COUNT          51    /* Error count for transfer exceeded */
#define USBE_ERROR_RATIO          52    /* Error/kbytes ratio for transfer exceeded */
#define USBE_ERROR_UNKNOWN        53    /* Unknown error */
#define USBE_HCI_BAD_TYPE         54    /* No type given for an HCI */
#define USBE_HCI_DATA_BUG         55    /* HC consistency check failed on transaction's HCI data */
#define USBE_HCI_DATA_TOO_BIG     56    /* HCI data is too big to fit into frame list */
#define USBE_HCI_HOST_NULL        57    /* NULL pointer to HCI host structure */
#define USBE_HCI_NAME_NULL        58    /* No name given for an HCI */
#define USBE_HCI_NOTFOUND         59    /* HCI controller not found as active */
#define USBE_HCI_NULL             60    /* NULL pointer passed as HCI structure */
#define USBE_HOSTSTRUCT_MISALIGN  61    /* Misaligned ts_usb_host* structure */
#define USBE_HOST_BAD_ADDR        62    /* Invalid host controller address */
#define USBE_HOST_BUSY            63    /* The host structure is still busy (use count non zero) */
#define USBE_HOST_BUS_ERROR       64    /* Bus error when accessing from host controller */
#define USBE_HOST_HCI_NO_FREE     65    /* The host structure has a HCI data but no method to free it */
#define USBE_HOST_NOT_FOUND       66    /* Host structure not found on the host list */
#define USBE_HOST_NO_ABORT        67    /* Host structure misses an abort method */
#define USBE_HOST_NO_RECEIVE      68    /* Host structure misses a receive method */
#define USBE_HOST_NO_XMIT         69    /* Host structure misses a transmit method */
#define USBE_HOST_NULL            70    /* NULL ts_usb_host structure pointer passed */
#define USBE_HUB_CANT_STOP        71    /* Couldn't stop the kernel hub driver daemon */
#define USBE_HUB_NO_ROOT          72    /* Couldn't find a root hub on the host */
#define USBE_HUB_NULL             73    /* NULL pointer to hub structure */
#define USBE_PORT_OUT_RANGE       74    /* Port number out of range */
#define USBE_PORT_RESET_INT       75    /* Port's reset process was interrupted or not completed */
#define USBE_PORT_STAT_C_UNK      76    /* Unknown hub's port status change code */
#define USBE_HUB_ROOT_NULL        77    /* Host doesn't have an assigned root hub */
#define USBE_HUB_STAT_C_UNK       78    /* Unknown hub status change code */
#define USBE_IFACE_BAD_INDEX      79    /* Invalid or out of range interface number */
#define USBE_IFACE_DESCR_BAD_TYPE 80    /* Bad descriptor type as interface descriptor */
#define USBE_IFACE_NULL           81    /* NULL pointer to interface structure */
#define USBE_IOREGION_USED        82    /* IO region already in use */
#define USBE_IRQ_USED             83    /* IRQ already in use */
#define USBE_MFAULT               84    /* Bad access to memory */
#define USBE_NOT_UHCI             85    /* Not an UHCI compliant USB controller */
#define USBE_NO_FREE_IRQ          86    /* Coulnd't find a free IRQ to use */
#define USBE_NO_IOREGION          87    /* Host controller has assigned IO region starting at zero! */
#define USBE_NO_MEMORY            88    /* No memory to allocate */
#define USBE_OPEN_CORRUPT         89    /* Corrupted ts_usb_open entry */
#define USBE_OPEN_NO_DEV          90    /* No USB destination device set */
#define USBE_OPEN_NO_ENDP         91    /* No USB destination endpoint set */
#define USBE_OPEN_NO_HOST         92    /* No USB destination bus set */
#define USBE_OPEN_NO_IFACE        93    /* No USB destination interface set */
#define USBE_OPEN_NULL            94    /* NULL ts_usb_open descriptor */
#define USBE_PCI_READ_ERROR       95    /* Error reading from PCI */
#define USBE_PCI_WRITE_ERROR      96    /* Error reading to PCI */
#define USBE_POINTER_NULL         97    /* NULL pointer passed as argument */
#define USBE_PORT_BAD_INDEX       98    /* Port index number out of range */
#define USBE_REQ_BAD_DIRECTION    99    /* Invalid direction for a USB request */
#define USBE_REQ_BAD_REC          100   /* Invalid recipient for a USB request */
#define USBE_REQ_BAD_TYPE         101   /* Invalid type for a USB request */
#define USBE_SHORT_READ           102   /* Readed less data than requested */
#define USBE_SHORT_WRITE          103   /* Wrote less data than requested */
#define USBE_TRBUF_NULL           104   /* NULL transfer buffer descriptor passed to function */
#define USBE_TR_ABORTED           105   /* Transaction aborted */
#define USBE_UHCI_BAD_DEFAULTS    106   /* Bad defaults values after reset */
#define USBE_UHCI_COULDNT_HALT    107   /* Couldn't halt UHCI controller */
#define USBE_UHCI_COULDNT_START   108   /* Coulnd't start UHCI controller */
#define USBE_UHCI_FL_BAD_ALIGN    109   /* Mis-aligned pointer for UHCI frame list */
#define USBE_UHCI_HOST_NULL       110   /* NULL ts_uhci_host structure pointer passed */
#define USBE_UHCI_INIT_ERROR      111   /* Error initializing UHCI controller */
#define USBE_UHCI_RESET_FAILED    112   /* Reset of UHCI controller failed */
#define USBE_UNSUPPORTED_RELEASE  113   /* Device uses an unsupported USB release */
#define USBE_ZERO_IRQ             114   /* USB controller configured to use IRQ 0 */


    /*
     * Lists declaration
     * -----------------
     *
     *  This declarations have to occur here due to the fairly
     * complexity of the definition process for macro'ed code
     */

    /*
     * Declaration of a triple node list
     *
     */

#define NLISTS 3

#define s_Nslist        slist_stk(s_,NLISTS,slist)
#define s_Nsnode        slist_stk(s_,NLISTS,snode)
#define ts_Nslist       slist_stk(ts_,NLISTS,slist)
#define ts_Nsnode       slist_stk(ts_,NLISTS,snode)

ts_msnode_declare;
ts_mslist_declare;

#ifdef __KERNEL__
#define ts_Nslist_nozero   slist_stk(ts_,NLISTS,slist_nozero)
#define ts_Nslist_clean    slist_stk(ts_,NLISTS,slist_clean)
#define ts_Nslist_head     slist_stk(ts_,NLISTS,slist_head)
#define ts_Nslist_del      slist_stk(ts_,NLISTS,slist_del)
#define ts_Nslist_del_fast slist_stk(ts_,NLISTS,slist_del_fast)
#define ts_Nslist_release  slist_stk(ts_,NLISTS,slist_release)
#define ts_Nslist_printk   slist_stk(ts_,NLISTS,slist_printk)
#define ts_Nslist_init     slist_stk(ts_,NLISTS,slist_init)
#define ts_Nslist_next     slist_stk(ts_,NLISTS,slist_next)
#define ts_Nslist_last     slist_stk(ts_,NLISTS,slist_last)
#define ts_Nslist_append   slist_stk(ts_,NLISTS,slist_append)
#define ts_Nslist_prep     slist_stk(ts_,NLISTS,slist_prep)
#define ts_Nslist_move     slist_stk(ts_,NLISTS,slist_move)
#define ts_Nslist_prev     slist_stk(ts_,NLISTS,slist_prev)
#define ts_Nsnode_nozero   slist_stk(ts_,NLISTS,snode_nozero)
#define ts_Nsnode_clean    slist_stk(ts_,NLISTS,snode_clean)
#define ts_Nsnode_release  slist_stk(ts_,NLISTS,snode_release)
#define ts_Nsnode_printk   slist_stk(ts_,NLISTS,snode_printk)
#define ts_Nsnode_init     slist_stk(ts_,NLISTS,snode_init)
#define ts_Nsnode_next     slist_stk(ts_,NLISTS,snode_next)
#define ts_Nsnode_ref      slist_stk(ts_,NLISTS,snode_ref)
#define ts_Nsnode_unref    slist_stk(ts_,NLISTS,snode_unref)
#define ts_Nsnode_unref_nd slist_stk(ts_,NLISTS,snode_unref_nd)

ts_msnode_declare_functions;
ts_mslist_declare_functions;

#undef ts_Nslist_nozero
#undef ts_Nslist_clean
#undef ts_Nslist_del
#undef ts_Nslist_del_fast
#undef ts_Nslist_release
#undef ts_Nslist_printk
#undef ts_Nslist_init
#undef ts_Nslist_next
#undef ts_Nslist_last
#undef ts_Nslist_append
#undef ts_Nslist_prep
#undef ts_Nslist_move
#undef ts_Nslist_prev
#undef ts_Nsnode_nozero
#undef ts_Nsnode_clean
#undef ts_Nsnode_release
#undef ts_Nsnode_printk
#undef ts_Nsnode_init
#undef ts_Nsnode_next
#undef ts_Nsnode_ref
#undef ts_Nsnode_unref
#undef ts_Nsnode_unref_nd

#endif /* __KERNEL__ */

    /* clear #defines which clutter */

#undef NLISTS
#undef s_Nslist
#undef s_Nsnode
#undef ts_Nslist
#undef ts_Nsnode


    /* USB data types
     * --------------
     *
     * Note that almost every ts_usb_* structure has a corresponding
     * set of functions to initialize/free them correctly. THESE
     * functions should be used, and never do it by hand. This is
     * intented to evitate confusions, as the code assumes certain
     * states of the variables as initilized/non-initialized values
     * that shoudl be freed, etc ... Take a look at generic.c
     */

    /*
     * USB escalar types
     * -----------------
     *
     * Ranges:
     *
     * t_usb_host   0-127; <0 invalid
     *
     * t_usb_addr   0-127; <0 invalid
     *
     * t_usb_cfg    0-255
     *
     * t_usb_iface  0-255
     *
     * t_usb_aset   0-255
     *
     * t_usb_endp   0-32; < 0 invalid
     *              Bit     0 = direction:
     *                          0 - upstream, out
     *                          1 - downstream, in
     *              Bits    1:4 = endp number
     */

typedef signed char     t_usb_host; /* USB host number */
typedef signed char     t_usb_addr; /* USB device's address */
typedef unsigned char   t_usb_cfg;  /* USB device's configuration */
typedef unsigned char   t_usb_iface;/* USB configurations' interface */
typedef unsigned char   t_usb_aset; /* USB interface's alternate setting */
typedef signed char     t_usb_endp; /* USB device's endpoint */

    /*
     * Tests for evaluating the endpoints direction
     * Note that USB endpoints can only be read or write, never both,
     * except for control endpoints, which are both (and thus,
     * the direction bit is ignored).
     */

#define USB_ENDP_READABLE(endpnumber) ((endpnumber) & 0x1)
#define USB_ENDP_WRITEABLE(endpnumber) (!((endpnumber) & 0x1))
/* note that the base endpoint number will always be an even number */
#define USB_ENDP_BASE_NO(endpnumber) ((endpnumber) & 0xE)
#define USB_DESC_ENDPNO2ADDR(endpn) ((((endpn)&0xf)<<1)|((endpn)&0x80?1:0))

    /*
     * USB compact address (host/device/endpoint)
     * ------------------------------------------
     *
     * NOTE: the first invalid address (-1) invalidates the remaining
     *       ones) and that endp 0 always exist.
     */

typedef struct
{
    t_usb_host  host;   /* Host number */
    t_usb_addr  device; /* Device address */
    t_usb_endp  endp;   /* Endpoint number */
} ts_usb_addr;

#ifdef __KERNEL__
#define ts_usb_addr_nozero(addr)                \
do                                              \
{                                               \
    (addr)->host = -1;                          \
    (addr)->device = -1;                        \
    (addr)->endp = -1;                          \
} while (0)

#define ts_usb_addr_clean(addr)                 \
do                                              \
{                                               \
/*     memset (addr, 0, sizeof (*(addr))); */   \
    ts_usb_addr_nozero (addr);                  \
} while (0)

extern ts_usb_addr* ts_usb_addr_init (ts_usb_addr*);
extern void ts_usb_addr_release (ts_usb_addr*);
#endif /* __KERNEL__ */


    /*
     * USB host controller types
     * -------------------------
     *
     */

typedef enum
{
    e_usb_hci_inv = 0,  /* Undefined */
    e_usb_hci_uhci,     /* UHCI controller */
    e_usb_hci_ohci      /* OHCI controller */
} e_usb_hci;


    /*
     * USB transfer types
     * ------------------
     *
     *  See USB Spec 1.0 4.7.[1234], 5.4, 8.5. These values match the
     * ones reported on the bmAttributes field of an endpoint
     * descriptor (See USB Spec 1.0 9.6.4).
     */

typedef enum
{
    e_usb_tr_type_cntl = 0, /* Control */
    e_usb_tr_type_isoc,     /* Isochronous */
    e_usb_tr_type_bulk,     /* Bulk */
    e_usb_tr_type_intr,     /* Interrupt */
    e_usb_tr_type_inv       /* Invalid type */
} e_usb_tr_type;


    /* Transfer/endpoint status
     * ------------------------
     *
     *  This enum applies for denoting transfers and endpoint (pipe)
     * status. It's used in ts_usb_endp and ts_usb_tr.
     *
     *  WARNING! The order is important, so greater/smaller than
     * relations can be used; thus, every transaction >= _status_ok
     * has finished, very transaction >= _status_queud but <
     * _status_ok is executing, queued for execution [to sum
     * up, running], and every transaction >= _status_clean <
     * _status_queued is being prepared.
     *
     *  Similarly, for pipe status, negative numbers mean errors
     * (trough only _babble or _stall may be seen), _ok means the pipe
     * is free for use, _executing means it's executing transfers, and
     * _queued means it's have transfers queued for execution, but
     * none is executing [this status should only be entered on few
     * times, when processing to start a new transaction].
     *
     *  See the diagram 'Transaction Request Life Cycle.fig' at the
     * docs.
     */

typedef enum
{
    e_usb_tr_status_clean = 0,  /* Clean */
    e_usb_tr_status_filled,     /* Filled */
    e_usb_tr_status_setup,      /* HCD Data structures are set up */
    e_usb_tr_status_inserted,   /* HCD Data structures are inserted */
    e_usb_tr_status_queued,     /* Queued for execution when previous xfers finish */
    e_usb_tr_status_executing,  /* Being executed */
    e_usb_tr_status_ok,         /* Ok, transfer is done and finished ok */
    e_usb_tr_status_aborted,    /* Aborted */
    e_usb_tr_status_error,      /* There was an error */
    e_usb_tr_status_max
} e_usb_tr_status;

extern int e_usb_tr_status_2_errno[];


    /*
     * Structure keeping info for each USB descriptor opened
     * -----------------------------------------------------
     *
     * WARNING! When a transfer is set down, it is removed from this
     *          list, so keep pointers to it somewhere
     *
     *   The transactions list is there so we can abort all
     * transactions (executing and aborted) when a ts_usb_open is
     * closed (anyother way, it'd be impossible to keep track of
     * them).
     *
     *  [For a description of states, see usbd-tr.c]
     *
     *  It is also used for queuing transfers. All the transactions
     * are appended to the list when moved to the SETUP state. When a
     * transaction is put to execute, it transitions to the QUEUED
     * state and moved up the list, after the last transaction in
     * EXECUTING or QUEUED state. When a transaction finishes
     * (moves to ERROR, DONE or ABORTED state), it is moved down to
     * the list, appended to the last transaction in ERROR, DONE or
     * ABORTED states. Below this are all the transactions, in order
     * of setup which still haven't been executed.
     *
     *  The 'running' pointer points to the last element in the list which
     * is in either EXECUTING or QUEUED state. If none is
     * queued for execution, then it will be NULL.
     *
     *  The 'finished' pointer points to the last element in the list
     * which is in either ERROR, DONE or ABORTED state. If there are
     * no elements in any of these states, it will point to the last
     * one in the executing states; if 'running' is also NULL, then it
     * will be NULL.
     *
     *  There must be done reference count on these pointers.
     *
     *  This way, the list is divided in three zones: executing
     * transactions, finished transactions and prepared transactions.
     *
     *  This is used by the HCD to, after completing a transaction,
     * see if there are more pending and to execute them.
     */

typedef struct s_usb_host ts_usb_host;
typedef struct s_usb_device ts_usb_device;
typedef struct s_usb_endp ts_usb_endp;
typedef struct s_usb_tr ts_usb_tr;
typedef struct s_usb_open ts_usb_open;

struct s_usb_open
{
    ts_usb_host     *host;      /* host */
    ts_usb_device   *dev;       /* device */
    ts_usb_endp     *endp;      /* device's endpoint we are using */
    ts_slist        list;       /* scheduled transfers (see above) */
    ts_snode        *running;   /* pointer to last transaction running */
    ts_snode        *finished;  /* pointer to last transaction finished */
};

#ifdef __KERNEL__
#define ts_usb_open_nozero(open)                \
do                                              \
{                                               \
    ts_slist_nozero (&(open)->list);            \
} while (0)

#define ts_usb_open_clean(open)                 \
do                                              \
{                                               \
    memset ((open), 0, sizeof (*(open)));       \
    ts_usb_open_nozero (open);                  \
} while (0)
extern ts_usb_open* ts_usb_open_init (ts_usb_open*);
extern void ts_usb_open_release (ts_usb_open*);
extern void ts_usb_open_printk (const ts_usb_open*);
#endif /* __KERNEL__ */


    /*
     * USB Endpoint structure
     * ----------------------
     *
     * See USB Spec 1.0 9.6.4
     *
     * NOTE: the pktsize is setable by non-USB defined means
     *
     * NOTE: the endpoint number is given in UUSBD semantics (in the
     *       format described for data type t_usb_endp, is is
     *       translated from the format described at the USB Spec
     *       9.6.4.
     *
     * FIXME: add statistics. See SCRATCH file.
     */

struct s_usb_endp
{
    t_usb_endp      addr;       /* endpoint number/address */
    e_usb_tr_type   type;       /* transfer type */
    e_usb_tr_status stat;       /* endpoint xfer status */
    __u16           maxpktsize; /* maximum packet size */
    __u8            interval;   /* polling interval for data transfers (ms) */
    ts_usb_open     *desc;      /* descriptor who's using us */
#if defined (CONFIG_USB_OHCI) || defined (CONFIG_USB_OHCI_MODULE)
    void            *hci;       /* HCI specific data */
#endif
};

#ifdef __KERNEL__
#define ts_usb_endp_nozero(endp)                \
do                                              \
{                                               \
    (endp)->addr = -1;                          \
    (endp)->type = e_usb_tr_type_inv;           \
    (endp)->stat = e_usb_tr_status_clean;       \
} while (0)

#define ts_usb_endp_clean(endp)                 \
do                                              \
{                                               \
    memset (endp, 0, sizeof (*(endp)));         \
    ts_usb_endp_nozero (endp);                  \
} while (0)

extern ts_usb_endp* ts_usb_endp_init (ts_usb_endp*);
extern ts_usb_endp* ts_usb_endp_a_init (unsigned);
extern ts_usb_endp** ts_usb_endp_pa_init (ts_usb_endp**, unsigned);
extern void ts_usb_endp_pa_clear (ts_usb_endp**, unsigned);
extern void ts_usb_endp_release (ts_usb_endp*);
extern void ts_usb_endp_a_release (ts_usb_endp*, unsigned);
extern void ts_usb_endp_pa_release (ts_usb_endp**, unsigned);
extern void ts_usb_endp_printk (const ts_usb_endp*);
#endif /* __KERNEL__ */


    /*
     * Configurable element structure
     * ------------------------------
     *
     *  This structure holds all the pointers to the data needed by
     * the drivers to handle configurations, as well as it's
     * attributes.
     *
     *  The configured flag is used to let know the cfel's status. It
     * may happen that the cfel pointers ain't needed, and still the
     * there must be a way to say it's configured. This flag must be
     * set by the driver's configuration methods.
     */

typedef struct s_usb_cfel ts_usb_cfel;
struct s_usb_cfel
{
    __u8    class;      /* USB class */
    __u8    subclass;   /* USB subclass */
    __u8    protocol;   /* USB protocol code */

        /* configured if not 0; usable to track as ID */

    unsigned long   configured;

        /* 
         *  Destructing this element means to unconfigure the part of
         * the device it configurse. Thus the destructor should be
         * something like CLASS_destroy (data).
         *
         *  Note the use of ts_snode* pointers to point to the driver
         * specific data. Though they  
         * are a ts_snode pointer, it is intended that they point to
         * any kind of structures, with the only condition of being
         * this structures beginning with a ts_snode structure, to
         * allow housekeeping. See usb-slist.h.
         *
         *  This approachs allows the USBD to let the driver know and
         * unconfigure the device on surprise disconnects cleanly.
         */

    ts_snode    *data;       /* driver specific data */
};

#ifdef __KERNEL__
#define ts_usb_cfel_nozero(cfel)                \
do                                              \
{                                               \
} while (0)

#define ts_usb_cfel_clean(cfel)                 \
do                                              \
{                                               \
    memset (cfel, 0, sizeof (*(cfel)));         \
    ts_usb_cfel_nozero (cfel);                  \
} while (0)
extern ts_usb_cfel* ts_usb_cfel_init (void);
extern void ts_usb_cfel_release (ts_usb_cfel*);
extern void ts_usb_cfel_printk (const ts_usb_cfel*);
#endif /* __KERNEL__ */


    /*
     * USB Alternate setting structure
     * -------------------------------
     *
     * See USB Spec 1.0 9.6.3
     *
     * In fact, each USB Interface is one of this structures. The
     * ts_usb_iface structure (described below) just has a collection
     * of all the available alternate settings and the active one.
     *
     * NOTE: the order of the endpoints MUST be the same order as
     *       they're declared on the descriptors.
     */

typedef struct
{
    t_usb_aset  addr;       /* alternate setting number */
    ts_usb_cfel cfel;       /* configurable element */
    ts_usb_endp *endp;      /* Array of endpoints */
    t_usb_endp  endps;      /* available endpoints */

    char        *descr;     /* string describing this alternate setting */
} ts_usb_aset;

#ifdef __KERNEL__
#define ts_usb_aset_clean(_aset)                \
do                                              \
{                                               \
    memset ((_aset), 0, sizeof (*(_aset)));     \
    ts_usb_cfel_clean (&(_aset)->cfel);         \
} while (0)
extern ts_usb_aset* ts_usb_aset_init (ts_usb_aset*);
extern ts_usb_aset* ts_usb_aset_a_init (unsigned);
extern void ts_usb_aset_a_release (ts_usb_aset*, unsigned);
extern void ts_usb_aset_release (ts_usb_aset*);
extern void ts_usb_aset_printk (const ts_usb_aset*);
#endif /* __KERNEL__ */


    /*
     * USB Interface structure
     * -----------------------
     *
     * See USB Spec 1.0 9.6.3
     */

typedef struct
{
    t_usb_iface addr;   /* interface number */

    ts_usb_aset *aset;  /* alternate settings for this interface */
    t_usb_aset  asets;  /* number of alternate settings for this interface */
    ts_usb_aset *active;/* ptr to the active alternate setting */
} ts_usb_iface;

#ifdef __KERNEL__
#define ts_usb_iface_nozero(iface)              \
do                                              \
{                                               \
} while (0)

#define ts_usb_iface_clean(iface)               \
do                                              \
{                                               \
    memset ((iface), 0, sizeof (*(iface)));     \
    ts_usb_iface_nozero (iface);                \
} while (0)
extern ts_usb_iface* ts_usb_iface_init (ts_usb_iface*);
extern ts_usb_iface* ts_usb_iface_a_init (unsigned);
extern void ts_usb_iface_release (ts_usb_iface*);
extern void ts_usb_iface_a_release (ts_usb_iface*, unsigned);
extern void ts_usb_iface_printk (const ts_usb_iface*);
#endif /* __KERNEL__ */


    /*
     * USB Configuration structure
     * ---------------------------
     *
     * See USB Spec 1.0 9.6.2
     *
     *  The ts_usb_cfg_pref is be used to select which configuration
     * is prefered at initial configuraiton time (this is used, for
     * example, by the hub driver in the hubd_cfg_prefs variable). The
     * initial configuration is tried to match as closest as possible
     * to the preference.
     */

typedef struct
{
    __u8    buspower:1;     /* bus powered: 1 - yes, 0 - no */
    __u8    selfpower:1;    /* self pwred:  1 - yes, 0 - no */
    __u8    remwakeup:1;    /* supports remote wake-up:1 - yes, 0 - no */
    __u8    min_maxpower;   /* Minimum Max pwr from bus (2mA units) */
    __u8    max_maxpower;   /* Minimum Max pwr from bus (2mA units) */
} ts_usb_cfg_pref;

typedef struct s_usb_cfg ts_usb_cfg;

struct s_usb_cfg
{
    t_usb_cfg       addr;       /* configuration number */

    __u8            failed:1;   /* last try to set to this cfg:
                                 * 0 - succeeded, 1 - failed */
    __u8            buspower:1; /* bus powered: 1 - yes, 0 - no */
    __u8            selfpower:1;/* self pwred:  1 - yes, 0 - no */
    __u8            remwakeup:1;/* supports remote wake-up:1 - yes, 0 - no */
    __u8            maxpower;   /* Max pwr consumption from bus (2mA units) */

    ts_usb_iface    *iface;     /* array of interfaces */
    t_usb_iface     ifaces;     /* number of different interfaces */

        /*
         * Array of pointers to available endpoints except endpoint 0
         * (numbering as in t_usb_endp, shifted one left to remove the
         * direction bit). The usb_dst_endp_set() routine already
         * knows that pendp[0] is for endpoint 1.
         */

    ts_usb_endp     **pendp;    /* array of pointers to endpoints */
    __u8            pendps;     /* number of pointers allocated */
    char            *descr;     /* string describing this configuration */
    atomic_t        usecnt;     /* references to this configuration */
};

#ifdef __KERNEL__
#define ts_usb_cfg_nozero(cfg)                  \
do                                              \
{                                               \
    atomic_set (&(cfg)->usecnt, 0);             \
} while (0)
#define ts_usb_cfg_clean(cfg)                   \
do                                              \
{                                               \
    memset ((cfg), 0, sizeof (*(cfg)));         \
    ts_usb_cfg_nozero (cfg);                    \
} while (0)
extern ts_usb_cfg* ts_usb_cfg_init (void);
extern void ts_usb_cfg_release (ts_usb_cfg*);
extern void ts_usb_cfg_printk (const ts_usb_cfg*);
#endif /* __KERNEL__ */


    /*
     * USB device structure
     * --------------------
     *
     *  See USB Spec 1.0 9.6.1)
     *
     * FIXME: add statistics which must be a sum of all the device's
     *        endpoints. However, it could enhance resource usage to
     *        have a function being called from the user to gather
     *        statistics, so that they'd be calculated only when
     *        called.
     */

    /* Lists in which the device is registered */

#define usb_dev_l_dev   0   /* list of configured devices */
#define usb_dev_l_host  1   /* list of devices registered on a host */
#define usb_dev_l_uncfg 2   /* list of unconfigured devices */

    /*
     * Tests for guessing the device type
     *
     * They'll evaluate to non-zero if true, zero if false
     */

#define USB_DEVICE_AG_INDEP(a) ((a)->cfel.class == 0)
#define USB_DEVICE_AG_DEP(a) ((a)->cfel.class != 0)

    /*
     *  Note: to get the host to which a device is plugged, we access
     * it trough the hub structure, thus host = device->hub->host.
     *
     * DESTRUCTOR: usb_device_destroy()
     */

typedef struct s_usb_hub ts_usb_hub;

struct s_usb_device
{                           /* 3snode is not a valid identifier! */
    ts_3snode   msnode;     /* multiple simple linked lists node */

    t_usb_addr  addr;       /* device's address */
    ts_usb_hub  *hub;       /* hub we are connected to */
    __u8        port;       /* port we are connected to in that hub */

    unsigned    remwakeup:1;/* remote wakeup 0 - disabled, 1 - enabled */
    unsigned    selfpwrd:1; /* device is 0 - bus powered, 1 - self powered */
    unsigned    cfgzero:1;  /* cfg index is 0 - 1 based */
    
    ts_usb_endp endp0;      /* Endpoint zero, always existant */
    ts_usb_open endp0d;     /* descriptor for endpoint 0, always open */

    ts_usb_cfg  **pcfg;     /* array of pointers to configurations */
    t_usb_cfg   pcfgs;      /* number of posible configurations */
    ts_usb_cfg  *active;    /* active configuration */

    int         cfels;      /* cfels (configurable elements) */
    int         cfelu;      /* unconfigured cfels */
    int         acfels;     /* active cfels (configurable elements) */
    int         acfelu;     /* unconfigured active cfels */

    ts_usb_cfel cfel;       /* device configurable element */

    __u16       usb_version;/* USB spec release version in BCD */
    __u16       vendor;     /* USB vendor ID */
    __u16       product;    /* product ID */
    __u16       version;    /* device release version in BCD */
    char        *mfrer;     /* manufacturer string */
    char        *descr;     /* product string */
    char        *serial;    /* serial number string */
};

#ifdef __KERNEL__
#ifdef __SMP__          /* FIXME: change this to rwlock_t_clean() */
#define __USB_SMP(a) a
#else
#define __USB_SMP(a)
#endif

#define ts_usb_device_nozero(device)            \
do                                              \
{                                               \
    ts_3snode_nozero (&(device)->msnode);       \
    (device)->addr = -1;                        \
    ts_usb_endp_nozero (&(device)->endp0);      \
    ts_usb_open_nozero (&(device)->endp0d);     \
    (device)->cfels = 1;                        \
    (device)->cfelu = 1;                        \
    (device)->acfels = 1;                       \
    (device)->acfelu = 1;                       \
    ts_usb_cfel_nozero (&(device)->cfel);       \
} while (0)

#define ts_usb_device_clean(device)             \
do                                              \
{                                               \
    memset ((device), 0, sizeof (*(device)));   \
    ts_usb_device_nozero (device);              \
} while (0)
extern ts_usb_device* ts_usb_device_init (void);
extern ts_usb_device* ts_usb_device_a_init (unsigned);
extern void ts_usb_device_release (ts_usb_device*);
extern void ts_usb_device_a_release (ts_usb_device*, unsigned);
extern void ts_usb_device_printk (const ts_usb_device*);
#endif /* __KERNEL__ */


    /*
     * USB requests (USB Spec 1.0 9.3)
     * ------------
     *
     *  Use ts_usb_request_compose() to compose the bmreqtype field
     * from direction, type and recipient.
     *
     *  Use ts_usb_request_(dir,rec,type)() to extract the direction,
     * recipient and request type from a bmreqtype bitmap.
     *
     *  No need to change byte order, all the functions take care of
     * it for you!!
     *
     * See usb-stddesc.h for the USB_REQ_* definitions
     *
     */

    /* Number of bytes for the minimal USB request */

#define USB_REQ_XFER_LENGTH 8

typedef struct
{
    __u8    bmreqtype;  /* Type of the request (USB_REQ_*_*) */
    __u8    request;    /* Specific request (USB_REQ_*) */
    __u16   value;      /* Value that varies with the request */
    __u16   index;      /* Value that varies with the request */
    __u16   length;     /* length of data phase */
} ts_usb_request;

#define ts_usb_request_rec(bmreqtype) ((bmreqtype)&0x1f)
#define ts_usb_request_type(bmreqtype) (((bmreqtype)>>5)&0x3)
#define ts_usb_request_dir(bmreqtype) (((bmreqtype)>>7)&0x1)
#define ts_usb_request_compose(direction, type, recipient)  \
    ((direction) & 0x1 << 7)                                \
        | ((type) & 0x3 << 6)                               \
        | ((recipient) & 0x1f))


    /*
     * Flags for ts_usb_tr callback
     * ----------------------------
     *
     *  When the callback is called, come flags are passed to indicate
     * the reason of the call. They are these:
     */

#define USB_CB_PKTS     0x1     /* tr.pkts have been transmited */
#define USB_CB_LOOP     0x2     /* the transaction has looped */
#define USB_CB_DONE     0x4     /* the transaction has finished */
#define USB_CB_ERROR    0x8     /* with _DONE -> transaction had errors */

    /*
     * Transaction buffers
     * -------------------
     *
     *  Describes and holds all the data for a transaction, as
     * requested by the tf_usb_* host methods. Also holds the status
     * and statistics of the transaction, so it may be consulted
     * there.
     *
     *  The use mode is simple. Create the structure, keep a reference
     * to it (with ts_snode_ref()) if you want to keep it around -for
     * example, to see the result- and fill it. Then feed it as
     * parameter to the corresponding usb_tr_* () functions. It'll be
     * added to the transaction list which the ts_usb_open struct
     * contains and sent to the HCD for processing. When done (wether
     * successfully, aborted or error), it'll be removed from the list
     * by the HCD and the warning method you chose (if any) will be
     * launched.
     *
     *  When a transaction is finished, wether done, finished with
     * error or aborted, there are two methods two warn about the
     * fact, with a function callback or a wait queue, selected with
     * data member `warn_method'.
     *
     *  Thus, the `warn' member is a union of fields `callback' and
     * `wq'. If non-null and callback selected (`warn_method' == 1)
     * the callback will be executed when the transaction is
     * finished. It is up to the called function to be careful on what
     * it does [FIXME: define careful and on what conditions is the
     * callback called] as it will probably be called from an IRQ or
     * Bottom Half handler. In the same fashion, if the wait queue
     * method is selected, then the pointer wait queue is woken up
     * when the transaction is finished [this way the process has more
     * freedom to do whatever it wants, but the drawback is it has
     * more latency].
     *
     *  The `usb_errno' and `errno' codes are aplicable only for the
     * transaction, and are undefined until the transaction is in one
     * of the aborted, done or error status). Also, they are ALWAYS
     * NEGATIVE or ZERO!! The `status' indicates on which stage is the
     * transaction currently in. When the transaction is finished, it
     * indicates on which state the transaction finished. Note the
     * codes are the same as for endpoint status on ts_usb_open
     * structures.
     *
     *  To sum up, to be able to use this struct, you gotta fill in
     * the indicated members. If you want the call to block until
     * done, set wq/callback to NULL. After cleaning it [with
     * ts_usb_tr_clean()], it is filled with a reasonable defaults [no
     * open descriptor, no data nor size -NULL USB Transfer, no loops,
     * defaults max error ratio and count, timeout; no packet count
     * warnings, refill activated, DATA0 toggle, blocking call, CLEAN
     * status, and no errnos].
     *
     *  If you want a transaction of no data, set data to NULL and
     * size to zero. Any other combination of NULL and zero will raise
     * an error.
     *
     *  The ts_usb_request* req member is a pointer to a USB Request
     * structure. If this transaction isn't intended to be a control
     * transfer, it will be NULL and ignored. If it is a control
     * transfer, if will be used UNTIL IT IS MOVED AGAIN TO THE FILLED
     * or CLEAN STATE, and will be modified from host order to little
     * endian if your host is big endian. After moving to FILLED
     * state, it'll be left on the original host order.
     *
     *  The semantics are these:
     *
     * size This is the size of the data buffer [the whole one].
     *
     * data_size
     *
     *      It is the size used for the Transfer Control Interface
     *      (Host Controller dependant) data structures. Instead of
     *      making HCI data structures for the whole buffer's 'size',
     *      they are made only for data_size, so we don't spend _too_
     *      much resources.
     *
     *      When the host is done transfering 'data_size' bytes, and
     *      if it is different (always smaller) than 'size', two
     *      things may happen: pointer advancement (refill) or pointer
     *      looping.
     *
     *      The goal is to transfer, in total 'size' bytes. If
     *      'refill' is required, it will automagically refill it's
     *      data structures, assuming the buffer is 'size' bytes in
     *      length, to make the transactions transfer from/to 'data +
     *      n*data_size', where 'n' is the number of times it has
     *      refilled. When it has transfered 'size' bytes, it will
     *      finish [ok, will process loops, etc ...].
     *
     *      If no refill is required ['refill' == 0], then it will
     *      loop over that buffer until the 'size' bytes are
     *      transfered, but the application will have to take care of
     *      refilling the buffer with appropiate data at the
     *      appropiate times. A nice way of doing this is setting
     *      callback warnings each half buffer, so when it's
     *      transferring the first half of the buffer you fill the
     *      second one and viceversa. To set up the warnings, you use
     *      the 'pkts' variable.
     *
     *      Note that the overheads of both methods may be
     *      balanced. The first method forces the HCD to rewrite it's
     *      data structures from time to time, what involves some
     *      pointer increments (fast), then converting from kernel or
     *      user virtual addresses to bus addresses [not that trivial]
     *      and then to little endian. The second one forces the user
     *      to actualize the buffer. If mmaping+locking from user
     *      land, perhaps the first one is preferible.
     *
     * loops
     *
     *      This variable indicates how many time is this transaction
     *      to be automatically repeated when done (done is considered
     *      when 'size' bytes have been transferred, without errors
     *      reaching the maximum values). If -1, it will be looped
     *      infinitely. If 0, no repetitions will be made. If 'n', it
     *      will be repeated 'n' times. Note you must make sure then
     *      endpoint will accept the data. The transaction statistics
     *      won't be cleared before each loop, so they are added to
     *      the totals.
     *
     * max_ratio
     *
     *      The errors ratio [aka: ratio] is defined as the number of
     *      errors multiplied by 1024 [FIXME] and divided by the
     *      number of _succesfully_ transferred bytes (note on
     *      isochronous transfers there's no distinction between
     *      succesful and failed transfered bytes).
     *
     *      The maximum ratio is the top value the ratio can reach
     *      before considering the transaction failed.
     *
     *      Note this magnitude considers errors in the long run as
     *      well as very puntual errors. If a lot of errors rise at
     *      the very beginning, it will shortly fail. If it sticks at
     *      some point of the transmission with a lot of errors, it
     *      will also make it fail. But if in the whole transaction,
     *      some errors happen, it won't. However, if a transaction
     *      has too much errors per byte, then we can consider the
     *      connection very flaky and warn the user.
     *
     * max_errors
     *
     *      The maximum number of errors we permit in a transaction,
     *      as an absolute error count.
     *
     * timeout
     *
     *      Timeout for the endpoint. If it takes more than the given
     *      milliseconds to answer, we consider it an error.
     *
     * pkts Each 'pkts' packets transfered, launch a warning. You can
     *      compute the packets number dividing the bytes by the
     *      packet size (ts_usb_endp.maxpktsize). When the warning is
     *      called, you can differentiate why are you called looking
     *      at the state. If EXECUTING, it is a packet or loop
     *      warning. To distinguish these two is quite difficult,
     *      if using wait queues (but easier with callbacks, as it is
     *      a parameter); it could be done seeing how much has been
     *      transfered, as this is updated each time a warning is
     *      launched. Anyhow, it is not a very good idea [neither
     *      perhaps neccessary] to use loop and packet warnings
     *      together.
     *
     *      NOTE: if a packet warning has to be launched at the same
     *            time than the done warning, it won't be, as there is
     *            no need for it.
     *
     *  This is computed each time the HCD scheduler has got a chance
     * [for example, when warning, handling errors or at end of
     * transfer], but it always is correctly computed when the
     * transaction is moved to the DONE, ABORTED or ERROR states.
     */

struct s_usb_tr             /* Transaction buffer descriptor */
{
    ts_snode    snode;      /* This is a single node in a list */

       /* Fill these in */

    ts_usb_open *open;      /* pointer to descriptor we use */
    ts_usb_request* req;    /* pointer to request data (if control xfer) */
    void        *data;      /* pointer to data buffer */
    size_t      data_size;  /* size of the data buffer/HCI basic buffer */
    size_t      size;       /* size of data to transmit/receive */
    size_t      loops;      /* repeat? -1 - infinitely, 0 - no, n - n times */
    size_t      max_ratio;  /* max errors/bytes*1024 ratio before failing */
    size_t      max_errors; /* max allowed errors before failing */
    size_t      timeout;    /* endpoint timeout [ms] */
    size_t      pkts;       /* warn each pkts packets? 0 - no, n - each n */

#if 0
    unsigned    user:1;     /* data is in 0 - kernel, 1 - user land */
#endif
    unsigned    toggle:1;   /* first data toggle used: DATA{0,1} */
    unsigned    refill:1;   /* refill the HCI's pointers? 0 - no, 1 - yes */
    unsigned    nak_error:1;/* NAK is error? 0 - no, 1 - yes */
    unsigned    method:1;   /* warn using 1 - callback, 0 - wait queue */

    unsigned    done_warn:1;/* warn at tr end? 0 - no, 1 - yes */
    unsigned    loop_warn:1;/* warn before looping: 0 - no, 1 - yes */

    union
    {
        void (*callback)(ts_usb_tr*, int);  /* int == flags, see above */
        struct wait_queue **wq;
    } warn;

        /*
         * STOP FILLING IN, THIS DATA IS READ-ONLY!
         *
         *  However, this data is constantly actualized, so you can
         * read it, read_lock(&this->snode.lock)ing, reading and
         * read_unlock(&this->snode.lock)ing it.
         *
         *  It's volatile because the HCD may update it anytime.
         */

    volatile e_usb_tr_status status;    /* Transaction status */
    volatile int    usb_errno;  /* USB errno code */
    volatile int    errno;      /* Errno */

    volatile size_t xfered;     /* Transferred bytes */
    volatile size_t errors;     /* Number of errors during transfer */

    void            *hci;       /* Pointer to HCI data (destructor at host) */
    struct timer_list timer;    /* timer */
};

#ifdef __KERNEL__
#define ts_usb_tr_nozero(_tr)                   \
do                                              \
{                                               \
    ts_snode_nozero (&(_tr)->snode);            \
    (_tr)->max_ratio = USB_DEFAULT_MAX_RATIO;   \
    (_tr)->max_errors = USB_DEFAULT_MAX_ERRORS; \
    (_tr)->timeout = USB_DEFAULT_TIMEOUT;       \
    (_tr)->refill = 1;                          \
    (_tr)->status = e_usb_tr_status_clean;      \
    init_timer (&(_tr)->timer);                 \
} while (0)

#define ts_usb_tr_clean(_tr)                    \
do                                              \
{                                               \
    memset ((_tr), 0, sizeof (*(_tr)));         \
    ts_usb_tr_nozero (_tr);                     \
} while (0)
extern ts_usb_tr* ts_usb_tr_init (void);
extern void ts_usb_tr_release (ts_usb_tr*);

#endif /* __KERNEL__ */


    /*
     * Hub structures
     * --------------
     *
     *  Yeah, this shouldn't belong here, but it is the only way to
     * allow the root struct to have an static hub struct. Anyway,
     * hubs are something _very_ needed on USB :)
     *
     *  This structure keeps all the info neccessary for a
     * hub. Basically you can find the array of ports that form this
     * hub, the hub characteristics and the two open descriptors to
     * the control and status change endpoints. Also there are
     * pointers (which inc their usecounts) to the device who holds
     * the hub and to the host the device is connected to.
     *
     * NOTE: the `local_power' and `overcurrent' fields are updated each
     *       time usb_hub_status_get() is called.
     */

typedef struct s_usb_port ts_usb_port;
typedef struct s_usb_hub_ops ts_usb_hub_ops;

struct s_usb_hub
{
    ts_snode        snode;          /* Magic destructor, usecount and list node */

    ts_usb_host     *host;          /* host we are connected to */
    ts_usb_device   *device;        /* our device */

    t_usb_cfg       cfg;            /* FIXME: guess this info isn't needed */
    t_usb_iface     iface;          /*        as hub driver always hooks at */
    t_usb_aset      aset;           /*        device->cfel->data */

    __u8            ports;          /* Number of down-stream ports */
    ts_usb_port     *port;          /* Array of ports' data */
    __u8            pwrontime;      /* Time (2ms units) to complete a power-on on each port */
    __u8            maxcurrent;     /* Maximum current requirement (mA units) */

    unsigned        compound:1;     /* Part of compound device: 0 - no, 1 - yes */
    unsigned        overc_prot:2;   /* Overcurrent protection capabilities: 0 - Global, 1 - Per Port, 2|3 - none */
    unsigned        power:2;        /* Power switching capabilities: 0 - ganged, 1 - individual, 2|3 - always on */

    unsigned        local_power:1;  /* Local power is: 1 - lost, 0 - active [only for self powered] */
    unsigned        overcurrent:1;  /* Power status: 0 - power ok, 1 - global overcurrent */

    unsigned        configuring:1;  /* we are: 0 - not, 1 - configuring the hub */
    atomic_t        in_progress;    /* we are: 0 - not, 1 - dealing with an status change */

        /* If statch_tr->open == NULL, this is not initialized */

    ts_usb_open     statch;         /* status change endpoint open descriptor */
    ts_usb_tr       statch_tr;      /* status change endpoint xfer buffer */

    const ts_usb_hub_ops *operations;/* operations on this hub */
};

#ifdef __KERNEL__
#define ts_usb_hub_nozero(_hub)                                 \
do                                                              \
{                                                               \
    ts_snode_nozero (&(_hub)->snode);                           \
    (_hub)->power = 3;      /* FIXME: #define/sysctl/proc */    \
    atomic_set (&(_hub)->in_progress, 0);                       \
    ts_usb_open_nozero (&(_hub)->statch);                       \
    ts_usb_tr_nozero (&(_hub)->statch_tr);                      \
    ts_snode_ref (&(_hub)->statch_tr.snode);                    \
} while (0)

#define ts_usb_hub_clean(_hub)                  \
do                                              \
{                                               \
    memset ((_hub), 0, sizeof (*(_hub)));       \
    ts_usb_hub_nozero (_hub);                   \
} while (0)
extern ts_usb_hub* ts_usb_hub_init (void);
extern void ts_usb_hub_release (ts_usb_hub*);
#endif /* __KERNEL__ */


    /*
     * USB Host data structure
     * -----------------------
     *
     *  Note the pointer to the hci data. It will be really a
     * ts_uhci_host or ts_ohci_host, but both begin with a ts_dtor to
     * allow housekeeping.
     *
     * DESTRUCTOR: usb_host_destroy()
     */

    /*
     * Methods for manipulating hosts
     *
     * all these return an errno if returning a signed type.
     */

typedef int (*tf_usb_reset)(ts_usb_host*);
typedef int (*tf_usb_setsof)(ts_usb_host*, unsigned);
typedef int (*tf_usb_ref)(ts_snode*, int);

#if defined (CONFIG_USB_OHCI) || defined (CONFIG_USB_OHCI_MODULE)
typedef int (*tf_usb_endp_setdn)(ts_usb_device*, ts_usb_endp*);
typedef int (*tf_usb_endp_setup)(ts_usb_device*, ts_usb_endp*);
#endif

typedef int (*tf_usb_tr_setdn)(ts_usb_tr*);
typedef int (*tf_usb_tr_setup)(ts_usb_tr*);

#if defined (CONFIG_USB_UHCI) || defined (CONFIG_USB_UHCI_MODULE)
typedef int (*tf_usb_tr_insert)(ts_usb_tr*);
typedef int (*tf_usb_tr_remove)(ts_usb_tr*);
#endif

typedef int (*tf_usb_tr_run)(ts_usb_tr*);
typedef int (*tf_usb_tr_reset)(ts_usb_tr*);
typedef int (*tf_usb_tr_abort)(ts_usb_tr*);

struct s_usb_host
{
    ts_snode        snode;      /* node for the hosts list (see usb-slist.h) */
    t_usb_host      addr;       /* host's address */
    rwlock_t        lock;       /* lock for twinkling with the host's data */

    ts_3slist       devices;    /* List of devices connected to this host */
    ts_usb_hub      root;       /* root hub structure */
    ts_usb_device   *master;    /* master device */
    unsigned        sof;        /* Start Of Frame value */

        /* HCI specific */

    ts_snode        *hci;       /* host controller interface (ts_*hci_host*) */
    e_usb_hci       hcitype;    /* host controller interface type */

        /* Generic HCI methods */

    tf_usb_reset    reset;      /* reset (host+bus+transactions) method */
    tf_usb_setsof   setsof;     /* set the SOF */
    tf_usb_ref      reference;  /* reference/unreference the driver */

#if defined (CONFIG_USB_OHCI) || defined (CONFIG_USB_OHCI_MODULE)
    tf_usb_endp_setdn   endp_setdn; /* destroy enpd's HCI data */
    tf_usb_endp_setup   endp_setup; /* setup endp's HCI data */
#endif
        /*
         * Transaction related methods
         *
         *  Note the tr_setdn, remove, tr_reset and tr_abort methods,
         * even when failing (retcode < 0), will be considered they
         * made their work until the possible extent and the
         * transaction state will be changed.
         */

    tf_usb_tr_setdn     tr_setdn;   /* destroy transactions's HCI data */
    tf_usb_tr_setup     tr_setup;   /* setup transaction's HCI data */

#if defined (CONFIG_USB_UHCI) || defined (CONFIG_USB_UHCI_MODULE)
    tf_usb_tr_insert    tr_insert;  /* insert a transaction */
    tf_usb_tr_remove    tr_remove;  /* remove a transaction */
#endif

    tf_usb_tr_run       tr_run;     /* run (execute) a transaction */
    tf_usb_tr_reset     tr_reset;   /* reset a transaction */
    tf_usb_tr_abort     tr_abort;   /* abort transactions */
};

#ifdef __KERNEL__
#define ts_usb_host_nozero(host)                    \
do                                                  \
{                                                   \
    ts_snode_nozero (&(host)->snode);               \
    __USB_SMP ((host)->lock = RW_LOCK_UNLOCKED);    \
    ts_3slist_nozero (&(host)->devices);            \
    (host)->devices.nlist = usb_dev_l_host;         \
    ts_usb_hub_nozero (&(host)->root);              \
    ts_snode_ref (&(host)->root.snode);             \
    (host)->hcitype = e_usb_hci_inv;                \
} while (0)

#define ts_usb_host_clean(host)                 \
do                                              \
{                                               \
    memset ((host), 0, sizeof (*(host)));       \
    ts_usb_host_nozero (host);                  \
} while (0)
extern ts_usb_host* ts_usb_host_init (void);
extern void ts_usb_host_release (ts_usb_host*);
#endif


    /* Structure for USB class/subclass/protocol/vendor drivers
     * --------------------------------------------------------
     *
     *  This structure holds the basic information for a USB
     * driver. Also some function pointers are included, which are
     * basic for driver operation.
     *
     *  The 'config' method is used to configure the device. If called
     * to configure the device cfel, it may configure as many
     * interfaces as it wishes. Else, it may only configure the
     * interface it's been given. Returns:
     *
     *  <0  error [it is considered that nothing has been configured]
     *  0   nothing configured
     *  N   N active configurable elements configured
     *
     *  The 'unconfig' method is used to unconfigure any cfel on a
     * device which is configured by the driver. It has got to search
     * trough it's data structures for entries for that device and
     * call the destructors on it. Returns:
     *
     *  <0  error [it is considered that everything has been
     *      unconfigured]
     *  0   nothing unconfigured
     *  N   N active configurable elements unconfigured
     *
     *  Neither configuration or unconfiguration method will add or
     * remove the device from any UUSBD list nor actualize
     * cfel-counts. That will be handled by the UUSBD.
     *
     *  The 'reference' method is used when we take any of the driver's
     * code or data structures to use it. If so, we call it with a
     * positive argument (one reference, regardless the value of the
     * _positive_ number). When we are done, we release the reference
     * calling it with a negative number. It always returns the number
     * of references or < 0 on error. Calling it with an argument of
     * zero will just return the number of references.
     *
     *  If the driver wishes to release itself, it'll run the list of
     * devices which use it (maintained by (*reference)()) calling the
     * destructor on each pointer it's got stored.
     *
     *  There's a default reference function, usb_drv_reference(), but
     * drivers may need to provide more specialized ones to fit their
     * needs.
     *
     *  See the document 'doc/Configuration.sgml' for more information
     * on the inner workings of the configuration system. FIXME: quite
     * out of date now.
     *
     *  The unique id field is a signature composed with the vendor,
     * product, class, subclass, protocol and the _contents_ of the
     * name copyright, *name ... strings. The idea behind this is to
     * be able to reproduce the same id *everytime* based on that
     * data. It is generated with the function usb_drv_id().
     *
     *  DESTRUCTOR: usb_drv_destroy (ts_usb_driver*)
     */

typedef struct s_usb_drv_use ts_usb_drv_use;
typedef struct s_usb_driver ts_usb_driver;

struct s_usb_drv_use
{
    ts_snode    snode;
    ts_snode    *user;
};

struct s_usb_driver
{
    ts_snode snode;

    __u16   vendor;     /* USB vendor ID this driver applies to, 0 - any */
    __u16   product;    /* Product ID this driver applies to, 0 - any */
    __u8    class;      /* USB class this driver applies to, 0 - any */
    __u8    subclass;   /* USB subclass this driver applies to, 0 - any */
    __u8    protocol;   /* USB protocol this driver applies to, 0 - any */

    __u32   id;         /* unique id, usb_drv_register() fills it */

        /* Optional, but recommended */

    const char  *name;          /* short name */
    const char  *copyright;     /* copyrighting stuff */
    const char  *vendorname;
    const char  *productname;
    const char  *classname;
    const char  *subclassname;
    const char  *protocolname;

        /*
         * driver methods and housekeepeing stuff
         *
         *  We don't relay on using the active configuration as it may
         * happen that in the middle of the configuration process,
         * this one is changed by somebody.
         *
         * ts_usb_device*   Device to configure/unconfigure.
         *
         * ts_usb_cfg*      Configuration of the alternate setting to
         *                  configure (NULL if none).
         *
         * t_usb_cfg        Configuration of the alternate setting to
         *                  configure (index number, only valid if
         *                  previous pointer is set). It is
         *                  &device->cfg[index].
         *
         * ts_usb_iface*    Interface of the alternate setting to
         *                  configure (NULL if configuration is
         *                  NULL).
         *
         * t_usb_iface      Interface of the alternate setting to
         *                  configure (index number, only valid if
         *                  previous pointer is set). It is
         *                  &device->cfg[cfg_index].iface[index].
         *
         * ts_usb_aset*     Alternate setting to configure (NULL if
         *                  interface is NULL).
         *
         * t_usb_aset       Alternate setting to configure (index
         *                  number, only valid if previous pointer is
         *                  set). It is
         *                  &device->cfg[cfg_index].iface[iface_idx]
         *                  .aset[index].
         *
         * __u32            Bitmap of reasons for the call (see
         *                  USB_CFG_*)
         *
         * NOTE: when calling (*unconfig)() it may happen that the
         *       device isn't reachable anymore [it may be unplugged,
         *       powered down ...], so the unconfiguration routines
         *       must be ready to handle this fact.
         *
         * NOTE: See structure description for the return values.
         *
         * NOTE: when calling the config method, it is guaranteed that
         *       an USB configuraton is already active. However, the
         *       code is free to change it.
         */

    int (*config) (ts_usb_cfel*, ts_usb_device*,
                   ts_usb_cfg*, t_usb_cfg,
                   ts_usb_iface*, t_usb_iface,
                   ts_usb_aset*, t_usb_aset, __u32);
    int (*unconfig) (ts_usb_device*, __u32);
    int (*reference) (ts_usb_device*, int);

    ts_slist uselist;   /* list of uses and use count (no. of elements) */
};

#ifdef __KERNEL__
#define ts_usb_drv_use_nozero(_use)             \
do                                              \
{                                               \
    ts_snode_nozero (&(_use)->snode);           \
} while (0)

#define ts_usb_drv_use_clean(_use)              \
do                                              \
{                                               \
    memset (_use, 0, sizeof (*(_use)));         \
    ts_usb_drv_use_nozero (_use);               \
} while (0)

    /* Rets int so it acts as destructor too */

static inline
void ts_usb_drv_use_release (ts_usb_drv_use* use)
{
    if (use == NULL)
        return;
    if (use->user)
        ts_snode_unref (use->user, NULL);
    return;
}

extern int ts_usb_drv_use_destroy (ts_usb_drv_use*);
static inline
ts_usb_drv_use* ts_usb_drv_use_init (void)
{
    ts_usb_drv_use* use;

    use = usb_kmalloc (sizeof (ts_usb_drv_use), GFP_KERNEL);
    if (use == NULL)
        return NULL;
    ts_usb_drv_use_clean (use);
    ts_dtor_fill (&use->snode.dtor, (tf_dtor)ts_usb_drv_use_destroy);
    return use;
}

#define ts_usb_driver_nozero(driver)            \
do                                              \
{                                               \
    ts_snode_nozero (&(driver)->snode);         \
    ts_slist_nozero (&(driver)->uselist);       \
} while (0)

#define ts_usb_driver_clean(driver)             \
do                                              \
{                                               \
    memset ((driver), 0, sizeof (*(driver)));   \
    ts_usb_driver_nozero (driver);              \
} while (0)
extern ts_usb_driver* ts_usb_driver_init (void);
extern void ts_usb_driver_release (ts_usb_driver*);
extern void ts_usb_driver_printk (const ts_usb_driver*);
#endif /* __KERNEL__ */


    /*
     * Host Controller Interfaces
     * --------------------------
     *
     *  The 'type' member is used to match when destroying the HCI
     * list of hosts.
     *
     * DESTRUCTOR: usb_hci_destroy()
     */

typedef struct s_usb_hci ts_usb_hci;
struct s_usb_hci
{
    ts_snode    snode;  /* Dtor, use count and node */
    const char  *name;  /* Name of this HCI */
    e_usb_hci   type;   /* Type of the HCI */
    ts_slist    hosts;  /* Controllers within this HCI */
        /* FIXME: add any statistics? */
};

#ifdef __KERNEL__
#define ts_usb_hci_nozero(hci)                  \
do                                              \
{                                               \
    ts_snode_nozero (&(hci)->snode);            \
    ts_slist_nozero (&(hci)->hosts);            \
} while (0)

#define ts_usb_hci_clean(hci)                   \
do                                              \
{                                               \
    memset ((hci), 0, sizeof (*(hci)));         \
    ts_usb_hci_nozero (hci);                    \
} while (0)
extern ts_usb_hci* ts_usb_hci_init (void);
extern void ts_usb_hci_release (ts_usb_hci*);
#endif /* __KERNEL__ */


    /*
     * Global variables
     * ----------------
     */

#ifdef __KERNEL__

    /*
     * error code set only on error, with negative values given by the
     * #define's USBE_*
     */

extern volatile int usb_errno;

#endif /* __KERNEL__ */

#endif /* _usb_h_ */


/* Log:
 *
 * $Log: usb.h,v $
 * Revision 1.3  1998/11/16 00:01:02  inaky
 * Added the number of maximum changes drivers can make on device cfels
 * when configuring before considering it's a race condition and
 * aborting.
 *
 * Revision 1.2  1998/11/14 17:22:05  inaky
 * Removed inclusion of modversions, which caused lots of problems when
 * compiling into the kernel, and fixed the definition of
 * tf_usb_tr_(insert|remove) when UHCI is being compiled.
 *
 * Revision 1.1  1998/11/14 00:38:27  inaky
 * Isolated all the header files
 *
 * Revision 1.83  1998/11/09 21:26:18  inaky
 * Changed the name of the configuration pointer array to pcfg. Added
 * references to members of structs which are snodes and static to the
 * structs, so they don't get freed [as they should not].
 *
 * Optimized some stuff to the case of being using only UHCI or OHCI.
 *
 * Revision 1.82  1998/10/31 02:09:58  inaky
 * Added special fields for endpoints on OHCI. Made them not be compiled
 * if OHCI support is not selected.
 *
 * Added the fix for devices using index 0 as first configuration
 * descriptor, with a bit which indicates it.
 *
 * Removed all the d_* pointers from ts_usb_cfel and collapsed them into
 * a single pointer; only one was really used.
 *
 * Revision 1.81  1998/10/21 23:59:40  inaky
 * Fixed the list numbers. I'm stupid. Multiple-list nodes are composed
 * of an array of links, and it is zero-based!
 *
 * Revision 1.80  1998/09/24 15:53:31  inaky
 * Moved 'hub_port' to 'port'. Finished the split of _clean into _nozero
 * and _clean.
 *
 * Revision 1.79  1998/09/14 02:35:21  inaky
 * Removed old hacky config system, moved the static pendp array to a
 * dynamic one [t ts_usb_cfg].
 *
 * Revision 1.78  1998/08/28 23:39:10  inaky
 * Changed USBE codes to decimal, to ease error tracking.
 *
 * Revision 1.77  1998/08/27 11:01:17  inaky
 * Moved ts_*_free() to ts_*_release(), made them return void and severed
 * assertions.
 *
 * Simplified ts_*_clean() macros. More space to simplification is
 * waiting, but for next time.
 *
 * Revision 1.76  1998/08/25 04:01:26  inaky
 * Moved the configured flag in a driver to be unsigned long, so it can
 * be used to put a big stamp (by the driver) to determine who owns the
 * cfel configuration.
 *
 * Revision 1.75  1998/08/24 15:47:23  inaky
 * Fixed little cosmetic stuff, removed old files.
 *
 * Revision 1.74  1998/08/24 02:22:49  inaky
 * Changed default timeout to be three seconds. Added structure_printk
 * functions and changed return type of some functions to void.
 *
 * Revision 1.73  1998/08/19 23:49:12  inaky
 * Added timeouts for transactions, adding means to easily change it's
 * units.
 *
 * Added a flag to hubs to see if they're been processed or not, to
 * evitate various change processing threads to fiddle at the same time
 * with a hub.
 *
 * Revision 1.72  1998/08/03 13:29:50  inaky
 * Added again timeouts with ms precission. Changed the name of the list
 * member in the open descriptor structure to 'list' (from 'tr'), as it
 * led to confusion as a ts_usb_tr descriptor is almost always called
 * 'tr', and 'tr->open->tr' was prone to misunderstandings.
 *
 * Changed also the wait queue pointer to be a pointer to pointer (struct
 * wait_queue **wq), as the wait queue is a 'struct wait_queue*', and we
 * multiple warn.wq to be able to refer to the same wait queue.
 *
 * Revision 1.71  1998/07/28 23:20:14  inaky
 * Renumbered error codes [hex admits a, b, c, d, e and f, remember? :)],
 * collected all error status for a transaction into a single one and
 * added an option to consider NAKs an error.
 *
 * Revision 1.70  1998/07/24 15:31:48  inaky
 * Removed the PAUSE state, it makes it quite difficult, and it has
 * almost no use. Should think things better ... :)
 *
 * Revision 1.69  1998/07/24 15:30:28  inaky
 * Removed programable timeout, added more error status and callback
 * second parameter to indicate callback call reason [on ts_usb_tr].
 *
 * Revision 1.68  1998/07/23 14:26:45  inaky
 * Fixed comments on e_usb_tr_status (transaction status), added PAUSED
 * state and removed BUSY state (useless).
 *
 * Fixed docs for ts_usb_open and added pointers into it's transaction
 * list of where does the running and finished subqueues start.
 *
 * Added HCI specific information to devices.
 *
 * Added the length field to the request data type.
 *
 * Renamed the ts_usb_trbuf struct to ts_usb_tr; fixed docs, added
 * timeout field, removed user field and added refill field. Also made
 * volatile all the fields the HCD code may update.
 *
 * Renamed the methods in the USB host structure to make them more
 * looking like what they should do.
 *
 * Made all errnos volatile.
 *
 * Revision 1.67  1998/07/16 14:09:14  inaky
 * Added entries for some default values.
 *
 * Modified enumeration elements to be _all_ prefixed with the
 * enumeration type name [to clean name space].
 *
 * Added some docs to the usage of ts_usb_tr structures
 * (transactions) and some comments on things to do. Also added a lot of
 * new interesting fields and functionality :).
 *
 * Moved from the old read/write interface to the new, request-based
 * interface; this implied changes in the host structure.
 *
 * Revision 1.66  1998/07/14 21:56:48  inaky
 * Added a retry parameter for resetting.
 * Fixed ts_usb_device_clean to also clean ENDPOINT 0's DESCRIPTOR!!
 *
 * Revision 1.65  1998/07/12 22:50:08  inaky
 * Moved in hub structure so the root hub can be used _statically_ in the
 * host structure. Added the reference method to the host structure.
 *
 * Revision 1.64  1998/06/21 20:14:39  inaky
 * Added a fix to clean namespace on ts_Nsnode stuff.
 *
 * Added the usb_drv_use structure to keep use lists on drivers. It may
 * be deprecated again really soon now.
 *
 * Revision 1.63  1998/06/20 21:54:52  inaky
 * Massive changes to split the old monolithic hubd driver into a hub
 * manager (hub wrappers and daemon) and a dynamic driver, which can be
 * replaced and its different parts reused.
 *
 * Revision 1.62  1998/05/01 19:08:13  inaky
 * Moved in the USB Packet ID definitions from the UHCI code.
 *
 * Revision 1.61  1998/04/26 03:28:54  inaky
 * Added the concept of max known USB Spec release and more USB errno
 * codes.
 *
 * Added a macro to convert from the endpoint addr in descriptor to the
 * format used by the UUSBD. Also added a comment on the transfer type
 * enumeration and forced it to begin in 0 for ctl xfers, for the sake of
 * clarity.
 *
 * Removed from the endpoint structure the pktsize field, which was
 * useless. Also removed the stall member, in favour of using the stat
 * member for that task.
 *
 * Moved the configurations array in a device to be an array of pointers,
 * so that unused configurations may be deallocated and thus waste less
 * space. Yet to be fully implemented.
 *
 * Revision 1.60  1998/04/13 22:03:58  inaky
 * Added new errno codes.
 *
 * Moved up the transfer status, which now is also endpoint status.
 *
 * Added status to the endpoint structure.
 *
 * Added comments and the configuration preferences datatype [still
 * unused, forthcoming].
 *
 * Added a 'failed' bit to the configuration, which is set if the set to
 * the configuration fails.
 *
 * Added status fields to the device structure and removed some others
 * which belonged to the port structure.
 *
 * Moved the trbuffers to use two methods, selectable.
 *
 * Fixed lot of comments on the driver structure [now autorithative
 * source of doc until new notice].
 *
 * Revision 1.59  1998/03/28 19:29:23  inaky
 * Fixed 3slist declaration code; moved ts_usb_open to use lists for
 * transfer scheduling, moved the wait queue in ts_usb_trbuf to a
 * callback fucntion. Also moved the tci member to be a ts_dtor, so it's
 * easy to destroy it.
 *
 * Revision 1.58  1998/03/26 11:23:34  inaky
 * Fixed some typos, finished obsoleting usb-methods.h, fixed docs for
 * ts_snode pointers on ts_usb_cfel and some more.
 *
 * Revision 1.57  1998/02/12 16:35:08  inaky
 * Finished the device ops, and moved synch_frame getting to usbd-endp,
 * where it belongs. Fixed some stuff here and there, and starting to get
 * worried on locks ... I'm gonna have to debug *lots* of deadlocks.
 *
 * Revision 1.56  1998/02/12 14:32:09  inaky
 * Added comments for the signature field of the ts_usb_driver structure.
 *
 * Revision 1.55  1998/02/10 17:30:22  inaky
 * Moved the 3slist declaration code to the top.
 *
 * Revision 1.54  1998/02/09 13:24:41  inaky
 * Moved to atomic_t use counts.
 *
 * Revision 1.53  1998/01/11 15:44:16  inaky
 * Finished functions from the descriptor, inteface and endpoint modules,
 * fixed docs, pretty printed and removed craps.
 *
 * Revision 1.52  1998/01/08 12:19:10  inaky
 * Added copyright strings to all important files
 *
 * Revision 1.51  1998/01/08 11:49:12  inaky
 * Finishing the configuration system. Only the final cfel-stuff is
 * missing to implement; however, nothing is tested.
 *
 * Revision 1.50  1998/01/02 20:51:55  inaky
 * Finished adding the control endpoints check support and reformatting
 * usb.h to be shorter than 80 columns in most places (except the errno
 * codes).
 *
 * Revision 1.49  1998/01/02 19:39:37  inaky
 * Added the definition for endpoint zero to the device structure and
 * removed it from the list of endpoints kept at each configuration. Also
 * updated it on places where needed and removed the concept of basic
 * configuration, as is no longer needed.
 *
 * Revision 1.48  1998/01/02 00:12:16  inaky
 * Finished the new driver scheme. Documented at usbd-drv.c
 *
 * Revision 1.47  1997/12/31 17:06:35  inaky
 * Removed some unneeded locks and fixed some others.
 *
 * Revision 1.46  1997/12/31 16:37:12  inaky
 * Finished the HCI interface
 *
 * Revision 1.45  1997/12/31 05:37:00  inaky
 * Almost finished all the lists and life cycle for hosts, drivers and
 * devices. Some craps to be fixed yet.
 *
 * Revision 1.44  1997/12/30 22:30:18  inaky
 * Added the foundations of the HCD and fixed some bugs.
 *
 * Revision 1.43  1997/12/30 20:15:43  inaky
 * Moved everything to the new lists scheme. Untested yet.
 *
 * Revision 1.42  1997/12/29 00:25:37  inaky
 * Fixed unimplemented functions to return -ENOSYS and moving the khubdd
 * to use signals as commands.
 * Also added function traces on most files that had them disabled (donkey!)
 *
 * Revision 1.41  1997/12/27 20:06:28  inaky
 * Finished fixinf the __KERNEL__ #ifdef wrapping and moved the ts_dtor.h
 * to usb-dtor.h
 *
 * Revision 1.40  1997/12/27 19:54:21  inaky
 * Added all the prototypes and skeleton definitions (incomplete) of all
 * of the USBD functions to each source file.
 *
 * Revision 1.39  1997/12/27 17:25:35  inaky
 * Finished moving to GDLF. Deprecated usb_printk and __LOCATION__
 *
 * Revision 1.38  1997/12/27 16:44:25  inaky
 * Moving to GDLF and fixinf docs. Removing more cruft.
 *
 * Revision 1.37  1997/12/27 13:18:28  inaky
 * Removed all the PCI stuff and moved it to usb-pci.h and usb-pci.c
 *
 * Revision 1.36  1997/12/27 03:24:23  inaky
 * Moved the monitored memory management code to it's own module
 *
 * Revision 1.35  1997/12/27 02:24:04  inaky
 * Moving everything to use the GDLF for logging and debug checkings.
 * Removed the global array of USB hosts and moved it to a linked
 * list. Added an address field to each USB host controller.
 * Removed lots of cruft that were accumulated and useless. Still some
 * missing, but may be useful someday.
 *
 * Revision 1.34  1997/12/22 02:44:31  inaky
 * Trying to move all the documentation to the source files from the
 * header files, as it was duplicated.
 *
 * Revision 1.33  1997/12/08 02:41:00  inaky
 * Fixed some stuff of the #includes, modified the makefile targets to be
 * more flexible, tried to finish the device configuration section.
 *
 * Revision 1.32  1997/12/06 01:55:48  inaky
 * Now finishing the configuration system
 *
 * Revision 1.31  1997/12/06 00:19:40  inaky
 * Almost finished the usbd/host, usbd/cfg and a lot of meat from
 * usbd/device. Also added the driver definition. Got to polish it yet
 * ...
 *
 * Revision 1.30  1997/11/29 20:21:48  inaky
 * Documented and almost fixes the read/write/abort API, documented a
 * damn lot and implemented some stuff
 *
 * Revision 1.29  1997/11/29 00:46:30  inaky
 * Documented almost all the hub basic API, added spin locks to
 * usb_host_register() and some more stuff. Going for more.
 *
 * Revision 1.28  1997/10/31 05:44:30  inaky
 * *** empty log message ***
 *
 * Revision 1.27  1997/10/24 04:20:20  inaky
 * Security backup
 *
 * Revision 1.26  1997/10/19 22:20:26  inaky
 * Modified some things, fixed some parameters to functions prototypes
 * and added a few FIXMEs
 *
 * Revision 1.25  1997/10/17 00:16:00  inaky
 * *** empty log message ***
 *
 * Revision 1.24  1997/10/02 21:29:03  inaky
 * Remade all the directory structure to make it cleaner when the USB
 * class drivers come.
 *
 * Revision 1.23  1997/10/01 21:28:11  inaky
 * Re-making a lot of stuff in the hadnling of host, hub and device
 * initialization and handling, following a more correct approach of
 * registering/unregistering. Also moved the usb_host array to be an
 * array of pointers, so dynamic reallocation of it is easier, and the
 * host structure needs not to be moved in memory. The same applies to
 * the device arrays for each host.
 *
 * Revision 1.22  1997/09/30 13:04:42  inaky
 * Moved the name of the package to uusbd: Universal USB Driver
 *
 * Revision 1.21  1997/09/25 00:12:30  inaky
 * At the very least I hunted down the bug which was causing all that
 * damned exceptions, all memory related. Well, I'm a cock-fingered. Ix
 * don't know how to code; I'm stupid; I'm dreadful; I'm ....
 *
 * Revision 1.20  1997/09/08 10:54:50  inaky
 * Modified the generic functions to gain some performance, also some other checks
 *
 * Revision 1.19  1997/09/01 19:14:43  inaky
 * Backup check in
 *
 * Revision 1.18  1997/07/03 20:00:05  inaky
 * Moved all the hw dependencies to the HCD module, modified the
 * controller detections stuff to use just PCI means (okay, other buses
 * to come, but when they come). General fixes, but no really new deep
 * enhancements.
 *
 * Revision 1.17  1997/06/22 19:54:31  inaky
 * Seems that the moving of all the hw dependants stuff to the HCI
 * modules is done. Got to test it yet, but at least it compiles cleanly.
 *
 * Revision 1.16  1997/06/19 20:38:18  inaky
 * Moving al hardware dependant stuff to the HCD modules
 *
 * Revision 1.15  1997/05/03 21:22:49  inaky
 * Adding a lot of code, with the only objetive of adding the first
 * transfer type: control :) Still there's a lot of stuff to do, mostly
 * concerning data structures handling.
 *
 * Revision 1.14  1997/04/20 21:25:57  inaky
 * Finished the spec for the Hub Class (now gotta implement it)
 * :). Started coding the Frame Lists code, hope UHCI support will be
 * finished in some sessions.
 * This code is breaking, I'm doing something bad with kmallocs. kFree()ing
 * something not kmalloc()ed.
 *
 * Revision 1.13  1997/04/14 16:56:02  inaky
 * *** empty log message ***
 *
 * Revision 1.12  1997/04/12 16:47:49  inaky
 * Added the code for buffering reads/writes to/from userland
 *
 * Revision 1.11  1997/04/10 21:30:49  inaky
 * Fixed the file code, so it opens/closes ok; also reads/writes ok ... ok, just theory
 *
 * Revision 1.10  1997/04/06 22:56:44  inaky
 * Added the functions to set destinations of an USB open descriptor
 * (ts_usb_open); however, I haven't tested it yet ... bugs flowing!
 *
 * Revision 1.9  1997/04/06 21:14:33  inaky
 * Still with the definition of the access interface. More or less it is
 * taking an stable form.
 *
 * Revision 1.8  1997/04/02 19:58:01  inaky
 * Adding and fixing lots of things related to the reference
 * counts.
 *
 * Revision 1.7  1997/03/31 20:24:46  inaky
 * Pfff ... a lot. Mostly added descriptors stuff (HUB class still is
 * missing a lot) and I think I've got a definitive idea of how to
 * implement the storing of configurations and interfaces.
 *
 * Revision 1.6  1997/03/14 20:47:30  inaky
 * Fixing auto-detection code. Still fails when the PCI USB device
 * isn't enabled on BIOS ??? Why???
 *
 * Revision 1.5  1997/03/09 23:10:32  inaky
 * *** empty log message ***
 *
 * Revision 1.4  1997/03/02 22:11:02  inaky
 * *** empty log message ***
 *
 * Revision 1.3  1997/02/26 22:28:36  inaky
 * Some little enchancements ... not too much
 *
 * Revision 1.2  1997/02/24 22:29:50  inaky
 * Hmmm ... I'm getting trouble with the IO ports. I guess I'm writing to
 * the bad places. Moved the PCI initialization code to a function (to
 * make it easier to read; on the other hand, it begins to seem Spaghetti
 * Basic :).
 *
 * Revision 1.1  1997/02/22 22:25:44  inaky
 * *** empty log message ***
 *
 *
 */

