
/* USB Driver Stack v$Revision: 1.53 $ 
 *
 * (C) 1997 by Iaky Prez Gonzlez <inaky@peloncho.fis.ucm.es>
 *
 * Distributed under the terms of the GNU Public License Version 2
 * See files COPYRIGHT and COPYING at the root distribution directory
 *
 * $Id: kernel-interface.c,v 1.53 1998/11/14 17:14:27 inaky Exp $
 *
 * Changelog at the bottom 
 *
 *
 *     Kernel Interface
 *     ----------------
 *
 * Functions to initialize the code and do basic interaction with the
 * rest of the Linux kernel.
 */

#include <linux/init.h>
#include <uusbd/usb.h>                /* general USB stuff */
#include <uusbd/usb-mm.h>             /* monitored mem management */
#ifdef CONFIG_PCI
#include <uusbd/usb-pci.h>            /* USB PCI stuff */
#endif
#include <uusbd/usbd.h>
#include <uusbd/usb-hub.h>
#include <uusbd/usb-slist.h>

#if defined (CONFIG_USB_UHCI_MODULE) \
    || defined (CONFIG_USB_OHCI_MODULE)
#include <linux/kmod.h>
#endif

#define GDLF_IDENT          "USB"
#define GDLF_LOCAL          /* Define if want local settings */
#define GDLF_DEF_LOGLEVEL   9
#define GDLF_ASSERTIONS     /* want assertions (#undef if not) */
#undef GDLF_FNTRACE        /* want function traces */
#define GDLF_LOCATIONS      /* want locations in log messages */
#include <uusbd/gdlf.h>


volatile int usb_errno = 0;    /* USB errno code */

    /* Misc declaration */

extern ts_3slist devices;
extern ts_3slist unconfs;
extern ts_slist drivers;
extern ts_slist hcis;
extern ts_slist hosts;

#ifdef CONFIG_USB_AUDD
extern int audd_init (void);
extern void audd_shutdown (void);
#endif

#ifdef CONFIG_USB_CDCD
extern int cdcd_init (void);
extern void cdcd_shutdown (void);
#endif

#ifdef CONFIG_USB_HIDBP
extern int hidbp_init (void);
extern void hidbp_shutdown (void);
#endif

#ifdef CONFIG_USB_HIDD
extern int hidd_init (void);
extern void hidd_shutdown (void);
#endif

#ifdef CONFIG_USB_HUBD
extern int hubd_init (void);
extern void hubd_shutdown (void);
#endif

#ifdef CONFIG_USB_MSDD
extern int msdd_init (void);
extern void msdd_shutdown (void);
#endif

#ifdef CONFIG_USB_OHCI
extern int ohci_init (void);
extern void ohci_shutdown (void);
#endif

#ifdef CONFIG_USB_UHCI
extern int uhci_init (void);
extern void uhci_shutdown (void);
#endif


    /* Universal USB Driver initialization
     *
     * Returns: 
     *
     *  int             0 if ok, errno code < 0 on error.
     */

int __init usb_init (void)
{
    gdlf_fntrace ("usb_init()");

    gdlf_info ("Universal USB Driver v$Revision: 1.53 $\n");
#ifdef MODULE
    gdlf_info ("(C) 1996-98 Inaky Perez Gonzalez <inaky@peloncho.fis.ucm.es>\n");
#endif

        /* 
         * Initialization of various data structures 
         */

    usb_mm_init();
    ts_3slist_clean (&devices);
    devices.nlist = usb_dev_l_dev;
    ts_3slist_clean (&unconfs);
    unconfs.nlist = usb_dev_l_uncfg;
    ts_slist_clean (&drivers);
    ts_slist_clean (&hcis);
    ts_slist_clean (&hosts);

        /* Start all the USB subsystems */

    usb_hub_init();             /* start the hub manager */

        /*
         * FIXME: When requesting the module, it looks like the
         *        symbols for the USB module are note exported yet,
         *        and thus uhci and ohci cannot be loaded.
         */

#warning FIXME: this does not work
#if 0
#ifdef CONFIG_USB_UHCI_MODULE   /* initialize UHCI controllers */
    request_module ("usb-uhci");
#endif
#ifdef CONFIG_USB_OHCI_MODULE   /* initialize OHCI controllers */
    request_module ("usb-ohci");
#endif
#endif

#ifdef CONFIG_USB_UHCI
    uhci_init();
#endif

#ifdef CONFIG_USB_OHCI
    ohci_init();
#endif

    gdlf_info ("%d USB host%s found\n",
               hosts.elements, hosts.elements == 1? "" : "s");

        /* USB drivers */

#ifdef CONFIG_USB_HUBD
    hubd_init();
#endif

#ifdef CONFIG_USB_AUDD
    audd_init();
#endif

#ifdef CONFIG_USB_HIDD
    hidd_init();
#endif

#ifdef CONFIG_USB_HIDBP
    hidbp_init();
#endif

#ifdef CONFIG_USB_CDCD
    cdcd_init();
#endif

#ifdef CONFIG_USB_MSDD
    msdd_init();
#endif

    return 0;
}   


    /* Release the USB driver
     *
     *  This function will force a release on the driver so that
     * it's use count is zero and if being a module, can be unloaded.
     *
     *  What it does is to force an unregistration of all the hosts.
     *
     *  The hub driver is stopped after the hosts are unregistered as
     * it is suppossed that everything happens in a controlled
     * way. Thus we follow the inverse of the initialization order.
     *
     * Returns:
     *
     *  int             0 if ok, errno code < 0 on error.
     *
     * Global:
     *
     *  int usb_errno   On error, error code.  
     */

int usb_release (void)
{
    gdlf_fntrace ("usb_release()");

    ts_slist_release (&hosts);  /* unregister all the hosts */
    return usb_hub_shutdown();  /* stop the hub driver */
}


    /* Shutdown the USB driver
     *
     * Shuts down the USB host, closing all connections and data
     * flows, frees all the memory and unregisters all the stuff it
     * has to.
     *
     * Returns:
     *
     *  int             0 if ok, errno code < 0 on error.
     */


int usb_shutdown (void)
{
    int errcode1;
    
    gdlf_fntrace ("usb_shutdown()");

    errcode1 = usb_release();   /* release everything */

        /* Release drivers */

    ts_slist_release (&drivers);

        /* Shutdown [compiled-in] drivers */

#ifdef CONFIG_USB_AUDD
    audd_shutdown();
#endif

#ifdef CONFIG_USB_HIDD
    hidd_shutdown();
#endif

#ifdef CONFIG_USB_HIDBP
    hidbp_shutdown();
#endif

#ifdef CONFIG_USB_CDCD
    cdcd_shutdown();
#endif

#ifdef CONFIG_USB_MSDD
    msdd_shutdown();
#endif

#ifdef CONFIG_USB_OHCI
    ohci_shutdown();
#endif

#ifdef CONFIG_USB_UHCI
    uhci_shutdown();
#endif

        /* 
         * Clean data structures 
         *
         * Not too useful, but nice checkpoint :), however, The list
         * of drivers is permament! we got to free it if we don't want
         * memory leaks :).
         */

    gdlf_passert (devices.elements == 0,
                  "Aieeee!! Number of elements in the devices "
                  "list is still %d!\n",         
                  devices.elements);             
    ts_3slist_release (&devices);                  
                                                 
    gdlf_passert (unconfs.elements == 0,         
                  "Aieeee!! Number of elements in the unconfigured "
                  "devices list is still %d!\n", 
                  unconfs.elements);             
    ts_3slist_release (&unconfs);                  
                                                 
    gdlf_passert (hcis.elements == 0,            
                  "Aieeee!! Number of elements in the HCIs "
                  "list is still %d!\n",
                  hcis.elements);
    ts_slist_release (&hcis);

    gdlf_passert (hosts.elements == 0,
                  "Aieeee!! Number of elements in the hosts "
                  "list is still %d!\n",
                  hosts.elements);
    ts_slist_release (&hosts);

    usb_mm_shutdown();
    gdlf_info ("shutting down ... Hasta luego Lucas!!\n");
    return errcode1;        /* and leave */
}


    /* Modular stuff */

#ifdef MODULE

int init_module (void)
{
    return usb_init();
}

void cleanup_module (void)
{
    usb_shutdown();
    return;
}

#endif


/*
 * $Log: kernel-interface.c,v $
 * Revision 1.53  1998/11/14 17:14:27  inaky
 * Moved the include files to uusbd/include/uusbd, to ease integration
 * into the Linux kernel. Linkage into the kernel made possible. Removed
 * copyright strings when linked into the kernel.
 *
 * Revision 1.52  1998/10/14 21:39:06  inaky
 * Added code for initialization of compiled-in modules. Untested.
 *
 * Revision 1.51  1998/09/14 02:30:00  inaky
 * Moved #endifs, so no confusing messages appear.
 *
 * Revision 1.50  1998/08/28 23:44:52  inaky
 * Moved the level of the informational messages to info, as it should
 * be.
 *
 * Moved order of lists initialization.
 *
 * Revision 1.49  1998/08/27 11:14:44  inaky
 * Moved ts_*_free() to ts_*_release(), made them return void and severed
 * assertions.
 *
 * Revision 1.48  1998/08/25 03:59:14  inaky
 * Added a warning on impossible loading of usb-[ou]hci from
 * usb_init(). Should fix this.
 *
 * Revision 1.47  1998/08/24 02:16:04  inaky
 * Added calls to clean all lists before using them. Not doing it was
 * causing memory corruption.
 *
 * Revision 1.46  1998/08/19 23:42:36  inaky
 * Removed spurious debug code.
 *
 * Revision 1.45  1998/08/19 23:35:46  inaky
 * Added the call to the memory management initialization routine on USB
 * startup. Added debug code, which is currently unused and will be
 * removed RSN. Added the call to shutdown the memory management stuff.
 *
 * Revision 1.44  1998/07/23 14:20:10  inaky
 * Made all errnos volatile.
 *
 * Revision 1.43  1998/06/20 21:54:49  inaky
 * Massive changes to split the old monolithic hubd driver into a hub
 * manager (hub wrappers and daemon) and a dynamic driver, which can be
 * replaced and its different parts reused.
 *
 * Revision 1.42  1998/04/13 21:50:27  inaky
 * Moved to use kmod instead of kerneld
 *
 * Revision 1.41  1998/03/28 19:24:46  inaky
 * Removed inclusion of usb-methods.h
 *
 * Revision 1.40  1998/02/12 14:29:46  inaky
 * Added a hack for simulating the inclusion of CONFIG_USB_UHCI as module
 * and removed the old cruft for having a modularised hub driver.
 *
 * Revision 1.39  1998/02/10 17:15:51  inaky
 * Moved up and commented the driver's list releasement. It is something
 * that *must* be done, and not just be checked on an assertion (however,
 * the others must be freed too).
 *
 * Revision 1.38  1998/01/08 12:19:07  inaky
 * Added copyright strings to all important files
 *
 * Revision 1.37  1998/01/02 00:22:08  inaky
 * Removing device reservation and file operations stuff, as they will
 * be part of the rawd IANF
 *
 * Revision 1.36  1998/01/02 00:17:20  inaky
 * Removed the char device allocation in order to move it to rawd
 *
 * Revision 1.35  1998/01/02 00:12:15  inaky
 * Finished the new driver scheme. Documented at usbd-drv.c
 *
 * Revision 1.34  1997/12/31 16:37:11  inaky
 * Finished the HCI interface
 *
 * Revision 1.33  1997/12/30 20:15:42  inaky
 * Moved everything to the new lists scheme. Untested yet.
 *
 * Revision 1.32  1997/12/30 00:26:55  inaky
 * Moved everything to the new scheme of lists
 *
 * Revision 1.31  1997/12/29 00:25:35  inaky
 * Fixed unimplemented functions to return -ENOSYS and moving the khubdd
 * to use signals as commands.
 * Also added function traces on most files that had them disabled (donkey!)
 *
 * Revision 1.30  1997/12/27 20:19:15  inaky
 * Added the usb_release() function and fixed dome things to use it.
 *
 * Revision 1.29  1997/12/27 03:21:37  inaky
 * Moved to GDLF and did a little cleanup. Still missing some
 * actualization when hubd is moved to fully modular.
 *
 * Revision 1.28  1997/12/27 02:07:00  inaky
 * Moving everything to use the GDLF for logging and debug checkings.
 * Removed the global array of USB hosts and moved it to a linked
 * list. Added an address field to each USB host controller.
 * Removed lots of cruft that were accumulated and useless. Still some
 * missing, but may be useful someday.
 *
 * Revision 1.27  1997/12/27 02:05:58  inaky
 * Moving everything to use the GDLF for logging and debug checkings.
 * Removed the global array of USB hosts and moved it to a linked
 * list. Added an address field to each USB host controller.
 * Removed lots of cruft that were accumulated and useless. Still some
 * missing, but may be useful someday.
 *
 * Revision 1.26  1997/12/08 02:40:57  inaky
 * Fixed some stuff of the #includes, modified the makefile targets to be
 * more flexible, tried to finish the device configuration section.
 *
 * Revision 1.25  1997/12/06 00:19:39  inaky
 * Almost finished the usbd/host, usbd/cfg and a lot of meat from
 * usbd/device. Also added the driver definition. Got to polish it yet
 * ...
 *
 * Revision 1.24  1997/11/29 20:21:47  inaky
 * Documented and almost fixes the read/write/abort API, documented a
 * damn lot and implemented some stuff
 *
 * Revision 1.23  1997/11/29 00:46:29  inaky
 * Documented almost all the hub basic API, added spin locks to
 * usb_host_register() and some more stuff. Going for more.
 *
 * Revision 1.22  1997/10/17 00:15:59  inaky
 * *** empty log message ***
 *
 * Revision 1.21  1997/10/02 22:29:36  inaky
 * Finally managed to understand the workings of the MODVERSIONS (ok, not
 * exactly the working, but how to make it work).
 *
 * Now I can go to sleep happy :) -- I know, I'm really sick --
 *
 * Revision 1.20  1997/10/02 21:29:00  inaky
 * Remade all the directory structure to make it cleaner when the USB
 * class drivers come.
 *
 * Revision 1.19  1997/10/01 21:28:10  inaky
 * Re-making a lot of stuff in the hadnling of host, hub and device
 * initialization and handling, following a more correct approach of
 * registering/unregistering. Also moved the usb_host array to be an
 * array of pointers, so dynamic reallocation of it is easier, and the
 * host structure needs not to be moved in memory. The same applies to
 * the device arrays for each host.
 *
 */
