
/* USB Driver Stack v$Revision: 1.23 $ 
 *
 * (C) 1997 by Iaky Prez Gonzlez <inaky@peloncho.fis.ucm.es>
 *
 * Distributed under the terms of the GNU Public License Version 2
 * See files COPYRIGHT and COPYING at the root distribution directory
 *
 * $Id: khubdd.c,v 1.23 1998/11/14 17:14:58 inaky Exp $
 *
 * Changelog at the bottom 
 *
 *
 *     Kernel Hub Driver Daemon
 *     ------------------------
 *
 *  This part of the hub manager controls the response to hub events,
 * such as device attachment, detachment and power changes.
 *
 *  Note that the code which handles port status changes is at
 * ports.c [port_status_change()].  
 */


#include <asm/spinlock.h>   /* spin_[un]lock() */
#include <linux/sched.h>    /* CLONE_VM for kernel_thread() */
#include <linux/types.h>    /* types for unistd.h */
#include <linux/string.h>   /* strncpy() */
#define __KERNEL_SYSCALLS__
#include <linux/unistd.h>   /* kernel_thread() */
#include <linux/smp_lock.h> /* [un]lock_kernel() */ 

#include <uusbd/usb.h>
#include <uusbd/usbd.h>
#include <uusbd/usb-hub.h>
#include "hubm.h"

#define GDLF_IDENT          "USB-HUBM"
#define GDLF_LOCAL          /* Define if want local settings */
#define GDLF_DEF_LOGLEVEL   9
#define GDLF_ASSERTIONS     /* want assertions (#undef if not) */
#undef GDLF_FNTRACE        /* want function traces */
#define GDLF_LOCATIONS      /* want locations in log messages */
#include <uusbd/gdlf.h>

    /* 
     * Use threads to handle hub changes or not.
     *
     *  By defining this, a new thread is spawned to handle each hub's
     * changes, otherway, it is done in the flow of khubdd. Which is
     * better? Depends. If you have a lot of hubs it is probably
     * better threaded, as response should be higher.
     *
     * WARNING: threading is currently quite broken, threads stick in
     *          zombie state after finishing, and don't know how to
     *          fix it.
     */

#warning FIXME: fix KHUBDD_THREAD_CHANGES (try using init contexts)
#undef KHUBDD_THREAD_CHANGES
/* #define KHUBDD_THREAD_CHANGES */

struct wait_queue *wq = NULL;       /* wait queue */
int pid = 0;                        /* khubdd's pid; > 0 running, <= not running */
struct task_struct *pkhubdd = NULL; /* khubdd's task_struct */

    /* Basic struct to pass hubs-in-need-of-attention data to khubdd */

char command;
volatile struct s_dqueue dqueue[USB_HUB_DQUEUE_SLOTS];
rwlock_t dqueue_lock = RW_LOCK_UNLOCKED;


    /* Do the work to handle an status change in the hub or hub ports.
     * 
     * Parameters:  
     *
     *  ts_usb_hub*     Pointer to hub we have to analize.
     */

int usb_hub_deal_status_change (ts_usb_hub* hub)
{
    int bit_top, bit_cnt, byte_cnt, byte, cnt;
    char name[10 + 2 * 20];   /* length should be safe enough */
    int namelen;
    ts_usb_tr *tr;

    gdlf_fntrace ("usb_hub_deal_status_change (0x%p)", hub);

    gdlf_aassert (hub != NULL, return -EINVAL);

        /* See if already being processed */

    atomic_inc (&hub->in_progress);
    if (atomic_read (&hub->in_progress) > 1)
    {
        gdlf_info ("A change is still being processed "
                   "in hub 0x%p (%d %d)\n",
                   hub, hub->host->addr, 
                   hub->device? hub->device->addr : -1);
        atomic_dec (&hub->in_progress);
        return EBUSY;
    }
                        
        /* Allow the change processing thread to be interrupted */

    sigdelset (&current->blocked, SIGTERM | SIGKILL | SIGQUIT);
    recalc_sigpending (current);

    tr = &hub->statch_tr;
    
        /* 
         * Set process name so we see it handles a change and where
         * (note root hubs have device address -1).
         *
         *  I know, this looks weird ... I miss snprintf()
         */

    memset (name, sizeof (name), 0);
	sprintf (name, "khubdd (%d %d)", hub->host->addr, 
             hub->device? hub->device->addr : -1);
    namelen = strlen (name);
    namelen = namelen >= sizeof (current->comm)?
        sizeof (current->comm) - 1: namelen;
    strncpy (current->comm, name, namelen);
    current->comm[namelen] = 0;

        /*
         * 1.Let's check if the transfer went ok; if not, complain and
         *   re-schedule it. Errors re-scheduling are handled by
         *   usb_hub_schedule_status_change().
         */

    if (tr->status > e_usb_tr_status_ok)        /* Error condition? */
    {
        gdlf_err ("Status change endpoint: error status (%d) "
                  "on hub 0x%p: errno/usb_errno %d/%d\n",
                  tr->status, hub, tr->errno, tr->usb_errno);
        atomic_dec (&hub->in_progress);
        usb_hub_schedule_status_change (hub);
        return 0;
    }
    else if (tr->status < e_usb_tr_status_ok)   /* Still on it? */
    {
        gdlf_warning ("Ein? Status change callback fired "
                      "without reason on hub 0x%p: status %d\n",
                      hub, tr->status);
        atomic_dec (&hub->in_progress);
        return 0;
    }

        /*
         * So it went ok ...
         *
         * 2.If it is a hub status change, call hub_status_change().
         *
         *     We assume tr->data is set, as at least, there is one
         *   byte in the status change data [even if the number of
         *   ports were zero, what looks quite stupid ... ].
         */

    if (((__u8*)tr->data)[0] & 0x1) /* First bit set? ... */
        usb_hub_status_change (hub);/* ... then we have a change on hub */

        /*
         * 3.If they're port status changes, call port_status_change()
         *   for each port.
         *
         *     The data read in from the status change endpoint is a
         *   bit array, described at the USB Spec 1.0 11.8.3, with
         *   byte granularity, starting on bit 0 with the hub change
         *   bit indicator, and following a bit change indicator for
         *   each port (starting on port 0). 
         */

    bit_top = hub->ports + 1;   /* +1 to get rid of hub status change bit */
    for (bit_cnt = byte_cnt = 0; bit_cnt < bit_top; byte_cnt++) 
        for (byte = ((__u8*)tr->data)[byte_cnt], cnt = 0; 
             cnt < 8 && bit_cnt < (hub->ports + 1); 
             cnt++, bit_cnt++)
        {
            if (bit_cnt == 0)               /* get rid of hub status ... */
                continue;                   /* ... change bit */

            if (byte & (0x1 << cnt))        /* port `bit_cnt' changed? ... */
                usb_port_status_change (hub, bit_cnt - 1);  /* notice it */
        }

        /* 
         * Now we are done, we re-schedule a read to the status change
         * endpoint. Errors are handled by it.
         */

    atomic_dec (&hub->in_progress);
    usb_hub_schedule_status_change (hub);
    return 0;
}


    /* Callback for hub endpoints
     *
     *  This is the callback function used by reads from the hub's
     * status change endpoint, so it will be under irq.
     *
     *  What it does is to find an empty slot in the dqueue and
     * fill it. Then it sets a command to order khubdd to run the
     * queue ASAP, and then wakes it up. If no slots are free, it
     * tells khubdd to complain.
     *
     * Parameters:  
     *
     *  ts_usb_tr*      Pointer to transaction buffer 
     *
     *  int             Cause of call (see USB_CB_* codes)
     *
     * Returns:
     *
     *  int             0 if ok, errno code < 0 on error.
     *
     * Global:
     *
     *  int usb_errno   On error, error code.  
     */

void usb_hub_status_callback (ts_usb_tr* tr, int cause)
{
    gdlf_fntrace ("usb_hub_status_callback (0x%p, %d)", tr, cause);
    gdlf_aassert (tr->open != NULL, return);
    gdlf_aassert (tr->open->dev != NULL, return);
    gdlf_aassert (tr->open->dev->cfel.data != NULL, return);
    gdlf_aassert (tr->open->dev->cfel.data->dtor.fn
                  == (tf_dtor)usb_hub_destroy, return);

        /*
         *  As the USB Spec 1.0 states on 11.8, the hub _must_ respond
         * with a NAK if all the change bits are zero (no changes),
         * thus, doing a "blocking" call. Let's hope this is well done
         * by all the hubs. 
         *
         *  In case some hub goes funny and doesn't do this, we could
         * optimize by seeing if there's any change (any bit of the
         * data is set to one). Thus, if the tr is finished
         * successfully and data is non-zero, we warn khubdd;
         * otherway, we tell khubdd just to re-schedule.
         */

    usb_hub_status_poll ((ts_usb_hub*)tr->open->dev->cfel.data);
    return;
}


    /* FIXME: Detailed desc
     *
     * Compliance:      USB Spec 1.0 xxxx
     *
     * Parameters:  
     *
     *  type            descr
     *
     * Global:
     *
     *  type name       descr
     *
     * Returns:
     *
     *  type            descr
     *
     * Global:
     *
     *  type name       descr
     */

void khubdd_insert (ts_usb_hub* hub, char com)
{
    volatile struct s_dqueue *itr, *top;

    gdlf_fntrace ("khubdd_insert (0x%p, %d)", hub, com);
    gdlf_aassert (hub != NULL, return);
    gdlf_aassert (com != 0, return);
    
        /* Search for a free slot on the queue and mark it */

    write_lock (&dqueue_lock);
    command = USB_HUB_COM_NO_SLOTS;
    for (itr = dqueue, top = itr + USB_HUB_DQUEUE_SLOTS; 
         itr < top; itr++)
        if (itr->hub == NULL)
        {
            itr->hub = hub;
            itr->command = com;
            command = 0;
            break;
        }
    write_unlock (&dqueue_lock);
    wake_up (&wq);
    return;
}


    /* Method for polling a hub
     *
     *  What it does is to find an empty slot in the dqueue and
     * fill it. Then it sets a command to order khubdd to run the
     * queue ASAP, and then wakes it up. If no slots are free, it
     * tells khubdd to complain.
     *
     * Parameters:  
     *
     *  ts_usb_hub*     Pointer to hub
     *
     *  int             Cause of call (see USB_CB_* codes)
     *
     * Returns:
     *
     *  int             0 if ok, errno code < 0 on error.
     */

void usb_hub_status_poll (ts_usb_hub* hub)
{
    gdlf_fntrace ("usb_hub_status_poll (0x%p)", hub);
    gdlf_aassert (hub != NULL, return);
    
    khubdd_insert (hub, USB_HUB_COM_HUB_PROCESS);
    return;
}


    /* FIXME: Detailed desc
     *
     * Compliance:      USB Spec 1.0 xxxx
     *
     * Parameters:  
     *
     *  type            descr
     *
     * Global:
     *
     *  type name       descr
     *
     * Returns:
     *
     *  type            descr
     *
     * Global:
     *
     *  type name       descr
     */

inline
void usb_hub_host_rm (ts_usb_host* host)
{
    gdlf_fntrace ("usb_hub_host_rm (0x%p)", host);
    gdlf_aassert (host != NULL, return);
    
    khubdd_insert (&host->root, USB_HUB_COM_HOST_RM);
    return;
}


    /* FIXME: Detailed desc
     *
     * Compliance:      USB Spec 1.0 xxxx
     *
     * Parameters:  
     *
     *  type            descr
     *
     * Global:
     *
     *  type name       descr
     *
     * Returns:
     *
     *  type            descr
     *
     * Global:
     *
     *  type name       descr
     */

inline
void usb_hub_host_enum (void)
{
    gdlf_fntrace ("usb_hub_host_enum (void)");

    write_lock (&dqueue_lock);
    command = USB_HUB_COM_HOST_ENUM;
    wake_up (&wq);
    write_unlock (&dqueue_lock);
    return;
}


    /* This is the kernel hub driver daemon. 
     *
     *  It will sit idle with real time priority, waiting for
     * something to happen. 
     *
     *  - it is told to exit, then it'll exit
     *
     *  - Else, it's been waken by the callbacks and there's data on
     *    the hub data queue, process it. It may be to process a hub,
     *    to enumerate buses searching for hosts or to remove hosts.
     *
     * Parameters:  
     *
     *  void*   Unused, but needed to comply with the prototyping.
     *
     * Global:
     *
     *  int pid         khubdd state
     *
     *  int command     Asynchronous commands to khubdd
     *
     *  struct wait_queue* hubd_wq
     *
     *                  Wait queue for the transactions to the hubs.
     *
     * Returns:
     *
     *  int             0 if ok, errno code < 0 on error.
     *
     */

int errno;              /* for close() syscall */

int khubdd (void* arg)
{
#if 0
    struct k_sigaction chld;
#endif
    volatile struct s_dqueue *itr;
    int errcode;
    int fd;
    ts_usb_host* host;
    
    gdlf_fntrace ("khubdd (0x%p)", arg);

    pkhubdd = current;
    pid = current->pid;
    gdlf_info ("Starting khubdd (pid %d)\n", pid);

        /*
         * FIXME: doing this is the only way I've found to make the
         *        daemon's name look ok. It's purely stetical, but I
         *        don't understand it. Anyone cares to explain it to
         *        me? (yeah, it puts us up in the kernel's mm context,
         *        ... but what does that exactly mean?)
         *
         * FIXME: if I do exit_files() and exit_fs(), the thing screws
         *        up and the child's kernel_thread() sicks both
         *        process (parent and child) into D state.
         *
         */

	lock_kernel();
	exit_mm (current);
#if 0
	exit_files (current);
	exit_fs (current);
#endif
	unlock_kernel();
    
	sprintf (current->comm, "khubdd");  /* we are called like this */
	current->session = 1;               /* and this is our session */
	current->pgrp = 1;                  /* Our group */
	current->leader = 1;
#warning FIXME: no use change decrement here?
	current->tty = NULL;
	current->tty_old_pgrp = 0;

        /* Ignore all signals, except HUP, we'll use it to poll */

    spin_lock (&current->sigmask_lock);
#if 0
    sigfillset (&current->blocked);         /* block _all_ */
    sigdelset (&current->blocked, SIGCHLD); /* but CHLD */ 
#endif
    flush_signals (current);                /* no signals _now_ */
#if 0
    chld.sa.sa_handler = NULL;              /* add a hanlder to SIGCHLD */
    sigemptyset (&chld.sa.sa_mask);         /* to evitate zombies */
    chld.sa.sa_flags = SA_NOCLDWAIT;
    if (do_sigaction (SIGCHLD, &chld, NULL) < 0 ) 
        gdlf_err ("Couldn't add handler for SIGCHLD! Expect zombies!\n");
#endif
    spin_unlock (&current->sigmask_lock);

        /* khubdd has all privileges ... FIXME: needed? */

	current->uid = current->fsuid = current->euid = 0;
	cap_set_full (current->cap_inheritable);
	cap_set_full (current->cap_effective);

        /* Close _all_ files (we don't need'em, anyway) :) */

	for (fd = 0; fd < current->files->max_fds; fd++ )
		if (current->files->fd[fd] != 0) 
            close (fd);

        /* khubdd response to hub changes must be fast */

        /* FIXME: is this the correct way to do it? I guess not */

	current->policy = SCHED_FIFO;       /* We want it to be a _fast_ RT task */
	current->priority = 32;             /* FIXME: adjust this priority */

        /* Now the main loop */

    while (command != USB_HUB_COM_STOP)
    {
        spin_lock (&current->sigmask_lock);
        if (signal_pending (current))
        {
            sigemptyset (&current->signal);
            recalc_sigpending (current);
        }
        spin_unlock (&current->sigmask_lock);
        interruptible_sleep_on (&wq);   /* wait for something to happen */
        
            /* 
             * Check for commands and hubs.
             *
             *  Each slot in the dqueue array has a hub pointer and a
             * command. We have to process the command. After
             * processing, we clean both the command and the pointer.
             */

        switch (command)
        {
          case 0:                       /* Nothing? see list */
            break;
            
          case USB_HUB_COM_STOP:        /* we should stop ... */
            continue;                   /* ... so let's go */
            
          case USB_HUB_COM_NO_SLOTS:    /* complain and run queue ... */
#warning FIXME: add a counter to evitate floodings
            gdlf_alert ("Ran out of hub queuing slots [with %d]\n",
                        USB_HUB_DQUEUE_SLOTS);
            break;            
            
                    /* This fires an enumeration of buses seaching hosts */

          case USB_HUB_COM_HOST_ENUM:
#warning FIXME: enumerate buses for hosts
            break;

          default:                      /* should not happen, jic */
            gdlf_err ("Unknown command 0x%x to khubdd\n", command);
            break;                      /* Don't check changes */
        }
        command = 0;
        
        write_lock (&dqueue_lock);            
        for (itr = dqueue; 
             itr < dqueue + USB_HUB_DQUEUE_SLOTS;
             itr++)
        {
            switch (itr->command)
            {
              case 0:                       /* empty */
                continue;
            
              case USB_HUB_COM_HUB_PROCESS: /* Run the queue ... */
                if (itr->hub == NULL)
                    gdlf_err ("dqueue slot 0x%p requests hub processing "
                              "with NULL hub!\n", itr);
                else
                {
#ifdef KHUBDD_THREAD_CHANGES
                    errcode =       /* it does, fire the handling thread */
                        kernel_thread ((int(*)(void*))
                                       usb_hub_deal_status_change, 
                                       itr->hub, 0);
#else
                    errcode =       /* it does, fire the handling thread */
                        usb_hub_deal_status_change (itr->hub);
                    sprintf (current->comm, "khubdd");
#endif
                    itr->hub = NULL;
                    
                    if (errcode < 0)
                    {
                            /*
                             * FIXME: Some more error handling here;
                             *        this could be used for DOS
                             *        attacks if the dqueue fills up.
                             */
                        gdlf_err ("Couldn't start hub status change "
                                  "handling thread: errno %d\n",
                                  errcode);
                        gdlf_err ("Will retry on next run\n");
                    }
                }
                break;

                    /* Remove the host for which `hub' is root */
            
              case USB_HUB_COM_HOST_RM:
                if (itr->hub == NULL)
                {
                    gdlf_err ("Pretended to remove a host "
                              "providing a NULL root hub\n");
                    break;
                }
                host = itr->hub->host;
                gdlf_aassert (host != NULL, break);
                if (&host->root != itr->hub)
                {
                    gdlf_err ("Cannot remove `root' hub 0x%p (%d %d), "
                              "it doesn't match host 0x%p (%d)'s "
                              "root hub 0x%p\n",
                              itr->hub, host->addr, 
                              itr->hub->device? itr->hub->device->addr : -1,
                              itr->hub->host, host->addr,
                              &host->root);
                    break;
                }
                usb_host_destroy (host);
                usb_kfree (host);
                break;
            
              default:                      /* should not happen, jic */
                gdlf_err ("Unknown specific command 0x%x to khubdd\n", 
                          command);            
                break;                      /* Don't check changes */
            }
            itr->command = 0;
        }
        write_unlock (&dqueue_lock);
    }

    pid = 0;         /* say we've stopped running */
    pkhubdd = NULL;
    return 0;
}


    /*  Start the kernel hubd driver daemon.
     *
     *  This will launch a new kernel thread who will be the hub
     * daemon. It will have a descriptor open for each hub on the
     * status change endpoint and will wait for any change to happen
     * on the hubs or to be sent a command. Then it will take the
     * corresponding actions. There is short of an state diagram at
     * the documentation directory of the uusbd distribution.
     *
     * Returns:
     *
     *  int             PID of the new thread if ok, errno code < 0 on
     *                  error. 
     *
     * Global:
     *
     *  int usb_errno   On error, error code.  
     */

int usb_hub_startd (void)
{
    gdlf_fntrace ("usb_hub_startd()");
    command = 0;
    init_waitqueue (&wq);
    return kernel_thread (khubdd, NULL, 
                          CLONE_FS | CLONE_FILES | CLONE_SIGHAND);
}


    /*  This function stops the kernel hub driver daemon.
     *
     *  It will signal the khubdd to stop and wait for it to die. If
     * it doesn't die, it will sleep and try again. If there's no way
     * to stop it, it will complain and return -1.
     *
     * Returns:
     *
     *  int             0 if ok, <0 on error.
     *
     * Global:
     *
     *  int usb_errno   On error, error code.  
     */

int usb_hub_stopd (void)
{
    int cnt = USB_HUB_WFHS_LOOPS;
    
    gdlf_fntrace ("usb_hub_stopd()");

    while (pid > 0 && cnt--)
    {
        command = USB_HUB_COM_STOP;             /* tell him to stop */
        wake_up (&wq);                          /* now make it notice it */
        schedule();                             /* ... c'mon! */
        if (pid > 0)                            /* and wait if not yet stopped, */
        {                                       /* then run the loop again */
            current->state = TASK_INTERRUPTIBLE;
            schedule_timeout (USB_HUB_WHFS_MSECS / 10);
            if (signal_pending (current))       /* Hmmm ... FIXME */
                return -EINTR;                  /* ... should I do this? */
        }    
    }
    command = 0;                    /* JIC */
    if (cnt == 0 && pid > 0)        /* Couldn't stop it? */
    {                               /* Ok, this should be impossible */
        gdlf_err ("Couldn't stop khubdd!\n");
        usb_errno = -USBE_HUB_CANT_STOP;
        return -EBUSY;
    }

    return 0;
}


/* Log:
 *
 * $Log: khubdd.c,v $
 * Revision 1.23  1998/11/14 17:14:58  inaky
 * Moved the include files to uusbd/include/uusbd, to ease integration
 * into the Linux kernel.
 *
 * Revision 1.22  1998/11/09 21:24:03  inaky
 * Moved to 2.1.127, no current->timeout, isntead, schedule_timeout().
 *
 * Revision 1.21  1998/10/31 02:07:56  inaky
 * Removed all the d_* pointers from ts_usb_cfel and collapsed them into
 * a single pointer; only one was really used.
 *
 * Revision 1.20  1998/10/21 23:58:07  inaky
 * Simplified expresions using a variable which was already there for
 * that ...
 *
 * Revision 1.19  1998/10/14 21:41:13  inaky
 * Made the khubdd be able of doing more things then processing changed
 * hubs, as removing hosts et all. Collected all the slot insertion into
 * a single function called by each public function.
 *
 * Moved the main khubdd loop to check for the main command [exit,
 * enumerate, error] and then each specific command on each slot.
 *
 * Revision 1.18  1998/09/24 15:51:01  inaky
 * Moved 'hub_port' to 'port'.
 *
 * Revision 1.17  1998/09/14 02:30:59  inaky
 * Did some more work on trying to fix the threaded zombies,
 * unsuccessfully.
 *
 * Revision 1.16  1998/09/02 21:08:36  inaky
 * Removed the polled hubs array in favour of timers in the HCDs. UHCI is
 * the only one requiring polling, so I remove it.
 *
 * Revision 1.15  1998/08/30 13:43:40  inaky
 * Fixed semaphore for hub modification.
 *
 * Revision 1.14  1998/08/28 14:30:22  inaky
 * Added a switch to have khubdd spawning threads for hadnling changes or
 * not. This is also a temporary fix for the zombies problem.
 *
 * Revision 1.13  1998/08/24 02:16:17  inaky
 * Removed fntraces.
 *
 * Revision 1.12  1998/08/19 23:44:47  inaky
 * Changed usb_hub_polled_del() behaviour to return the position the hub
 * was deleted from, zero based, or error (<0) if not found or any
 * error. This eased things.
 *
 * Added a flag to hubs to see if changes are still been processed in
 * them, so that slow processing hubs don't get overhelmed with various
 * change process threads.
 *
 * Tried to fiddle the signals to get the khubdd not going zombie, but
 * couldn't do ...
 *
 * Revision 1.11  1998/08/03 13:26:17  inaky
 * Moved the command variable to be volatile. Enabled the signals TERM,
 * KILl and QUIT to interrupt to interrupt the change processing
 * thread. This should enable interrupting it if it hangs with something
 * more than SIGHUP.
 *
 * Revision 1.10  1998/07/30 17:49:31  inaky
 * Added callback reason argument, and some debug stuff.
 *
 * Revision 1.9  1998/07/28 23:18:12  inaky
 * Silly pretty-printing fix :)
 *
 * Revision 1.8  1998/07/23 14:20:24  inaky
 * Renamed the ts_usb_trbuf struct to ts_usb_tr.
 *
 * Revision 1.7  1998/07/16 14:04:44  inaky
 * Moved from the old read/write interface to the new, request-based
 * interface.
 *
 * Revision 1.6  1998/07/14 21:45:38  inaky
 * Removed debug stuff and an unneeded FIXME.
 *
 * Revision 1.5  1998/07/14 14:53:11  inaky
 * Fixed khubdd to answer to SIGHUP for polling hubs.
 *
 * Fixed the process name of the status change dealing thread.
 *
 * Revision 1.4  1998/07/12 23:00:50  inaky
 * Fixed bits in the polleds code.
 *
 * Revision 1.3  1998/07/01 21:59:59  inaky
 * Added the list of polled hub and functions to deal with it.
 *
 * Modified the status change thread function to make it accept the hub
 * and then be directly spawned from the daemon loop.
 *
 * Fixed argv[0] setting, blocked all signals, made it go to no-tty
 * state, session-leader, FIFO RT priority and be fully privileged. Also
 * made the polling code.
 *
 * Revision 1.2  1998/06/21 20:10:27  inaky
 * Changed the flag arguments to kernel_thread(), but still
 * unlucky. Should try to look with care at docs and sources :)
 *
 * Revision 1.1  1998/06/20 21:54:50  inaky
 * Massive changes to split the old monolithic hubd driver into a hub
 * manager (hub wrappers and daemon) and a dynamic driver, which can be
 * replaced and its different parts reused.
 *
 * Revision 1.24  1998/05/24 12:07:13  inaky
 * Fixed spinlock declaration not to break on compilation.
 *
 * Revision 1.23  1998/05/02 00:52:58  inaky
 * Re-merged in the khubdd and modified the callback function to use it.
 *
 * Revision 1.22  1998/05/01 19:17:41  inaky
 * Fixed device address reporting on messages.
 *
 * Revision 1.21  1998/04/26 03:40:58  inaky
 * Fixed the callback to spawn a thread to the status change
 * handler. Guess it won't work.
 *
 * Revision 1.20  1998/04/13 22:18:40  inaky
 * Finished most of the hub driver; some updates still needed.
 *
 * Revision 1.19  1998/01/08 12:19:18  inaky
 * Added copyright strings to all important files
 *
 * Revision 1.18  1998/01/08 11:50:03  inaky
 * Adapted to the new configuration scheme
 *
 * Revision 1.17  1998/01/02 00:07:29  inaky
 * Moved to the new driver registration scheme
 *
 * Revision 1.16  1997/12/31 17:06:39  inaky
 * Removed some unneeded locks and fixed some others.
 *
 * Revision 1.15  1997/12/31 05:37:41  inaky
 * Fixed some craps.
 *
 * Revision 1.14  1997/12/30 20:14:59  inaky
 * Moved to the new lists code. Untested yet.
 *
 * Revision 1.13  1997/12/29 03:20:06  inaky
 * Using commands again. Compiles, unchecked.
 *
 * Revision 1.12  1997/12/29 03:02:09  inaky
 * Droping the use of signals as commands. It is too OS dependant and I'm
 * doing it wrong for sure. It's simpler using a command variable, and
 * enough for our needs.
 *
 * Revision 1.11  1997/12/29 00:26:15  inaky
 * Moving the khubdd to use signals as commands.
 *
 * Revision 1.10  1997/12/28 03:47:52  inaky
 * Some trouble here, catching a GPF :(
 *
 * Revision 1.9  1997/12/27 17:25:42  inaky
 * Finished moving to GDLF. Deprecated usb_printk and __LOCATION__
 *
 * Revision 1.8  1997/12/27 02:06:13  inaky
 * Moving everything to use the GDLF for logging and debug checkings.
 * Removed the global array of USB hosts and moved it to a linked
 * list. Added an address field to each USB host controller.
 * Removed lots of cruft that were accumulated and useless. Still some
 * missing, but may be useful someday.
 *
 * Revision 1.7  1997/12/13 20:56:37  inaky
 * Moving all the documentation to the .c files except the ones for data
 * structures and inline functions. Helps making the header files cleaner.
 *
 * Revision 1.6  1997/12/08 02:41:53  inaky
 * Moving the hubd to the new driver style
 *
 * Revision 1.5  1997/12/06 01:56:00  inaky
 * Added some notes I should remember
 *
 * Revision 1.4  1997/12/06 00:19:51  inaky
 * Almost finished the usbd/host, usbd/cfg and a lot of meat from
 * usbd/device. Also added the driver definition. Got to polish it yet
 * ...
 *
 * Revision 1.3  1997/11/29 20:21:59  inaky
 * Documented and almost fixes the read/write/abort API, documented a
 * damn lot and implemented some stuff
 *
 * Revision 1.2  1997/11/29 00:46:46  inaky
 * Documented almost all the hub basic API, added spin locks to
 * usb_host_register() and some more stuff. Going for more.
 *
 * Revision 1.1  1997/10/31 05:52:44  inaky
 * *** empty log message ***
 *
 * Revision 1.1  1997/10/02 21:29:17  inaky
 * Remade all the directory structure to make it cleaner when the USB
 * class drivers come.
 *
 * Revision 1.1  1997/02/22 22:25:44  inaky
 * *** empty log message ***
 *
 *
 */

