
/* USB Driver Stack v$Revision: 1.16 $ 
 *
 * (C) 1997 by Iaky Prez Gonzlez <inaky@peloncho.fis.ucm.es>
 *
 * Distributed under the terms of the GNU Public License Version 2
 * See files COPYRIGHT and COPYING at the root distribution directory
 *
 * $Id: mm.c,v 1.16 1998/11/14 17:15:00 inaky Exp $
 *
 * Changelog at the bottom 
 *
 *
 *     Memory management helpers
 *     -------------------------
 *
 * Functions and variables to help in tracking and debugging memory
 * failures. 
 */

#include <linux/slab.h>     /* kmalloc() && kfree() */

#include <uusbd/usb-mm.h>

#define GDLF_IDENT          "USB-MM"
#define GDLF_LOCAL          /* Define if want local settings */
#define GDLF_DEF_LOGLEVEL   gdlf_level_info
/* #define GDLF_DEF_LOGLEVEL   9 */
#define GDLF_ASSERTIONS     /* want assertions (#undef if not) */
#undef  GDLF_FNTRACE        /* want function traces */
#define GDLF_LOCATIONS      /* want locations in log messages */
#include <uusbd/gdlf.h>

#ifdef USB_KMALLOC
unsigned kmalloccnt = 0;    /* count of kmalloc call times */
unsigned kpagecnt = 0;      /* count of get_free_page call times */
#endif

int usb_mm_init (void)
{
    int errcode = 0;
    return errcode;
}

int usb_mm_shutdown (void)
{
    int errcode = 0;

#ifdef USB_KMALLOC          /* print statistics if requested */
    gdlf_info ("kmalloc()/kfree() count: %u\n", kmalloccnt);
    gdlf_info ("get_free_pages()/free_pages() count: %u\n", kpagecnt);
#endif
    return errcode;
}


    /*
     *  The way I'm using to get aligned memory is very risky. I'm
     * assuming I'm getting chunks of 2^n bytes in size. The Linux
     * Kernel allocator for i386 seems to follow this, so I'll assume
     * it's right, but however, it should be fixed to be more
     * generic. 
     *
     *  Ok, the recursion is a really _ugly_ trick. I'll just allocate
     * until I get an aligned one. Then I'll free the unaligned
     * ones. It may work, it may not ... 
     */

#define MAX_LEVELS 16

void* __usb_kmalloc_align_r (size_t size, int alignment, 
                             int prio, int depth)
{
    void *ptr, *rptr;

    gdlf_fntrace ("__usb_kmalloc_align_r (%u, %d, %d, %d)", 
                  size, alignment, prio, depth);
    if (depth > MAX_LEVELS)
        return NULL;
    
    ptr = kmalloc (size < alignment? alignment : size, prio);
    if (((unsigned long)ptr & (alignment-1)) != 0)
    {
        rptr = __usb_kmalloc_align_r (size, alignment, prio, depth+1);
        usb_kfree (ptr);
        ptr = rptr;
    }

#ifdef USB_KMALLOC
    if (ptr)
        kmalloccnt++;
#endif
    return ptr;
}

void* __usb_kmalloc_align (size_t size, int alignment, int prio)
{
    void* ptr;
    void* rptr;
    
    gdlf_fntrace ("__usb_kmalloc_align (%u, %d, %d)", 
                  size, alignment, prio);
    ptr = kmalloc (size < alignment? alignment : size, prio);
    if (((unsigned long)ptr & (alignment-1)) != 0)
    {
        rptr = __usb_kmalloc_align_r (size, alignment, prio, 0);
        usb_kfree (ptr);
        ptr = rptr;
    }
    
#ifdef USB_KMALLOC
    if (ptr)
        kmalloccnt++;
#endif
    return ptr;
}


void* __usb_realloc (void* old, size_t oldsize, size_t newsize, int pri)
{
	void *ptr;

    gdlf_fntrace ("__usb_realloc (0x%p, %u, %u, %d)", 
                  old, oldsize, newsize, pri);

	ptr = usb_kmalloc (newsize, pri);
    if (ptr == NULL)
		return NULL;

	memcpy (ptr, old, oldsize < newsize? oldsize : newsize);
	if (old)
        usb_kfree (old);
	return ptr;
}



    /* 
     * Routines for monitored memory "allocation" 
     */

#ifdef USB_KMALLOC

void* __usb_kmalloc (size_t size, int prio, const char* file, int line)
{
    void* ptr;

    gdlf_fntrace ("__usb_kmalloc (%u, %d)", size, prio);

    ptr = kmalloc (size, prio);
    if (ptr)
    {
        kmalloccnt++;
        gdlf_debug ("%s:%d: 0x%p: allocated %u bytes\n", 
                    file, line, ptr, size);
    }
    return ptr;
}


void* __usb_kmalloc_align_d (size_t size, int alignment, int prio, 
                             const char* file, int line)
{
    void *ptr;
	ptr = __usb_kmalloc_align (size, alignment, prio);
    if (ptr)
        gdlf_debug ("%s:%d: 0x%p: allocated %u bytes aligned to %d bytes\n", 
                    file, line, ptr, size, alignment);
    return ptr;
}


void* __usb_realloc_d (void* old, size_t oldsize, size_t newsize, int pri, 
                       const char* file, int line)
{
    gdlf_fntrace ("__usb_realloc_d (0x%p, %u, %u, %d)", 
                  old, oldsize, newsize, pri);

    gdlf_debug ("%s:%d: 0x%p: reallocating %u bytes to %u bytes\n", 
                file, line, old, oldsize, newsize);

	return __usb_realloc (old, oldsize, newsize, pri);
}


void __usb_kfree (void *ptr, const char *file, int line)
{
    gdlf_fntrace ("__usb_kfree (0x%p)", ptr);
    gdlf_debug ("%s:%d: 0x%p: freeing\n", 
                file, line, ptr);
    if (ptr)
        kmalloccnt--;
    else
        return;

    kfree (ptr);
    return;
}
 

    /* 
     * The order parameter is the exponent of two for the number of
     * pages we want to get (aka 2^order, 0 -> 1 page, 2 -> 2 pages
     * ... 
     */

unsigned long __usb_get_free_pages (int pri, unsigned long order, 
                                    int dma, const char *file, int line)
{
    unsigned long page;

    gdlf_fntrace ("__usb_get_free_pages (%d, %lu, %d)",
                  pri, order, dma);
    kpagecnt++;
    page = dma? __get_free_pages (pri, order) 
        : __get_dma_pages (pri, order);
    gdlf_debug ("get_free_page(): got order %ld page(s) at %lx (%s:%d)\n", 
                order, page, file, line);
    return page;
}


void __usb_free_pages (unsigned long addr, unsigned long order, 
                       const char *file, int line)
{
    gdlf_fntrace ("__usb_free_pages (%lu, %lu)",
                  addr, order);
    kpagecnt--;
    gdlf_debug ("free_page(): freeing page(s) at %lx (%s:%d)\n", 
                addr, file, line);
    free_pages (addr, order);
    return;
}


#endif /* USB_KMALLOC */




/*
 * $Log: mm.c,v $
 * Revision 1.16  1998/11/14 17:15:00  inaky
 * Moved the include files to uusbd/include/uusbd, to ease integration
 * into the Linux kernel.
 *
 * Revision 1.15  1998/10/14 21:41:55  inaky
 * Fixed function trace to include the depth.
 *
 * Revision 1.14  1998/09/14 02:31:37  inaky
 * Added a fallback for getting aligned memory, using recursive
 * allocation until an aligned one is obtained, then freeing the other
 * ones.
 *
 * Revision 1.13  1998/08/24 02:16:51  inaky
 * Changed debug levels.
 *
 * Revision 1.12  1998/08/19 18:58:56  inaky
 * The debug align allocation request function was increasing the
 * variable which kept the number of successful allocations.
 *
 * Revision 1.11  1998/08/19 11:34:28  inaky
 * Removed all the try to make an efence like mechanism for memory
 * allocation. It didn't work.
 *
 * Revision 1.10  1998/08/19 11:26:13  inaky
 * Tried to code an allocation system which took care of allocating the
 * memory in such a way it prevented drunk pointers to affect by hitting
 * protected pages above or below the reserved address. Failed. I'm
 * dedicating too much time to this and getting no results, so I'm
 * reverting back the changes.
 *
 * Revision 1.9  1998/08/03 14:14:11  inaky
 * Removed debug messages, as this module seems fine enough as for it.
 *
 * Revision 1.8  1998/07/16 14:05:12  inaky
 * Modifed the log messages to be more useful and easy to track.
 *
 * Revision 1.7  1998/07/12 22:45:41  inaky
 * Only free when reallocating if not reallocating from NULL.
 *
 * Revision 1.6  1998/07/01 21:36:17  inaky
 * Added a realloc function
 *
 * Revision 1.5  1998/02/08 20:58:00  inaky
 * Fixed for Linux 2.1.85+ protocol for getting memory pages (moved out
 * the dma parameter in favour of an allocation flag, splitting it in two
 * macros; using the macros).
 *
 * Revision 1.4  1998/01/08 12:19:07  inaky
 * Added copyright strings to all important files
 *
 * Revision 1.3  1997/12/29 00:25:36  inaky
 * Fixed unimplemented functions to return -ENOSYS and moving the khubdd
 * to use signals as commands.
 * Also added function traces on most files that had them disabled (donkey!)
 *
 * Revision 1.2  1997/12/27 14:09:11  inaky
 * Fixed the #includes
 *
 * Revision 1.1  1997/12/27 03:24:23  inaky
 * Moved the monitored memory management code to it's own module
 *
 * Revision 1.28  1997/12/27 02:07:00  inaky
 * Moving everything to use the GDLF for logging and debug checkings.
 * Removed the global array of USB hosts and moved it to a linked
 * list. Added an address field to each USB host controller.
 * Removed lots of cruft that were accumulated and useless. Still some
 * missing, but may be useful someday.
 *
 * Revision 1.27  1997/12/27 02:05:58  inaky
 * Moving everything to use the GDLF for logging and debug checkings.
 * Removed the global array of USB hosts and moved it to a linked
 * list. Added an address field to each USB host controller.
 * Removed lots of cruft that were accumulated and useless. Still some
 * missing, but may be useful someday.
 *
 * Revision 1.26  1997/12/08 02:40:57  inaky
 * Fixed some stuff of the #includes, modified the makefile targets to be
 * more flexible, tried to finish the device configuration section.
 *
 * Revision 1.25  1997/12/06 00:19:39  inaky
 * Almost finished the usbd/host, usbd/cfg and a lot of meat from
 * usbd/device. Also added the driver definition. Got to polish it yet
 * ...
 *
 * Revision 1.24  1997/11/29 20:21:47  inaky
 * Documented and almost fixes the read/write/abort API, documented a
 * damn lot and implemented some stuff
 *
 * Revision 1.23  1997/11/29 00:46:29  inaky
 * Documented almost all the hub basic API, added spin locks to
 * usb_host_register() and some more stuff. Going for more.
 *
 * Revision 1.22  1997/10/17 00:15:59  inaky
 * *** empty log message ***
 *
 * Revision 1.21  1997/10/02 22:29:36  inaky
 * Finally managed to understand the workings of the MODVERSIONS (ok, not
 * exactly the working, but how to make it work).
 *
 * Now I can go to sleep happy :) -- I know, I'm really sick --
 *
 * Revision 1.20  1997/10/02 21:29:00  inaky
 * Remade all the directory structure to make it cleaner when the USB
 * class drivers come.
 *
 * Revision 1.19  1997/10/01 21:28:10  inaky
 * Re-making a lot of stuff in the hadnling of host, hub and device
 * initialization and handling, following a more correct approach of
 * registering/unregistering. Also moved the usb_host array to be an
 * array of pointers, so dynamic reallocation of it is easier, and the
 * host structure needs not to be moved in memory. The same applies to
 * the device arrays for each host.
 *
 */
