
/* USB Driver Stack v$Revision: 1.3 $ 
 *
 * (C) 1997 by Iaky Prez Gonzlez <inaky@peloncho.fis.ucm.es>
 *
 * Distributed under the terms of the GNU Public License Version 2
 * See files COPYRIGHT and COPYING at the root distribution directory
 *
 * $Id: driver.c,v 1.3 1998/11/16 00:03:05 inaky Exp $
 *
 * Changelog at the bottom 
 *
 *
 *     OHCI stuff - Driver and host initialization
 *     -------------------------------------------
 */

#include <linux/init.h>         /* __init */
#include <linux/config.h>       /* Configuration */
#include <linux/module.h>       /* MOD_INC_USE_COUNT */
#include <linux/malloc.h>       /* GFP_KERNEL */
#ifdef CONFIG_PCI
#include <linux/pci.h>          /* PCI definitions */
#endif
#include <linux/ioport.h>       /* check_region(), release_region() */
#include <asm/errno.h>          /* -E* codes */

#include <uusbd/usbd.h>
#include <uusbd/usb-misc.h>
#ifdef CONFIG_PCI
#include <uusbd/usb-pci.h>
#endif
#include <uusbd/usb-hcd.h>


#define GDLF_IDENT          "USB-OHCI"
#define GDLF_LOCAL          /* Define if want local settings */
#define GDLF_DEF_LOGLEVEL   9
#define GDLF_ASSERTIONS     /* want assertions (#undef if not) */
#define GDLF_FNTRACE        /* want function traces */
#define GDLF_LOCATIONS      /* want locations in log messages */
#include <uusbd/gdlf.h>

#include "ohci.h"           /* this has to be include after GDLF */


ts_usb_hci* ohci = NULL;    /* pointer to the HCI structure */
atomic_t usecnt =           /* use count */
    ATOMIC_INIT (0);
const char *ohci_ctl_str = OHCI_CTL_STR; /* for request_irq() */
struct timer_list timer;    /* timer for clean ups */


    /*
     * OHCI timer function
     *
     *  This does cleanup tasks
     */

void ohci_timer (unsigned long last_time)
{
    ts_ohci_host *itr;
    
        /* We walk the list w/o locking, this may be dangerous */

    itr = (ts_ohci_host*)ts_slist_head (&ohci->hosts);
    while (itr)
    {
        itr = (ts_ohci_host*)ts_slist_next (&itr->snode);
        ohci_tb_hash_cleanup (&itr->hash, last_time);
    }
    timer.data = jiffies;
    mod_timer (&timer, jiffies + OHCI_TIMER*HZ);
    return;
}

    /*
     * Host timer function
     *
     *  This does some debug stuff when compiled in.
     */

#ifdef OHCI_DEBUG_TIMER
void ohci_hci_timer (unsigned long addr)
{
    gdlf_fntrace ("ohci_hci_timer (0x%lx)", addr);
#warning DEBUG: remove me
#if 0
    ts_ohci_host_printk (OHCI (addr));
#endif
    mod_timer (& OHCI (addr)->timer, jiffies + OHCI_DEBUG_TIMER*HZ);
    return;
}
#endif /* OHCI_DEBUG_TIMER */


    /* Destroy and free resources of an OHCI host
     *
     *  This must leave the OHCI host struct as unreferenced as
     * possible to the OHCI sub-sub-system (aka: removing it from all
     * lists). This way, after running this, the only reference should
     * be the one kept from the ts_usb_host structure.
     *
     * Parameters:  
     *
     *  ts_ohci_host*   Pointer to the OHCI host structure to
     *                  destroy. 
     *
     * Returns:
     *
     *  int             0 if ok, errno code < 0 on error.
     */

int ohci_host_destroy (ts_ohci_host* hci)
{
    int errcode;    
    gdlf_fntrace ("ohci_host_destroy (0x%p)", hci);

    ts_snode_ref (&hci->snode);
#ifdef OHCI_DEBUG_TIMER
    if (hci->timer.function)
        del_timer (&hci->timer);
#endif

        /* Ok, let's reset and disable IRQs */

    if (hci->regs.control)
    {
        le_writel (OHCI_CMDSTAT_HCR, hci->regs.control);
        le_writel (OHCI_IRQDIS_MIE, hci->regs.irq_disable);
    }
    errcode = usb_hci_unregister_host (ohci, &hci->snode);
    ts_snode_unref_nd (&hci->snode);
    ts_ohci_host_release (hci);
    return errcode;
}


    /* Shutdown an USB controller and it's bus. 
     *
     *
     * Parameters:  
     *
     *  ts_ohci_host*   Pointer to the host controller structure of
     *                  the host.
     *
     * Returns:
     *
     *  int             0 if ok, errno code < 0 on error.
     */

int ohci_hci_shutdown (ts_ohci_host* hci)
{
    gdlf_fntrace ("ohci_hci_shutdown (0x%p)", hci);
#warning FIXME: unimplemented
    return -ENOSYS;
}


    /* Take ride of device quirks ...
     *
     * Parameters:  
     *
     *  struct pci_dev* Pointer to PCI structure
     */


#ifdef CONFIG_PCI
void ohci_pci_quirks (struct pci_dev* dev)
{
    __u8 revision;
    
    gdlf_fntrace ("ohci_pci_quirks (0x%p)", dev);
    gdlf_aassert (dev != NULL, return);

    switch (dev->vendor)
    {
      case PCI_VENDOR_ID_CMD:
        switch (dev->device)
        { 
                /* disable debugging mode on CMD 670 rev 4 */

          case PCI_DEVICE_ID_CMD_670:
            if (pci_read_config_byte (dev, PCI_REVISION_ID, &revision)
                != PCIBIOS_SUCCESSFUL)
                gdlf_err ("Cannot determine revision for CMD 670 "
                          "controller at PCI bus %d dev %d "
                          "fn %d; if revision is 4, expect "
                          "problems, I cannot fix'em\n", 
                          dev->bus->number, PCI_SLOT (dev->devfn), 
                          PCI_FUNC (dev->devfn));
            else if (revision == 0x4)    /* disable debugging mode */
            {
                gdlf_info ("CMD 670rev4 found, disabling debug mode; "
                           "(PCI bus %d dev %d fn %d)\n", 
                           dev->bus->number, PCI_SLOT (dev->devfn), 
                           PCI_FUNC (dev->devfn));
                usb_pci_bits32 (dev, 0x4c, 0x0, 0x1);
            }
            break;

          default:
            gdlf_warning ("Unkown PCI USB device ID 0x%d from vendor 0x%d: "
                      "won't fine tune\n",
                      dev->device, dev->vendor);
        }
        break;

      default:
        gdlf_err ("Unkown vendor 0x%d for PCI USB device ID 0x%d: "
                  "won't fine tune\n",
                  dev->vendor, dev->device);
    }
    
    return;
}
#endif /* CONFIG_PCI */


    /* Initialize the structures for a PCI OHCI controller, filling the 
     * hci and hcitype parameters. Note that it will release nothing
     * previously adquired (aka: it assumes the hci member of the
     * ts_usb_host points to NULL).
     *
     * Parameters:  
     *
     *  ts_usb_host*    Pointer to the host structure which will hold
     *                  it
     *
     *  struct pci_dev* Pointer to PCI device structure.
     *
     * Returns:
     *
     *  int             0 if ok, errno code < 0 on error.
     */


#ifdef CONFIG_PCI
int ohci_pci_init (ts_usb_host *host, struct pci_dev* dev)
{
    ts_ohci_host* hci;
    unsigned long membase;
    int irq;        /* IRQ # */
    int errcode;
    
    gdlf_fntrace ("ohci_pci_init (0x%p, 0x%p)", host, dev);
    gdlf_aassert (dev != NULL, return -EINVAL);
    gdlf_aassert (host != NULL, return -EINVAL);

    hci = OHCI (host->hci);

    gdlf_aassert (hci != NULL, return -EINVAL);

        /* get iobase and irq */

    membase = dev->base_address[0] & PCI_BASE_ADDRESS_MEM_MASK;
    irq = dev->irq;
    if (irq == 0)   /* None assigned? */
    {               /* NOTE: Assuming zero means "unassigned" */ 
        gdlf_err ("Host controller "
                  "at PCI bus %d, function %d "
                  "has assigned IRQ 0!\n", 
                  dev->bus->number, dev->devfn);
        return -ENXIO;
    }

    hci->membase = (unsigned long)ioremap_nocache (membase, 4 * 1024);
    if (hci->membase == 0)
    {
        gdlf_err ("Cannot ioremap 4kb for OHCI controler 0x%p "
                  "registers at 0x%08lx\n", hci, membase);
        return -EFAULT;
    }
    
    errcode = 
        request_irq (irq, 
                     (void(*)(int,void*,struct pt_regs*))ohci_irqhandler,
                     SA_SHIRQ, ohci_ctl_str, host);
    if (errcode < 0)
    {
        gdlf_err ("Cannot obtain IRQ %i requested by host controller "
                  "at PCI bus %d, function %d; errno code %d\n",
                  hci->irq, dev->bus->number, dev->devfn, errcode);
        return errcode;
    }    

        /* Let's set PCI stuff, enable mastering + mem access */

    usb_pci_bits16 (dev, PCI_COMMAND, 
                    PCI_COMMAND_MEMORY | PCI_COMMAND_MASTER,
                    PCI_COMMAND_MEMORY | PCI_COMMAND_MASTER);

        /* Set latency to 24 PCI Clocks [OHCI 1.0 A, PCI CONFIGURATION] */

#warning FIXME: I get PCI write errors here ... 
#if 0
    pci_write_config_word (dev, PCI_LATENCY_TIMER, 0x16);
#endif
        /* all ok, set it */

    hci->pci = dev;                 /* store the PCI info */
    hci->irq = irq;                 /* ... data */
    hci->host = host;

        /* Let's go for the HW quirks */

    ohci_pci_quirks (dev);    
    return 0;
}
#endif /* CONFIG_PCI */



    /* Initialize a given OHCI compliant USB host controller.
     *
     *  If the OHCI is detected and properly initialized, it will
     * register the host controller within the OHCI host controllers
     * list.
     *
     * Parameters:  
     *
     *  ts_usb_host*    The USB Host structure; it expects that the
     *                  HCI pointer of the structure points to an HCI
     *                  struct, filled up with the PCI, IO and IRQ
     *                  data, as it is used for initialization. The
     *                  other field must be clean (ohas left by the
     *                  ts_ohci_host_clean() macro).
     *
     * Returns:     
     *
     *  int             0 if ok (OHCI compliant controller detected
     *                  and initialized); < 0 errno code on error.
     */

int ohci_host_init (ts_usb_host* host)
{
    __u32 val;
    int cnt, order, errcode;
    ts_ohci_host *hci;
    
    gdlf_fntrace ("ohci_host_init (0x%p)", host);
    gdlf_aassert (host != NULL, return -EINVAL);
    gdlf_aassert (host->hci != NULL, return -EINVAL);
    
    hci = OHCI (host->hci);
    host->hcitype = e_usb_hci_ohci;
    gdlf_aassert (hci->membase != 0, return -EINVAL);
    
        /* init host->regs _this is crappy :)_ */
    
    for (cnt = 0; cnt < OHCI_REGS_DWORDS; cnt++)
        ((unsigned long*)&hci->regs)[cnt] = hci->membase + cnt*0x4;
    
        /* check revision */

    val = le_readl (hci->regs.revision);
    if ((val & OHCI_REVISION) != 0x10)
    {
        gdlf_err ("Unsupported OHCI version %x.%x on "
                  "controller 0x%p\n",
                  val & OHCI_REVISION & 0xf0 >> 4,
                  val & OHCI_REVISION & 0x0f, hci);
        return -EINVAL;
    }
    
        /* allocate HCCA on imposed alignment [OHCI 1.0 7.2.1] */

    le_writel (~0x0, hci->regs.hcca);
    val = le_readl (hci->regs.hcca);
    hci->hcca = ts_ohci_hcca_init (~val + 1);
    if (hci->hcca == NULL)
    {
        gdlf_err ("Cannot allocate an HCCA for OHCI "
                  "host 0x%p\n", host);
        return -ENOMEM;
    }

    hci->iso_head = ts_ohci_ed_init();
    if (hci->iso_head == NULL)
    {
        gdlf_err ("Cannot allocate an isochronous list "
                  "head for OHCI host 0x%p\n", host);
        return -ENOMEM;
    }

        /* 
         * allocate dummy EDs for queueing interrupt EDs 
         *
         *  We have six lists [1, 2, 4, 8, 16 and 32 ms], queued in
         * descending order [32, 16, 8, 4, 2, 1]. 1 points to ISO
         * head. Thus, intr_head[0] is for order 0, interval 1 ms.
         */

    for (order = 0; order > OHCI_INTR_ORDER + 1; order++)
    {
        hci->intr_head[order] = ts_ohci_ed_init();
        if (hci->intr_head[order] == NULL)
        {
            gdlf_err ("Cannot allocate interrupt queue head "
                      "for interval %d on OHCI host 0x%p\n", 
                      1 << order, host);
            return -ENOMEM;
        }
            /* ed->skip bit already set on init() */
        hci->intr_head[order]->nexted =         /* link to smaller interval */
            kvoid2ptr (order == 0? hci->iso_head : 
                       hci->intr_head[order-1]) >> 4;
    }

        /* clear interrupt routing and reset */

    val = le_readl (hci->regs.control);
    le_writel (OHCI_CMDSTAT_HCR, hci->regs.cmd_status);
    usb_waitms (1);     /* max 10 us */

        /*
         * now we are suspended; in least then 2 ms we _must_ move to
         * operational state 
         *
         *  Initialize HCCA. Assign to each entry in such a way that
         * each 32 frames, one is called to intr_head[0], each 16 to
         * intr_head[1], each 8 to intr_head[2] ... and so on.
         */

    for (cnt = 0; cnt < OHCI_INTR_SLOTS; cnt++)
        for (order = OHCI_INTR_ORDER+1; order >= 0; order --)
            if (cnt % (1 << order) == 0)
            {
                hci->hcca->interrupt[cnt] = 
                    kvoid2ptr (hci->intr_head[OHCI_INTR_ORDER + 1 - order]);
                break;
            }

        /*
         * init registers 
         *
         *  Set the control register to have 2 control per bulk, no
         * lists enabled [none set, right? :)], no interrupt routing,
         * and remote wakeup connected and enabled.
         */

    val = le_readl (hci->regs.control);
    le_writel_mask (val, 
                    0x1 | OHCI_CONTROL_RWC | OHCI_CONTROL_RWE,
                    OHCI_CONTROL_CBSR | OHCI_CONTROL_PLE
                    | OHCI_CONTROL_IE | OHCI_CONTROL_CLE
                    | OHCI_CONTROL_BLE | OHCI_CONTROL_IR
                    | OHCI_CONTROL_RWC | OHCI_CONTROL_RWE,
                    hci->regs.control);

        /*
         * Interrupts on scheduling overrun, done_head writeback,
         * resume detect, unrecoverable error and root hub status
         * change. However, we keep them disabled until we have all
         * setup.
         */

    le_writel (~0, hci->regs.irq_disable);  /* disable all */
    le_writel (OHCI_IRQDIS_MIE, hci->regs.irq_disable);  /* disable all */
    le_writel (OHCI_IRQEN_SO | OHCI_IRQEN_WDH | OHCI_IRQEN_RD
               | OHCI_IRQEN_UE | OHCI_IRQEN_RHSC,
               hci->regs.irq_enable);

        /* Set HCCA, clean control and bulk heads */

    le_writel (kvoid2ptr (hci->hcca), hci->regs.hcca);
    le_writel (0, hci->regs.ctl_head);
    le_writel (0, hci->regs.bulk_head);
    le_writel (0, hci->regs.done_head);

        /* set largest data packet */

    val = le_readl (hci->regs.fm_interval);
#warning FIXME: find a suitable value for this, 1024 is random
    le_writel_mask (val, 1024 << 16, OHCI_FMINT_FSMPS, 
                    hci->regs.fm_interval);

        /*
         * Periodic xfers start at 10% of frame time [OHCI 1.0 7.3.4] 
         * [we set at 90% of fminterval cause fmremaining is a
         * down-counter. Please see OHCI 1.0 7.3.2 and 7.3.4 for
         * reference. Thanks to Paul Stewart <stewart@parc.xerox.com>
         * for pointing this out.
         */

    le_writel (((90*(val & OHCI_FMINT_FI))/100) & OHCI_PERSTART_PS, 
               hci->regs.periodic_start);

        /* disable legacy if present */
    
    if ((le_readl (hci->regs.revision) & OHCI_REVISION_LEG))
    {
        gdlf_info ("Disabling OHCI legacy support on "
                   "controller 0x%p\n", hci);        
        le_writel (0, hci->membase + OHCI_EMUCTL_OFFSET);
    }

        /* set the methods */

    host->reset = ohci_reset;
    host->setsof = ohci_setsof;
    host->reference = ohci_reference;

    host->endp_setdn = ohci_endp_setdn;
    host->endp_setup = ohci_endp_setup;

    host->tr_setdn = ohci_tr_setdn;
    host->tr_setup = ohci_tr_setup;

#if defined (CONFIG_USB_UHCI) || defined (CONFIG_USB_UHCI_MODULE)
    host->tr_insert = NULL;
    host->tr_remove = NULL;
#endif

    host->tr_run = ohci_tr_run;
    host->tr_reset = ohci_tr_reset;
    host->tr_abort = ohci_tr_abort;

        /* Now set up the root hub */

    host->root.host = host;                 /* our host */
    usb_host_ref (host);                    /* useless? perhaps */
        /* no device (it's already NULL) */
    host->root.cfg = host->root.iface = host->root.aset = 0;
    host->root.operations = &ohci_hub_operations;
        /* We don't call ohci_reference() to keep ref count at zero */

    errcode = ohci_hub_descriptor_fill (&host->root);   /* fill it's data */
    if (errcode < 0)                        /* trouble? */
    {
        gdlf_err ("Could not initialize root hub 0x%p "
                  "(%d -1)'s data: errno %d\n",
                  &host->root, host->addr,
                  errcode);
        host->root.statch_tr.data = NULL;
        host->root.statch_tr.size = 0;
            /* don't destroy, ohci_pci_enumerate() will do */
        return errcode;
    }

        /*
         *  The status change transaction is quite special in OHCI. We
         * aren't going to transfer anything trough the wire while
         * messing with the OHCI root hub, so we don't need to hold a
         * ts_usb_open descriptor. We only need to know that the hub
         * manager requires the transaction status to be DONE to
         * process it; so we leave it in that state; there's nothing
         * more we should care about. The hub destructor will notice
         * the device-less hubs and allow them to have this special
         * characteristic.
         */

    host->root.statch_tr.status = e_usb_tr_status_ok;
#if 0
    host->root.statch_tr.user = 0;
#endif
    host->root.statch_tr.size = sizeof (hci->tr_buffer);
    host->root.statch_tr.data = &hci->tr_buffer;
    host->root.statch_tr.open = &host->root.statch;
        /* do status check on _all_ root hub fields */
    hci->tr_buffer = ~0;

        /* Register into the HCI */
          
    errcode = usb_hci_register_host (ohci, &hci->snode);
    if (errcode < 0)
    {
        gdlf_err ("Could not register OHCI host controller "
                  "within the OHCI list of controllers: "
                  "errno: %d\n",
                  errcode);
        return errcode;
    }

        /* 
         *  Put it to run! 
         *
         *  We don't enable the short packet interrupt, because we are
         * going to get that data when accounting the finished TD.
         */

#ifdef OHCI_DEBUG_TIMER
    init_timer (&hci->timer);
    hci->timer.expires = jiffies + OHCI_DEBUG_TIMER*HZ;
    hci->timer.data = (unsigned long)hci;
    hci->timer.function = ohci_hci_timer;
    add_timer (&hci->timer);
#endif

            /* enable interrupts, start host */

    le_writel (OHCI_IRQEN_MIE, hci->regs.irq_enable);
    val = le_readl (hci->regs.control);
    le_writel_mask (val, OHCI_STATE_OPERATIONAL << 6, 
                    OHCI_CONTROL_HCFS, hci->regs.control);

    return 0;
}


    /* Enumerate the PCI buses
     *
     * 1. Walk the buses for OHCI controllers. 
     *
     * 2. On each found host, allocate a host and hci structs with
     *    the hw data filled in and call ohci_host_init
     *
     * 3. if succesful, usb_host_register() each host.
     *
     * Returns:
     *
     *  int             0 if ok, errno code < 0 on error.
     */

#ifdef CONFIG_PCI
int ohci_pci_enumerate (void)
{
    int bioscode;
    int errcode;
    int pcihosts;
    struct pci_dev *device;
    ts_usb_host *host;
    ts_ohci_host *itr;
    
    gdlf_fntrace ("ohci_pci_enumerate (void)");

    bioscode = 0;
    pcihosts = 0;

    if (!pci_present())
    {
        gdlf_err ("No PCI BIOS present, no PCI USB-OHCI hosts\n");
        return -ENODEV;
    }

    device = NULL;
    while ((device = pci_find_class (USB_PCI_CLASS | 0x10, device)))
    {
            /* Check it is not already allocated */

        itr = (ts_ohci_host*)ts_slist_head (&ohci->hosts);
        while (itr)
        {
            if (itr->pci == device)     /* same? */
                break;
            itr = (ts_ohci_host*)ts_slist_next (&itr->snode);
        }
        if (itr)            /* used? ... */
            continue;       /* ... oops */

        pcihosts++;
        gdlf_debug ("Host controller found at PCI bus %d dev "
                    "%d fn %d\n", 
                    device->bus->number, PCI_SLOT (device->devfn),
                    PCI_FUNC (device->devfn));

            /* Allocate USB and OHCI host structures */

        host = ts_usb_host_init();
        if (host == NULL)
        {
            gdlf_err ("Cannot get memory for USB host on OHCI "
                      "controller on PCI bus %d dev %d fn %d\n",
                      device->bus->number, PCI_SLOT (device->devfn),
                      PCI_FUNC (device->devfn));
            return -ENOMEM;
        }
        ts_dtor_fill (&host->snode.dtor, (tf_dtor)usb_host_destroy);
        
        host->hci = (ts_snode*)ts_ohci_host_init();
        if (host->hci == NULL)
        {
            usb_kfree (host);
            gdlf_err ("Cannot get memory for OHCI controller on "
                      "PCI bus %d dev %d fn %d\n",
                      device->bus->number, PCI_SLOT (device->devfn), 
                      PCI_FUNC (device->devfn));
            return -ENOMEM;
        }
        ts_dtor_fill (&host->hci->dtor, (tf_dtor)ohci_host_destroy);
                
            /* 
             * Assign hw resources to the PCI host (irqs and
             * ioregions)
             */
            
        errcode = ohci_pci_init (host, device);
        if (errcode < 0)                    /* PCI init failed? */
        {
            gdlf_err ("Failed doing PCI initialization of OHCI "
                      "host 0x%p on PCI bus %d dev %d fn %d: %d\n",
                      host, 
                      device->bus->number, PCI_SLOT (device->devfn),
                      PCI_FUNC (device->devfn), errcode);
            usb_host_destroy (host);
            usb_kfree (host);
            continue;
        }

            /* 
             * Assign USB resources, fill the structures and register
             * within the HCI
             */
        
        errcode = ohci_host_init (host);
        if (errcode < 0)                    /* init failed? */
        {
            gdlf_err ("Could not initialize OHCI host 0x%p "
                      "on PCI bus %d dev %d fn %d: %d\n",
                      host, 
                      device->bus->number, PCI_SLOT (device->devfn),
                      PCI_FUNC (device->devfn), errcode);
            usb_host_destroy (host);
            usb_kfree (host);
            continue;
        }
        
            /* init ok, register USB host */
        
        errcode = usb_host_register (host);
        if (errcode < 0)                    /* Register failed? */
        {
            gdlf_debug ("Couldn't register OHCI host 0x%p on "
                        "PCI bus %d dev %d fn %d: %d\n",
                        host, 
                        device->bus->number, PCI_SLOT (device->devfn),
                        PCI_FUNC (device->devfn), errcode);
            usb_host_destroy (host);
            usb_kfree (host);
            continue;
        }
        
            /* Register ok */

        gdlf_info ("Host Controller 0x%p (%d) at "
                   "PCI bus %d dev %d fn %d initialized\n",
                   host, host->addr,
                   device->bus->number, 
                   PCI_SLOT (device->devfn),
                   PCI_FUNC (device->devfn));
        ts_snode_ref (host->hci);               /* for `host->hci' */
        usb_hub_status_poll (&host->root);
        host = NULL;                        /* clean for next loop */
    }               /* PCI loop for hosts */
    return 0;
}
#endif /* CONFIG_PCI */


    /* Enumerate all the OHCI controllers in all buses you see.
     *
     * Returns:
     *
     *  int             0 if ok, < 0 on error.
     */

int ohci_enumerate (void)
{
    int errcode;
    
    gdlf_fntrace ("ohci_enumerate (void)");

#ifdef CONFIG_PCI
    errcode = ohci_pci_enumerate();
#else
    /* 
     * We still want to permit being able of initializing the USB
     * subsystem on a PCI-less system, thus the #warning instead
     * of an #error. I know it sounds stupid, but it doesn't hurt.
     */
#warning Actually there are no ways but PCI to get information 
#warning about OHCI USB hosts! You should enable PCI.
    errcode = -ENODEV;
 #endif /* CONFIG_PCI */
    return errcode;
}


    /* Initialize all the OHCI hosts found on the system, scanning all
     * the known bus types (currently, just PCI).
     *
     * Returns:
     *
     *  int             0 if ok, errno code < 0 on error.
     */

int __init ohci_init (void)
{   
    int errcode;
    
    gdlf_fntrace ("ohci_init()");

    gdlf_info ("USB Open Host Controller Interface Driver\n");
#ifdef MODULE
    gdlf_info ("(C) 1998 Inaky Perez Gonzalez "
               "<inaky@peloncho.fis.ucm.es>\n");
#endif

        /* Allocate and register an HCI */

    ohci = ts_usb_hci_init();           /* create one */
    if (ohci == NULL)                   /* oops? */
    {
        gdlf_debug ("HCI entry allocation failed\n");
        return -ENOMEM;
    }
    ts_snode_ref (&ohci->snode);        /* we keep a reference to it */
    ohci->name = OHCI_NAME;             /* set name ... */
    ohci->type = e_usb_hci_ohci;        /* ... type */
    
    errcode = usb_hci_register (ohci);  /* register it */
    if (errcode)                        /* oops? */
    {
        gdlf_err ("Couldn't register OHCI HCI: "
                  "errno/usb %d/%d\n",
                  errcode, usb_errno);
        ts_usb_hci_release (ohci);          /* free resources */
        usb_kfree (ohci);                   /* free */
        ohci = NULL;
        return errcode;                     /* sorry :) */
    }
    
        /* start the cleanup timer */
    
    init_timer (&timer);
    timer.expires = jiffies + OHCI_TIMER*HZ;
    timer.data = (unsigned long)jiffies;
    timer.function = ohci_timer;
    add_timer (&timer);

        /* Enumerate the known buses searching for OHCI */

    errcode = ohci_enumerate();
    if (errcode < 0)
        ohci_shutdown();
    return errcode;
}


    /* Force the release of all OHCI hosts
     *
     *  This function will walk the list of USB hosts searching for
     * OHCI ones and force it's unregistration, thus making the UUSBD
     * free of OHCI hosts. After this, it is guaranteed that the
     * USB-OHCI subsystem is completely free and unused, no matter if
     * any error code is returned.
     *
     *  However, we use a helper function :)
     *
     * Returns:
     *
     *  int             0 if ok, errno code < 0 on error.
     */

int ohci_release (void)
{
    int errcode = 0;
    
    gdlf_fntrace ("ochi_release (void)");

    if (ohci)
        errcode = usb_hci_release (ohci);
    return errcode;
}


    /*  Shutdown the USB-OHCI subsystem
     *
     *  This will clean up everything and shutdown the UCHI
     * subsystem. In fact, it will call ohci_release() to ensure that
     * all the OHCI hosts are freed and then exit.
     *
     * Returns:
     *
     *  int             0 if ok, errno code < 0 on error.
     */

int ohci_shutdown (void)
{
    int errcode = 0;

    gdlf_fntrace ("ohci_shutdown (void)");

    ohci_release();
    if (ohci)
    {   
        ts_snode_unref (&ohci->snode, NULL);
        errcode = usb_hci_destroy (ohci);
        usb_kfree (ohci);
    }
    del_timer (&timer);
    return errcode;
}


    /* Add/remove a reference to the driver
     *
     *  This function is used to tell the ohci driver if we are or not
     * using any of it's data structures or functions. When we take a
     * pointer to any of the driver's function in the operations
     * structure, we must reference (use) it, so we call this with a
     * positive argument. When we are done using them, we call it with
     * a negative argument.
     *
     * Parameters:  
     *
     *  ts_snode*       Pointer to the element with which we are using
     *                  the driver.  
     *
     *  __u8            Positive for ONE reference, negative for ONE
     *                  unreference. If zero, return use count.
     *
     * Returns:
     *
     *  int             0 if ok, errno code < 0 on error.
     */
    
int ohci_reference (ts_snode* snode, int no)
{
    gdlf_fntrace ("ohci_reference (0x%p, %d)", snode, no);
    
    if (no > 0)             /* reference */
    {
        atomic_inc (&usecnt);           /* a new use */
        MOD_INC_USE_COUNT;              /* module's too */
    }
    else if (no < 0)        /* unreference */
    {
        atomic_dec (&usecnt);           /* a new use */
        MOD_DEC_USE_COUNT;              /* module's too */
    }
    return atomic_read (&usecnt);
}


    /* Modular stuff */

#ifdef MODULE
int init_module (void)
{
    return ohci_init();
}

void cleanup_module (void)
{
    ohci_shutdown();
    return;
}
#endif


/*
 * $Log: driver.c,v $
 * Revision 1.3  1998/11/16 00:03:05  inaky
 * Moved the include files to uusbd/include/uusbd, to ease integration
 * into the Linux kernel. Added the tb hash cleaning up to timers, fixed
 * the initialization of periodic_start register, added the endpoint
 * setup/dn stuff and removed device setup/dn stuff. Moved all host
 * ref/unref to usb_host_[un]ref().
 *
 * Revision 1.2  1998/10/14 21:50:38  inaky
 * Fixed PCI error/warning messages when fixing quirks.
 *
 * Revision 1.1  1998/09/14 20:31:12  inaky
 * Added all the code for OHCI, a lot missing to complete yet.
 *
 */
