
/* USB Driver Stack v$Revision: 1.3 $ 
 *
 * (C) 1997 by Iaky Prez Gonzlez <inaky@peloncho.fis.ucm.es>
 *
 * Distributed under the terms of the GNU Public License Version 2
 * See files COPYRIGHT and COPYING at the root distribution directory
 *
 * $Id: ohci.h,v 1.3 1998/11/16 00:04:26 inaky Exp $
 *
 * Changelog at the bottom 
 *
 *
 *     OHCI stuff
 *     ----------
 *
 * Stuff related to OHCI compliant controllers.
 *
 */

#ifndef _ohci_h_
#define _ohci_h_

#include <linux/config.h>
#include <linux/timer.h>

    /*
     *  On little endian, gcc aligns first fields at the lowest
     * addresses (that's little endian :). Still have to see if on big
     * endian it does it the other way around or what happens ... :(
     */

#ifdef __BIG_ENDIAN
#warning FIXME: Bitfields may break! Big endian untested!
#endif

#include <uusbd/usb.h>
#include <uusbd/usb-pci.h>
#include <uusbd/usb-hub.h>
#include <uusbd/usb-misc.h>


    /* 
     * Control of the OHCI module 
     * --------------------------
     */

#define OHCI_DEBUG              /* Want debug stuff on OHCI */
#undef OHCI_DEBUG_TIMER         /* Want debug stuff on OHCI? (each seconds) */
#if 0
#define OHCI_DEBUG_TIMER 10
#endif
#undef OHCI_SBYS                /* Want step by step */
#define OHCI_TIMER      120     /* Cleanup tasks fires each X seconds */
#define OHCI_RHSC_BUG           /* enable bug workaround */

    /* definitions */

#define OHCI_NAME       "OHCI"
#define OHCI_CTL_STR    USB_CTL_STR "-ohci"
#define OHCI_INTR_SLOTS 32
#define OHCI_INTR_ORDER 5
#define OHCI(hci) ((ts_ohci_host*)(hci))


    /*
     * OHCI Registers [OHCI 1.0 7]
     * ---------------------------
     *
     *  These are memory mapped, so access them with le_readl()/le_writel()
     * [only dword access allowed]. The values in the structure are
     * the register's addresses, after translation with the base
     * memory address [so, only i.e. `le_readl (hci->regs.revision)'
     * reads the value]. Hope it's faster then reading
     * membase+adding.
     *
     *  Port status is an array of as many dwords as ports in the
     * system. To access these ones, you have to add the port number
     * (zero based) to the base address rhub_pstatus_base.
     *
     *  Reserved field's values _MUST_ be preserved.
     */

typedef struct
{
    unsigned long   revision;
    unsigned long   control;
    unsigned long   cmd_status;     /* command status */
    unsigned long   irq_status;     /* interrupt status */
    unsigned long   irq_enable;
    unsigned long   irq_disable;

        /* These pointers are _physical_ (cpu_to_le32 (virt_to_bus())) */

    unsigned long   hcca;           /* ptr to hcca */
    unsigned long   periodic_cur;   /* ptr to current periodic ED */
    unsigned long   ctl_head;       /* head ptr of control list */
    unsigned long   ctl_cur;        /* ptr to current control ed */
    unsigned long   bulk_head;
    unsigned long   bulk_cur;
    unsigned long   done_head;      /* ptr to last done TD */

    unsigned long   fm_interval;    /* frame interval? */
    unsigned long   fm_remaining;   /* */
    unsigned long   fm_number;      /* frame number? */
    unsigned long   periodic_start; /* */
    unsigned long   ls_threshold;   /* */

    unsigned long   rhub_descr_a;   /* Root Hub descriptor A */
    unsigned long   rhub_descr_b;   /* Root Hub descriptor B */
    unsigned long   rhub_status;    /* Root Hub status */
    unsigned long   rhub_pstatus_base;  /* Port's status base */
} ts_ohci_regs;

#define OHCI_REGS_DWORDS (sizeof (ts_ohci_regs)/sizeof (unsigned long))

#ifdef __KERNEL__
#define ts_ohci_regs_nozero(ed)                 \
do                                              \
{                                               \
} while (0)

#define ts_ohci_regs_clean(ed)                  \
do                                              \
{                                               \
    memset ((ed), 0, sizeof (*(ed)));           \
    ts_ohci_regs_nozero (ed);                   \
} while (0)

static inline
ts_ohci_regs* ts_ohci_regs_init (void)
{
    ts_ohci_regs* regs;
    gdlf_fntrace ("ts_ohci_regs_init (void)");

    regs = usb_kmalloc (sizeof (*regs), GFP_KERNEL);
    if (regs == NULL)
        return NULL;
    ts_ohci_regs_clean (regs);
    return regs;
}

static inline
void ts_ohci_regs_release (ts_ohci_regs* regs)
{
    gdlf_fntrace ("ts_ohci_regs_release (0x%p)", regs);
    gdlf_aassert (regs != NULL, return);
    return;
}
#endif /* __KERNEL__ */

    /*
     * These are the masks to access specific bit(s) on registers. On
     * docs, `n' is the value obtained when reading the register and
     * masking it with _that_ mask.
     */

    /* HcRevision - revision R [OHCI 1.0 7.1.1] */

#define OHCI_REVISION       0x000000ff  /* mask to read BCD revision */
#define OHCI_REVISION_LEG   0x00000100  /* 1 - legacy supported */

    /* HcControl - control RW [OHCI 1.0 7.1.2] */

#define OHCI_CONTROL_CBSR   0x00000003  /* control:bulk service ratio;
                                         * it's n+1:1 */
#define OHCI_CONTROL_PLE    0x00000004  /* periodic (intr/iso) enable */
#define OHCI_CONTROL_IE     0X00000008  /* iso enable [on nxt frame] */
#define OHCI_CONTROL_CLE    0x00000010  /* ctl enable [on nxt frame] */
#define OHCI_CONTROL_BLE    0x00000020  /* bulk enable [on nxt frame] */
#define OHCI_CONTROL_HCFS   0x000000c0  /* hc state: 0 RESET, 1 RESUME,
                                         * 2 OPERATIONAL, 3 SUSPEND */
#define OHCI_CONTROL_IR     0x00000100  /* IRQ routing 0 - normal, 1 - SMM */
#define OHCI_CONTROL_RWC    0x00000200  /* remote wake up 1 - yes */
#define OHCI_CONTROL_RWE    0x00000400  /* remote wake up enable */

#define OHCI_STATE_RESET        0x0
#define OHCI_STATE_RESUME       0x1
#define OHCI_STATE_OPERATIONAL  0x2
#define OHCI_STATE_SUSPEND      0x3

    /* HcCommandStatus - cmd_status RWC [OHCI 1.0 7.1.3] */

#define OHCI_CMDSTAT_HCR    0x00000001  /* HC reset (10us), moves to SUSPEND */
#define OHCI_CMDSTAT_CLF    0x00000002  /* ctl list filled 1 - there are
                                         * TDs ready to go on ctl list */
#define OHCI_CMDSTAT_BLF    0x00000004  /* bulk list filled 1 - ... */
#define OHCI_CMDSTAT_OCR    0x00000008  /* Ownership change request */
#define OHCI_CMDSTAT_SOC    0x00030000  /* Scheduling overrun count */

    /* HcInterruptStatus - irq_status RWC [OHCI 1.0 7.1.4] */

#define OHCI_IRQSTAT_SO     0x00000001  /* scheduling overrun */
#define OHCI_IRQSTAT_WDH    0x00000002  /* done_head written */
#define OHCI_IRQSTAT_SF     0x00000004  /* SOF generated */
#define OHCI_IRQSTAT_RD     0x00000008  /* upstream resume detected */
#define OHCI_IRQSTAT_UE     0x00000010  /* system error [reset me] */
#define OHCI_IRQSTAT_FNO    0x00000020  /* frame # overflow */
#define OHCI_IRQSTAT_RHSC   0x00000040  /* root hub status change */
#define OHCI_IRQSTAT_OC     0x40000000  /* ownership change requested */

    /*
     * HcInterruptEnable - irq_enable RWC [OHCI 1.0 7.1.5]
     *
     *  Direct correspondence with irq_stat, to enable them when set
     * to one.
     */

#define OHCI_IRQEN_SO       0x00000001
#define OHCI_IRQEN_WDH      0x00000002
#define OHCI_IRQEN_SF       0x00000004
#define OHCI_IRQEN_RD       0x00000008
#define OHCI_IRQEN_UE       0x00000010
#define OHCI_IRQEN_FNO      0x00000020
#define OHCI_IRQEN_RHSC     0x00000040
#define OHCI_IRQEN_OC       0x40000000
#define OHCI_IRQEN_MIE      0x80000000  /* Master enable */

    /*
     * HcInterruptDisable - irq_disable RWC [OHCI 1.0 7.1.5]
     *
     *  Direct correspondence with irq_stat, to disable them when set
     * to one.
     */

#define OHCI_IRQDIS_SO      0x00000001
#define OHCI_IRQDIS_WDH     0x00000002
#define OHCI_IRQDIS_SF      0x00000004
#define OHCI_IRQDIS_RD      0x00000008
#define OHCI_IRQDIS_UE      0x00000010
#define OHCI_IRQDIS_FNO     0x00000020
#define OHCI_IRQDIS_RHSC    0x00000040
#define OHCI_IRQDIS_OC      0x40000000
#define OHCI_IRQDIS_MIE     0x80000000  /* Master disable */

    /* HcHCCA - hcca RW [OHCI 1.0 7.2.1] */

#define OHCI_HCCA           0xfffffff00

    /* HcPeriodicCurrentED - periodic_cur R [OHCI 7.2.2] */

#define OHCI_PCED           0xffffffff0 /* ptr to current iso/intr ED */

    /* HcControlHeadED - ctl_head RW [OHCI 7.2.3] */

#define OHCI_CHED           0xffffffff0 /* ptr to ctl list first ED */

    /* HcControlCurrentED - ctl_cur R [OHCI 7.2.4] */

#define OHCI_CCED           0xffffffff0 /* ptr to current control ED */

    /* HcBulkHeadED - bulk_head RW [OHCI 7.2.5] */

#define OHCI_CHED           0xffffffff0 /* ptr to bulk list first ED */

    /* HcBulkCurrentED - bulk_cur R [OHCI 7.2.6] */

#define OHCI_BCED           0xffffffff0 /* ptr to current bulk ED */

    /* HcDoneHead - done_head R [OHCI 7.2.7] */

#define OHCI_BCED           0xffffffff0 /* ptr to last done TD */

    /* HcFmInterval - fm_interval RW [OHCI 1.0 7.3.1] */

#define OHCI_FMINT_FI       0x00003fff  /* interval between SOFs */
#define OHCI_FMINT_FSMPS    0x7fff0000  /* largest data pkt (in bits) */
#define OHCI_FMINT_FIT      0x80000000  /* Frame interval toggle [HCD sets
                                         * when changing _FI] */

    /* HcFmRemaining - fm_remaining R [OHCI 1.0 7.3.2] */

#define OHCI_FMREM_FR       0x00003fff  /* bit times to end frame */
#define OHCI_FMREM_FRT      0x80000000  /* loaded from OHCI_FMINT_FIT
                                         * when _FR reaches zero; HCD uses
                                         * it to sync FMINT_FI and _FR */

    /* HcFmNumber - fm_number R [OHCI 1.0 7.3.3] */

#define OHCI_FMNUMBER_FN    0x0000ffff  /* Current frame number */

    /* HcPeriodicStart - periodic_start RW [OHCI 1.0 7.3.4] */

#define OHCI_PERSTART_PS    0x00003fff  /* start iso/intr after this
                                         * bit time in frame (10%
                                         * fm_interval) */

    /* HcLSThreshold - ls_threshold RW [OHCI 1.0 7.3.5] */

#define OHCI_LSTH_LST       0x00000fff  /* bit time to start LS
                                         * transfers; HCD sets it up
                                         * in function of
                                         * xmission/setup overhead */

    /* HcRhDescriptorA - rhub_descr_a RW [OHCI 1.0 7.4.1] */

#define OHCI_RHDA_NDP       0x0000000f  /* Number of ports (1-15) */
#define OHCI_RHDA_NPS       0x00000010  /* pwr is 0 - switched, [see 
                                         * _PSM], 1 - on if HC on */
#define OHCI_RHDA_PSM       0x00000020  /* pwr is 0 - global, 1 - per-port */
#define OHCI_RHDA_DT        0x00000040  /* 0 - root hub is not compound */
#define OHCI_RHDA_OCPM      0x00000080  /* Overcurrent report [if
                                         * !_NOCP]: 0 - global, 1 -
                                         * per port */ 
#define OHCI_RHDA_NOCP      0x00000100  /* overcurrent protection: 0 -
                                         * no, 1 - yes */
#define OHCI_RHDA_POTPGT    0xff000000  /* time in 2*ms to wait for
                                         * pwr being ok */

    /* HcRhDescriptorB - rhub_descr_b RW [OHCI 1.0 7.4.2]
     *
     *  Two bitfields of 15 elements for a max of 15 ports. bit 0
     * reserved. 
     */

#define OHCI_RHDA_DR        0x0000ffff  /* 1 - device non removable  */
#define OHCI_RHDA_PPCM      0xffff0000  /* 1 - power controlled
                                         * individually, 0 - power
                                         * controlled globally */

    /* HcRhStatus - rhub_status RW [OHCI 1.0 7.4.3] */

#define OHCI_RHSTAT_LPS     0x00000001  /* read 0 (local pwr status),
                                         * write 1 turns global port's
                                         * pwr off */
#define OHCI_RHSTAT_OCI     0x00000002  /* 1 - global overcurrent */
#define OHCI_RHSTAT_DRWE    0x00008000  /* 1 - connect status
                                         * change sends a resume wake
                                         * up, 0 - doesn't */
#define OHCI_RHSTAT_LPSC    0x00010000  /* read 0, write 1 to turn pwr
                                         * on on all ports */
#define OHCI_RHSTAT_OCIC    0x00020000  /* RWC, over current indicator
                                         * change, ack writing 1 */
#define OHCI_RHSTAT_CRWE    0x80000000  /* write 1 to clear _DRWE */

    /* HcRhPortStatus - rhub_pstatus RW [OHCI 1.0 7.4.4] */

#define OHCI_RHPSTAT_CCS    0x00000001  /* (read) device is 1- not, 0
                                         * - connected, (write) 1 -
                                         * clear port enable status */
#define OHCI_RHPSTAT_PES    0x00000002  /* (read) port is 0 -
                                         * disabled, 1 - enabled,
                                         * (write) 1 - enable port */
#define OHCI_RHPSTAT_PSS    0x00000004  /* (read) port is 0 - not, 1 -
                                         * suspended, (write) 1 -
                                         * suspend port */
#define OHCI_RHPSTAT_POCI   0x00000008  /* (read) port is 0 - not, 1 -
                                         * in overcurrent, (write) 1 -
                                         * resume port */
#define OHCI_RHPSTAT_PRS    0x00000010  /* (read) port is 0 - not, 1 -
                                         * being reset, (write) 1 -
                                         * reset port */
#define OHCI_RHPSTAT_PPS    0x00000100  /* (read) power is 0 - off, 1
                                         * - on, (write) 1 - power on */
#define OHCI_RHPSTAT_LSDA   0x00000200  /* (read) device is 0 - full,
                                         * 1 - low speed, (write) 1 -
                                         * clear port power status */
#define OHCI_RHPSTAT_CSC    0x00010000  /* 0 - no, 1 - change in
                                         * connect status */ 
#define OHCI_RHPSTAT_PESC   0x00020000  /* port enable status 0 - not,
                                         * 1 - changed.*/
#define OHCI_RHPSTAT_PSSC   0x00040000  /* port resume is 0 - not, 1 -
                                         * completed */
#define OHCI_RHPSTAT_OCIC   0x00080000  /* overcurrent indicator 0 -
                                         * not, 1 - changed */
#define OHCI_RHPSTAT_PRSC   0x00100000  /* port reset is 0 - not, 1 -
                                         * completed */

    /* HceControl Emulation RW [OHCI 1.0 B] */

#define OHCI_EMUCTL_OFFSET  0x00000100  /* offset of EMUCTL register */
#define OHCI_EMUCTL_EE      0x00000001  /* 1 - enable, 0 - disable emulation */
    /* rest of bits ignored by me :), we don't use'em */


    /* 
     * OHCI Endpoint Descriptor 
     * ------------------------
     *
     *  Set SKIP to one to modify HEADP, to remove from ed_list, 
     * set TOGGLECARRY or setting HALTED.
     *  
     * WARNING: Must be 16-byte aligned!!
     */

#define OHCI_ED_DIR_FTD1  0x0     /* get from TD */
#define OHCI_ED_DIR_OUT   0x1
#define OHCI_ED_DIR_IN    0x2
#define OHCI_ED_DIR_FTD2  0x3     /* get from TD */

typedef struct s_ohci_ed ts_ohci_ed;
struct s_ohci_ed
{
        /* dword 0 */

    __u32   address:7;      /* fa - device address 0-127 */
    __u32   endpoint:4;     /* en - endpoint number 0-15 */
    __u32   direction:2;    /* d - direction code, use OHCI_ED_DIR_* */
    __u32   speed:1;        /* s - 0 - full speed, 1 - low speed */
    __u32   skip:1;         /* k - 0 - don't, 1 - do  skip this ED */
    __u32   format:1;       /* f - 0 - ctl/bulk/interrupt, 1 - iso */
    __u32   maxpktsize:11;  /* mps - max packet size */
    __u32   unused:5;
    
        /* dword 1 */
    
    __u32   unused2:4;
    __u32   tailp:28;       /* ptr to last queued TD */

        /* dword 2 */

    __u32   halted:1;       /* 1 = ed processing halted by TD error */
    __u32   togglecarry:1;  /* c - data toggle of last TD */
    __u32   zero:2;
    __u32   headp:28;       /* ptr to first queued TD */
    
        /* dword3 */

    __u32   unused3:4;
    __u32   nexted:28;      /* ptr to next ED */

        /* admin */
    
    rwlock_t    lock;
};

#ifdef __KERNEL__
#define ts_ohci_ed_nozero(ed)                   \
do                                              \
{                                               \
    (ed)->skip = 1;                             \
} while (0)

#define ts_ohci_ed_clean(ed)                    \
do                                              \
{                                               \
    memset ((ed), 0, sizeof (*(ed)));           \
    ts_ohci_ed_nozero (ed);                     \
} while (0)

static inline
ts_ohci_ed* ts_ohci_ed_init (void)
{
    ts_ohci_ed* ed;
    gdlf_fntrace ("ts_ohci_ed_init (void)");

    ed = usb_kmalloc_align (sizeof (*ed), 16, GFP_KERNEL);
    if (ed == NULL)
        return NULL;
    ts_ohci_ed_clean (ed);
    return ed;
}

static inline
void ts_ohci_ed_release (ts_ohci_ed* ed)
{
    gdlf_fntrace ("ts_ohci_ed_release (0x%p)", ed);
    gdlf_aassert (ed != NULL, return);

    gdlf_assert (ed->tailp == 0);
    gdlf_assert (ed->headp == 0);
    gdlf_assert (ed->nexted == 0);
    return;
}
#endif /* __KERNEL__ */


    /* OHCI condition codes */

typedef enum
{
    e_ohci_cc_noerror = 0,
    e_ohci_cc_crc,
    e_ohci_cc_bitstuff,
    e_ohci_cc_dt_mismatch,
    e_ohci_cc_stall,
    e_ohci_cc_no_response,
    e_ohci_cc_pid_check,
    e_ohci_cc_pid_unexpected,
    e_ohci_cc_data_overrun,
    e_ohci_cc_data_underrun,
    e_ohci_cc_buffer_overrun = 0xc,
    e_ohci_cc_buffer_underrun,
    e_ohci_cc_not_accesed,
    e_ohci_cc_not_accesed2
} e_ohci_cc;


    /* 
     * OHCI Asynchronous Transfer Descriptor
     * -------------------------------------
     *
     *  Data toggle source (datatogglesrc) means get the data toggle
     * bit from the endpoint descriptor (ts_ohci_ed) or the transfer
     * descriptor (ts_ohci_atd.datatoggle).
     *
     *  Max size of a data buffer is 8k (two 4k pages).
     *
     * WARNING: Must be 16-byte aligned!!
     */

typedef enum
{
    e_ohci_td_type_async = 0,
    e_ohci_td_type_iso
} e_ohci_td_type;

#define OHCI_ATD_DIR_SETUP 0x0
#define OHCI_ATD_DIR_OUT   0x1
#define OHCI_ATD_DIR_IN    0x2

#define OHCI_ATD_DELAY_NOIRQ    0x7

typedef struct s_ohci_atd ts_ohci_atd;
struct s_ohci_atd
{
        /* dword 0 */

    __u32       index:12;       /* endpoint index in host hash */
    __u32       unused:4;
    __u32       type:2;         /* type of TD: 00 = ts_ohci_atd */ 
    __u32       rounding:1;     /* r - 0 - last data packet MUST fit buffer */
    __u32       direction:2;    /* dp - direction/pid to be used, OHCI_ATD_DIR* */
    __u32       delay:3;        /* di - frames to delay IRQ on complete TD */
    __u32       datatoggle:1;   /* d0 - data toggle to use (0 or 1) */
    __u32       datatogglesrc:1;/* d1 - datatoggle source is 0 - ED, 1 - TD */
    __u32       errors:2;       /* ec - inc on error, if 2+1 -> error fired */
    e_ohci_cc   cc:4;           /* cc - status of last attempted trasnaction */
    
        /* dword 1 */

    __u32       currentbp;      /* cbp - current buffer pointer; 0 means
                                 * zero length data packet/all xfered */

        /* dword 2 */

    __u32       zero:4;
    __u32       nexttd:28;      /* next TD in list */

        /* dword 3 */

    __u32       endbp;          /* physical address of last byte to xfer */
};

#ifdef __KERNEL__
#define ts_ohci_atd_nozero(atd)                 \
do                                              \
{                                               \
    (atd)->cc = e_ohci_cc_not_accesed;          \
} while (0)

#define ts_ohci_atd_clean(atd)                  \
do                                              \
{                                               \
    memset ((atd), 0, sizeof (*(atd)));         \
    ts_ohci_atd_nozero (atd);                   \
} while (0)

static inline
ts_ohci_atd* ts_ohci_atd_init (void)
{
    ts_ohci_atd* atd;
    gdlf_fntrace ("ts_ohci_atd_init (void)");

    atd = usb_kmalloc_align (sizeof (*atd), 16, GFP_KERNEL);
    if (atd == NULL)
        return NULL;
    ts_ohci_atd_clean (atd);
    return atd;
}

static inline
ts_ohci_atd* ts_ohci_atd_a_init (size_t size)
{
    ts_ohci_atd *tds;
    ts_ohci_atd *itr;
    
    gdlf_fntrace ("ts_ohci_atd_a_init (%u)", size);

    tds = usb_kmalloc_align (sizeof (ts_ohci_atd) * size, 16, GFP_KERNEL);
    if (tds == NULL)
        return NULL;
    
    for (itr = tds + size - 1; itr >= tds; itr--)
        ts_ohci_atd_clean (itr);

    return tds;
}

static inline
void ts_ohci_atd_release (ts_ohci_atd* atd)
{
    gdlf_fntrace ("ts_ohci_atd_release (0x%p)", atd);
    gdlf_aassert (atd != NULL, return);
    gdlf_aassert (atd->nexttd == 0, return);
    return;
}

static inline
void ts_ohci_atd_a_release (ts_ohci_atd* tds, size_t size)
{
    ts_ohci_atd *itr;
    
    gdlf_fntrace ("ts_ohci_atd_a_release (0x%p, %u)", tds, size);

    gdlf_aassert (tds != NULL, return);
    gdlf_aassert (size != 0, return);

    for (itr = tds + size - 1; itr >= tds; itr--)
        ts_ohci_atd_release (itr);

    return;
}
#endif /* __KERNEL__ */


    /* 
     * OHCI Isochronous Transfer Descriptor
     * ------------------------------------
     *
     *  Data toggle source (datatogglesrc) means get the data toggle
     * bit from the endpoint descriptor (ts_ohci_ed) or the transfer
     * descriptor (ts_ohci_itd.datatoggle).
     *
     *  Max size of a data buffer is 8k (two 4k pages).
     *
     * WARNING: Must be 32-byte aligned!!
     *          The allocated size will depend on how many packets to
     *          we have to send, so the array offpsw[] may be not
     *          full. The number of packets is at field `frame_count'.
     */

#define OHCI_ITD_DIR_SETUP 0x0
#define OHCI_ITD_DIR_OUT   0x1
#define OHCI_ITD_DIR_IN    0x2

#define OHCI_ITD_DELAY_NOIRQ    0x7

typedef struct s_ohci_itd ts_ohci_itd;
struct s_ohci_itd
{
        /* dword 0 */
    
    __u32       start_frame:16; /* sf - starting frame */
    __u32       type:2;         /* type of TD: 01 = ts_ohci_itd */ 
    __u32       unused:3;
    __u32       delay:3;        /* di - frames to delay IRQ on complete TD */
    __u32       frame_count:3;  /* fc - no. of frames - 1 described by TD */
    __u32       unused2:1;
    e_ohci_cc   cc:4;           /* cc - status of last attempted transaction */
    
        /* dword 1 */

    __u32       index:12;       /* endpoint index in host hash */
    __u32       bp0:20;         /* pointer to page where buffer starts */

        /* dword 2 */

    __u32       zero:4;
    __u32       nexttd:28;      /* next TD in list */

        /* dword 3 */

    __u32       endbp;          /* physical address of last byte to xfer */

        /* dwords 4, 5, 6 and 7  */
        /* See OHCI Spec 1.0, 4.3.2.3.1 */

    struct          /* description of offset/results for eight packets */
    {
        __u16       offset:11;    
        __u16       relative_to:1;
        e_ohci_cc   cc:4;
    } offpsw[8];
};

#ifdef __KERNEL__
#define ts_ohci_itd_nozero(itd)                 \
do                                              \
{                                               \
    (itd)->cc = e_ohci_cc_not_accesed;          \
    (itd)->type = e_ohci_td_type_iso;           \
} while (0)

#define ts_ohci_itd_clean(itd)                  \
do                                              \
{                                               \
    memset ((itd), 0, sizeof (*(itd)));         \
    ts_ohci_itd_nozero (itd);                   \
} while (0)

static inline
ts_ohci_itd* ts_ohci_itd_init (void)
{
    ts_ohci_itd* itd;
    gdlf_fntrace ("ts_ohci_itd_init (void)");

    itd = usb_kmalloc_align (sizeof (*itd), 32, GFP_KERNEL);
    if (itd == NULL)
        return NULL;
    ts_ohci_itd_clean (itd);
    return itd;
}

static inline
void ts_ohci_itd_release (ts_ohci_itd* itd)
{
    gdlf_fntrace ("ts_ohci_itd_release (0x%p)", itd);
    gdlf_aassert (itd != NULL, return);
    gdlf_assert (itd->nexttd == 0);
    return;
}
#endif /* __KERNEL__ */


    /* 
     * OHCI Transfer Block 
     * -------------------
     *
     *  This structure groups all the TDs needed to complete a
     * transfer.  
     */

typedef struct s_ohci_tb ts_ohci_tb;
struct s_ohci_tb
{
    ts_usb_tr*  tr;     /* Unreferenced! don't ts_snode_[un]ref() it! */
    size_t      index;  /* index in host hash list */
    union               /* First [placeholder] TD descriptor */
    {
        ts_ohci_atd *a; /* aynch tds */
        ts_ohci_itd *i; /* iso tds */
    } ftd;
    union               /* td descriptors */
    {
        ts_ohci_atd *a; /* aynch tds */
        ts_ohci_itd *i; /* iso tds */
    } td;
    size_t      tds;
};

#ifdef __KERNEL__
#define ts_ohci_tb_nozero(tb)                   \
do                                              \
{                                               \
    (tb)->index = -1;                           \
} while (0)

#define ts_ohci_tb_clean(tb)                    \
do                                              \
{                                               \
    memset ((tb), 0, sizeof (*(tb)));           \
    ts_ohci_tb_nozero (tb);                     \
} while (0)

static inline
ts_ohci_tb* ts_ohci_tb_init (void)
{
    ts_ohci_tb* tb;
    gdlf_fntrace ("ts_ohci_tb_init (void)");

    tb = usb_kmalloc_align (sizeof (*tb), 16, GFP_KERNEL);
    if (tb == NULL)
        return NULL;
    ts_ohci_tb_clean (tb);
    return tb;
}

static inline
void ts_ohci_tb_release (ts_ohci_tb* tb)
{
    gdlf_fntrace ("ts_ohci_tb_release (0x%p)", tb);
    gdlf_aassert (tb != NULL, return);

        /* makes no difference to use td.i or td.a */

    if (tb->td.i)
        usb_kfree (tb->td.i);
    return;
}
#endif /* __KERNEL__ */


    /* 
     * OHCI Host Controller Communication Area
     * ---------------------------------------
     *
     *  `frame' is updated before processing the periodic (interrupt
     * and iso) lists.
     *
     *  `done_head' is a physical pointer to the last retired TD, and
     * it is not written again until HCD clears the OHCI_IRQSTAT_WDH
     * bit. If on IRQ it is zero, then the interrupt cause is in the
     * OHCI_IRQ_STAT register. If non-zero, the IRQ is caused by a
     * `done_head' actualization by the HCD. If non-zero and bit 0 is
     * set to one, then aside from the done list actualization, you
     * have to take a look to the OHCI_IRQ_STAT register.
     *
     * WARNING: must be aligned to the alignment restrictions imposed
     *          by the hcca register!
     */

typedef struct s_ohci_hcca ts_ohci_hcca;
struct s_ohci_hcca
{
    __u32   interrupt[OHCI_INTR_SLOTS];  /* ptrs to interupt heads */
    __u16   frame;          /* current frame being processed */
    __u16   pad;            /* set to zero by host when `frame' is updated */
    __u32   done_head;      /* head of list of done TDs */
    __u8    reserved[116];  /* reserved for use by HC */
};

#ifdef __KERNEL__
#define ts_ohci_hcca_nozero(hcca)               \
do                                              \
{                                               \
} while (0)

#define ts_ohci_hcca_clean(hcca)                \
do                                              \
{                                               \
    memset ((hcca), 0, sizeof (*(hcca)));       \
    ts_ohci_hcca_nozero (hcca);                 \
} while (0)

static inline
ts_ohci_hcca* ts_ohci_hcca_init (size_t alignment)
{
    ts_ohci_hcca* hcca;
    gdlf_fntrace ("ts_ohci_hcca_init (%u)", alignment);
    gdlf_assert (alignment != 0);

    hcca = usb_kmalloc_align (sizeof (*hcca), alignment, GFP_KERNEL);
    if (hcca == NULL)
        return NULL;
    ts_ohci_hcca_clean (hcca);
    return hcca;
}

static inline
void ts_ohci_hcca_release (ts_ohci_hcca* hcca)
{
    gdlf_fntrace ("ts_ohci_hcca_release (0x%p)", hcca);
    gdlf_aassert (hcca != NULL, return);
    gdlf_assert (hcca->done_head == 0);
    return;
}
#endif /* __KERNEL__ */


    /* 
     * OHCI endpoint data
     * ------------------
     */

typedef struct
{
    ts_ohci_ed  ed;     /* MUST be first */
} ts_ohci_endp;

#ifdef __KERNEL__
#define ts_ohci_endp_nozero(endp)               \
do                                              \
{                                               \
    ts_ohci_ed_nozero (&(endp)->ed);            \
} while (0)

#define ts_ohci_endp_clean(endp)                \
do                                              \
{                                               \
    memset ((endp), 0, sizeof (*(endp)));       \
    ts_ohci_endp_nozero (endp);                 \
} while (0)

static inline
ts_ohci_endp* ts_ohci_endp_init (void)
{
    ts_ohci_endp* endp;
    gdlf_fntrace ("ts_ohci_endp_init (void)");

    endp = usb_kmalloc (sizeof (*endp), GFP_KERNEL);
    if (endp == NULL)
        return NULL;
    ts_ohci_endp_clean (endp);
    return endp;
}

static inline
void ts_ohci_endp_release (ts_ohci_endp* endp)
{
    gdlf_fntrace ("ts_ohci_endp_release (0x%p)", endp);
    gdlf_aassert (endp != NULL, return);

    ts_ohci_ed_release (&endp->ed);
    return;
}
extern void ts_ohci_endp_printk (ts_ohci_endp*);
#endif /* __KERNEL__ */


    /* 
     * OHCI endpoint hash
     * ------------------
     */

#define OHCI_TB_HASH_MAX_TBS 4096   /* max number of addressable entries */

typedef struct
{
    ts_ohci_tb  **ptb;
    size_t      size;
    rwlock_t    lock;
} ts_ohci_tb_hash;

#ifdef __KERNEL__
#define ts_ohci_tb_hash_nozero(tb_hash)             \
do                                                  \
{                                                   \
    __USB_SMP ((host)->lock = RW_LOCK_UNLOCKED);    \
} while (0)

#define ts_ohci_tb_hash_clean(tb_hash)          \
do                                              \
{                                               \
    memset ((tb_hash), 0, sizeof (*(tb_hash))); \
    ts_ohci_tb_hash_nozero (tb_hash);           \
} while (0)

static inline
ts_ohci_tb_hash* ts_ohci_tb_hash_init (void)
{
    ts_ohci_tb_hash* tb_hash;
    gdlf_fntrace ("ts_ohci_tb_hash_init (void)");

    tb_hash = usb_kmalloc (sizeof (*tb_hash), GFP_KERNEL);
    if (tb_hash == NULL)
        return NULL;
    ts_ohci_tb_hash_clean (tb_hash);
    return tb_hash;
}

static inline
void ts_ohci_tb_hash_release (ts_ohci_tb_hash* tb_hash)
{
    gdlf_fntrace ("ts_ohci_tb_hash_release (0x%p)", tb_hash);
    gdlf_aassert (tb_hash != NULL, return);

#ifdef GDLF_ASSERTIONS
    {
        int cnt;
        for (cnt = 0; cnt < tb_hash->size; cnt++)
            gdlf_passert (tb_hash->ptb[cnt] != NULL,
                          "hash 0x%p index %d\n",
                          tb_hash, cnt);
    }
#endif

    if (tb_hash->ptb)
        usb_kfree (tb_hash->ptb);
    return;
}
#endif /* __KERNEL__ */


    /* 
     * OHCI host
     * ---------
     */

typedef struct
{
    ts_snode        snode;
    unsigned long   membase;
    unsigned int    irq;    /* IRQ line */
    ts_usb_host     *host;  /* USB host pointer. RO only, no reference counting here */

#ifdef CONFIG_PCI
    struct pci_dev  *pci;
#endif
    atomic_t        sem;    /* semaphore for IRQ handling */
    unsigned        errors; /* hard errors (HCPE and PCIBE) */

    ts_ohci_regs    regs;
    ts_ohci_hcca    *hcca;          /* host controller communication area */
    __u16           tr_buffer;      /* to compose the change bitmap for hub */

        /* Endpoint lists */

    ts_ohci_ed      *intr_head[OHCI_INTR_ORDER+1];  /* dummy ed for intr heads */
    ts_ohci_ed      *iso_head;      /* dummy ed for ISO list head */
    ts_ohci_tb_hash hash;           /* tb hash for ts_ohci_[ai]td.index */
    rwlock_t        ctl_lock;
    rwlock_t        bulk_lock;
    rwlock_t        iso_lock;
    rwlock_t        intr_lock;

#ifdef OHCI_RHSC_BUG
    __u32           rhstatus;   /* hub status */
    __u32           *rhpstatus; /* ports status */
#endif
#ifdef OHCI_DEBUG_TIMER
    struct timer_list timer;
#endif
} ts_ohci_host;

#ifdef __KERNEL__
#define ts_ohci_host_nozero(host)               \
do                                              \
{                                               \
    ts_snode_nozero (&(host)->snode);           \
    ts_ohci_regs_nozero (&(host)->regs);        \
    ts_ohci_tb_hash_nozero (&(host)->hash);     \
    __USB_SMP (ctl_lock = RWLOCK_UNLOCKED);     \
    __USB_SMP (bulk_lock = RWLOCK_UNLOCKED);    \
    __USB_SMP (intr_lock = RWLOCK_UNLOCKED);    \
} while (0)

#define ts_ohci_host_clean(host)                \
do                                              \
{                                               \
    memset ((host), 0, sizeof (*(host)));       \
    ts_ohci_host_nozero (host);                 \
} while (0)

static inline
ts_ohci_host* ts_ohci_host_init (void)
{
    ts_ohci_host* host;
    gdlf_fntrace ("ts_ohci_host_init (void)");

    host = usb_kmalloc (sizeof (*host), GFP_KERNEL);
    if (host == NULL)
        return NULL;
    ts_ohci_host_clean (host);
    return host;
}

static inline
void ts_ohci_host_release (ts_ohci_host* host)
{
    int cnt;
    
    gdlf_fntrace ("ts_ohci_host_release (0x%p)", host);
    gdlf_aassert (host != NULL, return);

    ts_snode_release (&host->snode);
    if (host->membase)
        iounmap ((void*)host->membase);
    if (host->irq)
        free_irq (host->irq, host->host);
    gdlf_assert (atomic_read (&host->snode.usecnt) == 0);
    gdlf_assert (atomic_read (&host->sem) == 0);
    ts_ohci_regs_release (&host->regs);

    if (host->hcca)
    {
        ts_ohci_hcca_release (host->hcca);
        usb_kfree (host->hcca);
    }

    for (cnt = 0; cnt < OHCI_INTR_ORDER; cnt++)
        if (host->intr_head[cnt])
        {
            gdlf_assert (host->intr_head[cnt]->nexted 
                         == (cnt == 0? kvoid2ptr (host->iso_head)
                             : kvoid2ptr (host->intr_head[cnt-1])));
            ts_ohci_ed_release (host->intr_head[cnt]);
            usb_kfree (host->intr_head[cnt]);
        }    

    if (host->iso_head)
    {
        gdlf_assert (host->iso_head->nexted == 0);
        ts_ohci_ed_release (host->iso_head);
        usb_kfree (host->iso_head);
    }
#ifdef OHCI_RHSC_BUG
    if (host->rhpstatus)
        usb_kfree (host->rhpstatus);
#endif
#ifdef OHCI_DEBUG_TIMER
    del_timer (&host->timer);
#endif
    return;
}
extern void ts_ohci_host_printk (ts_ohci_host*);
#endif /* __KERNEL__ */


    /* 
     * OHCI functions 
     * --------------
     */

#ifdef __KERNEL__

    /* driver.c */

extern int ohci_init (void);
extern int ohci_shutdown (void);
extern int ohci_reference (ts_snode*, int);

    /* irq-handler.c */

extern void ohci_irqhandler (int, ts_usb_host*, struct pt_regs*);

    /* operations.c */

extern int ohci_reset (ts_usb_host*);
extern int ohci_setsof (ts_usb_host*, unsigned);

    /* scheduler.c */

extern int ohci_endp_setdn (ts_usb_device*, ts_usb_endp*);
extern int ohci_endp_setup (ts_usb_device*, ts_usb_endp*);

extern int ohci_tr_abort (ts_usb_tr*);
extern int ohci_tr_insert (ts_usb_tr*);
extern int ohci_tr_remove (ts_usb_tr*);
extern int ohci_tr_reset (ts_usb_tr*);
extern int ohci_tr_run (ts_usb_tr*);
extern int ohci_tr_setdn (ts_usb_tr*);
extern int ohci_tr_setup (ts_usb_tr*);

    /* lists */

extern ts_ohci_endp* ed_list_prev (ts_ohci_endp*, ts_ohci_endp*);

    /* vhub.c */

extern ts_usb_hub_ops ohci_hub_operations;
extern int ohci_hub_descriptor_fill (ts_usb_hub*);

    /* endp-hash.c */

extern int ohci_tb_hash_grow (ts_ohci_tb_hash*);
extern int ohci_tb_hash_set_slot (ts_ohci_tb_hash*, void*);
extern int ohci_tb_hash_cleanup (ts_ohci_tb_hash*, unsigned long);
static inline
void ohci_tb_hash_release_slot (ts_ohci_tb_hash* hash, size_t slot)
{
    gdlf_fntrace ("ohci_tb_hash_release_slot (0x%p, %u)", hash, slot);
    gdlf_aassert (hash != NULL, return);
    gdlf_aassert (slot < hash->size, return);
    hash->ptb[slot] = NULL;
}
static inline
ts_ohci_tb* ohci_tb_hash_get_slot (ts_ohci_tb_hash* hash, size_t slot)
{
    gdlf_fntrace ("ohci_tb_hash_get_slot (0x%p, %u)", hash, slot);
    gdlf_aassert (hash != NULL, return NULL);
    gdlf_aassert (slot < hash->size, return NULL);
    return hash->ptb[slot];
}

#endif /* __KERNEL__ */

#endif /* _ohci_h_ */

/* Log:
 *
 * $Log: ohci.h,v $
 * Revision 1.3  1998/11/16 00:04:26  inaky
 * Moved the include files to uusbd/include/uusbd, to ease integration
 * into the Linux kernel. Added the tb hash to TDs, endpoint data and
 * removed old device stuff.
 *
 * Revision 1.2  1998/09/14 20:31:13  inaky
 * Added all the code for OHCI, a lot missing to complete yet.
 *
 */

