
/* USB Driver Stack v$Revision: 1.2 $ 
 *
 * (C) 1997 by Iaky Prez Gonzlez <inaky@peloncho.fis.ucm.es>
 *
 * Distributed under the terms of the GNU Public License Version 2
 * See files COPYRIGHT and COPYING at the root distribution directory
 *
 * $Id: scheduler.c,v 1.2 1998/11/16 00:05:15 inaky Exp $
 *
 * Changelog at the bottom 
 *
 *
 *     OHCI stuff - Schedulement
 *     -------------------------
 *
 *
 *  Each ED has a TD which is a placeholder. Initially we allocate
 * one, and new scheduled TDs use that one as first and then append a
 * new one to use as placeholder. At the very end, when then endp is
 * closed, we have a placeholder left there by the last TD. We free
 * it [this is all done in ohci_endp_setup(), ohci_endp_setdn() and
 * taken into account in ohci_tb_fill(), ohci_tr_run().
 */


#include <uusbd/usbd.h>
#include <uusbd/usb-hub.h>
#include <uusbd/usb-misc.h>

#define GDLF_IDENT          "USB-OHCI"
#define GDLF_LOCAL          /* Define if want local settings */
#define GDLF_DEF_LOGLEVEL   9
#define GDLF_ASSERTIONS     /* want assertions (#undef if not) */
#define GDLF_FNTRACE        /* want function traces */
#define GDLF_LOCATIONS      /* want locations in log messages */
#include <uusbd/gdlf.h>

#include "ohci.h"


    /* Fill up a set of TDs for transfer
     *
     *  The only thing I don't like about OHCI is this point :). At
     * the endpoint's TD list, the last one is always unprocessed and
     * servers as a place holder. You fill in that one and append a
     * new empty one to serve as a placeholder. Then you update
     * endp->tailp and the old placeholder is ready to be processed.
     *
     *  So, to set up the xfer, I need to fill a list of TDs, but the
     * last one has to be allocated alone to be left as placeholder,
     * as I'll take over the ownership of the `old' placeholder I used
     * as my first TD, and so that last one will be left to the
     * following one to use it.
     *
     *  What I do then is to allocate a single TD to be the new
     * placeholder and I'll fill it like the first TD:
     *
     *  ts_ohci_td* ftd;
     *
     * then I allocate an array with the rest of the TDs [1..n-1]:
     *
     *  ts_ohci_td td[];
     *
     * so that the complete sequence of TDs is:
     *
     *  ftd, td[0] ... td[n-2]
     *
     * where 'n' is the total number of TDs we need. 'ftd' links to
     * 'td[0]', 'td[0]' to 'td[1]' ... until 'td[n-2]'. The
     * 'td[n-2].nexttd' is linked to 'ftd'; this way, when inserting
     * for execution, we just copy the contents of 'ftd' into the
     * placeholder and leave 'ftd' as new placeholder [there's no need
     * to clear it], without having to do any work, and being able to
     * refer to 'ftd' as 'td[n-2].nexted'.
     *
     *  When we just have a single TD, what we have is the tb->td
     * pointer pointing to the only TD which will serve as new
     * placeholder. 
     *
     *  The ISOCHRONOUS tb filling is not done neither designed yet,
     * but it should be similar.
     *
     * Parameters:  
     *
     *  ts_ohci_tb*     Pointer to tb structure to fill in, with
     *                  tb->tr pointing to the ts_usb_tr to be used as
     *                  filling stuff.
     *
     * Returns:
     *
     *  int             < 0 errno code on error, else ok.
     */

#define OHCI_ATD_PAGE (4096)
#define OHCI_ATD_PAYLOAD (2*OHCI_ATD_PAGE)

int ohci_atb_fill (ts_ohci_tb* tb)
{
    e_usb_tr_type type;
    int cnt;
    ts_ohci_atd *ftd;       /* first TD */
    ts_ohci_atd *td;        /* array of TDs */
    ts_ohci_atd *ltd;       /* last TD */
    ts_usb_tr *tr;
    void *data, *top;
    size_t tds;
    
    gdlf_fntrace ("ohci_atb_fill (0x%p)", tb);

    tr = tb->tr;
    type = tr->open->endp->type;

#ifdef GDLF_ASSERTIONS
    if (type == e_usb_tr_type_cntl)
        gdlf_aassert (tr->req != NULL, return -EINVAL);
#endif
    
        /* 
         *  Let's calculate the number of tds we're going to
         * require. We start with the size of the data to transfer [if
         * control xfer, the data phase]. If there's no data to
         * transfer, we have to transfer one NULL packet at least
         * [except on control].
         *
         *  Each TD carries as much as 8k, so we divide by
         * OHCI_ATD_PAYLOAD.
         *
         *  Also, if control transfer, we have to add a td for the setup
         * phase and another for the status phase, but if there's no
         * data, we don't need the NULL data packet.
         */

    if (tr->data == NULL)               /* no data? */
        tds = type == e_usb_tr_type_cntl? 0 : 1;
    else                                /* there's data */
        tds = (tr->size / OHCI_ATD_PAYLOAD)
            + ((tr->size & OHCI_ATD_PAYLOAD) != 0? 1 : 0);

    if (type == e_usb_tr_type_cntl)     /* control setup and ack? */
        tds += 2;

        /* get the ftd */

    ftd = ts_ohci_atd_init();
    if (ftd == NULL)
    {
        usb_errno = -USBE_NO_MEMORY;
        return -ENOMEM;
    }
    
        /* Ok, let's allocate or reallocate the tds array */

    if (tds > 1)    /* do we need more tds? */
    {
        if (tb->td.a == NULL)               /* do we have to allocate? */
        {                                   
            td = ts_ohci_atd_a_init (tds - 1);  /* allocate the tds array */
            if (td == NULL)                 /* Error? */
            {
                usb_errno = -USBE_NO_MEMORY;
                return -ENOMEM;             /* and return */
            }
        }                                   
        else                                /* No need to allocate, reuse */
        {                
            td = tb->td.a;                  /* get original */
            if (tb->tds == tds)             /* same space we need? ... */
                ts_ohci_atd_a_release (td, tds);    /* ... yep, clean */
            else                            /* no, reallocate the array */
            {
                ts_ohci_atd_a_release (td, tds);/* release */
                usb_kfree (td);                 /* free */
                tb->td = NULL;                  /* jic */
                td = ts_ohci_atd_a_init (tds);  /* reinitialize the array */
                if (td == NULL)                 /* error? ... */
                {
                    usb_errno = -USBE_NO_MEMORY;
                    return -ENOMEM;             /* ... return it */
                }
            }
        }
    }
    
        /* So, now we have the first TD and the other ones */

    ftd->delay = OHCI_ATD_DELAY_NOIRQ;      /* irq only on last */
    ftd->datatoggle = tr->toggle;
    ftd->datatogglesrc = 1;         /* we always take toggle from tr */
    ftd->errors = 0;
    ftd->cc = e_ohci_cc_not_accesed;
    if (tds > 1)
        ltd = &td[tds-2];               /* we allocated td-1! */
    else
        ltd = ftd;

        /*
         *  ok, now handle the case of this being a control xfer. If
         * it is, first TD is the setup packet and last TD is the
         * status packet. If not, ftd is the first packet of data
         * transmitted. 
         */

    data = tr->data;
    top = data + tr->size;
    if (type == e_usb_tr_type_cntl)
    {
        ftd->direction =    /* we'll change this later, used to copy */
            ts_usb_request_dir (tr->req->bmreqtype) == e_usb_req_dir_h2d?
            OHCI_ATD_DIR_IN : OHCI_ATD_DIR_OUT;

        memcpy (ltd, ftd, sizeof (ts_ohci_atd));
        ltd->direction =    /* invert sense to ack request */
            ts_usb_request_dir (tr->req->bmreqtype) == e_usb_req_dir_h2d?
            OHCI_ATD_DIR_OUT : OHCI_ATD_DIR_IN;

        ftd->currentbp = cpu_to_le32 (virt_to_bus (tr->req));
        ftd->endbp = cpu_to_le32 (
            virt_to_bus (tr->req + sizeof (*tr->req)));
            /* current and endbp are already zero */        
    }
    else    /* This is not a control transfer */
    {
            /* we don't care td->direction, it is taken from ED */
        ftd->currentbp = cpu_to_le32 (virt_to_bus (data));
        data += OHCI_ATD_PAYLOAD;
        if (data > top)
            data = top;
        ftd->endbp = cpu_to_le32 (virt_to_bus (data));
    }

        /* now do the linkings for the first TD */

    if (tds > 1)
    {
        ftd->nexttd = cpu_to_le32 (virt_to_bus (&td[0])) >> 4;
        ltd->nexttd = cpu_to_le32 (virt_to_bus (ftd)) >> 4;
    }

        /* setup rest of tds */

    cnt = 0;
    while (data < top)
    {
        gdlf_aassert (cnt < tds-1, break);

        memcpy (&td[cnt], ftd, sizeof (ts_ohci_atd));

        td[cnt].currentbp = (__u32)data;
        data += OHCI_ATD_PAYLOAD;
        if (data >= top)
        {
            data = top;
            td[cnt].endbp = (__u32)data;
            break;
        }
        td[cnt].endbp = cpu_to_le32 (virt_to_bus (data));
        td[cnt].nexttd = cpu_to_le32 (virt_to_bus (&td[cnt+1])) >> 4;
        cnt++;
    }

    if (type == e_usb_tr_type_cntl)
        ftd->direction = OHCI_ATD_DIR_SETUP;

        /* setup last TD bits */

    ltd->rounding = 1;
    ltd->delay = 0;         /* interrupt on end */

    tb->ftd.a = ftd;
    tb->td.a = td;
    tb->tds = tds;
    return 0;    
}

int ohci_itb_fill (ts_ohci_tb* tb)
{
    gdlf_fntrace ("ohci_itb_fill (0x%p)", tb);
#warning FIXME: ohci_itb_fill() not implemented
    return -ENOSYS;
}

    /* split isochronous and other tb generation */

static inline
int ohci_tb_fill (ts_ohci_tb* tb)
{
    gdlf_fntrace ("ohci_tb_fill (0x%p)", tb);
    gdlf_aassert (tb != NULL, return -EINVAL);
    gdlf_aassert (tb->tr != NULL, return -EINVAL);
    gdlf_aassert (tb->tr->open != NULL, return -EINVAL);
    gdlf_aassert (tb->tr->open->endp != NULL, return -EINVAL);

    switch (tb->tr->open->endp->type)
    {
      case e_usb_tr_type_cntl:
      case e_usb_tr_type_bulk:
      case e_usb_tr_type_intr:
        return ohci_atb_fill (tb);
        break;

      case e_usb_tr_type_isoc:
        return ohci_itb_fill (tb);
        break;
    
      default:
        gdlf_err ("What the hell am I doing here?\n");
    }
    return -EINVAL;
}


    /* Set down an endpoint's HCI data.
     *
     *  Sets down and destroys the endpoint's HCI data. It will unlink
     * from everywhere and leave everything in a nice state.
     *
     * Parameters:  
     *
     *  ts_usb_device*  Pointer to device structure
     *
     *  ts_usb_endp*    Pointer to endpoint to set down
     *
     * Returns:
     *
     *  int             < 0 errno code on error, 0 if ok. 
     */

int ohci_endp_setdn (ts_usb_device* device, ts_usb_endp* endp)
{
    ts_ohci_endp *ehci, *prev, *head;
    ts_ohci_host *hhci;
    union 
    {
            ts_ohci_atd *a;
            ts_ohci_itd *i;
    } ltd;
    int order;
    
    gdlf_fntrace ("ohci_endp_setdn (0x%p, 0x%p)", device, endp);
    gdlf_aassert (device != NULL, return -EINVAL);
    gdlf_aassert (endp != NULL, return -EINVAL);
    gdlf_aassert (endp->hci != NULL, return -EINVAL);
    gdlf_aassert (device->hub != NULL, return -EINVAL);
    gdlf_aassert (device->hub->host != NULL, return -EINVAL);
    gdlf_aassert (device->hub->host->hci != NULL, return -EINVAL);

        /* Release the place holder */

    ehci = endp->hci;
    if (ehci->ed.nexted != 0)   /* 0 == le32 (0) == be32 (0) */
    {
        ltd.a = bus_to_virt (le32_to_cpu (ehci->ed.nexted << 4));
        if (ltd.a->type == e_ohci_td_type_async)
            ts_ohci_atd_release (ltd.a);
        else
            ts_ohci_itd_release (ltd.i);
        usb_kfree (ltd.a);
    }

        /* Unlink from the corresponding list */

    hhci = OHCI (device->hub->host->hci);
    gdlf_aassert (hhci != NULL, return -EINVAL);
    gdlf_aassert (hhci->iso_head != NULL, return -EINVAL);
    
    switch (endp->type)
    {
            /* list heads for cntl and bulk are registers */

      case e_usb_tr_type_cntl:
        write_lock (&hhci->ctl_lock);
        gdlf_aassert (le_readl (hhci->regs.ctl_head) != 0, 
                      write_unlock (&hhci->ctl_lock); break);
        head = ptr2kvoid (le_readl (hhci->regs.ctl_head));
        if (head == ehci)
            le_writel (ehci->ed.nexted << 4, hhci->regs.ctl_head);
        else
        {
            prev = ed_list_prev (head, ehci);
            gdlf_aassert (prev != NULL, 
                          write_unlock (&hhci->ctl_lock); break);
            prev->ed.nexted = ehci->ed.nexted;
        }
        write_unlock (&hhci->ctl_lock);
        break;
        
      case e_usb_tr_type_bulk:
        write_lock (&hhci->bulk_lock);
        gdlf_aassert (le_readl (hhci->regs.bulk_head) != 0, 
                      write_unlock (&hhci->bulk_lock); break);
        head = ptr2kvoid (le_readl (hhci->regs.bulk_head));
        if (head == ehci)
            le_writel (ehci->ed.nexted << 4, hhci->regs.bulk_head);
        else
        {
            prev = ed_list_prev (head, ehci);
            gdlf_aassert (prev != NULL, 
                          write_unlock (&hhci->bulk_lock); break);
            prev->ed.nexted = ehci->ed.nexted;
        }
        write_unlock (&hhci->bulk_lock);
        break;
        
            /* Isocrhonous have a 'normal' head at ts_ohci_host */
        
      case e_usb_tr_type_isoc:
        write_lock (&hhci->iso_lock);
        gdlf_aassert (hhci->iso_head->nexted != 0, 
                      write_unlock (&hhci->bulk_lock); break);
        head = ptr2kvoid (hhci->iso_head->nexted << 4);
        if (head == ehci)
            hhci->iso_head->nexted = ehci->ed.nexted;
        else
        {
            prev = ed_list_prev (head, ehci);
            gdlf_aassert (prev != NULL, 
                          write_unlock (&hhci->bulk_lock); break);
            prev->ed.nexted = ehci->ed.nexted;
        }
        write_unlock (&hhci->iso_lock);
        break;

            /* Intr have heads which depend on the interval */

      case e_usb_tr_type_intr:
        write_lock (&hhci->intr_lock);
        order = log2 (endp->interval);
        gdlf_aassert (hhci->intr_head[order] != NULL, 
                      write_unlock (&hhci->intr_lock); return -EINVAL);
        head = ptr2kvoid (hhci->intr_head[order]->nexted << 4);
        if (head == ehci)
            hhci->intr_head[order]->nexted = ehci->ed.nexted;
        else
        {
                /*
                 * ed_list_prev() will only stop on nexted ==
                 * 0, thus, if unfound, it will scan all the interval
                 * lists after the head, until the end of the
                 * isochronous list. It should make no difference, as
                 * the ED should be found before passing to another
                 * interval list.
                 */

            prev = ed_list_prev (head, ehci);
            gdlf_aassert (prev != NULL, 
                          write_unlock (&hhci->bulk_lock); break);
            prev->ed.nexted = ehci->ed.nexted;            
        }
        write_unlock (&hhci->intr_lock);
        break;

      default:
        gdlf_apassert (0, 
                       {
                           ts_ohci_endp_release (ehci); 
                           usb_kfree (ehci);
                           return -EINVAL;
                       },
                       "What the hell am I doing here?\n");
    }

        /* release, free and leave */

    ts_ohci_ed_release (endp->hci);
    usb_kfree (endp->hci);
    endp->hci = NULL;
    return 0;
}


    /* Initialize the HCI data for an endpoint
     *
     *  This will allocate and setup the HCI data for the endpoint.
     *
     * Parameters:  
     *
     *  ts_usb_device*  Pointer to device structure.
     *
     *  ts_usb_endp*    Pointer to endpoint of that device.
     *
     * Returns:
     *
     *  int             < 0 errno code on error, 0 if ok. 
     */

int ohci_endp_setup (ts_usb_device* device, ts_usb_endp* endp)
{
    ts_ohci_endp* ehci;
    ts_ohci_host* hhci;
    union 
    {
            ts_ohci_atd *a;
            ts_ohci_itd *i;
    } ltd;
    int order;
    
    gdlf_fntrace ("ohci_endp_setup (0x%p, 0x%p)", device, endp);
    gdlf_aassert (device != NULL, return -EINVAL);
    gdlf_aassert (endp != NULL, return -EINVAL);
    gdlf_aassert (endp->hci == NULL, return -EINVAL);
    gdlf_aassert (device->hub != NULL, return -EINVAL);
    gdlf_aassert (device->hub->host != NULL, return -EINVAL);
    gdlf_aassert (device->hub->host->hci != NULL, return -EINVAL);
    
    ehci = ts_ohci_endp_init();
    if (ehci == NULL)
    {
        gdlf_err ("Cannot allocate memory for HCI data "
                  "of endpoint %d, device 0x%p (%d %d)\n",
                  usb_endp_addr (endp), device, usb_host_addr (device),
                  usb_dev_addr (device));
        return -ENOMEM;
    }
    
        /* Allocate placeholder TD */

    if (endp->type == e_usb_tr_type_isoc)
        ltd.i = ts_ohci_itd_init();
    else
        ltd.a = ts_ohci_atd_init();
    if (ltd.i == NULL)
    {
        gdlf_err ("Cannot allocate placeholder TD for HCI data "
                  "of endpoint %d, device 0x%p (%d %d)\n",
                  usb_endp_addr (endp), device, usb_host_addr (device),
                  usb_dev_addr (device));
        usb_kfree (ehci);
        return -ENOMEM;
    }

        /* Ok, fill in the data */

    ehci->ed.address = device->addr;
    ehci->ed.endpoint = usb_endp_addr (endp);
    if (endp->type == e_usb_tr_type_cntl)
        ehci->ed.direction = OHCI_ED_DIR_FTD1;
    else
        ehci->ed.direction = USB_ENDP_READABLE (endp->addr)? 
            OHCI_ED_DIR_IN /* 1 */ : OHCI_ED_DIR_OUT /* 0 */;
    ehci->ed.speed = device->hub->port[device->port].lowspeed;
        /* ehci->ed.skip already set to 1 on _clean */
    ehci->ed.format = endp->type == e_usb_tr_type_isoc? 1 : 0;
    ehci->ed.maxpktsize = endp->maxpktsize;

        /* Setup placeholder TD */

    ehci->ed.tailp = ehci->ed.headp = 
        cpu_to_le32 (virt_to_bus (ltd.a)) >> 4;

        /* Link to the corresponding list head */

#warning FIXME: dont prepend, APPEND

    hhci = OHCI (device->hub->host->hci);
    gdlf_aassert (hhci != NULL, return -EINVAL);
    gdlf_aassert (hhci->iso_head != NULL, return -EINVAL);

    switch (endp->type)
    {
            /* list heads for cntl and bulk are registers */

      case e_usb_tr_type_cntl:
        write_lock (&hhci->ctl_lock);
        ehci->ed.nexted = le_readl (hhci->regs.ctl_head) >> 4;
        le_writel (kvoid2ptr (&ehci->ed) & ~0xf, hhci->regs.ctl_head);
        write_unlock (&hhci->ctl_lock);
        break;
        
      case e_usb_tr_type_bulk:
        write_lock (&hhci->bulk_lock);
        ehci->ed.nexted = le_readl (hhci->regs.bulk_head) >> 4;
        le_writel (kvoid2ptr (&ehci->ed) & ~0xf, hhci->regs.bulk_head);
        write_unlock (&hhci->bulk_lock);
        break;

            /* Isocrhonous have a normal head at ts_ohci_host */
        
      case e_usb_tr_type_isoc:
        write_lock (&hhci->iso_lock);
        ehci->ed.nexted = hhci->iso_head->nexted;
        hhci->iso_head->nexted = kvoid2ptr (&ehci->ed) >> 4;
        write_unlock (&hhci->iso_lock);
        break;

            /* Intr have heads which depend on the interval */

      case e_usb_tr_type_intr:
        write_lock (&hhci->intr_lock);
        order = log2 (endp->interval);
        gdlf_aassert (hhci->intr_head[order] != NULL, 
                      write_unlock (&hhci->intr_lock); return -EINVAL);
        ehci->ed.nexted = hhci->intr_head[order]->nexted;
        hhci->intr_head[order]->nexted = kvoid2ptr (&ehci->ed) >> 4;
        write_unlock (&hhci->intr_lock);
        break;

      default:
        gdlf_apassert (0, 
                       {
                           ts_ohci_endp_release (ehci); 
                           usb_kfree (ehci);
                           return -EINVAL;
                       },
                       "What the hell am I doing here?\n");
    }

    endp->hci = ehci;
    return 0;
}


    /* Setup a tr using OHCI means
     *
     * Parameters:  
     *
     *  type            descr
     *
     * Global:
     *
     *  type name       descr
     *
     * Returns:
     *
     *  type            descr
     *
     * Global:
     *
     *  type name       descr
     */

int ohci_tr_setup (ts_usb_tr* tr)
{
    ts_ohci_tb *tb;
    int errcode;
    
    gdlf_fntrace ("ohci_tr_setup (0x%p)", tr);

    gdlf_aassert (tr != NULL, return -EINVAL);
    gdlf_aassert (tr->status == e_usb_tr_status_filled, 
                  return -EBADFD);


        /* Allocate tr HCI */

    tb = ts_ohci_tb_init();
    if (tb == NULL)
    {
        gdlf_err ("Unable to allocate memory for an OHCI tr structure\n");
        return -ENOMEM;
    }
    
        /* Setup the tds */

    tb->tr = tr;
    errcode = ohci_tb_fill (tb);
    if (errcode < 0)
    {
        ts_ohci_tb_release (tb);
        usb_kfree (tb);
        return errcode; /* ohci_tb_fill already complained */
    }
 
    tr->hci = tb;
    return 0;
}


    /* FIXME: Detailed desc
     *
     * Compliance:      USB Spec 1.0 xxxx
     *
     * Parameters:  
     *
     *  type            descr
     *
     * Global:
     *
     *  type name       descr
     *
     * Returns:
     *
     *  type            descr
     *
     * Global:
     *
     *  type name       descr
     */

int ohci_tr_setdn (ts_usb_tr* tr)
{
    gdlf_fntrace ("ohci_tr_setdn (0x%p)", tr);

    gdlf_aassert (tr != NULL, return -EINVAL);
    gdlf_aassert (tr->status == e_usb_tr_status_setup, return -EBADFD);

    if (tr->hci)
        ts_ohci_tb_release (tr->hci);

    return 0;
}


    /* FIXME: Detailed desc
     *
     * Compliance:      USB Spec 1.0 xxxx
     *
     * Parameters:  
     *
     *  type            descr
     *
     * Global:
     *
     *  type name       descr
     *
     * Returns:
     *
     *  type            descr
     *
     * Global:
     *
     *  type name       descr
     */

int ohci_tr_run (ts_usb_tr* tr)
{
    int errcode;
    ts_ohci_tb *tb;
    ts_ohci_host *hhci;
    
    gdlf_fntrace ("ohci_tr_run (0x%p)", tr);
    gdlf_aassert (tr->hci != NULL, return -EINVAL);
    gdlf_aassert (tr->open->endp != NULL, return -EINVAL);
    gdlf_aassert (tr->open->host != NULL, return -EINVAL);
    gdlf_aassert (tr->open->host->hci != NULL, return -EINVAL);
    gdlf_aassert (tr->status == e_usb_tr_status_inserted, return -EBADFD);

    tb = (ts_ohci_tb*)tr->hci;
    hhci = (ts_ohci_host*)tr->open->host->hci;
    
    errcode = ohci_tb_hash_set_slot (&hhci->hash, tb);
    if (errcode < 0)
        return errcode;
    else
        tb->index = errcode;

#warning FIXME: set index on TDs

        /* copy "last" td into tail td */
        /* clean last td [FIXME: needed?] */

#warning FIXME: ohci_tr_run() not implemented
    return -ENOSYS;
}


    /* FIXME: Detailed desc
     *
     * Compliance:      USB Spec 1.0 xxxx
     *
     * Parameters:  
     *
     *  type            descr
     *
     * Global:
     *
     *  type name       descr
     *
     * Returns:
     *
     *  type            descr
     *
     * Global:
     *
     *  type name       descr
     */

int ohci_tr_abort (ts_usb_tr* tr)
{
    gdlf_fntrace ("ohci_tr_setdn (0x%p)", tr);
#warning FIXME: ohci_tr_abort() not implemented
    return -ENOSYS;
#if 0
    ohci_tb_hash_release_slot (&hhci->hash, tb->index);
#endif
}


    /* FIXME: Detailed desc
     *
     * Compliance:      USB Spec 1.0 xxxx
     *
     * Parameters:  
     *
     *  type            descr
     *
     * Global:
     *
     *  type name       descr
     *
     * Returns:
     *
     *  type            descr
     *
     * Global:
     *
     *  type name       descr
     */

int ohci_tr_reset (ts_usb_tr* tr)
{
    gdlf_fntrace ("ohci_tr_reset (0x%p)", tr);
#warning FIXME: ohci_tr_reset() not implemented
    return -ENOSYS;
}


/*
 * $Log: scheduler.c,v $
 * Revision 1.2  1998/11/16 00:05:15  inaky
 * Moved the include files to uusbd/include/uusbd, to ease integration
 * into the Linux kernel. Removed old dev stuff, added endp setup/dn
 * stuff, tb filling functions [untested] and tr setup/dn.
 *
 * Revision 1.1  1998/09/14 20:31:13  inaky
 * Added all the code for OHCI, a lot missing to complete yet.
 *
 */
