
/* USB Driver Stack v$Revision: 1.1 $ 
 *
 * (C) 1997 by Iaky Prez Gonzlez <inaky@peloncho.fis.ucm.es>
 *
 * Distributed under the terms of the GNU Public License Version 2
 * See files COPYRIGHT and COPYING at the root distribution directory
 *
 * $Id: tb-hash.c,v 1.1 1998/11/14 00:43:10 inaky Exp $
 *
 * Changelog at the bottom 
 *
 *     OHCI - Endpoint hash list
 *     -------------------------
 */

#include <uusbd/usb.h>

#define GDLF_IDENT          "USB-OHCI"
#define GDLF_LOCAL          /* Define if want local settings */
#define GDLF_DEF_LOGLEVEL   9
#define GDLF_ASSERTIONS     /* want assertions (#undef if not) */
#define GDLF_FNTRACE        /* want function traces */
#define GDLF_LOCATIONS      /* want locations in log messages */
#include <uusbd/gdlf.h>

#include "ohci.h"


    /* Resize the hash endpoint to double it's actual size
     *
     * WARNING: No locks acknowledged or held. This function assumes
     *          the hash struct is locked; also it is also supposed to
     *          be called ONLY from ohci_tb_hash_set_slot().
     *
     * Parameters:  
     *
     *  ts_ohci_tb_hash*  Pointer to the hash structure.
     *
     * Returns:
     *
     *  int             < 0 errno code on error; new size if ok.
     */

int ohci_tb_hash_grow (ts_ohci_tb_hash* hash)
{
    size_t new;
    ts_ohci_tb **new_ptb;
    
    gdlf_fntrace ("ohci_tb_hash_grow (0x%p)", hash);
    gdlf_aassert (hash != NULL, return -EINVAL);

    new = hash->size == 0? 2 : 2 * hash->size;
    new_ptb = usb_realloc (hash->ptb, 
                           hash->size * sizeof (ts_ohci_endp*), 
                           new * sizeof (ts_ohci_endp*),  
                           GFP_KERNEL);

    if (new_ptb != NULL)  /* Uh? ... */
    {
        hash->ptb = new_ptb;
        memset (new_ptb + hash->size, 0,
                new / (hash->size? 2 : 1) * sizeof (ts_ohci_endp*));
        hash->size = new;           /* set the new size */
        return hash->size;          /* go away */
    }
    else
        return -ENOMEM;
}


    /* Adquire an empty slot index.
     *
     *  This will search for an empty slot [hash->ptb[index] ==
     * NULL] and return it's index. If none found, it'll resize.
     *
     *  To avoid locking problems, we take a second parameter which is
     * the pointer we are going to write into the index. This function
     * will write it for us, at least the first time, so no other
     * caller will think the index is empty.
     *
     *  Note that setting the index pointer to NULL means you release
     * it, and it will be re-used or even freed.
     *
     * Parameters:  
     *
     *  ts_ohci_tb_hash*  Pointer to the hash structure.
     *
     *  void*           Pointer to set the index to.
     *
     * Returns:
     *
     *  int             < 0 errno code on error; assigned index if
     *                  ok.
     */

int ohci_tb_hash_set_slot (ts_ohci_tb_hash* hash, void* ptr)
{
    int cnt;
    
    gdlf_fntrace ("ohci_tb_hash_set_slot (0x%p)", hash);
    gdlf_aassert (hash != NULL, return -EINVAL);
    gdlf_aassert (ptr != NULL, return -EINVAL);
    
        /* Search a NULL pointer in the array */

    write_lock (&hash->lock);
    for (cnt = 0; cnt < hash->size; cnt++)
        if (hash->ptb[cnt] == NULL)
        {
            hash->ptb[cnt] = ptr;
            write_unlock (&hash->lock);
            return cnt;
        }

        /* no room left, resize and get first */

    if (hash->size == OHCI_TB_HASH_MAX_TBS)
        return -ENFILE;

    cnt = ohci_tb_hash_grow (hash);
    if (cnt >= 0)
    {
        cnt = hash->size / 2 + 1;
        hash->ptb[cnt] = ptr;
    }
    write_unlock (&hash->lock);
    return cnt;
}


    /* Cleanup the hash
     *
     *  This will cleanup the hash, compacting it as much as
     * possible. Note the compactation is only done for contiguous
     * empty spaces at the top [that is, it will only free if the top
     * is empty, and only the empty space].
     *
     *  It starts on top and searches for how many contiguous NULL
     * pointers are on the array. If there are as many as 
     *
     * Parameters:  
     *
     *  ts_ohci_tb_hash*  Pointer to the hash structure.
     *
     *  unsigned long   Last time it was called [for statistical
     *                  purposes, currently unused] expressed in
     *                  jiffies. If zero, will completely free the
     *                  hash. 
     *
     * Returns:
     *
     *  int             < 0 errno code on error, >= 0 new size if ok.
     */

int ohci_tb_hash_cleanup (ts_ohci_tb_hash* hash, 
                          unsigned long last_time)
{
    ssize_t cnt, new;

    gdlf_fntrace ("ohci_tb_hash_cleanup (0x%p, %lu)", 
                  hash, last_time);
    gdlf_aassert (hash != NULL, return -EINVAL);

    write_lock (&hash->lock);
    for (cnt = hash->size - 1; cnt > 0; cnt--)
        if (hash->ptb[cnt] != NULL)
            break;
    
    if (last_time == 0) /* free everything */
    {
        if (cnt != hash->size)
            gdlf_err ("slots still used at endp hash 0x%p; "
                      "won't free, expect memory leaks\n", hash);
        else
        {
            usb_kfree (hash->ptb);
            hash->ptb = NULL;
            hash->size = 0;
        }
    }
    else
    {
            /*
             *  get new size [next power of two of 'cnt'] and if
             * different, reallocate 
             */

        for (new = 1; new < cnt; new *= 2);
        gdlf_assert (new <= hash->size);    /* jic */
        if (new < hash->size)
        {
            ts_ohci_tb** ptb;
        
            ptb = usb_realloc (hash->ptb, 
                               hash->size * sizeof (ts_ohci_endp*),
                               new * sizeof (ts_ohci_endp*), GFP_KERNEL);
            if (ptb != NULL)
            {
                hash->ptb = ptb;
                hash->size = new;
            }
            else
                gdlf_err ("Cannot shrink endp hash 0x%p from %d to "
                          "%d positions\n",
                          hash, hash->size, new);
        }
    }
    write_unlock (&hash->lock);
    return hash->size;
}


/*
 * $Log: tb-hash.c,v $
 * Revision 1.1  1998/11/14 00:43:10  inaky
 * *** empty log message ***
 *
 */
