

/* USB Driver Stack v$Revision: 1.3 $ 
 *
 * (C) 1997 by Iaky Prez Gonzlez <inaky@peloncho.fis.ucm.es>
 *
 * Distributed under the terms of the GNU Public License Version 2
 * See files COPYRIGHT and COPYING at the root distribution directory
 *
 * $Id: file-operations.c,v 1.3 1998/09/14 20:32:26 inaky Exp $
 *
 * Changelog at the bottom 
 *
 *
 *     File operations
 *     ---------------
 *
 * Here is contained all the code related to VFS interface (file
 * operations, aka fops), that will be used for the kernel interface.
 *
 *
 */

#include <linux/malloc.h>
#include <linux/fs.h>
#include <linux/errno.h>
#include <asm/uaccess.h>

#include "usb.h"
#include "usbd.h"

#define GDLF_IDENT          "USB"
#define GDLF_LOCAL          /* Define if want local settings */
#define GDLF_DEF_LOGLEVEL   9
#define GDLF_ASSERTIONS     /* want assertions (#undef if not) */
#define GDLF_FNTRACE        /* want function traces */
#define GDLF_LOCATIONS      /* want locations in log messages */
#include "gdlf.h"


    /* Open a file descriptor.
     *
     * Major number is USB, minor number is bus address. U select
     * inside the host using IOCTLs.
     *
     * Parameters:  
     *
     *  struct inode*   i-node description structure.
     *
     *  struct file*    file description structure.
     *
     * Returns:
     *
     *  int             0 if ok, errno code < 0 on error.
     *
     * Global:
     *
     *  int usb_errno   On error, error code.  
     */


int usb_file_open (struct inode *inode, struct file *file)
{
    unsigned        minor;                  /* the USB host number */
    ts_usb_addr     addr;                   /* destination address */
    ts_usb_open     *open;                  /* open file structure */
    int             result;
    
    gdlf_fntrace ("usb_file_open (0x%p, 0x%p)", inode, file);
    
    file->private_data = NULL;              /* clean (jic) */
    minor = MINOR (inode->i_rdev);          /* get the minor number */

    open = ts_usb_open_init (NULL);         /* allocate an open file struct */
    if (open == NULL)                       /* error? ... */    
        return -ENOMEM;                     /* ... error */

    ts_usb_addr_clean (&addr);              /* init the addr */
    addr.host = minor;                      /* set the host number */

    result = usb_open (&addr, &open);       /* set a ts_usb_open descriptor for this file */
    if (result < 0)                         /* on error ... */
    {
        ts_usb_open_free (open);                /* free every resource */
        usb_kfree (open);
        return result;                          /* and return it */
    }
    file->private_data = open;              /* set the private data */
    return 0;
}


    /* Close a file descriptor.
     *
     * Parameters:  
     *  struct inode*   i-node description structure.
     *  struct file*    file description structure.
     *
     * Returns:
     *  int             0 no error detected, != errno error code.
     *
     * Global:
     *  unsigned int usb_errno  Error code, only set on error.
     */

int usb_file_close (struct inode *inode, struct file *file)
{
    ts_usb_open *open;
    
    gdlf_fntrace ("usb_file_release (0x%p, 0x%p)", inode, file);   

    open = file->private_data;          /* get the private data */

    gdlf_apassert (open != NULL,
                   { usb_errno = -USBE_DEVICE_NULL; return -EINVAL; },
                   "File structure has a NULL open descriptor\n");
    
    usb_close (open);                 /* release the data ... */
    ts_usb_open_free (open);
    usb_kfree (open);
    return 0;
}


    /* Write data to a USB file descriptor. What has to be done is to
     * set up a buffer in a per FILE descriptor basis. Thus, when user
     * requests a write, we take her data and copy it to our buffer
     * (we should use an small buffer, with a size configurable via
     * /proc and sysctl). Then call usb_write() with out buffer. If
     * still there's data remaining (and no error), then we copy the
     * next data chunk and write it ... and so on, until we are
     * done. Ok, I know it's not a new procedure but hey! I'm
     * documenting a function :).
     *
     * FIXME: What'd happen if we write more data then accepted?
     *
     * FIXME: make sure all the checks we make are enough and not
     *        duplicated anywhere. The same applies to file_read().
     *
     * Parameters:
     *
     *  struct file*    file description structure.
     *
     *  const char*     pointer to the data to write IN USER-LAND
     *
     *  unsigned long   length of the data to write
     *
     * Returns:
     *
     *  long            Bytes written, errno code < 0 on error.
     *
     * Global:
     *
     *  int usb_errno   On error, error code.  
     */

ssize_t usb_file_write (struct file *file, const char *data, 
                        size_t len, loff_t *offset)
{
    ts_usb_open     *open;          /* USB open descriptor */

    gdlf_fntrace ("usb_file_write (0x%p, 0x%p, %u, 0x%p)", 
                  file, data, len, offset);

    if (len == 0 || len > SSIZE_MAX)                    /* Incorrect argument? ... */
        return -EINVAL;                                 /* ... forget it */

    if (!access_ok (VERIFY_READ, data, len))            /* Is it a nice pointer? ... */
        return -EFAULT;                                 /* ... nope! */

    open = (ts_usb_open*) file->private_data;           /* get the USB open descriptor */

    gdlf_apassert (open != NULL,
                   { usb_errno = -USBE_DEVICE_NULL; return -EINVAL; },
                   "File structure has a NULL open descriptor\n");    
    
    {
        ts_usb_trbuf    tr;
        
        ts_usb_trbuf_clean (&tr);
        tr.data = (void*)data;      /* though it's const, won't be written :) */
        tr.user = 1;
        tr.size = len;
        
        return usb_write_nochk (open, &tr); /* write it */
    }
}


    /* Read data from a USB open file descriptor. What has to be done
     * is to set up a buffer in a per FILE descriptor basis. Thus,
     * when user requests a read, we read data until filling up our
     * buffer, and then we copy it to her buffer; we continue reading
     * until we fill her buffer. If there's still more data, I don't
     * know what should be done. Ok, we should make it wait over
     * there, but I dunno if the device will be happy. FIXME.
     *
     * Parameters:  
     *  struct inode*   i-node description structure.
     *  struct file*    file description structure.
     *  const char*     pointer to the data to write IN USER-LAND
     *  unsigned long   length of the data to write
     *
     * Returns:
     *
     *  long            Bytes readen, errno code < 0 on error.
     *
     * Global:
     *
     *  int usb_errno   On error, error code.  
     */

ssize_t usb_file_read (struct file *file, char *data, size_t len, loff_t * offset)
{
    ts_usb_open     *open;          /* USB open descriptor */
    
    gdlf_fntrace ("usb_file_read (0x%p, 0x%p, %u, 0x%p)", 
                  file, data, len, offset);

    if (len == 0 || len > SSIZE_MAX)            /* Incorrect argument? ... */
        return -EINVAL;                         /* ... forget it */

    if (!access_ok (VERIFY_WRITE, data, len))   /* Is it a nice pointer? ... */
        return -EFAULT;                         /* ... nope! */

    open = (ts_usb_open*) file->private_data;       /* get the file USB data */

    gdlf_apassert (open != NULL,
                   { usb_errno = -USBE_DEVICE_NULL; return -EINVAL; },
                   "File structure has a NULL open descriptor\n");

    {
        ts_usb_trbuf    tr;
        
        ts_usb_trbuf_clean (&tr);
        tr.data = data;
        tr.user = 1;
        tr.size = len;
        
        return usb_read_nochk (open, &tr); /* read it */
    }
}


    /* Control the USB file descriptor from the VFS interface
     *
     * Parameters:  
     *
     *  struct inode*   Inode structure
     *
     *  struct file*    File structure
     *
     *  unsigned int    Command code
     *
     *  unsigned long   Arguments to the command
     *
     * Returns:
     *
     *  int             0 if ok, errno code < 0 on error.
     *
     * Global:
     *
     *  int usb_errno   On error, error code.  
     */


int usb_file_ioctl (struct inode *inode, struct file *file, 
                    unsigned int cmd, unsigned long arg)
{
    ts_usb_open *open;
    
    gdlf_fntrace ("usb_ioctl (0x%p, 0x%p, %u, %lu)", 
                  inode, file, cmd, arg);

    open = file->private_data;

    gdlf_apassert (open != NULL,
                   { usb_errno = -USBE_DEVICE_NULL; return -EINVAL; },
                   "File structure has a NULL open descriptor\n");
    
        /* Let's see ... what do we have to do? */

    switch (cmd)
    {

            /* Commands for destination control */

      case USB_IOCTL_SET_DST:       /* Set full destination */
      {
          ts_usb_addr addr;
          if (copy_from_user (&addr, (ts_usb_addr*)arg, sizeof (ts_usb_addr)))
              return -EFAULT;
          return usb_dst_set (open, &addr);
          break;
      }      

      case USB_IOCTL_SET_DST_HOST:  /* Set destination host */
        return usb_dst_host_set (open, arg);
        break;
        
      case USB_IOCTL_SET_DST_DEV:   /* Set destination device */
        return usb_dst_dev_set (open, arg);
        break;

      case USB_IOCTL_SET_DST_ENDP:  /* Set destination endpoint */
        return usb_dst_endp_set (open, arg);
        break;                      /* check read/write with endp's direction */

#warning FIXME: ioctl interface unfinished
#if 0
            /* Commands for device control */

      case USB_IOCTL_SET_DEV_ASET:  /* Set alternate setting */
        break;
        
      case USB_IOCTL_SET_DEV_CFG:   /* Set configuration */
        break;
        
      case USB_IOCTL_SET_DEV_ADDR:  /* Set device address */
        break;
#endif

      default: 
        gdlf_err ("Unknown IOCTL command (%u,%lu)\n", cmd, arg);
        break;
    }
    
    return -EINVAL;
}


struct file_operations fops = 
{
    NULL,               /* llseek */
    usb_file_read,      /* read */
    usb_file_write,     /* write */
    NULL,               /* readdir */
    NULL,               /* select */
    usb_file_ioctl,     /* ioctl */
    NULL,               /* mmap FIXME: would be cool for isochronous transfers */
    usb_file_open,      /* open */
    NULL,               /* flush */
    usb_file_close,     /* release */
    NULL,               /* fsync */
    NULL,               /* fasync */
    NULL,               /* check_media_change */
    NULL                /* revalidate */
};

/* Log:
 *
 * $Log: file-operations.c,v $
 * Revision 1.3  1998/09/14 20:32:26  inaky
 * Added the flush method.
 *
 * Revision 1.2  1998/01/08 12:19:19  inaky
 * Added copyright strings to all important files
 *
 * Revision 1.1  1998/01/02 00:17:46  inaky
 * Moving here all the stuff to make rawd IANF
 *
 * Revision 1.23  1997/12/30 20:15:41  inaky
 * Moved everything to the new lists scheme. Untested yet.
 *
 * Revision 1.22  1997/12/29 00:25:34  inaky
 * Fixed unimplemented functions to return -ENOSYS and moving the khubdd
 * to use signals as commands.
 * Also added function traces on most files that had them disabled (donkey!)
 *
 * Revision 1.21  1997/12/27 18:12:23  inaky
 * *** empty log message ***
 *
 * Revision 1.20  1997/12/27 17:25:35  inaky
 * Finished moving to GDLF. Deprecated usb_printk and __LOCATION__
 *
 * Revision 1.19  1997/12/27 16:44:23  inaky
 * Moving to GDLF and fixinf docs. Removing more cruft.
 *
 * Revision 1.18  1997/12/27 02:46:05  inaky
 * Moved to GDLF for debugging messages and logs. Fixed some docs.
 *
 * Revision 1.17  1997/12/27 02:07:00  inaky
 * Moving everything to use the GDLF for logging and debug checkings.
 * Removed the global array of USB hosts and moved it to a linked
 * list. Added an address field to each USB host controller.
 * Removed lots of cruft that were accumulated and useless. Still some
 * missing, but may be useful someday.
 *
 * Revision 1.16  1997/12/27 02:05:58  inaky
 * Moving everything to use the GDLF for logging and debug checkings.
 * Removed the global array of USB hosts and moved it to a linked
 * list. Added an address field to each USB host controller.
 * Removed lots of cruft that were accumulated and useless. Still some
 * missing, but may be useful someday.
 *
 * Revision 1.15  1997/12/08 02:40:57  inaky
 * Fixed some stuff of the #includes, modified the makefile targets to be
 * more flexible, tried to finish the device configuration section.
 *
 * Revision 1.14  1997/11/29 20:21:47  inaky
 * Documented and almost fixes the read/write/abort API, documented a
 * damn lot and implemented some stuff
 *
 * Revision 1.13  1997/11/29 00:46:29  inaky
 * Documented almost all the hub basic API, added spin locks to
 * usb_host_register() and some more stuff. Going for more.
 *
 * Revision 1.12  1997/10/02 22:29:36  inaky
 * Finally managed to understand the workings of the MODVERSIONS (ok, not
 * exactly the working, but how to make it work).
 *
 * Now I can go to sleep happy :) -- I know, I'm really sick --
 *
 * Revision 1.11  1997/09/08 10:54:47  inaky
 * Modified the generic functions to gain some performance, also some other checks
 *
 * Revision 1.10  1997/07/03 20:00:02  inaky
 * Moved all the hw dependencies to the HCD module, modified the
 * controller detections stuff to use just PCI means (okay, other buses
 * to come, but when they come). General fixes, but no really new deep
 * enhancements.
 *
 * Revision 1.9  1997/06/22 19:54:30  inaky
 * Seems that the moving of all the hw dependants stuff to the HCI
 * modules is done. Got to test it yet, but at least it compiles cleanly.
 *
 * Revision 1.8  1997/04/14 16:56:01  inaky
 * *** empty log message ***
 *
 * Revision 1.7  1997/04/12 16:47:49  inaky
 * Added the code for buffering reads/writes to/from userland
 *
 * Revision 1.6  1997/04/10 21:30:48  inaky
 * Fixed the file code, so it opens/closes ok; also reads/writes ok ... ok, just theory
 *
 * Revision 1.5  1997/04/06 22:56:44  inaky
 * Added the functions to set destinations of an USB open descriptor
 * (ts_usb_open); however, I haven't tested it yet ... bugs flowing!
 *
 * Revision 1.4  1997/04/06 21:14:31  inaky
 * Still with the definition of the access interface. More or less it is
 * taking an stable form.
 *
 * Revision 1.3  1997/03/31 20:24:46  inaky
 * Pfff ... a lot. Mostly added descriptors stuff (HUB class still is
 * missing a lot) and I think I've got a definitive idea of how to
 * implement the storing of configurations and interfaces.
 *
 * Revision 1.2  1997/03/09 23:10:31  inaky
 * *** empty log message ***
 *
 * Revision 1.1  1997/02/22 22:25:44  inaky
 * *** empty log message ***
 *
 *
 */

