
/* USB Driver Stack v$Revision: 1.22 $ 
 *
 * (C) 1997 by Iaky Prez Gonzlez <inaky@peloncho.fis.ucm.es>
 *
 * Distributed under the terms of the GNU Public License Version 2
 * See files COPYRIGHT and COPYING at the root distribution directory
 *
 * $Id: slist.c,v 1.22 1998/11/16 18:29:31 inaky Exp $
 *
 * Changelog at the bottom 
 *
 *
 *     USB Simply linked Lists
 *     -----------------------
 *
 * These are the implementations of singly-linked lists for the USB.
 *
 */

#include <uusbd/usb-misc.h>
#include <uusbd/usb-slist.h>

#define GDLF_IDENT          "USB-SLIST"
#define GDLF_LOCAL          /* Define if want local settings */
#define GDLF_DEF_LOGLEVEL   9
#define GDLF_ASSERTIONS     /* want assertions (#undef if not) */
#undef GDLF_FNTRACE        /* want function traces */
#define GDLF_LOCATIONS      /* want locations in log messages */
#include <uusbd/gdlf.h>


    /* Allocate and initialize a ts_snode
     *
     * Parameters:  
     *
     *  tf_dtor         Pointer to the destructor fucntion. A tf_dtor
     *                  is an int(*)(void*). May be NULL, but if NULL,
     *                  you can't then let the ts_snode be destroyed
     *                  by it's means, as you'll get a GPF. 
     *
     * Returns:
     *
     *  ts_snode*       New snode, with the ts_dtor initialized with
     *                  the parameter given. NULL on allocation error.
     */

ts_snode* ts_snode_init (tf_dtor dtor)
{
    ts_snode* snode;
    
    gdlf_fntrace ("ts_snode_init (0x%p)", dtor);

    snode = ts_slist_malloc (sizeof (ts_snode));
    if (snode == NULL)
        return NULL;
    
    ts_snode_clean (snode);
    ts_dtor_fill (&snode->dtor, dtor);
    return snode;
}


    /* Destroy a ts_node, calling it's destructor.
     *
     *  Will only destroy the snode's contents, not the space taken by
     * the structure itself!
     *
     * Parameters:  
     *
     *  ts_snode*       Pointer to snode to destroy.
     *
     * Returns:
     *
     *  int             0 if ok, errno code < 0 on error. In fact,
     *                  this is the destructor's return code.
     */

void ts_snode_release (ts_snode* snode)
{
    gdlf_aassert (snode != NULL, return);
    gdlf_passert (atomic_read (&snode->usecnt) == 0,
                  "Aieee! Freeing an snode with use count %d!\n",
                  atomic_read (&snode->usecnt));
    gdlf_passert (snode->next == NULL,
                  "Aieee! Freeing an snode with `next' = 0x%p!\n",
                  snode->next);    
    return;
}


    /* Print a snode data
     *
     * Parameters:  
     *
     *  ts_snode*       Pointer to snode to print.
     */

void ts_snode_printk (const ts_snode* snode)
{
    gdlf_fntrace ("ts_snode_printk (0x%p)", snode);

    gdlf_debug ("ts_snode 0x%p\n", snode);
    gdlf_debug ("  dtor   0x%p\n", snode->dtor.fn);
    gdlf_debug ("  usecnt %d\n", atomic_read (&snode->usecnt));
    gdlf_debug ("  next:  0x%p\n", snode->next);
    return;
}

    /* Allocate and initialize an slist structure.
     *
     * Returns:
     *
     *  ts_slist*       Pointer to the new allocated and initialized
     *                  slist, ready for use. NULL on allocation
     *                  error. 
     */

ts_slist* ts_slist_init (void)
{
    ts_slist* slist;    

    gdlf_fntrace ("ts_slist_init()");

    slist = ts_slist_malloc (sizeof (ts_slist));
    if (slist == NULL)
    {
        gdlf_err ("Couldn't allocate a ts_slist structure\n");
        return NULL;
    }
    
    ts_slist_clean (slist);
    return slist;
}


    /* Free an slist
     *
     *  Destroys an slist by walking the list down and unlinking all
     * the elements out of it (calling the destructors on each node if
     * the use count reaches zero).
     *
     *  When destroying nodes, if any returns an error (!= 0), it will
     * be saved to be returned, but the destruction of the list will
     * continue until the end. In fact, if multiple errors are
     * returned, only the last one will be kept and returned. However,
     * the list ALWAYS will be destroyed.
     *
     * Parameters:  
     *
     *  ts_slist*       Pointer to the slist to destroy.
     *
     * Returns:
     *
     *  int             0 if ok, errno code < 0 on error.
     */

void ts_slist_release (ts_slist* slist)
{
    int count;
    ts_snode *itr, *next;
    
    gdlf_fntrace ("ts_slist_release (0x%p)", slist);    
    gdlf_aassert (slist != NULL, return);
        
    ts_slist_write_lock (&slist->lock);
    itr = slist->head;      /* get the first */
    slist->head = NULL;     /* there's nothing on the list */
    while (itr)             /* while there're elements */
    {
        next = itr->next;       /* get next one */
        itr->next = NULL;       /* unlink it */
        ts_slist_write_lock (&itr->lock);
        count = atomic_read (&itr->usecnt);
        ts_snode_unref (itr, NULL); /* remove one reference */
        if (count != 1)         /* else, it's been freed */
            ts_slist_write_unlock (&itr->lock);
        itr = next;             /* let's check next */
    }
    ts_slist_write_unlock (&slist->lock);
    ts_slist_clean (slist); /* clean */
    return;
}


void ts_slist_printk (const ts_slist* slist)
{
    gdlf_fntrace ("ts_slist_printk (0x%p)", slist);    
    gdlf_aassert (slist != NULL, return);
        
    gdlf_debug ("ts_slist 0x%p\n", slist);
    gdlf_debug ("  usecnt:    %d\n", atomic_read (&slist->usecnt));
    gdlf_debug ("  elements:  %u\n", slist->elements);
    gdlf_debug ("  head:      0x%p\n", slist->head);
    return;
}


    /* Remove a node from a list.
     *
     * Parameters:  
     *
     *  ts_slist*       Pointer to the list to remove a node from.
     *
     *  ts_snode*       Pointer to the snode to remove from the list. 
     *
     * Returns:
     *
     *  int             0 if ok, errno code < 0 on error.
     */

int ts_slist_del (ts_slist* slist, ts_snode* snode)
{
    int errcode;
    ts_snode *prev;                 /* previous ptr */

    gdlf_fntrace ("ts_slist_del (0x%p, 0x%p)", slist, snode);
    gdlf_aassert (slist != NULL, return -EINVAL);
    gdlf_aassert (snode != NULL, return -EINVAL);

        /* 
         * Don't check prev != NULL as ts_slist_del_fast() will do it
         * for us.
         */

    ts_slist_write_lock (&slist->lock);
    prev = ts_slist_prev (slist, snode);
    if (prev == NULL)
        errcode = -ENOENT;
    else
        errcode = ts_slist_del_fast (slist, snode, prev);
    ts_slist_write_unlock (&slist->lock);
    return errcode;
}



    /* Fast Removal of one list element.
     *
     *  This function performs the same as ts_slist_del(), but works
     * faster, in the case we know which is the previous element to
     * the one we want to remove.
     *
     *  WARNING!!! This function *DOES*NOT* lock the list. It must be
     *             already locked. However, it does lock `snode' and
     *             `prev'.
     *
     * Parameters:  
     *
     *  ts_slist*       Pointer to the list to remove a node from.
     *
     *  ts_snode*       Pointer to the snode to remove from the list. 
     *
     *  ts_snode*       Pointer to the previous snode to the one to
     *                  remove. If equal to the one to remove, it
     *                  means it is list's head. 
     * Returns:
     *
     *  int             0 if ok, errno code < 0 on error.
     */

int ts_slist_del_fast (ts_slist* slist, ts_snode* snode, ts_snode* prev)
{
    int count;
    
    gdlf_fntrace ("ts_slist_del_fast (0x%p, 0x%p, 0x%p)", 
                  slist, snode, prev);
    gdlf_aassert (slist != NULL, return -EINVAL);
    gdlf_aassert (snode != NULL, return -EINVAL);
    gdlf_aassert (prev != NULL, return -EINVAL);

    ts_slist_write_lock (&snode->lock);
    ts_slist_write_lock (&prev->lock);
    if (prev == snode)              /* if it's list's head ... */
        slist->head = snode->next;  /* ... set head to next */
    else                            /* if not head ... */
    {
        gdlf_apassert (prev->next == snode,
                       {
                           ts_slist_write_unlock (&prev->lock);
                           ts_slist_write_unlock (&snode->lock);
                           return -EINVAL;
                       },
                       "Element 0x%p is not the previous element "
                       "to 0x%p on ts_slist 0x%p\n",
                       prev, snode, slist);
        prev->next = snode->next;   /* ... set prev's next to snode's next */
    }
    slist->elements--;
    snode->next = NULL;             /* clean */
    count = atomic_read (&snode->usecnt);
    ts_snode_unref (snode, NULL);   /* unreference it */
    ts_slist_write_unlock (&prev->lock);
    if (count != 1)                 /* if not freed ... */
        ts_slist_write_unlock (&snode->lock);
    return 0;
}


    /* Prepend a node into the list
     *
     * Parameters:  
     *
     *  ts_slist*       Pointer to the list in which we want to
     *                  prepend a node.
     *
     *  ts_snode*       Pointer to the node we want to insert.
     *
     * Returns:
     *
     *  int             0 if ok, errno code < 0 on error.
     */

int ts_slist_prep (ts_slist* slist, ts_snode* snode)
{
    gdlf_fntrace ("ts_list_prep (0x%p, 0x%p)", 
                  slist, snode);
    gdlf_aassert (slist != NULL, return -EINVAL);        
    gdlf_aassert (snode != NULL, return -EINVAL);
    gdlf_aassert (snode->next == NULL, return -EBUSY);
        
    ts_slist_write_lock (&slist->lock);
    ts_slist_write_lock (&snode->lock);
    snode->next = slist->head;          /* insert it ... */
    slist->head = ts_snode_ref (snode); /* it's the head of the list now */
    slist->elements++;                  /* one element more in the list */
    ts_slist_write_unlock (&snode->lock);
    ts_slist_write_unlock (&slist->lock);
    return 0;
}


    /* Get previous element in a list
     *
     * Parameters:  
     *
     *  ts_slist*       List where to find the previous element.
     *
     *  ts_snode*       Node of that list for whom we want to search
     *                  the previous element. If NULL, you'll get the
     *                  last element of the list.
     *
     * Returns:
     *
     *  ts_snode*       The previous node if not NULL. NULL if the
     *                  node is not on this list. Itself if it is the
     *                  first element.
     *
     * Global:
     *
     *  int usb_errno   On error, error code.  
     */

ts_snode* ts_slist_prev (ts_slist* slist, ts_snode* snode)
{
    ts_snode *itr;
    
    gdlf_fntrace ("ts_slist_prev (0x%p, 0x%p)", slist, snode);
    
    read_lock (&slist->lock);
    itr = slist->head;
    if (itr == snode)   /* If this one is the first ... */
    {
        read_unlock (&slist->lock);
        return itr;     /* ... then we return it */
    }
    while (itr)         /* while there are elements to search */
    {
        if (itr->next == snode) /* if next is the searched ... */
            break;              /* ... we're done */
        itr = itr->next;        /* it isn't, get next */
    }
    read_unlock (&slist->lock);

    return itr;
}


    /* Find last element in the list.
     *
     * Parameters:  
     *
     *  ts_slist*       Pointer to the list.
     *
     * Returns:
     *
     *  ts_snode*       Pointer to the last element or NULL if the
     *                  list is empty.
     */

ts_snode* ts_slist_last (ts_slist* slist)
{
    ts_snode *itr;
    
    gdlf_fntrace ("ts_slist_last (0x%p)", slist);
    
    read_lock (&slist->lock);
    itr = slist->head;
    if (itr == NULL)        /* empty list? */
    {
        read_unlock (&slist->lock);
        return NULL;
    }
    while (itr->next)       /* while there are elements to search */
        itr = itr->next;    /* it isn't, get next */
    read_unlock (&slist->lock);
    return itr;
}


    /* Append an element to the list
     *
     * Parameters:  
     *
     *  ts_slist*       Pointer to the list to append to.
     *
     *  ts_snode*       Node to append to the list.
     *
     * Returns:
     *
     *  int             0 if ok, errno code < 0 on error.
     */

int ts_slist_append (ts_slist* slist, ts_snode* snode)
{
    ts_snode *last;
    
    gdlf_fntrace ("ts_slist_append (0x%p, 0x%p)", slist, snode);
    gdlf_aassert (slist != NULL, return -EINVAL);
    gdlf_aassert (snode != NULL, return -EINVAL);

    ts_slist_write_lock (&slist->lock);
    ts_slist_write_lock (&snode->lock);

    last = ts_slist_last (slist);   /* get list's tail */
    if (last == NULL)               /* The list is empty ... */
        slist->head = snode;        /* ... so node is list's head now */
    else                            /* List isn't empty ... */
        last->next = snode;         /* ... so we append it */
    ts_snode_ref (snode);           /* one more using the node */
    snode->next = NULL;             /* no node is next */
    slist->elements++;              /* one element more in the list */

    ts_slist_write_unlock (&snode->lock);
    ts_slist_write_unlock (&slist->lock);   
    return 0;
}


    /* Move an element in the list
     *
     *  Move element 'snode' after 'prev' in the list.
     *
     * Parameters:  
     *
     *  ts_slist*       Pointer to the list.
     *
     *  ts_snode*       Node to move around the list.
     *
     *  ts_snode*       Node to move it before. If NULL, we want to
     *                  move it to the beginning of the list.
     *
     * Returns:
     *
     *  int             0 if ok, errno code < 0 on error.
     */

int ts_slist_move (ts_slist* slist, ts_snode* snode, ts_snode* prev)
{
    ts_snode* snode_prev;
    
    gdlf_fntrace ("ts_slist_move (0x%p, 0x%p, 0x%p)", 
                  slist, snode, prev);
    gdlf_aassert (slist != NULL, return -EINVAL);
    gdlf_aassert (snode != NULL, return -EINVAL);

        /* moving to the same place? */

    if (snode == prev)
        return -EMLINK; /* would cause circular link */

        /* both in list? */

    snode_prev = ts_slist_prev (slist, snode);
    gdlf_aassert (snode_prev != NULL, return -EINVAL);
    gdlf_aassert (prev == NULL || ts_slist_prev (slist, prev) != NULL, return -EINVAL);

        /* no change in position? */

    if ((snode_prev == prev)    /* in the middle of the list */
        || (prev == NULL        /* or at the beginning */
            && slist->head == snode))
        return 0;

        /* Ok, we can proceed, let's go */

    ts_slist_write_lock (&slist->lock);
    ts_slist_write_lock (&snode->lock);
        
        /* Unlink snode */

    if (snode_prev == snode)    /* First one? */
        slist->head = snode->next;
    else
        snode_prev->next = snode->next;

        /* Relink it */
    
    if (prev == NULL)       /* At beginning of list */
    {
        snode->next = slist->head;
        slist->head = snode;
    }
    else                    /* in the middle/end of the list */
    {
        snode->next = prev->next;
        prev->next = snode;
    }    
    
    ts_slist_write_unlock (&snode->lock);
    ts_slist_write_unlock (&slist->lock);   
    return 0;
}


/* Log:
 *
 * $Log: slist.c,v $
 * Revision 1.22  1998/11/16 18:29:31  inaky
 * Fixed the move function, it was blatantly broken.
 *
 * Revision 1.21  1998/11/15 23:58:29  inaky
 * Fixed the ts_slist_move function, as it wasn't working very
 * well. Untested deeply.
 *
 * Revision 1.20  1998/11/14 17:15:04  inaky
 * Moved the include files to uusbd/include/uusbd, to ease integration
 * into the Linux kernel.
 *
 * Revision 1.19  1998/08/30 13:44:12  inaky
 * Severed assertions. Made ts_*_release() return void.
 *
 * Revision 1.18  1998/08/27 11:08:47  inaky
 * Moved ts_*_free() to ts_*_release(), made them return void and severed
 * assertions.
 *
 * Revision 1.17  1998/08/25 04:00:04  inaky
 * Fixed ts_snode_printk() to take const pointer as argument.
 *
 * Revision 1.16  1998/08/24 02:17:17  inaky
 * Added functions to print node and list information.
 *
 * Revision 1.15  1998/08/20 23:40:53  inaky
 * Synched both codes (Nslist was getting quite obsolete).
 *
 * Revision 1.14  1998/08/20 17:02:25  inaky
 * Removed spurious debug messages.
 *
 * Added a check to ts_slist_move() not to move the element to the same
 * place it is [faster than actually modifying the code to take it into
 * account].
 *
 * Revision 1.13  1998/08/19 23:46:15  inaky
 * Simplified some on function entry assertions, made clearer the
 * assignment to seek the previous element on ts_slist_del() and added
 * (disabled) debug messages to ts_slist_prep(). Also fixed a drunk lock.
 *
 * Revision 1.12  1998/07/23 14:20:54  inaky
 * Added a function to move an element from one position to another
 * inside a list.
 *
 * Revision 1.11  1998/06/20 21:54:50  inaky
 * Massive changes to split the old monolithic hubd driver into a hub
 * manager (hub wrappers and daemon) and a dynamic driver, which can be
 * replaced and its different parts reused.
 *
 * Revision 1.10  1998/04/13 21:51:50  inaky
 * Added a function to find last element in a list, another for
 * appending.
 *
 * Revision 1.9  1998/02/12 14:31:05  inaky
 * Fixed the delete functions to return error if item to delete isn't on
 * the list instead of failing miserably.
 *
 * Revision 1.8  1998/02/10 17:20:13  inaky
 * Fixed ts_slist_free(), to make it NULLify the head pointer, so that
 * when unreferencing pointers which drop to 0 it's use count, it doesn't
 * try to remove them again from the list if the destructor does so.
 *
 * Also fixed function traces and an assertion which had gone crazy.
 *
 * Revision 1.7  1998/02/10 15:07:05  inaky
 * All referencing and unreferencing done now via the ts_snode_[un]ref()
 * functions. Fixed ts_snode_free(), as a NULL dtor would cause a NULL
 * ptr dereference. Added the ts_slist_del_fast() function to remove
 * elements knowing it's predecesors. Modified ts_slist_del() to use
 * it. Modified ts_slist_free() to use ts_snode_unref(). Added write
 * locks for snodes in variuos places. Fixed check in ts_slist_prev() to
 * allow it return the previous element to the list's head element
 * (itself).
 *
 * Revision 1.6  1998/02/09 13:20:41  inaky
 * Moved to atomic use counts
 *
 * Revision 1.5  1998/01/08 12:19:08  inaky
 * Added copyright strings to all important files
 *
 * Revision 1.4  1998/01/02 00:12:15  inaky
 * Finished the new driver scheme. Documented at usbd-drv.c
 *
 * Revision 1.3  1997/12/31 15:39:01  inaky
 * Added a function to search previous elements
 *
 * Revision 1.2  1997/12/30 20:15:42  inaky
 * Moved everything to the new lists scheme. Untested yet.
 *
 * Revision 1.1  1997/12/30 00:03:45  inaky
 * Adding a generic system to handle singly linked lists and the HCI
 * registration control system
 *
 *
 */
