
/* USB Driver Stack v$Revision: 1.30 $ 
 *
 * (C) 1997 by Iaky Prez Gonzlez <inaky@peloncho.fis.ucm.es>
 *
 * Distributed under the terms of the GNU Public License Version 2
 * See files COPYRIGHT and COPYING at the root distribution directory
 *
 * $Id: driver.c,v 1.30 1998/11/14 17:45:40 inaky Exp $
 *
 * Changelog at the bottom 
 *
 *
 *     USB Universal Host Controller Interface
 *     ---------------------------------------
 *
 *   This is the code to support the host controller in a way as the
 * one proposed by Intel (and some others). All the docs for this can
 * be obtained from the Intel's website at http://www.intel.com. It is
 * a BIG .pdf file. Also available at the project's home page,
 * http://peloncho.fis.ucm.es/~inaky. 
 *
 *   The mode of operation for this module is at the
 * documentation. Please refer to it
 *
 *  This file is the one who takes care of basic registration,
 * initialization and USB bus enumeration of the UHCI module and host
 * controller. 
 */

#include <linux/config.h>   /* Configuration */
#include <linux/module.h>   /* MOD_INC_USE_COUNT */
#include <linux/kernel.h>   /* printk */
#include <linux/init.h>     /* __init */
#include <linux/malloc.h>   /* GFP_KERNEL */
#include <linux/mm.h>       /* GFP_KERNEL */
#include <linux/pci.h>      /* PCI definitions */
#include <linux/ioport.h>   /* check_region(), release_region() */
#include <asm/errno.h>      /* -E* codes */
#include "piix3.h"

#include <uusbd/usbd.h>
#include <uusbd/usb-misc.h>
#ifdef CONFIG_PCI
#include <uusbd/usb-pci.h>
#endif
#include <uusbd/usb-hcd.h>

#define GDLF_IDENT          "USB-UHCI"
#define GDLF_LOCAL          /* Define if want local settings */
#define GDLF_DEF_LOGLEVEL   9
#define GDLF_ASSERTIONS     /* want assertions (#undef if not) */
#undef GDLF_FNTRACE        /* want function traces */
#define GDLF_LOCATIONS      /* want locations in log messages */
#include <uusbd/gdlf.h>

#include "uhci.h"           /* this has to be include after GDLF */


ts_usb_hci* uhci = NULL;    /* pointer to the HCI structure */
volatile int uhci_errno = 0;/* uhci errno */
atomic_t usecnt =           /* use count */
    ATOMIC_INIT (0);
const char *uhci_ctl_str = UHCI_CTL_STR; /* for request_(irq|region)() */

struct timer_list timer;    /* timer for clean ups */

    /*
     * UHCI timer function
     *
     *  This does cleanup tasks and polls the root hub
     */

void uhci_timer (unsigned long last_time)
{
    static unsigned long last_cleanup = 0;
    ts_uhci_host *itr;

    gdlf_fntrace ("uhci_timer (%lu)", last_time);

        /* poll the hub */
    
    itr = (ts_uhci_host*)ts_slist_head (&uhci->hosts);
    while (itr)
    {
        if (itr->host != NULL)
            usb_hub_status_poll (&itr->host->root);
        else
            gdlf_err ("UHCI controller 0x%p has no USB host!\n", itr);
        itr = (ts_uhci_host*)ts_slist_next (&itr->snode);
    }
    timer.data = jiffies;
    mod_timer (&timer, jiffies + UHCI_HUB_T*HZ);
    if (jiffies - last_cleanup < UHCI_CLEANUP_T*HZ)
        return;
    
        /* We walk the list w/o locking, this may be dangerous */

    last_cleanup = jiffies;
    itr = (ts_uhci_host*)ts_slist_head (&uhci->hosts);
    while (itr)
    {
        iso_hash_cleanup (&itr->iso, timer.data);
        intr_hash_cleanup (itr, timer.data);
        itr = (ts_uhci_host*)ts_slist_next (&itr->snode);
    }
    return;
}

    /*
     * Host timer function
     *
     *  This does some debug stuff when compiled in.
     */

#ifdef UHCI_DEBUG_TIMER
void uhci_hci_timer (unsigned long addr)
{
#ifdef UHCI_SBYS
    __u16 w;
#endif

    gdlf_fntrace ("uhci_hci_timer (0x%lu)", addr);
#ifdef UHCI_DEBUG_TIMER_FL
    uhci_fl_search (UHCI (addr), 
                    inw (UHCI (addr)->iobase + USBFRNUM) & USBFRNUM_M,
                    inw (UHCI (addr)->iobase + USBSTS),
                    UHCI_FL_ALL);
#endif
    mod_timer (& UHCI (addr)->timer, jiffies + UHCI_DEBUG_TIMER*HZ);
#if 0
    uhci_dumpreg (UHCI (addr));
#endif
#if 0
    ts_uhci_fl_printk (UHCI (addr));
#endif
#ifdef UHCI_SBYS
    w = inw (UHCI (addr)->iobase + USBCMD);
    outw (w | USBCMD_RS | USBCMD_SWDBG, UHCI (addr)->iobase + USBCMD);
#endif
    return;
}
#endif /* UHCI_DEBUG_TIMER */


    /* Destroy and free resources of an UHCI host
     *
     *  This must leave the UHCI host struct as unreferenced as
     * possible to the UHCI sub-sub-system (aka: removing it from all
     * lists). This way, after running this, the only reference should
     * be the one kept from the ts_usb_host structure.
     *
     *  NOTE: this assumes everything on the upper layer [the USBD] is
     *        cancelled.  
     *
     * Parameters:  
     *
     *  ts_uhci_host*   Pointer to the UHCI host structure to
     *                  destroy. 
     *
     * Returns:
     *
     *  int             0 if ok, errno code < 0 on error.
     *
     * Global:
     *
     *  int usb_errno   On error, error code.  
     *  int uhci_errno  On error, error code.  
     */

int uhci_host_destroy (ts_uhci_host* hci)
{
    int errcode;
    
    gdlf_fntrace ("uhci_host_destroy (0x%p)", hci);

        /* Ok, let's stop the controller and reset it */

    ts_snode_ref (&hci->snode);
    if (hci->iobase)
        uhci_hci_shutdown (hci);

    errcode = usb_hci_unregister_host (uhci, &hci->snode);

#ifdef UHCI_DEBUG_TIMER
    if (hci->timer.function)
        del_timer (&hci->timer);
#endif
    
    intr_hash_cleanup (hci, 0);
    iso_hash_cleanup (&hci->iso, 0);

    ts_snode_unref_nd (&hci->snode);
    ts_uhci_host_release (hci);

        /*
         *  This must be cleaned in order for usb_host_destroy()
         * not to repeat the unregistration in certain steps.
         */

    hci->fl = NULL;
    hci->iobase = 0;
    hci->ioextent = 0;
    
    return errcode;
}


    /* Shutdown an USB controller and it's bus. 
     *
     * Frees data structures and shuts the controller down. All the
     * administrativia stuff relative to the USB must be already done
     * (aka: all devices unregistered). 
     *
     * Note that it won't free the space taken by the ts_uhci_host
     * structure, just it's contents.
     *
     * On error, it will keep going on. The worst thing it may happen
     * is the host controller not being properly closed, but the last
     * reset will shut it up.
     *
     * Parameters:  
     *
     *  ts_uhci_host*   Pointer to the host controller structure of
     *                  the host.
     *
     * Returns:
     *
     *  int             0 if ok, errno code < 0 on error.
     *
     * Global:
     *
     *  int usb_errno   On error, error code.  
     */

int uhci_hci_shutdown (ts_uhci_host* hci)
{
    int retval1, retval2 = 0;
    
    gdlf_fntrace ("uhci_hci_shutdown (0x%p)", hci);

        /* 
         * On error, we'll ignore it, as we can do nothing and the
         * functions will complain themselves 
         */

    retval1 = uhci_runstop (hci, 0);
    if (retval1)
        retval2 = retval1;

    retval1 = uhci_bus_reset (hci);
    if (retval1)
        retval2 = retval1;

    return retval2;
}


#ifdef CONFIG_PCI
void uhci_pci_quirks (struct pci_dev* pci)
{
    gdlf_fntrace ("uhci_pci_quirks (0x%p)", pci);
    gdlf_aassert (pci != NULL, return);

        /* 
         * Ok, now, if this is an Intel's PIIX3 or PIIX4, we got to do some
         * tweaks to have the HC really delivering interrupts and
         * being a master.
         */

    switch (pci->vendor)
    {
      case PCI_VENDOR_ID_INTEL:
        switch (pci->device)
        { 
          case PCI_DEVICE_ID_INTEL_82371AB_0:
          case PCI_DEVICE_ID_INTEL_82371AB:
          case PCI_DEVICE_ID_INTEL_82371AB_2:
          case PCI_DEVICE_ID_INTEL_82371AB_3:
          case PCI_DEVICE_ID_INTEL_82371SB_2:
                /* Enable interrupts */
            usb_pci_bits16 (pci, PCILEGSUP, PCILEGSUP_USBPIRQDEN,
                            PCILEGSUP_USBPIRQDEN);
                /* Enable bus mastering */
            usb_pci_bits16 (pci, PCICMD, PCICMD_BME, PCICMD_BME);
                /* #warning FIXME: check sillicon revision? (PCI reg 0x08) */
            if (usb_pci_bits32 (pci, PCILEGSUP, 0, 0) & 0xf)
            {
                gdlf_warning ("PIIX: Legacy support was enabled, "
                              "disabling\n");
                    /* 0xaf0f writes 0 on R/WC bits */
                usb_pci_bits32 (pci, PCILEGSUP, 0, 0xaf0f);
            }
            break;

          default:
            gdlf_err ("Unkown PCI USB device ID 0x%d from vendor 0x%d: "
                      "won't fine tune\n",
                      pci->device, pci->vendor);
        }
        break;

      default:
        gdlf_err ("Unkown vendor 0x%d for PCI USB device ID 0x%d: "
                  "won't fine tune\n",
                  pci->vendor, pci->device);
    }
    return;
}
#endif

    /* Initialize the structures for a PCI UHCI controller, filling the 
     * hci and hcitype parameters. Note that it will release nothing
     * previously adquired (aka: it assumes the hci member of the
     * ts_usb_host points to NULL).
     *
     * Parameters:  
     *
     *  ts_usb_host*    Pointer to the host structure which will hold
     *                  it
     *
     *  struct pci_dev* Pointer to PCI device structure.
     *
     * Returns:
     *
     *  int             0 if ok, errno code < 0 on error.
     *
     * Global:
     *
     *  int usb_errno   On error, error code.  
     */


#ifdef CONFIG_PCI
int uhci_pci_init (ts_usb_host *host, struct pci_dev* dev)
{
    ts_uhci_host*   hci;
    int             iobase;     /* Base of io region */
    int             ioextent;   /* Length of io region */
    int             irq;        /* IRQ # */
    int             errcode;

    gdlf_fntrace ("uhci_pci_init (0x%p, 0x%p)", host, dev);
    gdlf_aassert (dev != NULL, return -EINVAL);
    gdlf_aassert (host != NULL, return -EINVAL);

    hci = UHCI (host->hci);

    gdlf_aassert (hci != NULL, return -EINVAL);

        /* get iobase */

    iobase = dev->base_address[4] & PCI_BASE_ADDRESS_IO_MASK;
    ioextent = UHCI_IOEXTENT;           /* get the io extension */
    if (iobase == 0)                    /* Hmmm ... uninitialized? */
    {                                   /* If this ever happens ... :) */
        gdlf_err ("Host controller "
                  "at PCI bus %d dev %d fn %d "
                  "has asigned IO address 0x0000!\n",
                  dev->bus->number, PCI_SLOT (dev->devfn),
                  PCI_FUNC (dev->devfn));
        uhci_errno = usb_errno = -USBE_NO_IOREGION;
        return -ENXIO;
    }
    
        /* IO zone is ok? check it ... */

    if (check_region (iobase, ioextent) != 0)
    {
        gdlf_err ("IO region 0x%x-0x%x for host controller "
                  "at PCI bus %d, function %d "
                  "is already in use!\n", 
                  iobase, iobase+ioextent, 
                  dev->bus->number, dev->devfn);
        uhci_errno = usb_errno = -USBE_IOREGION_USED;
        return -EBUSY;
    }

        /* Get an IRQ */

    irq = dev->irq;
    if (irq == 0)   /* None assigned? */
    {               /* NOTE: Assuming zero means "unassigned" */ 
        gdlf_err ("Host controller "
                  "at PCI bus %d, function %d "
                  "has assigned IRQ 0!\n", 
                  dev->bus->number, dev->devfn);
        uhci_errno = usb_errno = -USBE_NO_FREE_IRQ;
        return -ENXIO;
    }

        /* 
         * All the checks are OK; we can register the IO region (on
         * error, ts_usb_host_release() will free the IO region)
         */  

    request_region (iobase, ioextent, uhci_ctl_str);

    hci->pci = dev;                 /* store the PCI info */
    hci->iobase = iobase;           /* Set ... */
    hci->ioextent = ioextent;       /* ... the ... */
    hci->irq = irq;                 /* ... data (this one's not confirmed!) */
    hci->host = host;
    
    errcode = 
        request_irq (hci->irq, 
                     (void(*)(int,void*,struct pt_regs*))uhci_irqhandler,
                     SA_SHIRQ, uhci_ctl_str, host);
    if (errcode < 0)
    {
        gdlf_err ("Cannot obtain IRQ %i requested by host controller "
                  "at PCI bus %d, function %d; errno code %d\n",
                  hci->irq, dev->bus->number, dev->devfn, errcode);
        hci->irq = 0;               /* clean (irq is not registered) */
        uhci_errno = usb_errno = -USBE_NO_FREE_IRQ;
        return errcode;
    }

        /* Let's go for the HW quirks */

    uhci_pci_quirks (dev);
    return 0;
}
#endif /* CONFIG_PCI */


    /* Initialize the frame list data
     *
     *  This just will link the Interrupt QH to the LS Control QH,
     * this one to the Control QH, which links to the Bulk QH. The
     * Bulk QH will be linked back to the Control QH, for bandwidth
     * reclamation.
     *
     *  All the entries in the frame list will point, by defect to the
     * Interrupt QH.
     *
     * Parameters:  
     *
     *  ts_uhci_host*   Pointer to UHCI host structure.
     *
     * Returns:
     *
     *  int 0 if ok, errno code < 0 on error.  
     */

int uhci_fl_init (ts_uhci_host *host)
{
    __u32 word;
    ts_uhci_flp *itr, *top;  /* FL QH iterator and top */

    gdlf_fntrace ("uhci_fl_init (0x%p)", host);    
    gdlf_aassert (host != NULL, return -EINVAL);
    gdlf_aassert (host->fl != NULL, return -EINVAL);

    host->ls_ctrl = ts_uhci_tb_list_init();
    if (host->ls_ctrl == NULL)
        return -ENOMEM;

    host->ctrl = ts_uhci_tb_list_init();
    if (host->ctrl == NULL)
        return -ENOMEM;
    
    host->bulk = ts_uhci_tb_list_init();
    if (host->bulk == NULL)
        return -ENOMEM;

        /*
         *  Set up the device list heads (the elements are already
         * terminated, as done by ts_uhci_qh_clean(). We link the Low
         * Speed Control Queue to the Control Queue. The Control Queue
         * to the Bulk Queue, and the Bulk Queue back to the Control
         * Queue, for bandwidth reclamation.
         */

    host->ls_ctrl->first.elt.word = cpu_to_le32 (0x3);
    host->ls_ctrl->first.head.word = 
        cpu_to_le32 ((virt_to_bus (&host->ctrl->first) & ~0xf) 
                     | 0x2 /* q == 1 */ 
                     | 0x0 /* t == 0 */);
    
    host->ctrl->first.elt.word = cpu_to_le32 (0x3);
    host->ctrl->first.head.word = 
        cpu_to_le32 ((virt_to_bus (&host->bulk->first) & ~0xf) 
                     | 0x2 /* q == 1 */ 
                     | 0x0 /* t == 0 */);

    host->bulk->first.elt.word = cpu_to_le32 (0x3);
#ifdef UHCI_DO_BWR
    host->bulk->first.head.word = 
        cpu_to_le32 ((virt_to_bus (&host->ctrl->first) & ~0xf) 
                     | 0x2 /* q == 1 */ 
                     | 0x0 /* t == 0 */);
#else
    host->bulk->first.head.word = cpu_to_le32 (0x3);
#endif
        /*
         *  Now setup all the frame list FLPs to point to the LS Control
         * QH, without terminating them. Remember they're raw, we
         * haven't initialized them when allocated.
         *
         *  We're gonna make them point to a QH, so Q = 1, T = 0 (we
         * don't want'em to terminate), and thus:
         *
         *  & 0xfffffff0 - Cleans lower four bits (T = 0)
         *  | 0x00000002 - Makes Q = 1.
         */

    word = (cpu_to_le32 (virt_to_bus (&host->ls_ctrl->first)) 
            & 0xfffffff0) | 0x2;
    for (itr = host->fl, top = itr + UHCI_FL_SIZE; itr < top; itr++)
        itr->word = word;
    
    return 0;
}

    /* Initialize a given UHCI compliant USB host controller.
     *
     *  If the UHCI is detected and properly initialized, it will
     * register the host controller within the UHCI host controllers
     * list.
     *
     * Parameters:  
     *
     *  ts_usb_host*    The USB Host structure; it expects that the
     *                  HCI pointer of the structure points to an HCI
     *                  struct, filled up with the PCI, IO and IRQ
     *                  data, as it is used for initialization. The
     *                  other field must be clean (as left by the
     *                  ts_uhci_host_clean() macro).
     *
     * Returns:     
     *
     *  int             0 if ok (UHCI compliant controller detected
     *                  and initialized); < 0 errno code on error.
     *
     * Global:
     *
     *  int usb_errno   On error, error code.  
     */

int uhci_host_init (ts_usb_host* host)
{
    int errcode;
    ts_uhci_host *hci;
    unsigned iobase;
    size_t bytes;

    gdlf_fntrace ("uhci_host_init (0x%p)", host);

    host->hcitype = e_usb_hci_uhci;
    hci = UHCI (host->hci);
    gdlf_aassert (hci != NULL, return -EINVAL);

        /* 
         * No need to lock here, not yet even in use.
         *
         * Hmmm ... for guessing if a controller is UHCI compatible,
         * I'll just try to reset it in the UHCI way and then check if
         * all the default values are the ones they oughta be. This
         * should be ok, as the UHCI Spec determines exactly which are
         * the default values.
         *
         * DAMNIT!! Something's wrong. The registers aren't the good
         *          ones, as I'm getting odd values when reading. Is
         *          it the correct IO port? 
         *
         * COOL: Today I learnt that readX/writeX are for memory
         *       mapped devices :) not for IO ports ... hee hee
         */

    uhci_bus_reset (hci);
    if (uhci_host_reset (hci) < 0)          /* Try to reset ... */
        return -ENODEV;                     /* ... exit if unable. No UHCI. */

    iobase = hci->iobase;    
    if (inw (iobase + USBCMD) != 0          /* Check defaults */
        || inw (iobase + USBSTS) != USBSTS_HCH
        || inw (iobase + USBINTR) != 0
        || inw (iobase + USBFRNUM) != 0
        || inb (iobase + USBSOF) != USB_UHCI_SOF_DEFAULT)
    {
        gdlf_err ("Host controller 0x%p "
                  "defaults after reset violation!\n", hci);
        uhci_errno = usb_errno = uhci_errno = -USBE_UHCI_BAD_DEFAULTS;
        return -EIO;
    }

        /* 
         *  Ok, it's fine, pretty UHCI host in the house. 
         * Let's start setting it up 
         */

    uhci_bus_reset (hci);                   /* reset the bus and hc */
    uhci_host_reset (hci);                  /* reset the bus and hc */
    uhci_configure (hci, 1);                /* we are configuring it */
    uhci_runstop (hci, 0);                  /* stop the controller (jic) */
    hci->sof_timing = uhci_get_sof (hci);   /* get the SOF (save it!) */
    uhci_intr (hci, 0,                      /* disable interrupts */
               USBINTR_SPIEN | USBINTR_IOCEN 
               | USBINTR_REN | USBINTR_TOCRCEN);

        /* 
         * Get a place to put the frame list in and init it.
         *
         * NOTE: Hmmm ... quite platform dependant; let's see when
         *       USB comes to other archs 
         */ 

    hci->fl = (ts_uhci_flp*)
        usb_get_free_pages (GFP_KERNEL, 0, 0);
    if (hci->fl == NULL)                    /* no memory? */
    {
        uhci_errno = usb_errno = -USBE_NO_MEMORY;
        gdlf_err ("Couldn't get a free page for "
                  "the frame list of host controller 0x%p!\n", 
                  hci);
        return -ENOMEM;
    }

    errcode = uhci_fl_init (hci);           /* fill in the frame list */
    if (errcode < 0)
    {
        uhci_errno = usb_errno = -USBE_NO_MEMORY;
        gdlf_err ("Couldn't initialize frame list for "
                  "UHCI host 0x%p: errno %d\n", 
                  hci, errcode);
        return -ENOMEM;
    }
    
    uhci_configure (hci, 1);                /* configuring */
    uhci_set_frnum (hci, 0);                /* clean everything */
    uhci_set_fl (hci, hci->fl);             /* set the frame list */

        /* set the methods */

    host->reset = uhci_hci_reset;
    host->setsof = uhci_hci_setsof;
    host->reference = uhci_reference;

#if defined (CONFIG_USB_OHCI) || defined (CONFIG_USB_OHCI_MODULE)
    host->endp_setdn = NULL;
    host->endp_setup = NULL;
#endif

    host->tr_setdn = uhci_tr_setdn;
    host->tr_setup = uhci_tr_setup;

    host->tr_insert = uhci_tr_insert;
    host->tr_remove = uhci_tr_remove;

    host->tr_run = uhci_tr_run;
    host->tr_reset = NULL;      /* unneeded */
    host->tr_abort = uhci_tr_abort;

    uhci_configure (hci, 0);                /* done configuring */

        /* Now set up the root hub */

    host->root.host = host;                 /* our host */
    usb_host_ref (host);                    /* useless? perhaps */
        /* no device (it's already NULL) */
    host->root.cfg = host->root.iface = host->root.aset = 0;
    host->root.operations = &uhci_hub_operations;
        /* We don't call uhci_reference() to keep ref count at zero */

    errcode = uhci_hub_descriptor_fill (&host->root); /* fill it's data */
    if (errcode < 0)                        /* trouble? */
    {
        gdlf_err ("Could not initialize root hub 0x%p "
                  "(%d -1)'s data: errno/uhci %d/%d\n",
                  &host->root, host->addr,
                  errcode, uhci_errno);
        host->root.statch_tr.size = 0;
        return errcode;
    }

        /*
         *  The status change transaction is quite special in UHCI. We
         * aren't going to transfer anything trough the wire while
         * messing with the UHCI root hub, so we don't need to hold a
         * ts_usb_open descriptor. We only need to know that the hub
         * manager requires the transaction status to be DONE to
         * process it; so we leave it in that state; there's nothing
         * more we should care about. The hub destructor will notice
         * the device-less hubs and allow them to have this special
         * characteristic.
         */

    bytes = (1 + host->root.ports)/8 + ((1 + host->root.ports) % 8? 1 : 0);
    host->root.statch_tr.data = usb_kmalloc (bytes, GFP_KERNEL);
    if (host->root.statch_tr.data == NULL)
    {
        gdlf_err ("No memory to set up a %u bytes data "
                  "buffer for root hub 0x%p (%d -1)'s status "
                  "change endpoint\n",
                  bytes, 
                  &host->root, host->root.host->addr);
        return -ENOMEM;
    }
    host->root.statch_tr.status = e_usb_tr_status_ok;
#if 0
    host->root.statch_tr.user = 0;
#endif
    host->root.statch_tr.size = bytes;
    host->root.statch_tr.open = &host->root.statch;

        /* 
         * Now set all bits to one (all ports require attention) but
         * bit zero (hub doesn't need it)
         */

    memset (host->root.statch_tr.data, 0xff, bytes);
    *(__u8*)host->root.statch_tr.data &= ~0x1;

        /* Register into the HCI */
          
    errcode = usb_hci_register_host (uhci, &hci->snode);
    if (errcode < 0)
    {
        gdlf_err ("Could not register UHCI host controller "
                  "within the UHCI list of controllers: "
                  "errno/uhci: %d/%d\n",
                  errcode, uhci_errno);
        return errcode;
    }

        /* 
         *  Put it to run! 
         *
         *  We don't enable the short packet interrupt, because we are
         * going to get that data when accounting the finished TD.
         */

#ifdef UHCI_DEBUG_TIMER
    init_timer (&hci->timer);
    hci->timer.expires = jiffies + UHCI_DEBUG_TIMER*HZ;
    hci->timer.data = (unsigned long)hci;
    hci->timer.function = uhci_hci_timer;
    add_timer (&hci->timer);
#endif
    
        /*
         * If on step-by-step mode, we have to change it with the
         * controller stopped. Last thing we do is enable interrupts.
         */
    
    outw (0xff, hci->iobase + USBSTS);      /* clear status */
    errcode = uhci_runstop (hci, 1);
    if (errcode < 0)
        gdlf_err ("Cannot put UHCI host 0x%p (%d) to run: "
                  "errno/uhci %d/%d\n", 
                  host, host->addr, errcode, uhci_errno);
    uhci_intr (hci, 1, USBINTR_IOCEN | USBINTR_REN | USBINTR_TOCRCEN);
    return errcode < 0? errcode : 0;
}


    /* Enumerate the PCI buses
     *
     * 1. Walk the buses for UHCI controllers. 
     *
     * 2. On each found host, allocate a host and hci structs with
     *    the hw data filled in and call uhci_host_init
     *
     * 3. if succesful, usb_host_register() each host.
     *
     * Returns:
     *
     *  int             0 if ok, errno code < 0 on error.
     *
     * Global:
     *
     *  int usb_errno   On error, error code.  
     *  int uhci_errno  On error, error code.  
     */

#ifdef CONFIG_PCI
int uhci_pci_enumerate (void)
{
    ts_uhci_host *itr;
    ts_usb_host *host;
    int errcode;
    int bioscode;
    int pcihosts;
    struct pci_dev *device;

    gdlf_fntrace ("uhci_pci_enumerate()");

    bioscode = 0;
    pcihosts = 0;

    if (!pci_present())
    {
        gdlf_err ("No PCI BIOS present, no USB UHCI PCI hosts\n");
        return -ENODEV;
    }

    device = NULL;
    while ((device = pci_find_class (USB_PCI_CLASS, device)))
    {
        pcihosts++;
        gdlf_debug ("Host controller found at "
                    "PCI bus %d, device %d, function %d\n",
                    device->bus->number, 
                    PCI_SLOT (device->devfn),
                    PCI_FUNC (device->devfn));

            /* Check it is not already allocated */

        itr = (ts_uhci_host*)ts_slist_head (&uhci->hosts);
        while (itr)
        {
            if (itr->pci == device)     /* same? */
                break;
            itr = (ts_uhci_host*)ts_slist_next (&itr->snode);
        }
        if (itr)            /* used? ... */
            continue;       /* ... oops */

            /* Allocate USB and UHCI host structures */

        host = ts_usb_host_init();
        if (host == NULL)
        {
            uhci_errno = usb_errno = -USBE_NO_MEMORY;
            gdlf_crit ("Cannot get memory for a "
                       "UHCI host structure!\n");
            return -ENOMEM;
        }
        ts_dtor_fill (&host->snode.dtor, (tf_dtor)usb_host_destroy);

        host->hci = (ts_snode*)ts_uhci_host_init();
        if (host->hci == NULL)
        {
            usb_kfree (host);
            uhci_errno = usb_errno = -USBE_NO_MEMORY;
            gdlf_crit ("Cannot get memory for a "
                       "host structure!\n");
            return -ENOMEM;
        }
        ts_dtor_fill (&host->hci->dtor, (tf_dtor)uhci_host_destroy);
                
            /* 
             * Assign hw resources to the PCI host (irqs and
             * ioregions)
             */
            
        errcode = uhci_pci_init (host, device);
        if (errcode < 0)                    /* PCI init failed? */
        {
            gdlf_debug ("Failed doing PCI initialization of UHCI "
                        "host on bus %d, function %d: %d\n",
                        device->bus->number, device->devfn, errcode);
            usb_host_destroy (host);
            usb_kfree (host);
            continue;
        }

            /* 
             * Assign USB resources, fill the structures and register
             * within the HCI
             */
        
        errcode = uhci_host_init (host);
        if (errcode < 0)                    /* init failed? */
        {
            gdlf_debug ("Failed doing initialization of UHCI "
                        "host 0x%p: %d\n", host, errcode);
            usb_host_destroy (host);
            usb_kfree (host);
            continue;
        }
        
            /* init ok, register USB host */
        
        errcode = usb_host_register (host);
        if (errcode < 0)                    /* Register failed? */
        {
            gdlf_debug ("Couldn't register UHCI host  "
                        "0x%p: errno %d\n", host, errcode);
            usb_host_destroy (host);
            usb_kfree (host);
            continue;
        }
        
            /* Register ok */
            
        gdlf_info ("Host controller 0x%p (%d) at "
                   "PCI bus %d, device %d, function %d\n",
                   host, host->addr,
                   device->bus->number, 
                   PCI_SLOT (device->devfn),
                   PCI_FUNC (device->devfn));
        ts_snode_ref (host->hci);           /* for `host->hci' */
        usb_hub_status_poll (&host->root);  /* force polling */
        host = NULL;                        /* clean for next loop */
    }               /* PCI loop for hosts */
    return 0;
}
#endif


    /* Initialize all the UHCI hosts found on the system, scanning all
     * the known bus types (currently, just PCI).
     *
     * Returns:
     *
     *  int             0 if ok, errno code < 0 on error.
     *
     * Global:
     *
     *  int usb_errno   On error, error code.  
     *  int uhci_errno  On error, error code.  
     */

int __init uhci_init (void)
{   
    int             errcode;
    
    gdlf_fntrace ("uhci_init()");

    gdlf_info ("USB Universal Host Controller Interface Driver\n");
#ifdef MODULE
    gdlf_info ("(C) 1996-98 Inaky Perez Gonzalez "
               "<inaky@peloncho.fis.ucm.es>\n");
#endif

        /* Allocate and register an HCI */

    uhci = ts_usb_hci_init();           /* create one */
    if (uhci == NULL)                   /* oops? */
    {
        gdlf_debug ("HCI entry allocation failed\n");
        uhci_errno = usb_errno = -USBE_NO_MEMORY;
        return -ENOMEM;
    }
    ts_snode_ref (&uhci->snode);        /* we keep a reference to it */
    uhci->name = UHCI_NAME;             /* set name ... */
    uhci->type = e_usb_hci_uhci;        /* ... type */
    
    errcode = usb_hci_register (uhci);  /* register it */
    if (errcode)                        /* oops? */
    {
        gdlf_err ("Couldn't register UHCI HCI: "
                  "errno/usb %d/%d\n",
                  errcode, usb_errno);
        ts_usb_hci_release (uhci);          /* free resources */
        usb_kfree (uhci);                   /* free */
        uhci = NULL;
        return errcode;                     /* sorry :) */
    }
    
        /* start the cleanup timer */
    
    init_timer (&timer);
    timer.expires = jiffies + UHCI_HUB_T*HZ;
    timer.data = (unsigned long)jiffies;    /* must not be zero! */
    timer.function = uhci_timer;
    add_timer (&timer);

        /* Enumerate the known buses searching for UHCI */

#ifdef CONFIG_PCI
    errcode = uhci_pci_enumerate();
#else
    /* 
     * We still want to permit being able of initializing the USB
     * subsystem on a PCI-less system, thus the #warning instead
     * of an #error. I know it sounds stupid, but it doesn't hurt.
     */
#warning Actually there are no ways but PCI to get information 
#warning about UHCI USB hosts! You should enable PCI.
#endif /* CONFIG_PCI */

    return errcode; 
}


    /* Force the release of all UHCI hosts
     *
     *  This function will walk the list of USB hosts searching for
     * UHCI ones and force it's unregistration, thus making the UUSBD
     * free of UHCI hosts. After this, it is guaranteed that the
     * USB-UHCI subsystem is completely free and unused, no matter if
     * any error code is returned.
     *
     *  However, we use a helper function :)
     *
     * Returns:
     *
     *  int             0 if ok, errno code < 0 on error.
     *
     * Global:
     *
     *  int usb_errno   On error, error code.  
     */

int uhci_release (void)
{
    int errcode = 0;
    
    gdlf_fntrace ("uchi_release()");

    if (uhci)
        errcode = usb_hci_release (uhci);
    return errcode;
}


    /*  Shutdown the USB-UHCI subsystem
     *
     *  This will clean up everything and shutdown the UCHI
     * subsystem. In fact, it will call uhci_release() to ensure that
     * all the UHCI hosts are freed and then exit.
     *
     * Returns:
     *
     *  int             0 if ok, errno code < 0 on error.
     *
     * Global:
     *
     *  int             0 if ok, errno code < 0 on error.
     */

int uhci_shutdown (void)
{
    int errcode = 0;
    
    gdlf_fntrace ("uhci_shutdown()");

    if (uhci)
    {   
        ts_snode_unref (&uhci->snode, NULL);
        errcode = usb_hci_destroy (uhci);
        usb_kfree (uhci);
    }
    del_timer (&timer);
    return errcode;
}


    /* Add/remove a reference to the driver
     *
     *  This function is used to tell the uhci driver if we are or not
     * using any of it's data structures or functions. When we take a
     * pointer to any of the driver's function in the operations
     * structure, we must reference (use) it, so we call this with a
     * positive argument. When we are done using them, we call it with
     * a negative argument.
     *
     * Parameters:  
     *
     *  ts_snode*       Pointer to the element with which we are using
     *                  the driver.  
     *
     *  __u8            Positive for ONE reference, negative for ONE
     *                  unreference. If zero, return use count.
     *
     * Returns:
     *
     *  int             0 if ok, errno code < 0 on error.
     */
    
int uhci_reference (ts_snode* snode, int no)
{
    gdlf_fntrace ("uhci_reference (0x%p, %d)", snode, no);
    
    if (no > 0)             /* reference */
    {
        atomic_inc (&usecnt);           /* a new use */
        MOD_INC_USE_COUNT;              /* module's too */
    }
    else if (no < 0)        /* unreference */
    {
        atomic_dec (&usecnt);           /* a new use */
        MOD_DEC_USE_COUNT;              /* module's too */
    }
    return atomic_read (&usecnt);
}


    /* Modular stuff */

#ifdef MODULE

int init_module (void)
{
    return uhci_init();
}

void cleanup_module (void)
{
    uhci_shutdown();
    return;
}

#endif



/* Log:
 *
 * $Log: driver.c,v $
 * Revision 1.30  1998/11/14 17:45:40  inaky
 * Moved the include files to uusbd/include/uusbd, to ease integration
 * into the Linux kernel.
 *
 * Revision 1.29  1998/11/09 21:43:14  inaky
 * Moved all host ref/unref to usb_host_[un]ref(). Removed the tr_buffer
 * from the ts_uhci_host structure and made it be `permanent' in
 * statch_tr.data. Changed message levels for first indication of host
 * controller detection. Made it poll the root hub ASA registered.
 *
 * Revision 1.28  1998/10/31 02:20:36  inaky
 * Removed the device setup/dn methods [unused] and added the endpoint
 * ones.
 *
 * Revision 1.27  1998/10/14 21:52:30  inaky
 * Extracted the PCI quirks fixing code to a single function.
 *
 * Revision 1.26  1998/09/24 16:12:11  inaky
 * Added cleanup of iso and intr hash tables before shutting down.
 *
 * Revision 1.25  1998/09/14 20:33:49  inaky
 * Fixed messages, fixed destructor to reference the object, fixed PCI
 * error messages; checked the PCI host is not already used before
 * initializing it [allow for re-enumeration].
 *
 * Revision 1.24  1998/09/02 21:19:11  inaky
 * Fixed headers and moved root hub polling to UHCI HCD. Severed some
 * assertions.
 *
 * Revision 1.23  1998/08/30 13:59:07  inaky
 * Finished new list's, linear topology.
 *
 * Revision 1.22  1998/08/29 00:09:33  inaky
 * Added comments, fntraces where missing (uhci_hci_timer), option to
 * remove bandwidth reclamation, and soem fixes.
 *
 * Revision 1.21  1998/08/27 22:20:26  inaky
 * Removed traces. Fixed some comments and made the shutdown function
 * remove the timer, or it would crash :)
 *
 * Revision 1.20  1998/08/27 14:20:32  inaky
 * Added the uhci global timer to do cleanups, moved the ts_*_free() to
 * ts_*_release(), severed some assertions, Added the bandwidth
 * reclamation code.
 *
 * Revision 1.19  1998/08/25 16:32:10  inaky
 * Removed remaining FIXMEs.
 *
 * Revision 1.18  1998/08/25 16:29:07  inaky
 * Added a check to disable legacy stuff on PIIX3.
 *
 * Revision 1.17  1998/08/25 04:10:03  inaky
 * Added some more verbosity to UHCI host detection.
 *
 * Revision 1.16  1998/08/24 15:46:51  inaky
 * Fixed little cosmetic stuff, removed old files.
 *
 * Revision 1.15  1998/08/24 02:29:42  inaky
 * Did the little debugging needed to make the functions work as they're
 * expected to.
 *
 * Revision 1.14  1998/08/19 23:54:54  inaky
 * Removed all the debug code and let the timer only to do step-by-step
 * execution.
 *
 * Added debug code to trap the memory corruption problem. Changed the
 * ts_snode_unref() call for usb_host_destroy()+usb_kfree() when any
 * step of the host initialization fails.
 *
 * Revision 1.13  1998/08/03 13:37:57  inaky
 * Added a debug per-host timer.
 *
 * Fixed the host destructor to call uhci_hci_shutdown, so it STOPS the
 * controller before exiting.
 *
 * Revision 1.12  1998/07/30 17:56:33  inaky
 * Added PIIX3 specific tweaks.
 *
 * Moved the host TD lists to be allocated, not static within the
 * structure, to have them aligned to 16 bytes.
 *
 * Added a check to the call of the frame list initialization function
 * and removed the UHCI reset method, it is unneeded.
 *
 * Temorarily removed host operation until I find that f**** drunk
 * pointer which is hyperspacing out.
 *
 * Revision 1.11  1998/07/28 23:13:09  inaky
 * Fixed some typos, moved the frame list size to be fixed; moved all io
 * operations not requiring read'n'write to use single in[bwl] or
 * out[bwl].
 *
 * Also enabled interrupts [it wouldn't work, otherwise :)] ...
 *
 * Revision 1.10  1998/07/23 14:41:43  inaky
 * Moved usb-uhci.h to uhci/uhci.h.
 *
 * Made all errnos volatile.
 *
 * Removed the IRQ handler from the hosts structure, it is useless. Also
 * made it have a pointer to the host structure, not to the HCI host.
 *
 * Added a function to initialize the frame list with the pointers to
 * interrupt, ls control, control and bulk lists.
 *
 * Revision 1.9  1998/07/16 15:59:39  inaky
 * Fixed the mess with transactions on device-less hubs.
 *
 * Revision 1.8  1998/07/16 14:16:27  inaky
 * Moved to the new enumeration elements naming scheme.
 *
 * Moved from the old read/write interface to the new, request-based
 * interface.
 *
 * Fixed the shutdown function to correctly destroy the HCI and _free_
 * it's used space.
 *
 * Revision 1.7  1998/07/14 22:03:26  inaky
 * Removed a MOD_DEC_USE_COUNT which I don't know what I was thinking on
 * when I put it there ...
 *
 * Revision 1.6  1998/07/14 14:57:56  inaky
 * Added a comment on locking.
 *
 * Revision 1.5  1998/07/12 22:58:20  inaky
 * Made all globals be correctly initialized.
 *
 * Added the code to initialize the root hub and to do HCI referencing.
 *
 * Revision 1.4  1998/06/21 20:03:23  inaky
 * Fixed the destroy method to clean after freeing used resources to fix
 * a problem which arised as it is called twice after cleaning the
 * struct. That caused a try to free resources twice.
 *
 * Revision 1.3  1998/06/20 21:55:52  inaky
 * Minor tweaks to fix bugs.
 *
 * Revision 1.2  1998/05/01 19:19:58  inaky
 * Fixed the abort method.
 *
 * Revision 1.1  1998/05/01 18:31:26  inaky
 * Moved everything to use pci_dev according to new kernel PCI API. Also
 * split all the code in several files, as old uhci.c was a complete mess.
 *
 */
