
/* USB Driver Stack v$Revision: 1.10 $
 *
 * (C) 1997 by Iaky Prez Gonzlez <inaky@peloncho.fis.ucm.es>
 *
 * Distributed under the terms of the GNU Public License Version 2
 * See files COPYRIGHT and COPYING at the root distribution directory
 *
 * $Id: intr-hash.c,v 1.10 1998/11/14 17:45:43 inaky Exp $
 *
 * Changelog at the bottom
 *
 *
 *     USB Universal Host Controller Interface
 *     ---------------------------------------
 *
 *  This file implements the interrupt transfers hash and lists table.
 *
 *  We implement interrupts as OHCI does. We only allow intervals of
 * 32, 16, 8, 4, 2 and 1 ms. We have 6 possible lists which don't have
 * to be linked into the frame list permanently, in a per host array
 * of 6 ts_uhci_tb_list's. With the function intr_hash_list() you
 * indicate you want a pointer to the list more suitable for the asked
 * interval [and if it wasn't inserted, it will be]. Then you have the
 * list and you can freely add/delete to/from it.
 *
 *  We have a list of sub-lists [one per interval] whose nodes are the
 * list's heads; the list->first.head pointer is used to point to the
 * next queue head, the list->first.element pointer is used to point
 * to the sub-list's elements.
 */

#include <uusbd/usbd.h>
#include <uusbd/usb-misc.h>

#include <asm/spinlock.h>
#include <linux/timer.h>

#define GDLF_IDENT          "USB-UHCI"
#define GDLF_LOCAL          /* Define if want local settings */
#define GDLF_DEF_LOGLEVEL   9
#define GDLF_ASSERTIONS     /* want assertions (#undef if not) */
#undef GDLF_FNTRACE        /* want function traces */
#define GDLF_LOCATIONS      /* want locations in log messages */
#include <uusbd/gdlf.h>

#include "uhci.h"           /* this has to be include after GDLF */


#define UHCI_INTR_MAX_INT 32    /* max interval allowed */
#define UHCI_INTR_MAX_IDX 5     /* log(2, MAX_INT) */


    /* Link a list into the frame list
     *
     * WARNING: doesn't hold/ack locks, call it with the list write
     *          locked!  
     *
     * Parameters:  
     *
     *  ts_uhci_host*   Host where we are inserting.
     *
     *  __u8            Index of the interval (log2 (interval)).
     */

void intr_hash_list_link (ts_uhci_host* hci, int index)
{
    ts_uhci_tb_list *list;
    int interval, skip_interval;
    int cnt;
    int frnum;
    __u32 ptr;
    
    gdlf_fntrace ("intr_hash_list_link (0x%p, %u)", hci, index);
    gdlf_aassert (hci != NULL, return);
    gdlf_aassert (hci->fl != NULL, return);
    gdlf_aassert (hci->intr.slot != NULL, return);
    gdlf_aassert (hci->intr.size > index, return);
    gdlf_aassert (hci->ls_ctrl != NULL, return);

    list = hci->intr.slot[index];
    gdlf_aassert (list != NULL, return);
    interval = 1 << index;

        /* get where should we link to and make the link */

    for (cnt = index - 1; cnt >= 0 && !hci->intr.slot[cnt]; cnt--);
    if (cnt == -1)        /* link to LS control list head */
        list->first.head.word = cpu_to_le32 (
            (virt_to_bus (&hci->ls_ctrl->first) & 0xfffffff0) | 0x2);
    else                    /* link to lower interval intr head */
        list->first.head.word = cpu_to_le32 (
            (virt_to_bus (&hci->intr.slot[cnt]->first) & 0xfffffff0) | 2);

        /*
         *  Now get who should we link to us; it'll be next highest
         * intr head or flp if this is the longer interval inserted.
         * Note this means we only link the flp/iso lp/td.ptr's whose
         * frame number is not a multiple of interval if we are the
         * longer interval registered.
         */

    for (cnt = index + 1; cnt < hci->intr.size && !hci->intr.slot[cnt]; 
         cnt++);
    if (cnt < hci->intr.size)
        skip_interval = 1 << cnt;
    else    /* link on all flp/iso lps */
    {
        skip_interval = UHCI_FL_SIZE;
        hci->intr.top = list;       /* we're the longer interval list */
    }

        /*
         *  Now walk all the frame list, and on each flp, if it would
         * match the interval we have to service, insert the QH. Note
         * we skip those flps for our interval which are also flps for
         * longer intervals, as they'll reach us anyway [as longer
         * intervals connect to shorter ones trough the list].
         */

    ptr = kvoid2ptr (list) >> 4;
    for (frnum = 0; frnum < UHCI_FL_SIZE; frnum += interval)
    {
        ts_uhci_td *itr, *prev;
        ts_uhci_qh *top;
        int context;
            
            /* 
             *  A longer interval present? we don't skip first frame
             * when the interval is the longer one.
             */

        if ((frnum % skip_interval == 0) 
            && !(frnum == 0 && skip_interval == UHCI_FL_SIZE))
            continue;

            /* Ok, walk from the flp until you find a QH */

        top = &hci->ls_ctrl->first;
        gdlf_aassert (top != NULL, continue);
        gdlf_aassert (hci->fl[frnum].bits.ptr != 0, continue);
        gdlf_aassert (hci->fl[frnum].bits.t != 1, continue);
        itr = ptr2kvoid (hci->fl[frnum].bits.ptr << 4);
        context = hci->fl[frnum].bits.q;
        prev = NULL;
            
        while (!context && (void*)itr != (void*)top)
        {
            prev = itr;
            gdlf_aassert (itr->lp.bits.ptr != 0, return);
            gdlf_aassert (itr->lp.bits.t != 1, return);
            itr = ptr2kvoid (itr->lp.bits.ptr << 4);
            context = itr->lp.bits.q;
        }

            /*
             * Ok, here we should have:
             *
             *  prev - the element previous to the QH, that means,
             *         the guy we should link to us. If NULL, it
             *         means we have to link to the flp.
             *
             *  itr - the victim, ignored
             */

        if (prev == NULL)   /* link to the flp */
            hci->fl[frnum].bits.ptr = ptr;
        else                /* link to the TD */
        {
#warning DEBUG: remove me
            gdlf_debug ("Linking interval idx 0x%x to "
                        "TD 0x%p of frame 0x%x\n", 
                        index, itr, frnum);
            itr->lp.bits.ptr = ptr;
        }
    }
    return;
}


    /* Unlink an interrupt head from the frame list
     *
     * WARNING: doesn't hold/ack locks, call it with the list write
     *          locked!  
     *
     * Parameters:  
     *
     *  ts_uhci_host*   Host we are removing from.
     *
     *  __u8            Index of the interval (log2 (interval)).
     */

void intr_hash_list_unlink (ts_uhci_host* hci, __u8 index)
{
    int cnt, interval, skip_interval;
    ts_uhci_tb_list *list;
    
    gdlf_fntrace ("intr_hash_list_unlink (0x%p, %u)", hci, index);
    gdlf_aassert (hci != NULL, return);
    gdlf_aassert (hci->fl != NULL, return);
    gdlf_aassert (hci->intr.slot != NULL, return);
    gdlf_aassert (hci->intr.size > index, return);

    list = hci->intr.slot[index];
    gdlf_aassert (list != NULL, return);
    gdlf_assert (list->first.head.word != cpu_to_le32 (0x3));
    gdlf_assert (list->first.elt.word == cpu_to_le32 (0x3));

        /*
         *  Firstly, seek the first interval longer than this one [if
         * any]. Make it point to the one next to this one and set
         * skip_interval to that interval, so we do not unlink it.
         */

    for (cnt = index + 1; cnt < hci->intr.size && !hci->intr.slot[cnt]; 
         cnt++);
    if (cnt < hci->intr.size)
    {
        gdlf_debug ("Making idx 0x%x link to my head, 0x%x\n",
                    cnt, list->first.head.word);
        hci->intr.slot[cnt]->first.head.word = list->first.head.word;
        skip_interval = 1 << cnt;   /* we'll have to skip each some frames */
    }
    else    /* nobody above us ... */
    {
        skip_interval = UHCI_FL_SIZE;    /* no need to skip */
        hci->intr.top = ptr2kvoid (list->first.head.word << 4);
    }
    interval = 1 << index;
    
        /* Ok, now walk the fl list and remove */

    for (cnt = 0; cnt < UHCI_FL_SIZE; cnt += interval)
    {
        ts_uhci_td *itr, *prev;
        int context;            

            /* 
             *  A longer interval present? we don't skip first frame
             * when the interval is the longer one.
             */

        if ((cnt % skip_interval == 0) 
            && !(cnt == 0 && skip_interval == UHCI_FL_SIZE))
            continue;

            /*
             *  Walk the fl chain, discarding iso TDs and seeking for
             * the first QH, that's the one we must remove.
             */

        gdlf_aassert (hci->fl[cnt].bits.ptr != 0, continue);
        gdlf_aassert (hci->fl[cnt].bits.t != 1, continue);
        itr = ptr2kvoid (hci->fl[cnt].bits.ptr << 4);
        context = hci->fl[cnt].bits.q;
        prev = NULL;
            
        while (!context)
        {
            gdlf_aassert (itr->lp.bits.t != 1, return);
            prev = itr;
            gdlf_aassert (itr->lp.bits.ptr != 0, return);
            gdlf_aassert (itr->lp.bits.t != 0, return);
            itr = ptr2kvoid (itr->lp.bits.ptr << 4);
            context = itr->lp.bits.q;
        }

            /*
             * Ok, here we should have:
             *
             *  prev - the element previous to the QH, that means,
             *         the guy we should unlink from us. If NULL, it
             *         means we have to unlink from the flp.
             */

        if (prev == NULL)   /* link to the flp */
            hci->fl[cnt].word = list->first.head.word;
        else                /* link to the TD */
            itr->lp.word = list->first.head.word;
    }
    list->first.head.word = cpu_to_le32 (0x3);  /* terminate the list */
    return;
}


    /* Do a cleanup of the intr lists
     *
     *  Seeks which ones are empty and removes them
     *
     * Parameters:  
     *
     *  ts_uhci_host*   Pointer to HCI host where to do the clean-up.
     *
     *  unsigned long   Last time [jiffies] we were called. May be
     *                  used IANF for an adaptative algorithm. If 0,
     *                  then it means we should do a general cleanup.
     */

void intr_hash_cleanup (ts_uhci_host* hci, unsigned long last_time)
{
    ts_uhci_intr_hash *hash;
    ts_uhci_tb_list **itr, **top;
    int index;

    gdlf_fntrace ("intr_hash_cleanup (0x%p, %lu)", hci, last_time);
    gdlf_aassert (hci != NULL, return);
    
    hash = &hci->intr;
    if (hash->slot == NULL)    /* Unneeded? */
        return;

        /* Hmmm ... no spinlocks on UP */

    atomic_inc (&hash->usecnt);
    if (atomic_read (&hash->usecnt) > 1)
    {
        atomic_dec (&hash->usecnt);
        gdlf_info ("Interrupt hash 0x%p is in use, cleanup "
                   "skipped until next run\n", hash);
        return;
    }

    write_lock (&hash->lock);
    
        /*
         *  Walk the array of lists and unlink + delete the empty
         * ones. If general cleanup, all lists must be already empty,
         * you know.
         */

    for (index = 0, itr = hash->slot, top = itr + hash->size; 
         itr < top; itr++, index++)
    {
        if (*itr        /* empty? */
            && ((*itr)->first.elt.bits.t == 1
                || last_time == 0))
        {
#warning FIXME: add check on insertion statistical data            
            intr_hash_list_unlink (hci, index);
            ts_uhci_tb_list_release (*itr);
            usb_kfree (*itr);
            *itr = NULL;
        }
    }

    if (last_time == 0)
    {
        usb_kfree (hash->slot);
        hash->size = 0;
        hash->slot = NULL;
    }
    
    write_unlock (&hash->lock);
    atomic_dec (&hash->usecnt);
    return;
}


    /* FIXME: Detailed desc
     *
     * Compliance:      USB Spec 1.0 xxxx
     *
     * Parameters:  
     *
     *  type            descr
     *
     * Global:
     *
     *  type name       descr
     *
     * Returns:
     *
     *  type            descr
     *
     * Global:
     *
     *  type name       descr
     */

ts_uhci_tb_list* intr_hash_list (ts_uhci_host* hci, __u8 interval)
{
    __u8 index;
    ts_uhci_intr_hash* hash;
    ts_uhci_tb_list* tb;
    
    gdlf_fntrace ("intr_hash_list (0x%p, %u)", hci, interval);
    gdlf_aassert (hci != NULL, return NULL);
    gdlf_aassert (interval != 0, return NULL);
    
    hash = &hci->intr;

    index = log2 (interval);
    if (index > UHCI_INTR_MAX_IDX)  /* more than UHCI_INTR_MAX_INT? ... */
        index = UHCI_INTR_MAX_IDX;  /* ... you'll have to stick with it .. */

        /* if array smaller, realloc, clean new ones */

    write_lock (&hash->lock);
    if (index >= hash->size)
    {
        ts_uhci_tb_list **new_slot;
        size_t new_size;

            /* 
             *  The idea in doing it this way is always allocating
             * powers of two, and for that to be accomplished,
             * new_size has to be also a power of two
             */

        new_size = hash->size;      /* actual size */
        do 
        {
            new_size = new_size == 0? 2 : 2 * new_size;
        } while (new_size <= index); /* if small try next */
        new_slot = usb_realloc (hash->slot, 
                                hash->size * sizeof (ts_uhci_tb*), 
                                new_size * sizeof (ts_uhci_tb*), 
                                GFP_KERNEL);

        if (new_slot == NULL)
        {
            gdlf_err ("Cannot reallocate lists for interval %d of "
                      "UHCI host 0x%p interrupts hash\n",
                      1 << index, hci);
            write_unlock (&hash->lock);
            return NULL;
        }
        hash->slot = new_slot;
        memset (new_slot + hash->size, 0, 
                (new_size - hash->size) * sizeof (ts_uhci_tb*));
        hash->size = new_size;      /* set the new size */
        tb = NULL;
    }
    else
        tb = hash->slot[index];

        /* if list unexistant, create, init it, link it */
    
    if (tb == NULL)
    {
        tb = ts_uhci_tb_list_init();
        if (tb == NULL)
        {
            gdlf_err ("Cannot allocate a list for interval %d of "
                      "UHCI host 0x%p interrupts hash\n",
                      1 << index, hci);
            write_unlock (&hash->lock);
            return NULL;
        }
        
        tb->first.elt.word = cpu_to_le32 (0x3);
        tb->first.head.word = cpu_to_le32 (0x3);
        tb->last = NULL;    /* needed */
        hash->slot[index] = tb;
        intr_hash_list_link (hci, index);
        gdlf_aassert (tb->first.head.word != cpu_to_le32 (0x3), 
                      {
                          usb_kfree (tb);
                          hash->slot[index] = NULL;
                          return NULL;
                      }); 
    }

        /* return the list */

    write_unlock (&hash->lock);
    return tb;
}


/* Log:
 *
 * $Log: intr-hash.c,v $
 * Revision 1.10  1998/11/14 17:45:43  inaky
 * Moved the include files to uusbd/include/uusbd, to ease integration
 * into the Linux kernel.
 *
 * Revision 1.9  1998/10/31 02:21:03  inaky
 * Using the log2e at the misc stuff.
 *
 * Revision 1.8  1998/10/22 00:10:46  inaky
 * Removed debug stuff, fixed the stupid _bug_ which made lots of fl
 * entries be set to zero [when setting list->head.word = 0x3 after the
 * first fl being set to list->first.head.word, the rest would be zero].
 *
 * Revision 1.7  1998/10/14 21:52:48  inaky
 * Comment fixes.
 *
 * Revision 1.6  1998/09/24 16:15:55  inaky
 * Added link on frame zero, made the cleanup function do a full cleanup
 * if last_time argument is zero.
 *
 * Revision 1.5  1998/09/14 20:35:14  inaky
 * Fixed counts in hash walks, fixed reallocator to take into account
 * pointers have a size (yo stupid!).
 *
 * Revision 1.4  1998/09/02 21:19:57  inaky
 * Moved uhci.h inclusion after gdlf.
 *
 * Revision 1.3  1998/08/30 13:59:46  inaky
 * Fixed and finished interrupt transactions in a linear list topology.
 *
 * Revision 1.2  1998/08/29 00:11:27  inaky
 * Fixed endianess on qhs and tds.
 *
 * Revision 1.1  1998/08/27 12:08:24  inaky
 * Seems to be ok, have to test it.
 *
 */
