
/* USB Driver Stack v$Revision: 1.18 $
 *
 * (C) 1997 by Iaky Prez Gonzlez <inaky@peloncho.fis.ucm.es>
 *
 * Distributed under the terms of the GNU Public License Version 2
 * See files COPYRIGHT and COPYING at the root distribution directory
 *
 * $Id: irq-handler.c,v 1.18 1998/11/14 17:45:44 inaky Exp $
 *
 * Changelog at the bottom
 *
 *
 *     USB Universal Host Controller Interface
 *     ---------------------------------------
 *
 *  This file contains the core logic for transfer irq service and
 * house keeping. 
 *
 *  See scheduler.c, at the top, for how to distinguish TB status.
 */

#include <uusbd/usbd.h>

#undef GDLF_UHCI_IRQ_HANDLER_C /* Define if want local settings override globals */
#define GDLF_IDENT          "USB-UHCI"
#define GDLF_LOCAL          /* Define if want local settings */
#define GDLF_DEF_LOGLEVEL   9
#define GDLF_ASSERTIONS     /* want assertions (#undef if not) */
#undef GDLF_FNTRACE        /* want function traces */
#define GDLF_LOCATIONS      /* want locations in log messages */
#include <uusbd/gdlf.h>

#include "uhci.h"           /* this has to be include after GDLF */


/*
 *        How to guess a TB status inside the Frame List
 *        ----------------------------------------------
 *
 * Status    Isochronous / TDs          Interrupt, Control & Bulk (QH+TDs)
 * --------- -------------------------- ----------------------------------
 * INSERTED 
 * QUEUED    td.ctstat.active == 0      qh.elt.t == 1
 *           td.ctstat-status == 0      qh.elt.q == 0
 * (1)       td.ctstat.actlen == 0x7fe  qh.elt.lp == &td[0]
 *           td.ctstat.c_err == 3       td.lp.vf == 1
 *                    
 * EXECUTING td.ctstat.active == 1      qh.elt.t == 0
 *           td.ctstat-status == 0      qh.elt.q == 0
 *           td.ctstat.actlen != 0x7fe  qh.elt.lp =? &td[0]
 *           td.ctstat.c_err =? 3       td.lp.vf == 1
 *                    
 * DONE      td.ctstat.active == 0      qh.elt.t == 1
 *           td.ctstat-status == 0      qh.elt.q == 0      
 *           td.ctstat.actlen != 0x7fe  (tds > 1 && qh.elt.lp == td[tds-2].lp && td[tds-1].ctstat.active == 0) || (tds == 1 && td[0].ctstat.active == 0)
 *           td.ctstat.c_err =? 3       td.lp.vf == 1
 *                                      td.cstat-status == 0
 *  This is wrong. On done, qh.elt = td[tds-1].lp
 *
 *                   
 * ERROR     td.ctstat.active == 0      qh.elt.t == 0 -> move to 1
 *           td.ctstat-status != 0      qh.elt.q == 0
 *           td.ctstat.actlen != 0x7fe  (tds > 1 && qh.elt.lp =? td[tds-2].lp && td.ctstat.active == 0) || (tds == 1 && td[0].ctstat.active == 0)
 *           td.ctstat.c_err =? 3       
 *
 *
 *   =?     Maybe equal, may be not ...
 *   td.    All TDs
 *   td[n]. TD number n
 *
 * (1) Max pkt size is 0x3ff (1024). 0x7ff means a NULL packet (no
 *     data transaction), so we use 0x7fe as water-mark value.
 */


    /* Check if a TD is in the inserted or queued states
     *
     *  The way to check it is, based on the table above, we AND to
     * select the bits we want to test (0x18fe07ff clears everything
     * but c_err, status and actlen) and then it's gotta be equal to
     * the values we want, this is, inactive, no transaction-based
     * errors, actlen UHCI_ACTLEN and c_err UHCI_RETRIES.
     *  
     * Parameters:  
     *
     *  ts_uhci_td*     Pointer to the TD
     *
     * Returns:
     *
     *  int             0 if not, !0 if in INSERTED or QUEUED states.
     */

inline
int td_is_inserted_or_queued_q (ts_uhci_td* td)
{
    return 
        (td->ctstat.word & cpu_to_le32 (0x18fe07ff)) 
        == cpu_to_le32 (0x180007fe);
}


    /* Check if a TD is executing
     *
     *  The way to check it is, based on the table above, we AND to
     * select the bits we want to test (0x00fe0000 clears everything
     * but status) and then it's gotta be equal to the values we want,
     * this is, active and no transaction-based errors (0x00800000).
     *  
     * Parameters:  
     *
     *  ts_uhci_td*     Pointer to the TD
     *
     * Returns:
     *
     *  int             0 if not, !0 if in EXECUTING state.
     */

inline
int td_is_executing_q (ts_uhci_td* td)
{
    return td == NULL? 
        0 : (td->ctstat.word & cpu_to_le32 (0x00fe0000)) 
        == cpu_to_le32 (0x00800000);
}


    /* Check if a TD is done
     *
     *  Using the table above, we AND to select the bits we want
     * (0x00fe0000 selects status bits). It must be zero (inactive and
     * no tranmision errors).
     *
     * Parameters:  
     *
     *  ts_uhci_td*     Pointer to the TD
     *
     * Returns:
     *
     *  int             0 if not, !0 if in DONE state.
     */

inline
int td_is_done_q (ts_uhci_td* td)
{
    return 
        (td->ctstat.word & cpu_to_le32 (0x00fe0000)) 
        == cpu_to_le32 (0x00000000);
}


    /* Check if a TD is in error state
     *
     *  Using the table above, we AND to select the bits we want
     * (0x00fe0000 selects the status bits). If it is not active
     * (& 0x00800000 == 0) and other status bits are set (& 0x007e0000
     * != 0), then posibly there is an error (it still may happen it's
     * a NAK).
     *
     * Parameters:  
     *
     *  ts_uhci_td*     Pointer to the TD
     *
     * Returns:
     *
     *  int             0 if not, !0 if in ERROR state.
     */

inline
int td_is_error_q (ts_uhci_td* td)
{
    return td == NULL? 1 : 
        ((td->ctstat.word & cpu_to_le32 (0x00800000)) 
         == cpu_to_le32 (0x00000000))
        && ((td->ctstat.word & cpu_to_le32 (0x007e0000)) 
            != cpu_to_le32 (0x00000000));   /* pun ... */
}


    /* See if a queued TB is in EXECUTING state
     *
     *  Using the table, we see that an EXECUTING TB has it's QH's
     * element t bits set to zero, and if the current TD is EXECUTING
     * (as determined by iso_td_is_executing_q()).
     *
     * Parameters:  
     *
     *  ts_uhci_tb*     Pointer to transactions's TB.
     *
     *  ts_uhci_td*     Pointer to last executed/processed TD.
     *
     * Returns:
     *
     *  int             !0 the TB is in EXECUTING state, 0 it is
     *                  not.
     */

inline
int tb_is_executing_q (ts_uhci_tb* tb, ts_uhci_td* td)
{
    return tb->qh.elt.bits.t == 0 && td_is_executing_q (td);
}


    /* Check if a queued TB is in ERROR state
     *
     *  Using the table, we see that an ERROneous TB has it's QH's
     * element t bits set to zero, and the current executing TD is in
     * ERROR state [as determined by td_is_error_q()].
     *
     * Parameters:  
     *
     *  ts_uhci_tb*     Pointer to transactions's TB.
     *
     *  ts_uhci_td*     Pointer to last executed/processed TD.
     *
     * Returns:
     *
     *  int             !0 the TB is in ERROR state, 0 it is
     *                  not.
     */

int tb_is_error_q (ts_uhci_tb* tb, ts_uhci_td* td)
{
    return tb->qh.elt.bits.t == 0 && (td == NULL || td_is_error_q (td));
}


    /* Check if a queued TB is in DONE state
     *
     *  When the HC completes a queue, the last TD's lp pointer bits
     * are copied to the QH's element pointer, and thus, as last
     * pointers are _always_ 0x1 [NULL ptr, Q = 0 and T = 1], it must
     * be 0x1. 
     *
     * Parameters:  
     *
     *  ts_uhci_tb*     Pointer to transactions's TB.
     *
     * Returns:
     *
     *  int             !0 the TB is in DONE state, 0 it is
     *                  not.
     */

int tb_is_done_q (ts_uhci_tb* tb)
{
    return (tb->qh.elt.word & cpu_to_le32 (0xfffffff3)) 
        == cpu_to_le32 (0x00000001);
}


    /* Check if a queued TB has interrupted because of a warning
     * packet.
     *
     *  A TB having interrupted in warning state is like a done one,
     * but the QH's element pointer _doesn't_ point to the last
     * pointer in the list of TD's to process. 
     *
     *  FIXME: last one will be missed, hack this, or better:
     *         implement it correctly. Also a single-TD transfer with
     *         packet warning [stupid, but possible] and no done
     *         warning won't work.
     *
     * Parameters:  
     *
     *  ts_uhci_tb*     Pointer to transactions's TB.
     *
     *  ts_uhci_td*     Pointer to last executed/processed TD.
     *
     * Returns:
     *
     *  int             !0 the TB was warning, 0 it is
     *                  not.
     */

int tb_is_warning_q (ts_uhci_tb* tb, ts_uhci_td* td)
{
    gdlf_aassert (tb != NULL, return 0);
    
        /* If TD not done, nothing */

    if (!(tb->qh.elt.bits.t == 0 && td_is_done_q (td)))
        return 0;
    
    if (tb->tds == 1)
    {
            /* Single packet and no done_warn? pkt warning */

        if (!tb->tr->done_warn)
            return 1;
        else
            return 0;
    }
        /* Multi packet not at end? pkt warning */

    else if (tb->qh.elt.bits.ptr != td[tb->tds - 2].lp.bits.ptr)
        return 1;

        /* Multi packet at end and no done_warn? pkt warning */

    else if (!tb->tr->done_warn)
        return 1;

        /* Nothing ... */

    else 
        return 0;
}


    /* Move a transaction to ERROR state.
     *
     *  A transaction moved to error status is as an INSERTED
     * transaction, but with the state set to e_usb_tr_status_error,
     * and usb_errno and errno set.
     *
     * Parameters:  
     *
     *  ts_uhci_tb*     Transfer block to move
     *
     *  int             USB Errno code for the error.
     *
     *  int             ERRNO code for the error.
     */

void move_to_error (ts_uhci_tb* tb, int usb_errno, int errno)
{
    ts_usb_tr *tr;
    ts_usb_tr *newtr;
    
    gdlf_fntrace ("move_to_error (0x%p, %d, %d)", 
                  tb, usb_errno, errno);
    gdlf_aassert (tb != NULL, return);

        /* First thing we do, terminate the TB, to keep the HC away */

    tb->qh.elt.bits.t = 1;
    
    tr = tb->tr;
    gdlf_aassert (tr != NULL, return);
    gdlf_aassert (tr->status == e_usb_tr_status_ok || 
                  tr->status == e_usb_tr_status_executing,
                  return);
 
#warning DEBUG: remove me
    gdlf_debug ("Transaction tr 0x%p (tb 0x%p) is on error: usb errno %d, errno %d\n", 
                tb->tr, tb, usb_errno, errno);
    ts_uhci_tb_printk (tb);

        /*
         *  If called with EXECUTING status, it means move_to_done()
         * has detected a transaction error and is forwarding the TR
         * to us, so we don't need to remove the timer, as it's
         * already done. 
         */

    if (tr->status == e_usb_tr_status_executing)
    {
        tr->status = e_usb_tr_status_error;     /* evitate timer jumpin in */    
        del_timer (&tr->timer);
    }
    else
        tr->status = e_usb_tr_status_error;     /* evitate timer jumpin in */    
        
        /* 
         * reset this one and move it to the done part of the
         * endpoints queue 
         *
         *  The check is needed. When a device is unplugged, the open
         * structure is cleaned to invalidate connections to it, and
         * we don't wan't it to crash. We use an special type then,
         * the emergency :)
         */

    uhci_tb_reset (tb, tr->open->endp? 
                   tr->open->endp->type : e_usb_tr_type_inv);
    ts_slist_move (&tr->open->list, &tr->snode, tr->open->finished);
    tr->open->finished = &tr->snode;

        /* 
         *  If this is true, tr was the only running transaction in
         * the open's list, so have to set it to NULL, to indicate
         * there are no more running tr's on the endp.
         */

    if (tr->open->running == &tr->snode)
        tr->open->running = NULL;
    
        /* 
         * activate next one for this endpoint if it is not the
         * transaction we are poking now 
         */ 

    newtr = (ts_usb_tr*)ts_slist_head (&tr->open->list);
    if (newtr != NULL && &((ts_uhci_tr_hci*)newtr->hci)->tb != tb) 
        uhci_tb_activate (&((ts_uhci_tr_hci*)newtr->hci)->tb);
    
            /* Set statistics and errors */

    tr->usb_errno = usb_errno;
    tr->errno = errno;
    tr->xfered = 0;
#warning FIXME: has to calculate the ratios and it may want to retry
    update_statistics (tr);         /* update statistics */

        /* Call [done] warnings */

    if (tr->done_warn && tr->warn.wq != NULL)
    {
        if (tr->method)
            (*tr->warn.callback) (tr, USB_CB_ERROR | USB_CB_DONE);
        else
            wake_up (tr->warn.wq);
    }
    return;
}


    /* Move a transaction to DONE state.
     *
     *  A transaction moved to DONE status is as an INSERTED
     * transaction, but with the state set to e_usb_tr_status_ok,
     * and usb_errno and errno set to zero, as well as the statistical
     * data filled in (xfered and errors).
     *
     * Parameters:  
     *
     *  ts_uhci_tb*     Transfer block to move
     */

void move_to_done (ts_uhci_tb* tb)
{
    ts_usb_tr *tr;
    ts_usb_tr *newtr;
    size_t xfered, errors, ratio;
    ts_uhci_td *itr, *top;
    
    gdlf_fntrace ("move_to_done (0x%p)", tb);

        /* No need to set tb->qh.elt.t bit here, it's already set */

    tr = tb->tr;                        /* set errors */
    tr->status = e_usb_tr_status_ok;    /* evitate timer jumpin in! */
    gdlf_aassert (tr != NULL, return);
    
        /* Remove the timer */
        
    del_timer (&tr->timer);

        /* 
         * Check short packets and NAKs. if any undesired one, it's an
         * error, call move_to_error(). Note we MUST do this before
         * resetting, as resetting will clear all the fields.
         *
         *  Also collect statistics on transfered bytes and errors
         * [mind the count held on the TD is actually 'UHCI_RETRIES -
         * number of errors', and thus, if I add all those counts, I
         * need to substract UHCI_RETRIES per TD to get the total
         * number of errors].
         *
         *  We can safely assume, when adding xfered bytes, that all
         * the actlen's are ok, so we just add one to the actlen and
         * and to 0x7ff. This way, when actlen=0x7ff (NULL packet),
         * 0x7ff + 1 & 0x7ff = 0x800 & 0x7ff == 0.
         */

    for (itr = tb->td, top = itr + tb->tds,
             xfered = 0, errors = 0; 
         itr < top; itr++)
    {
        if (itr->ctstat.bits.nak    /* Unwanted NAK? */
            && tr->nak_error)
        {
            move_to_error (tb, -USBE_BUS_NAK_RECEIVED, -ENODATA);
            return;
        }
        xfered += (itr->ctstat.bits.actlen + 1) & 0x7ff;
        errors += itr->ctstat.bits.c_err;
    }
    errors = tb->tds * UHCI_RETRIES - errors;
        /* short packets will be catched by client if xfered < scheduled */
    tr->xfered= xfered;
    tr->errors= errors;
    
#warning FIXME: polish this of error ratios some more ...

    if (errors > tr->max_errors)
    {
        gdlf_err ("%u errors for tr 0x%p on device 0x%p (%d %d) "
                  "(limit was %u)\n",
                  errors, tr, tr->open->dev, 
                  tr->open->dev->hub->host->addr,
                  tr->open->dev->addr,
                  tr->max_errors);
        move_to_error (tb, -USBE_ERROR_COUNT, -EIO);
        return;        
    }

    ratio = xfered != 0? errors * 1024 / xfered : (size_t)-1;
    if (ratio > tr->max_ratio)
    {
        gdlf_err ("%u errors/k for tr 0x%p on "
                  "device 0x%p (%d %d) "
                  "(limit was %u errors/k)\n",
                  ratio, tr, tr->open->dev, 
                  tr->open->dev->hub->host->addr,
                  tr->open->dev->addr,
                  tr->max_ratio);
        move_to_error (tb, -USBE_ERROR_RATIO, -EIO);
        return;        
    }

        /* 
         *  It seems it's ok, so we can keep on. We can move it to the
         * done part of the endpoint's transactions list. 
         */

    uhci_tb_reset (tb, tr->open->endp->type);
    ts_slist_move (&tr->open->list, &tr->snode, tr->open->finished);
    tr->open->finished = &tr->snode;

        /* 
         *  If this is true, tr was the only running transaction in
         * the open's list, so have to set it to NULL, to indicate
         * there are no more running tr's on the endp.
         */

    if (tr->open->running == &tr->snode)
        tr->open->running = NULL;
    
        /* 
         * activate next for this endpoint if it is not the same one
         * we are poking with now.
         */

    newtr = (ts_usb_tr*)ts_slist_head (&tr->open->list);
    if (newtr != NULL && &((ts_uhci_tr_hci*)newtr->hci)->tb != tb) 
        uhci_tb_activate (&((ts_uhci_tr_hci*)newtr->hci)->tb);
        
        /*
         *  Check if gotta loop; if so, reprogram a loop. You'll need
         * to reset the warnings and some more fields.
         */

#warning FIXME: reprogram loops

    tr->usb_errno = 0;
    tr->errno = 0;
    update_statistics (tr);         /* update statistics */

        /* Call [done] warnings */

    if (tr->done_warn && tr->warn.wq)
        if (tr->method)
            (*tr->warn.callback) (tr, USB_CB_DONE);
        else
            wake_up (tr->warn.wq);

    return;
}


    /* FIXME: Detailed desc
     *
     * Compliance:      USB Spec 1.0 xxxx
     *
     * Parameters:  
     *
     *  type            descr
     *
     * Global:
     *
     *  type name       descr
     *
     * Returns:
     *
     *  type            descr
     *
     * Global:
     *
     *  type name       descr
     */

void update_statistics (ts_usb_tr* tr)
{
    gdlf_fntrace ("update_statistics (0x%p)", tr);
#warning FIXME: update_statistics() not implemented
    return;
}


    /* See what's up with a queued TB
     *
     *  We will look, in a TB for it being in any of the DONE or ERROR
     * states. If so, we'll move it's tr out to that states. Else,
     * we'll let them be.
     *
     * Parameters:  
     *
     *  ts_uhci_tb*     Pointer to the TB to process.
     *
     *  int             HC status
     *
     *  int             UHCI_FL_FAST
     *
     *                     exit as soon as you find and repair the
     *                     cause of a PCIBE or HCPE error [searching
     *                     the full frame, but skipping it entirely
     *                     once found the problem].
     *
     *                  Other
     *
     *                      Proceed normally.
     *
     * Returns:
     *
     *  int             0   if ok
     *
     *                  !0  if a TB which halted the HC was found
     */

int cook_queued_tb (ts_uhci_tb* tb, int stat, int mode)
{
    ts_uhci_td* td; /* last processed TB's TD */
    
    gdlf_fntrace ("cook_queued_tb (0x%p, 0x%x, %d)", tb, stat, mode);
    gdlf_aassert (tb != NULL, return -EINVAL);
    
        /*
         *  First, check if the transaction has finished. This should
         * be the most common case.
         */

    if (tb_is_done_q (tb) && (stat & USBSTS_USBINT))
    {
        move_to_done (tb);
        return 0;
    }
    else        /* Transaction not finished or any error? */
        td = tb->qh.elt.bits.ptr == 0? 
            NULL : bus_to_virt (le32_to_cpu (tb->qh.elt.bits.ptr << 4));

        /*
         *  Here we use this evaluation order because we know that if
         * there's an HCPE or PCIBE error, the TD which will have
         * caused it is the following to the one active in the QH's
         * element pointer. We have to make sure that the td we think
         * may have caused the error is really here, so if the queue
         * is already completed, this one cannot be; that's why the
         * check in the middle line is for.
         */

    if ((stat & (USBSTS_PCIBE | USBSTS_HCPE))
        && (tb_is_executing_q (tb, td) 
            || tb_is_error_q (tb, td)))
    {
        gdlf_debug ("tr 0x%p tb 0x%p td 0x%p finished on HCPE/PCIBE error\n", 
                    tb->tr, tb, td);
            /* HC halted error, move to error */
        move_to_error (tb, 
                       stat & USBSTS_PCIBE? 
                       -USBE_HOST_BUS_ERROR : -USBE_HCI_DATA_BUG,
                       stat & USBSTS_PCIBE? -EIO : -EINVAL);
        return 1;
    }
    else if (tb_is_error_q (tb, td))
    {
        int errno, usb_errno;         
            /*
             * This guy has an error inside it! 
             *
             *  On serious errors, we complain at klog.
             *
             *  The NAK is not a `true' error except when on
             * a SETUP packet transaction, when the timeout
             * error is also set, so we don't worry for it (and
             * anyhow, it wouldn't trigger an error interrupt).
             *
             *  We only check the last TD, as the UHCI controller will
             * STOP processing a QUEUE with an inactive TD on it
             * (state to which TDs are moved if errors are found).
             */

#warning FIXME: Indicate each error, and if ctl xfer, phase

#warning DEBUG+FIXME: this case should not happen
        if (td == NULL)
        {
            gdlf_emerg ("BUUUG!!! tb is 0x%p, td 0x%p, executing %d, error %d\n",
                        tb, td, tb_is_executing_q (tb, td), 
                        tb_is_error_q (tb, td));
            errno = -EINVAL;
            usb_errno = -USBE_BUG;
        }
        else if (td->ctstat.bits.crcto_err)
        {
            if (td->token.bits.pid == USB_PID_SETUP 
                && td->ctstat.bits.nak)
            {
                gdlf_err ("NAK received for SETUP PID "
                          "tb 0x%p td 0x%p device 0x%p (%d %d) "
                          "endp %d\n",
                          tb, td, tb->tr->open->dev,
                          tb->tr->open->dev->hub->host->addr,
                          tb->tr->open->dev->addr,
                          tb->tr->open->endp->addr >> 1);
                errno = -EPROTO;
            }
            else
            {
                gdlf_err ("Timeout/CRC error "
                          "tb 0x%p td 0x%p device 0x%p (%d %d) "
                          "endp %d\n",
                          tb, td, tb->tr->open->dev,
                          tb->tr->open->dev->hub->host->addr,
                          tb->tr->open->dev->addr,
                          tb->tr->open->endp->addr >> 1);
                errno = -EIO;
            }
            usb_errno = -USBE_BUS_CRCTO;
        }
        else if (td->ctstat.bits.databuf_err)
        {
            gdlf_err ("Data buffer error, td 0x%p, tb 0x%p, "
                      "tr 0x%p, device 0x%p (%d %d), "
                      "endpoint %d\n",
                      td, tb, tb->tr, tb->tr->open->dev,
                      tb->tr->open->dev->hub->host->addr, 
                      tb->tr->open->dev->addr,
                      tb->tr->open->endp->addr >> 1);
            errno = -ENODATA;
            usb_errno = -USBE_DATA_BUFFER_ERROR;
        }  
        else if (td->ctstat.bits.babble)
        {
            gdlf_err ("Babble error, td 0x%p, tb 0x%p, "
                      "tr 0x%p, device 0x%p (%d %d), "
                      "endpoint %d\n",
                      td, tb, tb->tr, tb->tr->open->dev,
                      tb->tr->open->dev->hub->host->addr, 
                      tb->tr->open->dev->addr,
                      tb->tr->open->endp->addr >> 1);
            errno = -EIO;
            usb_errno = -USBE_BUS_BABBLE;
        }
        else if (td->ctstat.bits.stalled)    /* stalled? */
        {
            errno = -EREMOTEIO;
            usb_errno = -USBE_ENDP_STALLED;
        }
        else if (td->ctstat.bits.bitst_err)
        {
            gdlf_err ("Bit Stuff error, td 0x%p, tb 0x%p, "
                      "tr 0x%p, device 0x%p (%d %d), "
                      "endpoint %d\n",
                      td, tb, tb->tr, tb->tr->open->dev,
                      tb->tr->open->dev->hub->host->addr, 
                      tb->tr->open->dev->addr,
                      tb->tr->open->endp->addr >> 1);
            errno = -EIO;
            usb_errno = -USBE_BUS_BITSTUFF;
        }
        else    /* Ignore if just a NAK and no errors on NAK */
        {
            errno = 0;
            usb_errno = 0;  /* shut up gcc */
        }
            
        if (errno != 0)
        {
            gdlf_debug ("tr 0x%p td 0x%p finished on USB error\n", 
                        tb->tr, td);
            move_to_error (tb, usb_errno, errno);
        }
    }
    else if (tb_is_warning_q (tb, td))
    {
            /* Warning packet, let's call it */
        update_statistics (tb->tr);
        if (td == NULL)     /* Hmm .. HCPE or PCIBE */
        {                   /* serious anyway, should not be here */
            gdlf_emerg ("BUG! I should not be here! NULL td "
                        "are HCPE or PCIBE\n");            
        }
        else if (tb->tr->pkts && tb->tr->warn.wq != NULL)   /* warning set? */
            if (tb->tr->method)
                (*tb->tr->warn.callback) (tb->tr, USB_CB_PKTS);
            else
                wake_up (tb->tr->warn.wq);
    }

        /* If here, the transaction didn't have any error or anything */
        
    return 0;
}


    /* See's what does it have to do with an ISO TD
     *
     *  If the TD is INSERTED, QUEUED or EXECUTING, it is left like
     * that, if it is on ERROR or DONE, it is moved.
     *
     * Parameters:  
     *
     *  ts_uhci_td*     Pointer to isochronous TD to process.
     *
     *  ts_uhci_host*   Pointer to the UHCI host.
     *
     *  int             HC status
     *
     *  int             UHCI_FL_FAST
     *
     *                     exit as soon as you find and repair the
     *                     cause of a PCIBE or HCPE error [searching
     *                     the full frame, but skipping it entirely
     *                     once found the problem].
     *
     *                  Other
     *
     *                      Proceed normally.
     *
     *  Returns:
     *
     *  int             0   if ok
     *
     *                  !0  if a TD which halted the HC was found
     */

int cook_iso_td (ts_uhci_td* td, ts_uhci_host* hci, int stat, int mode)
{
    ts_uhci_tb* tb;
    
    gdlf_fntrace ("cook_iso_td (0x%p, 0x%p, 0x%x, %d)", 
                  td, hci, stat, mode);
    gdlf_aassert (td != NULL, return -EINVAL);
    gdlf_aassert (hci != NULL, return -EINVAL);
    
        /* 
         *  If this TD is executing and we have the host halted, then
         * there's a PCIBE or HCPE error, so we move it to error; if
         * we were called with 'mode == UHCI_FL_FAST', then we exit so
         * the caller can restart the controller ASAP, else we just
         * continue seeing the place.  
         */
            
    if (stat & (USBSTS_PCIBE | USBSTS_HCPE)
        && td_is_executing_q (td))
    {
        tb = get_iso_tb_by_td (&hci->iso, td);  /* locate the TB */
        if (tb->tr->status == e_usb_tr_status_executing)
            move_to_error (tb, 
                           stat & USBSTS_PCIBE? 
                           -USBE_HOST_BUS_ERROR : -USBE_HCI_DATA_BUG,
                           stat & USBSTS_PCIBE? 
                           -EIO : -EINVAL);
        return 1;
    }   
    else if (td_is_error_q (td))
    {
        int errno, usb_errno;         
                
            /*
             * This guy has an error inside it! 
             *
             *  On serious errors, we complain at klog.
             *
             *  The NAK is not a `true' error except when on
             * a SETUP packet transaction, when the timeout
             * error is also set, but however, this should only be
             * processing ISO transaction ... so we don't worry for
             * it. 
             */

        tb = get_iso_tb_by_td (&hci->iso, td);     /* locate the TB */

        if (tb->tr->status == e_usb_tr_status_executing)
        {               /* if executing, do nothing */
            errno = 0;
            usb_errno = 0;
        }
        else if (td->ctstat.bits.crcto_err)
        {
            if (td->token.bits.pid == USB_PID_SETUP 
                && td->ctstat.bits.nak)
            {
                errno = -EPROTO;
                gdlf_err ("Ein? ... SETUP pid on iso transaction? "
                          "tb 0x%p td 0x%p device 0x%p (%d %d) "
                          "endp %d\n",
                          tb, td, tb->tr->open->dev,
                          tb->tr->open->dev->hub->host->addr,
                          tb->tr->open->dev->addr,
                          tb->tr->open->endp->addr >> 1);
            }
            else
                errno = -EIO;
            usb_errno = -USBE_BUS_CRCTO;
        }
        else if (td->ctstat.bits.databuf_err)
        {
            gdlf_err ("Data buffer error, td 0x%p, tb 0x%p, "
                      "tr 0x%p, device 0x%p (%d %d), "
                      "endpoint %d\n",
                      td, tb, tb->tr, tb->tr->open->dev,
                      tb->tr->open->dev->hub->host->addr, 
                      tb->tr->open->dev->addr,
                      tb->tr->open->endp->addr >> 1);
            errno = -ENODATA;
            usb_errno = -USBE_DATA_BUFFER_ERROR;
        }  
        else if (td->ctstat.bits.babble)
        {
            gdlf_err ("Babble error, td 0x%p, tb 0x%p, "
                      "tr 0x%p, device 0x%p (%d %d), "
                      "endpoint %d\n",
                      td, tb, tb->tr, tb->tr->open->dev,
                      tb->tr->open->dev->hub->host->addr, 
                      tb->tr->open->dev->addr,
                      tb->tr->open->endp->addr >> 1);
            errno = -EIO;
            usb_errno = -USBE_BUS_BABBLE;
        }
        else if (td->ctstat.bits.stalled)    /* stalled? */
        {
            errno = -EREMOTEIO;
            usb_errno = -USBE_ENDP_STALLED;
        }
        else if (td->ctstat.bits.bitst_err)
        {
            gdlf_err ("Bit Stuff error, td 0x%p, tb 0x%p, "
                      "tr 0x%p, device 0x%p (%d %d), "
                      "endpoint %d\n",
                      td, tb, tb->tr, tb->tr->open->dev,
                      tb->tr->open->dev->hub->host->addr, 
                      tb->tr->open->dev->addr,
                      tb->tr->open->endp->addr >> 1);
            errno = -EIO;
            usb_errno = -USBE_BUS_BITSTUFF;
        }
        else
        {
            errno = 0;
            usb_errno = 0;  /* shut up gcc */
        }
        
        if (errno != 0)
            move_to_error (tb, usb_errno, errno);
    }
    else if (td_is_done_q (td))
    {
            /* 
             *  This may be a warning set on this packet, or that
             * the full transaction is finished. 
             *
             *  This is easy. If the TD is the last one, the
             * transaction is over, otherwise, it's a warning
             * packet. If there's a warning packet at the end of the
             * transaction, it is discarded in favour of the done
             * warning. However if pkt warnings are activated and done
             * warnings aren't, the packet warning is called.  
             */

        tb = get_iso_tb_by_td (&hci->iso, td); /* locate the TB */
#ifdef GDLF_ASSERTIONS
        if (tb == NULL || tb->tr == NULL)
            gdlf_err ("tb 0x%p tb->tr 0x%p\n",
                      tb, tb? tb->tr : NULL);
        else
#endif
        if (tb->tr->status == e_usb_tr_status_executing)
        {
            ts_usb_tr *tr;
            
            tr = tb->tr;        /* get the transaction */

            if (td == &tb->td[tb->tds]) /* TD is last one? */
                move_to_done (tb);      /* transaction's done */

            if ((td == &tb->td[tb->tds] /* TD is last one? */
                 && !tr->done_warn)      /* & there's no pkt warning */
                || td != &tb->td[tb->tds])  /* or TD isn't last */
            {
                    /* this is a warning packet */
                update_statistics (tr);
                if (tr->warn.wq != NULL)    /* warning set? */
                    if (tr->method)
                        (*tr->warn.callback) (tr, USB_CB_PKTS);
                    else
                        wake_up (tr->warn.wq);
            }
        }
    }

        /*
         *  The TD still may be in INSERTED or QUEUED state,
         * states on which we'll do _nothing_ 
         */

    return 0;
}


    /* Search the frame list from the specified point for transfers
     * which are no longer executing or errors.
     *
     *  Note: a 'frame chain' is all the chain which is generated by
     * following the link from the frame list pointer until the end
     * (until the Bulk QH points back to the Control QH).
     * 
     *  We're going to check the frame chain. We start on isochronous
     * transfers, then interrupt transfers, then control and then
     * bulk.
     *
     *  Thus, we have two specific parts: iso and interrupts and
     * control and bulk. This helps our behaviour, as when
     * exploring, we may not wish to re-explore the control and bulk
     * queues [as all the frame list entries generate frame chains
     * which finish there]. We can control how the function behaves
     * with the third parameter ['fast'].
     *
     * Parameters:  
     *
     *  ts_uhci_host*   Pointer to UHCI host where to search
     *
     *  unsigned        Frame number to start looking at. Once this
     *                  frame is visited fully (including all the
     *                  queues), the next frame to visit (previous
     *                  one) will be only examined skipping the
     *                  queues. 
     *
     *  unsigned        Flags given by the caller indicating what's
     *                  got to be searched and how to interpretate
     *                  it. In fact, it is the HC status, and we
     *                  interpretate it as:
     *
     *                  USBSTS_HCPE
     *
     *                      Search for the first active TD and
     *                      move to e_usb_tr_status_bug ERROR the
     *                      full TB. Return if no USBSTS_USBEI and no
     *                      USBSTS_USBINT.
     *                  
     *                  USBSTS_PCIBE
     *
     *                      Search for the first active TD and
     *                      move to e_usb_tr_status_memerr ERROR the
     *                      full TB. Return if no USBSTS_USBEI and no
     *                      USBSTS_USBINT.
     *                  
     *                  USBSTS_USBEI
     *
     *                      Error on USB transaction. When walking the
     *                      list, search for a failed transaction and
     *                      move it to the e_usb_tr_status_ERROR state
     *                      which requires it.
     *                  
     *                  USBSTS_USBINT
     *
     *                      Some transactions have finished. Search
     *                      for them while walking the list.
     *
     *                  IT IS ALWAYS ASSUMED ONE OF THESE IS ACTIVE!!!
     *
     *  int             UHCI_FL_FAST
     *
     *                     exit as soon as you find and repair the
     *                     cause of a PCIBE or HCPE error [searching
     *                     the full frame, but skipping it entirely
     *                     once found the problem].
     *
     *                  UHCI_FL_ALL
     *
     *                     process the full frame chain.
     *
     *                  UHCI_FL_INI
     *
     *                     process only the iso and interrupt part of
     *                     the frame chain. 
     *
     * Returns:
     *
     *  int             Number of TBs processed [aka: sent to DONE or
     *                  ERROR states]. 0 if none, < 0 on error.
     *
     *  I started coding this ~ July 24 9:35 CEST 1998; these are my
     * desperate remarks on it :):
     *
     *  11:55 - So, should I cook lunch?
     *
     *  14:35 - GEEE!!! This function is _*UGLY*_!!! 
     *
     *  15:45 - Huh ... It's beginning to look like black magic
     *
     *  18:01 - some rework ... looks clearer now, agfs!  
     *
     *  This function is a mess, it should be rewritten to use a
     * linear list.  
     *
     *  August 27 14:30 CEST 1998, moving to a linear list.
     */

void uhci_fl_search (ts_uhci_host* host, unsigned frnum, __u8 stat,
                     int mode) 
{
    void* ptr;
    ts_uhci_td *td;     /* td iterator */
    ts_uhci_qh *qh;     /* qh iterator */
    ts_uhci_flp *flp;   /* frame list pointer */
    char qcontext;      /* 0 - TD context, !0 - QH context */
    char runs;          /* times the search ran on control and bulk queues */
    
    gdlf_fntrace ("uhci_fl_search (0x%p, %u, 0x%x, %d)",
                  host, frnum, stat, mode);
    gdlf_aassert (host != NULL, return);
    gdlf_aassert (frnum < UHCI_FL_SIZE, return);
    
    flp = &host->fl[frnum];
    gdlf_aassert (flp->bits.t != 1, return);

    td = NULL;          /* shut up gcc? */
    qh = NULL;
    if ((qcontext = flp->bits.q))   /* qh context or td context? */
        qh = ptr2kvoid (flp->bits.ptr << 4);
    else
        td = ptr2kvoid (flp->bits.ptr << 4);
    
        /* 
         *  Ok, let's check what we've got at this first frame chain;
         * If there are isochronous TDs, let's advance until we
         * find TDs in EXECUTING, DONE or ERROR status or a QH. When
         * we reach a QH, we just exit this loop and let the flow
         * continue handling it.
         */

    while (!qcontext)
    {
        gdlf_debug ("Processing isochronous TD 0x%p\n", td);
        gdlf_aassert (td != NULL, return);
        gdlf_aassert (td->lp.bits.t != 1, return);
        
        ptr = ptr2kvoid (td->lp.bits.ptr << 4);

            /* The check to see if executing is done by cook_iso_td */

        if (cook_iso_td (td, host, stat, mode)
            && mode == UHCI_FL_FAST)
            return;     /* MIND THIS!!! */

            /* Now let's get next and see what is it */

        if (td->lp.bits.q)  /* qh context or td context? */
        {
            qh = ptr;
            break;
        }
        else
            td = ptr;
    }

#warning FIXME: add an assertion for the qh being an interrupt/ls_ctrl
            
        /*
         * Ok, now these are the interrupt, control and bulk queues, for
         * sure. As we check them every time we are called with an
         * argument different of UHCI_FL_INI, we must check if we have
         * to check.
         */

    if (mode == UHCI_FL_INI)    /* Check all queues? ... */
        return;                 /* ... nope */

        /*
         *  Now we are on QH context for sure; we may be in one of the
         * interrupt queues or at the LS Control queue. How to guess
         * it is done by checking the element pointer; if non
         * terminated and pointing to a QH, it is an interrupt list.
         *
         *  So, what we do at the very end is skip all the QHs with
         * terminated element and dig into the ones with .q bit
         * set. If we find one with the .q bit unset, we exit this
         * loop and continue (it means that qh is a TB, so that's
         * control/bulk transactions).
         *
         *  Note we don't follow the frame list, but our own composed
         * list, not to miss anyone. 
         */

    if (host->intr.top != NULL)         /* do we have to check interrupts? */
        qh = &host->intr.top->first;    /* yes, service them */
    else                                /* no, skip to control */
        qh = &host->ls_ctrl->first;

    while (qh != &host->ls_ctrl->first)
    {
        ts_uhci_qh *itr;    /* queue iterator*/
        
            /* Empty QHs have elt set to 0x3 */

        if (qh->elt.word == cpu_to_le32 (0x3))  /* empty one? check next */
        {
            qh = ptr2kvoid (qh->head.bits.ptr << 4);
            gdlf_aassert (qh != ptr2kvoid (qh->head.bits.ptr << 4), break);
            continue;
        }
        gdlf_aassert (qh->elt.bits.t != 1,      /* make it sure */
                      {
                          ts_uhci_qh_printk (qh);
                          qh = ptr2kvoid (qh->head.bits.ptr << 4);
                          continue;
                      });
        
            /* 
             * Ok, walk this queue 
             *
             *  First we process first element, and then we use the
             * sequence. It could be done more elegantly, sure, but
             * hey, I don't want to get all the glory myself and thus
             * is left as an exercise to the smart reader :)
             */

        itr = ptr2kvoid (qh->elt.bits.ptr << 4);
        gdlf_aassert (itr->elt.bits.q == 0, break);
        gdlf_aassert (((ts_uhci_tb*)itr)->tr != NULL, break);
        gdlf_aassert (((ts_uhci_tb*)itr)->tds != 0, break);
        gdlf_aassert (((ts_uhci_tb*)itr)->td != NULL, break);

        cook_queued_tb ((ts_uhci_tb*)itr, stat, mode);
        gdlf_aassert (itr != ptr2kvoid (itr->head.bits.ptr << 4), break);

        while (itr->head.bits.t != 1)
        {
            itr = ptr2kvoid (itr->head.bits.ptr << 4);
                /* make sure this is a TB */

            gdlf_aassert (itr->elt.bits.q == 0, break);
            gdlf_aassert (((ts_uhci_tb*)itr)->tr != NULL, break);
            gdlf_aassert (((ts_uhci_tb*)itr)->tds != 0, break);
            gdlf_aassert (((ts_uhci_tb*)itr)->td != NULL, break);

            cook_queued_tb ((ts_uhci_tb*)itr, stat, mode);
            gdlf_aassert (itr != ptr2kvoid (itr->head.bits.ptr << 4), break);
        }
        gdlf_assert (itr->head.word == cpu_to_le32 (0x3));
        gdlf_aassert (qh != ptr2kvoid (qh->head.bits.ptr << 4), break);
        qh = ptr2kvoid (qh->head.bits.ptr << 4);
    }

        /* Now we _SHOULD_ be at the ls_ctrl list */

    gdlf_aassert (qh == &host->ls_ctrl->first, return);

        /*
         *  When running the LS Control, Control and Bulk queues, we
         * have to be careful with one thing: the last entry in bulk
         * queue points back to the control queue, and we could be
         * looping forever. Thus, each time we pass trough the Control
         * queue, we inc the 'runs' variable. If it reaches two, it'll
         * mean we have finished.
         */

    runs = 0;
    qh = ptr2kvoid (qh->head.bits.ptr << 4);    /* get first after ls_ctrl */
    do
    {
            /* Empty QHs have elt to 0x3!! */

        if (qh->elt.word != cpu_to_le32 (0x3))  /* if not empty ... */
        {
                /* Make sure this is a TB */
            gdlf_aassert (qh->elt.bits.q == 0, break);
            gdlf_aassert (((ts_uhci_tb*)qh)->tr != NULL, break);
            gdlf_aassert (((ts_uhci_tb*)qh)->tds != 0, break);
            gdlf_aassert (((ts_uhci_tb*)qh)->td != NULL, break);

            if (cook_queued_tb ((ts_uhci_tb*)qh, stat, mode)
                && mode == UHCI_FL_FAST)
                return;
        }

            /* Check next is ok */

#ifdef UHCI_DO_BWR
        gdlf_aassert (qh->head.bits.t != 1, break);
#warning FIXME: bandwidth reclamation slows system down!!
#else
        if (qh->head.bits.t == 1)
        {
            gdlf_assert (qh->head.word == cpu_to_le32 (0x3));
            break;
        }
#endif
            /* next is ok, get it */

        gdlf_aassert (qh != ptr2kvoid (qh->head.bits.ptr << 4), break);
        qh = ptr2kvoid (qh->head.bits.ptr << 4);
        if (qh == &host->ctrl->first)
            runs++;
    } while (runs < 2);
    return;
}


    /* IRQ handler for the host controller
     *
     * Note that we have to ensure that none of the functions called
     * by this one do printk something, neither do anything nasty to
     * an IRQ handler.
     *
     *  We need the USB host controller address instead of the HCI
     * host controller to do the locking et al ... 
     *
     * Parameters:
     *
     *  int             IRQ which triggered it
     *
     *  ts_usb_host*    Private IRQ handler data, being the USB Host
     *                  which caused this IRQ.
     *
     *  struct pt_regs* Registers
     *
     * Global:
     *
     *  int usb_errno   On error, error code.
     */

#warning FIXME: move this to uhci.h
    /* not too low, polling devs may hit it when unplugging */
#define UHCI_MAX_ERRORS 100

void uhci_irqhandler (int irq, ts_usb_host* host, struct pt_regs* regs)
{
    unsigned frcnt;     /* frame counter */
    unsigned frbot;     /* frame bottom */
    __u8 stat;          /* current HC status */
    __u16 frnum;        /* Frame number currently executing */
    ts_uhci_host *hci;

    gdlf_fntrace ("uhci_irqhandler (0x%x, 0x%p, 0x%p)", 
                  irq, host, regs);
    gdlf_aassert (host != NULL, return);
    gdlf_aassert (host->hci != NULL, return);

    hci = UHCI (host->hci);
    gdlf_assert (irq == hci->irq);  /* paranoia? */    
    gdlf_aassert (hci->fl != NULL, return);

    atomic_inc (&hci->sem);
    if (atomic_read (&hci->sem) > 1)
    {
        gdlf_debug ("Missed IRQ, still busy!!\n");
        atomic_dec (&hci->sem);
        return;
    }

        /*
         *  First thing we do is to get what's the host status and
         * which frame is it executing.
         */
    
    frnum = inw (hci->iobase + USBFRNUM) & USBFRNUM_M;
    stat = inw (hci->iobase + USBSTS);
    if (stat == 0)  /* Nothing? ... */
    {
#warning DEBUG: remove me
        static int cnt = 0;
        if (cnt++ == 0)
            gdlf_debug ("host 0x%p (hci 0x%p), shared interrupt?\n",
                        host, hci);
        atomic_dec (&hci->sem);
        return;     /* ... ok, do nothing */        
    }
    outw (stat, hci->iobase + USBSTS);  /* ACK _everything_ (jic) */

    write_lock (&host->snode.lock); /* FIXME: huh? */

        /* 
         * Let's see what do we have 
         *
         *  Host controller halting is _bad_, as we will suffer
         * synchronization problems in isochronous endpoints, so we
         * have to solve it fast; however, got to solve first any
         * error which could have caused it. When we solve all the
         * errors, we can run the controller again.
         *
         *  Normally, any event we have (process error, PCI error,
         * USB Error Interrupt and USB interupt) will need a list
         * search for the offending/finished TD, so if the error is
         * the only onle which doesn't require a list search, we
         * process it separately. Later we restart the controller if
         * it had stopped.
         */


        /*
         *  Resume received, wake up you lazy!
         */
    
    if (stat & USBSTS_RR)
    {
        gdlf_info ("Host 0x%p (%d): resume received\n",
                   host, host->addr);
#warning FIXME: send an APM wake up event?
#warning FIXME: Why a global resume? no need to wake up everybody!
        uhci_global_resume (hci);
    }

        /* Other reasons for calling? */

    if (stat & 
        (USBSTS_HCPE | USBSTS_PCIBE 
         | USBSTS_USBEI | USBSTS_USBINT))
    {
            /*
             *  Huh ... consistency check, let's warn the search
             * function to expect the first active TD it finds to be
             * the cause of the error.
             */

        if (stat & USBSTS_HCPE)
        {
            hci->errors++;
            gdlf_err ("Host 0x%p (%d): process error (%d); severing heads\n",
                      host, host->addr, hci->errors);
            
#warning FIXME: make hci->errors a density in time
            if (hci->errors > UHCI_MAX_ERRORS)
            {
                gdlf_err ("Host controller 0x%p (%d) reached maximal "
                          "error water-mark, resetting\n", host, host->addr);
                uhci_host_reset (hci);
                usb_hub_host_rm (host);
            }
        }   

            /*
             *  Oh ohhhh ... this is _bad_. However, it seems a hw reset
             * is issued to the whole system. Ok I assume that when this
             * happen, we just retire the current active TD's TB, like
             * on HCPE.
             */
    
        if (stat & USBSTS_PCIBE)
        {
            hci->errors++;
            gdlf_err ("Host 0x%p (%d): system error (%d); severing heads\n",
                      host, host->addr, hci->errors);
#warning FIXME: be more verbose with the error (PIIX3/PCIDS register)
            
#warning FIXME: make hci->errors a density in time
            if (hci->errors > UHCI_MAX_ERRORS)
            {
                gdlf_err ("Host controller 0x%p (%d) reached maximal "
                          "error water-mark, resetting\n", host, host->addr);
                uhci_host_reset (hci);
                usb_hub_host_rm (host);
            }
#warning FIXME: when these happen too much, I should somehow _fully_ reset
        }
        
            /*
             *  Error on a USB transaction? ok, search for it, mark the
             * transaction as finished on ERROR state [put your error type
             * here] and let the next one execute.
             */

        if (stat & USBSTS_USBEI)
        {
            hci->errors++;
            gdlf_info ("Host 0x%p (%d): transaction error\n",
                       host, host->addr);
            
#warning FIXME: make hci->errors a density in time
            if (hci->errors > UHCI_MAX_ERRORS)
            {
                gdlf_err ("Host controller 0x%p (%d) reached maximal "
                          "error water-mark, resetting\n", host, host->addr);
                uhci_host_reset (hci);
                usb_hub_host_rm (host);
            }
        }
        

            /* 
             *  Transaction finished, search it, mark it as finished ok
             * and let next one execute.
             *
             *  Also, when short packet detect ... hmm. This is
             * dangerous; if short packet is last packet, then that's
             * _ok_, BUT if it is one of the middle packets, that's
             * _bad_, because it means that it's corresponding part of
             * the data buffer hasn't been filled entirely.
             *
             *  What we'll do is, at end of transaction, we'll walk
             * all the TDs from last to first, severing the 'holes'
             * left because of short packets [detect them with
             * td.actlen] and writing the final total data count,
             * letting the user/client deal with that.  
             */

        if (stat & USBSTS_USBINT)
        {
#warning DEBUG: uncomment me
#if 0
            gdlf_debug ("Host 0x%p (%d): transaction finished\n",
                        host, host->addr);
#endif
        }
        
            /*
             *  Once informed the pertinent autoritires, we can deal
             * with the interrupt. It may be caused by an error or a
             * normal completion.
             *
             *  If the error was one which caused the host to halt, we
             * just call usb_fl_search() on the current frame chain and
             * tell it to locate the error, move it [the TB] to ERROR
             * status and exit, so we can restart the controller
             * ASAP. We don't actualize 'frbot', so next time this
             * frame is checked too, as we only check for the errors
             * which halted the HC (PCIBE and HCPE).
             *
             *  In the other cases (other errors and normal completion
             * or warning packet), we just search the first frame
             * chain (the one at 'frnum') and if no cause is found on
             * the first one, the other ones until 'frbot' or a reason
             * for the interrupt is found [however, on the second and
             * following we just check the iso and interrupt queues,
             * not the control and bulk queues [it is not too probable
             * they'll change in the time it takes us to check every
             * frame.
             *
             *  FIXME: this has a side effect: on a lightly loaded
             * bus, the time between interrupts could be more than a
             * couple seconds [easily], so the frnum will pass frbot;
             * how to deal with this case? It is inneficient.
             *
             *  [NOTE we make the count modulo-1024, what's the frame
             * list size!].  
             */
        
        if (!(stat & (USBSTS_USBEI | USBSTS_USBINT)))
        {
#ifndef UHCI_DEBUG_TIMER_FL
            frbot = frnum - 1;
            uhci_fl_search (hci, frnum, stat, UHCI_FL_FAST);
#endif
        }
        else
        {    
#ifndef UHCI_DEBUG_TIMER_FL
            frbot = hci->frbot; /* get last one to check */
            hci->frbot = frnum; /* next time check only til here */
            uhci_fl_search (hci, frnum, stat, UHCI_FL_ALL);
#warning DEBUG: remove me, I will not scan all frames
#if 0
            for (frcnt = frnum - 1;  frcnt > frbot; 
                 frcnt++, frcnt %= UHCI_FL_SIZE)
                uhci_fl_search (hci, frnum, stat, UHCI_FL_INI);
#endif
#endif  /* !UHCI_DEBUG_TIMER_FL */
        }        

            /*
             *  This indicates if the host controller halted;
             * We must have solved any error cause when here, so just 
             * restart the controller. On too much errors, we must
             * reset the controller.
             */

#ifndef UHCI_SBYS
        if (stat & USBSTS_HCH)
        {
#warning FIXME: make this a density in time, and it should reset
            if (hci->errors > 10)
            {
                gdlf_err ("Host 0x%p (%d): too many errors, "
                          "resetting controller\n",
                          host, host->addr);
#warning FIXME: save state (ports), reset and reload state
            }
            else
            {
                gdlf_err ("Host 0x%p (%d): restarting controller\n",
                          host, host->addr);
                uhci_runstop (hci, 1);
            }
        }
#endif
    }

    write_unlock (&host->snode.lock); /* FIXME: huh? */
    atomic_dec (&hci->sem);
    return;
}


/* Log:
 *
 * $Log: irq-handler.c,v $
 * Revision 1.18  1998/11/14 17:45:44  inaky
 * Moved the include files to uusbd/include/uusbd, to ease integration
 * into the Linux kernel.
 *
 * Revision 1.17  1998/11/09 21:44:02  inaky
 * Disabled spurious debug messages.
 *
 * Revision 1.16  1998/10/22 00:11:20  inaky
 * Moved max of errors to 100, to leave time to the programs using
 * devices to notice.
 *
 * Revision 1.15  1998/10/14 21:53:36  inaky
 * Added _basic_ code for handling excesive errors by resetting the
 * controller. Theoretically it should reenumerate later, but I have to
 * code it :).
 *
 * Revision 1.14  1998/09/24 16:18:11  inaky
 * Terminate QHs on error as first step when moving_to_error, to evitate
 * HC jumping in. Fixed interrupt queues walking.
 *
 * Revision 1.13  1998/09/02 21:20:31  inaky
 * Moved uhci.h inclusion after gdlf. Added assertions on
 * uhci_irqhandler(),
 *
 * Revision 1.12  1998/08/30 14:00:13  inaky
 * Finished new list's linear topology.
 *
 * Revision 1.11  1998/08/29 00:11:03  inaky
 * Fixed endianess on qhs and tds, and option to disable bandwidth
 * reclamation.
 *
 * Revision 1.10  1998/08/27 22:22:52  inaky
 * Moved to a linear list, simpler and more elegant. Took out timeouts to
 * the USBD layer.
 *
 * Revision 1.9  1998/08/25 16:45:16  inaky
 * Removed spurios debug messages.
 *
 * Revision 1.8  1998/08/24 02:30:53  inaky
 * Did the little debugging needed to make the functions work as they're
 * expected to.
 *
 * Revision 1.7  1998/08/19 23:57:57  inaky
 * Change in the interaction between move_to_done() and
 * move_to_error(). Also added debug code. Got to test this a lot more.
 *
 * Revision 1.6  1998/08/03 13:55:32  inaky
 * Removed some debug info.
 *
 * Revision 1.5  1998/08/03 13:46:29  inaky
 * Added status assertion checks on move_to_*(), ability to call
 * move_to_error() from move_to_done(), primitive statistics calculation,
 * finished move_to_done(), list transversing code. Added a check for
 * correctly initialized host (host->fl != NULL), HC status
 * acknowledgement, and step-by-step execution support.
 *
 * Revision 1.4  1998/07/30 17:59:23  inaky
 * Added some stuff for debugging.
 *
 * Revision 1.3  1998/07/28 23:15:56  inaky
 * Added a comment of dislikeness.
 *
 * Moved the ISO td state checking functions to be valid for generic TDs,
 * and not only ISOs.
 *
 * Added functions to check TB's state. Finished the
 * move_to_(error|done)() functions and the TB processing functions,
 * cook_queued_tb() and cook_iso_tb(). Also fixed the IRQ handler and
 * frame list search function to use all this.
 *
 * Revision 1.2  1998/07/24 15:42:24  inaky
 * Removed the PAUSED state, it was quite useless ...
 *
 * Revision 1.1  1998/07/24 15:41:19  inaky
 * Handles IRQ stuff. Still unfinished.
 *
 */

