
/* USB Driver Stack v$Revision: 1.8 $
 *
 * (C) 1997 by Iaky Prez Gonzlez <inaky@peloncho.fis.ucm.es>
 *
 * Distributed under the terms of the GNU Public License Version 2
 * See files COPYRIGHT and COPYING at the root distribution directory
 *
 * $Id: iso-hash.c,v 1.8 1998/11/14 17:45:45 inaky Exp $
 *
 * Changelog at the bottom
 *
 *
 *     USB Universal Host Controller Interface
 *     ---------------------------------------
 *
 *  This file implements the Isochronous transfers hash table, used to
 * locate TBs from a given TD.
 */

#include <asm/spinlock.h>
#include <linux/types.h>
#include <linux/timer.h>

#include <uusbd/usbd.h>

#define GDLF_IDENT          "USB-UHCI"
#define GDLF_LOCAL          /* Define if want local settings */
#define GDLF_DEF_LOGLEVEL   9
#define GDLF_ASSERTIONS     /* want assertions (#undef if not) */
#undef GDLF_FNTRACE        /* want function traces */
#define GDLF_LOCATIONS      /* want locations in log messages */
#include <uusbd/gdlf.h>

#include "uhci.h"           /* this has to be include after GDLF */


    /* Squeeze the iso hash table
     *
     *  This will be called each period, look into the hash table, and
     * if empty in a fifty something percent, it'll shrink it.
     *
     * Parameters:  
     *
     *  ts_uhci_iso_hash*
     *
     *                  Pointer to isochronous hash table.
     *
     *  unsigned long   Last time it was called, in jiffies. Useful if
     *                  sometime we add an statistical algorithm. if
     *                  0, do a general cleanup.
     */

void iso_hash_cleanup (ts_uhci_iso_hash* hash, unsigned long last)
{
    ts_uhci_tb **itr, **top;

    gdlf_fntrace ("iso_hash_cleanup (0x%p, %lu)", hash, last);
    gdlf_aassert (hash != NULL, return);

    if (hash->hash == NULL) /* Unneeded? */
        return;

        /* Hmmm ... no spinlocks on UP */

    atomic_inc (&hash->usecnt);
    if (atomic_read (&hash->usecnt) > 1)
    {
        atomic_dec (&hash->usecnt);
        gdlf_info ("Isochronous hash 0x%p is in use, cleanup "
                   "skiped until next run\n", hash);
        return;
    }

        /* 
         *  We have a base-of-two optimal allocation constraint, check
         * if we ain't using optimally and resize if so. We put the
         * water low mark in using less than half the size of the hash
         * table. Less means having two free as a threshold. Whatever
         * we do, we have to make the desired size a power of two.
         */

    write_lock (&hash->lock);
    if (last == 0)
    {
        gdlf_assert (hash->used == 0);
        usb_kfree (hash->hash);
        hash->slots = 0;
        hash->hash = NULL;
    }
    else if (hash->used + 2 < hash->slots/2)
    {
            /* FIXME: GFP_ATOMIC? we're called from a BH... */

        ts_uhci_tb **dst;

        gdlf_debug ("Squeezing ISO hash table 0x%p from %d\n",
                    hash, hash->slots);
        
        dst = usb_kmalloc (hash->slots * sizeof(ts_uhci_tb*) / 2, GFP_ATOMIC);
        if (dst == NULL)
        {
            gdlf_err ("Squeeze of ISO hash table 0x%p failed\n", hash);
            atomic_dec (&hash->usecnt);
            return;
        }     
        
            /* Now clean the new array and write-compact the old one */
   
        memset (dst, 0, hash->slots * sizeof (ts_uhci_tb*) / 2);
        for (itr = hash->hash, hash->hash = dst, top = itr + hash->slots; 
             itr < top; itr++)
            if (*itr)               /* used? then copy */
                *dst++ = *itr;
        hash->slots /= 2;
        usb_kfree (hash->hash);     /* free the old one */
        hash->hash = dst;           /* get the new one */
    }
    write_unlock (&hash->lock);
    atomic_dec (&hash->usecnt);
    return;
}

    /* Add an isochronous TB to the iso hash table
     *
     *  This will resize the array if needed, so you don't need to
     * care about anything.
     *
     * Parameters:  
     *
     *  ts_uhci_iso_hash*
     *
     *                  Pointer to isochronous hash table.
     *
     *  ts_uhci_tb*     Pointer to the TB to add.
     *
     * Returns:
     *
     *  int             0 if ok, < 0 on error. It is not fine to
     *                  return the position in which inserted, because
     *                  when resizing, that changes.
     *
     */

int iso_hash_add (ts_uhci_iso_hash* hash, ts_uhci_tb* tb)
{
    ts_uhci_tb **itr, **top;
    size_t new;
    
    gdlf_fntrace ("iso_hash_add (0x%p)", tb);
    gdlf_aassert (hash != NULL, return -EINVAL);
    gdlf_aassert (tb != NULL, return -EINVAL);

    write_lock (&hash->lock);

    for (itr = hash->hash, top = itr + hash->slots; itr < top; itr++)
    {
        if (*itr == NULL)    /* free place? ... */
        {
            *itr = tb;
            hash->used++;
            write_unlock (&hash->lock);    
            return 0;
        }
    }

        /* 
         * If here, we ran all the array and found no free places, so
         * we gotta make more room ... we double the size.
         */
    
    new = hash->slots == 0? 2 : 2 * hash->slots;
    itr = usb_realloc (hash->hash, hash->slots * sizeof (ts_uhci_tb*), 
                       new * sizeof (ts_uhci_tb*), GFP_KERNEL);
    if (itr != NULL)    /* Uh? ... */
    {
            /* 
             * Clean new storage and set the hash entry 
             *
             *  Note: if we _reallocated_ from zero size, we got to
             * set to zero the full array, not the new, upper
             * half. That's what (hash->slots? 2 : 1) is for.
             */

        hash->hash = itr;
        memset (itr + hash->slots, 0,
                new / (hash->slots? 2 : 1) * sizeof (ts_uhci_tb*));
        itr += hash->slots;         /* go to the first, new, free one */
        *itr = tb;                  /* set the new tb */
        hash->slots = new;          /* set the new size */
        hash->used++;
        write_unlock (&hash->lock); /* unlock */
        return 0;                   /* go away */
    }
    else
    {
        write_unlock (&hash->lock);
        return -ENOMEM;
    }
}


    /* Delete an ISO TB from the hash table
     *
     * Parameters:  
     *
     *  ts_uhci_iso_hash*
     *
     *                  Pointer to isochronous hash table.
     *
     *  ts_uhci_tb*     Pointer to the TB to delete.
     */

void iso_hash_del (ts_uhci_iso_hash* hash, ts_uhci_tb* tb)
{
    ts_uhci_tb  **itr, **top;
    
    gdlf_fntrace ("iso_hash_del (0x%p, 0x%p)", hash, tb);
    gdlf_aassert (hash != NULL, return);
    gdlf_aassert (tb != NULL, return);
    
    write_lock (&hash->lock);
    for (itr = hash->hash, top = itr + hash->slots;
         itr < top; itr++)
        if (*itr == tb)
        {
            hash->used--;
            write_unlock (&hash->lock);
            return;
        }
    write_unlock (&hash->lock);
    gdlf_debug ("TB 0x%p not found on iso hash 0x%p\n", tb, hash);
    return;
}


    /* Search an isochronous TB by TD
     *
     *  Given one of it's TDs, search an isochronous's transaction
     * TB. This is accomplished using the ISO hash table, which keeps
     * a pointer to the TB and the bottom and top addresses of it's TD
     * array. By seeing in which interval of TD addresses is the
     * searched TD, we locate the TB.
     *  
     * Parameters:  
     *
     *  ts_uhci_iso_hash
     *  ts_uhci_td*     Pointer to the TD to locate.
     *
     * Returns:
     *
     *  ts_uhci_tb*     Pointer to the TB the TD belongs to or, if
     *                  not found (hmmm, hash incoherency), NULL.
     */

ts_uhci_tb* get_iso_tb_by_td (ts_uhci_iso_hash* hash, ts_uhci_td* td)
{
    ts_uhci_tb **itr, **top;

    gdlf_fntrace ("get_iso_tb_by_td (0%p)", td);
    gdlf_aassert (hash != NULL, return NULL);
    gdlf_aassert (td != NULL, return NULL);
    
        /*
         *  Walk the array (if any allocated) and if you find a
         * non-null entry (*itr), then if the the td pointer (td)
         * minus the base of the td array for that tb (tb->td) is
         * lower than the tb size (tb->tds), then thats is our kid.
         */

    read_lock (&hash->lock);
    for (itr = hash->hash, top = itr + hash->slots; 
         itr < top; itr++)
        if (*itr && (td - (*itr)->td) < (*itr)->tds)
        {
            read_unlock (&hash->lock);
            return *itr;
        }
    read_unlock (&hash->lock);
    gdlf_debug ("iso td 0x%p not found in iso hash 0x%p\n", hash, td);
    return NULL;
}


/* Log:
 *
 * $Log: iso-hash.c,v $
 * Revision 1.8  1998/11/14 17:45:45  inaky
 * Moved the include files to uusbd/include/uusbd, to ease integration
 * into the Linux kernel.
 *
 * Revision 1.7  1998/09/24 16:16:28  inaky
 * Made the cleanup function do a full cleanup if the 'last' argument is
 * zero.
 *
 * Revision 1.6  1998/09/14 20:35:45  inaky
 * Fixed reallocator to take into account pointers have a size.
 *
 * Revision 1.5  1998/09/02 21:22:04  inaky
 * Moved uhci.h inclusion after gdlf.
 *
 * Revision 1.4  1998/08/30 14:00:29  inaky
 * Added semaphores.
 *
 * Revision 1.3  1998/08/26 18:07:56  inaky
 * Reworked to make the hash tables hci dependant and smaller [against
 * some speed]. Optimized to use always full memory buffers of power of
 * two sizes.
 *
 * Revision 1.2  1998/08/19 23:59:12  inaky
 * Suppossedly ptrdiff_t are units ... so no need to divide by unit size.
 *
 * Revision 1.1  1998/07/28 23:10:18  inaky
 * Contains the hash table code for the ISO TBs location from td
 * pointers.
 *
 *
 */

