
/* USB Driver Stack v$Revision: 1.15 $ 
 *
 * (C) 1997 by Iaky Prez Gonzlez <inaky@peloncho.fis.ucm.es>
 *
 * Distributed under the terms of the GNU Public License Version 2
 * See files COPYRIGHT and COPYING at the root distribution directory
 *
 * $Id: operations.c,v 1.15 1998/11/14 17:45:46 inaky Exp $
 *
 * Changelog at the bottom 
 *
 *
 *     USB Universal Host Controller Interface
 *     ---------------------------------------
 *  
 *  This file contains all basic operations on the UHCI host
 * controllers. Some of them are inlined in the UHCI internal header
 * file uhci.h.
 */

#include <uusbd/usb.h>

#define GDLF_IDENT          "USB-UHCI"
#define GDLF_LOCAL          /* Define if want local settings */
#define GDLF_DEF_LOGLEVEL   9
#define GDLF_ASSERTIONS     /* want assertions (#undef if not) */
#undef GDLF_FNTRACE        /* want function traces */
#define GDLF_LOCATIONS      /* want locations in log messages */
#include <uusbd/gdlf.h>

#include "uhci.h"           /* this has to be include after GDLF */




    /* Print the hex registers of a UHCI compliant controller
     *
     * Parameters: 
     *
     *  unsignedt int   Number of the controller in the usb_host array
     *
     * Returns:
     *
     *  int             0 if ok, errno code < 0 on error.
     *
     * Global:
     *
     *  int usb_errno   On error, error code.  
     */

#ifdef UHCI_DEBUG
int uhci_dumpreg (ts_uhci_host* hci)
{
    unsigned iobase;
    
    gdlf_fntrace ("uhci_dumpreg (0x%p)", hci);

    read_lock (&hci->snode.lock);
    iobase = hci->iobase;
    gdlf_info ("Register dump of controller 0x%p\n"
               "  USBCMD       0x%04x\n"
               "  USBSTS       0x%04x\n" 
               "  USBINTR      0x%04x\n" 
               "  USBFRNUM     0x%04x\n"
               "  USBFLBASEADD 0x%08lx\n" 
               "  USBSOF       0x%02x\n"
               "  USBPORTSC1   0x%04x\n"
               "  USBPORTSC2   0x%04x\n",
               hci,
               (unsigned)inw (iobase + USBCMD),
               (unsigned)inw (iobase + USBSTS),
               (unsigned)inw (iobase + USBINTR),
               (unsigned)inw (iobase + USBFRNUM),
               (unsigned long) inl (iobase + USBFLBASEADD),
               (unsigned)inb (iobase + USBSOF),
               (unsigned)inw (iobase + USBPORTSC1),
               (unsigned)inw (iobase + USBPORTSC2)
        );
    read_unlock (&hci->snode.lock);
    return 0;
}   
#endif


    /* Reset an UHCI compliant controller. To reset the controller, it
     * must be written a 1 to the HCReset bit and then wait until
     * reading a 0 from this bit (when the resetting is finished).
     *
     * Parameters:  
     *
     *  ts_uhci_host*   Pointer to the host controller structure of
     *                  the host.
     *
     * Returns:
     *
     *  int             0 if ok, errno code < 0 on error.
     *
     * Global:
     *
     *  int usb_errno   On error, error code.  
     */

int uhci_host_reset (ts_uhci_host* hci)
{
    int cnt;
    unsigned iobase;

    gdlf_fntrace ("uhci_host_reset (0x%p)", hci);
    gdlf_aassert (hci != NULL, return -EINVAL);

    write_lock (&hci->snode.lock);
    iobase = hci->iobase;
    usb_bits16 (iobase + USBCMD, 0xffff, USBCMD_HCRESET);

    cnt = 100;
    usb_waitms (15);    /* 10 ms gives some problems */
    while ((inw (iobase + USBCMD) & USBCMD_HCRESET) && --cnt)
    {
        if (cnt % 10 == 0)
            gdlf_debug ("Waited %d times for UHCI "
                        "reset to finish\n", cnt);
        usb_waitms (1);
    }
    
    write_unlock (&hci->snode.lock);

    if (cnt == 0)   /* Count reached zero ... too bad, we are off */
    {        
        gdlf_err ("Reset failed on controller 0x%p\n", hci);
        uhci_errno = usb_errno = -USBE_UHCI_RESET_FAILED; 
        return -EIO;
    }

    return 0;
}


    /* 
     * This function cannot be inlined, as it is the one used for the
     * ts_usb_host structure 
     */

int uhci_hci_reset (ts_usb_host* host)
{
    int errcode;
    
    gdlf_fntrace ("uhci_hci_host_reset (0x%p)", host);

    errcode = uhci_host_reset (UHCI (host->hci));
    host->sof = uhci_get_sof (UHCI (host->hci));
    return errcode;
}


    /* 
     * This function cannot be inlined, as it is the one used for the
     * ts_usb_host structure 
     */

int uhci_hci_busreset (ts_usb_host* host)
{
        /* we don't want the HC to be fully resetted, just the bus */

    return uhci_host_reset (UHCI (host->hci));
}


    /* 
     * This function cannot be inlined, as it is the one used fot the
     * ts_usb_host structure.
     */ 

int uhci_hci_setsof (ts_usb_host* host, unsigned sof)
{
    return uhci_set_sof (UHCI (host->hci), sof);
}



    /* Set the maximum packet size that may be used for band-width
     * reclamation for full-speed transfers.
     *
     * Parameters:  
     *
     *  ts_uhci_host*   Pointer to the host controller structure of
     *                  the host.
     *
     *  int             0 - 32 bytes, 1 - 64 bytes, -1 - Read
     *
     * Returns:
     *
     *  int             The state of the bit before changing it or the
     *                  bit itself if read requested (-1), or < 0
     *                  errno code on error. 
     *
     * Global:
     *
     *  int usb_errno   On error, error code.  
     */

int uhci_maxp (ts_uhci_host* hci, int maxsize)
{
    int retval;
    
    gdlf_fntrace ("uhci_maxp (0x%p)", hci);
    
    gdlf_apassert (hci != NULL,
                   { 
                       uhci_errno = usb_errno = -USBE_HCI_NULL; 
                       return -EINVAL; 
                   },
                   "NULL pointer to UHCI host controller structure\n");    

    write_lock (&host->snode.lock);
    retval = (usb_bits16 (hci->iobase + USBCMD, maxsize << 7,
                          maxsize == -1? 0 : USBCMD_MAXP) 
              & USBCMD_MAXP) >> 7;
    write_unlock (&host->snode.lock);
    return retval;
}


    /* Halts (0) or runs (1) the USB controller (checking for it).
     *
     * Parameters:
     *
     *  ts_uhci_host*   Pointer to the host controller structure of
     *                  the host.
     *
     *  int             -1  get actual status USBSTS_HCH.
     *                  0   halt the controller.
     *                  1   start the controller.
     *
     * Returns:
     *
     *  int             Status of the controller (previous status if
     *                  modified). 0 stopped, 1 running; on error < 0
     *                  errno code.  
     */

int uhci_runstop (ts_uhci_host* hci, int action)
{
    int      val;   /* actual controller's state */
    int      cnt;   /* down counter */
    unsigned iobase;    

    gdlf_fntrace ("uhci_runstop (0x%p, %d)", hci, action);

    gdlf_aassert (hci != NULL, return -EINVAL);
    
    write_lock (&host->snode.lock);
    iobase = hci->iobase;
    
        /* Get actual status */

    val = (inw (iobase + USBSTS) & USBSTS_HCH)? 0 : 1;
    if (val == action)  /* If we're trying to let it in the same state ... */
    {
        write_unlock (&host->snode.lock);
        return val;     /* ... do nothing */
    }
    
    switch (action)     /* What do we want to do? */
    {
      case -1:          /* get value */
        break;

            /* 
             * Stop the controller
             *
             * Tell the controller to stop, wait until it is stopped,
             * if too much waiting, error.
             */

      case 0:

        cnt = 100;
        usb_bits16 (iobase + USBCMD, 0, USBCMD_RS);
        while (!(inw (iobase + USBSTS) & USBSTS_HCH) && --cnt)
        {
            if (cnt % 10 == 0)
                gdlf_debug ("Waited %d times for "
                            "controller 0x%p to stop\n", 
                            cnt, hci);
            usb_waitms (1);
        }
        if (cnt == 0)   /* Too much waiting? ... */
        {
            gdlf_err ("Couldn't halt UHCI controller 0x%p\n", hci);
            uhci_errno = usb_errno = -USBE_UHCI_COULDNT_HALT;
            val = -EIO;
        }
        break;
        
            /*
             *  Run the controller
             * 
             * Tell the controller to run, wait until it is running,
             * if too much waiting, error.
             */

      case 1:

#ifdef UHCI_SBYS             /* Activate step by step? */
        cnt = 100;
        usb_bits16 (iobase + USBCMD, 0, USBCMD_RS);
        while (!(inw (iobase + USBSTS) & USBSTS_HCH) && --cnt)
        {
            if (cnt % 10 == 0)
                gdlf_debug ("Waited %d times for "
                            "controller 0x%p to stop\n", 
                            cnt, hci);
            usb_waitms (1);
        }
        if (cnt == 0)   /* Too much waiting? ... */
        {
            gdlf_err ("Couldn't halt UHCI controller 0x%p\n", hci);
            uhci_errno = usb_errno = -USBE_UHCI_COULDNT_HALT;
            val = -EIO;
        }
        break;
        usb_bits16 (iobase + USBCMD, USBCMD_SWDBG, USBCMD_SWDBG);
#endif
        cnt = 100;
        usb_bits16 (iobase + USBCMD, 0, USBCMD_SWDBG);
        usb_bits16 (iobase + USBCMD, USBCMD_RS, USBCMD_RS);
        while ((inw (iobase + USBSTS) & USBSTS_HCH) && --cnt)
        {
            if (cnt % 10 == 0)
                gdlf_debug ("Waited %d times for "
                            "controller 0x%p to start\n", 
                            cnt, hci);
            usb_waitms (1);
        }
        if (cnt == 0)   /* Too much waiting? ... */
        {
            gdlf_err ("Couldn't start UHCI controller 0x%p\n", hci);
            uhci_errno = usb_errno = -USBE_UHCI_COULDNT_START;
            val = -EIO;
        }
        break;

      default:          /* Bad command */
        gdlf_err ("Bad action command %u passed to "
                  "uhci_runstop()\n", action);
        uhci_errno = usb_errno = -USBE_BUG;
        val = -EINVAL;
    }
    write_unlock (&host->snode.lock);
    return val;
}


    /* Get and set the frame number from the UHCI controller. For
     * setting the frame number we've got to stop it first, and then
     * restart it.
     *
     * Parameters:
     *
     *  ts_uhci_host*   Pointer to the host controller structure of
     *                  the host.
     *
     *  unsigned int    The frame number to set (for usb_uhci_set_frnum()),
     *                  which ranges from 0 to USB_UHCI_FL_SIZE.
     * Returns:     
     *
     *  __u16           [get_frnum()] The actual frame number.
     *
     *  int             0 on success; on error error, < 0 errno code.
     *
     * Global:
     *
     *  int usb_errno   On error, error code.  
     */

__u16 uhci_get_frnum (ts_uhci_host* hci)
{
    __u16 retval;
    
    gdlf_fntrace ("uhci_get_frnum (0x%p)", hci);
    gdlf_aassert (hci != NULL, return -EINVAL);

    read_lock (&host->snode.lock);
    retval = inw (hci->iobase + USBFRNUM) & USBFRNUM_M;
    read_unlock (&host->snode.lock);
    return retval;
}

int uhci_set_frnum (ts_uhci_host* hci, unsigned int frame)
{
    gdlf_fntrace ("uhci_set_frnum (0x%p, %u)", hci, frame);
    gdlf_aassert (hci != NULL, return -EINVAL);
    
    if (frame >= UHCI_FL_SIZE)
    {
        gdlf_err ("Frame number %u out of range; cannot set\n", frame);
        return -EINVAL;
    }

        /* Set the frame number */

    write_lock (&host->snode.lock);
    usb_bits16 (hci->iobase+USBFRNUM, frame, USBFRNUM_M);
    write_unlock (&host->snode.lock);

    return 0;
}


/* Log:
 *
 * $Log: operations.c,v $
 * Revision 1.15  1998/11/14 17:45:46  inaky
 * Moved the include files to uusbd/include/uusbd, to ease integration
 * into the Linux kernel.
 *
 * Revision 1.14  1998/09/24 16:19:21  inaky
 * Added a wait after asserting reset before checking if it is done.
 *
 * Revision 1.13  1998/09/02 21:22:17  inaky
 * Moved uhci.h inclusion after gdlf.
 *
 * Revision 1.12  1998/08/25 16:45:43  inaky
 * Removed unneeded FIXMEs.
 *
 * Revision 1.11  1998/08/25 16:18:59  inaky
 * Fixed some reads from ports, to have less overhead and severed some
 * assertions.
 *
 * Revision 1.10  1998/08/25 04:08:40  inaky
 * Removed debug register dump.
 *
 * Revision 1.9  1998/08/24 02:31:13  inaky
 * Did the little debugging needed to make the functions work as they're
 * expected to.
 *
 * Revision 1.8  1998/08/19 23:59:37  inaky
 * Moved casts to use the UHCI macro.
 *
 * Revision 1.7  1998/08/03 13:49:28  inaky
 * Fixed code which does recycling to stop/reset host controller, moved
 * the bus reset function to just reset the bus, not the bus and
 * host. Added support for step-by-step execution.
 *
 * Revision 1.6  1998/07/28 23:16:11  inaky
 * Moved the frame list size to be fixed.
 *
 * Revision 1.5  1998/07/23 14:43:37  inaky
 * Moved ../usb-uhci.h to uhci.h.
 *
 * Renamed the uhc_hci_host_reset function to uhci_hci_reset.
 *
 * Revision 1.4  1998/07/14 14:58:29  inaky
 * Fixed locking acknowledges of the whole uhci subsystem.
 *
 * Revision 1.3  1998/07/12 22:59:03  inaky
 * Fixed the code to do ms waits and ack locks.
 *
 * Revision 1.2  1998/05/01 19:21:07  inaky
 * Added the HCI setsof function (uhci_hci_setsof()) to be called from a
 * pointer on the host structure.
 *
 * Revision 1.1  1998/05/01 18:31:27  inaky
 * Moved everything to use pci_dev according to new kernel PCI API. Also
 * split all the code in several files, as old uhci.c was a complete mess.
 *
 */
