
/* USB Driver Stack v$Revision: 1.20 $
 *
 * (C) 1997 by Iaky Prez Gonzlez <inaky@peloncho.fis.ucm.es>
 *
 * Distributed under the terms of the GNU Public License Version 2
 * See files COPYRIGHT and COPYING at the root distribution directory
 *
 * $Id: scheduler.c,v 1.20 1998/11/14 17:45:47 inaky Exp $
 *
 * Changelog at the bottom
 *
 *
 *     USB Universal Host Controller Interface
 *     ---------------------------------------
 *
 *  This file contains the core logic for transfer schedulement [is
 * this word correct?] and execution ... it is the guy who really
 * talks to USB on UHCI hosts. All the handling since an irq breaks in
 * is done at irq-handler.c
 *
 *  Note that list heads are distinguished because (XXXXXXXX means
 * non-zero):  
 *  
 * - Empty list heads: qh->elt.word == 0x3
 *
 * - Non empty list head: qh->elt.word = 0xXXXXXXX{2,6}
 *
 * - Inserted TBs (waiting to be activated for execution):
 *   tb->qh.elt.word = 0xXXXXXXX1
 * 
 * - Finished TBs: tb->qh.elt.word = 0x0000000{1,5}
 *
 * - Executing TBs: tb->qh.elt.word = 0xXXXXXXX{0,4}
 *                  pointed TD is not on error (active set and no
 *                  errors set)
 *
 * - Executing TBs: tb->qh.elt.word = 0xXXXXXXX{0,4}
 *                  pointed TD is on error (active unset and
 *                  errors set)
 *
 * - Invalid head pointers in QHs are 0x3
 */

#warning FIXME: resynch support? USB Spec 1.0 8.4.4/ACK, 8.6

#include <uusbd/usbd.h>

#undef  GDLF_UHCI_SCHEDULER_C
#define GDLF_IDENT          "USB-UHCI"
#define GDLF_LOCAL          /* Define if want local settings */
#define GDLF_DEF_LOGLEVEL   9
#define GDLF_ASSERTIONS     /* want assertions (#undef if not) */
#undef GDLF_FNTRACE        /* want function traces */
#define GDLF_LOCATIONS      /* want locations in log messages */
#include <uusbd/gdlf.h>

#include "uhci.h"           /* this has to be include after GDLF */


    /* 
     *  qh_list_*() operate on the head pointer and use as list head an
     * element, the tb_list's QH.
     *
     *  OTOH, qh_ilist_*() use the tb_list's head element pointer as
     * head ... mind it's different :) 
     */


    /* Append a QH to a QH list
     *
     *  Will hold a write lock on the list. What this function do is
     * to append an element to a list of QHs. Thus, if the list is
     * empty, it is hooked to the element pointer of the head QH. If
     * the list isn't empty, it is appended at the head pointer of the
     * last QH in the list.
     *
     * Parameters:  
     *
     *  ts_uhci_tb_list*    
     *
     *                  Pointer to the QH list structure.
     *
     *  ts_uhci_qh*     Pointer to node to add to the QH list.
     *
     * Returns:
     *
     *  int             0 if ok, errno code < 0 on error.
     */

void qh_list_append (ts_uhci_tb_list* list, ts_uhci_qh* node)
{
    ts_uhci_qh* tmp;
    gdlf_fntrace ("qh_list_append (0x%p, 0x%p)", list, node);

    gdlf_aassert (list != NULL, return);
    gdlf_aassert (node != NULL, return);
    gdlf_aassert (list->last != NULL, return);
    
        /*
         * ((& 0xfffffff0) | 0x2) makes qtselect = 1 and terminate = 0
         */                                  

    write_lock (&list->lock);
    node->head.word = list->last->head.word;
    tmp = list->last;
    list->last = node;
    tmp->head.word = 
        cpu_to_le32 ((virt_to_bus (node) & 0xfffffff0) | 0x2);
    write_unlock (&list->lock);
    return;
}

void qh_ilist_append (ts_uhci_tb_list* list, ts_uhci_qh* node)
{
    ts_uhci_qh* tmp;
    
    gdlf_fntrace ("qh_ilist_append (0x%p, 0x%p)", list, node);
    gdlf_aassert (list != NULL, return);
    gdlf_aassert (node != NULL, return);

        /* 
         * prepare the node, so as soon as we put it, the HC sees a
         * nice one.
         *
         *  Note the difference. These ilists are terminated (end is
         * 0x1), normal lists aren't.
         */

    write_lock (&list->lock);
    node->head.word = cpu_to_le32 (0x3);
    tmp = list->last;
    list->last = node;
    if (tmp)                                /* non-empty list */
        tmp->head.word =                    /* append on head of last */
            cpu_to_le32 ((virt_to_bus (node)
                          & 0xfffffff0)     /* terminate = 0 */
                         | 0x2);            /* qtselect = 1 */
    else                                    /* empty list */
        list->first.elt.word =              /* append on head's element */
            cpu_to_le32 ((virt_to_bus (node)
                          & 0xfffffff0)     /* terminate = 0 */
                         | 0x2);            /* qtselect = 1 */
    write_unlock (&list->lock);
    return;
}


    /* Extract previous element to a node in the QH list
     *
     *  Won't hold nor acknowledge locks! You must call it with the
     * list locked.
     *
     * Parameters:  
     *
     *  ts_uhci_tb_list*    
     *
     *                  Pointer to the QH list structure.
     *
     *  ts_uhci_qh*     Pointer to node to search.
     *
     * Returns:
     *
     *  ts_uhci_qh*     Pointer to the previous element. If NULL, the
     *                  node hasn't been found in the list. If the
     *                  node is the first one, it'll return a pointer
     *                  to the list head.
     */

ts_uhci_qh* qh_list_prev (ts_uhci_tb_list* list, ts_uhci_qh* node)
{
    ts_uhci_qh* itr;
    __u32       bus;    /* xlated physical address */
    
    gdlf_fntrace ("qh_list_prev (0x%p, 0x%p)", list, node);    
    gdlf_aassert (list != NULL, return NULL);
    gdlf_aassert (node != NULL, return NULL);
    gdlf_aassert (list->last != NULL, return NULL);

        /*
         *  The list ends when a new one starts. A new one starts when
         * itr is a head of list, and that's indicated because it's
         * elt.word is 0x3
         */

    bus = kvoid2ptr (node) >> 4;
    itr = &list->first;
    do
    {
        if (itr->head.bits.ptr == bus)      /* if next one is 'node' ... */
            return itr;                     /* ... return this one */
        else                                /* it is not, get next */
            itr = ptr2kvoid (itr->head.bits.ptr << 4);
    } while (itr->elt.word != cpu_to_le32 (0x3)
#ifdef UHCI_DO_BWR
             && gdlf_c1assert (itr->head.bits.t != 1)
#else
             && itr->head.word != cpu_to_le32 (0x3)
#endif
        );

    gdlf_debug ("node 0x%p not found on list 0x%p!\n", node, list);
    return NULL;
}

ts_uhci_qh* qh_ilist_prev (ts_uhci_tb_list* list, ts_uhci_qh* node)
{
    ts_uhci_qh* itr;
    __u32       bus;    /* xlated physical address */
    
    gdlf_fntrace ("qh_ilist_prev (0x%p, 0x%p)", list, node);    
    gdlf_aassert (list != NULL, return NULL);
    gdlf_aassert (node != NULL, return NULL);

        /*
         *  The addresses in the QH are stored in little endian, and in
         * bus "syntax", so we have to xlate them. The <<>>4 is caused
         * because the pointers are only the upper 28 bits.
         */

        /* Ok, let's see if the list is empty */

    if (list->last == NULL)     /* empty list? */
    {
        gdlf_assert (list->first.elt.word == cpu_to_le32 (0x1));
        return NULL;
    }
    else
    {
        gdlf_aassert (list->first.elt.bits.t != 1, return NULL);
        gdlf_aassert (list->first.elt.bits.ptr != 0, return NULL);
    }
        
        /* Non-empty list, get first element */

    bus = kvoid2ptr (node) >> 4;
    itr = ptr2kvoid (list->first.elt.bits.ptr << 4);

    if (itr == node)        /* NODE is the first one? ... */
        return node;        /* ... ok, return it */

    while (itr->head.word != cpu_to_le32 (0x3)) /* while the QH isn't finished */
        if (itr->head.bits.ptr == bus)          /* if next one is 'node' ... */
            return itr;                         /* ... return this one */
        else                                    /* it is not, get next */
            itr = ptr2kvoid (itr->head.bits.ptr << 4);
    gdlf_debug ("node 0x%p not found on list 0x%p!\n", node, list);
    return NULL;
}


    /* Remove a QH from a QH list
     *
     *  Note it is _VERY_ important we make atomic the removal
     * operation (the copy).
     *
     *  Holds a write lock on the list.
     *
     * Parameters:  
     *
     *  ts_uhci_tb_list*    
     *
     *                  Pointer to the QH list structure.
     *
     *  ts_uhci_qh*     Pointer to node to remove.
     *
     * Returns:
     *
     *  int             0 if ok, errno code < 0 on error.
     */

void qh_list_delete (ts_uhci_tb_list* list, ts_uhci_qh* node)
{
    ts_uhci_qh* prev;

    gdlf_fntrace ("qh_list_delete (0x%p, 0x%p)", list, node);
    gdlf_aassert (list != NULL, return);
    gdlf_aassert (node != NULL, return);
    gdlf_aassert (list->last != NULL, return);
    
    prev = qh_list_prev (list, node);
    if (prev == NULL)           /* not found */
        return;

        /*
         *  Copy next pointer (node->head) to the previous's, so the
         * chain is kept. if 'prev == node', 'node' is the first one
         * in the list, else it's somewhere in the middle.  
         *
         * Try to be as atomic as possible, so that the host
         * controller doesn't get a half changed structure and stops
         * with a consistency error.
         */

    write_lock (&list->lock);
    if (list->last == node)
        list->last = prev;
    prev->head.word = node->head.word;
    node->head.word = cpu_to_le32 (0x3);
    write_unlock (&list->lock);
    return;
}

void qh_ilist_delete (ts_uhci_tb_list* list, ts_uhci_qh* node)
{
    ts_uhci_qh* prev;

    gdlf_fntrace ("qh_ilist_delete (0x%p, 0x%p)", list, node);    
    gdlf_aassert (list != NULL, return);
    gdlf_aassert (node != NULL, return);
    
    write_lock (&list->lock);
    prev = qh_ilist_prev (list, node);
    if (prev == NULL)           /* not found */
        return;

        /*
         *  Copy next pointer (node->head) to the previous's, so the
         * chain is kept. if 'prev == node', 'node' is the first one
         * in the list, else it's somewhere in the middle.  
         * 
         *  However, take into account that, if this element is the
         * first one of the list, then the pointer to it is the
         * element pointer of the list's head QH.
         *
         * Try to be as atomic as possible, so that the host
         * controller doesn't get a half changed structure and stops
         * with a consistency error.
         *
         *  Note, when actualizing the list->last entry, we must make
         * sure we're consistent. If we are the last ones, we must set
         * list->last to be our previous, _unless_ we are the first
         * ones, in which case, we set last to NULL, as the list is
         * empty. 
         */

    if (prev == node)                           /* node is 1st? */
    {
        gdlf_assert (ptr2kvoid (list->first.elt.bits.ptr << 4) == node);
        if (node == list->last)     /* list is empty */
        {
            list->last = NULL;
            list->first.elt.word = cpu_to_le32 (0x3);
        }
        else                        /* list is not empty */
        {
            list->last = prev;
            list->first.elt.word = node->head.word; /* removed */
        }
    }
    else                                        /* node is not the 1st */
    {
        prev->head.word = node->head.word;
        if (node == list->last)                 /* actualize last */
            list->last = prev;
    }
    node->head.word = cpu_to_le32 (0x3);        /* invalidate */
    write_unlock (&list->lock);
    return;
}


    /* Destroy the HCI data pending from a transaction descriptor
     *
     * Parameters:  
     *
     *  ts_usb_tr*      Pointer to the transaction data.
     *
     * Returns:
     *
     *  int             0 if ok, errno code < 0 on error.
     */

int uhci_tr_setdn (ts_usb_tr* tr)
{
    gdlf_fntrace ("uhci_tr_setdn (0x%p)", tr);

    gdlf_aassert (tr != NULL, return -EINVAL);
    gdlf_aassert (tr->status == e_usb_tr_status_setup, return -EBADFD);
    if (tr->hci)
        ts_uhci_tr_hci_release (tr->hci);
    return 0;
}


    /* Setup the HCI data of a transaction
     *
     * Parameters:  
     *
     *  ts_usb_tr*      Pointer to the transaction data in FILLED
     *                  state. 
     *
     * Returns:
     *
     *  int             0 if ok, errno code < 0 on error.
     */

int uhci_tr_setup (ts_usb_tr* tr)
{
    int errcode;
    ts_uhci_tr_hci *hci;
    
    gdlf_fntrace ("uhci_tr_setup (0x%p)", tr);

    gdlf_aassert (tr != NULL, return -EINVAL);
    gdlf_aassert (tr->status == e_usb_tr_status_filled, 
                  return -EBADFD);

    hci = tr->hci = ts_uhci_tr_hci_init();
    if (hci == NULL)
    {
        gdlf_err ("No memory to set up the HCI information "
                  "for transaction 0x%p\n", tr);
        return -ENOMEM;
    }

    hci->tb.tr = tr;
    errcode = uhci_tb_fill (&hci->tb);
    if (errcode < 0)
    {
        gdlf_err ("Cannot prepare HCI data for "
                  "transaction 0x%p\n", tr);
        usb_kfree (hci);
        tr->hci = NULL;
    }
    return errcode;
}


    /* Insert TBs into the transfer list
     *
     * Parameters:  
     *
     *  ts_usb_tr*      Pointer to the transaction data in SETUP
     *                  state. 
     *
     * Returns:
     *
     *  int             0 if ok, errno code < 0 on error.
     */

#warning DEBUG: UHCI_SLOW_CTL
#undef UHCI_SLOW_CTL

int uhci_tr_insert (ts_usb_tr* tr)
{
    ts_uhci_tr_hci *thci;       /* transaction's HCI */
    ts_uhci_host *hci;          /* host's HCI */
    ts_uhci_tb_list *hlist;     /* host-wide list in which we hook dlist*/
    
    gdlf_fntrace ("uhci_tr_insert (0x%p)", tr);

    gdlf_aassert (tr != NULL, return -EINVAL);
    gdlf_aassert (tr->open != NULL, return -EINVAL);
    gdlf_aassert (tr->open->dev != NULL, return -EINVAL);
    gdlf_aassert (tr->open->dev->hub != NULL, return -EINVAL);
    gdlf_aassert (tr->open->dev->hub->port != NULL, return -EINVAL);
    gdlf_aassert (tr->open->endp != NULL, return -EINVAL);
    gdlf_aassert (tr->open->host != NULL, return -EINVAL);
    gdlf_aassert (tr->open->host->hci != NULL, return -EINVAL);
    gdlf_aassert (tr->status == e_usb_tr_status_setup, return -EBADFD);

        /* Get the HCIs */

    thci = (ts_uhci_tr_hci*)tr->hci;
    hci = UHCI (tr->open->host->hci);

        /* Which list do we have to link into? */

    switch (tr->open->endp->type)
    {
      case e_usb_tr_type_cntl:      /* Control list */
        gdlf_aassert (hci->ls_ctrl != NULL, return -EINVAL);
        gdlf_aassert (hci->ctrl != NULL, return -EINVAL);
#ifndef UHCI_SLOW_CTL
        if (tr->open->dev->hub->port[tr->open->dev->port].lowspeed)
            qh_list_append (hci->ls_ctrl, &thci->tb.qh);
        else
            qh_list_append (hci->ctrl, &thci->tb.qh);
#else /* UHCI_SLOW_CTL */
        hlist = intr_hash_list (UHCI(tr->open->host->hci), 
                                1 + tr->open->dev->hub->
                                port[tr->open->dev->port].try);
        if (hlist == NULL)          /* mem alloc error */
        {
            gdlf_err ("Couldn't allocate memory for an "
                      "Interrupt HCI list on host 0x%p (%d)\n",
                      tr->open->host, tr->open->host->addr);
            return -ENOMEM;
        }
        qh_ilist_append (hlist, &thci->tb.qh);
#endif /* UHCI_SLOW_CTL */
#warning FIXME: update statistical insertion data
        break;
        
      case e_usb_tr_type_bulk:      /* Bulk list */
        gdlf_aassert (hci->bulk != NULL, return -EINVAL);
        qh_list_append (hci->bulk, &thci->tb.qh);    
#warning FIXME: update statistical insertion data
        break;
        
      case e_usb_tr_type_intr:      /* Interrupt list */
        hlist = intr_hash_list (UHCI(tr->open->host->hci), 
                                tr->open->endp->interval);
        if (hlist == NULL)          /* mem alloc error */
        {
            gdlf_err ("Couldn't allocate memory for an "
                      "Interrupt HCI list on host 0x%p (%d)\n",
                      tr->open->host, tr->open->host->addr);
            return -ENOMEM;
        }
        qh_ilist_append (hlist, &thci->tb.qh);
#warning FIXME: update statistical insertion data
        break;
        
      case e_usb_tr_type_isoc:      /* iso */
#warning FIXME: Isochronous transfers unimplemented
        return -ENOSYS;
        break;

      default:
#ifdef GDLF_ASSERTIONS
        gdlf_emerg ("What am I doing here? " gdlf_location);
#endif
        return -EINVAL;
    }
    return 0;
}


    /* Remove transaction's HCI data from the frame list
     *
     * Parameters:  
     *
     *  ts_usb_tr*      Pointer to the transaction data in INSERTED
     *                  state. 
     *
     * Returns:
     *
     *  int             0 if ok, errno code < 0 on error.
     */

int uhci_tr_remove (ts_usb_tr* tr)
{
    ts_uhci_host* hci;
    
    gdlf_fntrace ("uhci_tr_remove (0x%p)", tr);
    gdlf_aassert (tr != NULL, return -EINVAL);
    gdlf_aassert (tr->hci != NULL, return -EINVAL);
    gdlf_aassert (tr->open->dev != NULL, return -EINVAL);
    gdlf_aassert (tr->open->dev->hub != NULL, return -EINVAL);
    gdlf_aassert (tr->open->dev->hub->port != NULL, return -EINVAL);
    gdlf_aassert (tr->open->host != NULL, return -EINVAL);
    gdlf_aassert (tr->open->host->hci != NULL, return -EINVAL);
    gdlf_aassert (tr->open->endp != NULL, return -EINVAL);
    gdlf_aassert (tr->status == e_usb_tr_status_inserted, return -EBADFD);

        /* Get the list we have to remove from */

    hci = UHCI (tr->open->host->hci);
    switch (tr->open->endp->type)
    {
      case e_usb_tr_type_cntl:      /* Control list */
        gdlf_aassert (hci->ls_ctrl != NULL, return -EINVAL);
        gdlf_aassert (hci->ctrl != NULL, return -EINVAL);
#ifndef UHCI_SLOW_CTL
        if (tr->open->dev->hub->port[tr->open->dev->port].lowspeed)
            qh_list_delete (hci->ls_ctrl, 
                            &((ts_uhci_tr_hci*)tr->hci)->tb.qh);
        else
            qh_list_delete (hci->ctrl, 
                            &((ts_uhci_tr_hci*)tr->hci)->tb.qh);
#else /* UHCI_SLOW_CTL */
        qh_ilist_delete (intr_hash_list (UHCI(tr->open->host->hci), 
                                         1 + tr->open->dev->hub->
                                         port[tr->open->dev->port].try),
                         &((ts_uhci_tr_hci*)tr->hci)->tb.qh);
#endif /* UHCI_SLOW_CTL */
        break;
        
      case e_usb_tr_type_bulk:      /* Bulk list */
        gdlf_aassert (hci->bulk != NULL, return -EINVAL);
        qh_list_delete (hci->bulk, &((ts_uhci_tr_hci*)tr->hci)->tb.qh);
        break;
        
      case e_usb_tr_type_intr:      /* Interrupt list */
        qh_ilist_delete (intr_hash_list (hci, tr->open->endp->interval), 
                         &((ts_uhci_tr_hci*)tr->hci)->tb.qh);
        break;
        
      case e_usb_tr_type_isoc:
#warning FIXME: isochronous removing not implemented
        return -ENOSYS;
        
      default:
#ifdef GDLF_ASSERTIONS
        gdlf_emerg ("What am I doing here? " gdlf_location);
#endif
        return -EINVAL;
    }
    return 0;
}


    /* Make the transaction execute
     *
     *  If the transaction queue is empty of transactions in the QUEUE
     * or EXECUTING states), then it will send it to run, otherwise,
     * it'll let it QUEUED until the host can service it.
     *
     *  If it is a control, bulk or interrupt transaction, we just
     * have to clear the terminate bit on the TB's QH element [it is
     * suppossed the TB is in the INSERTED state]. This will cause the
     * host to get into it next time it runs over the TB's QH.
     *
     *  If it is an interrupt transaction, we have to clear the
     * inactive bits in all the TB's TDs.
     *
     *  Here we have to deal with locks too. If the transaction is to
     * be synchronous, we set a wait queue and wait on it.
     *
     *  We assume the transaction is in the INSERTED state.
     *
     * Parameters:  
     *
     *  ts_usb_tr*      Pointer to the transaction data in INSERTED
     *                  state. 
     *
     * Returns:
     *
     *  int             0 if ok, errno code < 0 on error.
     */

int uhci_tr_run (ts_usb_tr* tr)
{
    int errcode = 0;
    char done_warn = 0;             /* to save original parameters */
    char sync = 0;                  /* synchronous call? */
    char method = 0;                /* save original warn method */
    struct wait_queue *wq;          /* if sync call ... */

    gdlf_fntrace ("uhci_tr_run (0x%p)", tr);

    gdlf_aassert (tr != NULL, return -EINVAL);
    gdlf_aassert (tr->hci != NULL, return -EINVAL);
    gdlf_aassert (tr->open->endp != NULL, return -EINVAL);
    gdlf_aassert (tr->status == e_usb_tr_status_inserted, return -EBADFD);

    sync = (tr->warn.wq == NULL);
    if (sync)                   /* synchronous? */
    {
        done_warn = tr->done_warn;  /* save old data */
        method = tr->method;
        tr->method = 0;             /* warn using wait queue */
        init_waitqueue (&wq);       /* set new data */
        tr->warn.wq = &wq;
        tr->done_warn = 1;
    }
    
        /* 
         *  Make it run only if the queue is empty, the first element
         * is our transaction or it is empty of running elements (this
         * last one is a security fallback). If not empty, the IRQ
         * handler will notice when it can execute more transactions
         * and will do it for us.  
         */

    if (ts_slist_head (&tr->open->list) == NULL
        || ts_slist_head (&tr->open->list) == &tr->snode
        || (
            ((ts_usb_tr*)ts_slist_head (&tr->open->list))->status != 
              e_usb_tr_status_executing
            &&
            ((ts_usb_tr*)ts_slist_head (&tr->open->list))->status != 
              e_usb_tr_status_queued)
        )
    {       /* Queue is empty, we've got to make it run */

        tr->status = e_usb_tr_status_executing;
        uhci_tb_activate (&((ts_uhci_tr_hci*)tr->hci)->tb);
    }
    else    /* Queue is not empty, we've just appended it */
        tr->status = e_usb_tr_status_queued;
    
        /* Synchronous call? wait for it and later restore */
    
    if (sync)
    {
        interruptible_sleep_on (&wq);
        if (signal_pending (current))
            errcode = -EINTR;
        else
            errcode = 0;
        tr->done_warn = done_warn;
        tr->method = method;
        tr->warn.wq = NULL;
    }
    return errcode;
}


    /* FIXME: Detailed desc
     *
     *  Deactivate all TDs as fast as possible? 
     *
     * Compliance:      USB Spec 1.0 xxxx
     *
     * Parameters:  
     *
     *  type            descr
     *
     * Global:
     *
     *  type name       descr
     *
     * Returns:
     *
     *  type            descr
     *
     * Global:
     *
     *  type name       descr
     */

int uhci_tr_abort (ts_usb_tr* tr)
{
    gdlf_fntrace ("uhci_tr_abort (0x%p)", tr);
    gdlf_aassert (tr != NULL, return -EINVAL);
    gdlf_aassert (tr->hci != NULL, return -EINVAL);
    move_to_error ((ts_uhci_tb*)tr->hci, -USBE_TR_ABORTED, -EINTR);
    return 0;
}


    /* Prepare a transfer block for a transaction
     *
     *  Create an array of ts_uhci_td (transfer descriptors)
     * partitioning the transfer for a given buffer, given a
     * ts_usb_tb*, from where the transactions' information is
     * obtained. It will fill in the structure.
     *
     *  When you're about to transfer on the USB bus, you've got to
     * transfer each packet with the maximum endpoint size, but the
     * last one, which doesn't need to be padded.
     * 
     *  There's no mention of endianess for UHCI, but I guess it'll
     * always be Little Endian (coming from Intel ...)
     *
     * NOTE: I speak here of 'data phase'. When I talk about it I mean
     *       the data phase on a control transaction or the whole
     *       transacion's data in a bulk, interrupt or isochronous
     *       one.
     *
     * Parameters:
     *
     *  ts_uhci_tb*     Pointer to the structure where to store the
     *                  data. It must be already allocated; if not
     *                  clean (the tds pointer is not NULL), the array
     *                  will be considered allocated, and it'll try to
     *                  re-use it (if the size is equal or more than
     *                  needed). If the size les than needed, the
     *                  array will be freed.
     *
     *                  If the structure is clean or the array has
     *                  been cleaned, a new array will be allocated of
     *                  the exact needed size.
     *  
     *                  YOU MUST FILL THE tr POINTER, SO THE CODE
     *                  KNOWS WHERE THE TRANSACTIONS'S INFORMATION IS
     *                  AT.
     *
     * Returns:
     *
     *  int             0 if ok, errno code < 0 on error.
     *
     * Global:
     *
     *  int uhci_errno  Unmodified if ok, USB errno code < 0 on
     *                  error.  
     */

int uhci_tb_fill (ts_uhci_tb* tb)
{
    size_t tds;         /* number of tds to allocate */
    ts_uhci_td *td;     /* array of tds */
    ts_uhci_td *dfirst; /* first td for data phase */
    ts_uhci_td *ditr;   /* td iterator for data phase */
    ts_uhci_td *dtop;   /* top td for data phase */
    e_usb_tr_type type; /* transfer type */
                        
    unsigned maxlen;    /* maximum length */
    __u8* buffer;       /* buffer pointer iterator */
                        
    ts_usb_tr* tr;      /* transaction [to clarify] */
    
    gdlf_fntrace ("uhci_tb_fill (0x%p)", tb);

    gdlf_aassert (tb != NULL, return -EINVAL);
    gdlf_aassert (tb->tr != NULL, return -EINVAL);
    gdlf_aassert (tb->tr->open != NULL, return -EINVAL);
    gdlf_aassert (tb->tr->open->endp != NULL, return -EINVAL);
    gdlf_aassert (tb->tr->open->dev != NULL, return -EINVAL);
    gdlf_aassert (tb->tr->open->dev->hub != NULL, return -EINVAL);
    gdlf_aassert (tb->tr->open->dev->hub->port != NULL, return -EINVAL);

#warning FIXME: code pkts warnings

    tr = tb->tr;
    type = tr->open->endp->type;

#ifdef GDLF_ASSERTIONS
    if (type == e_usb_tr_type_cntl)
        gdlf_aassert (tr->req != NULL, return -EINVAL);
#endif
    
        /* 
         *  Let's calculate the number of tds we're going to
         * require. We start with the size of the data to transfer [if
         * control xfer, the data phase]. If there's no data to
         * transfer, we have to transfer one NULL packet at least
         * [except on control].
         *
         *  Also, if control transfer, we have to add a td for the setup
         * phase and another for the status phase, but if there's no
         * data, we don't need the NULL data packet.
         */

    if (tr->data == NULL)               /* no data? */
        tds = type == e_usb_tr_type_cntl? 0 : 1;
    else                                /* there's data */
        tds = (tr->size / tr->open->endp->maxpktsize)
            + ((tr->size % tr->open->endp->maxpktsize) != 0? 1 : 0);

    if (type == e_usb_tr_type_cntl)     /* control setup and ack? */
        tds += 2;

    if (type == e_usb_tr_type_isoc      /* verify ISO size */
        && tds > UHCI_FL_SIZE)
    {
        usb_errno = -USBE_HCI_DATA_TOO_BIG;
        return -EFBIG;
    }

        /* Ok, let's allocate or reallocate the tds array */

    if (tb->td == NULL)                 /* do we have to allocate? */
    {                                   
        td = ts_uhci_td_a_init (tds);   /* allocate the tds array */
        if (td == NULL)                 /* Error? */
        {
            usb_errno = -USBE_NO_MEMORY;
            return -ENOMEM;             /* and return */
        }
    }                                   
    else                                /* No need to allocate, reuse */
    {                
            /*
             * Warning! When recoding to make reuse of TD arrays
             * bigger then needed take into account that other
             * functions have to be modified to evitate having the TDs
             * which ain't used activated/resetted ... etc
             */
                   
        td = tb->td;                    /* get original */
        if (tb->tds == tds)             /* same space we need? ... */
            ts_uhci_td_a_release (td, tds);/* ... yep, clean */
        else                            /* no, reallocate the array */
        {
            ts_uhci_td_a_release (td, tds); /* release */
            usb_kfree (td);                 /* free */
            tb->td = NULL;                  /* jic */
            td = ts_uhci_td_a_init (tds);   /* reinitialize the array */
            if (td == NULL)                 /* error? ... */
            {
                usb_errno = -USBE_NO_MEMORY;
                return -ENOMEM;             /* ... return it */
            }
        }
    }

        /*
         *  So we have the number of tds we need and space allocated
         * for them. Let's fill it.
         *
         *  Initialize the first data td with the data it we'll use during
         * all operations. We'll clone that td with memcpy, and then
         * initialize the per-td specific stuff. Note we'll make a couple
         * special cases on control transfers.
         *
         *  The buffer pointer is defined to NULL, initially.
         */

    td[0].buffer = (__u32)NULL;
    
        /*  Token fields:
         *
         *  PID will be constant all the time, so we set it
         * here. Control transfers will modify it later to make the
         * first one a setup packet.
         *
         *  The address is just device's address, endpoint's is
         * endpoint's address >> 1, as we use first bit to identify
         * endpoint directions (as used to get the PID); however, this
         * is problematic on endpoint zero, as it is _always_ zero,
         * for in and out, so on control transfers, we _retouch_ a
         * little later, when setting up the setup and status
         * packets. 
         *
         *  Toggle is the programmed one unless control transfers,
         * which always start on DATA0.
         */

    td[0].token.bits.pid = USB_ENDP_READABLE (tr->open->endp->addr)? 
         USB_PID_IN : USB_PID_OUT;
    td[0].token.bits.addr = tr->open->dev->addr;
    td[0].token.bits.endp = tr->open->endp->addr >> 1;
    td[0].token.bits.dt = type == e_usb_tr_type_cntl? 
        0 : tr->toggle? 1 : 0;

        /*
         * Status fields
         *
         *  The fields the host touches are the c_err (error count),
         * status and actlen. We write the others, and set c_err to
         * UHCI_RETRIES and actlen to UHCI_ACTLEN. This is a magic
         * number to help determining the TDs status.
         *  
         *  Isochronous transfers aren't put as active directly,
         * because in the INSERTED state they'd start transfering. We
         * don't want interrupts at end until the last TD. We also
         * want three retries on error after taking care of it.
         *
         *  Short packet detection is currently enabled. When we
         * collect at the end of transfer, if short packets, we drop
         * an IO error.
         */

    td[0].ctstat.word = 0;              /* clean */
    td[0].ctstat.bits.actlen = UHCI_ACTLEN; /* TD is inserted/queued */
    td[0].ctstat.bits.active = type == e_usb_tr_type_isoc? 0 : 1;
    td[0].ctstat.bits.ioc = 0;          /* no interrupt at end of xfer */
    td[0].ctstat.bits.iso = type == e_usb_tr_type_isoc? 1 : 0;  /* isochr? */
    td[0].ctstat.bits.ls =              /* Is it low speed? */
        tb->tr->open->dev->hub->port[tb->tr->open->dev->port].lowspeed;
    td[0].ctstat.bits.c_err = UHCI_RETRIES; /* Want three error retries (max) */
    td[0].ctstat.bits.spd = 1;          /* Want short packet detection */

        /*
         * Link pointer
         *
         *  Only the last one terminates, and rest have as LP the next
         * TD (in LE32, bus address map), the Q bit reset and the VF
         * bit set [next TD] for control, bulk and interrupt.
         *  
         *  What's next (Q bit) for an isochronous is determined when
         * it is inserted, so we leave it at 1 [normally everything is
         * liked to TDs in queues, but isochronous use to link to
         * interrupt queues].
         */

    if (type == e_usb_tr_type_isoc)
        td[0].lp.word = cpu_to_le32 (0x2);
    else
        td[0].lp.word = cpu_to_le32 (0x4);
    
        /* 
         *  For control transfers, first and last tds are special, so
         * let's set'em up.
         */

    if (type == e_usb_tr_type_cntl)
    {
#ifdef UHCI_SLOW_CTL
        td[0].lp.bits.vf = 1;
#endif

            /* 
             * This is the setup request packet (8 bytes data) 
             *
             *  Data phase setup loop only sets up link pointers for
             * the data phase, not for this one, so we have to set it
             * to the next TD [it's always sure there are at least two
             * on a control transfer].
             */

        dfirst = &td[1];        /* set new reference one */
        memcpy (dfirst, &td[0], sizeof (td[0]));
        dfirst->token.bits.pid = 
            ts_usb_request_dir (tr->req->bmreqtype) == e_usb_req_dir_h2d?
            USB_PID_OUT : USB_PID_IN;
        dfirst->token.bits.dt = dfirst->token.bits.dt? 0 : 1;
        
        td[0].buffer = cpu_to_le32 (virt_to_bus (tr->req));
        td[0].token.bits.pid = USB_PID_SETUP;
        td[0].token.bits.maxlen = cpu_to_le16 (USB_REQ_XFER_LENGTH-1);
        td[0].lp.bits.ptr = cpu_to_le32 (virt_to_bus (&td[1])) >> 4;

            /* 
             * This is the status packet 
             *
             *  We copy from the first td, already setup with the
             * common data, and setup the top TD (dtop) in a way it is
             * this one, as the data phase doesn't have to touch it.
             *
             *  The direction is the inverse of the data stage one, or
             * IN if there's no data stage, and it is a NULL data
             * packet (see UHCI spec 3.2.3 ). The DATA toggle is
             * always 1.
             *
             *  This is the last packet, so we terminate it directly.
             */

        dtop = &td[tds-1];      /* we don't want last to be touched */
        if (dtop != dfirst)     /* don't ... copy twice */
            memcpy (dtop, dfirst, sizeof (td[0]));

        if (ts_usb_request_dir (tr->req->bmreqtype) == e_usb_req_dir_h2d
            || tr->data == NULL)
            dtop->token.bits.pid = USB_PID_IN;
        else
            dtop->token.bits.pid = USB_PID_OUT;
        dtop->token.bits.dt = 1;
        dtop->token.bits.maxlen = cpu_to_le16 (0x7ff); /* no data */
        dtop->lp.bits.t = 1;
    }
    else    /* this is not a control transfer, all is data */
    {
        dfirst = &td[0];
        dtop = &td[tds];    /* now we need last being touched */
    }
    
        /* 
         *  After this, the reference TD for copying is the one
         * pointer to by 'dfirst'.
         *
         *  Set up the data stage (if it exists) 
         */

    if (dtop > dfirst)  /* jump in if there's data phase */
    {   
            /* 
             *  If buffer is longer than the max packet size, size is 
             * maxpktsize, else, buffer size. Note we store it in the
             * TD as n-1, and don't set maxlen to maxlen-1, as we need
             * it later to incrementing the buffer variable. There's
             * no performance loop as the TD is memcpy()ed.
             *
             *  The buffer 'pointer' is used because we store the
             * others on little endian format, and we've gotta be
             * safe. We cannot convert it to bus directly because the
             * page allocation doesn't have to be continuous.  
             */

        maxlen = tr->size > tr->open->endp->maxpktsize?
            tr->open->endp->maxpktsize : tr->size;
        buffer = tr->data;
        dfirst->token.bits.maxlen = cpu_to_le16 (maxlen - 1);
        dfirst->buffer = cpu_to_le32 (virt_to_bus (buffer));
        dfirst->lp.bits.ptr = cpu_to_le32 (virt_to_bus (dfirst + 1)) >> 4;

            /* 
             *  So, now we run along the array [skipping the first
             * one, as we already filled it] and copy the first one to
             * the TD in turn and actualize the buffer pointer and the
             * DATA toggle.
             *
             * NOTE: this comment is actually deprecated 
             *
             *  Host reads the memory lines, so that's _bus_
             * addresses. In kernel land we've got physical address
             * and in user land, virtual addresses. Using tr->user, we
             * guess which one we have to use (note that if using
             * user-land pointers, you should _lock_ the page(s) where
             * the memory is until the transaction is done).
             */

        buffer += maxlen;       /* we've already set first one */        
        for (ditr = dfirst + 1; ditr < dtop; 
             ditr++, buffer += maxlen)
        {
            memcpy (ditr, dfirst, sizeof (*ditr));
            ditr->buffer = cpu_to_le32 (virt_to_bus (buffer));
            ditr->token.bits.dt = !(ditr-1)->token.bits.dt;
            ditr->lp.bits.ptr = cpu_to_le32 (virt_to_bus (ditr+1)) >> 4;
        }
        ditr--;      /* Now ditr is the last TD of the data phase */
        
            /* 
             *  Set the size of the last packet (on a multi packet data
             * phase, all have same size except the last one, which
             * may transfer less than that). Remember size on TDs is
             * coded as n-1, and that after the for(;;) loop, ditr ==
             * dtop. Note, however, if the remainder of tr->size %
             * maxlen is zero, then we don't have to modify it.
             */

        if (dtop > dfirst + 1 && (tr->size % maxlen != 0))
            ditr->token.bits.maxlen = cpu_to_le16 (tr->size % maxlen - 1);

            /* 
             *  Put the lp.ptr to the last TD of a data phase on
             * control to last TD [ACK], which is already terminated.
             */

        if (type == e_usb_tr_type_cntl)
            ditr->lp.bits.ptr = cpu_to_le32 (virt_to_bus (dtop)) >> 4;
    }
    else    /* there's no data phase, this is a ctl xfer */
        ditr = dfirst;

        /* Last TD terminated at the end */

        /*
         *  Now we setup QH's element pointer. We put it terminated,
         * so that it doesn't start to transfer as soon as it is
         * inserted into the transfer list. Of course, it doesn't
         * apply to isoc transactions, they ain't queued.
         *
         *  & 0xfffffff0 -> makes q = 0
         *  | 0x00000003 -> makes t = 1.
         */

    if (type != e_usb_tr_type_isoc)
    {
        tb->qh.elt.word = (cpu_to_le32 (virt_to_bus (td)) 
                           & 0xfffffff0) | 0x1;
        tb->qh.head.word = cpu_to_le32 (0x3);
    }

    td[tds-1].ctstat.bits.ioc = 1;          /* last packet interrupts */
    td[tds-1].lp.word = cpu_to_le32 (0x1);  /* last packet signaling */
        /* (see uhci_fl_search()) */

            /*
             * Warning! When recoding to make reuse of TD arrays
             * bigger then needed take into account that other
             * functions have to be modified to evitate having the TDs
             * which ain't used activated/resetted ... etc
             */
                   
    tb->td = td;
    tb->tds = tds;

    return 0;
}


    /* Reset a transaction block
     *
     *  We have the transaction in any of the OK, ABORTED or ERROR
     * states, and we have to take it to the INSERTED state.
     *
     *  A transaction being resetted means it get's all it's HCI data
     * fields back to as they were when uhci_tb_fill() was called on
     * it. 
     *
     *  The UHCI host controller _ONLY_ touches the status, actlen and
     * c_err fields of the ctstat field of a TD, so we are only going
     * to touch those ones and activate the active bit on
     * non-isochronous transfers. We are also going to update the
     * element pointer in the QH, set the Q bit to zero and terminate
     * (t = 1) the QH.
     *
     * Parameters:  
     *
     *  ts_usb_tb*      Pointer to the transaction block data.
     *
     *  e_usb_tr_type   Endpoint type.
     *
     * Returns:
     *
     *  int             0 if ok, errno code < 0 on error.  
     */

int uhci_tb_reset (ts_uhci_tb* tb, e_usb_tr_type type)
{
    ts_uhci_td *itr, *top;  /* td iterators */
    
    gdlf_fntrace ("uhci_tb_reset (0x%p, %d)", tb, type);
    gdlf_aassert (tb != NULL, return -EINVAL);

    tb->qh.elt.word = cpu_to_le32 
        ((0xfffffff0 & (virt_to_bus (tb->td))) | 0x1);

        /*
         *  We walk all the TDs, and mask off to zero all modified
         * fields and then recode them.
         *
         * & 0xe700f800 clear c_err, status and actlen
         * | 0x00800000 activate TD
         * | 0x18000000 = UHCI_RETRIES << 27, set c_err to
         *                                    UHCI_RETRIES
         *
         *  Note UHCI doesn't need OR'ing the active bit, because we
         * don't want to start transmiting inmediately.  
         */

    switch (type)
    {
            /* We don't know the type, apply isochronous case only */

      case e_usb_tr_type_inv:
      case e_usb_tr_type_isoc:
        for (itr = tb->td, top = itr + tb->tds; itr < top; itr++)
            itr->ctstat.word = (itr->ctstat.word & 0xe700f800)
                | UHCI_RETRIES << 27;
        break;
        
      case e_usb_tr_type_cntl:
      case e_usb_tr_type_bulk:
      case e_usb_tr_type_intr:
        for (itr = tb->td, top = itr + tb->tds; itr < top; itr++)
            itr->ctstat.word = (itr->ctstat.word & 0xe700f800)
                | 0x00800000 | UHCI_RETRIES << 27;  
        break;


      default:
#ifdef GDLF_ASSERTIONS
        gdlf_err ("What am I doing here? type = %d\n", type);
#endif
    }
    return 0;
}


    /* Activate a TB to allow it to execute
     *
     * Parameters:  
     *
     *  ts_uhci_tb*     Pointer to the TB to activate.
     */

void uhci_tb_activate (ts_uhci_tb* tb)
{
    gdlf_fntrace ("uhci_tb_activate (0x%p)", tb);

    if (tb->tr->open->endp->type == e_usb_tr_type_isoc)
    {
        ts_uhci_td *itr, *top;  /* TD iterator and top */
#warning FIXME: huh ... have to keep track of where hc is reading
#warning FIXME: isochronous activation not implemented
            /* 
             *  If we don't it _will_ start transferring to/from the
             * wrong place.  
             */
        for (itr = tb->td, top = itr + tb->tds; itr < top; itr++)
            itr->ctstat.bits.active = 1;
    }
    else        /* control, bulk or interrupt, reset qh.bits.t */
        tb->qh.elt.bits.t = 0;
    tb->tr->status = e_usb_tr_status_executing;
    return;
}


/* Log:
 *
 * $Log: scheduler.c,v $
 * Revision 1.20  1998/11/14 17:45:47  inaky
 * Moved the include files to uusbd/include/uusbd, to ease integration
 * into the Linux kernel.
 *
 * Revision 1.19  1998/11/09 21:44:27  inaky
 * Changed to use macros to detect endpoint sense.
 *
 * Revision 1.18  1998/09/24 16:19:56  inaky
 * Removed setting the T bit on the TBs QH, as move_to_error does it.
 *
 * Revision 1.17  1998/09/14 20:37:25  inaky
 * Added an slow control mode [compile time] on which control transfers
 * are scheduled as interrupt transfers with an interval dependant of the
 * port's try count and with Vf set to one, so each packet is scheduled
 * each X frames. For debugging only.
 *
 * Revision 1.16  1998/08/30 14:01:34  inaky
 * Finished new list's linear topology. Added option to enable/disable bw
 * reclamation at compile time. Should move it to be run-time
 * selectable.
 *
 * Revision 1.15  1998/08/29 00:11:51  inaky
 * Fixed endianess on qhs and tds, and option to disable bandwidth
 * reclamation.
 *
 * Revision 1.14  1998/08/27 22:23:39  inaky
 * Moved to a linear list, more simple and elegant.
 *
 * Revision 1.13  1998/08/24 02:31:51  inaky
 * Removed spurious debug info. Changed name of all enumerations to be
 * prefixed with the enumeration type name.
 *
 * Did the little debugging needed to make the functions work as they're
 * expected to.
 *
 * Revision 1.12  1998/08/20 00:00:46  inaky
 * Added some (disabled) debug code, fixed timeouts and PID
 * generation.
 *
 * Revision 1.11  1998/08/03 13:52:53  inaky
 * Made the dev_setdn function remove the device list heads from the host
 * lists when destroying them.
 *
 * Fixed the qh_list_prev() function to work correctly, not considering
 * the list head as a possible previous element.
 *
 * Fixed qh_list_delete() to take in account this last change. Added a
 * per-transaction-timer to timeout hanged transfers.
 *
 * Revision 1.10  1998/07/30 23:00:51  inaky
 * Removed unused debug information.
 *
 * Revision 1.9  1998/07/30 18:00:55  inaky
 * Fixed the qh list insertion function. Finished uhci_tb_activate().
 *
 * Revision 1.8  1998/07/28 23:09:36  inaky
 * Redone the reset method to be split in two parts. It is making merits
 * to be eliminated, as it is quite unneeded in the UHCI side of the
 * problem (though still should be there for USBD/HCD interface).
 *
 * Disabled short packet interrupt. Short packet detection is done when
 * accounting at end of transfer, signalling an error if less data than
 * expected is transfered.
 *
 * Revision 1.7  1998/07/24 15:39:25  inaky
 * Removed the PAUSED state, it was quite useless ...
 *
 * Revision 1.6  1998/07/24 15:38:36  inaky
 * Move the IRQ handling to irq-handler.c, added checks to fix max ISO
 * size, usb_errno updating on errors and new flag for detecting
 * INSERTED/QUEUED states in TDs.
 *
 * Revision 1.5  1998/07/23 14:46:48  inaky
 * Moved ../usb-uhci.h to uhci.h
 *
 * Added the main scheduler functions: set up and down device HCI
 * information, QH list management, transaction HCI information set up
 * and down, transaction insertion and removal from the frame list,
 * transaction execution, pause, continue and reset. Also finished the
 * function to set up transaction blocks from the transaction
 * information.
 *
 * Revision 1.4  1998/07/16 14:17:52  inaky
 * Moved from the old read/write interface to the new, request-based
 * interface.
 *
 * Added some docs on the type of schedulement we are going to implement.
 *
 * Reworked some of the uhci_buftds_fill function, but much more working
 * is needed.
 *
 * Revision 1.3  1998/05/24 12:08:54  inaky
 * Added some comments to issues I have to keep track of.
 *
 * Revision 1.2  1998/05/01 19:21:27  inaky
 * Added the usbd.h header file inclusion, as it is needed.
 *
 * Revision 1.1  1998/05/01 18:31:28  inaky
 * Moved everything to use pci_dev according to new kernel PCI API. Also
 * split all the code in several files, as old uhci.c was a complete mess.
 *
 */

