
/* USB Driver Stack v$Revision: 1.20 $ 
 *
 * (C) 1997 by Iaky Prez Gonzlez <inaky@peloncho.fis.ucm.es>
 *
 * Distributed under the terms of the GNU Public License Version 2
 * See files COPYRIGHT and COPYING at the root distribution directory
 *
 * $Id: vhub.c,v 1.20 1998/11/14 17:45:50 inaky Exp $
 *
 * Changelog at the bottom 
 *
 *
 *     Virtual Root Hub
 *     ----------------
 *
 *  This is the infamous virtual root hub. Instead of simulating a
 * full USB device, I'm just going to simulate the hub (perhaps
 * overseeing the USB Spec). I'm doing it this way because I'll have
 * to treat specially that 'device' _anyway_, so I can save an
 * indirection level if I work only at the hub level.
 *
 *  Thus, vhub simulates _all_ the operations given at the structure
 * ts_usb_hub_ops to behave as a normal hub. Root hubs will be
 * distinctive because their device member is NULL. 
 * 
 *  Note we don't check if the hub we are passed as parameter has the
 * same operations structure as uhci_hub_operations. This is because
 * other hub type may use our methods, and thus, this check renders
 * useless.
 */

#include <uusbd/usb.h>
#include <uusbd/usb-hub.h>

#define GDLF_IDENT          "USB-UHCI"
#define GDLF_LOCAL          /* Define if want local settings */
#define GDLF_DEF_LOGLEVEL   9
#define GDLF_ASSERTIONS     /* want assertions (#undef if not) */
#undef GDLF_FNTRACE        /* want function traces */
#define GDLF_LOCATIONS      /* want locations in log messages */
#include <uusbd/gdlf.h>

#include "uhci.h"


    /* 
     * UHCI Root Hub descriptor data
     *
     * NOTE: There is no way to know in advance the number of ports,
     *       so I assume it is 2; neither can I find a release number
     *       or UHCI version to help keeping track on it ... funny.
     *        
     *  WARNING: This descriptor must contain little endian values
     *           so, on uhci initialization, the wHubCharacteristics
     *           field is swapped. 
     */

#define UHCI_HUB_DESC_TYPE      e_usb_desc_dev
#define UHCI_PORTS          2   /* Just two ports :) */
#define UHCI_HUB_CHARACS                                                \
        cpu_to_le16 (                                                   \
            2                       /*  No power switching */           \
            | 0 << 2                /*  Not part of compound device */  \
            | 2 << 4)               /*  No over current protection */
#define UHCI_HUB_PWRON2GOOD     0   /* always powered on */
#define UHCI_HUB_CURRENT        0   /* bHubContrCurrent (no pwr required) */
#define UHCI_PORTDATA           /* port data for two ports */            \
    {                                                                        \
        0,                          /* 1 and 2 allow removable attachment */ \
        0,                          /* 1 and 2 dont mask gang pwr ctl */     \
                                                                             \
              0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,                      \
        0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,                      \
        0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,                      \
        0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0                       \
    }


    /* Get the UHCI hub descriptor
     *
     *  This function returns the UHCI hub descriptor. As this is a
     * virtual hub, we `make' it to match our needs. 
     *
     *  Note the ts_usb_hub_desc structure has more space than needed,
     * in order to fit all possible ports (as stated on the USB Spec,
     * chapter 11). We won't copy that unused space, just what's
     * needed. The calculus of the unused space is as follows:
     * 
     *  We reserve bit 0 of the bitfield array, so we sum one to get a
     * 'total' of ports. Thus we have 1 + UHCI_PORTS bits per
     * bitfield. As the bitfield is in byte granularity, to convert to
     * bytes, we need to divide by eight and add one if the remainder
     * of the division is nonzero, nothing otherwise. 
     *
     *  So, we are using this per port on each array. As we 'waste' 32
     * bytes per array, then the wasted space is:
     *
     *
     *                   UHCI_PORTS
     *   64 - 2 (0.125 + -------------- + ((1 + UHCI_PORTS) % 8) ? 1 : 0)
     *                         8
     *   
     *  Also note we don't relay in this `descriptor' for the
     * usb_hub_descriptor_fill() function, as we will fill it directly
     * to fit our needs. This function is only put here because it may
     * be needed by other parts of the UUSBD system.
     *
     * Compliance:      USB Spec 1.0 11.11.2.1
     *
     * Parameters:  
     *
     *  ts_usb_hub*     Pointer to hub struct whose descriptor we want
     *                  to set
     *
     *  __u8            Index number of the hub descriptor. Normally zero.
     *
     *  const void*     Pointer to data we'll use to set the
     *                  descriptor.
     *
     *  size_t          Data's size
     *
     * Returns:
     *
     *  int             0 if ok, errno code < 0 on error.
     *
     * Global:
     *
     *  int usb_errno       On error, error code.
     *
     *  int uhci_errno      On error, error code.
     *
     *  int usb_hub_errno   On error, error code.  
     */


ssize_t uhci_hub_descriptor_get (ts_usb_hub* hub, __u8 index, 
                                 void* data, size_t size)
{
    ssize_t minimum;
    ts_usb_hub_desc descriptor =    /* UHCI Hub Descriptor */
    {
        {
            sizeof (ts_usb_hub_desc) +  /* bDescLength, without the wasted space */
            - (64 - 2*(0.125 + UHCI_PORTS / 8 
                       + (1 + UHCI_PORTS) % 8 ? 1 : 0)),
            UHCI_HUB_DESC_TYPE      /* bDescriptorType */
        },
        UHCI_PORTS,                 /* bNbrPorts */
        UHCI_HUB_CHARACS,           /* wHubCharacteristics */
        UHCI_HUB_PWRON2GOOD,        /* bPwrOn2PwrGood (always powered :) */
        UHCI_HUB_CURRENT,           /* bHubContrCurrent (no pwr required) */
        UHCI_PORTDATA               /* DeviceRemovable & PortPwrCtlMask */
#warning FIXME: grrr! I only use two of 64 bytes!!
    };
    
    gdlf_fntrace ("uhci_hub_descriptor_get (0x%p, %u, 0x%p, %d)", 
                  hub, index, data, size);
    gdlf_aassert (data != NULL, return -EINVAL);

    minimum = size > descriptor.head.bLength? size : descriptor.head.bLength;
    memcpy (data, &descriptor, minimum);
    return minimum;
}


    /* Fill a ts_usb_hub structure from the hub descriptor data
     *
     *  However, we'll skip the getting of the descriptor data and
     * fill directly from our data.
     *
     * Parameters:  
     *
     *  ts_usb_hub*     Pointer to the hub structure to fill
     *
     * Returns:
     *
     *  int             0 if ok, errno code < 0 on error.
     *
     * Global:
     *
     *  int usb_hub_errno   On error, error code.  
     *
     *  int uhci_errno      On error, error code.  
     */


int uhci_hub_descriptor_fill (ts_usb_hub* hub)
{
    gdlf_fntrace ("uhci_hub_descriptor_fill (0x%p)", hub);
    gdlf_aassert (hub != NULL, return -EINVAL);

    hub->ports = UHCI_PORTS;
    hub->port = ts_usb_port_a_init (hub->ports);
    if (hub->port == NULL)
    {
        ts_usb_hub_release (hub);
        return -ENOMEM;
    }        

    hub->pwrontime = UHCI_HUB_PWRON2GOOD;
    hub->maxcurrent = UHCI_HUB_CURRENT;

    hub->power = (le16_to_cpu (UHCI_HUB_CHARACS)
                  | USB_HUB_CRTS_PWR_CTL_MASK);
    hub->compound = (le16_to_cpu (UHCI_HUB_CHARACS)
                     | USB_HUB_CRTS_COMPD_MASK)? 1 : 0;
    hub->overc_prot = (le16_to_cpu (UHCI_HUB_CHARACS)
                       | USB_HUB_CRTS_CURRENT_MASK) 
                           >> USB_HUB_CRTS_CURRENT_MASK_S;

        /* 
         * Initialize the ports' data 
         *
         *  We followe the same approach as in a normal hub, so we
         * don't have to worry if sometime we change the number of
         * ports or its characteristics. Anyway, it isn't that
         * expensive.
         */
    
    {
        int     byte;               /* byte counter */
        int     port, top;          /* port counter, top */
        int     bit;                /* bit counter */
        __u8    bits1, bits2;       /* Data for each byte of the bit arrays */
        __u8    portdata[] = UHCI_PORTDATA;
        
        for (port = byte = 0, top = hub->ports; port < top; byte++)
        {
            bits1 = portdata[byte];          /* Device removable array */
            bits2 = portdata[byte + top];    /* Power Ctl Mask array */
            for (bit = 0; 
                 bit < sizeof (char) && port < top;
                 port++, bit++, bits1 >>= 1, bits2 >>= 1)
            {
                if (byte == 0 && port == 0)  /* This one is reserved */
                    continue;
                hub->port[port].removable = bits1 & 0x1;
                hub->port[port].pwrctl = bits2 & 0x1;
            }
        }
    }

    return 0;
}


    /* Returns hub status
     *
     *  As we are always on the same mood, we don't change it, and
     * always return the same status (local power ok, power ops ok, no
     * change in local power status and no change on overcurrent
     * indicator).
     *
     * NOTE: this funcion will update the hub structure's fields
     *       `local_power' and `overcurrent'.
     *
     *  See hubd_status_get() and usb_hub_status_get()
     *
     * Compliance:      USB Spec 1.0 11.12.2.5
     *
     * Parameters:  
     *
     *  ts_usb_hub*     Pointer to the hub structure whose status we
     *                  want to check out.
     *
     * Returns:
     *
     *  __s32           If > 0, status code (and with HUBD_STAT_* to
     *                  get the bits) [NOTE: things may get hairy if
     *                  sometime someone decides to use bit 31 for
     *                  status ...]. On error, <0 errno code.
     */

__s32 uhci_hub_status_get (ts_usb_hub* hub)
{
    gdlf_fntrace ("uhci_hub_status_get (0x%p)", hub);
    gdlf_aassert (hub != NULL, return -EINVAL);

    return 
        (0 & USB_HUB_STATUS_C_LOCAL_PWR) |      /* no change on local power status */
        (0 & USB_HUB_STATUS_C_OVER_CURRENT) |   /* no change on current status */
        (0 & USB_HUB_STATUS_LOCAL_PWR) |        /* local pwr ok */
        (0 & USB_HUB_STATUS_OVER_CURRENT);      /* no overcurrent */
}


    /* Return the bus status as sampled on last EOF2
     *
     *  This function will read the bus status on fiven port as
     * sampled at the last EOF2 hub state. It will be returned on bits
     * 0 and 1 of the return value if > 0. You can then use the mask
     * values USB_HUB_BUS_STAT_DP and USB_HUB_BUS_STAT_DN to sample the
     * values of the D- and D+ lines.
     *
     * NOTE: This is a very lowlevel function, intended as a debugging
     *       aid.
     *
     * Compliance:      USB Spec 1.0 - 11.12.2.3
     *
     * Parameters:  
     *
     *  ts_usb_hub*     Pointer to the hub structure. If NULL, -EINVAL
     *                  will be returned.  
     *
     *  __u8            Port number.
     *
     * Returns:
     *
     * __s8             > 0 if ok, encoding the bus state (and with
     *                  USB_HUB_BUS_STAT_* to get the bits). <0 errno
     *                  code on error. If the hub doesn't support this
     *                  operation, it will answer with an stall and
     *                  return. The code returned will be -EOPNOTSUPP.
     */

__s8 uhci_hub_bus_status_get (ts_usb_hub* hub, __u8 port)
{    
    __s8 val;
    
    gdlf_fntrace ("uhci_hub_bus_status_get (0x%p, %u)", hub, port);
    gdlf_aassert (hub != NULL, return -EINVAL);
    gdlf_aassert (port < UHCI_PORTS, return -EINVAL);
                       
    read_lock (&hci->snode.lock);
    val = (inw (UHCI (hub->host->hci)->iobase + USBPORTSC1 + 2 * port)
        & (USBPORTSC_LS_DN | USBPORTSC_LS_DP)) >> 4;
    read_unlock (&hci->snode.lock);
    return val;
}


    /* Set any of the hub features
     *
     *  We don't support none, as there's no hw for it, so we check
     * this function can handle the requested parameters. If so, we
     * just return 0, otherwise, -ENOSYS.
     *
     * Parameters:  
     *
     *  ts_usb_hub*     Pointer to hub structure
     *
     *  e_usb_featsel   Feature
     *
     *  e_usb_req       USB request code
     *
     * Returns:
     *
     *  int             0 if ok, errno code < 0 on error.
     */

int uhci_hub_feature (ts_usb_hub* hub, e_usb_hub_featsel feature,
                      e_usb_req request)
{
    gdlf_fntrace ("uhci_hub_feature (0x%p, %u, %u)", hub, feature, request);

        /* 
         * Only hub features implemented, for the other ones, use the
         * wrappers 
         */  

    if ((request != e_usb_req_set_feature 
         && request == e_usb_req_clear_feature)
        || (feature != e_usb_hub_fes_c_local_pwr 
            && feature != e_usb_hub_fes_c_over_current))
        return -ENOSYS;

        /* 
         * We only can be requested to ack current and pwr change
         * indicators; as we don't support that, we just say okay 
         */

    return 0;
}


    /* Set/clear hub features
     *
     *  This function will set/clear hub features, but filtering only
     * those which the UHCI spec allows.
     *
     * Parameters:  
     *
     *  ts_usb_hub*     Pointer to hub structure
     *
     *  __u8            Port number
     *
     *  e_usb_port_featsel
     *
     *                  Port feature selector.
     *
     *  e_usb_req       Request (only USB_REQ_(SET|CLEAR)_FEATURE.
     *
     * Returns:
     *
     *  int             0 if ok, errno code < 0 on error.
     */

int uhci_port_feature (ts_usb_hub* hub, __u8 port, 
                       e_usb_port_featsel feature, e_usb_req request)
{
    gdlf_fntrace ("uhci_port_feature (0x%p, %u, %u, %u)", 
                  hub, port, feature, request);
    gdlf_aassert (hub != NULL, return -EINVAL);
    gdlf_aassert (port < hub->ports, return -EINVAL);
    
        /* 
         * Only hub port features implemented, for the other ones, use the
         * wrappers 
         */  

    switch (feature)
    {
            /*
             * NOYE: I can't find any valid use for this as a feature
             *       being set or cleared in the spec. 
             */

      case e_usb_port_fes_connection:
        return -ENOSYS;
        
            /*
             * Enable or disable a port 
             *
             *  The _PEDC and _CSC stuff is needed because they're R/WC
             * and we don't  want them to be set/reset.
             */

      case e_usb_port_fes_enable:
        write_lock (& UHCI (hub->host->hci)->snode.lock);
        switch (request)
        {
          case e_usb_req_clear_feature:
            hub->port[port].enable = 0;
            usb_bits16 (UHCI (hub->host->hci)->iobase + USBPORTSC1 + 2 * port,
                        ~USBPORTSC_PED & ~USBPORTSC_PEDC & ~USBPORTSC_CSC, 
                        USBPORTSC_PED | USBPORTSC_PEDC | USBPORTSC_CSC);
            break;
            
          case e_usb_req_set_feature:
            hub->port[port].enable = 1;
            usb_bits16 (UHCI (hub->host->hci)->iobase + USBPORTSC1 + 2 * port,
                        USBPORTSC_PED, 
                        USBPORTSC_PED | USBPORTSC_PEDC | USBPORTSC_CSC);
            break;
            
          default:
            write_unlock (& UHCI (hub->host->hci)->snode.lock);
            return -EINVAL;
        }
        write_unlock (& UHCI (hub->host->hci)->snode.lock);
        break;
        

            /* 
             * Resume or suspend a port 
             *
             *  We can only write the suspend bit if the host is not
             * entering the global suspend mode. That's why the if is
             * there.
             */
        
      case e_usb_port_fes_suspend:
        write_lock (& UHCI (hub->host->hci)->snode.lock);
        switch (request)
        {
          case e_usb_req_clear_feature:
            if ((inw (UHCI (hub->host->hci)->iobase + USBCMD) 
                 & USBCMD_EGSM) == 0)
                usb_bits16 
                    (UHCI (hub->host->hci)->iobase + USBPORTSC1 + 2 * port,
                     ~USBPORTSC_SUSP & ~USBPORTSC_PEDC & ~USBPORTSC_CSC, 
                     USBPORTSC_SUSP | USBPORTSC_PEDC | USBPORTSC_CSC);
            break;
            
          case e_usb_req_set_feature:
            hub->port[port].suspend = 1;
            usb_bits16 (UHCI (hub->host->hci)->iobase + USBPORTSC1 
                        + 2 * port, USBPORTSC_SUSP, 
                        USBPORTSC_SUSP | USBPORTSC_PEDC | USBPORTSC_CSC);
            break;
            
          default:
            write_unlock (& UHCI (hub->host->hci)->snode.lock);
            return -EINVAL;
        }
        write_unlock (& UHCI (hub->host->hci)->snode.lock);
        break;
        

            /* UHCI root hubs don't support overcurrent protection */

      case e_usb_port_fes_over_current:
        return -ENOSYS;

        
            /* There's an specific function for this */
        
      case e_usb_port_fes_reset:
        return -EPERM;


            /* UHCI root hubs don't support power switching */

      case e_usb_port_fes_power:
        return -ENOSYS;

        
            /* NOTE: What's this feature for? */

      case e_usb_port_fes_low_speed:
        return -ENOSYS;

            /* 
             *  Ack a connection change 
             *
             * WARNING: This is a write-one-to clear port!
             */

      case e_usb_port_fes_c_connection:
        switch (request)
        {
          case e_usb_req_clear_feature:
            write_lock (& UHCI (hub->host->hci)->snode.lock);
            usb_bits16 (UHCI (hub->host->hci)->iobase + USBPORTSC1 + 2 * port,
                        USBPORTSC_CSC, 
                        USBPORTSC_CSC | USBPORTSC_PEDC | USBPORTSC_CSC);
            write_unlock (& UHCI (hub->host->hci)->snode.lock);
            break;
            
          default:
            return -EINVAL;
        }
        break;
        
        
            /* 
             *  Ack a enable/disable change
             *
             * WARNING: This is a write-one-to clear port!
             */

      case e_usb_port_fes_c_enable:
        switch (request)
        {
          case e_usb_req_clear_feature:
            write_lock (& UHCI (hub->host->hci)->snode.lock);
            usb_bits16 (UHCI (hub->host->hci)->iobase + USBPORTSC1 + 2 * port,
                        USBPORTSC_PEDC, 
                        USBPORTSC_PEDC | USBPORTSC_PEDC | USBPORTSC_CSC);
            write_unlock (& UHCI (hub->host->hci)->snode.lock);
            break;
            
          default:
            return -EINVAL;
        }
        break;
        
        
            /* 
             *  Ack a resume/suspend change
             *
             *  This is a no-op, as clear isn't needed.
             */

      case e_usb_port_fes_c_suspend:
        break;


            /* Ack over-current, unsupported */

      case e_usb_port_fes_c_over_current:
        return -ENOSYS;


            /* Ack reset, a function exists for this */

      case e_usb_port_fes_c_reset:
        return -EPERM;
        

      default:
    
            /*
             *  We don't support the requested feature/request pair,
             * so we fail.
             */
        return -ENOSYS;
    }
    return 0;
}


    /* Return UHCI hub port status in USB format
     *
     * Parameters:  
     *
     *  ts_usb_hub*     Pointer to hub structure
     *
     *  __u8            Port number
     *
     *
     * Returns:
     *
     *  int             0 if ok, errno code < 0 on error.
     */

__s32 uhci_port_status_get (ts_usb_hub* hub, __u8 port)
{
    __u32 status;
    
    gdlf_fntrace ("uhci_port_status_get (0x%p, %u)", hub, port);
    gdlf_aassert (hub != NULL, return -EINVAL);
    gdlf_aassert (port < UHCI_PORTS, return -EINVAL);
                       
    read_lock (& UHCI (hub->host->hci)->snode.lock);
    status = 
        inw (UHCI (hub->host->hci)->iobase + USBPORTSC1 + 2 * port);
    read_unlock (& UHCI (hub->host->hci)->snode.lock);

        /*
         * Put them in their right positions as if read via a
         * request.
         */

    return
            /* USB_PORT_STAT_C_RESET -- no reset changes */
            /* USB_PORT_STAT_C_OVER_CURRENT - no overcurrent */
        ((status & USBPORTSC_RD)?   USB_PORT_STAT_C_SUSPEND : 0) | 
        ((status & USBPORTSC_PEDC)? USB_PORT_STAT_C_ENABLE : 0) | 
        ((status & USBPORTSC_CSC)?  USB_PORT_STAT_C_CONNECTION : 0) | 
        ((status & USBPORTSC_LSDA)? USB_PORT_STAT_LOW_SPEED : 0) | 
            /* always powered on */ USB_PORT_STAT_POWER |
        ((status & USBPORTSC_PR)?   USB_PORT_STAT_RESET : 0) | 
            /* USB_PORT_STAT_OVER_CURRENT - no overcurrent */ 
        ((status & USBPORTSC_SUSP)? USB_PORT_STAT_SUSPEND : 0) | 
        ((status & USBPORTSC_PED)?  USB_PORT_STAT_ENABLE : 0) | 
        ((status & USBPORTSC_CCS)?  USB_PORT_STAT_CONNECTION : 0);
}


    /* Reset an UHCI Root Hub port
     *
     *  We reset by setting to one the port reset bit on port status
     * and control register. We gotta hold it like that at least 10
     * ms (see USB Spec 1.0 11.6.2). To be sure we'll push it to 13 ms
     * [so we have better luck :]
     *
     *  But note, that according to the USB Spec, chapter 11, figure
     * 11-5, a reset must carry the port to the enabled state, so we
     * also have to assert the enable/disable bit.
     *
     *  In case we failed last time, we'll use the 'try' counter to
     * increase the assertion times.
     *
     * Parameters:  
     *
     *  ts_usb_hub*     Pointer to the hub structure. If NULL, -EINVAL
     *                  will be returned.  
     *
     *  __u8            Port number.
     *
     * Returns:
     *
     *  int             0 if ok, errno code < 0 on error.
     */

int uhci_port_reset (ts_usb_hub* hub, __u8 port)
{
    unsigned io;
    
    gdlf_fntrace ("uhci_port_reset (0x%p, %u)", hub, port);    
    gdlf_aassert (hub != NULL, return -EINVAL); 
    gdlf_aassert (port < hub->ports, return -EINVAL); 

        /* Tell the port to reset */

    write_lock (& UHCI (hub->host->hci)->snode.lock);
    io = UHCI (hub->host->hci)->iobase + USBPORTSC1 + 2 * port;
    usb_bits16 (io, USBPORTSC_PR,       /* Assert reset signaling */
                USBPORTSC_PR | USBPORTSC_PEDC | USBPORTSC_CSC);
    usb_waitms (USB_PORT_RESET_TIME * hub->port[port].try);
    usb_bits16 (io, 0,                  /* Finish reset signalling */
                USBPORTSC_PR | USBPORTSC_PEDC | USBPORTSC_CSC);
    udelay (5 * hub->port[port].try);   /* Wait for device to catch up */
    usb_bits16 (io, USBPORTSC_PED,      /* Enable port */
                USBPORTSC_PED | USBPORTSC_PEDC | USBPORTSC_CSC);
    usb_waitms (USB_PORT_AFTER_RESET_TIME * hub->port[port].try);
    write_unlock (& UHCI (hub->host->hci)->snode.lock);

    return 0;
}


    /* Free a reference to the UHCI hub driver
     *
     *  We do nothing here. When we register the root hubs, we will
     * not increase the UHCI reference count (by calling
     * uhci_reference()), because that way the module is kept at a
     * zero-reference count state _until_ there are devices and they
     * are really used.
     *
     * Parameters:  
     *
     *  ts_usb_hub*     Hub for which we want to release this
     *                  reference. 
     *
     * Returns:
     *
     *  int             0 if ok, errno code < 0 on error.
     */

int uhci_hub_free (ts_usb_hub* hub)
{
    gdlf_fntrace ("uhci_hub_free (0x%p)", hub);
    return 0;
}


    /* 
     * UHCI Hub Operations
     *
     */

const ts_usb_hub_ops uhci_hub_operations = 
{
    NULL,
    uhci_hub_descriptor_get,
    NULL,
    uhci_hub_descriptor_fill,
    uhci_hub_status_get,
    uhci_hub_bus_status_get,
    uhci_hub_feature,
    uhci_port_feature,
    uhci_port_status_get,
    uhci_port_reset,
    uhci_hub_free
};


/* Log:
 *
 * $Log: vhub.c,v $
 * Revision 1.20  1998/11/14 17:45:50  inaky
 * Moved the include files to uusbd/include/uusbd, to ease integration
 * into the Linux kernel.
 *
 * Revision 1.19  1998/11/09 21:45:56  inaky
 * Fixed to use the ts_usb_desc_head on fake hub descriptor.
 *
 * Revision 1.18  1998/10/18 15:12:57  inaky
 * Fixed the connection sequence to remove the time between reset and
 * enable to make it in the microseconds order. Actually it is 5 us
 * mutiplied by the retry count, and seems to work ok with all my
 * devices. The time it was used (around 12 ms) was causing some devices
 * to feel really confused. The CATC traffic analyzer showed that time as
 * a SUSPEND.
 *
 * Revision 1.17  1998/10/16 16:39:04  inaky
 * Fixed the problem which prevented many devices from being correctly
 * resetted. The cause was the wait-after-deasserting reset time, which
 * looks like [CATC Tracer shows it that way] was generating a SUSPEND
 * interval. This was confusing the hell out of the devices.
 *
 * Revision 1.16  1998/09/24 16:21:17  inaky
 * Moved 'hub_port' to 'port'.
 *
 * Revision 1.15  1998/09/14 20:37:44  inaky
 * Removed test code, fixed macro names.
 *
 * Revision 1.14  1998/09/02 21:23:11  inaky
 * Moved uhci.h inclusion after gdlf.
 *
 * Revision 1.13  1998/08/30 14:03:21  inaky
 * Added an attachment-try counter for the ports. Made the reset/enable
 * wait times be a direct function of the failed tries.
 *
 * Revision 1.12  1998/08/27 22:24:53  inaky
 * Moved _free to _release.
 *
 * Revision 1.11  1998/08/25 16:46:40  inaky
 * Removed uneeded FIXMEs and fixed warning tags.
 *
 * Revision 1.10  1998/08/25 16:14:13  inaky
 * Moved all the port operations to mind the R/WC bits (CSC and PEDC), by
 * always writing 0 to them when accesing the io port.
 *
 * Changed the debug stuff (assertions et al) to be shorter.
 *
 * Revision 1.9  1998/08/24 02:32:57  inaky
 * Changed name of all enumerations to be prefixed with the enumeration
 * type name.
 *
 * Revision 1.8  1998/08/19 18:49:45  inaky
 * Added debug stuff for ports and still playing with times.
 *
 * Removed reference counting on UHCI root hubs, as we want the UHCI
 * module to be kept at a zero use count until devices are used.
 *
 * Revision 1.7  1998/08/03 13:54:04  inaky
 * Added an enable-after-reset to UHCI root hub ports.
 *
 * Revision 1.6  1998/07/23 14:47:49  inaky
 * Moved ../usb-uhci.h to uhci.h
 *
 * Revision 1.5  1998/07/14 22:04:38  inaky
 * Port access is by WORDS, so it's TWO bytes (+2*port).
 *
 * Also fixed the Hub style status generation in status_get, as negating
 * something yields a lot of ones which are NOT needed :)
 *
 * Revision 1.4  1998/07/14 14:59:49  inaky
 * Fixed locking acknowledges of the whole uhci subsystem; added comments
 * to functions which lacked them. Finished usb_hub_port_feature()
 * ... praying now.
 *
 * Revision 1.3  1998/07/12 22:59:59  inaky
 * Removed the function for fake status change read schedulement, it
 * isn't needed anymore.
 *
 * Revision 1.2  1998/07/02 01:45:54  inaky
 * Added some comments on checking.
 *
 * Finished status change endpoint read schedulement, reset, descriptor
 * obtention doc and some checks, descriptor_fill, all status get,
 * feature and still missing port feature.
 *
 * Revision 1.1  1998/06/25 23:11:08  inaky
 * Virtual hub driver for UHCI root hubs.
 *
 * Revision 1.2  1997/12/27 19:54:46  inaky
 * Removed .txt file who got in by mistake and added some stuff to the
 * templates.
 *
 * Revision 1.1  1997/10/02 21:29:17  inaky
 * Remade all the directory structure to make it cleaner when the USB
 * class drivers come.
 *
 * Revision 1.1  1997/02/22 22:25:44  inaky
 * *** empty log message ***
 *
 *
 */
