
/* Universal USB Driver Stack v$Revision: 1.43 $ 
 *
 * (C) 1997 by Iaky Prez Gonzlez <inaky@peloncho.fis.ucm.es>
 *
 * Distributed under the terms of the GNU Public License Version 2
 * See files COPYRIGHT and COPYING at the root distribution directory
 *
 * $Id: usbd-cfg.c,v 1.43 1998/11/15 23:59:38 inaky Exp $
 *
 * Changelog at the bottom 
 *
 *
 *  USB Driver - Devices section
 *  -----------------------------
 *
 * This is the part of the UUSBD which takes care of device
 * configuration and unconfiguration at a USB device level.
 *
 * NOTE: Read the document `doc/Configuration.sgml'.
 *
 *  Take into account this thing: if you read this source code without
 * having a background on the terms used, you'll probably get lost
 * soon, so that's why I wrote the docs referenced above :) They're
 * not candidates to the Nobel on Literature, but they should help!
 *
 * FIXME: make sysctl/proc control over trying to configure
 *        unconfigured active configurable elements when changing
 *        configuration or loading new drivers.  
 */


    /* 
     *  Define this is if we assume that the configuration values all
     * start at one and finish at n, where n is the number of
     * configurations.
     */

#define USBD_CFG_ACTIVE_IDX


#include <uusbd/usb.h>
#include <uusbd/usbd.h>

#define GDLF_IDENT          "USBD"
#define GDLF_LOCAL          /* Define if want local settings */
#define GDLF_DEF_LOGLEVEL   9
#define GDLF_ASSERTIONS     /* want assertions (#undef if not) */
#undef GDLF_FNTRACE         /* want function traces */
#define GDLF_LOCATIONS      /* want locations in log messages */
#include <uusbd/gdlf.h>

extern ts_3slist unconfs;   /* list of unconfigured devices */


    /* Update the cfel counts (active and inactive)
     *
     *  Updates the cfel counts by scanning all the configurations,
     * interfaces and alternate settings.
     *
     *  For unloaded configurations, it will just skip them.
     *
     * WARNING! No locks are used, they must be already held (wether
     *          spin, read or write) on the device.
     *
     * Parameters:  
     *
     *  ts_usb_device*  Pointer to the device whose cfel counts
     *                  update. 
     *
     * Returns:
     *
     *  int             0 if ok, errno code < 0 on error.
     *
     * Global:
     *
     *  int usb_errno   On error, error code.  
     */

int usb_cfel_update (ts_usb_device* device)
{
    int cfels = 1, cfelu = 0;
    ts_usb_cfg *cfg, **pcfg, **top_pcfg;
    ts_usb_iface *iface, *top_iface;
    ts_usb_aset *aset, *top_aset;
    
    gdlf_fntrace ("usb_cfel_update (0x%p)", device);
    gdlf_aassert (device != NULL, return -EINVAL);
    gdlf_aassert (device->pcfg != NULL, return -EINVAL);
    
        /*
         *  If the device is aggregate dependant, then the device cfel
         * counts as active or (un)configured
         */

    if (USB_DEVICE_AG_DEP (device))
        if (!device->cfel.configured)   /* device cfel */
            cfelu++;

    pcfg = device->pcfg;
    if (device->pcfg)
    {
        top_pcfg = pcfg + device->pcfgs;
        while (pcfg < top_pcfg)          /* walk USB configurations */
        {
            cfg = *pcfg;
            if (cfg == NULL)                /* unloaded cfg? */
            {   
                pcfg++;
                continue;                   /* ... skip it */
            }                               
            iface = cfg->iface;             /* get first interface */
            gdlf_aassert (iface != NULL,    /* iface cannot be NULL */
                          cfg++; continue);
            top_iface = iface + cfg->ifaces;
            while (iface < top_iface)       /* walk interfaces */
            {
                aset = iface->aset;
                gdlf_aassert (aset != NULL, iface++; continue);
                top_aset = aset + iface->asets;
                while (aset < top_aset)         /* walk alternate settings */
                {
                    if (!aset->cfel.configured)     /* unconfigured? ... */
                        cfelu++;                    /* ... one more */
                    cfels++;                        /* for the totals */
                    aset++;
                }
                iface++;
            }
            pcfg++;
        }
    }
    
    device->cfels = cfels;          /* update */
    device->cfelu = cfelu;
    return 0;
}


    /* Update the cfel counts of configured active cfels
     *
     *  Note this function is needed because some times we just touch
     * the active configurable elements, not the inactive, and it'd be
     * quite stupid to recalculate all for it.
     *
     * WARNING! No locks are used, they must be already held (wether
     *          spin, read or write) on the device.
     *
     * Parameters:  
     *
     *  ts_usb_device*  Pointer to the device whose active cfel counts
     *                  to update
     *
     * Returns:
     *
     *  int             0 if ok, errno code < 0 on error.
     *
     * Global:
     *
     *  int usb_errno   On error, error code.  
     */

int usb_cfel_aupdate (ts_usb_device* device)
{
    int acfels = 0, acfelu = 0;
    ts_usb_iface *iface, *top_iface;
    
    gdlf_fntrace ("usb_cfel_aupdate (0x%p)", device);
    gdlf_aassert (device != NULL, return -EINVAL);
    
        /*
         *  If the device is aggregate dependant, then the device cfel
         * counts as active or (un)configured
         */

    if (USB_DEVICE_AG_DEP (device))
    {
        if (!device->cfel.configured)   /* device cfel */
            acfelu++;
        acfels++;
    }

    if (device->active && (iface = device->active->iface) != NULL)
    {
        top_iface = iface + device->active->ifaces;
        while (iface < top_iface)   /* walk ifaces on active USB cfg */
        {
            if (iface->active)
            {
                if (!iface->active->cfel.configured)
                    acfelu++;
                acfels++;
            }
            iface++;
        }                           /* interfaces */
    }                               /* active configuration */

    device->acfels = acfels;        /* update */
    device->acfelu = acfelu;

    return 0;
}


    /* Try to configure a configurable element
     *
     *  This function will take the configurable element, and based on
     * it's codes, search for the most suitable driver in the list and
     * use it to try configure the configurable element, by calling
     * it's configuration method. If the cfel is finally configured
     * (cfel->configured == 1), then the cfel counts will be updated.
     *
     *  If after the configuration, the pointers point to 'something'
     * and the cfel isn't configured (cfel->configured == 0), then it
     * will be asumed that someting went really wrong, the destructors
     * will be called and the pointers freed. Note this is a bad
     * behaviour from the drivers.
     *
     * WARNING! Device must be already locked, as no locks are done
     *          nor acknowledged.
     *
     * Parameters:  
     *  
     *  ts_usb_cfel*    Configurable element we want to configure.
     *
     *  ts_usb_device*  Pointer to the device structure to configure.
     *
     *  ts_usb_cfg*     USB configuration where the configurable
     *                  element is (NULL if it is the Device
     *                  Configurable Element).
     *
     *  t_usb_cfg       Configuration of the alternate setting to
     *                  configure (index number, only valid if
     *                  previous pointer is set; it is
     *                  device->cfg[index]).
     *
     *  ts_usb_iface*   Interface where the configurable element is
     *                  (NULL if it is the Device Configurable
     *                  Element).
     *
     *  t_usb_iface     Interface of the alternate setting to
     *                  configure (index number, only valid if
     *                  previous pointer is set; it is
     *                  &device->cfg[cfg_index].iface[index]).
     *
     *  ts_usb_aset*    Alternate setting where the configurable
     *                  element is (NULL if it is the Device
     *                  Configurable Element).
     *
     *  t_usb_aset      Alternate setting to configure (index
     *                  number, only valid if previous pointer is
     *                  set; it is 
     *                  &device->cfg[cfg_index].iface[iface_idx]
     *                  .aset[index]). 
     *
     *  __u32           USB_CFG_* flags indicating why is this
     *                  function called.  
     *
     * Returns:
     *
     *  int             0 if configured, 1 if not configured, errno
     *                  code < 0 on error. 
     *
     * Global:
     *
     *  int usb_errno   On error, error code.  
     */

#define cook_mask(d,c)                          \
((((d)->vendor? 1 : 0) << 4)                    \
    | (((d)->product? 1 : 0) << 3)              \
    | (((c)->class? 1 : 0) << 2)                \
    | (((c)->subclass? 1 : 0) << 1)             \
    | (((c)->protocol? 1 : 0)))

int usb_cfel_config (ts_usb_cfel* cfel, 
                     ts_usb_device* device, 
                     ts_usb_cfg* cfg, t_usb_cfg cfg_idx, 
                     ts_usb_iface* iface,t_usb_iface iface_idx,
                     ts_usb_aset* aset, t_usb_aset aset_idx, 
                     __u32 flags)
{
    ts_usb_driver* driver;
    int errcode;
    int vicsp, ovicsp, mask;
    int cfel_changes;           /* number of cfel data changes */
    ts_usb_cfel old;            /* keep old cfel data */

#warning DEBUG: remove me (disabled)    
#if 0
    gdlf_debug ("usb_cfel_config (0x%p, 0x%p, "
                "0x%p, %u, 0x%p, %u, 0x%p, %u, %u)\n",
                cfel, device, cfg, cfg_idx, iface, iface_idx, 
                aset, aset_idx, flags);
#endif
    gdlf_fntrace ("usb_cfel_config (0x%p, 0x%p, "
                  "0x%p, %u, 0x%p, %u, 0x%p, %u, %u)",
                  cfel, device, cfg, cfg_idx, iface, iface_idx, 
                  aset, aset_idx, flags);
    gdlf_aassert (cfel != NULL, return -EINVAL);
    gdlf_aassert (device != NULL, return -EINVAL);

    if (cfel->configured)       /* already configured? */
    {
        gdlf_debug ("Trying to configure an aset already "
                    "conifgurated; device 0x%p (%d %d), "
                    "cfel 0x%p of cfg 0x%p (%d), "
                    "iface 0x%p (%d), aset 0x%p (%d)\n",    
                    device, device->hub->host->addr, device->addr,
                    cfel, cfg, cfg_idx, iface, iface_idx,
                    aset, aset_idx);
        return 0;               /* nothing to do */
    }

        /* 
         * Get first matching driver 
         *
         *  We are going to try first the drivers already loaded. If
         * these fail, we'll try loading. When it fails configuring,
         * we just complain and keep on. If success, we update the
         * cfel counts and get out.
         */

  restart:
#warning DEBUG: remove me (disabled)
#if 0
    gdlf_debug ("Starting match ... \n");
#endif
    errcode = usb_drv_match (&driver, device->vendor, 
                             device->product, cfel, NULL);
    gdlf_aassert (errcode >= 0, return errcode);
    driver = usb_drv_reload (driver);   /* reload if unloaded */
    cfel_changes = 0;
    
    while (driver != NULL)
    {
        gdlf_aassert (driver->config != NULL, break);
        memcpy (&old, cfel, sizeof (ts_usb_cfel));  /* keep actual settings */
        errcode =                           /* configure */
            (*driver->config) (cfel, device, cfg, cfg_idx,
                               iface, iface_idx, aset, aset_idx, flags);
        if (errcode < 0)                    /* error? */
        {
            gdlf_err ("Error while trying to configure device "
                      "(%d %d) with driver `%s' (0x%p): "
                      "errno/usb %d/%d. Trying next\n",
                      device->hub->host->addr, device->addr,
                      driver->name, driver,
                      errcode, usb_errno);
        }
        if (cfel->configured)               /* configured? */
        {
            device->cfelu--;                /* ok, done */
            if (cfg == device->active && iface && iface->active == aset)
                device->acfelu--;
            return 0;
        }

            /* 
             *  Check if this driver has changed device vicsp codes,
             * if so, restart the search process.
             *
             *  We add a high-water mark to evitate deadlocks on bad
             * behaving drivers or malicious ones.
             *
             *  This code is repeated below, in the module loading
             * loop. Be sure to keep'em synchronized.
             */

        if (old.class != cfel->class
            || old.subclass != cfel->subclass
            || old.protocol != cfel->protocol)
        {
            cfel_changes++;
            if (cfel_changes > USB_CFEL_CHANGES_MAX)
            {
                gdlf_err ("Too many cfel vicsp code changes on "
                          "device 0x%p (%d %d), aborting "
                          "configuration\n", device,
                          usb_host_addr (device), 
                          usb_dev_addr (device)); 
                return -EDEADLK;    
            }
#warning DEBUG: remove me (disabled)    
#if 0
            gdlf_debug ("cfel vicsp changed, restarting configuration\n");
#endif
            goto restart;
        }

        errcode = usb_drv_match (&driver,   /* ops, search another */
                                 device->vendor, device->product, 
                                 cfel, driver);
        gdlf_aassert (errcode >= 0, return errcode);
        driver = usb_drv_reload (driver);   /* reload if unloaded */
    }

        /*
         * None ... let's try loading.
         *
         *  We'll use this mask order:
         *
         *  vicsp  1f
         *  vics0  1e
         *  vic00  1c
         *  vi000  18
         *  v0csp  17
         *  v0cs0  16
         *  v0c00  14
         *  v0000  10
         *  00csp  07
         *  00cs0  06
         *  00c00  04
         *  00000  00
         *
         *  Invalids: 0-3, 5, 8-f, 11-13, 15, 19-1b, 1d  
         */

#warning DEBUG: remove me (disabled)    
#if 0
    gdlf_debug ("Tryng to load drivers ...\n");
#endif
    vicsp = 0x1f+1;           /* start on most specific */
    mask = cook_mask (device, cfel);
    cfel_changes = 0;
    do
    {
            /* dec vicsp (next more generic) and adjust it */

        ovicsp = vicsp & mask;
        vicsp--;
        if (vicsp == 0x1d)
            vicsp = 0x1c;
        else if (vicsp >= 0x19 && vicsp <= 0x1b)
            vicsp = 0x18;
        else if (vicsp == 0x15)
            vicsp = 0x14;
        else if (vicsp >= 0x11 && vicsp <= 0x13)
            vicsp = 0x10;
        else if (vicsp >= 0x8 && vicsp <= 0xf)
            vicsp = 0x7;
        else if (vicsp == 0x5)
            vicsp = 0x4;
        else if (vicsp >= 0x0 && vicsp <= 0x3)
            break;

        if ((vicsp & mask) == ovicsp)           /* new one same as prev? */
            continue;                           /* try next */
        if (usb_drv_find (device->vendor,       /* already loaded ... */
                          device->product,      /* ... (and thus, tested)? */
                          cfel))                
            continue;
        driver = usb_drv_load (vicsp & mask,    /* load */
                               device->vendor, device->product, cfel);
        if (driver == NULL)                     /* failed? ... */
            continue;                           /* ... try next */        
        gdlf_aassert (driver->config != NULL, break);

        memcpy (&old, cfel, sizeof (ts_usb_cfel));  /* keep actual settings */
        errcode =                               /* configure */
            (*driver->config) (cfel, device, cfg, cfg_idx,
                               iface, iface_idx, aset, aset_idx, flags);
        if (errcode < 0)                        /* error? */
        {
            gdlf_err ("Error while trying to configure device "
                      "(%d %d) with driver `%s' (0x%p): "
                      "errno/usb %d/%d. Trying next\n",
                      device->hub->host->addr, device->addr,
                      driver->name, driver,
                      errcode, usb_errno);
        }
        if (cfel->configured)                   /* configured? */
        {
            device->cfelu--;
            if (cfg == device->active && iface && iface->active == aset)
                device->acfelu--;
            return 0;       /* configured, we are done */
        }

            /* 
             *  Check if this driver has changed device vicsp codes,
             * if so, restart the search process.
             *
             *  We add a high-water mark to evitate deadlocks on bad
             * behaving drivers or malicious ones.
             */

        if (old.class != cfel->class
            || old.subclass != cfel->subclass
            || old.protocol != cfel->protocol)
        {
            cfel_changes++;
            if (cfel_changes > USB_CFEL_CHANGES_MAX)
            {
                gdlf_err ("Too many cfel vicsp code changes on "
                          "device 0x%p (%d %d), aborting "
                          "configuration\n", device,
                          usb_host_addr (device), 
                          usb_dev_addr (device)); 
                return -EDEADLK;    
            }
#warning DEBUG: remove me (disabled)    
#if 0
            gdlf_debug ("cfel vicsp changed, restarting configuration\n");
#endif
            goto restart;
        }
    } while (vicsp > 0);
    return 1;   /* not configured */
}


    /* Unconfigure a configurable element
     *
     *  This function will call the destructors of non-NULL pointers
     * contained inside a configured configurable element. After this,
     * all the data must have been freed, remaining only the space
     * allocated by the structures being pointed by the pointers (but
     * it's contents must be free!). Then, a memory free (kfree())
     * will be done on those pointers. After it, the configurable
     * element state will be reverted to unconfigured and the
     * unconfigured cfel counts (acfelu, cfelu) incremented by one.
     *
     * WARNING! Device must be already locked, as no locks are done
     *          nor acknowledged.
     *
     * Parameters:  
     *
     *  ts_usb_cfel*    Configurable element we want to unconfigure.
     *
     *  ts_usb_device*  Pointer to the device structure to
     *                  where the configurable element is. 
     *
     *  ts_usb_cfg*     USB configuration where the configurable
     *                  element is (NULL if it is the Device
     *                  Configurable Element).
     *
     *  ts_usb_iface*   Interface where the configurable element is
     *                  (NULL if it is the Device Configurable
     *                  Element).
     *
     *  ts_usb_aset*    Alternate setting where the configurable
     *                  element is (NULL if it is the Device
     *                  Configurable Element).
     *
     *  __u32           USB_CFG_* flags indicating why is this
     *                  function called.  
     *
     * Returns:
     *
     *  int             0 if ok, errno code < 0 on error.
     *
     * Global:
     *
     *  int usb_errno   On error, error code.  
     */

int usb_cfel_unconfig (ts_usb_cfel* cfel, ts_usb_device* device, 
                       ts_usb_cfg* cfg, ts_usb_iface* iface,
                       ts_usb_aset* aset, __u32 flags)
{
    gdlf_fntrace ("usb_cfel_unconfig (0x%p, 0x%p, 0x%p, "
                  "0x%p, 0x%p, %u)", 
                  cfel, device, cfg, iface, aset, flags);
    gdlf_aassert (cfel != NULL, return -EINVAL);
    gdlf_aassert (device != NULL, return -EINVAL);

#warning FIXME: Check driver is loaded, so it can unconfigure

        /* 
         * If the cfel is the device cfel and it is always
         * considered configured so we cannot unconfigure it. Also if
         * it is unconfigured we'll do nothing.
         */

    if ((&device->cfel == cfel && cfel->class == 0) 
        || !cfel->configured)
        return 0;

        /* 
         * We won't check if the configure flag is set, as
         * what we've gotta do is to free all the pointers. We'll set
         * the configured flag later.  
         *
         *  Note the `if (cfel->...)'. If the device destructor is
         * smart enough [perhaps too much] to free the structure, that
         * will be signaled with a NULL pointer where we had a
         * non-null one.
         */

    if (cfel->data)
    {
        ts_dtor_exec (&cfel->data->dtor);
        if (cfel->data)
        {
            usb_kfree (cfel->data);
            cfel->data = NULL;
        }
    }

    cfel->configured = 0;   /* no longer is configured */
    
        /* if active or device configurable element ... */

    if ((cfg == device->active && iface && aset == iface->active)
        || (cfel == &device->cfel))
        device->acfelu++;   /* another active one is unconfigured */
    device->cfelu++;        /* one more is unconfigured */
    return 0;               /* done */
}


    /* Configure a device
     *
     *  Tries to configure all the unconfigured configurable
     * elements of a device, starting by the device configurable
     * element (if configurable -- aka: if aggregate dependant
     * device) and continuing with all the interfaces. If the
     * interfaces are independent (non-aggregate dependant device),
     * they'll be called to be configured separately, in increasing
     * order of index (interface 0, 1, 2 ...).
     *
     *  Note that the first time we invoke this function on a device
     * (and probably some others) the device will be fully
     * unconfigured, what means that not even a configuration will be
     * selected as `active'. In this case we'll choose a default
     * configuration using the persistence data or a best match.
     *
     *  This function will remove the device from the list of
     * unconfigured devices if it is fully configured at
     * completion. Also will update the cfel counts.
     *
     *  This function is called on device attach and when it's USB
     * configuration changes; note anyhow, that it passed the flags to
     * the configuration routines to indicate the reason of the config
     * call.
     *
     * Parameters:  
     *
     *  ts_usb_device*  Device to try to configure
     *
     *  __u32           Flags to pass to the driver routines 
     *                  (USB_CFG_*).
     *
     * Returns:
     *
     *  int             0 if ok, errno code < 0 on error.
     *                  Mind in case of error, some configurable
     *                  elements may be configured, some not ... in
     *                  any case, the cfel counts are updated.
     *
     * Global:
     *
     *  int usb_errno   On error, error code.  
     */

int usb_device_config (ts_usb_device* device, __u32 flags)
{
    int errcode = 0;
    ts_usb_iface *iface, *top_iface;
    
    gdlf_fntrace ("usb_device_config (0x%p)", device);

    gdlf_aassert (device != NULL, return -EINVAL);

    if (device->acfelu == 0)        /* doesn't need config? */
        return 0;                   /* ... so what? */

        /* 
         * Configures the device cfel if it has to [aggregate
         * dependant].  Try to configure device cfel. Note this step
         * may change the device's USB configuration.  
         */

    if (USB_DEVICE_AG_DEP (device))     /* aggregate dependant? */
    {                                   
        flags |= USB_CFG_AGTYPE;        /* (add it to the flags) */
        errcode = usb_cfel_config (&device->cfel, device, 
                                   NULL, 0, NULL, 0, 
                                   NULL, 0, flags);
    }
    else
        flags &= ~USB_CFG_AGTYPE;       /* aggregate independant */

        /* 
         * Activate an USB configuration if none active
         *
         *  Now we try to search for a default USB configuration to
         * activate if none is. What we do is select with
         * usb_configuration_best() a suitable configuration and try
         * to set it. If it fails, we warn about it and request
         * another configuration [that's what the second parameter is
         * for]. This way, is guaranteed that we are going to try
         * every possible cfg without repeating them. If we don't find
         * any, we are out of luck.
         */

    if (device->active == NULL)     /* huh? fully unconfigured? */
    {
        ts_usb_cfg *cfg;
        int cnt;

            /* 
             * Make all configurations be equally choosable for
             * usb_configuration_best(), by clearing the failed bit, as it
             * is used by it to skip the configurations that previously
             * failed to be set.
             */
        
        for (cnt = 0; cnt < device->pcfgs; cnt++)
            if (device->pcfg[cnt] != NULL)      /* on some error cases, this happens */
                device->pcfg[cnt]->failed = 0;

            /* Now seek for a nice default configuration */

        do
        {            
            cfg = usb_configuration_best (device);  /* get a cfg */
            if (cfg == NULL)                        /* no way? */
            {
                gdlf_err ("Couldn't obtain a suitable "
                          "default configuration for "
                          "device (%d %d)\n",
                          device->hub->host->addr, device->addr);
                return -ENODATA;    /* FIXME: better code? */
            }
            errcode = usb_configuration_set (device, cfg->addr, 0); /* set it */
            if (errcode < 0)
                gdlf_warning ("Couldn't set to default configuration "
                              "%d for device (%d %d): "
                              "errno/usb_errno %d/%d; trying next\n",
                              cfg->addr, 
                              device->hub->host->addr, device->addr,
                              errcode, usb_errno);
        } while (errcode);
    }     

        /* 
         * Now try to configure remaining active configurable elements
         *
         *  If a call to usb_cfel_config changes the alternate
         * setting, it should happen nothing as it has already been
         * called 
         */

    gdlf_assert (device->active->iface);

    if (device->active && (iface = device->active->iface) != NULL)
    { 
        int iface_idx = 0;
        int active_idx = 0;

            /* get active's idx */

        for (active_idx = 0; 
             active_idx < device->pcfgs
                 && device->pcfg[active_idx] != device->active; 
             active_idx++);

        gdlf_apassert (active_idx < device->pcfgs,
                       return -EINVAL, /* aqu se va a liar finaaa ... */
                       "JAAARL!!! Cannot find active's "
                       "configuration index (0x%p)\n",
                       device->active);

        top_iface = iface + device->active->ifaces;
        while (iface < top_iface)
        {
            if (iface->active && !iface->active->cfel.configured)
                usb_cfel_config (&iface->active->cfel, device, 
                                 device->active, 
                                 active_idx,
                                 iface, 
                                 iface_idx,
                                 iface->active, 
                                 iface->active - iface->aset, 
                                 flags);
            iface++;
            iface_idx++;
        }
    }

        /* 
         * On error, some cfels may be configured, some not, so, to be
         * sure, we recalculate everything.
         */

    usb_cfel_aupdate (device);      /* recalculate active cfel counts */
    if (device->acfelu == 0)        /* fully configured? remove from */
        ts_3slist_del (&unconfs, &device->msnode); /* unconfs list */    

    return errcode;
}


    /*  Unconfigures a device
     *
     *  This is accomplished by calling the unconfiguration methods of
     * each configurable element, either inactive or active (as each
     * cfel pointer is a pointer to a ts_snode and has a destructor).
     *
     *  After unconfiguring, updates the cfel counts and adds the
     * device to the unconfigured's list. 
     *
     *  When calling the destructor, this must be an special function
     * of each driver WHO MUST KNOW what's the address of the
     * ts_usb_driver structure of the driver, to decrement it's use
     * and module use count.
     *
     *  If the device is an aggregate with dependant interfaces
     * device, we'll first call the destructor on the device
     * configurable element and then on each configurable element
     * which remains unconfigured (no matter active or not). Other
     * way, we'll unconfigure directly configurable element per
     * configurable element.
     *
     * Parameters:  
     *
     *  ts_usb_device*  Device to unconfigure
     *
     *  __u32           Flags to indicate what caused the
     *                  unconfiguration and some other info (See
     *                  USB_CFG_* at usbd.h) 
     *
     * Returns:
     *
     *  int             0 if ok, errno code < 0 on error.
     *
     *                  Note that, in case of error, it is guaranteed
     *                  that *all* the configurable elements are
     *                  unconfigured, and the error code returned,
     *                  unless on parameters error, will be the error
     *                  code returned by the last failed
     *                  unconfiguration of the configurable elements.
     *
     * Global:
     *
     *  int usb_errno   On error, error code.  
     */

int usb_device_unconfig (ts_usb_device* device, __u32 flags)
{
    t_usb_cfg cnt;
    ts_usb_iface *iface, *top_iface;
    ts_usb_aset *aset, *top_aset;
    
    int errcode = 0, errcode2;
    
    gdlf_fntrace ("usb_device_unconfig (0x%p)", device);
    gdlf_aassert (device != NULL, return -EINVAL);

    if (device->cfels == device->cfelu) /* doesn't need unconfig? ... */
        return 0;                       /* ... do nothing */
    
    write_lock (&device->lock);
    
    if (device->acfelu == 0)            /* not on the unconfs list? ... */
        ts_3slist_prep (&unconfs, &device->msnode); /* ... add */

        /* 
         * dependent aggregate device; thus it has the device cfel
         * configured, and we must unconfigure it.
         */

    if (USB_DEVICE_AG_DEP(device)) 
        usb_cfel_unconfig (&device->cfel, device, NULL, 
                           NULL, NULL, flags);

        /*
         * Now we step around the interfaces which still are
         * configured after [posibly] having unconfigured the device
         * cfel.
         */

    gdlf_aassert (device->pcfg != NULL, return -ENODATA);

    if (device->pcfgs)                  /* if there are configurations */
    {
        for (cnt = 0; cnt < device->pcfgs; cnt++)   /* walk the configurations */
        {
            if (device->pcfg[cnt] == NULL)  /* unloaded cfg? ... */
                continue;                   /* ... skip it */
            
            iface = device->pcfg[cnt]->iface;   /* first inteface */
            gdlf_aassert (iface != NULL, continue); /* iface cannot be null */
            top_iface = iface 
                + device->pcfg[cnt]->ifaces;/* get top interface */
            while (iface < top_iface)       /* walk the interfaces */
            {
                aset = iface->aset;         /* first alternate setting */
                if (aset == NULL)           /* if none get next interface */
                    continue;
                top_aset = aset + iface->asets; /* get top aset */
                while (aset < top_aset)     /* walk the alternate settings */
                {
                    if (aset->cfel.configured)
 /* I */            {
 /* know */             errcode2 =              /* unconfig the aset */
 /* this looks */           usb_cfel_unconfig (&aset->cfel, device,
 /* quite */                                   device->pcfg[cnt], iface, 
 /* ugly */                                    aset, flags);
 /* ... */              if (errcode2)           /* if error ... */
                            errcode = errcode2; /* ... keep last only */
                    }
                    aset++;
                }
                iface++;
            }               /* ifaces while loop */
        }                   /* for loop */
    }
    
    write_unlock (&device->lock);
    return errcode;                     /* return last error */
}


    /* FIXME: Detailed desc
     *
     *  Call with locks held!
     *
     * Parameters:  
     *
     *  type            descr
     *
     * Global:
     *
     *  type name       descr
     *
     * Returns:
     *
     *  type            descr
     *
     * Global:
     *
     *  type name       descr
     */

void usb_cfg_pendp_unset (ts_usb_device* device)
{
    t_usb_iface ifacen, top;

    gdlf_fntrace ("usb_cfg_pendp_unset (0x%p)", device);
    gdlf_aassert (device != NULL, return);
    gdlf_aassert (device->active != NULL, return);

    for (ifacen = 0, top = device->active->ifaces; ifacen < top; ifacen++)
        usb_iface_pendp_unset (device->active, ifacen);
    return;
}


    /* FIXME: Detailed desc
     *
     *  Call with locks held!
     *
     * Parameters:  
     *
     *  type            descr
     *
     * Global:
     *
     *  type name       descr
     *
     * Returns:
     *
     *  type            descr
     *
     * Global:
     *
     *  type name       descr
     */

int usb_cfg_pendp_set (ts_usb_device* device)
{
    int errcode = 0;
    t_usb_iface ifacen, top;
    gdlf_fntrace ("usb_cfg_pendp_set (0x%p)", device);
    gdlf_aassert (device != NULL, return -EINVAL);
    gdlf_aassert (device->active != NULL, return -EINVAL);

    for (ifacen = 0, top = device->active->ifaces; 
         errcode == 0 && ifacen < top; ifacen++)
        errcode = usb_iface_pendp_set (device->active, ifacen);
    return errcode;
}


    /* FIXME: Detailed desc
     *
     * Compliance:      USB Spec 1.0 xxxx
     *
     * Parameters:  
     *
     *  type            descr
     *
     * Global:
     *
     *  type name       descr
     *
     * Returns:
     *
     *  type            descr
     *
     * Global:
     *
     *  type name       descr
     */

int usb_cfg_pendps_resize (ts_usb_cfg* cfg, t_usb_endp pendps)
{
    ts_usb_endp** pendp;
    size_t nsize;

    gdlf_fntrace ("usb_cfg_pendps_resize (0x%p, %d)", cfg, pendps);
    gdlf_aassert (cfg != NULL, return -EINVAL);
    
      /* calc next power-of-two which is ok */

    nsize = cfg->pendps != 0? 2 * cfg->pendps : 2;
    while (nsize <= pendps)
        nsize *= 2;
    pendp = usb_realloc (cfg->pendp, cfg->pendps * sizeof (ts_usb_endp*), 
                         nsize * sizeof (ts_usb_endp*), GFP_KERNEL);
    if (pendp == NULL)
    {
        gdlf_err ("Couldn't rellocate endpoints array "
                  "(from %u to %u), cfg 0x%p\n",
                  cfg->pendps, nsize, cfg);
        return -ENOMEM;
    }
    
        /* clean upper half of the array */

    memset (pendp + cfg->pendps,
            0, (cfg->pendps? cfg->pendps : nsize) * sizeof (ts_usb_endp*));
    cfg->pendp = pendp;
    cfg->pendps = nsize;
    return 0;
}


    /* Set the active configuration on a device
     *
     * Check nobody uses the configuration and then change. Actualize
     * ts_usb_device->active, try to configure unconfigured acfels.
     *
     * Parameters:  
     *
     *  ts_usb_device*  Pointer to the device structure whose
     *                  configuration we want to set.
     *
     *  t_usb_cfg       Configuration number to set to (the one in
     *                  cfg->addr), not the order number in the
     *                  array. Use 0 to let the device fully
     *                  unconfigured [without any active USB
     *                  configuration].
     *
     *  int do_config   0 if you don't want this function to
     *                  automagically call the device configuration
     *                  process (usb_device_config()) for the new USB
     *                  configuration. != if you want it.
     *
     * Returns:
     *
     *  int             0 if ok, errno code < 0 on error.
     *
     * Global:
     *
     *  int usb_errno   On error, error code.  
     */

int usb_configuration_set (ts_usb_device* device, t_usb_cfg cfgn,
                           int do_config)
{
    int errcode;
    unsigned cnt;
    ts_usb_cfg *cfg, *old_cfg;
    char in_unconfs;
    ts_usb_request request;
    
    gdlf_fntrace ("usb_configuration_set (0x%p, %u, %d)", 
                  device, cfgn, do_config);
    gdlf_aassert (device != NULL, return -EINVAL);

#warning FIXME: Reimplement using upgradeable read locks

    write_lock (&device->msnode.lock);
    
    old_cfg = device->active;

        /* Search the pointer for the new configuration */

    if (cfgn != 0)              /* if not wanting to fully unconfigure */
    {
        cnt = 0;
        while (cnt < device->pcfgs
               && (cfg = device->pcfg[cnt]) != NULL 
               && cfg->addr != cfgn)
            cnt++;
        if (cnt >= device->pcfgs)
        {
#warning FIXME: reload the configuration tree with usb_descriptor_cfg_fill()
            gdlf_err ("Invalid or not available USB configuration "
                      "number %d for device (%u %u)\n", 
                      cfgn, device->hub->host->addr,
                      device->addr);
            write_unlock (&device->msnode.lock);
            usb_errno = -USBE_CFG_BAD_INDEX;
            return -EINVAL;
        }
        cfg->failed = 1;        /* by default, it failed */
    }
    else
        cfg = NULL;
    
        /* 
         *  Search all the endpoints, and if any's got an open
         * descriptor on it, we can't change.
         *
         *  First we check endpoint 0.
         */

    if (device->endp0.desc 
        && device->endp0.desc->list.elements != 0)
    {
        write_unlock (&device->msnode.lock);
        return -EBUSY;
    }

    if (old_cfg)
    {
        for (cnt = 0; cnt < old_cfg->pendps; cnt++)
            if (old_cfg->pendp[cnt] && old_cfg->pendp[cnt]->desc != NULL)
            {
                write_unlock (&device->msnode.lock);
#warning DEBUG: remove me
                gdlf_assert (old_cfg->pendp[cnt] && old_cfg->pendp[cnt]->desc == NULL);
                return -EBUSY;
            }

            /* Check how many people is using the active configuration */

#warning FIXME: stablish the patterns for usage of this, they are not clear

        if (atomic_read (&device->active->usecnt) != 1) /* ourselves and? */
        {
            write_unlock (&device->msnode.lock);
            return -EBUSY;
        }
    }
    
        /* Check new configuration constraints */

    if (cfg)
    {    
#warning FIXME: Check there is enough bandwidth -(ENOSR,EDQUOT)
#warning FIXME: Check the power -ENOMEDIUM, taking into account hub power    
    }
    
        /* Ok, everything is fine, let's try to change the USB configuration */
    
    request.bmreqtype = 0;
    request.request = e_usb_req_set_configuration;
    request.value = cfgn;   /* cfg == 0 means unconfigured state */
    request.index = 0;      /* see USB Spec 1.0 + clarifications 9.4.7 */
    
    errcode = usb_request_dataless (device, &request, 
                                    "configuration change");
#warning FIXME: && cfgn is a hack, it should not fail!
    if (errcode < 0 && cfgn)
    {
        gdlf_err ("Cannot change configuration; request failed: "
                  "errno/usb %d/%d\n", errcode, usb_errno);
        write_unlock (&device->msnode.lock);
        return errcode;
    }
    
        /* Ok, now we can change the configuration in soft */

    in_unconfs = device->acfelu != 0;    /* is it on the unconfs list? */
    
    if (device->active)
    {
        usb_cfg_pendp_unset (device);     /* update device->active->pendp */
        usb_kfree (device->active->pendp);
        device->active->pendp = NULL;
        device->active->pendps = 0;
        usb_configuration_unref (device->active);
    }
    device->active = cfg;
    if (cfg)
    {
        usb_configuration_ref (device->active);   /* reference it */

            /*
             * Now we reconfigure
             *
             *  Note we have to do the full reconfiguration (device
             * level included), so that the posible drivers used as
             * hooks to kill quirks at device level are called and do
             * their settings on the device data structures.
             */

        cfg->failed = 0;                /* we know it changed ok */
        usb_cfel_aupdate (device);      /* actualize active cfel counts */
        usb_cfg_pendp_set (device);     /* update device->active->pendp */
        if (do_config                   /* shall we configure it? */
            && device->acfelu != 0)     /* and partially or fully unconfigured? */
        {                               /* partially */
            if (!in_unconfs)            /* previously not on the unconf list? */
                ts_3slist_prep (&unconfs, &device->msnode);
            usb_device_config (device, USB_CFG_CHANGE);
        }
    }
    write_unlock (&device->msnode.lock);
#warning FIXME: Tell the persistence system this new USB configuration
    return 0;
}


    /* Search a suitable default USB configuration
     *
     *  When a device is initially plugged in or resetted, it won't
     * even have an active USB configuration, and the configuration
     * routine has to provide one.
     *
     *  The ideal thing would be to let the user choose it's best
     * configuration, but we want automation. What we do is to choose
     * a configuration from the following sources:
     *
     *  - Persistence database [currently unimplemented]
     *
     *  - Default UUSBD Configuration Preferences variable
     *    (usb_cfg_prefs)
     *
     *  Then the user may, interactively configure the device to it's
     * likeliness. This configuration is then stored in the
     * persistence system, which keeps [at options] it's databases
     * between reboots. Next time a default configuration is requested
     * it is retrieved from the database [if available] or recomputed
     * from the usb_cfg_prefs variable.
     *
     *  What this function does, then, is to seek the first default
     * configuration for the given device. As topologies may change,
     * it may happen that a previously valid configuration doesn't
     * work on other moment, and thus is necessary to search for
     * another. This is what the 'failed' member of the ts_usb_cfg
     * structure is used for. When a configuration failed to be used,
     * it is set, and this function ignores it when searching a new
     * suitable configuration.
     *
     *  That's one of the reasons you should clear the failed bits on
     * all the configurations before running the search loop with this
     * function.
     *
     * Parameters:  
     *
     *  ts_usb_device*  Pointer to the device structure for which we
     *                  want to get a suitable configuration.
     *
     * Global:
     *
     *  ts_usb_cfg_prefs usb_cfg_prefs 
     *
     *                  Global configuration preferences.
     *
     * Returns:
     *
     *  ts_usb_cfg*     Pointer to new recommended configuration. NULL
     *                  if none found [rare ... should only happen if
     *                  you fiddle too much with the usb_cfg_prefs
     *                  variable and make it too restrictive, and even
     *                  so ... it'd always return one].
     *
     * Global:
     *
     *  int usb_errno On error, error code.  */

ts_usb_cfg* usb_configuration_best (ts_usb_device* device)
{
    t_usb_cfg cnt;
    
    gdlf_fntrace ("usb_configuration_best (0x%p)", device);

#warning FIXME: Persistence not implemented
#warning FIXME: Match on usb_cfg_prefs not implemented

        /*
         * Ok, this is a quick hack. We just run sequentially on the
         * configurations and return the first one which hasn't
         * previously failed. Gotta make the real ones :)
         *
         *  Under some cases, device->cfg[cnt] is NULL (on errors
         * ... that crawl around the call chain :)
         */

#warning FIXME: Crappy algorithm used :-P

    cnt = 0;
    while (device->pcfg[cnt] && device->pcfg[cnt]->failed)
    {
        cnt++;
        if (cnt >= device->pcfgs)
            return NULL;
    }
    return device->pcfg[cnt];
}

/* Log:
 *
 * $Log: usbd-cfg.c,v $
 * Revision 1.43  1998/11/15 23:59:38  inaky
 * Fixed some of the configuration and driver woes, as there was
 * problems, mainly caused by the ts_slist_move() function. They seem
 * fixed now. Added a code to skip when loading already loaded
 * modules.
 *
 * Revision 1.42  1998/11/14 17:15:07  inaky
 * Moved the include files to uusbd/include/uusbd, to ease integration
 * into the Linux kernel.
 *
 * Revision 1.41  1998/11/09 21:26:52  inaky
 * Changed the name of the configuration pointer array to pcfg.
 *
 * Revision 1.40  1998/10/31 02:11:04  inaky
 * Fixed the cfels counting function (usb_cfel_update()). When the
 * device->cfel was not 00.00.00, it was counting the device cfel twice,
 * as it initially is set as one.
 *
 * Removed all the d_* pointers from ts_usb_cfel and collapsed them into
 * a single pointer; only one was really used.
 *
 * Revision 1.39  1998/10/21 23:59:59  inaky
 * Removed assertions.
 *
 * Revision 1.38  1998/09/24 15:56:48  inaky
 * Moved the configuration of the device cfel _before_ selecting a
 * configuration [if none active], so that the device cfel configurator
 * may take charge of what it needs.
 *
 * Revision 1.37  1998/09/14 05:15:58  inaky
 * Added checks for errors on usb_cfg_pendp_set(), a dynamic resize
 * function for the pendps array and freeing of it when the configuration
 * is unset.
 *
 * Revision 1.36  1998/08/30 13:49:51  inaky
 * Added two methods to actualize the endpoint pointers on a cfg
 * structure.
 *
 * Revision 1.35  1998/08/28 22:31:30  inaky
 * Fixed the driver loading stuff, moving it from the mess in usbd-drv.c
 * to usb_cfel_config, doing it in two loops. One loops is for the list
 * of drivers, the other one, to go loading them.
 *
 * Revision 1.34  1998/08/27 22:15:21  inaky
 * Renamed driver_vicsp() to usb_drv_vicsp(), as now it is public.
 *
 * Revision 1.33  1998/08/25 04:02:13  inaky
 * Fixed usb_drv_match() to take a vicsp mask. This helps when doing
 * module load scan.
 *
 * Revision 1.32  1998/08/24 15:47:23  inaky
 * Fixed little cosmetic stuff, removed old files.
 *
 * Revision 1.31  1998/08/24 02:23:42  inaky
 * Done the little debugging to make the functions work as they are
 * expected to.
 *
 * Revision 1.30  1998/08/03 13:30:18  inaky
 * Moved ts_usb_tr.tr -> ts_usb_tr.list
 *
 * Revision 1.29  1998/07/16 14:09:39  inaky
 * Moved from the old read/write interface to the new, request-based
 * interface. Fixed some locks.
 *
 * Revision 1.28  1998/07/14 21:57:07  inaky
 * Fixed assertion, it was the other way around.
 *
 * Revision 1.27  1998/07/12 22:51:49  inaky
 * Fixed the aggregate type configuration code, it was swapped.
 *
 * Revision 1.26  1998/04/26 03:33:21  inaky
 * Generally fixed to the new configurations array. Fixed docs
 * [prettyprinting].
 *
 * Added check code in usb_device_config() for the case the
 * configuration's index numbers aren't zero-based sequentials.
 *
 * Revision 1.25  1998/04/13 22:08:06  inaky
 * Added the position parameters to the configuration routines.
 *
 * Fixed return value for usb_cfel_config().
 *
 * Fixed flags arguments to *_config() and *_unconfig().
 *
 * Fixed the unconfiguration code to allow more flexibility in structure
 * freeing to the drivers.
 *
 * Fixed the usb_device_config() function to make a configuration active
 * on startup if none is.
 *
 * Added posibility to fully unconfigure the device using
 * usb_configuration_set().
 *
 * Added a function to search for the most suitable configuration, which
 * is used to select initial configuration.
 *
 * Revision 1.24  1998/03/28 19:30:36  inaky
 * Fixed to work with callbacks and new transfer lists.
 *
 * Revision 1.23  1998/03/02 22:22:55  inaky
 * Checking in for administrative purposes
 *
 * Revision 1.22  1998/02/12 16:35:09  inaky
 * Finished the device ops, and moved synch_frame getting to usbd-endp,
 * where it belongs. Fixed some stuff here and there, and starting to get
 * worried on locks ... I'm gonna have to debug *lots* of deadlocks.
 *
 * Revision 1.21  1998/02/12 14:33:07  inaky
 * Added the code to handle the case of drivers registered-but-not loaded
 * and force it's reload using usb_drv_reload().
 *
 * Revision 1.20  1998/02/10 17:30:42  inaky
 * Fixed a crap using atomic_read()
 *
 * Revision 1.19  1998/02/09 13:27:27  inaky
 * Moved all the use counts to be atomic
 *
 * Revision 1.18  1998/01/11 15:44:17  inaky
 * Finished functions from the descriptor, inteface and endpoint modules,
 * fixed docs, pretty printed and removed craps.
 *
 * Revision 1.17  1998/01/08 12:19:11  inaky
 * Added copyright strings to all important files
 *
 * Revision 1.16  1998/01/08 11:49:12  inaky
 * Finishing the configuration system. Only the final cfel-stuff is
 * missing to implement; however, nothing is tested.
 *
 * Revision 1.15  1998/01/02 20:51:55  inaky
 * Finished adding the control endpoints check support and reformatting
 * usb.h to be shorter than 80 columns in most places (except the errno
 * codes).
 *
 * Revision 1.14  1998/01/02 19:39:37  inaky
 * Added the definition for endpoint zero to the device structure and
 * removed it from the list of endpoints kept at each configuration. Also
 * updated it on places where needed and removed the concept of basic
 * configuration, as is no longer needed.
 *
 * Revision 1.13  1998/01/02 13:31:22  inaky
 * Moving the docs to a separate SGML document
 *
 * Revision 1.12  1998/01/02 00:12:17  inaky
 * Finished the new driver scheme. Documented at usbd-drv.c
 *
 * Revision 1.11  1997/12/31 04:58:51  inaky
 * Stablished the proper return codes for the config and unconfig
 * functions.
 *
 * Revision 1.10  1997/12/30 20:15:44  inaky
 * Moved everything to the new lists scheme. Untested yet.
 *
 * Revision 1.9  1997/12/29 00:25:38  inaky
 * Fixed unimplemented functions to return -ENOSYS and moving the khubdd
 * to use signals as commands.
 * Also added function traces on most files that had them disabled (donkey!)
 *
 * Revision 1.8  1997/12/27 20:06:28  inaky
 * Finished fixinf the __KERNEL__ #ifdef wrapping and moved the ts_dtor.h
 * to usb-dtor.h
 *
 * Revision 1.7  1997/12/27 19:54:21  inaky
 * Added all the prototypes and skeleton definitions (incomplete) of all
 * of the USBD functions to each source file.
 *
 * Revision 1.6  1997/12/27 17:25:35  inaky
 * Finished moving to GDLF. Deprecated usb_printk and __LOCATION__
 *
 * Revision 1.5  1997/12/27 02:07:03  inaky
 * Moving everything to use the GDLF for logging and debug checkings.
 * Removed the global array of USB hosts and moved it to a linked
 * list. Added an address field to each USB host controller.
 * Removed lots of cruft that were accumulated and useless. Still some
 * missing, but may be useful someday.
 *
 * Revision 1.4  1997/12/27 02:06:00  inaky
 * Moving everything to use the GDLF for logging and debug checkings.
 * Removed the global array of USB hosts and moved it to a linked
 * list. Added an address field to each USB host controller.
 * Removed lots of cruft that were accumulated and useless. Still some
 * missing, but may be useful someday.
 *
 * Revision 1.3  1997/12/13 20:56:32  inaky
 * Moving all the documentation to the .c files except the ones for data
 * structures and inline functions. Helps making the header files cleaner.
 *
 * Revision 1.2  1997/12/08 02:41:01  inaky
 * Fixed some stuff of the #includes, modified the makefile targets to be
 * more flexible, tried to finish the device configuration section.
 *
 * Revision 1.1  1997/12/06 01:55:48  inaky
 * Now finishing the configuration system
 *
 */

