
/* Universal USB Driver Stack v$Revision: 1.23 $ 
 *
 * (C) 1997 by Iaky Prez Gonzlez <inaky@peloncho.fis.ucm.es>
 *
 * Distributed under the terms of the GNU Public License Version 2
 * See files COPYRIGHT and COPYING at the root distribution directory
 *
 * $Id: usbd-descr.c,v 1.23 1998/11/14 17:15:08 inaky Exp $
 *
 * Changelog at the bottom 
 *
 *
 *  USB Driver - Devices section - USB Descriptor operations
 *  --------------------------------------------------------
 *
 *  Standard and common operations on USB descriptors.
 *
 *  A descriptor block is a set of descriptors:
 *
 *  - Configuration descriptor block [CDB]:
 *
 *      Set of descriptors composed of a configuration descriptor,
 *    possibly one or more device/vendor specific descriptors and 
 *    one or more interface descriptor blocks [IDB].
 *
 *  - Interface descriptor block [idb]:
 *
 *      Set of descriptors composed of an interface/alternate setting
 *    descriptor, possibly one or more device/vendor specific
 *    descriptors and normally one or more endpoint descriptors.
 *
 *  As stated by the USB Spec 1.0 9.6.3, a configuration descriptor
 * block will return all the descriptors for the configuration,
 * followed by interface descriptors, starting with alternate setting
 * zero and following alternate settings of the same interface
 * ... etc. See it there, it's explained much better :)
 */

#include <uusbd/usb.h>
#include <uusbd/usbd.h>
#include <uusbd/usb-stddesc.h>
#include <uusbd/usb-misc.h>

#define GDLF_IDENT          "USBD"
#define GDLF_LOCAL          /* Define if want local settings */
#define GDLF_DEF_LOGLEVEL   9
#define GDLF_ASSERTIONS     /* want assertions (#undef if not) */
#undef GDLF_FNTRACE        /* want function traces */
#define GDLF_LOCATIONS      /* want locations in log messages */
#include <uusbd/gdlf.h>


    /* Get a USB descriptor
     *
     *  Does a request to get a descriptor, as given indicated in the
     * parameters. If there's more data than space, only the first
     * fitting bytes are written. If the inverse, the rest of the
     * bytes are left untouched.
     *
     *  Mind this function only gets device descriptors, so, for other
     * kind of USB descriptors (with a bmRequestType different to
     * 0x80), you need to make your own requests.
     *
     * Parameters:  
     *
     *  ts_usb_device*  Pointer to the device structure.
     *
     *  e_usb_desc      e_usb_desc_* which describes which descriptor do
     *                  we want.
     *
     *  __u8            Descriptor index number.
     *
     *  __u16           Language identification (valid for string
     *                  descriptors only, set to zero for others).
     *
     *  void*           Pointer to the data buffer.
     *
     *  size_t          Size of the data buffer.
     *
     * Returns:
     *
     *  ssize_t         errno code < 0 on error. If the error
     *                  code is -EREMOTEIO (endpoint stall), the
     *                  device doesn't support getting that
     *                  descriptor. If no error, it returns the amount
     *                  of data readen, without counting the data used
     *                  for the request (just the descriptor
     *                  data). Take into account, that according to
     *                  USB Spec 1.0 9.5 ('Descriptors'), if the
     *                  length is less than expected, the descriptor
     *                  is considered invalid. If more than expected
     *                  (the length field is greater), the extra bytes
     *                  must be ignored, but this greater length must
     *                  be used to calculate next descriptor's
     *                  position.
     *
     * Global:
     *
     *  int usb_errno   On error, error code.     
     */

ssize_t usb_descriptor_get (ts_usb_device* device, e_usb_desc descr, 
                            __u8 index, __u16 langid, 
                            void* data, size_t size)
{
    int errcode;
    ts_usb_request req;
        
    gdlf_fntrace ("usb_descriptor_get (0x%p, %d, %u, %u, 0x%p, %u)", 
                  device, descr, index, langid, data, size);

    req.bmreqtype = 0x80;
    req.request = e_usb_req_get_descriptor;
    req.value = ((descr & 0xff) << 8) | index;
    req.index = langid;
    req.length = size;
    
    errcode = usb_request_data (device, &req, data, "get descriptor");
    if (errcode < 0)
        gdlf_err ("Cannot get descriptor; request failed: "
                  "errno/usb %d/%d\n", errcode, usb_errno);
    return errcode;
}


    /* Set an USB descriptor
     *
     *  This function will, on DEVICES which support it, set a given
     * USB descriptor.
     *
     *  Mind this function only sets DEVICE descriptors, so, for other
     * kind of USB descriptors (with a bmRequestType different to
     * 0x0), you need to make your own requests.
     *
     * Parameters:  
     *
     *  ts_usb_device*  Pointer to the device structure.
     *
     *  e_usb_desc      e_usb_descr_* which describes which descriptor do
     *                  we want to set.
     *
     *  __u8            Descriptor index number.
     *
     *  __u16           Language identification (valid for string
     *                  descriptors only, set to zero for others).
     *
     *  const void*     Pointer to the data buffer with the new
     *                  descriptor data.
     *
     *  size_t          Size of the data buffer.
     *
     * Returns:
     *
     *  ssize_t         errno code < 0 on error. If the error
     *                  code is -EREMOTEIO (endpoint stall), the
     *                  device doesn't support setting that
     *                  descriptor. If no error, it returns the amount
     *                  of data readen, without counting the data used
     *                  for the request (just the descriptor data). 
     *
     * Global:
     *
     *  int usb_errno   On error, error code.     
     */

ssize_t usb_descriptor_set (ts_usb_device* device, e_usb_desc descr, 
                            __u8 index, __u16 langid, 
                            const void* data, size_t size)
{
    int errcode;
    ts_usb_request req;
        
    gdlf_fntrace ("usb_descriptor_set (0x%p, %d, %u, %u, %u, 0x%p)", 
                  device, descr, index, langid, size, data);

    req.bmreqtype = 0x0;
    req.request = e_usb_req_set_descriptor;
    req.value = ((descr & 0xff) << 8) | index;
    req.index = langid;
    req.length = size;
    
    errcode = usb_request_data (device, &req, (void*)data, "set descriptor");
    if (errcode < 0)
    {
        gdlf_err ("Cannot set descriptor; request failed: "
                  "errno/usb %d/%d\n", errcode, usb_errno);
        write_unlock (&device->msnode.lock);
        return errcode;
    }
    return errcode;
}


#define piface(a) ((ts_usb_iface_descr*)(a))
#define pendp(a) ((ts_usb_endp_descr*)(a))
#define phead(a) ((ts_usb_desc_head*)(a))


    /* Fill in an endpoint structure from it's descriptor data
     *
     * Parameters:  
     *
     *  ts_usb_endp*    Pointer to the endpoint structure to fill in.
     *
     *  __u8*           Pointer to where the endpoint descriptor
     *                  begins inside an interface descriptor block.  
     *
     *  size_t          Size of the buffer given in previous
     *                  parameter. 
     *
     * Returns:
     *
     *  ssize_t         errno code < 0 on error, size of endpoint and
     *                  device/vendor specific descriptors.
     *
     * Global:
     *
     *  int usb_errno   On error, error code.  
     */

ssize_t usb_descriptor_endp_fill (ts_usb_endp* endp, 
                                  __u8* buffer, size_t buflength)
{
    size_t cumlen = 0;      /* accumulative length */

    gdlf_fntrace ("usb_descriptor_endp_fill (0x%p, 0x%p, %u)",
                  endp, buffer, buflength);

    gdlf_aassert (endp != NULL, return -EINVAL);
    gdlf_aassert (buffer != NULL, return -EINVAL);
    gdlf_assert (buflength >= USB_ENDP_DESCR_LEN_1_0);

    if (phead (buffer)->bDescriptorType != e_usb_desc_endp)
    {
        gdlf_err ("Pretended to process a non-endpoint "
                  "descriptor [endp 0x%p, buffer 0x%p, size %u]\n",
                  endp, buffer, buflength);
        usb_errno = -USBE_ENDP_DESCR_BAD_TYPE;
        return -ENODATA;
    }   

        /* Fill in endpoint data */

    endp->addr = USB_DESC_ENDPNO2ADDR (pendp(buffer)->bEndpointAddress);
    endp->type = pendp(buffer)->bmAttributes & 0x3;
    switch (endp->type)
    { 
      case e_usb_tr_type_cntl:
        endp->addr &= 0xFE;     /* clean direction bit for cntl */
        break;
        
      case e_usb_tr_type_intr:
        endp->addr |= 0x01;     /* set direction bit for cntl */
        break;

      case e_usb_tr_type_isoc:
      case e_usb_tr_type_bulk:
      case e_usb_tr_type_inv:
        break;

      default:
        gdlf_passert (0, "Curious enpoint type %d\n", endp->type);
    }
    endp->maxpktsize = le16_to_cpu (pendp(buffer)->wMaxPacketSize);
    endp->interval = pendp(buffer)->bInterval;

        /*
         * Skip vendor/device specific descriptors and
         * position into next endpoint or endpoint descriptor 
         */

    cumlen += phead (buffer)->bLength;  /* One for our descriptor */
    while (cumlen + 2 < buflength       /* bDescriptorType is at cumlen+2 */
           && phead (buffer+cumlen)->bDescriptorType != e_usb_desc_endp 
           && phead (buffer+cumlen)->bDescriptorType != e_usb_desc_iface)
        cumlen += phead (buffer+cumlen)->bLength;

    return cumlen;
}



    /* Fill an alternate setting structure from it's descriptor
     * information
     *
     * Parameters:  
     *
     *  const ts_usb_iface*   
     *
     *                  Pointer to interface structure which owns this
     *                  alternate setting [unmodified].
     *
     *  ts_usb_aset*    Pointer to the alternate setting structure to
     *                  fill in.
     *
     *  __u8*           Pointer to where the interface descriptor
     *                  block begins inside a configuration descriptor
     *                  block. 
     *
     *  size_t          Size of the buffer given in previous
     *                  parameter. 
     *
     * Returns:
     *
     *  ssize_t         errno code < 0 on error, size of all interface
     *                  [alternate setting], endpoint and
     *                  vendor/device specific descriptors.
     *
     * Global:
     *
     *  int usb_errno   On error, error code.  
     */

ssize_t usb_descriptor_aset_fill (ts_usb_device* device,
                                  const ts_usb_iface* iface, 
                                  ts_usb_aset* aset,
                                   __u8* buffer, size_t buflength)
{
    size_t cumlen;              /* accumulative length */
    ssize_t length;             /* processed length */
    ts_usb_endp *endp, *top;    /* endpoint iterator and top */

    gdlf_fntrace ("usb_descriptor_aset_fill (0x%p, 0x%p, 0x%p, %u)",
                  iface, aset, buffer, buflength);

    gdlf_aassert (iface != NULL, return -EINVAL);
    gdlf_aassert (buffer != NULL, return -EINVAL);
    gdlf_aassert (buflength >= USB_IFACE_DESCR_LEN_1_0, return -EINVAL);

    if (phead (buffer)->bDescriptorType != e_usb_desc_iface)
    {
        gdlf_err ("Pretended to process a non-interface "
                  "descriptor for device 0x%p (%d %d) "
                  "[iface 0x%p, aset 0x%p, buffer "
                  "0x%p, size %u]\n",
                  device, device->hub->host->addr, device->addr, 
                  iface, aset, buffer, buflength);
        usb_errno = -USBE_IFACE_DESCR_BAD_TYPE;
        return -ENODATA;
    }   

        /*
         *  So we allocate the endpoints array, fill in the data and
         * fill in endpoints data. If this iface uses no endpoints, it
         * means it only uses endpoint 0.
         */

    aset->endps = piface(buffer)->bNumEndpoints;
    if (aset->endps)
    {
        aset->endp = ts_usb_endp_a_init (aset->endps);
        if (aset->endp == NULL)
        {
            aset->endps = 0;
            usb_errno = -USBE_NO_MEMORY;
            return -ENOMEM;
        }
    }
    
        /* Fill in rest of the data */

    aset->addr = piface(buffer)->bAlternateSetting;
    if (aset->addr == iface->asets) /* undo the asets number hack */
        aset->addr = 0;
    aset->cfel.class = piface(buffer)->bInterfaceClass;
    aset->cfel.subclass = piface(buffer)->bInterfaceSubClass;
    aset->cfel.protocol = piface(buffer)->bInterfaceProtocol;
    aset->descr = usb_string_get (device, piface(buffer)->iInterface, -1);

    gdlf_debug ("Processing interface %d, alternate setting %d%s%s\n",
                iface->addr, aset->addr, aset->descr? ": " : "",  
                aset->descr? aset->descr : "");

        /* Now scan the endpoint's data (if it exists) */

    endp = aset->endp;
    top = endp + aset->endps;
    cumlen = 0;
    cumlen += phead (buffer)->bLength;  /* one for our descriptor */

    while (endp < top)                  /* fill in all endps */
    {

            /* Skip to next endpoint descriptor */

        while (cumlen+2 < buflength     /* bDescriptorType is at +2 */
               && phead (buffer+cumlen)->bDescriptorType != e_usb_desc_endp
               && phead (buffer+cumlen)->bDescriptorType != e_usb_desc_iface)
            cumlen += phead (buffer+cumlen)->bLength;

            /* Fill in endpoint data */

        length = usb_descriptor_endp_fill (endp, buffer + cumlen,
                                           buflength - cumlen);
        if (length < 0)                 /* error? cleanup */
        {
            ts_usb_endp_a_release (aset->endp, aset->endps);
            usb_kfree (aset->endp);
            ts_usb_aset_clean (aset);
            return length;
        }
        endp++;                         /* fetch next */
        cumlen += length;
    }

        /*
         * Skip vendor/device specific descriptors and
         * position into next interface descriptor 
         */

    while (cumlen + 2 < buflength           /* bDescriptorType at cumlen+2 */
           && phead (buffer+cumlen)->bDescriptorType != e_usb_desc_iface)
        cumlen += phead (buffer+cumlen)->bLength;

    return cumlen;
}


    /* Fill an interface structure from it's descriptor information
     *
     *  This function will scan an interface descriptor and fill in
     * the interface structure given.
     *
     *  Note the interface descriptor block must have already been
     * cooked by a call to usb_descriptor_cfg_cook(). Note after going
     * trough this function, it'll be left untouched.
     *
     *  What we do is to cycle trough the alternate settings and
     * let'em fill. This may look quite suboptimal, but looks clearer.
     *
     * Parameters:  
     *
     *  ts_usb_iface*   Pointer to interface structure to fill in.
     *
     *  __u8*           Pointer to where the interface descriptor
     *                  block begins inside a configuration descriptor
     *                  block. Note this has to point to the first
     *                  alternate setting of the interface.
     *
     *  size_t          Size of the buffer given in previous
     *                  parameter. 
     *
     * Returns:
     *
     *  ssize_t         errno code < 0 on error, size of all interface
     *                  and endpoint descriptors corresponding to the
     *                  alternate settings of the same interface.
     *
     * Global:
     *
     *  int usb_errno   On error, error code.  
     */

ssize_t usb_descriptor_iface_fill (ts_usb_device* device,
                                   ts_usb_iface* iface, 
                                   __u8* buffer, size_t buflength)
{
    size_t cumlen;              /* accumulative length */
    ts_usb_aset *aset, *top;    /* aset iterator and top iteration */
    ssize_t length;             /* size of aset descriptor block */
    
    gdlf_fntrace ("usb_descriptor_iface_fill (0x%p, 0x%p, 0x%p, %u)",
                  device, iface, buffer, buflength);

    gdlf_aassert (iface != NULL, return -EINVAL);
    gdlf_aassert (buffer != NULL, return -EINVAL);
    gdlf_aassert (buflength >= USB_IFACE_DESCR_LEN_1_0, return -EINVAL);

    if (phead (buffer)->bDescriptorType != e_usb_desc_iface)
    {
        gdlf_err ("Pretended to process a non-interface "
                  "descriptor for device 0x%p (%d %d) "
                  "[iface 0x%p, buffer 0x%p, size %u]\n",
                  device, device->hub->host->addr, device->addr,
                  iface, buffer, buflength);
        usb_errno = -USBE_IFACE_DESCR_BAD_TYPE;
        return -ENODATA;
    }   

        /*
         *  Having no alternate settings means there's already one,
         * and that's this one 
         */

    iface->asets = piface(buffer)->bAlternateSetting;
    if (iface->asets == 0)
        iface->asets = 1;
    iface->addr = piface(buffer)->bInterfaceNumber;
    iface->aset = ts_usb_aset_a_init (iface->asets);    /* alloc asets */
    iface->active = iface->aset;            /* first one is the active! */

    piface(buffer)->bAlternateSetting = 0;  /* make it sane again */

        /* Now fill in each alternate setting */

    aset = iface->aset;                     /* get first aset */
    top = iface->aset + iface->asets;       /* get last one */
    cumlen = 0;                             /* init cumulative length */

    while (aset < top)                      /* fill in all asets */
    {        
        length = usb_descriptor_aset_fill (device, iface, aset, 
                                           buffer + cumlen,
                                           buflength - cumlen);
        if (length < 0)                     /* error? cleanup */
        {
            ts_usb_aset_a_release (iface->aset, iface->asets);
            usb_kfree (iface->aset);
            piface(buffer)->bAlternateSetting = iface->asets;
            ts_usb_iface_clean (iface);
            return length;
        }
        aset++;                             /* fetch next */
        cumlen += length;
    }                           

        /*
         * No need to skip to next interface descriptor, as
         * usb_descriptor_aset_fill has already done 
         */

    piface(buffer)->bAlternateSetting = iface->asets;   /* undo sanity */ 
    return cumlen;
}


    /* Cook a configuration descriptor block
     *
     * THIS FUNCTION MODIFIES THE CONFIGURATION DESCRIPTOR BLOCK!!!
     *
     *  This function is used to scan a configuration descriptor block
     * for the number of different interfaces [not counting each
     * alternate setting of the same interface] and alternate settings
     * of each interface. This way it's easier to build the
     * configuration tree [see USB Spec 1.0 9.5].
     *
     *  We also scan for each interface the number of alternate
     * settings and store it into iface->bAlternateSetting [as we
     * know, by USB 1.0 Spec 9.6.3, that first alternate setting is
     * always number zero, we can recover later from this
     * modification].
     *
     * Parameters:  
     *
     *  const ts_usb_device*
     *
     *                  Pointer to the device structure which gave
     *                  this configuration descriptor block [only used
     *                  for diagnostic purposes].
     *
     *  __u8*           Pointer to where the configuration descriptor
     *                  block begins.
     *
     *  size_t          Maximum size [from where the cfg descr block
     *                  begins]. This is a safety lock, as no scanning
     *                  will be done beyond the limit set by adding
     *                  this value to previous pointer.
     *
     * Returns:
     *
     *  int             Errno code < 0 on error; in this case, nothing
     *                  is modified. Else, number of different
     *                  interfaces found. 
     *
     * Global:
     *
     *  int usb_errno   On error, error code.
     */

int usb_descriptor_cfg_cook (const ts_usb_device* device,
                             __u8* buffer, size_t length)
{
    int ifaces;     /* number of different interfaces found */
    __u8 *pasets;   /* Position where the bAlternateSetting of the
                     * first alternate setting descriptor for
                     * current interface is located at */
    int last_iface; /* number of last interface */
                    
    __u8 *data;     /* ptr for walking on the buffer */
    __u8 *top;      /* buffer's top */
    

    gdlf_fntrace ("usb_descriptor_cfg_cook (0x%p, %u)", 
                  buffer, length);
    
        /* Check configuration descriptor's bLength [USB Spec 1.0 9.6.2] */

    data = buffer;
    if (length < phead (data)->bLength)
    {
        usb_errno = -USBE_CFG_DESCR_SIZE;
        return -ENODATA;
    }
    
        /* 
         *  Now let's walk the block and count how many interfaces we
         * do have [again, skipping alternate settings of the same
         * interface]. So we'll just walk up from the very beginning,
         * checking at each iteration if we are walking over the
         * limit. 
         *
         *  We start at first interface, so we skip the configuration
         * descriptor. 
         *
         *  Note the top test is done against data+3, as we'll access
         * elements data+1, data+2 and data+3, and we want to be sure
         * we won't screw anything up there.
         */

    data = buffer + phead (data)->bLength;
    top = buffer + length;
    ifaces = 0;             /* initially we've got no ifaces */
    last_iface = -1;        /* last interface's number */
    pasets = NULL;

    while (data+3 < top)    /* +3 -> as we access +1, +2 and +3 */
    {                       /* what's at bDescriptorType? */
        switch (phead (data)->bDescriptorType)
        {
          case e_usb_desc_iface:    /* This is an iface descriptor */
            if (last_iface              /* new interface? */
                != piface(data)->bInterfaceNumber)
            {
                last_iface = piface(data)->bInterfaceNumber;
                pasets = &piface(data)->bAlternateSetting;
                gdlf_assert (piface(data)->bAlternateSetting == 0);
                *pasets = 1;
                ifaces++;                   /* 've got one more */
            }
            else                /* no, it's an aset */
            {
                (*pasets)++;
            }
            break;
            
          case e_usb_desc_endp: /* This is an endpoint descriptor */
            break;

          default:              /* This is something unknown? */
            gdlf_debug ("Unknown descriptor type 0x%x from device "
                        "0x%p (%d %d)'s configuration %d descriptor "
                        "block at offset 0x%x\n", 
                        (unsigned)phead (data)->bDescriptorType,
                        device, device->hub->host->addr, device->addr,
                        (unsigned)((ts_usb_cfg_descr*)buffer)->bConfigurationValue,
                        data - buffer);
            break;
        }
        gdlf_aassert (phead (data)->bLength != 0, break);
        data += phead (data)->bLength;   /* get nxt one */
    }
    return ifaces;
}


    /* Get a configuration descriptor and fill the structure
     *
     *  This function will read in a configuration descriptor [which
     * includes information for its interfaces and endpoints] and with
     * it, fill in a tree of configuration information, allocating the
     * interfaces, alternate settings and endpoints arrays.
     *
     *  The configuration descriptor includes the descriptor itself,
     * followed by the interface descriptors [which in turn are
     * followed by it's endpoint descriptors] and class
     * specific/vendor specific descriptors. See USB Spec 1.0 9.6.2.
     *
     *  What we do is to read firstly the configuration descriptor,
     * parse it's interesting stuff and then get the full package.
     *
     * NOTE: We assume that ts_usb_iface_desc->bNumInterfaces includes
     *       all alternate settings of a single interface [I mean,
     *       alternate settings zero and one of interface 0 are counted
     *       as two different interfaces]. Hope this is right.
     *
     * Parameters:  
     *
     *  ts_usb_device*  Pointer to USB device structure to fill in.
     *
     *  t_usb_cfg       Configuration index number.
     *
     * Returns:
     *
     *  int             0 if ok, errno code < 0 on error.
     *
     * Global:
     *
     *  int usb_errno   On error, error code.  
     */

int usb_descriptor_cfg_fill (ts_usb_device* device, 
                             t_usb_cfg idx)
{
    __u8 *buffer;
    int cnt;
    char *str;
    ssize_t length;     /* read length [bytes] */
    size_t cumlen;      /* accumulative length */
    ts_usb_cfg *cfg;  
    ts_usb_cfg_descr desc;

    gdlf_fntrace ("usb_descriptor_cfg_fill (0x%p, %u)", device, idx);

    gdlf_aassert (device != NULL, return -EINVAL);
    gdlf_aassert (idx < device->pcfgs, return -EINVAL);
    cfg = device->pcfg[idx];
    gdlf_aassert (cfg != NULL, return -EINVAL);

        /* Read just the configuration descriptor */

    idx += device->cfgzero;     /* cfg index one or zero based? */
    length = usb_descriptor_get (device, e_usb_desc_cfg, idx,
                                 0, &desc, USB_CFG_DESCR_LEN_1_0);
    if (length < 0)         /* if error, usb_descriptor_get() ... */
        return length;      /* ... has already complained */
    if (length != USB_CFG_DESCR_LEN_1_0)
    {
        gdlf_err ("Invalid USB configuraton descriptor index %d "
                  "returned by device 0x%p (%d %d); %d bytes expected, "
                  "%d readen\n", 
                  idx, device, device->hub->host->addr, device->addr,
                  USB_CFG_DESCR_LEN_1_0, length);
        usb_errno = -USBE_CFG_DESCR_SIZE;
        return -EIO;
    }
    
        /* Do some checks */

    if (desc.head.bDescriptorType != e_usb_desc_cfg)
    {
        gdlf_err ("Invalid USB configuration descriptor index %d "
                  "returned by device 0x%p (%d %d): invalid descriptor "
                  "type %u\n", 
                  idx, device, device->hub->host->addr, device->addr,
                  desc.head.bDescriptorType);
        usb_errno = -USBE_CFG_DESCR_BAD_TYPE;
        return -ENODATA;
    }
    desc.wTotalLength = le16_to_cpu (desc.wTotalLength);

        /* Print some pretty info */

    str = usb_string_get (device, desc.iConfiguration, -1);
    gdlf_info ("Processing USB device 0x%p (%d %d)%s%s%s, "
               "configuration index %d%s%s\n",
               device, device->hub->host->addr, device->addr,
               device->descr? " '" : "", 
               device->descr? device->descr : "", 
               device->descr? "'" : "",               
               idx, str? ": " : "", str? str : "");

        /* Allocate the buffer for getting the whole data */

    buffer = usb_kmalloc (desc.wTotalLength * sizeof (__u8), GFP_KERNEL);
    if (buffer == NULL)
    {
        gdlf_err ("Couldn't allocate %d bytes to read descriptor "
                  "for configuration index %d of device 0x%p (%d %d)\n",
                  desc.wTotalLength, idx, device,
                  device->hub->host->addr, device->addr);
        usb_kfree (str);
        usb_errno = -USBE_NO_MEMORY;
        return -ENOMEM;
    }
    
        /* Get the whole descriptors block */

    length = usb_descriptor_get (device, e_usb_desc_cfg, idx, 
                                 0, buffer, desc.wTotalLength);
    if (length < 0)                     /* error? usb_descriptor_get() ... */
        return length;                  /* ... has already complained */
    if (length != desc.wTotalLength)    /* For the USB 1.0 Spec */
    {
        gdlf_err ("Invalid USB configuration descriptor index %d block "
                  "returned by device 0x%p (%d %d); %d bytes "
                  "expected, %d readen\n", 
                  idx, device, device->hub->host->addr, device->addr,
                  desc.wTotalLength, length);
        usb_kfree (str);
        usb_kfree (buffer);
        usb_errno = -USBE_CFG_DESCR_SIZE;
        return -EIO;
    }

        /* Compute number of interfaces */

    length = usb_descriptor_cfg_cook (device, buffer, desc.wTotalLength);
    if (length < 0)
    {
        gdlf_err ("Error computing number of interfaces on "
                  "configuration index %d descriptor block returned "
                  "by device 0x%p (%d %d): errno %d\n",
                  idx, device, device->hub->host->addr, 
                  device->addr, length);
        usb_kfree (str);
        usb_kfree (buffer);
        return length;
    }
    cfg->ifaces = length;
    
        /* Allocate the interfaces array */

    cfg->iface = ts_usb_iface_a_init (cfg->ifaces);
    if (cfg->iface == NULL)
    {
        gdlf_err ("Couldn't allocate %d bytes to store interfaces' "
                  "information for configuration index %d of device 0x%p "
                  "(%d %d)\n",
                  sizeof (ts_usb_iface) * cfg->ifaces, idx, device,
                  device->hub->host->addr, device->addr);
        cfg->ifaces = 0;
        usb_kfree (str);
        usb_kfree (buffer);
        usb_errno = -USBE_NO_MEMORY;
        return -ENOMEM;
    }

        /*
         * Skip vendor/device specific descriptors and
         * position and position to first interface descriptor 
         */

    cumlen = 0;
    while (phead (buffer+cumlen)->bDescriptorType != e_usb_desc_iface 
           && cumlen < desc.wTotalLength)
        cumlen += phead (buffer+cumlen)->bLength;

        /* Fill in each interface's data */

    for (cnt = 0; cnt < cfg->ifaces && cumlen < desc.wTotalLength; 
         cnt++, cumlen += length)
    {
        length = usb_descriptor_iface_fill (device, &cfg->iface[cnt], 
                                            buffer + cumlen,
                                            desc.wTotalLength - cumlen);
        if (length < 0)
        {
            gdlf_err ("Couldn't fill in information for "
                      "interface %d of configuration index %d "
                      "of device 0x%p (%d %d): errno %d\n",
                      cnt, idx, device, device->hub->host->addr, 
                      device->addr, length);
                /* usb_errno set by usb_descriptor_iface_fill() */
            usb_kfree (str);
            usb_kfree (buffer);
            ts_usb_cfg_release (cfg);
            return length;
        }
    }

    usb_kfree (buffer);

        /* Fill in the configuration's data */

    cfg->addr = desc.bConfigurationValue;
    cfg->buspower = desc.bmAttributes & USB_CFG_DESCR_ATTR_BUS_PWR? 1 : 0;
    cfg->selfpower = desc.bmAttributes & USB_CFG_DESCR_ATTR_SELF_PWR? 1 : 0;
    cfg->remwakeup = desc.bmAttributes & USB_CFG_DESCR_ATTR_REM_WAKEUP? 1 : 0;

    cfg->maxpower = desc.MaxPower;
    
    cfg->ifaces = desc.bNumInterfaces;
    cfg->descr = str;
    return 0;
}


    /* Fill in a device structure with the data from the USB
     * descriptors. 
     *
     * This function is also known as 'La Dolorosa'.
     *
     *  This function firstly gets the first 8 bytes of device
     * descriptor. With them we do the most basic checks and get the
     * maximum packet size for endpoint 0 [initially we consider it to
     * be eight].
     *
     *  Then we get the whole device descriptor and fill the device
     * info. After that we start getting the configurations array. We
     * allocate the array of pointers to configurations and allocate
     * one structure for each, calling the configuration descriptor
     * fill function on it. This will fill all the interface and
     * endpoint information for it.
     *
     * NOTE: we do no locking as it is suppossed only on process gets
     *       the notification for device attachment.
     *
     * Parameters:  
     *
     *  ts_usb_device*  Pointer to USB device structure to fill in.
     *
     * Returns:
     *
     *  int             0 if ok, errno code < 0 on error.
     *
     * Global:
     *
     *  int usb_errno   On error, error code.  
     */

extern int usb_cfel_update (ts_usb_device*);

int usb_descriptor_fill (ts_usb_device* device)
{
    int errcode;
    int cfg_cnt;    /* need to be int, so -1 < device->cfgs */
    ssize_t length;
    ts_usb_dev_descr desc;        

        /* Get first eight bytes to obtain basic info */
        
    length = usb_descriptor_get (device, e_usb_desc_dev, 0, 0, &desc, 8);

    if (length < 0)         /* if error, usb_descriptor_get() ... */
        return length;      /* ... has already complained */

        /* Check descriptor length */

    if (length != 8)        /* For the USB 1.0 Spec */
    {
        gdlf_err ("Invalid USB Device Descriptor head returned "
                  "by device 0x%p (%d %d): 8 bytes expected, "
                  "%d readen\n", 
                  device, device->hub->host->addr, 
                  device->addr, length);
        usb_errno = -USBE_DEV_DESCR_SIZE;
        return -ENODATA;
    }

        /* Check descriptor type */

    if (desc.head.bDescriptorType != e_usb_desc_dev)
    {
        gdlf_err ("Invalid USB Device Descriptor returned "
                  "by device 0x%p (%d %d): type 0x%u it's not a "
                  "device descriptor\n", 
                  device, device->hub->host->addr, device->addr, 
                  (unsigned) desc.head.bDescriptorType);
        usb_errno = -USBE_DEV_DESCR_BAD_TYPE;
        return -ENODATA;
    }

        /* Check USB Protocol Revision version */

    device->usb_version = le16_to_cpu (desc.bcdUSB);        
    if (device->usb_version > USB_MAX_RELEASE)   /* do we know about it? */
    {
        gdlf_err ("Device 0x%p (%d %d) uses protocol "
                  "specification %d.%d; I only understand "
                  "until %d.%d; cannot use it\n",
                  device, device->hub->host->addr,
                  device->addr, 
                  (device->usb_version & 0xff00) >> 8, 
                  device->usb_version & 0x00ff,
                  (USB_MAX_RELEASE & 0xff00) >> 8,
                  USB_MAX_RELEASE & 0x00ff);
        usb_errno = -USBE_UNSUPPORTED_RELEASE;
        return -EINVAL;
    }            

    device->endp0.maxpktsize = desc.bMaxPacketSize;
        
        /* ok, now really get the full descriptor */

    length = usb_descriptor_get (device, e_usb_desc_dev, 
                                 0, 0, &desc, USB_DEV_DESCR_LEN_1_0);

        /* Check descriptor length */

    if (length != USB_DEV_DESCR_LEN_1_0)    /* For the USB 1.0 Spec */
    {
        if (length < 0)
        {
            gdlf_err ("Error reading USB Device Descriptor from "
                      "device 0x%p (%d %d): errno/usb %d/%d\n", 
                      device, device->hub->host->addr, device->addr,
                      length, usb_errno);
            return length;   
        }
        else
        {
            gdlf_err ("Invalid USB Device Descriptor returned by "
                      "device 0x%p (%d %d); %d bytes expected, "
                      "%dreaden\n", 
                      device, device->hub->host->addr, device->addr,
                      USB_DEV_DESCR_LEN_1_0, length);
            usb_errno = -USBE_DEV_DESCR_SIZE;
            return -EIO;
        }
    }

        /* Convert endianess, except fields we already know */

        /* device->usb_version is already set */
    device->cfel.class = desc.bDeviceClass;
    device->cfel.subclass = desc.bDeviceSubClass;
    device->cfel.protocol = desc.bDeviceProtocol;
        /* device->endp0d.maxpktsize is already set */
    device->vendor = le16_to_cpu (desc.idVendor);
    device->product = le16_to_cpu (desc.idProduct);
    device->version = le16_to_cpu (desc.bcdDevice);
    device->pcfgs = desc.bNumConfigurations;

        /* 
         * Get the strings stuff 
         *
         *  Note this stuff is encoded using Unicode. Problem is we
         * don't know from the kernel which language to get [in case
         * the device supports them]. We can default to get the first
         * one [index 1], but we should fix it.
         *
         *  For converting from Unicode to a local charset [let's say
         * latin1, i.e.] we could use the routines which handle
         * Unicode for the Joliet extensions to iso9660.
         */

    device->mfrer = usb_string_get (device, desc.iManufacturer, -1);
    device->descr = usb_string_get (device, desc.iProduct, -1);
    device->serial = usb_string_get (device, desc.iSerialNumber, -1);

#warning FIXME: add version information on banners
    if (device->descr || device->mfrer || device->serial)
    {
        gdlf_info ("Device 0x%p (%d %d): %s%s%s%s%s\n",
                   device, device->hub->host->addr, device->addr,
                   device->mfrer? device->mfrer : "",
                   device->mfrer? " " : "",
                   device->descr? device->descr : "",
                   device->serial? " s/n " : "",
                   device->serial? device->serial : "");
    }
    else
        gdlf_info ("Device 0x%p (%d %d): (no information available)\n",
                   device, device->hub->host->addr, device->addr);
        
        /*
         * Ok, now we have a perfect device which is fully
         * unconfigured [no configuration is selected]. Now we gotta
         * allocate the configurations array, an structure per
         * configuration and call usb_descriptor_cfg_fill() function
         * on each.
         */

        /* allocate cfg pointers array */

    device->pcfg = usb_kmalloc (sizeof (ts_usb_cfg*) * device->pcfgs,
                                GFP_KERNEL);
    if (device->pcfg == NULL)
    {
        gdlf_err ("Couldn't allocate %u bytes of memory for "
                  "the configurations array of device 0x%p (%d %d)\n", 
                  sizeof (ts_usb_cfg*) * device->pcfgs,
                  device, device->hub->host->addr, device->addr);
        usb_errno = -USBE_NO_MEMORY;
        return -ENOMEM;
    }

        /*
         * Now we ask for each configuration descriptor and fill it. 
         *
         *  However, we may find trouble. Some devices expect
         * configuration descriptor indexes ranging from 1 to n,
         * others [most common I've seen so far] from 0 to n-1, and
         * they may or not be related to the configuration value.
         *
         *  What we do is start in configuration index 0. If that
         * stalls (errno == -EREMOTEIO), then it is 1 based. We set a
         * little flag and retry first descriptor
         */

    device->cfgzero = 0;
    for (cfg_cnt = 0; cfg_cnt < device->pcfgs; cfg_cnt++)
    {
        device->pcfg[cfg_cnt] = ts_usb_cfg_init();  /* allocate */
        if (device->pcfg[cfg_cnt] == NULL)
        {
            gdlf_err ("Cannot allocate USB configuration %d "
                      "structure for device 0x%p (%d %d); skipping; "
                      "THIS IS VERY BAD!\n",
                      cfg_cnt, device, device->hub->host->addr,
                      device->addr);
                /* FIXME: in fact, I think I'd better abort */
            continue;
        }

        errcode =                                   /* fill it */
            usb_descriptor_cfg_fill (device, cfg_cnt);

        if (cfg_cnt == 0 && errcode == -EREMOTEIO 
            && device->cfgzero == 0) 
        {
            device->cfgzero = 1;
            errcode =
                usb_descriptor_cfg_fill (device, cfg_cnt);
        }

        if (errcode < 0)
        {
            gdlf_err ("Could not fill USB configuration %d for "
                      "device 0x%p (%d %d): errno/usb_errno %d/%d; "
                      "skipping; THIS IS VERY BAD!\n",
                      cfg_cnt, device, device->hub->host->addr,
                      device->addr, errcode, usb_errno);
            ts_usb_cfg_clean (device->pcfg[cfg_cnt]);
            usb_kfree (device->pcfg[cfg_cnt]);
            device->pcfg[cfg_cnt] = NULL;
                /* FIXME: in fact, I think I'd better abort */
            continue;
        }
    }
    

    if (device->cfel.class == 0)
        device->cfel.configured = 1;    
    usb_cfel_update (device);
    return 0;   /* so ... that's it, we're done */
}


/* Log:
 *
 * $Log: usbd-descr.c,v $
 * Revision 1.23  1998/11/14 17:15:08  inaky
 * Moved the include files to uusbd/include/uusbd, to ease integration
 * into the Linux kernel.
 *
 * Revision 1.22  1998/11/09 21:28:06  inaky
 * Fixed informational messages. Changed the name of the configuration
 * pointer array to pcfg.
 *
 * Revision 1.21  1998/10/31 02:12:04  inaky
 * Added the fix for devices using index 0 as first configuration
 * descriptor, with a bit which indicates it. Fixed the error messages to
 * refer to the 'configuration index', not just to the 'configuration'
 * [which could be confused with the configuration value].
 *
 * Revision 1.20  1998/10/29 22:21:05  inaky
 * Fixed a subttle bug ... the product code was being filled in with the
 * product code :) silly me. Thx to Charles Salmon for noticing.
 *
 * Revision 1.19  1998/08/28 23:45:42  inaky
 * Removed traces.
 *
 * Revision 1.18  1998/08/27 11:14:23  inaky
 * Moved ts_*_free() to ts_*_release(), made them return void and severed
 * assertions.
 *
 * Made it print always the info, writing "(not available)" on
 * unavailable items.
 *
 * Revision 1.17  1998/08/24 02:24:27  inaky
 * Changed name of all enumerations to be prefixed with the enumeration
 * type name.
 *
 * Did the little debugging needed to make the functions work as they're
 * expected to.
 *
 * Revision 1.16  1998/08/19 23:50:52  inaky
 * Removed some calls to ts_usb_device_clean() which were wrong, as
 * usb_descriptor_fill() doesn't have to _clean_ the device on error, but
 * comunicate and let the caller do the proper cleaning with
 * usb_device_destroy().
 *
 * Revision 1.15  1998/07/23 14:27:22  inaky
 * Added the length parameter to the request structure.
 *
 * Revision 1.14  1998/07/16 14:10:01  inaky
 * Moved from the old read/write interface to the new, request-based
 * interface and fixed locking.
 *
 * Revision 1.13  1998/05/01 19:09:05  inaky
 * Fixed the descriptor filling functions for interfaces and alternate
 * settings, to take a non-constant ts_usb_device* as parameter.
 *
 * Revision 1.12  1998/04/26 14:55:45  inaky
 * Fixed error messages in usb_descriptor_set().
 *
 * Added device argument to many functions which needed it to correctly
 * report errors and get string descriptors. Also added the code to get
 * the string descriptors and print them while scanning if available.
 *
 * Fixed usb_descriptor_fill() to actualize cfel counts.
 *
 * Revision 1.11  1998/04/26 03:34:51  inaky
 * Added some docs to clarify concepts.
 *
 * Finished 'La Dolorosa'. It fills recursively, by letting it slip in
 * four levels: configuration, interface, alternate setting and
 * endpoints, which are handled each one by it's corresponding function.
 *
 * Revision 1.10  1998/04/19 22:43:50  inaky
 * Completed the descriptor filling function. Still missing the interface
 * filling function and a couple helpers.
 *
 * Revision 1.9  1998/04/13 22:08:56  inaky
 * Fixed error messages when requesting USB descriptors, as well as
 * fixing warning methods in trbuffers.
 *
 * Revision 1.8  1998/03/28 19:31:13  inaky
 * Fixed to work with callbacks and starting to do the descriptor fill
 * function.
 *
 * Revision 1.7  1998/01/11 15:44:17  inaky
 * Finished functions from the descriptor, inteface and endpoint modules,
 * fixed docs, pretty printed and removed craps.
 *
 * Revision 1.6  1998/01/09 17:02:52  inaky
 * Basic descriptor functions finished. Untested.
 *
 * Revision 1.5  1998/01/08 12:19:12  inaky
 * Added copyright strings to all important files
 *
 * Revision 1.4  1998/01/08 11:49:12  inaky
 * Finishing the configuration system. Only the final cfel-stuff is
 * missing to implement; however, nothing is tested.
 *
 * Revision 1.3  1998/01/02 22:56:03  inaky
 * Added the descriptor fill function prototype and part of the
 * documentation
 *
 * Revision 1.2  1997/12/29 00:25:39  inaky
 * Fixed unimplemented functions to return -ENOSYS and moving the khubdd
 * to use signals as commands.
 * Also added function traces on most files that had them disabled (donkey!)
 *
 * Revision 1.1  1997/12/27 19:54:21  inaky
 * Added all the prototypes and skeleton definitions (incomplete) of all
 * of the USBD functions to each source file.
 *
 */

