
/* Universal USB Driver Stack v$Revision: 1.48 $ 
 *
 * (C) 1997 by Iaky Prez Gonzlez <inaky@peloncho.fis.ucm.es>
 *
 * Distributed under the terms of the GNU Public License Version 2
 * See files COPYRIGHT and COPYING at the root distribution directory
 *
 * $Id: usbd-device.c,v 1.48 1998/11/14 17:15:10 inaky Exp $
 *
 * Changelog at the bottom 
 *
 *
 *  USB Driver - Devices section
 *  -----------------------------
 *
 * This is the part of the UUSBD who takes care of device registration
 * and managing.
 *
 */

#define __NO_VERSION__
#include <linux/module.h>   /* MOD_INC_USE_COUNT() */
#undef __NO_VERSION__
#include <uusbd/usb.h>
#include <uusbd/usbd.h>

#define GDLF_IDENT          "USBD"
#define GDLF_LOCAL          /* Define if want local settings */
#define GDLF_DEF_LOGLEVEL   9
#define GDLF_ASSERTIONS     /* want assertions (#undef if not) */
#undef GDLF_FNTRACE        /* want function traces */
#define GDLF_LOCATIONS      /* want locations in log messages */
#include <uusbd/gdlf.h>


    /* 
     * System lists of devices and unconfigured devices 
     *
     *  Every device registered into the system is into the list of
     * devices, no matter it is fully unconfigured, partially
     * configured or totally configured. 
     *
     *  A device is on the list of unconfigured devices if it has any
     * unconfigured configurable element, wether it is the main
     * configurable element (the one at ts_usb_device.cfel, unless
     * it's attributes are all zero) or anyone in the ACTIVE
     * CONFIGURATION. 
     */

#define NLISTS 3
ts_3slist devices;  /* list of devices */
ts_3slist unconfs;  /* list of unconfigured devices */


    /* Get a device pointer, knowing it's address and bus.  Return a
     * pointer to it's control structure.
     *
     * Warning! It is advisable to get a shared lock on the host
     *          structure we are searching for, to evitate unexpected
     *          changes.
     *
     * FIXME: these functions could be deprecated, but
     *        usb_set_dst_dev() uses it. Sounds like the whole
     *        addressing system should be revised.
     *
     * Parameters:  
     *
     *  t_usb_host   Host number [usb_device_hn_ptr]
     *
     *  ts_usb_host* Host pointer  [usb_device_hp_ptr]
     *
     *  t_usb_addr   Device's address
     *
     * Returns:
     *
     *  ts_usb_device*  Pointer to the device's data structure, or
     *                  NULL if not found (address not asigned or bad
     *                  host number ).  
     */

ts_usb_device* usb_device_ptr_hn (t_usb_host hostn, t_usb_addr addrn)
{
    ts_usb_host *host;
    
    gdlf_fntrace ("usb_device_ptr_hn (%d, %d)", hostn, addrn);

    host = usb_host_ptr (hostn);
    if (host == NULL)   /* usb_host_ptr() already complains */
        return NULL;
    
    return usb_device_ptr_hp (host, addrn);
}


ts_usb_device* usb_device_ptr_hp (ts_usb_host *host, t_usb_addr addrn)
{
    ts_usb_device *dev;   /* device iterator */

    gdlf_fntrace ("usb_device_ptr_hp (0x%p, %d)", host, addrn);
    
    gdlf_apassert (host != NULL,
                   { usb_errno = -USBE_HOST_NULL; return NULL; },
                   "Tried to search a device in a NULL host!\n");

    if (host->devices.elements == 0)    /* No devices? ... */
    {
        gdlf_debug ("No devices on host %d ... \n", host->addr);
        return NULL;                    /* ... we won't find it, will we? */
    }

        /* Iterate from top element of the array until the first */

    read_lock (&host->lock);
    dev = (ts_usb_device*)ts_3slist_head (&host->devices);
    while (dev != NULL)
    {
        if (dev->addr == addrn)
        {
            read_unlock (&host->lock);
            return dev;
        }
        dev = (ts_usb_device*)
            ts_3slist_next (&host->devices, &dev->msnode);
    }
    
    read_unlock (&host->lock);
    return NULL;
}


    /*  Register a device into the UUSBD system.
     *
     *  This function performs no configuration tasks!!
     *
     * Compliance:      USB Spec 1.0 10.5.2.5
     * 
     * Parameters:  
     *
     *  ts_usb_device*  Device structure, with all the relevant
     *                  information obtained from the descriptors
     *                  loaded.
     *
     * Returns:
     *
     *  int             0 if ok, errno code < 0 on error.
     *
     * Global:
     *
     *  int usb_errno   On error, error code.  
     */

int usb_device_register (ts_usb_device *device)
{
    int errcode;
 
    gdlf_fntrace ("usb_device_register (0x%p)", device);    

        /* Add it to the list of devices */

    write_lock (&device->lock);
    errcode = ts_3slist_prep (&devices, &device->msnode);
    write_unlock (&device->lock);

    if (errcode)
        gdlf_err ("Couldn't add device 0x%p to the "
                  "system list of devices: "
                  "errno/usb %d/%d\n",
                  device, errcode, usb_errno);
    return errcode;
}


    /* Unregister a device from the USB system. 
     *
     *  Removes a device from the system list of devices.
     *
     * Compliance:      USB Spec 1.0 10.5.2.6
     * 
     * Parameters:  
     *
     *  ts_usb_device*  Device structure, with all the relevant
     *                  information obtained from the descriptors
     *                  loaded.
     *
     * Returns:
     *
     *  int usb_errno   0 if ok, errno code < 0 On error.
     *
     * Global:
     *
     *  int usb_errno   On error, error code.  
     */

int usb_device_unregister (ts_usb_device *device)
{
    int errcode;
    
    gdlf_fntrace ("usb_device_unregister (0x%p)", device);

        /* Remove it from the list of devices */

    write_lock (&device->lock);
    errcode = ts_3slist_del (&devices, &device->msnode);
    write_unlock (&device->lock);

        /* 
         *  In normal operation, this error only'd mean the device
         * wasn't at the list, something that can happen in some
         * situations 
         */

    if (errcode)
        gdlf_debug ("Couldn't remove device 0x%p from the "
                    "system list of devices: "
                    "errno/usb %d/%d\n",
                    device, errcode, usb_errno);
    return errcode;
}


    /* Attach a device to the UUSBD
     *
     *  On connection event, khubdd calls this function with a freshly
     * allocated and cleaned ts_usb_device structure and the hub and
     * port where the attachment has been detected. 
     *
     *  NOTE: configuration tree = device->cfg + device->cfgs +
     *        device->active 
     *
     *  Then we initialize the device structure, register it within
     * the hub; we search for a free address in the bus and assign it
     * to the device. 
     *
     *  Now we read in the device descriptors and make a full USB
     * configuration tree. 
     *
     *  Is time then to register the device into it's host and then
     * into the UUSBD and to try to configure it. The device is now
     * ready for operation. 
     *
     *  When a device is detached, the function usb_device_detach()
     * undoes what usb_device_attach() does.
     *
     * NOTE: we set here the destructor for device structures, which
     *       is the usb_device_detach() function. Thus, when a device
     *       is destroyed, it will cleanly remove any track it may
     *       leave in the system.
     *
     * Parameters:  
     *
     *  ts_usb_device*  Pointer to an allocated and cleaned device
     *                  struct. 
     *
     *  ts_hubd*        Hub where the device will be attached to.
     *
     *  __u8            Port number to which the device is attached.
     *
     * Returns:
     *
     *  int             0 if ok, errno code < 0 on error.
     *
     * Global:
     *
     *  int usb_errno   On error, error code.  
     */

#warning FIXME: speed this up using a single tr to do everything, _fast_

int usb_device_attach (ts_usb_device* device, 
                       ts_usb_hub* hub, __u8 port, __u32 flags)
{
    ts_usb_addr endp0_addr;
    int errcode, newaddr;
    
    gdlf_fntrace ("usb_device_attach (0x%p, %u)", hub, port);

    gdlf_aassert (device != NULL, return -EINVAL);
    gdlf_aassert (hub != NULL, return -EINVAL);
    gdlf_aassert (hub->host != NULL, return -EINVAL);
    gdlf_aassert (port < hub->ports, return -EINVAL);

        /* set the destructor */

    write_lock (&device->lock);
    ts_dtor_fill (&device->msnode.dtor, (tf_dtor)usb_device_destroy);
    write_unlock (&device->lock);

        /* 
         * usb_hub_register_device() only fails if asserting, so it makes
         * no sense to do a permanent check
         */

    errcode = usb_hub_register_device (hub, device, port);
    gdlf_apassert (errcode >= 0,
                   {
                           /* NO need to ts_usb_device_release, nothing used */
                       usb_kfree (device);
                       return errcode;
                   },
                   "Cannot register device 0x%p into "
                   "hub 0x%p, port %u: errno/usb "
                   "%d/%d\n",
                   device, hub, port,
                   errcode, usb_errno);

        /* Now register into into the host */

    device->addr = 0;
    errcode = usb_host_register_device (device);
    if (errcode)
    {   /* FIXME: does usb_host_register_device() complain? */
        gdlf_err ("Couldn't register device 0x%p "
                  "into host: errno/usb %d/%d\n",
                  device, errcode, usb_errno);
        usb_device_destroy (device);
        return errcode;
    }

        /* Register the device into the UUSBD system */

    errcode = usb_device_register (device);
    if (errcode)
    {
        gdlf_err ("Cannot register device: "
                  "errno/usb %d/%d\n",
                  errcode, usb_errno);
        usb_device_destroy (device);
        return errcode;
    }

        /*
         *  Let's open the endpoint 0 descriptor 
         *   
         *  Fill in endpoint zero data, then address to open. Open it
         * using the default address (address zero); then set it's
         * address, close it and reopen using the new address.
         */

    write_lock (&device->lock);
    device->endp0.addr = 0;
    device->endp0.type = e_usb_tr_type_cntl;
    device->endp0.maxpktsize = 8;   /* After reset, we read only 8 bytes */
    write_unlock (&device->lock);   /* (USB Spec 1.0 5.5.3 pr 2) */

    endp0_addr.host = hub->host->addr;
    endp0_addr.device = 0;
    endp0_addr.endp = 0;
    errcode = usb_open (&endp0_addr, &device->endp0d);
    if (errcode < 0)
    {
        gdlf_err ("Cannot open device's default pipe: "
                  "errno/usb %d/%d\n",
                  errcode, usb_errno);
        usb_device_destroy (device);
        return errcode;
    }

        /* 
         * Get a new address and check if the host is full 
         * (maximum devices reached!) If ok, set the address.
         */

    read_lock (&host->lock);
    gdlf_passert (hub->host != NULL, 
                  "Hub has a NULL host, bus will seem full\n");
    newaddr = usb_host_free_address (hub->host);
    if (newaddr < 0)
    {
        gdlf_err ("The host 0x%p (bus %u) is full and doesn't admit "
                  "more devices; move it to other bus\n",
                  hub->host, hub->host->addr);
        read_unlock (&host->lock);
        usb_device_destroy (device);
        return errcode;
    }

    errcode = usb_device_address_set (device, newaddr);
    read_unlock (&host->lock);
    if (errcode)
    {
        gdlf_err ("Cannot set the device's address: "
                  "errno/usb %d/%d\n",
                  errcode, usb_errno);
        usb_device_destroy (device);
        return errcode;
    }
    
        /* Now we changed the address, reopen */

    errcode = usb_close (&device->endp0d);
    if (errcode < 0)
    {
        gdlf_err ("Cannot close device's default pipe (to re-open): "
                  "errno/usb %d/%d\n",
                  errcode, usb_errno);
        usb_device_destroy (device);
        return errcode;
    }
    endp0_addr.device = newaddr;    
    errcode = usb_open (&endp0_addr, &device->endp0d);
    if (errcode < 0)
    {
        gdlf_err ("Cannot re-open device's default pipe: "
                  "errno/usb %d/%d\n",
                  errcode, usb_errno);
        usb_device_destroy (device);
        return errcode;
    }

        /* Fill in the configuration tree from the descriptors. */

    errcode = usb_descriptor_fill (device);
    if (errcode < 0)
    {
        gdlf_err ("Cannot get device descriptors: "
                  "errno/usb %d/%d\n",
                  errcode, usb_errno);
        usb_device_destroy (device);
        return errcode;
    }

        /* Let's try to configure it now.  */

    errcode = usb_device_config (device, flags);
    if (errcode < 0)            /* oops ... error */
        gdlf_err ("Error configuring device: "
                  "errno/usb %d/%d; leaving unconfigured\n",
                  errcode, usb_errno);
    
    if (device->acfels == device->acfelu)
        gdlf_warning ("Device 0x%p (%d %d) could not be configured; "
                      "newly loaded drivers will try to "
                      "configure it\n",
                      device, hub->host->addr, device->addr);
    else if (device->acfelu)
        gdlf_warning ("Device 0x%p (%d %d) is partially configured; "
                      "newly loaded drivers will try to "
                      "fully configure it\n",
                      device, hub->host->addr, device->addr);
    ts_usb_device_printk (device);    
    return 0;
}


    /* Detach a device from the UUSBD
     *
     *  This function is called when a device is deteched (or it is
     * going to be detached) from the UUSBD.
     *
     *  The steps followed are:
     *
     *  1. unconfigure the device (usb_device_unconfig())
     *
     *     Note that this is a crucial step for the topology of the
     *     USB buses. If the device we are detaching is a hub, it will
     *     destroy itself, but previously it will detach all it's
     *     attached devices using this same function call and thus it
     *     will do a recursive detach of all the bus tree starting at
     *     the hub.
     *
     *      Also take into account that when unconfiguring, the
     *     destructors for the configuration data in the ts_usb_cfel
     *     structures are all called, what means that the drivers will
     *     (well, should) cleanly close everything.
     *
     *  2. abort all the remaining opened transfers on the device by
     *     unsetting (usb_dst_unset()) the destinations of all the
     *     open descriptors who still are open on the endpoints
     *     (this way the code using it will fail *softly* and will be
     *     able of recovering and aborting).
     *
     *      This step should not be necessary, as previous step should
     *     have left all the endpoints unused but perhaps the control
     *     endpoint. Should assert this. FIXME.
     *
     *  3. remove the device from the unconfigured device's list
     *
     *  4. unregister from it's host (usb_host_unregister_device_())
     *
     *  5. unregister (usb_device_unregister()
     *
     *  6. unregister from the hub (hubd_device_unregister())
     *
     *  7. free device's contents and structure space
     *
     * Parameters:  
     *
     *  ts_usb_device*  Pointer to the device to detach
     *
     *  __u32           Flags given to the unconfiguration routine as
     *                  reason for the detachment.
     *
     * Returns:
     *
     *  int             0 if ok, errno code < 0 on error.
     *                  However, even on error, it is guaranteed that the
     *                  device is going to be detached from the UUSBD
     *                  system. 
     *
     * Global:
     *
     *  int usb_errno   On error, error code.  
     */

int usb_device_detach (ts_usb_device* device, __u32 flags)
{
    unsigned cnt;
    ts_usb_endp **pendp;
    
    int errcode = 0, errcode2;
    
    gdlf_fntrace ("usb_device_detach (0x%p)", device);
    gdlf_aassert (device != NULL, return -EINVAL);
    
    usb_device_ref (device);

    errcode2 = usb_device_unconfig (device, flags);
    if (errcode2)
    {
        gdlf_err ("Couldn't unconfigure device 0x%p (%d %d): "
                  "errno/usb %d/%d; "
                  "detaching anyway\n",
                  device, device->hub->host->addr, device->addr,
                  errcode2, usb_errno);
        errcode = errcode2;
    }

        /* Unset the active USB configuration :) */

    if (device->active)
        if ((flags & (USB_CFG_DETACH | USB_CFG_HUB_PWR_DN 
                      | USB_CFG_HUB_OVR_CUR | USB_CFG_HUB_DESTROY)) == 0)
        {   /* If physically attached, unset it */
            errcode2 = usb_configuration_set (device, 0, 0);
            if (errcode2)
            {
                gdlf_err ("Couldn't reset USB configuration "
                          "for device 0x%p (%d %d): "
                          "errno/usb %d/%d; "
                          "detaching anyway\n",
                          device, device->hub->host->addr, device->addr,
                          errcode2, usb_errno);
                errcode = errcode2;
            }
        }
        else    /* no device connected, do it manually */
        {
            usb_configuration_unref (device->active);
            device->active = NULL;
        }
    
        /* We ignore errors here ... it may not be in the list */

    errcode2 = ts_3slist_del (&unconfs, &device->msnode);
    if (errcode2)
    {
        gdlf_debug ("Couldn't remove device 0x%p (%d %d) from the "
                    "list of unconfigured devices: "
                    "errno/usb %d/%d\n",
                    device, device->hub->host->addr, device->addr,
                    errcode2, usb_errno);
    }

    write_lock (&device->lock);
    usb_close (&device->endp0d);

        /* 
         * close descriptors to device!
         *
         *  The active configuration is the only one allowed to have
         * open descriptors, so we only look on it.
         *
         *  I could use pointer iterators, but it looks cleaner like
         * this, and I guess that optimization would yield in the same
         * results.
         */

    if (device->active)     /* we may have no cfg active yet */
    {
        pendp = device->active->pendp;
        for (cnt = 0; cnt < device->active->pendps; cnt++)
            if (pendp[cnt] && pendp[cnt]->desc)
                usb_close (pendp[cnt]->desc);
    }

#warning FIXME: free the active configuration and other configuraitons?

    write_unlock (&device->lock);

    errcode2 = usb_device_unregister (device);  /* unregister from list */
    if (errcode2)
    {
        gdlf_err ("Couldn't unregister device 0x%p (%d %d): "
                  "errno/usb %d/%d. "
                  "Detaching anyway\n",
                  device, device->hub->host->addr, device->addr,
                  errcode2, usb_errno);
        errcode = errcode2;
    }

    errcode2 = usb_host_unregister_device (device); /* unregister from host */
    if (errcode2)
    {
        gdlf_err ("Couldn't unregister device 0x%p (%d %d) "
                  "from host 0x%p: "
                  "errno/usb %d/%d. "
                  "Detaching anyway\n",
                  device, device->hub->host->addr, device->addr, 
                  device->hub->host, errcode2, usb_errno);
        errcode = errcode2;
    }

    errcode2 = usb_hub_unregister_device (device); /* Unregister from hub */
    if (errcode2)
    {
        gdlf_err ("Couldn't unregister device 0x%p (%d %d) "
                  "from hub 0x%p: "
                  "errno/usb %d/%d. "
                  "Detaching anyway\n",
                  device, device->hub->host->addr, device->addr,
                  device->hub, errcode2, usb_errno);
        errcode = errcode2;
    }

    usb_device_unref_nd (device);
    ts_usb_device_release (device);        /* free data */
    return errcode;
}


    /* Device destructor
     *
     *  This is merely a wrapper to usb_device_detach(), which adds
     * the destruction flag.
     *
     * Parameters:  
     *
     *  ts_usb_device*  Pointer to the device to detach
     *
     * Returns:
     *
     *  int             0 if ok, errno code < 0 on error.
     *                  However, even on error, it is guaranteed that the
     *                  device is going to be detached from the UUSBD
     *                  system. 
     *
     * Global:
     *
     *  int usb_errno   On error, error code.  
     */

int usb_device_destroy (ts_usb_device* device)
{
    return usb_device_detach (device, USB_CFG_DESTROY);
}


/* Log:
 *
 * $Log: usbd-device.c,v $
 * Revision 1.48  1998/11/14 17:15:10  inaky
 * Moved the include files to uusbd/include/uusbd, to ease integration
 * into the Linux kernel.
 *
 * Revision 1.47  1998/11/09 21:29:23  inaky
 * Moved all device ref/unref to usb_device_[un]ref(). Moved all
 * configuration ref/unref to usb_configuration_[un]ref(). Fixed the
 * limits on the endpoint closing function.
 *
 * Revision 1.46  1998/10/31 02:12:37  inaky
 * Removed the device setup stuff, unused.
 *
 * Revision 1.45  1998/08/30 13:50:06  inaky
 * Fixed checks on configuration.
 *
 * Revision 1.44  1998/08/28 23:46:29  inaky
 * Removed traces, made usb_device_detach() only unset the configuration
 * sending a USB request _if_ the device is still physically attached
 * [and reacheable].
 *
 * Revision 1.43  1998/08/27 11:13:57  inaky
 * Moved ts_*_free() to ts_*_release(), made them return void and severed
 * assertions.
 *
 * Added a check to USB unconfigure when detaching only if USB
 * configured. Also made it call dev_setup only if method present in
 * host.
 *
 * Revision 1.42  1998/08/24 02:24:54  inaky
 * Did the little debugging needed to make the functions work as they're
 * expected to.
 *
 * Revision 1.41  1998/08/19 23:51:18  inaky
 * Added a check for device pointer's NULLiness at usb_device_detach()
 *
 * Revision 1.40  1998/08/03 13:31:08  inaky
 * Changed the device attachmen function to opent the descriptor using
 * the normal means (at last, it could be done :). Also the detachment
 * function has been updated to the correct order of events.
 *
 * Revision 1.39  1998/07/30 17:52:47  inaky
 * Moved the device attach function to use the normal means to open a
 * device's endpoint 0 descriptor.
 *
 * Revision 1.38  1998/07/23 14:37:56  inaky
 * We got to free HCI info after releasing it.
 *
 * Revision 1.37  1998/07/23 14:29:49  inaky
 * Updated the device attach function to initialize HCI information. Made
 * the endpoint 0 be closed instead of unset. Device detach releases HCI
 * information.
 *
 * Revision 1.36  1998/07/16 14:10:19  inaky
 * Moved to new enumeration elements name scheme.
 *
 * Revision 1.35  1998/07/14 21:58:51  inaky
 * Fixed usb_device_(de|at)tach(), removing some unnecesary errors (left
 * them as debug messages), fixing the raw open sequence for the endpoint
 * 0 descriptor, fixed some swapped assertions and destruction of device
 * sequences on error (some of them where even missing!).
 *
 * Revision 1.34  1998/07/12 22:52:26  inaky
 * Fixed the destructor to keep the object alive during destruction.
 *
 * Revision 1.33  1998/06/20 21:54:52  inaky
 * Massive changes to split the old monolithic hubd driver into a hub
 * manager (hub wrappers and daemon) and a dynamic driver, which can be
 * replaced and its different parts reused.
 *
 * Revision 1.32  1998/05/24 12:04:14  inaky
 * Fixed some error codes on usb_device_attach(), not to mess with hub
 * driver (manager?) error code
 *
 * Revision 1.31  1998/04/26 03:36:01  inaky
 * Added endpoint 0 data basic initialization and a note for a possible
 * problem.
 *
 * Revision 1.30  1998/04/13 22:10:06  inaky
 * Added flags argument to the usb_device_attach() and _detach()
 * functions. Fixed the destructor, comments and flags passing.
 *
 * Added the device destuctor.
 *
 * Revision 1.29  1998/03/28 19:31:44  inaky
 * Fixed to use the new transfer list code and data.
 *
 * Revision 1.28  1998/02/12 16:35:09  inaky
 * Finished the device ops, and moved synch_frame getting to usbd-endp,
 * where it belongs. Fixed some stuff here and there, and starting to get
 * worried on locks ... I'm gonna have to debug *lots* of deadlocks.
 *
 * Revision 1.27  1998/02/12 13:52:42  inaky
 * Moved the 3slist implementation code to a separate file
 *
 * Revision 1.26  1998/02/10 17:32:11  inaky
 * Moved to use ts_[N]snode_[un]ref() for messing with snodes.
 *
 * Revision 1.25  1998/02/09 13:28:23  inaky
 * Moved all the use counts to atomic
 *
 * Revision 1.24  1998/01/11 15:44:18  inaky
 * Finished functions from the descriptor, inteface and endpoint modules,
 * fixed docs, pretty printed and removed craps.
 *
 * Revision 1.23  1998/01/08 12:19:12  inaky
 * Added copyright strings to all important files
 *
 * Revision 1.22  1998/01/08 11:49:13  inaky
 * Finishing the configuration system. Only the final cfel-stuff is
 * missing to implement; however, nothing is tested.
 *
 * Revision 1.21  1998/01/02 20:51:56  inaky
 * Finished adding the control endpoints check support and reformatting
 * usb.h to be shorter than 80 columns in most places (except the errno
 * codes).
 *
 * Revision 1.20  1998/01/02 19:39:37  inaky
 * Added the definition for endpoint zero to the device structure and
 * removed it from the list of endpoints kept at each configuration. Also
 * updated it on places where needed and removed the concept of basic
 * configuration, as is no longer needed.
 *
 * Revision 1.19  1998/01/02 00:12:17  inaky
 * Finished the new driver scheme. Documented at usbd-drv.c
 *
 * Revision 1.18  1997/12/31 17:06:35  inaky
 * Removed some unneeded locks and fixed some others.
 *
 * Revision 1.17  1997/12/31 16:37:12  inaky
 * Finished the HCI interface
 *
 * Revision 1.16  1997/12/31 05:37:01  inaky
 * Almost finished all the lists and life cycle for hosts, drivers and
 * devices. Some craps to be fixed yet.
 *
 * Revision 1.15  1997/12/31 05:01:48  inaky
 * Moved the list definitions into usbd-device.c
 *
 * Revision 1.14  1997/12/31 04:58:16  inaky
 * Finished defining the scheme for device creation and destruction,
 * via the attach and detach functions.
 *
 * Revision 1.13  1997/12/30 20:15:44  inaky
 * Moved everything to the new lists scheme. Untested yet.
 *
 * Revision 1.12  1997/12/29 00:25:40  inaky
 * Fixed unimplemented functions to return -ENOSYS and moving the khubdd
 * to use signals as commands.
 * Also added function traces on most files that had them disabled (donkey!)
 *
 * Revision 1.11  1997/12/27 19:54:22  inaky
 * Added all the prototypes and skeleton definitions (incomplete) of all
 * of the USBD functions to each source file.
 *
 * Revision 1.10  1997/12/27 17:25:36  inaky
 * Finished moving to GDLF. Deprecated usb_printk and __LOCATION__
 *
 * Revision 1.9  1997/12/27 16:44:25  inaky
 * Moving to GDLF and fixinf docs. Removing more cruft.
 *
 * Revision 1.8  1997/12/27 02:07:04  inaky
 * Moving everything to use the GDLF for logging and debug checkings.
 * Removed the global array of USB hosts and moved it to a linked
 * list. Added an address field to each USB host controller.
 * Removed lots of cruft that were accumulated and useless. Still some
 * missing, but may be useful someday.
 *
 * Revision 1.7  1997/12/27 02:06:00  inaky
 * Moving everything to use the GDLF for logging and debug checkings.
 * Removed the global array of USB hosts and moved it to a linked
 * list. Added an address field to each USB host controller.
 * Removed lots of cruft that were accumulated and useless. Still some
 * missing, but may be useful someday.
 *
 * Revision 1.6  1997/12/08 02:41:01  inaky
 * Fixed some stuff of the #includes, modified the makefile targets to be
 * more flexible, tried to finish the device configuration section.
 *
 * Revision 1.5  1997/12/06 00:19:41  inaky
 * Almost finished the usbd/host, usbd/cfg and a lot of meat from
 * usbd/device. Also added the driver definition. Got to polish it yet
 * ...
 *
 * Revision 1.4  1997/11/30 23:32:35  inaky
 * Finished the basic API. Still missing lotsa functions, but hey
 * ... they'll be shortcoming :)
 *
 * Revision 1.3  1997/10/17 00:16:00  inaky
 * *** empty log message ***
 *
 * Revision 1.2  1997/10/02 21:29:03  inaky
 * Remade all the directory structure to make it cleaner when the USB
 * class drivers come.
 *
 * Revision 1.1  1997/10/01 21:28:12  inaky
 * Re-making a lot of stuff in the hadnling of host, hub and device
 * initialization and handling, following a more correct approach of
 * registering/unregistering. Also moved the usb_host array to be an
 * array of pointers, so dynamic reallocation of it is easier, and the
 * host structure needs not to be moved in memory. The same applies to
 * the device arrays for each host.
 *
 */

