
/* Universal USB Driver Stack v$Revision: 1.34 $ 
 *
 * (C) 1997 by Iaky Prez Gonzlez <inaky@peloncho.fis.ucm.es>
 *
 * Distributed under the terms of the GNU Public License Version 2
 * See files COPYRIGHT and COPYING at the root distribution directory
 *
 * $Id: usbd-drv.c,v 1.34 1998/11/15 23:59:49 inaky Exp $
 *
 * Changelog at the bottom 
 *
 *
 *  USB Driver - Devices - Configuration
 *  ------------------------------------
 *
 *  This is the part of the UUSBD who takes care of device
 * configuration. Please take a look at the document
 * 'Driver-matching.smgl' for reference on inner working of the driver
 * subsystem.
 *
 */

#include <uusbd/usb-misc.h>           /* usb_signature() */
#include <uusbd/usbd.h>

#ifdef CONFIG_KMOD
#include <linux/kmod.h>     /* request_module() */
#endif

#define GDLF_IDENT          "USBD"
#define GDLF_LOCAL          /* Define if want local settings */
#define GDLF_DEF_LOGLEVEL   9
#define GDLF_ASSERTIONS     /* want assertions (#undef if not) */
#undef GDLF_FNTRACE        /* want function traces */
#define GDLF_LOCATIONS      /* want locations in log messages */
#include <uusbd/gdlf.h>


extern ts_3slist devices;   /* reference to the list of devices */
extern ts_3slist unconfs;   /* reference to the list of unconfigured devices */
ts_slist drivers;           /* list of drivers */


    /* Generate an unique id.
     *
     *  This function will take a ts_usb_driver and generate an unique
     * signature for it from it's vicsp and string fields.
     *
     *  If the string pointers are NULL or empty ("\0"), they wont be
     * taken in consideration.
     *
     *  Note that if everything is zero (the vicsp codes, the strings
     * ...), the signature will also be zero, what is marked as an
     * error.
     *
     * Parameters:  
     *
     *  ts_usb_driver*  Pointer to the driver struct for which we want
     *                  to generate the signature.
     *
     * Returns:
     *
     *  int             0 if ok, errno code < 0 on error.
     *
     * Global:
     *
     *  int usb_errno   On error, error code.  
     */

__u32 usb_drv_id (ts_usb_driver* driver)
{
    void *seed, *seedptr;
    size_t seed_len;
    size_t name_len;
    size_t copyright_len;
    size_t vendorname_len;
    size_t productname_len;
    size_t classname_len;
    size_t subclassname_len;
    size_t protocolname_len;    

    gdlf_fntrace ("usb_drv_id (0x%p)", driver);
    
    gdlf_aassert (driver != NULL, return -EINVAL);

    name_len = driver->name? strlen (driver->name) : 0;
    copyright_len = 
        driver->copyright? strlen (driver->copyright) : 0;    
    vendorname_len = 
        driver->vendorname? strlen (driver->vendorname) : 0;
    productname_len = 
        driver->productname? strlen (driver->productname) : 0;
    classname_len = 
        driver->classname? strlen (driver->classname) : 0;
    subclassname_len =
        driver->subclassname? strlen (driver->subclassname) : 0;
    protocolname_len =
        driver->protocolname? strlen (driver->protocolname) : 0;

    seed_len = sizeof (driver->vendor) + sizeof (driver->product) 
        + sizeof (driver->class) + sizeof (driver->subclass)
        + sizeof (driver->protocol) + name_len + copyright_len
        + vendorname_len + productname_len + classname_len
        + subclassname_len + protocolname_len;

    seed = usb_kmalloc (seed_len, GFP_KERNEL);
    if (seed == NULL)
    {
        gdlf_err ("Cannot get %d bytes of memory to "
                  "generate driver's id\n", seed_len);
        usb_errno = -USBE_NO_MEMORY;
        return -ENOMEM;
    }

        /* 
         * I know, this way of copying is awkward, but it evitates
         * having to keep a constant eye on the ts_usb_driver
         * struct. If ya know a better (and faster) way, tell me.
         */

    seedptr = seed;
    memcpy (seedptr, &driver->vendor, sizeof (driver->vendor));
    seedptr += sizeof (driver->vendor);
    memcpy (seedptr, &driver->product, sizeof (driver->product));
    seedptr += sizeof (driver->product);

    memcpy (seedptr, &driver->class, sizeof (driver->class));
    seedptr += sizeof (driver->class);
    memcpy (seedptr, &driver->subclass, sizeof (driver->subclass));
    seedptr += sizeof (driver->subclass);
    memcpy (seedptr, &driver->protocol, sizeof (driver->protocol));
    seedptr += sizeof (driver->protocol);

    if (driver->name)
    {
        memcpy (seedptr, driver->name, name_len);
        seedptr += name_len;
    }    
    if (driver->copyright)
    {
        memcpy (seedptr, driver->copyright, copyright_len);
        seedptr += copyright_len;
    }
    
    if (driver->vendorname)
    {
        memcpy (seedptr, driver->vendorname, vendorname_len);
        seedptr += vendorname_len;
    }
    if (driver->productname)
    {
        memcpy (seedptr, driver->productname, productname_len);
        seedptr += productname_len;
    }
    
    if (driver->classname)
    {
        memcpy (seedptr, driver->classname, classname_len);
        seedptr += classname_len;
    }
    if (driver->subclassname)
    {
        memcpy (seedptr, driver->subclassname, subclassname_len);
        seedptr += subclassname_len;
    }
    if (driver->productname)
    {
        memcpy (seedptr, driver->productname, protocolname_len);
        seedptr += protocolname_len;
    }

    driver->id = usb_signature (seed, seed_len);
    usb_kfree (seed);
    return 0;
}


    /* Calculate the weight of a driver and a vendor/product and cfel codes. 
     *
     * Parameters:  
     *
     *  ts_usb_driver*  Pointer to the driver structure.
     *
     *  __u16           USB Vendor code.
     *
     *  __u16           Product code.
     *
     *  ts_usb_cfel*    Pointer to the configurable element structure.
     *
     * Returns:
     *
     *  int             >= 0 weight if ok, errno code < 0 on error.
     *  
     * Global:
     *
     *  int usb_errno   On error, error code.  
     */

static inline
int usb_drv_weight (ts_usb_driver* driver, __u16 vendor, 
                    __u16 product, ts_usb_cfel* cfel)
{
    int weight;
    int v, i, vi;

    gdlf_fntrace ("usb_drv_weight (0x%p, 0x%04x, 0x%04x, 0x%p)",
                  driver, vendor, product, cfel);
    gdlf_aassert (cfel != NULL, return -EINVAL);
    gdlf_aassert (driver != NULL, return -EINVAL);

    v = driver->vendor == vendor;
    i = driver->product == product;
    vi = (v || !driver->vendor) && (i || !driver->product);
        
#define c (cfel->class                                  \
           && ((driver->class == cfel->class)? 1 : 0)   \
           && vi)

#define s (cfel->subclass                                       \
           && ((driver->subclass == cfel->subclass)? 1 : 0)     \
           && vi)

#define p (cfel->protocol                                       \
           && ((driver->protocol == cfel->protocol)? 1 : 0)     \
           && vi)

    weight = v << 4 | i << 3 | c << 2 | s << 1 | p;

#undef c
#undef s
#undef p

    return weight;
}


    /* Compute the best weight for a configurable element
     *
     * Parameters:  
     *
     *  ts_usb_driver** Pointer to where to store the pointer to the
     *                  driver who better matches (gives best
     *                  weight). If on return, it is NULL, it means
     *                  none matched and the weight returned is zero.
     *
     *  __u16           USB Vendor ID
     *
     *  __u16           USB Product id
     *
     *  ts_usb_cfel*    Pointer to the configurable element.
     *
     *  ts_usb_driver*  Pointer to last driver seen. This is used to
     *                  mark where to start searching drivers. Thus,
     *                  if a call to usb_drv_match() yields driver A,
     *                  A will be the first driver which matched, but
     *                  there may be more behind in the list. Thus,
     *                  next time you call usb_drv_match() with this
     *                  parameter set to A. Then the search will start
     *                  at this driver and not at the head of the
     *                  list. Thus, you can guarantee you can reach
     *                  all drivers. If NULL, it'll start on the head
     *                  of the list.
     *
     * Returns:
     *
     *  int             <0  error, errno code
     *                  N   Maximum weight given by the driver
     *                      returned in the location pointed to by the
     *                      first parameter. If zero, none matched and
     *                      the returned driver doesn't need to be
     *                      zero, as it may be one of the generic
     *                      drivers (example rawd or hooks).
     *
     * Global:
     *
     *  int usb_errno   On error, error code.  
     */

int usb_drv_match (ts_usb_driver** pdriver,
                   __u16 vendor, __u16 product, 
                   ts_usb_cfel* cfel, 
                   ts_usb_driver* last)
{
    ts_usb_driver *max_driver, *driver;
    int weight, max_weight, zero;

    gdlf_fntrace ("usb_drv_match (0x%p, 0x%04x, 0x%04x, 0x%p, 0x%p)",
                  pdriver, vendor, product, cfel, last);
    gdlf_aassert (cfel != NULL, return -EINVAL);
    gdlf_aassert (pdriver != NULL, return -EINVAL);

        /* check 'last' driver is in the drivers list */
    
    if (last && ts_slist_prev (&drivers, &last->snode) == NULL)
    { 
        gdlf_err ("Driver 0x%p isn't registered "
                  "within the drivers list; "
                  "Cannot use it as start point.\n",
                  last);
        usb_errno = -USBE_DRIVER_UNREFERENCED; 
        return -ENXIO; 
    }

        /* let's check the list for a suitable driver */

    read_lock (&drivers->lock);
    max_driver = NULL;
    weight = max_weight = 0;
    
    if (last)
        driver = (ts_usb_driver*)ts_slist_next (&last->snode);
    else    /* last != null, start on head */
        driver = (ts_usb_driver*)ts_slist_head (&drivers);
    
    while (driver)
    {
        zero = !driver->vendor && !driver->product
            && !driver->class && !driver->subclass
            && !driver->protocol;
        if (zero)       /* fully generic driver! try it */
        {
            max_weight = 0;
            max_driver = driver;
            break;
        }
        weight = usb_drv_weight (driver, vendor, product, cfel);
#warning DEBUG: remove me
        gdlf_debug ("Considering driver 0x%p `%s', weight %d\n",
                    driver, driver->name? driver->name : "(name n/a)", 
                    weight);
        
            /* 
             *  Note this. If we just did it doing an >= comparison,
             * we would get the same effect, but we'd also break the
             * driver's list selection order for same weights and
             * wouldn't be able of using the `last' parameter.
             */

        if (weight > max_weight)    /* a maximum? */
        {
            max_weight = weight;        /* got it */
            max_driver = driver;        /* actualize */
        }
        driver = (ts_usb_driver*) ts_slist_next (&driver->snode);
    }
    read_unlock (&drivers->lock);

    *pdriver = max_driver;
    return max_weight;
}


    /* Find an exact driver
     *
     *  Feed it with the USB vendor ID, product ID and
     * class/subclass/protocol codes and it'll find a driver exactly
     * matching that.
     *
     * Parameters:  
     *
     *  __u16           USB Vendor ID
     *
     *  __u16           USB Product id
     *
     *  ts_usb_cfel*    Pointer to the configurable element with the
     *                  vicsp data.
     *
     * Returns:
     *
     *  ts_usb_driver*  Pointer to the driver which matchs exactly the
     *                  data given or NULL if not present in the
     *                  drivers list.
     */

ts_usb_driver* usb_drv_find (__u16 vendor, __u16 product, 
                             ts_usb_cfel* cfel)
{
    ts_usb_driver *driver;

    gdlf_fntrace ("usb_drv_find (0x%04x, 0x%04x, 0x%p)",
                  vendor, product, cfel);
    gdlf_aassert (cfel != NULL, return NULL);

        /* let's check the list for a suitable driver */

    read_lock (&drivers->lock);

    driver = (ts_usb_driver*)ts_slist_head (&drivers);
    while (driver && 
           (driver->vendor != vendor
            || driver->product != product
            || driver->class != cfel->class
            || driver->subclass != cfel->subclass
            || driver->protocol != cfel->protocol)
        )
        driver = (ts_usb_driver*) ts_slist_next (&driver->snode);
    read_unlock (&drivers->lock);

    return driver;
}


    /*  Check a USB driver structure to be ok for
     * registration/unregistration.
     *
     * Parameters:  
     *
     *  ts_usb_driver*  Pointer to the driver structure
     *
     * Returns:
     *
     *  int             0 if ok, errno code < 0 on error.
     *
     * Global:
     *
     *  int usb_errno   On error, error code.  
     */

int usb_drv_check (ts_usb_driver *driver)
{
    gdlf_fntrace ("usb_drv_check (0x%p)", driver);

    gdlf_aassert (driver != NULL, return -EINVAL);
    
    if (driver->config == NULL)
    {
        gdlf_err ("Driver 0x%p doesn't have a "
                  "configuration method\n", driver);
        usb_errno = -USBE_DRIVER_NULL_CONFIG;
        return -EINVAL;
    }
    
    if (driver->unconfig == NULL)
    {
        gdlf_err ("Driver 0x%p doesn't have an "
                  "unconfiguration method\n", driver);
        usb_errno = -USBE_DRIVER_NULL_UNCONFIG;
        return -EINVAL;
    }

    return 0;
}


    /* Walk the list of unconfigured devices with a driver
     *
     *  This function takes a driver and walks the list of
     * unconfigured devices, looking for anything that could be
     * configured by it.
     *
     * Parameters:  
     *
     *  ts_usb_driver*  Pointer to the driver structure to use for
     *                  configuring.
     *
     * Returns:
     *
     *  int             0 if ok, errno code < 0 on error.
     *
     * Global:
     *
     *  int usb_errno   On error, error code.  
     */

int usb_drv_walk_unconfs (ts_usb_driver* driver)
{
#if 0
    ts_usb_device *itr;
    int zero;
#endif
    
    gdlf_fntrace ("usb_drv_walk_unconfs (0x%p)", driver);
#warning FIXME: usb_drv_walk_unconfs() unfinished!
    return -ENOSYS;
    
#if 0
    zero = !driver->vendor && !driver->product
        && !driver->class && !driver->subclass
        && !driver->protocol;

    if (driver->config == NULL)
    {
        gdlf_err ("Driver 0x%p doesn't have a "
                  "configuration method\n", driver);
        usb_errno = -USBE_DRIVER_NULL_CONFIG;
        return -EINVAL;
    }

    write_lock (&unconfs->lock);
    itr = (ts_usb_device*) ts_3slist_head (&unconfs);
    while (itr)
    {
        if (!zero && !)
    }
    write_unlock (&unconfs->lock);
    return 0;
#endif
}


    /* Searches the list of drivers for one with same attributes
     *
     *  This function will take driver's signature id and searchs for an
     * already registered driver structure, returning it. NULL if none
     * found.
     *
     * Parameters:  
     *
     *  __u32           Id code for the driver.
     *
     * Global:
     *
     *  ts_slist drivers    Drivers list.
     *
     * Returns:
     *
     *  ts_usb_driver*  Pointer to the driver structure found or NULL
     *                  if none found.
     *
     */

ts_usb_driver* usb_drv_registered (__u32 id)
{
    ts_usb_driver* itr;

    gdlf_fntrace ("usb_drv_registered (0x%08x)", id);

    read_lock (&drivers->lock);
    itr = (ts_usb_driver*) ts_slist_head (&drivers);
    while (itr)
    {
        if (itr->id == id)
        {
            gdlf_debug ("Driver for ID 0x%08x "
                        "already known as "
                        "0x%p (%04x.%04x.%02x.%02x.%02x)\n", 
                        id, itr, itr->vendor, itr->product, 
                        itr->class, itr->subclass, itr->protocol);
            read_unlock (&drivers->lock);
            return itr;
        }
        itr = (ts_usb_driver*) ts_slist_next (&itr->snode);
    }
    read_unlock (&drivers->lock);
    return NULL;
}


    /* Registers a driver into the USB system.
     *
     *  This function is used to register a driver into the USB system
     * by giving a driver structure filled in. The structure MUST be
     * dinamically allocated, and will be owned by the driver
     * subsystem. It'll be freed automagically, so once you register a
     * driver structure, you don't need to worry about it any more.
     *
     *  However, on driver unload, you still have to unregister, to
     * let UUSBD know you're not available any more.
     *
     *  After succesful registration, the function will walk the list
     * of unconfigured devices to try to look for something that could
     * be configured by it. Note that we will ignore any kind of
     * errors on this stage, as they "should" be unrelated to the act
     * of driver loading.
     *
     * NOTE: It may seem impractical or suboptimal to allocate and
     *       fill an structure and after, if an already registered one
     *       is found, use the new one and free the old. This is done
     *       because the driver itself may be helding pointers to the
     *       structure, and thus, the cleaner way is to use the new
     *       one. Also is a way to insure that always the check is
     *       done (by this function) to wether the structure is
     *       already registered [and so to evitate mis-behaving
     *       drivers].
     *
     * Parameters:  
     *
     *  ts_usb_driver*  Pointer to the driver structure to register.
     *
     * Returns:
     *
     *  int             0 if ok, errno code < 0 on error.
     *
     * Global:
     *
     *  int usb_errno   On error, error code.  
     */

int usb_drv_register (ts_usb_driver* driver)
{
    int errcode;
    ts_usb_driver* oth;
    
    gdlf_fntrace ("usb_drv_register (0x%p)", driver);

    errcode = usb_drv_check (driver);   /* check it */
    if (errcode)
        return errcode;
    errcode = usb_drv_id (driver);      /* calculate id */
    if (errcode)
        return errcode;

        /* 
         * See if we've got to replace the driver structure with a new
         * one. Note we first search for it, _insert_ the new one (so
         * it's used instead of the old one if someone happens to
         * request it in the meantime) and then remove the old one if
         * we have to. If we searched for the old one after adding the
         * new, we'd always find and try to replace it.
         */
    
    oth = usb_drv_registered (driver->id);
    ts_dtor_fill (&driver->snode.dtor,          /* add destructor */
                  (tf_dtor) usb_drv_destroy);
    ts_slist_prep (&drivers, &driver->snode);   /* add the driver */

    if (oth != NULL)                            /* remove previous */
    {
        if (oth->config != NULL)                    /* registered, but */
        {                                           /* not unloaded? */
            gdlf_err ("Jarl? The same driver (%s) "
                      "has been reloaded twice ???\n",
                      driver->name);
        }
#warning DEBUG: remove me
        gdlf_debug ("Use count for oth 0x%p is %d\n",
                    oth, atomic_read (&oth->snode.usecnt));
        gdlf_assert (atomic_read (&oth->snode.usecnt) == 1);
        ts_slist_del (&drivers, &oth->snode);   /* usecnt should drop to 0 */
    }

        /*
         *  Now move the new driver entry to the list's tail. When
         * matching a driver against a cfel, if we reach the tail, we
         * load a new, possibly suitable one and it that one fails, we
         * need to be able to check which was the last one loaded so
         * we try next possibly suitable driver.
         */

    usb_drv_walk_unconfs (driver);
    return 0;
}


    /* Enforce the unconfiguration of any device/interface using a
     * driver. 
     *
     *  This function will enforce the unconfiguration of any device
     * or interface which is using the given driver, calling the
     * destructors of the structures which uses it.
     *
     *  This function may be safely used previous to a forced
     * usb_drv_unregister() to unload a driver, as the transfers on
     * the affected target are cleanly aborted. Also this function
     * guarantees that the use count for the driver will be one (the
     * reference which keeps the drivers list). However,
     * the operation of release && unregister should be atomic to
     * prevent possible uses from any other devices.
     *
     * Parameters:  
     *
     *  ts_usb_driver*  Pointer to the driver structure to register.
     *
     * Returns:
     *
     *  int             0 if ok, errno code < 0 on error.
     *
     * Global:
     *
     *  int usb_errno   On error, error code.  
     */

int usb_drv_release (ts_usb_driver* driver)
{
    int             errcode;
    ts_usb_device   *device;

    gdlf_fntrace ("usb_drv_release (0x%p)", driver);
    
    gdlf_apassert (driver != NULL,
                   {
                       usb_errno = -USBE_DRIVER_NULL; 
                       return -EINVAL; 
                   },
                   "Cannot release a NULL ts_usb_driver\n");
    
    if (atomic_read (&driver->snode.usecnt) < 2)/* if only used by the list or unused */
        return 0;                               /* ... we can move */    

    read_lock (&devices->lock);
    device =                        /* get the head of the device list */
        (ts_usb_device*)ts_3slist_head (&devices);
    while (device)
    {
        errcode = driver->unconfig (device, USB_CFG_RELEASE);/* unconfigure */
        
        if (errcode >= 1                        /* partially unconfigured */
            && device->cfelu != 1)              /* and not on unconf list */
            ts_3slist_prep (&unconfs, &device->msnode);  /* add to unconfs */

        if (errcode < 0)
            gdlf_err ("Couldn't determine if unconfiguration "
                      "was neccesary on device 0x%p by driver 0x%p: "
                      "errno %d\n", device, driver, errcode);

        device = 
            (ts_usb_device*)ts_3slist_next (&devices, &device->msnode);
    }
    read_unlock (&devices->lock);

    return 0;
}


    /* Unregister a driver from the system.
     *
     *  This function will just check the driver is more or less ok
     * and unset all the fields that could cause trouble if let set on
     * module unload. Then it will exit.
     *
     *  Remember that on the module's system, driver's entry is kept
     * in main memory to keep track of which drivers are available.
     *
     * Parameters:  
     *
     *  ts_usb_driver*  Pointer to the driver structure to register.
     *
     * Returns:
     *
     *  int             0 if ok, errno code < 0 on error.
     *
     * Global:
     *
     *  int usb_errno   On error, error code.  
     */

int usb_drv_unregister (ts_usb_driver *driver)
{
    gdlf_fntrace ("usb_drv_unregister (0x%p)", driver);
    gdlf_aassert (driver != NULL, return -EINVAL);

    ts_slist_prev (&drivers, &driver->snode);
    driver->name =
        driver->vendorname = 
        driver->productname = 
        driver->classname = 
        driver->subclassname = 
        driver->protocolname = NULL;    
    driver->config = NULL; 
    driver->unconfig = NULL;
    return 0;
}


    /* Release and unregister a USB driver
     *
     *  This is a destructor, and so, it's got to completely destroy a
     * ts_usb_driver struct, releasing, removing it from the drivers
     * list, and later freeing it. There's no need to unregister, as
     * it will only clean some fields.
     *
     * Parameters:  
     *
     *  ts_usb_driver*  Pointer to the driver structure.
     *
     * Returns:
     *
     *  int             0 if ok, errno code < 0 on error.
     *
     * Global:
     *
     *  int usb_errno   On error, error code.  
     */

int usb_drv_destroy (ts_usb_driver *driver)
{
    gdlf_fntrace ("usb_drv_destroy (0x%p)", driver);

    gdlf_apassert (driver != NULL,
                   {
                       usb_errno = -USBE_DRIVER_NULL; 
                       return -EINVAL; 
                   },
                   "Cannot destroy of a NULL "
                   "ts_usb_driver\n");

        /* All the functions complain themselves */

    ts_snode_ref (&driver->snode);
    usb_drv_release (driver);
    ts_slist_del (&drivers, &driver->snode);
    ts_snode_unref_nd (&driver->snode);
    ts_usb_driver_release (driver);
    return 0;
}


    /* Load a driver suitable for a given vendor, product id, class,
     * subclass and protocol codes masked with a vicsp mask.
     *
     *  Given the vicsp codes and a vicsp mask (`vicsp' parameter),
     * compose a module name and try to load it.
     *
     * Parameters:  
     *
     *  int             vicsp mask for composing the module name.
     *
     *  __u16           USB vendor code
     *
     *  __u16           vendor product code
     *
     *  ts_usb_cfel*    Configurable element. Class, subclass and
     *                  protocol codes are taken from here.
     *
     * Returns:
     *
     *  ts_usb_driver*  Pointer to the new driver structure if
     *                  successfully loaded, NULL if couldn't load
     *                  it.
     */

ts_usb_driver* usb_drv_load (int vicsp, __u16 vendor, __u16 product,
                             ts_usb_cfel* cfel)
{
    __u16 v;
    __u16 i;
    __u8 c;
    __u8 s;
    __u8 p;
    
    char module_name[4+4+1+4+1+2+1+2+1+2+1];

    gdlf_fntrace ("usb_drv_load (0x%02x, 0x%04x, 0x%04x, 0x%p)", 
                  vicsp, vendor, product, cfel);
    gdlf_aassert (cfel != NULL, return NULL);

#ifdef CONFIG_KMOD
    v = (vicsp & 0x10)? vendor : 0;
    i = (vicsp & 0x08)? product : 0;
    c = (vicsp & 0x04)? cfel->class : 0;
    s = (vicsp & 0x02)? cfel->subclass : 0;
    p = (vicsp & 0x01)? cfel->protocol : 0;
    
    sprintf (module_name, "usb-%04x.%04x.%02x.%02x.%02x",
             v, i, c, s, p);
    if (request_module (module_name) == 0)
    {
        ts_usb_driver *driver;
        ts_usb_cfel cfel;
        
        cfel.class = c;
        cfel.subclass = s;
        cfel.protocol = p;
        
        driver = usb_drv_find (v, i, &cfel);  /* huh? not found */
        if (driver == NULL)
            gdlf_err ("Failed to load driver "
                      "'usb-%04x.%04x.%02x.%02x.%02x'; "
                      "it is not on the drivers list\n", 
                      v, i, c, s, p);
        return driver;
    }
    else    /* Load failed, driver isn't available anymore */
    {
        gdlf_err ("Failed to load driver "
                  "'usb-%04x.%04x.%02x.%02x.%02x'\n",
                  v, i, c, s, p);
        return NULL;
    }
#else
#warning KMOD disabled, dynamic load of USB drivers disabled
    return NULL;
#endif /* #ifdef CONFIG_KMOD */
}



    /* Reload a driver which isn't currently loaded, but we know about
     * it.
     *
     *  This function will take a driver structure in the unloaded
     * state and will try to load the module whose driver it refers
     * to. If failure loading, it'll be assumed the module is not
     * available anymore and will be removed from the list of
     * available drivers.
     *
     * Parameters:  
     *
     *  ts_usb_driver*  Pointer to the driver structure we want to
     *                  reload/check. 
     *
     * Returns:
     *
     *  ts_usb_driver*  Pointer to the new driver structure if
     *                  reloaded, the same one if already loaded, NULL
     *                  if couldn't reload it (and get's removed from
     *                  the list of drivers.
     */

ts_usb_driver* usb_drv_reload (ts_usb_driver* driver)
{
    char module_name[4+4+1+4+1+2+1+2+1+2+1];

    gdlf_fntrace ("usb_drv_reload (0x%p)", driver);

#ifdef CONFIG_KMOD
    if (driver == NULL)
        return NULL;
    
    if (driver->config != NULL) /* check it's already loaded */
        return driver;
    
    ts_snode_ref (&driver->snode);
    sprintf (module_name, "usb-%04x.%04x.%02x.%02x.%02x",
             driver->vendor, driver->product,
             driver->class, driver->subclass, driver->protocol);
    if (request_module (module_name) == 0)
    {
        ts_usb_driver *itr;

            /*
             *  Reloaded OK, now we gotta search new structure. We'll
             * just walk the driver's list searching for the driver
             * ID, as we've got it from the actual driver.
             *
             *  When reloaded, the old structure (our driver) has been
             * already removed from the drivers list and only the new
             * one is there.
             */

        itr = (ts_usb_driver*) ts_slist_head (&drivers);
        while (itr && itr->id != driver->id)
            itr = (ts_usb_driver*) ts_slist_next (&itr->snode);
        if (itr == NULL)            /* huh? not found */
        {
            gdlf_err ("Error re-loading driver "
                      "'usb-%04x.%04x.%02x.%02x.%02x', seemed "
                      "to re-load fine, but now it is not on the "
                      " drivers list\n",
                      driver->vendor, driver->product,
                      driver->class, driver->subclass,
                      driver->protocol);
        }
        ts_snode_unref (&driver->snode, NULL);
        return itr;
    }
    else    /* Load failed, driver isn't available anymore */
    {
        gdlf_warning ("Driver 'usb-%04x.%04x.%02x.%02x.%02x' isn't "
                      "available for loading anymore\n",
                      driver->vendor, driver->product,
                      driver->class, driver->subclass,
                      driver->protocol);
        ts_snode_unref (&driver->snode, NULL);
        usb_drv_destroy (driver);
        return NULL;
    }
#else
#warning KMOD disabled, dynamic load of USB drivers disabled
    return NULL;
#endif /* #ifdef CONFIG_KMOD */
}


    /* Add/remove a reference to a driver
     *
     *  This function is used to tell a driver if someone is or not
     * using any of it's data structures or functions. When we take a
     * pointer to any of the driver's stuff, we must reference (use)
     * it, so we call this with a positive argument. When we are done
     * using it, we call it with a negative argument.
     *
     *  Note this won't affect the module use count if using
     * modules. For that, wrap this function (use as an example what
     * the hub driver does)
     *
     * Parameters:  
     *
     *  ts_usb_driver*  Pointer to the driver structure which we use.
     *
     *  ts_snode*       Pointer to the element which uses the driver's
     *                  stuff.
     *
     *  __u8            Positive for ONE reference, negative for ONE
     *                  unreference. If zero, return use count.
     *
     * Returns:
     *
     *  int             0 if ok, errno code < 0 on error.
     */
    
void usb_drv_release_use (ts_usb_driver* driver,
                          ts_usb_drv_use* use, ts_usb_drv_use* prev)
{    
    gdlf_fntrace ("usb_drv_release_use (0x%p, 0x%p, 0x%p)", 
                  driver, use, prev);

    ts_slist_del_fast (&driver->uselist,    /* remove from list */
                       &use->snode, 
                       prev? &prev->snode : &use->snode);
}

int usb_drv_reference (ts_usb_driver* driver, ts_snode* user, int no)
{
    ts_usb_drv_use *use;
    
    gdlf_fntrace ("usb_drv_reference (0x%p, 0x%p, %d)", 
                  driver, user, no);
    gdlf_aassert (driver != NULL, return -EINVAL);
    gdlf_aassert (user != NULL, return -EINVAL);
    
    if (no > 0)             /* reference */
    {
        use = ts_usb_drv_use_init();       /* create a use node */
        if (use == NULL)
            gdlf_err ("Cannot add user 0x%p to the "
                      "user's list of driver 0x%p (%s). "
                      "Expect possible memory leaks\n", 
                      user, driver, driver->name);
        else
        {
            ts_snode_ref (user);                /* reference the device */
            use->user = user;                   /* store it */
            ts_slist_prep (&driver->uselist,    /* add to the list */
                           &use->snode);
        }
    }
    else if (no < 0)        /* unreference */
    {
        ts_usb_drv_use *prv;

            /* Search the device in the list, ignore if unfound */
        
        use = (ts_usb_drv_use*) ts_slist_head (&driver->uselist);
        prv = NULL;
        while (use && use->user != user)
        {
            prv = use;
            use = (ts_usb_drv_use*) ts_slist_next (&use->snode);
        }

        if (use)                            /* found, let's remove it */
            usb_drv_release_use (driver, use, prv);
    }
    return atomic_read (&driver->uselist.usecnt);
}

/* Log:
 *
 * $Log: usbd-drv.c,v $
 * Revision 1.34  1998/11/15 23:59:49  inaky
 * Fixed some of the configuration and driver woes, as there was
 * problems, mainly caused by the ts_slist_move() function. They seem
 * fixed now. Added a code to skip when loading already loaded
 * modules.
 *
 * Revision 1.33  1998/11/14 17:15:11  inaky
 * Moved the include files to uusbd/include/uusbd, to ease integration
 * into the Linux kernel.
 *
 * Revision 1.32  1998/11/09 21:30:09  inaky
 * Disabled some debug output until I have time to finish fixing it.
 *
 * Revision 1.31  1998/08/30 13:50:35  inaky
 * Added some debug code to try to track the driver problem when it is
 * loaded and another device searches it.
 *
 * Revision 1.30  1998/08/28 23:47:00  inaky
 * Removed traces, severed assertions and a check.
 *
 * Revision 1.29  1998/08/28 22:32:37  inaky
 * Removed all the loading logic from usb_drv_match() and moved it to
 * usb_cfel_config(), being it now much cleaner and well-working.
 *
 * Revision 1.28  1998/08/27 22:15:35  inaky
 * Removed traces.
 *
 * Revision 1.27  1998/08/27 11:12:34  inaky
 * Moved ts_*_free() to ts_*_release(), made them return void and severed
 * assertions.
 *
 * Revision 1.26  1998/08/25 04:03:34  inaky
 * Fixed usb_drv_match() to take a vicsp mask. This helps when doing
 * module load scan. Now when a device cannot be configured from the
 * driver's list, it scans for a suitable driver, requesting the various
 * possible modules to be loaded. Added a driver loading function to
 * accomplish this.
 *
 * Revision 1.25  1998/08/24 02:25:15  inaky
 * Did the little debugging needed to make the functions work as they're
 * expected to.
 *
 * Revision 1.24  1998/07/12 22:52:49  inaky
 * Fixed the destructor to keep the object alive during destruction.
 *
 * Revision 1.23  1998/06/21 15:30:27  inaky
 * Added the reference handling code.
 *
 * Revision 1.22  1998/06/20 21:54:52  inaky
 * Massive changes to split the old monolithic hubd driver into a hub
 * manager (hub wrappers and daemon) and a dynamic driver, which can be
 * replaced and its different parts reused.
 *
 * Revision 1.21  1998/04/13 22:10:35  inaky
 * Moved to use kmod instead of kerneld.
 *
 * Revision 1.20  1998/03/28 19:33:46  inaky
 * Added the hex number codes to the snprintf that makes up the module
 * name for requestion [they were missing].
 *
 * Revision 1.19  1998/03/02 22:22:56  inaky
 * Checking in for administrative purposes
 *
 * Revision 1.18  1998/02/12 13:48:33  inaky
 * Added the function to generate unique id's for each driver, and
 * modified to use it in searching for registered driver structures when
 * reloading them. Also fixed to check on reloading of a driver if the
 * old driver structure is really loaded or not. Also added the specific
 * driver reload function.
 * Fixed the hook for generic (0.0.0.0.0) drivers.
 *
 * Revision 1.17  1998/02/10 17:34:58  inaky
 * Seems it works now. Some stuff is incomplete yet, but the basics seem
 * to go ok.
 *
 * Revision 1.16  1998/01/11 15:44:18  inaky
 * Finished functions from the descriptor, inteface and endpoint modules,
 * fixed docs, pretty printed and removed craps.
 *
 * Revision 1.15  1998/01/08 12:19:12  inaky
 * Added copyright strings to all important files
 *
 * Revision 1.14  1998/01/08 11:49:13  inaky
 * Finishing the configuration system. Only the final cfel-stuff is
 * missing to implement; however, nothing is tested.
 *
 * Revision 1.13  1998/01/02 19:39:38  inaky
 * Added the definition for endpoint zero to the device structure and
 * removed it from the list of endpoints kept at each configuration. Also
 * updated it on places where needed and removed the concept of basic
 * configuration, as is no longer needed.
 *
 * Revision 1.12  1998/01/02 00:12:17  inaky
 * Finished the new driver scheme. Documented at usbd-drv.c
 *
 * Revision 1.11  1998/01/01 21:36:59  inaky
 * Removed an stupid check to avoid driver structures with an usecount >
 * 0 not to be allowed to register.
 *
 * Revision 1.10  1997/12/31 05:22:04  inaky
 * Removed some cruft from the old scheme of devices list
 *
 * Revision 1.9  1997/12/31 05:08:31  inaky
 * Added the destructor for the ts_usb_driver structure
 *
 * Revision 1.8  1997/12/30 20:15:44  inaky
 * Moved everything to the new lists scheme. Untested yet.
 *
 * Revision 1.7  1997/12/29 00:25:38  inaky
 * Fixed unimplemented functions to return -ENOSYS and moving the khubdd
 * to use signals as commands.
 * Also added function traces on most files that had them disabled (donkey!)
 *
 * Revision 1.6  1997/12/27 17:25:35  inaky
 * Finished moving to GDLF. Deprecated usb_printk and __LOCATION__
 *
 * Revision 1.5  1997/12/27 02:07:03  inaky
 * Moving everything to use the GDLF for logging and debug checkings.
 * Removed the global array of USB hosts and moved it to a linked
 * list. Added an address field to each USB host controller.
 * Removed lots of cruft that were accumulated and useless. Still some
 * missing, but may be useful someday.
 *
 * Revision 1.4  1997/12/13 20:56:32  inaky
 * Moving all the documentation to the .c files except the ones for data
 * structures and inline functions. Helps making the header files cleaner.
 *
 * Revision 1.3  1997/12/08 02:41:00  inaky
 * Fixed some stuff of the #includes, modified the makefile targets to be
 * more flexible, tried to finish the device configuration section.
 *
 * Revision 1.2  1997/12/06 01:55:48  inaky
 * Now finishing the configuration system
 *
 * Revision 1.1  1997/12/06 00:19:40  inaky
 * Almost finished the usbd/host, usbd/cfg and a lot of meat from
 * usbd/device. Also added the driver definition. Got to polish it yet
 * ...
 *
 */
