
/* Universal USB Driver Stack v$Revision: 1.14 $ 
 *
 * (C) 1997 by Iaky Prez Gonzlez <inaky@peloncho.fis.ucm.es>
 *
 * Distributed under the terms of the GNU Public License Version 2
 * See files COPYRIGHT and COPYING at the root distribution directory
 *
 * $Id: usbd-endp.c,v 1.14 1998/11/14 17:15:12 inaky Exp $
 *
 * Changelog at the bottom 
 *
 *
 *  USB Driver - Devices section - USB Endpoint operations
 *  ------------------------------------------------------
 *
 * Standard and common operations on USB endpoints.
 */

#include <uusbd/usbd.h>

#define GDLF_IDENT          "USBD"
#define GDLF_LOCAL          /* Define if want local settings */
#define GDLF_DEF_LOGLEVEL   9
#define GDLF_ASSERTIONS     /* want assertions (#undef if not) */
#undef GDLF_FNTRACE        /* want function traces */
#define GDLF_LOCATIONS      /* want locations in log messages */
#include <uusbd/gdlf.h>


    /* Dirty-work function for usb_endp_feature_*()
     *
     *  THIS FUNCTION IS FOR INTERNAL USE, DO NOT CALL IT DIRECTLY
     *
     * Parameters:  
     *
     *  ts_usb_device*  Pointer to the device structure where the
     *                  endpoint is located.
     *
     *  t_usb_endp      Endpoint number
     *
     *  e_usb_featsel   Feature
     *
     *  e_usb_req       USB request code
     *
     * Returns:
     *
     *  int             0 if ok, errno code < 0 on error.
     *
     * Global:
     *
     *  int usb_errno   On error, error code.  
     */

int usb_endp_feature (ts_usb_device* device, t_usb_endp endpn, 
                       e_usb_featsel feature, e_usb_req request)
{
    ssize_t errcode;
    ts_usb_request req;
    
    gdlf_fntrace ("usb_endp_feature (0x%p, %u, %d, %u)",
                  device, endpn, feature, request);

    gdlf_aassert (device != NULL, return -EINVAL); 

    read_lock (&device->msnode.lock);
    if (device->active == NULL)     /* unconfigured? */
    {
        gdlf_err ("Device 0x%p isn't configured yet.\n",
                  device);
        usb_errno = -USBE_CFG_NULL;
        read_unlock (&device->msnode.lock);
        return -EPERM;
    }

    if (device->active->pendp[endpn] == NULL)
    {
        gdlf_err ("Endpoint with index %d undefined "
                  "for USB configuration %d "
                  "of device (%d, %d)\n",
                  endpn, device->active->addr, 
                  device->hub->host->addr, 
                  device->addr);
        usb_errno = -USBE_ENDP_BAD_INDEX;
        read_unlock (&device->msnode.lock);
        return -EINVAL; 
    }
    
    req.bmreqtype = 0x2;
    req.request = request;
    req.value = feature;
    req.index = endpn;

    errcode = usb_request_dataless (device, &req, 
                                    "endpoint feature change");
    if (errcode < 0)
        gdlf_err ("Cannot change endpoint feature; request failed: "
                  "errno/usb %d/%d\n", errcode, usb_errno);
    read_unlock (&device->msnode.lock);
    return errcode;
}


    /* Return the status of an endpoint
     *
     *  Obtains the endpoints status, doing byte order conversion if
     * needed and actualizing endpoint status data.
     *
     * Parameters:  
     *
     *  ts_usb_open*    Open descriptor of the endpoint whose status
     *                  we want to know.
     *  
     * Returns:
     *
     *  __s32           If > 0, the status code (and with
     *                  USB_ENDP_STAT_* to get individual bits). Errno
     *                  code < 0 on error. 
     *
     * Global:
     *
     *  int usb_errno   On error, error code.  
     */

__s32 usb_endp_status_get (ts_usb_open* open)
{
    int errcode;
    ts_usb_request req;
    __u16 status;
    
    gdlf_fntrace ("usb_endp_status_get (0x%p)", open);

    gdlf_aassert (open != NULL, return -EINVAL);

    errcode = usb_dst_check (open);
    if (errcode)
        return errcode;
    
    req.bmreqtype = 0x82;
    req.request = e_usb_req_get_status;
    req.value = 0;
    req.index = open->endp->addr;
    req.length = sizeof (status);
    
    write_lock (&open->dev->msnode.lock);
    status = 0;
    errcode = usb_request_data (open->dev, &req, &status, 
                                "get endpoint status");
    if (errcode < 0)
    {
        gdlf_err ("Cannot get endpoint status; request failed: "
                  "errno/usb %d/%d\n", errcode, usb_errno);
        write_unlock (&open->dev->msnode.lock);
        return errcode;
    }

    status = le16_to_cpu (status);      /* change endianess */
    if (status & USB_ENDP_STAT_STALL)
        open->endp->stat = e_usb_tr_status_error;
    write_unlock (&open->dev->msnode.lock);
    return status;
}


    /* Get the synchronization frame on an isochronous endpoint
     *
     *  This function is used to get the synch frame on an
     * isochronous endpoint. 
     *
     * Parameters:  
     *
     *  ts_usb_open*    Open descriptor of the endpoint whose
     *                  synch frame we want to set.
     *
     * Returns:
     *
     *  __s32           __u16 synch frame if ok, errno code < 0 on
     *                  error. 
     *
     * Global:
     *
     *  int usb_errno   On error, error code.  
     */

__s32 usb_synch_frame (ts_usb_open* open)
{
    int errcode;
    __u16 frame;
    ts_usb_request req;
    
    gdlf_fntrace ("usb_synch_frame (0x%p)", open);

    gdlf_apassert (open != NULL,
                   { 
                       usb_errno = -USBE_OPEN_NULL; 
                       return -EINVAL; 
                   },
                   "Cannot set synch frame on a NULL "
                   "ts_usb_open!\n");

    errcode = usb_dst_check (open);
    if (errcode)
        return errcode;

    if (open->endp->type != e_usb_tr_type_isoc)
    {
        gdlf_err ("Only isochronous endpoints support synch frames\n");
        usb_errno = -USBE_ENDP_NO_SYNCH;
        return -EINVAL;
    }
    
    req.bmreqtype = 0x82;
    req.request = e_usb_req_synch_frame;
    req.value = 0;
    req.index = open->endp->addr;
    req.length = sizeof (frame);
    
    read_lock (&open->dev->msnode.lock);
    frame = 0;
    errcode = usb_request_data (open->dev, &req, &frame,
                                "get synch frame");
    if (errcode < 0)
        gdlf_err ("Cannot get sync frame; request failed: "
                  "errno/usb %d/%d\n", errcode, usb_errno);
    read_unlock (&open->dev->msnode.lock);
    return errcode < 0? errcode : le16_to_cpu (frame);
}


/* Log:
 *
 * $Log: usbd-endp.c,v $
 * Revision 1.14  1998/11/14 17:15:12  inaky
 * Moved the include files to uusbd/include/uusbd, to ease integration
 * into the Linux kernel.
 *
 * Revision 1.13  1998/09/14 19:44:26  inaky
 * Removed function traces.
 *
 * Revision 1.12  1998/08/24 02:25:33  inaky
 * Changed name of all enumerations to be prefixed with the enumeration
 * type name.
 *
 * Revision 1.11  1998/07/28 23:20:40  inaky
 * Collected all transaction error states into a single error state.
 *
 * Revision 1.10  1998/07/23 14:30:07  inaky
 * Added the length parameter to the request structure.
 *
 * Revision 1.9  1998/07/16 14:10:43  inaky
 * Moved from the old read/write interface to the new, request-based
 * interface, fixed locks.
 *
 * Revision 1.8  1998/04/26 03:36:42  inaky
 * Fixed to make usb_endp_status_get() actualize endpoint's status field
 * on stall.
 *
 * Revision 1.7  1998/04/13 22:11:58  inaky
 * Fixed warn methods in trbuffers.
 *
 * Fixed some docs, made *_status_get() function return status values
 * instead of returning them via pointers, and also made them actualize
 * the status fields.
 *
 * Revision 1.6  1998/03/28 19:34:15  inaky
 * Removed inclusion of usb-methods.h and fixed to work with callbacks.
 *
 * Revision 1.5  1998/02/12 16:35:10  inaky
 * Finished the device ops, and moved synch_frame getting to usbd-endp,
 * where it belongs. Fixed some stuff here and there, and starting to get
 * worried on locks ... I'm gonna have to debug *lots* of deadlocks.
 *
 * Revision 1.4  1998/01/11 15:44:18  inaky
 * Finished functions from the descriptor, inteface and endpoint modules,
 * fixed docs, pretty printed and removed craps.
 *
 * Revision 1.3  1998/01/08 12:19:12  inaky
 * Added copyright strings to all important files
 *
 * Revision 1.2  1997/12/29 00:25:39  inaky
 * Fixed unimplemented functions to return -ENOSYS and moving the khubdd
 * to use signals as commands.
 * Also added function traces on most files that had them disabled (donkey!)
 *
 * Revision 1.1  1997/12/27 19:54:21  inaky
 * Added all the prototypes and skeleton definitions (incomplete) of all
 * of the USBD functions to each source file.
 *
 */
