
/* Universal USB Driver Stack v$Revision: 1.43 $ 
 *
 * (C) 1997 by Iaky Prez Gonzlez <inaky@peloncho.fis.ucm.es>
 *
 * Distributed under the terms of the GNU Public License Version 2
 * See files COPYRIGHT and COPYING at the root distribution directory
 *
 * $Id: usbd-host.c,v 1.43 1998/11/14 17:15:13 inaky Exp $
 *
 * Changelog at the bottom 
 *
 *
 *  USB Driver - Hosts section
 *  --------------------------
 *
 * This is the part of the UUSBD who takes care of host registration
 * and managing.
 *
 *
 *  FIXME: {read,write}_lock -> {read,write}_lock_irq?
 */

#include <uusbd/usb-hub.h>
#include <uusbd/usbd.h>

#define GDLF_IDENT          "USBD"
#define GDLF_LOCAL          /* Define if want local settings */
#define GDLF_DEF_LOGLEVEL   9
#define GDLF_ASSERTIONS     /* want assertions (#undef if not) */
#undef GDLF_FNTRACE        /* want function traces */
#define GDLF_LOCATIONS      /* want locations in log messages */
#include <uusbd/gdlf.h>


ts_slist hosts;             /* list of hosts */


    /* Reset a host controller.
     *
     * Parameters:  
     *
     *  ts_usb_host*    USB host structure (aka: handle)
     *
     * Returns:
     *
     *  int     0 if ok, -1 on error. In case of error, the
     *          reset may have taken place incompletely, so a second
     *          reset could be neccesary.
     *
     * Global:
     *
     *  int usb_errno   Error code if != 0.
     */

int usb_host_reset (ts_usb_host *host)
{   
    gdlf_fntrace ("usb_host_reset (0x%p)", host);
    
        /* 1. Check all the data being OK if requested
         * 
         * 2. Reset each device plugged to this host, aborting all the
         *    pending transactions
         *
         * 3. Reset the device via an HCI method
         */

        /* Check the descriptors */
    
    gdlf_apassert (host != NULL,
                   { 
                       usb_errno = -USBE_HOST_NULL; 
                       return -EINVAL; 
                   },
                   "Tried to reset a NULL USB host controller structure\n");    
    
    gdlf_apassert (host->reset != NULL,
                   {    
                       usb_errno = -USBE_BAD_HOST_STRUCT; 
                       return -EBADF; 
                   },
                   "Host controller doesn't have a HCI reset method\n");

        /* 
         * We are going to reset the host, thus all the devices
         * pending have to be detached; any device which is
         * resetted goes into the default state (see USB Spec 1.0
         * Figure 9-1), and when resetting a host, the root hub is
         * resetted and set to the default state. Thus, previous to
         * resetting, we detach everyone; later we reset all the root
         * hub ports. 
         *
         * FIXME: I am worried, because a large USB tree could cause
         *        an stack overflow. The maximum hub deep is five hubs
         *        linked one after the other, so we can count on a
         *        safe limit ... I hope 
         */

    usb_hub_detach_devices (&host->root, USB_CFG_HOST_RESET);


        /* Reset the host. There's no need to do any locking here, as
         * it is done at UHCI level.
         *
         * Remember that the hci-reset method MUST actualize the sof
         * at ts_usb_host after the reset (and any other field which
         * needs actualization). UHCI already does, I leave the FIXME
         * below for when I implement OHCI.
         *
         * FIXME: does hci-reset actualize the sof at ts_usb_host?
         */

    return host->reset (host);      /* reset the host */
}


    /* Search an unused address number for a host (not a device)
     *
     *  Based on the way the addresses are assigned, this routine
     * searches for an unused address number and returns it. The idea
     * is that we register hosts in the host list by inserting them
     * into the top of the list, and we assign it the address of the
     * last host registered plus one (thus, the address numbers
     * decrease from the head of the list to the tail).
     *
     *  However, given the case of a system with 345 hosts being
     * registered and unregistered constantly [I know I'm pedantic
     * when thinking about almost impossible scenarios :)], we should
     * try to reuse freed addreses. Thus, to comply with this, we'll
     * explore begining on the head of the list and taking the initial
     * address as the one of the host on the top. We'll go down the
     * list, and on each new list link we follow, we'll decrement the
     * address. On each list node, we check if the address number
     * we've got and the address number of the host in that node are
     * equal. If they're, it's used (really? :), if not, it's free and
     * we can use it. It is guaranteed that our address number will be
     * greater or equal than the address used by the host at that node
     * because of the order of registration we are using.
     *
     *  In the case of being all the numbers used (we reach the last
     * node), then we'll use the address of the host at the head of
     * the list plus one.
     *
     * Returns:
     *
     *  t_usb_addr      Address to use, 0 on error.
     *
     * Global:
     *
     *  int usb_errno   On error, error code.  
     */

t_usb_host usb_host_addr_assign (void)
{
    ts_usb_host *host;
    t_usb_host  addr = 0;
    
    gdlf_fntrace ("usb_host_addr_assign()");

    read_lock (ts_slist_lock (&hosts));
    host = (ts_usb_host*)ts_slist_head (&hosts);
    if (host)                       /* if there are hosts registered... */
    {
        addr = host->addr;              /* fetch head's address */
        do                              /* while we have hosts */
        {
            if (host->addr != addr--)       /* this one unused? ... */
                break;                      /* .. we got it */
            host = (ts_usb_host*)ts_slist_next (&host->snode);
        } while (host != NULL && addr >= 0);

        if (host == NULL)               /* if none is free ... */
            addr = ((ts_usb_host*)      /* ... get last one */
                ts_slist_head (&hosts))->addr;

        addr += 1;                      /* address is the one we got + 1 */
    }
    else
        addr = 0;                       /* none registered, use 0 */

    read_unlock (ts_slist_lock (&hosts));
    return addr;                        
}


    /* Registers a filled in host structure into the global array of
     * working usb hosts; after that, starts to enumerate the bus by
     * telling the hub driver to use the root hub at the given
     * host. This will make the hub driver enumerate the whole bus.
     *
     *  Note you must initialize _all_ fields, and the root hub 
     * must be ready to work. Note a failing hub root will invalidate
     * the host registration process.
     *
     * Parameters:  
     *
     *  ts_usb_host*    Host structure to register
     *
     * Returns:
     *
     *  int             0 if ok, errno code < 0 on error.
     *
     * Global:
     *
     *  int usb_errno   On error, error code.  
     */

int usb_host_register (ts_usb_host *host)
{   
    int errcode;
    
    gdlf_fntrace ("usb_host_register (0x%p)", host);
    
    gdlf_aassert (host != NULL, return -EINVAL);

        /*
         *  HCI modules should be careful enough to have everything
         * defined ... so we'll assume they're good guys and the host
         * structure is good enough not to check it.
         *
         *  Take into account the zillion checks made on all the
         * routines which use the host pointers, this is the reason
         * not to make them here.
         */
    
    ts_dtor_fill (&host->snode.dtor, (tf_dtor)usb_host_destroy);

    host->addr = usb_host_addr_assign();    /* get a free address */
    errcode = ts_slist_prep (&hosts, &host->snode);
    if (errcode)
    {
        gdlf_err ("Couldn't add host 0x%p to the "
                  "list of hosts: errno/usb %d/%d\n",
                  host, errcode, usb_errno);
        return errcode;
    }

        /* 
         *  Now we enumerate the bus by starting the hub root and
         * letting it enumerate the bus behind it. The only thing we
         * gotta do is registering it as a hub; the khubdd will take
         * care of the rest.
         */

    errcode = usb_hub_register (&host->root);
    if (errcode < 0)
    {
        gdlf_err ("Error registering root hub "
                  "for the host controller 0x%p (%d): "
                  "errno/usb %d/%d. "
                  "Registration failed.\n", 
                  host, host->addr, errcode, usb_errno);
        ts_slist_del (&hosts, &host->snode);
        return errcode;
    }
    return 0;
}


    /* Unregister a host structure and all the devices pending from
     * it. 
     *
     *  Note this function doesn't free the host data
     * structures. It'll just free everything pending from it and then
     * let the caller do the final clean on the host. Note that the
     * HCI part is already cleaned.
     *
     *  It will also stop the scheduling on the root hub by
     * unregistering it.
     *
     * Parameters:  
     *
     *  ts_usb_host*    Pointer to the host structure.
     *
     * Returns:
     *
     *  int             0 if ok, <0 on error.
     *
     * Global:
     *
     *  int usb_errno Error code if != 0.  */

int usb_host_unregister (ts_usb_host *host)
{
    int errcode;
    
    gdlf_fntrace ("usb_host_unregister (0x%p)", host);
    gdlf_aassert (host != NULL, return -EINVAL);    
    
        /* 
         * 1. Detach all the devices in the host. However, it'll
         *    easier to just unregister the root hub, as it will
         *    detach each device pending from itself, and each hub
         *    will do the same with all the devices plugged to it.
         */

    errcode = usb_hub_destroy (&host->root);
    if (errcode < 0)
        gdlf_err ("Error destroying root hub 0x%p (%d %d) "
                  "for host 0x%p (%d) unregistration: "
                  "errno/usb %d/%d; "
                  "expect inconsistencies\n",
                  &host->root,
                  host->addr, 
                  host->root.device? host->root.device->addr : -1,
                  host, host->addr,
                  errcode, usb_errno);

        /*
         * 2. Remove the host structure from the host list.
         *    [Note we don't care if there's an error -- should only
         *     happen if the host is not on the list --, as would be
         *     common on some error handling situations not to have
         *     the host on the list].
         *
         * NOTE: What may happen if here somebody registers a new root
         *       hub? Should not happen, as the hubd driver won't do
         *       it auto-magically, but is interesting ...
         */

    ts_slist_del (&hosts, &host->snode);
    return 0;
}


    /* Destructor: Unregister and free a ts_usb_host 
     *
     * Parameters:  
     *
     *  ts_usb_host*    Pointer to the host structure.
     *
     * Returns:
     *
     *  int             0 if ok, <0 on error.
     *
     * Global:
     *
     *  int usb_errno   Error code if != 0.  
     */

int usb_host_destroy (ts_usb_host *host)
{
    int errcode;
    gdlf_fntrace ("usb_host_destroy (0x%p)", host);

    usb_host_ref (host);
    
        /* Unregister */

    errcode = usb_host_unregister (host);
    gdlf_passert (errcode == 0,
                  "Couldn't unregister host 0x%p: "
                  "errno/usb %d/%d. Destroying anyway\n",
                  host, errcode, usb_errno);

    if (host->hci)
    {
        ts_snode_unref_nd (host->hci);
        ts_dtor_exec (&host->hci->dtor);
        usb_kfree (host->hci);
    }

    usb_host_unref_nd (host);
    ts_usb_host_release (host);
    return errcode;
}


    /*  Register a device into a host.
     *
     *  Registers a device within a host, inserting it into the top of
     * the list of devices related to that host.
     *
     * Parameters:  
     *
     *  ts_usb_device*  Pointer to the device structure to register
     *                  into the host.
     *
     * Returns:
     *
     *  int             0 if ok, errno code < 0 on error.
     *
     * Global:
     *
     *  int usb_errno   On error, error code.  
     */

int usb_host_register_device (ts_usb_device* device)
{
    int errcode;
    ts_usb_host *host;

    gdlf_fntrace ("usb_host_register_device (0x%p)", device);

    gdlf_apassert (device != NULL,
                   { 
                       usb_errno = -USBE_DEVICE_NULL; 
                       return -EINVAL; 
                   },
                   "Tried to register a NULL device into host\n");

    host = device->hub->host;
    write_lock (&device->lock);
    errcode = ts_3slist_prep (&host->devices, &device->msnode);
    write_unlock (&device->lock);
    return errcode;
}


    /*  Unregister a device from a host.
     *
     *  Unregisters a device from a host, removing it from the list of
     * devices related to that host.
     *
     * Parameters:  
     *
     *  ts_usb_device*  Pointer to the device structure to register
     *                  into the host.
     *
     * Returns:
     *
     *  int             0 if ok, errno code < 0 on error.
     *
     * Global:
     *
     *  int usb_errno   On error, error code.  
     */

int usb_host_unregister_device (ts_usb_device* device)
{
    ts_usb_host *host;

    gdlf_fntrace ("usb_host_unregister_device (0x%p)", device);

    gdlf_apassert (device != NULL,
                   { 
                       usb_errno = -USBE_DEVICE_NULL; 
                       return -EINVAL; 
                   },
                   "Tried to unregister a NULL device from host\n");
        
        /* 
         * Ignore errors on master drop, as they'll mean we weren't
         * master 
         */ 

    usb_master_drop (device);

    host = device->hub->host;
    write_lock (&device->lock);
    ts_3slist_del (&host->devices, &device->msnode);
    write_unlock (&device->lock);
    return 0;
}


    /* Get a pointer to a host controller structure known it's
     * address.  
     *
     * Parameters:  
     *
     *  t_usb_addr      Address number of the host controller.
     *
     * Returns:
     *
     *  ts_usb_host*    Pointer to the host controller structure, NULL
     *                  on error.
     *
     * Global:
     *
     *  int usb_errno   On error, error code.  
     */

ts_usb_host* usb_host_ptr (t_usb_host addr)
{
    ts_usb_host *host;

    gdlf_fntrace ("usb_host_ptr (%d)", addr);

    gdlf_aassert (addr >= 0, return NULL);
    
    read_lock (&hosts->lock);
    host = (ts_usb_host*)ts_slist_head (&hosts);
    while (host != NULL)
    {
        if (host->addr == addr)
        {
            read_unlock (&hosts->lock);
            return host;
        }
        host = (ts_usb_host*)ts_slist_next (&host->snode);
    }
    read_unlock (&hosts->lock);

    gdlf_warning ("Unexistant host controller with address %d\n", 
                  addr);
    usb_errno = -USBE_HOST_NOT_FOUND;
    return NULL;
}


    /* Get a pointer to a host controller structure known it's
     * HCI host address.  
     *
     * Parameters:  
     *
     *  ts_snode*       Address of the HCI controller.
     *
     * Returns:
     *
     *  ts_usb_host*    Pointer to the host controller structure, NULL
     *                  on error.
     *
     * Global:
     *
     *  int usb_errno   On error, error code.  
     */

ts_usb_host* usb_host_hci (ts_snode* hci)
{
    ts_usb_host *host;

    gdlf_fntrace ("usb_host_hci (0x%p)", hci);

    read_lock (&hosts->lock);
    host = (ts_usb_host*)ts_slist_head (&hosts);
    while (host != NULL)
    {
        if (host->hci == hci)
        {
            read_unlock (&hosts->lock);
            return host;
        }
        host = (ts_usb_host*)ts_slist_next (&host->snode);
    }
    read_unlock (&hosts->lock);

    gdlf_warning ("Unexistant host with HCI address 0x%p\n", 
                  hci);
    usb_errno = -USBE_HOST_NOT_FOUND;
    return NULL;
}


    /* Seek a free device address and return it.
     *
     * NOTE: The method for doing this search could be enhanced, as it
     *       is no secret I haven't squeezed my brains finding one :)
     *
     *  This is simple. It starts on address 1 [first allowed USB
     * address] and walks the list. If it finds any device using the
     * address, then it increments and tries again from the head of
     * the list. If it finishes the list without having found any
     * device using the address, then that address is given.
     *
     * WARNING! The host must be already locked, this function won't
     *          do nor acknowledge any locking.
     *
     * Returns:
     *
     *  int             Free address if ok, errno code < 0 on
     *                  error. Address is guaranteed to be between 1
     *                  and 127 (the allowed addresses for USB
     *                  devices). If no address is available, -ENOSPC
     *                  (no space left on device) will be returned, as
     *                  there's no more space in the bus for another
     *                  device (according to the USB spec, not to the
     *                  bus topology).
     *
     * Global:
     *
     *  int usb_errno   On error, error code.  
     */

int usb_host_free_address (ts_usb_host* host)
{
    ts_usb_device* itr;
    int addr = 1;       /* default initial address */
    
    gdlf_fntrace ("usb_host_free_address (0x%p)", host);
    gdlf_aassert (host != NULL, return -EINVAL); 

    if (host->devices.elements == USB_DEVICES_MAX)
        return -ENOSPC;

    itr = (ts_usb_device*)              /* get last registered device */
        ts_3slist_head (&host->devices);
    while (itr)                         /* if there're devices ... */
        if (addr == itr->addr)          /* used? */
        {
            addr++;                         /* used */
            if (addr == USB_DEVICES_MAX)    /* should not happen */
                return -ENOSPC;
            itr = (ts_usb_device*) ts_3slist_head (&host->devices);
            continue;
        }
        else
            itr = (ts_usb_device*)          /* get next dev */
                ts_3slist_next (&host->devices, &itr->msnode);
    return addr; 
}


/* Log:
 *
 * $Log: usbd-host.c,v $
 * Revision 1.43  1998/11/14 17:15:13  inaky
 * Moved the include files to uusbd/include/uusbd, to ease integration
 * into the Linux kernel.
 *
 * Revision 1.42  1998/11/09 21:30:34  inaky
 * Moved all host ref/unref to usb_host_[un]ref().
 *
 * Revision 1.41  1998/09/24 15:57:16  inaky
 * Fixed the address assigning function, it was broken.
 *
 * Revision 1.40  1998/09/14 19:45:16  inaky
 * Removed usb_hub_errno; moved the HCI desctruction to be _after_ the
 * host unregistration.
 *
 * Revision 1.39  1998/08/30 13:50:51  inaky
 * Severed assertions.
 *
 * Revision 1.38  1998/08/27 22:15:47  inaky
 * Removed traces.
 *
 * Revision 1.37  1998/08/27 11:12:24  inaky
 * Moved ts_*_free() to ts_*_release(), made them return void and severed
 * assertions.
 *
 * Revision 1.36  1998/08/24 02:25:48  inaky
 * Removed spurious debug messages.
 *
 * Revision 1.35  1998/08/19 23:52:00  inaky
 * Added some debug code for memory corruption problems. Fixed
 * usb_host_hci()'s fntrace to have the code for a pointer, not for a
 * decimal.
 *
 * Revision 1.34  1998/08/03 13:34:03  inaky
 * Severed assertions to make the code clearer. Fixed the host
 * destructor to destroy the HCI prior to unregister the USB host, so the
 * HCI destructor can use the list of registered hosts to determine which
 * is it's ts_usb_host from it's HCI pointer. Added the usb_host_hci()
 * function for this.
 *
 * Revision 1.33  1998/07/30 17:53:48  inaky
 * Fixed usb_host_ptr() to allow hosts with address 0 and tweaked a
 * format specifier in warning argument.
 *
 * Revision 1.32  1998/07/16 14:11:51  inaky
 * Fixed lockings (removed some, as when adding/removing from lists, the
 * list code takes care of locking).
 *
 * Fixed the host destructor to _destroy_ the HCI and then free it,
 * instead of destroy it and then unreference it.
 *
 * Revision 1.31  1998/07/14 14:54:16  inaky
 * Fixed the address assign function, as it was assigning 1 as the first
 * address 1 instead or 0.
 *
 * Revision 1.30  1998/07/12 22:54:13  inaky
 * Fixed the host reset function to use the new function the hub manager
 * provides to detach all devices from a hub. Moved from unregistering
 * (_bad_) to destroying, and fixed the destructor to keep the object
 * alive while destroying.
 *
 * Revision 1.29  1998/07/10 12:06:04  inaky
 * Finished the root hub registration and start stuff.
 *
 * Revision 1.28  1998/06/21 20:15:28  inaky
 * Fix docs and add a _clean() call to usb host, just in case.
 *
 * Revision 1.27  1998/06/20 21:54:53  inaky
 * Massive changes to split the old monolithic hubd driver into a hub
 * manager (hub wrappers and daemon) and a dynamic driver, which can be
 * replaced and its different parts reused.
 *
 * Revision 1.26  1998/05/24 12:04:43  inaky
 * Fixed error messages not to mess with hub driver's error code.
 *
 * Revision 1.25  1998/04/13 22:13:23  inaky
 * Polished the detachment of the root usb hub and pending devices when
 * resetting or unregistering a host
 *
 * Revision 1.24  1998/03/28 19:34:53  inaky
 * Removed inclusion of usb-methods.h and some linux specific headers
 * which weren't needed.
 *
 * Revision 1.23  1998/02/12 14:37:10  inaky
 * Removed the inclusion of the atomic declarations, as now it's all
 * handled via ts_[N]snode_[un]ref(). Added some comments, fixed to
 * ignore errors when unregistering if host is not on the list (ditto
 * when unregistering devices within the host). Make usb_host_destroy()
 * clean the hci field after unreferencing it, so ts_usb_host_free()
 * doesn't get confused about that.
 *
 * Revision 1.22  1998/02/09 13:30:36  inaky
 * Moved all use counts to atomic
 *
 * Revision 1.21  1998/01/11 15:44:19  inaky
 * Finished functions from the descriptor, inteface and endpoint modules,
 * fixed docs, pretty printed and removed craps.
 *
 * Revision 1.20  1998/01/08 12:19:12  inaky
 * Added copyright strings to all important files
 *
 * Revision 1.19  1998/01/02 00:12:17  inaky
 * Finished the new driver scheme. Documented at usbd-drv.c
 *
 * Revision 1.18  1998/01/01 19:14:49  inaky
 * Added the usb_host_free_address(), to search for free device address
 * on a host device list.
 *
 * Revision 1.17  1997/12/31 16:37:12  inaky
 * Finished the HCI interface
 *
 * Revision 1.16  1997/12/31 05:37:01  inaky
 * Almost finished all the lists and life cycle for hosts, drivers and
 * devices. Some craps to be fixed yet.
 *
 * Revision 1.15  1997/12/31 05:17:51  inaky
 * Added the destructors for hosts
 *
 * Revision 1.14  1997/12/30 20:15:45  inaky
 * Moved everything to the new lists scheme. Untested yet.
 *
 * Revision 1.13  1997/12/30 04:14:38  inaky
 * Moved to the new lists code. Unchecked.
 *
 * Revision 1.12  1997/12/29 00:25:40  inaky
 * Fixed unimplemented functions to return -ENOSYS and moving the khubdd
 * to use signals as commands.
 * Also added function traces on most files that had them disabled (donkey!)
 *
 * Revision 1.11  1997/12/27 20:06:28  inaky
 * Finished fixinf the __KERNEL__ #ifdef wrapping and moved the ts_dtor.h
 * to usb-dtor.h
 *
 * Revision 1.10  1997/12/27 02:07:04  inaky
 * Moving everything to use the GDLF for logging and debug checkings.
 * Removed the global array of USB hosts and moved it to a linked
 * list. Added an address field to each USB host controller.
 * Removed lots of cruft that were accumulated and useless. Still some
 * missing, but may be useful someday.
 *
 * Revision 1.9  1997/12/27 02:06:01  inaky
 * Moving everything to use the GDLF for logging and debug checkings.
 * Removed the global array of USB hosts and moved it to a linked
 * list. Added an address field to each USB host controller.
 * Removed lots of cruft that were accumulated and useless. Still some
 * missing, but may be useful someday.
 *
 * Revision 1.8  1997/12/08 02:41:02  inaky
 * Fixed some stuff of the #includes, modified the makefile targets to be
 * more flexible, tried to finish the device configuration section.
 *
 * Revision 1.7  1997/12/06 00:19:41  inaky
 * Almost finished the usbd/host, usbd/cfg and a lot of meat from
 * usbd/device. Also added the driver definition. Got to polish it yet
 * ...
 *
 * Revision 1.6  1997/11/29 20:21:48  inaky
 * Documented and almost fixes the read/write/abort API, documented a
 * damn lot and implemented some stuff
 *
 * Revision 1.5  1997/11/29 00:46:31  inaky
 * Documented almost all the hub basic API, added spin locks to
 * usb_host_register() and some more stuff. Going for more.
 *
 * Revision 1.4  1997/10/24 04:20:20  inaky
 * Security backup
 *
 * Revision 1.3  1997/10/02 22:29:37  inaky
 * Finally managed to understand the workings of the MODVERSIONS (ok, not
 * exactly the working, but how to make it work).
 *
 * Now I can go to sleep happy :) -- I know, I'm really sick --
 *
 * Revision 1.2  1997/10/02 21:29:03  inaky
 * Remade all the directory structure to make it cleaner when the USB
 * class drivers come.
 *
 * Revision 1.1  1997/10/01 21:28:12  inaky
 * Re-making a lot of stuff in the hadnling of host, hub and device
 * initialization and handling, following a more correct approach of
 * registering/unregistering. Also moved the usb_host array to be an
 * array of pointers, so dynamic reallocation of it is easier, and the
 * host structure needs not to be moved in memory. The same applies to
 * the device arrays for each host.
 *
 */

