
/* Universal USB Driver Stack v$Revision: 1.16 $ 
 *
 * (C) 1997 by Iaky Prez Gonzlez <inaky@peloncho.fis.ucm.es>
 *
 * Distributed under the terms of the GNU Public License Version 2
 * See files COPYRIGHT and COPYING at the root distribution directory
 *
 * $Id: usbd-iface.c,v 1.16 1998/11/14 17:15:14 inaky Exp $
 *
 * Changelog at the bottom 
 *
 *
 *  USB Driver - Devices section - USB Interface operations
 *  -------------------------------------------------------
 *
 * Standard and common operations on USB interfaces.
 */

#include <uusbd/usbd.h>

#define GDLF_IDENT          "USBD"
#define GDLF_LOCAL          /* Define if want local settings */
#define GDLF_DEF_LOGLEVEL   9
#define GDLF_ASSERTIONS     /* want assertions (#undef if not) */
#undef GDLF_FNTRACE        /* want function traces */
#define GDLF_LOCATIONS      /* want locations in log messages */
#include <uusbd/gdlf.h>


    /* Dirty-work function for usb_iface_feature_*()
     *
     *  THIS FUNCTION IS FOR INTERNAL USE, DO NOT CALL IT DIRECTLY
     *
     * Parameters:  
     *
     *  ts_usb_device*  Pointer to the device structure where the
     *                  interface is located.
     *
     *  t_usb_iface     Interface number.
     *
     *  e_usb_featsel   Feature
     *
     *  e_usb_req       USB request code
     *
     * Returns:
     *
     *  int             0 if ok, errno code < 0 on error.
     *
     * Global:
     *
     *  int usb_errno   On error, error code.  
     */

int usb_iface_feature (ts_usb_device* device, t_usb_iface ifacen, 
                       e_usb_featsel feature, e_usb_req request)
{
    ssize_t errcode;
    ts_usb_request req;
    
    gdlf_fntrace ("usb_iface_feature (0x%p, %u, %d, %u)",
                  device, ifacen, feature, request);

    gdlf_apassert (device != NULL,
                   { 
                       usb_errno = -USBE_DEVICE_NULL; 
                       return -EINVAL; 
                   },
                   "Cannot set/clear features on NULL ts_usb_device!\n");

    write_lock (&device->msnode.lock);
    if (device->active == NULL)     /* unconfigured? */
    {
        write_unlock (&device->msnode.lock);
        gdlf_err ("Device 0x%p isn't configured yet.\n",
                  device);
        usb_errno = -USBE_CFG_NULL;
        return -EPERM;
    }

    if (ifacen >= device->active->ifaces)
    {
        gdlf_err ("Interface index %d out of range "
                  "for USB configuration %d "
                  "of device (%d %d)\n",
                  ifacen, device->active->addr, 
                  device->hub->host->addr, 
                  device->addr);
        write_unlock (&device->msnode.lock);
        usb_errno = -USBE_IFACE_BAD_INDEX;
        return -EINVAL; 
    }
    
    req.bmreqtype = 0x1;
    req.request = request;
    req.value = feature;
    req.index = ifacen;

    errcode = usb_request_dataless (device, &req, 
                                    "inteface feature change");
    if (errcode < 0)
        gdlf_err ("Cannot change interface feature; request failed: "
                  "errno/usb %d/%d\n", errcode, usb_errno);
    write_unlock (&device->msnode.lock);
    return errcode;
}

#if 0 

    /* 
     * Currently there's no iface status ... so 
     *
     * FIXME: make it behave as usb_(device|endp)_status_get()
     */

    /* Returns an interface's status
     *
     * NOTE: As of USB 1.0, this call will return all zeroes, as the
     *       interface status is not used.
     *
     * NOTE: the fields in the status structure (ok, none - jic) are
     *       given in host order.
     *
     * Parameters:  
     *
     *  ts_usb_device*  Pointer to the device structure where the
     *                  interface is located.
     *
     *  ts_usb_iface_stat*
     *
     *                  Pointer to where to store the interface
     *                  status.
     *
     * Returns:
     *
     *  int             0 if ok, errno code < 0 on error.
     *
     * Global:
     *
     *  int usb_errno   On error, error code.  
     */

int usb_iface_status_get (ts_usb_device* device, t_usb_iface ifacen,
                          ts_usb_iface_stat* pstat)
{
    ssize_t errcode;
    ts_usb_request req;
    
    gdlf_fntrace ("usb_iface_status_get (0x%p, %u, 0x%p)",
                  device, ifacen, pstat);

    gdlf_apassert (device != NULL,
                   { 
                       usb_errno = -USBE_DEVICE_NULL; 
                       return -EINVAL; 
                   },
                   "Cannot get interface status on NULL "
                   "ts_usb_device!\n");

    if (device->active == NULL)     /* unconfigured? */
    {
        gdlf_err ("Device 0x%p isn't configured yet.\n",
                  device);
        usb_errno = -USBE_CFG_NULL;
        return -EPERM;
    }

    if (ifacen >= device->active->ifaces)
    {
        gdlf_err ("Interface index %d out of range "
                  "for USB configuration %d "
                  "of device (%d %d)\n",
                  ifacen, device->active->addr, 
                  device->hub->host->addr, 
                  device->addr);
        usb_errno = -USBE_IFACE_BAD_INDEX;
        return -EINVAL; 
    }
    
    req.bmreqtype = 0x81;
    req.request = USB_REQ_GET_STATUS;
    req.value = 0;
    req.index = ifacen;
    req.length = sizeof (ts_usb_iface_stat);

    errcode = usb_request_data (device, &req, pstat, 
                                "get interface status");
    if (errcode < 0)
    {
        gdlf_err ("Cannot get interface status; request failed: "
                  "errno/usb %d/%d\n", errcode, usb_errno);
        write_unlock (&device->msnode.lock);
        return errcode;
    }
    ts_usb_iface_stat_le2h (&status);
    return 0;
}
#endif /* #if 0 */


    /* FIXME: Detailed desc
     *
     *  Call with locks held!
     *
     *  Beware endpoint addresses have direction in bit 0, and the
     * pendps array is for all endpoints but 0. Thus, we have to cut
     * away bit 0 (>>1) and substract one to get the correct pointer.
     *
     * Parameters:  
     *
     *  type            descr
     */

void usb_iface_pendp_unset (ts_usb_cfg* cfg, t_usb_iface ifacen)
{
    ts_usb_endp *itr, *top;
    
    gdlf_fntrace ("usb_iface_pendp_unset (0x%p, %u)", cfg, ifacen);
    gdlf_aassert (cfg != NULL, return);
    gdlf_aassert (cfg->iface != NULL, return);
    gdlf_aassert (ifacen < cfg->ifaces, return);
    gdlf_aassert (cfg->iface[ifacen].active != NULL, return);

    for (itr = cfg->iface[ifacen].active->endp,
             top = itr + cfg->iface[ifacen].active->endps;
         itr < top; itr++)
    {
        gdlf_assert (itr->desc == NULL);
        cfg->pendp[(itr->addr >> 1) - 1] = NULL;
    }    
    return;
}


    /* FIXME: Detailed desc
     *
     *  Call with locks held!
     *
     * 
     *
     * Parameters:  
     *
     *  type            descr
     *
     * Returns:
     *
     *  int             Active alternate setting number if ok, errno
     *                  code < 0 on error. 
     */

int usb_iface_pendp_set (ts_usb_cfg* cfg, t_usb_iface ifacen)
{
    int errcode = 0;
    t_usb_endp addr;
    ts_usb_endp *itr, *top;
    
    gdlf_fntrace ("usb_iface_pendp_set (0x%p, %u)", cfg, ifacen);
    gdlf_aassert (cfg != NULL, return -EINVAL);
    gdlf_aassert (cfg->iface != NULL, return -EINVAL);
    gdlf_aassert (ifacen < cfg->ifaces, return -EINVAL);
    gdlf_aassert (cfg->iface[ifacen].active != NULL, return -EINVAL);
    
    for (itr = cfg->iface[ifacen].active->endp,
             top = itr + cfg->iface[ifacen].active->endps;
         itr < top; itr++)      /* set in the cfg's endps array */
    {
        addr = (itr->addr >> 1) - 1;

        if (cfg->pendps <= addr)
        {
            errcode = usb_cfg_pendps_resize (cfg, addr);
            if (errcode)
                break;
        }
        cfg->pendp[addr] = itr;
    }
    return errcode;
}


    /* Set an alternate setting of an interface
     *
     * Check nobody uses the interface and then set it. Actualize
     * ts_usb_iface->active  
     *
     * Parameters:  
     *
     *  type            descr
     *
     * Returns:
     *
     *  int             Active alternate setting number if ok, errno
     *                  code < 0 on error. 
     *
     * Global:
     *
     *  int usb_errno   On error, error code.  
     */

int usb_iface_set (ts_usb_device* device, t_usb_iface ifacen, 
                   t_usb_aset asetn)
{
    int errcode;
    ts_usb_request req;        
    ts_usb_cfg *cfg;
    ts_usb_endp *itr, *top;     /* endpoint iterator, top */

    gdlf_fntrace ("usb_iface_set (0x%p, %u, %u)",
                  device, ifacen, asetn);
    gdlf_aassert (device != NULL, return -EINVAL);

    write_lock (&device->msnode.lock); /* FIXME: read upgradeable */

        /* Check USB configuration and interface/aset numbers */
    
    cfg = device->active;
    if (cfg == NULL)     /* unconfigured? */
    {
        write_unlock (&device->msnode.lock);
        gdlf_err ("Device 0x%p isn't configured yet.\n",
                  device);
        usb_errno = -USBE_CFG_NULL;
        return -EPERM;
    }

    if (ifacen >= cfg->ifaces)
    {
        gdlf_err ("Interface number %u out of "
                  "range for USB configuration %d "
                  "on device (%d %d)\n", 
                  ifacen, cfg->addr,
                  device->hub->host->addr, device->addr);
        write_unlock (&device->msnode.lock);
        usb_errno = -USBE_IFACE_BAD_INDEX;
        return -EINVAL;
    }
    
    if (asetn >= cfg->iface[ifacen].asets) 
    {
        gdlf_err ("Alternate setting number %u out of "
                  "range for interface %d of "
                  "USB configuration %d "
                  "on device (%d %d)\n", 
                  asetn, ifacen, cfg->addr,
                  device->hub->host->addr, device->addr);
        write_unlock (&device->msnode.lock);
        usb_errno = -USBE_ASET_BAD_INDEX;
        return -EINVAL;
    }
    
#warning FIXME: Check there is enough bandwidth -(ENOSR,EDQUOT)

        /* 
         * Ok, data is correct, now we'll check we can set the
         * alternate setting. This will be only possible if all the
         * endpoints for the interface's actual active setting aren't
         * used (ts_usb_endp.desc == NULL). 
         */
        
    itr = cfg->iface[ifacen].active->endp;
    top = itr + cfg->iface[ifacen].active->endps;
    for (; itr < top; itr++)        /* cycle all aset's endpoints */
        if (itr->desc != NULL)      /* active? */
        {
            write_unlock (&device->msnode.lock);
            return -EBUSY;          /* ... exit, cannot set */
        }

        /* Now we tell the device to set it */

    req.bmreqtype = 0x1;
    req.request = e_usb_req_set_interface;
    req.value = asetn;
    req.index = ifacen;
        
    errcode = usb_request_dataless (device, &req, 
                                    "set alternate setting");
    if (errcode < 0)
    {
        gdlf_err ("Cannot set alternate setting; request failed: "
                  "errno/usb %d/%d\n", errcode, usb_errno);
        write_unlock (&device->msnode.lock);
        return errcode;
    }

        /* Now we actualize the pointers */

    usb_iface_pendp_unset (cfg, ifacen);    /* remove old endp pointers */
    cfg->iface[ifacen].active =             /* set the active pointer */
        &cfg->iface[ifacen].aset[asetn];
    errcode = usb_iface_pendp_set (cfg, ifacen); /* put new endp pointers */
    write_unlock (&device->msnode.lock);
    return errcode;                         /* return ok */
}


    /* Get the active alternate setting on an interface.
     *
     *  This function will read in from the device which is the actual
     * alternate setting on the given interface and then perform a
     * consistency check. If everything is ok, the active alternate
     * setting' number will be returned.
     *
     * Parameters:  
     *
     *  ts_usb_device*  Pointer to the device structure.
     *
     *  t_usb_iface     Interface number for which we want to get the
     *                  alternate setting.
     *
     * Returns:
     *
     *  int             Active alternate setting number if ok, errno
     *                  code < 0 on error. 
     *
     * Global:
     *
     *  int usb_errno   On error, error code.  
     */

int usb_iface_get (ts_usb_device* device, t_usb_iface ifacen)
{
    int errcode;
    ts_usb_request req;
    ts_usb_cfg *cfg;
    ts_usb_endp *itr, *top;
    __u8 asetn;
    
    gdlf_fntrace ("usb_iface_get (0x%p, %u)", device, ifacen);

    gdlf_apassert (device != NULL,
                   { 
                       usb_errno = -USBE_DEVICE_NULL; 
                       return -EINVAL; 
                   },
                       "Cannot get interface on NULL ts_usb_device!\n");

    write_lock (&device->msnode.lock); /* FIXME: use upgradeable read locks */

    cfg = device->active;
    if (cfg == NULL)     /* unconfigured? */
    {
        write_unlock (&device->msnode.lock);
        gdlf_err ("Device 0x%p isn't configured yet.\n",
                  device);
        usb_errno = -USBE_CFG_NULL;
        return -EPERM;
    }

    if (ifacen >= cfg->ifaces)
    {
        write_unlock (&device->msnode.lock);
        gdlf_err ("Interface number %u out of "
                  "range for USB configuration %d "
                  "on device (%d %d)\n", 
                  ifacen, cfg->addr,
                  device->hub->host->addr, device->addr);
        usb_errno = -USBE_IFACE_BAD_INDEX;
        return -EINVAL;
    }

        /* 
         * FIXME: The actual idea on having it on memory was to save
         *        bandwidth. This way we're screwing it down. All
         *        these checks (asking to the device, doing the
         *        consistency check and fixing it) should be
         *        considered if is needed to always make'em or just
         *        leave'em as a debug aid.
         */
    
        /* All checks ok, request actual alternate setting */

        
    req.bmreqtype = 0x81;
    req.request = e_usb_req_get_interface;
    req.value = 0;
    req.index = ifacen;
    req.length = sizeof (asetn);
    
    asetn = 0;
    errcode = usb_request_data (device, &req, &asetn,
                                "get alternate setting");
    if (errcode < 0)
    {
        gdlf_err ("Cannot get alternate setting; request failed: "
                  "errno/usb %d/%d\n", errcode, usb_errno);
        write_unlock (&device->msnode.lock);
        return errcode;
    }

        /* 
         * Now do a consistency check 
         *
         *  If it fails, it's bad. We'll try to get on, by setting the
         * software to reflect hardware's state. We'll close any
         * descriptor on the fake active aset and move everything to
         * the new one.
         */

    if (asetn != cfg->iface[ifacen].active->addr)
    {
        gdlf_err ("Consistency check failure! Device (%d %d) reports "
                  "active alternate setting %d on interface %d of "
                  "USB configuration %d. "
                  "I thought it was %d.\n",
                  device->hub->host->addr, device->addr,
                  asetn, ifacen, cfg->addr,
                  cfg->iface[ifacen].active->addr);
        gdlf_err ("Actualizing to the actual device's alternate setting\n");
        
            /* Close descriptors on fake alternate setting */

        itr = cfg->iface[ifacen].active->endp;
        top = itr + cfg->iface[ifacen].active->endps;
        for (; itr < top; itr++)            /* cycle all old aset's endpoints */
            if (itr->desc != NULL)          /* active? */
                usb_dst_unset (itr->desc);

            /* actualize new alternate setting */

        cfg->iface[ifacen].active =         /* set the active pointer */
            &cfg->iface[ifacen].aset[asetn];

        itr = cfg->iface[ifacen].active->endp;
        top = itr + cfg->iface[ifacen].active->endps;
        for (; itr < top; itr++)            /* cycle all pointers */
            cfg->pendp[itr->addr-2] = itr;  /* set in the cfg's endps array */
    }

    write_unlock (&device->msnode.lock);
    return asetn;
}

/* Log:
 *
 * $Log: usbd-iface.c,v $
 * Revision 1.16  1998/11/14 17:15:14  inaky
 * Moved the include files to uusbd/include/uusbd, to ease integration
 * into the Linux kernel.
 *
 * Revision 1.15  1998/11/09 21:30:48  inaky
 * Removed traces.
 *
 * Revision 1.14  1998/10/31 02:13:26  inaky
 * Misc esthetical fixes and little bug.
 *
 * Revision 1.13  1998/09/14 19:46:16  inaky
 * Moved from the static endp pointers scheme to the dynamic one. Fixed
 * the alternate setting change function to take into account the
 * actualization of endpoint pointers and added a fixme to check for
 * bandwidth.
 *
 * Revision 1.12  1998/08/30 13:51:14  inaky
 * Added methods to actualize endpoint pointers on cfg array when
 * changing the alternate setting.
 *
 * Revision 1.11  1998/08/24 02:25:56  inaky
 * Changed name of all enumerations to be prefixed with the enumeration
 * type name.
 *
 * Revision 1.10  1998/07/23 14:30:18  inaky
 * Added the length parameter to the request structure.
 *
 * Revision 1.9  1998/07/16 14:12:11  inaky
 * Moved from the old read/write interface to the new, request-based
 * interface and fixed locks.
 *
 * Revision 1.8  1998/04/13 22:14:20  inaky
 * Fixed to use the new callbacks scheme on trbuffers, removed [#ifdef'd
 * to zero] the function to get interface status, as it is useless.
 *
 * Revision 1.7  1998/03/28 19:35:25  inaky
 * Removed inclusion of usb-methods.h and fixed to use callbacks.
 *
 * Revision 1.6  1998/02/12 16:35:10  inaky
 * Finished the device ops, and moved synch_frame getting to usbd-endp,
 * where it belongs. Fixed some stuff here and there, and starting to get
 * worried on locks ... I'm gonna have to debug *lots* of deadlocks.
 *
 * Revision 1.5  1998/01/11 15:44:19  inaky
 * Finished functions from the descriptor, inteface and endpoint modules,
 * fixed docs, pretty printed and removed craps.
 *
 * Revision 1.4  1998/01/09 15:11:02  inaky
 * Finished implementing; untested.
 *
 * Revision 1.3  1998/01/08 12:19:13  inaky
 * Added copyright strings to all important files
 *
 * Revision 1.2  1997/12/29 00:25:39  inaky
 * Fixed unimplemented functions to return -ENOSYS and moving the khubdd
 * to use signals as commands.
 * Also added function traces on most files that had them disabled (donkey!)
 *
 * Revision 1.1  1997/12/27 19:54:22  inaky
 * Added all the prototypes and skeleton definitions (incomplete) of all
 * of the USBD functions to each source file.
 *
 */

