
/* Universal USB Driver Stack v$Revision: 1.24 $ 
 *
 * (C) 1997 by Iaky Prez Gonzlez <inaky@peloncho.fis.ucm.es>
 *
 * Distributed under the terms of the GNU Public License Version 2
 * See files COPYRIGHT and COPYING at the root distribution directory
 *
 * $Id: usbd-ops.c,v 1.24 1998/11/14 17:15:15 inaky Exp $
 *
 * Changelog at the bottom 
 *
 *
 *  USB Driver - Devices section - USB standard operations
 *  ------------------------------------------------------
 *
 * Standard and common operations on USB hosts.
 */

#include <uusbd/usbd.h>
#include <uusbd/usb-nls.h>

#define GDLF_IDENT          "USBD"
#define GDLF_LOCAL          /* Define if want local settings */
#define GDLF_DEF_LOGLEVEL   9
#define GDLF_ASSERTIONS     /* want assertions (#undef if not) */
#undef GDLF_FNTRACE        /* want function traces */
#define GDLF_LOCATIONS      /* want locations in log messages */
#include <uusbd/gdlf.h>


__u16 usb_lang =            /* Language ID used by default */
    USB_LNG_BUILD (USB_LNG_ENGLISH, USB_SLNG_ENGLISH_US);


    /* Do a data-less USB request
     *
     *  This function is used to make a data-less request. It is used
     * to simplify the repetitive process of request-making. It makes
     * all the error handling, complaining and transaction fill up,
     * set up, insertion, execution and removal. However, you're free
     * to use your own methods to do requests. This is only provided
     * as a facility. 
     *
     * Parameters:  
     *
     *  ts_usb_device*  Device to make the request to.
     *
     *  ts_usb_request* Pointer to request structure. You must fill it
     *                  with all the appropiate data for the
     *                  request. Note that DURING the execution of
     *                  this function, which BLOCKS until the request
     *                  finishes, this structure may be translated to
     *                  little endian. When finished, it'll be
     *                  translated back to host endian order.
     *
     *  const char*     Request name. This is used for the error
     *                  messages.
     *
     * Returns:
     *
     *  int             0 if ok, errno code < 0 on error.
     *
     * Global:
     *
     *  int usb_errno   Unmodified if ok, USB errno code < 0 on
     *                  error.  
     */

int usb_request_dataless (ts_usb_device* device, 
                          ts_usb_request* req, const char* name)
{
    ssize_t errcode;
    ts_usb_tr tr;
    
    gdlf_fntrace ("usb_request_dataless (0x%p, 0x%p, \"%s\")", 
                  device, req, name);

    gdlf_aassert (req != NULL, return -EINVAL);
    gdlf_aassert (name != NULL, return -EINVAL);

    ts_usb_tr_clean (&tr);          /* tr in CLEAN state */
    
    tr.open = &device->endp0d;      /* transaction in FILLED status */
    tr.req = req;
    req->length = 0;                /* _empty_, no data */
    tr.done_warn = 1;
    tr.status = e_usb_tr_status_filled;
    errcode = usb_tr_setup (&tr);
    if (errcode < 0)
    {
        gdlf_err ("Cannot setup request '%s' to device "
                  "0x%p (%d %d): errno/usb %d/%d\n",
                  name, device, device->hub->host->addr, 
                  device->addr, errcode, usb_errno);
        usb_tr_destroy (&tr);
        return errcode;
    }

    errcode = usb_tr_insert (&tr);
    if (errcode < 0)
    {
        gdlf_err ("Cannot insert request '%s' to device "
                  "0x%p (%d %d): errno/usb %d/%d\n",
                  name, device, device->hub->host->addr, 
                  device->addr, errcode, usb_errno);
        usb_tr_destroy (&tr);
        return errcode;
    }

    errcode = usb_tr_run (&tr);
    if (errcode < 0)
    {
        gdlf_err ("Cannot run request '%s' to device "
                  "0x%p (%d %d): errno/usb %d/%d\n",
                  name, device, device->hub->host->addr, 
                  device->addr, errcode, usb_errno);
        usb_tr_destroy (&tr);
        return errcode;
    }

        /* 
         * After this, the transaction must be in the ERROR, DONE
         * or ABORTED states, as the usb_tr_run() call was a
         * blocking call.
         */
        
    gdlf_assert (tr.status >= e_usb_tr_status_ok &&
                 tr.status < e_usb_tr_status_max);
        
    switch (tr.status)
    {
      case e_usb_tr_status_ok:  /* Transfered ok? check it */
        if (tr.xfered != USB_REQ_XFER_LENGTH)
        {
            gdlf_err ("Error doing request '%s' to device "
                      "0x%p (%d %d): only %d bytes of " 
                      gdlf_stringfy (USB_REQ_XFER_LENGTH) 
                      " written in setup phase\n",
                      name, device, device->hub->host->addr, 
                      device->addr, tr.xfered);
            usb_errno = -USBE_SHORT_WRITE;
            errcode = -EIO;
            break;
        }
        else
            errcode = 0;
        break;
            
      case e_usb_tr_status_aborted:
        errcode = -EINTR;
        usb_errno = -USBE_TR_ABORTED;
        gdlf_err ("Error doing request '%s' to device "
                  "0x%p (%d %d): errno/usb %d/%d\n", 
                  name, device, device->hub->host->addr,
                  device->addr, tr.errno, tr.usb_errno); 
        break;
        
            /* complain in the default: case */
            
      default:                  /* all error cases + abort*/
        gdlf_err ("Error doing request '%s' to device "
                  "0x%p (%d %d): errno/usb %d/%d\n", 
                  name, device, device->hub->host->addr,
                  device->addr, tr.errno, tr.usb_errno); 
        errcode = tr.errno;
        usb_errno = tr.usb_errno;
    }
    usb_tr_destroy (&tr);
    return errcode;
}


    /* Do a USB request (with data phase)
     *
     *  This function is used to make a request with data. It is used
     * to simplify the repetitive process of request-making. It makes
     * all the error handling, complaining and transaction fill up,
     * set up, insertion, execution and removal. However, you're free
     * to use your own methods to do requests. This is only provided
     * as a facility. 
     *
     * Parameters:  
     *
     *  ts_usb_device*  Device to make the request to.
     *
     *  ts_usb_request* Pointer to request structure. You must fill it
     *                  with all the appropiate data for the
     *                  request. Note that DURING the execution of
     *                  this function, which BLOCKS until the request
     *                  finishes, this structure may be translated to
     *                  little endian. When finished, it'll be
     *                  translated back to host endian order.
     *
     *                  The size of the data buffer is implied in the
     *                  request structure. 
     *  
     *  void*           Data for the data phase. If this is a read
     *                  request, data will be placed there, else it'll
     *                  be read from there.
     *
     *  const char*     Request name. This is used for the error
     *                  messages.
     *
     * Returns:
     *
     *  ssize_t         Bytes transfered (in data phase) if ok, errno
     *                  code < 0 on error.  
     *
     * Global:
     *
     *  int usb_errno   Unmodified if ok, USB errno code < 0 on
     *                  error.  
     */

ssize_t usb_request_data (ts_usb_device* device, ts_usb_request* req, 
                          void* data, const char* name)
{
    ssize_t errcode;
    ts_usb_tr tr;
    
    gdlf_fntrace ("usb_request_data (0x%p, 0x%p, 0x%p, \"%s\")", 
                  device, req, data, name);

    gdlf_aassert (data != NULL, return -EINVAL);
    gdlf_aassert (req != NULL, return -EINVAL);
    gdlf_aassert (name != NULL, return -EINVAL);

    if (req->length == 0 && data != NULL)
    {
        gdlf_err ("Cannot transfer %d bytes from a NULL "
                  "buffer to device 0x%p (%d %d)\n",
                  req->length, device, device->hub->host->addr, 
                  device->addr);
        return -EINVAL;
    }

    ts_usb_tr_clean (&tr);          /* tr in CLEAN state */

    tr.open = &device->endp0d;      /* transaction in FILLED status */
    tr.req = req;
    tr.data = data;
    tr.data_size = tr.size = req->length;
    tr.done_warn = 1;
    tr.status = e_usb_tr_status_filled;
        
    errcode = usb_tr_setup (&tr);
    if (errcode < 0)
    {
        gdlf_err ("Cannot setup request '%s' to device "
                  "0x%p (%d %d): errno/usb %d/%d\n",
                  name, device, device->hub->host->addr, 
                  device->addr, errcode, usb_errno);
        usb_tr_destroy (&tr);
        return errcode;
    }

    errcode = usb_tr_insert (&tr);
    if (errcode < 0)
    {
        gdlf_err ("Cannot insert request '%s' to device "
                  "0x%p (%d %d): errno/usb %d/%d\n",
                  name, device, device->hub->host->addr, 
                  device->addr, errcode, usb_errno);
        usb_tr_destroy (&tr);
        return errcode;
    }

    errcode = usb_tr_run (&tr);
    if (errcode < 0)
    {
        gdlf_err ("Cannot run request '%s' to device "
                  "0x%p (%d %d): errno/usb %d/%d\n",
                  name, device, device->hub->host->addr, 
                  device->addr, errcode, usb_errno);
        usb_tr_destroy (&tr);
        return errcode;
    }

        /* 
         * After this, the transaction must be in the ERROR, DONE
         * or ABORTED states, as the usb_tr_run() call was a
         * blocking call.
         */
        
    gdlf_assert (tr.status >= e_usb_tr_status_ok &&
                 tr.status < e_usb_tr_status_max);
        
    switch (tr.status)
    {
      case e_usb_tr_status_ok:  /* Transfered ok? check it */
        if (tr.xfered < USB_REQ_XFER_LENGTH)
        {
            gdlf_err ("Error doing request '%s' to device "
                      "0x%p (%d %d): only %u bytes of " 
                      gdlf_stringfy (USB_REQ_XFER_LENGTH) 
                      " written in setup phase\n",
                      name, device, device->hub->host->addr, 
                      device->addr, tr.xfered);
            usb_errno = -USBE_SHORT_WRITE;
            errcode = -EIO;
            break;
        }
        else if (tr.xfered < USB_REQ_XFER_LENGTH + req->length)
        {
            gdlf_err ("Error doing request '%s' to device "
                      "0x%p (%d %d): only %u bytes of " 
                      "%u transferred in data phase\n",
                      name, device, device->hub->host->addr, 
                      device->addr, tr.xfered, req->length);
            usb_errno = -USBE_SHORT_WRITE;
            errcode = -EIO;
            break;
        }
        else
            errcode = 0;
        break;
            
      case e_usb_tr_status_aborted:
        errcode = -EINTR;
        usb_errno = -USBE_TR_ABORTED;
        gdlf_err ("Error doing request '%s' to device "
                  "0x%p (%d %d): errno/usb %d/%d\n", 
                  name, device, device->hub->host->addr,
                  device->addr, tr.errno, tr.usb_errno); 
        break;
        
            /* complain in the default: case */
            
      default:                  /* all error cases + abort*/
        gdlf_err ("Error doing request '%s' to device "
                  "0x%p (%d %d): errno/usb %d/%d\n", 
                  name, device, device->hub->host->addr,
                  device->addr, tr.errno, tr.usb_errno); 
        errcode = tr.errno;
        usb_errno = tr.usb_errno;
    }
    errcode = (errcode < 0)? errcode : tr.xfered - USB_REQ_XFER_LENGTH;
    usb_tr_destroy (&tr);
    return errcode;
}


    /*  Soft reset a USB device.
     *
     *  Resetting a USB device brings it to a default state, in which
     * it `dissapears' from the system. Thus, we'd had to re-add it to
     * the USB system, reconfigure it, etc. Ok. What this function
     * should do is to reset a device keeping it into the USB
     * system. It should detach it, reset it, and attach it again,
     * once resetted, keeping the same structure for the device.
     *
     *  In this case, the most likely approach is to unconfigure the
     * device and send a reset on the port. As the port structure will
     * be still keeping the pointer to the device structure, the
     * device is still registered at the host and hub, and we just
     * have to configure it again.
     *
     * Parameters:  
     *
     *  ts_usb_device*  Device structure of the device we wish to
     *                  reset.
     *
     * Returns:
     *
     *  int usb_errno   0 if ok, errno code < 0 On error.
     *
     * Global:
     *
     *  int usb_errno On error, error code.  
     */

int usb_device_reset (ts_usb_device* device)
{
    int errcode;

    gdlf_fntrace ("usb_device_reset (0x%p)", device);
    
    errcode = usb_device_unconfig (device, USB_CFG_RESET);
    if (errcode)
        gdlf_err ("Couldn't unconfigure properly the device. "
                  "Continuing anyway with the reset process. "
                  "Expect memory leaks.\n");

    errcode = usb_port_reset (device->hub, device->port);
    if (errcode)
        gdlf_err ("Couldn't reset properly the device's port. "
                  "Continuing anyway with the reset process. "
                  "Expect memory leaks.\n");

#warning FIXME: fix this, reassign same address to the device, how?
    errcode = usb_device_config (device, USB_CFG_RESET);
    if (errcode)
        gdlf_err ("Couldn't configure properly the device "
                  "after resetting it.\n");

    return errcode;
}


    /* Dirty-work function for usb_device_feature_*()
     *
     *  THIS FUNCTION IS FOR INTERNAL USE, DO NOT CALL IT DIRECTLY
     *
     * Parameters:  
     *
     *  ts_usb_device*  Pointer to the device structure whose feature
     *                  we want to set.
     *
     *  e_usb_featsel   Feature
     *
     *  e_usb_req       USB request code
     *
     * Returns:
     *
     *  int             0 if ok, errno code < 0 on error.
     *
     * Global:
     *
     *  int usb_errno   On error, error code.  
     */

int usb_device_feature (ts_usb_device* device, 
                        e_usb_featsel feature, e_usb_req request)
{
    ssize_t errcode;
    ts_usb_request req;
    
    gdlf_fntrace ("usb_device_feature (0x%p, %d, %u)",
                  device, feature, request);

    gdlf_aassert (device != NULL, return -EINVAL);

    req.bmreqtype = 0x0;    /* device request */
    req.request = request;
    req.value = feature;
    req.index = 0;

    errcode = usb_request_dataless (device, &req, 
                                    "feature change");
    if (errcode < 0)
    {
        gdlf_err ("Cannot clear/set feature; request failed: "
                  "errno/usb %d/%d\n", errcode, usb_errno);
        return errcode;
    }    
    return errcode;
}


    /* Get device's status
     *
     *  Return the device's status and actualize in the device's
     * structure. No need to do byte-order conversions, the function  
     * already does.
     *
     * Parameters:  
     *
     *  ts_usb_device*  Pointer to the device whose status we want to
     *                  obtain.
     *
     * Returns:
     *
     *  __s32           If > 0, 16 bits status code (and with the
     *                  USB_DEV_STAT_* macros to get the bits). Errno
     *                  code < 0 on error. 
     *
     * Global:
     *
     *  int usb_errno   On error, error code.  
     */

__s32 usb_device_status_get (ts_usb_device* device)
{
    int errcode;
    ts_usb_request req;
    __u16 status;
    
    gdlf_fntrace ("usb_device_status_get (0x%p)", device);

    gdlf_aassert (device != NULL, return -EINVAL);

    req.bmreqtype = 0x80;   /* Get device ... */
    req.request = e_usb_req_get_status;
    req.value = 0;
    req.index = 0;
    req.length = sizeof (status);
    
    write_lock (&device->msnode.lock);
    status = 0;
    errcode = usb_request_data (device, &req, &status, 
                                "get device status");
    if (errcode < 0)
    {
        gdlf_err ("Cannot get device status; request failed: "
                  "errno/usb %d/%d\n", errcode, usb_errno);
        write_unlock (&device->msnode.lock);
        return errcode;
    }

    status = le16_to_cpu (status);          /* change endianess */
    device->selfpwrd = status & USB_DEV_STAT_SELF_PWR? 1 : 0;
    device->remwakeup = status & USB_DEV_STAT_REM_WAKEUP? 1 : 0;
    write_unlock (&device->msnode.lock);
    return status;
}


    /* Set device's address
     *
     *  Note that in case of any error, the device's address will be
     * kept unchanged. 
     *
     *  Note you must have the device locked! This will do no locking
     * now acknowledge them.
     *
     * Parameters:  
     *
     *  ts_usb_device*  Pointer to the device whose address we want to
     *                  set.
     *
     *  t_usb_addr      New device's address.
     *
     * Returns:
     *
     *  int             0 if ok, errno code < 0 on error.
     *
     * Global:
     *
     *  int usb_errno   On error, error code.  
     */

int usb_device_address_set (ts_usb_device* device, t_usb_addr addr)
{
    int errcode;
    ts_usb_request req;
    
    gdlf_fntrace ("usb_device_address_set (0x%p, %d)",
                  device, addr);

    gdlf_aassert (device != NULL, return -EINVAL);
    gdlf_passert (addr != 0,
                  "Sure you want to set device 0x%p's "
                  "address to zero?\n", device);

    req.bmreqtype = 0x0;
    req.request = e_usb_req_set_address;
    req.value = addr;
    req.index = 0;
    req.length = 0;

    errcode = usb_request_dataless (device, &req, 
                                    "set device address");
    if (errcode < 0)
        gdlf_err ("Cannot set device address; request failed: "
                  "errno/usb %d/%d\n", errcode, usb_errno);
    else
        device->addr = addr;
    return errcode;
}


    /* Allocate and fill a string from a USB string descriptor
     *
     *  This function will get a USB string descriptor, allocate space
     * for it and return it's pointer. You're responsible to free it
     * after use.
     *
     * FIXME: Should we translate it? [from UTF8/UNICODE to
     *        ISO-ANYTHING].
     *
     * Parameters:  
     *
     *  ts_usb_device*  Pointer to the device structure from where to
     *                  get the string.
     *
     *  __u8            String's descriptor index. If requested index
     *                  zero, the device will return an array of
     *                  supported string codes. As this funciton is
     *                  made to return an string, it is considered
     *                  that a zero index means that string is not
     *                  available.
     *
     *  int             Language ID. It's a 16 bit code, as defined by
     *                  a doc by M$. See USB Spec 1.0 9.6.5. If < 0,
     *                  it'll assumed you want to use default USB
     *                  language code, contained in the variable
     *                  usb_lang. See also usb-nls.h
     *
     * Returns:
     *
     *  int             0 if ok, errno code < 0 on error.
     *
     * Global:
     *
     *  int usb_errno   On error, error code.  
     */

char* usb_string_get (ts_usb_device* device, __u8 index, int lang)
{
    __u8 *str, *itr;
    __u16 *unival, *top;

    ssize_t errcode;
    size_t length;
    ts_usb_str_descr desc;
    __u8 strdata[2];
    ts_usb_request req;
    
    if (index == 0)         /* Index zero? ... */
        return NULL;        /* ... means unavailable string */

    req.bmreqtype = 0x80;
    req.request = e_usb_req_get_descriptor;
    req.value = e_usb_desc_str << 8 | index;
    req.index = lang < 0? usb_lang : lang;
    req.length = sizeof (strdata);
    
#warning FIXME: optimize using a single transaction

        /* Get first two bytes to guess how big is the string */

    errcode = usb_request_data (device, &req, &strdata, 
                                "get string header");
    if (errcode < 0)
    {
        gdlf_err ("Cannot get string index %u [lang %d] "
                  "from device 0x%p (%d %d): "
                  "errno/usb_errno %d/%d\n",
                  index, lang, device, device->hub->host->addr,
                  device->addr, errcode, usb_errno);
        return NULL;
    }
    if (errcode < sizeof (strdata))   /* less then expected? */
    {
        gdlf_err ("Cannot get string index %u [lang %d] "
                  "from device 0x%p (%d %d): short read, "
                  "readen %d bytes, expected %d\n",
                  index, lang, device, device->hub->host->addr,
                  device->addr, errcode, strdata[1]);
        usb_errno = -USBE_SHORT_READ;
        return NULL;
    }

        /* Ok, get the full descriptor */
    
    req.length = strdata[0];
    errcode = usb_request_data (device, &req, &desc, "get string");
    if (errcode < 0)
    {
        gdlf_err ("Cannot get string index %u [lang %d] "
                  "from device 0x%p (%d %d): "
                  "errno/usb_errno %d/%d\n",
                  index, lang, device, device->hub->host->addr,
                  device->addr, errcode, usb_errno);
        return NULL;
    }
    if (errcode < strdata[0])   /* less then expected? */
    {
        gdlf_err ("Cannot get string index %u [lang %d] "
                  "from device 0x%p (%d %d): short read, "
                  "readen %d bytes, expected %d\n",
                  index, lang, device, device->hub->host->addr,
                  device->addr, errcode, strdata[0]);
        usb_errno = -USBE_SHORT_READ;
        return NULL;
    }

        /* Mind length already counts one char for trailing \0 */

    length = sizeof (__u8) * (desc.head.bLength - 1);
    str = usb_kmalloc (length, GFP_KERNEL);
    if (str == NULL)
    {
        gdlf_err ("Cannot allocate %u bytes for string "
                  "index %u [lang %u] from device 0x%p (%d %d): "
                  "errno/usb_errno %d/%d\n",
                  length, index, lang, device, 
                  device->hub->host->addr, device->addr, 
                  errcode, usb_errno);
        return NULL;
    }    
    str[length - 1] = 0;
    
#warning FIXME: UNICODE xlation not implemented (quick hack done)
#undef CONFIG_NLS
#ifdef CONFIG_NLS
#else
        /* quick dirty method :) strip high byte */

    unival = (__u16*)&desc.bString;
    top = unival + length / 2;
    itr = str;
    while (unival < top)
        *itr++ = (__u8)le16_to_cpu (*unival++) & 0x00ff;
    *itr = 0;
#endif
    return str;
}


/* Log:
 *
 * $Log: usbd-ops.c,v $
 * Revision 1.24  1998/11/14 17:15:15  inaky
 * Moved the include files to uusbd/include/uusbd, to ease integration
 * into the Linux kernel.
 *
 * Revision 1.23  1998/10/22 00:00:16  inaky
 * Removed traces.
 *
 * Revision 1.22  1998/09/24 15:57:34  inaky
 * Moved 'hub_port' to 'port'.
 *
 * Revision 1.21  1998/08/28 23:47:50  inaky
 * Fixed request! they have not to be xlated to little endian,
 * usb_request_* does it.
 *
 * Revision 1.20  1998/08/27 22:16:12  inaky
 * Changed address to make it little endian when requesting address set.
 *
 * Revision 1.19  1998/08/24 02:26:46  inaky
 * Changed name of all enumerations to be prefixed with the enumeration
 * type name.
 *
 * Did the little debugging needed to make the functions work as they're
 * expected to. Fixed usb_string_get() to return NULL when requested
 * string with descriptor index 0.
 *
 * Revision 1.18  1998/08/19 23:52:36  inaky
 * Fixed comments for usb_request_data(); Added assertions.
 *
 * Revision 1.17  1998/08/03 13:55:17  inaky
 * Fixed error messages to report error codes correctly and made
 * usb_device_address_set() to set the address at the device structure.
 *
 * Revision 1.16  1998/07/30 22:57:52  inaky
 * No changes.
 *
 * Revision 1.15  1998/07/23 14:30:54  inaky
 * Renamed the ts_usb_trbuf struct to ts_usb_tr.
 *
 * Added the length parameter to the request structure.
 *
 * Revision 1.14  1998/07/16 14:13:07  inaky
 * Added two functions for making requests (one without data phase, other
 * one with data phase).
 *
 * Moved from the old read/write interface to the new, request-based
 * interface and fixed locks.
 *
 * Revision 1.13  1998/06/20 21:54:54  inaky
 * Massive changes to split the old monolithic hubd driver into a hub
 * manager (hub wrappers and daemon) and a dynamic driver, which can be
 * replaced and its different parts reused.
 *
 * Revision 1.12  1998/05/01 19:10:06  inaky
 * Added the USB NLS include file; added the default USB language for
 * selection; fixed the crappy [and temporary] UNICODE to ASCII
 * converter.
 *
 * Revision 1.11  1998/04/26 03:37:32  inaky
 * Added the high-level, easy string retrieving function.
 *
 * Revision 1.10  1998/04/13 22:15:35  inaky
 * Fixed docs, fixed to use the new callbacks scheme on trbuffers, moved
 * status functions to return status instead of copying it via
 * pointers. Also made them actualize status on structures.
 *
 * Revision 1.9  1998/03/28 19:35:47  inaky
 * Removed inclusion of usb-methods.h and fixed to use callbacks.
 *
 * Revision 1.8  1998/02/12 16:35:10  inaky
 * Finished the device ops, and moved synch_frame getting to usbd-endp,
 * where it belongs. Fixed some stuff here and there, and starting to get
 * worried on locks ... I'm gonna have to debug *lots* of deadlocks.
 *
 * Revision 1.7  1998/01/11 15:44:20  inaky
 * Finished functions from the descriptor, inteface and endpoint modules,
 * fixed docs, pretty printed and removed craps.
 *
 * Revision 1.6  1998/01/08 12:19:13  inaky
 * Added copyright strings to all important files
 *
 * Revision 1.5  1998/01/08 11:49:13  inaky
 * Finishing the configuration system. Only the final cfel-stuff is
 * missing to implement; however, nothing is tested.
 *
 * Revision 1.4  1997/12/31 17:06:35  inaky
 * Removed some unneeded locks and fixed some others.
 *
 * Revision 1.3  1997/12/31 05:37:01  inaky
 * Almost finished all the lists and life cycle for hosts, drivers and
 * devices. Some craps to be fixed yet.
 *
 * Revision 1.2  1997/12/29 00:25:40  inaky
 * Fixed unimplemented functions to return -ENOSYS and moving the khubdd
 * to use signals as commands.
 * Also added function traces on most files that had them disabled (donkey!)
 *
 * Revision 1.1  1997/12/27 19:54:22  inaky
 * Added all the prototypes and skeleton definitions (incomplete) of all
 * of the USBD functions to each source file.
 *
 */

