
/* Universal USB Driver Stack v$Revision: 1.19 $ 
 *
 * (C) 1997 by Iaky Prez Gonzlez <inaky@peloncho.fis.ucm.es>
 *
 * Distributed under the terms of the GNU Public License Version 2
 * See files COPYRIGHT and COPYING at the root distribution directory
 *
 * $Id: usbd-tr.c,v 1.19 1998/11/17 10:04:35 inaky Exp $
 *
 * Changelog at the bottom 
 *
 *
 *  USB Driver
 *  ----------
 *
 *  Transactions API; these are all the operations to do the USB
 * transactions on the bus (deprecating the old rw/crw interfaces).
 *
 * WARNING!! All these functions do not do any locking. They expect
 *           the device to be already locked for them!!!
 *
 *  See the doc file 'Transaction Request Life Cycle.fig' for more
 * graphical info :).
 *
 *  The transactions may be in one of these states; note that all the
 * up transitions (CLEAN -> FILLED -> SETUP -> INSERTED -> QUEUED ->
 * EXECUTING -> {ABORTED, DONE, ERROR}) happen one step
 * at a time, whereas the down transitions (the inverse), may happen
 * various steps at the time (for example, they'll bring an ABORTED
 * state to FILLED with a usb_tr_setdn() call):
 * 
 * CLEAN:       It is the status after creation. There is no
 *              information filled in now HCD data, anything. The
 *              usb_tr_destroy() function takes down a transaction in any
 *              state to this one. Filling in the data fields moves it
 *              to the FILLED status [and exceptionally, you must
 *              set tr->status = e_usb_tr_status_filled to reflect the
 *              status chane].
 * 
 * FILLED:      Information is filled in, but no data structured
 *              created by HCD. The transaction is not in the
 *              ts_usb_open descriptor list. usb_tr_setdn() brings
 *              down a transaction to this state. Calling
 *              usb_tr_setup() brings that transaction to the SETUP
 *              state. 
 * 
 * SETUP:       HCD, taken the information on the struct, has created
 *              it's data structures and they are hanging from the
 *              structure. Also the transaction has been added to the
 *              list of transactions for the ts_usb_open descriptor
 *              queue [appended in cronological order]. usb_tr_setdn()
 *              will move to FILLED a transaction in any state, by
 *              freeing and clearing the HCI data. Calling
 *              usb_tr_insert() will move to the INSERTED state. 
 * 
 * INSERTED:    The HCD has inserted the request's HCI data structures
 *              into the transfer list, but they are
 *              inactive. usb_tr_remove() moves down to SETUP state. A
 *              call to usb_tr_execute() moves to the EXECUTING state.
 * 
 * QUEUED:      A transaction is QUEUED when it is ready to run, but
 *              there is still a transaction to the same endpoint
 *              executing and it has to wait for it to
 *              finish. usb_tr_run() moves a transaction in the
 *              INSERTED state to this state. A call to usb_tr_abort()
 *              moves to the ABORTED state. It will transition to the
 *              EXECUTING state when the HCD successfully finishes the
 *              last executing transaction. 
 *
 * EXECUTING:   The HCD activates the in-transfer-list-inserted data
 *              structures so that the transfer begins. If the loop is
 *              active, when completed, it'll be activated (executed)
 *              again; the various warnings described in the
 *              transaction structure will be activated when due. A
 *              call to usb_tr_reset() will bring it back to INSERTED
 *              state after going trough the ABORTED state. A call to
 *              usb_tr_abort() will move it to the ABORTED state. When
 *              the transaction finishes (all the loops et al), it
 *              moves to the DONE state. If any error happens, the
 *              transaction is moved to the ERROR state. 
 *
 * ABORTED:     The transaction has been aborted. This state is the
 *              same as inserted, but the transfer is not
 *              executing. You must usb_tr_reset() it to INSERTED
 *              state to be able to execute it again.
 *
 * DONE:        The transacion finished succesfully, after all the
 *              loops, et al. You must You must usb_tr_reset() it to
 *              INSERTED state to be able to execute it again.
 *
 * ERROR:       The transacion finished on error. You must usb_tr_reset() it
 *              to INSERTED state to be able to execute it again
 *              (after, of course, solving the problem which caused
 *              the error).  
 *
 *  To optimize for certain HCIs, some steps are omitted. They're
 * #if'ed to evitate silly delays. They should be optimized better.
 */

#warning FIXME: all this code needs locking

#include <linux/kernel.h>
#include <asm/uaccess.h>
#include <asm/current.h>

#include <uusbd/usbd.h>

#define GDLF_IDENT          "USBD"
#define GDLF_LOCAL          /* Define if want local settings */
#define GDLF_DEF_LOGLEVEL   9
#define GDLF_ASSERTIONS     /* want assertions (#undef if not) */
#undef GDLF_FNTRACE        /* want function traces */
#define GDLF_LOCATIONS      /* want locations in log messages */
#include <uusbd/gdlf.h>

    /* Negative values mean errors, positive conditions */

int e_usb_tr_status_2_errno[e_usb_tr_status_max] =
{   
    EINVAL,
    ENODATA,
    ENOTCONN,
    EISCONN,
    EBUSY,
    EINPROGRESS,
    0,
    EINTR,
    -EIO
};


    /* Checks the request parameters to a USB request call to be ok to
     * do the request. In case of any error, it will return an errno
     * code and modify the errno and usb_errno fields of the
     * ts_usb_tr structure.
     *
     * NOTE: this function is not intended to be public, but to be
     *       called by usb_tr_setup().
     *
     * Parameters:  
     *
     *  ts_usb_tr*      Transfer buffer descriptor to the hub's control
     *                  endpoint. Will have the errno and usb_errno
     *                  fields updated in case of error.
     *
     *  const ts_usb_request*
     *
     *                  Pointer to the data for the USB request.
     *
     * Returns:
     *
     *  int             0 if ok, < 0 errno code on error and the errno
     *                  and usb_errno fields of the ts_usb_tr*
     *                  structure updated.
     *
     *                  -EINVAL     Some parameter of the 
     *                              type or recipient fields is not
     *                              correct.
     * 
     *                  -EOPNOTSUPP The destination endpoint is not a
     *                              control endpoint and doesn't
     *                              support the operation.
     *
     * Global:
     *
     *  int usb_errno   On error, error code.
     */

int usb_request_check (ts_usb_tr* tr)
{
    gdlf_fntrace ("usb_request_check (0x%p)", tr);
    
    gdlf_aassert (tr != NULL, return -EINVAL);
    gdlf_aassert (tr->req != NULL, return -EINVAL);
    
        /* Check the request */

    if (ts_usb_request_type (tr->req->bmreqtype) > e_usb_req_type_vendor)
    {
        usb_errno = tr->usb_errno = -USBE_REQ_BAD_TYPE;
        return tr->errno = -EINVAL;
    }
    
    if (ts_usb_request_rec (tr->req->bmreqtype) > e_usb_req_rec_other)
    {
        usb_errno = tr->usb_errno = -USBE_REQ_BAD_REC;
        return tr->errno = -EINVAL;
    }

    return 0;
}


    /* Move from the FILLED state to the SETUP state
     *
     *  We just have to call the host's setup function on the
     * transaction and see the result code. We have to handle it the
     * possible USB request data.
     *
     * Parameters:  
     *
     *  ts_usb_tr*      Pointer to transaction structure.
     *
     * Returns:
     *
     *  int             0 if ok, errno code < 0 on error.
     *
     * Global:
     *
     *  int usb_errno   Unmodified if ok, USB errno code < 0 on
     *                  error. 
     */

int usb_tr_setup (ts_usb_tr* tr)
{
    int errcode;
    
    gdlf_fntrace ("usb_tr_setup (0x%p)", tr);

    gdlf_aassert (tr != NULL, return -EINVAL);
    gdlf_aassert (tr->open != NULL, return -EINVAL);

        /*
         *  Check the open descriptor is ok 
         *
         *  This step is not needed on following steps, as the state
         * will put a barrier to it.
         */

    if (tr->open->endp == NULL)
        return -EBADFD;
    if (tr->open->host == NULL)
        return -EBADFD;
    if (tr->open->host->tr_setup == NULL)
        return -EBADFD;

    if (tr->status != e_usb_tr_status_filled)
    {
        gdlf_err ("Cannot setup transaction 0x%p "
                  "as it is not in the FILLED state "
                  "(it is in state %d)\n",
                  tr, tr->status);
        return -EBADFD;
    }
    
        /* Check the request (if any) */

    if (tr->open->endp->type == e_usb_tr_type_cntl)
        if (tr->req)
        {
            errcode = usb_request_check (tr);
            if (errcode < 0)
            {
                gdlf_debug ("Cannot setup transaction 0x%p, "
                            "bad request: errno/usb %d/%d\n", 
                            tr, errcode, usb_errno);
                return errcode;
            }
        }
        else
        {
            gdlf_debug ("Cannot setup transaction 0x%p, required "
                        "USB request data missing\n", tr);
            return -EINVAL;
        }

        /* Check if trying to do OUTs on Interrupt */

    if (tr->open->endp->type == e_usb_tr_type_intr)
        gdlf_apassert ((tr->open->endp->addr & 0x1) != 0,
                       return -EINVAL,
                       "Huh ... trying to do OUT on interrupt "
                       "endpoint %d from device 0x%p (%d %d)\n",
                       tr->open->endp->addr, tr->open->dev, 
                       tr->open->dev->hub->host->addr >> 1,
                       tr->open->dev->addr);

        /* Ok, set it up */

    errcode = (*tr->open->host->tr_setup) (tr);
    if (errcode < 0)
    {
        gdlf_debug ("Cannot setup transaction 0x%p: "
                    "errno/usb %d/%d\n", 
                    tr, errcode, usb_errno);
        return errcode;
    }
    ts_dtor_fill (&tr->snode.dtor, (tf_dtor)usb_tr_destroy);
    
        /* Add it to the descriptor's list */

    errcode = ts_slist_append (&tr->open->list, &tr->snode);
    gdlf_passert (errcode == 0, 
                  "Couldn't add transaction 0x%p to open "
                  "descriptors' 0x%p list!: errno %d\n",
                  tr, tr->open, errcode);

        /* prepare the timer */

    tr->timer.data = (unsigned long)tr;
    tr->timer.function = (void(*)(unsigned long))usb_tr_timer;

        /* We can set the request to little endian if it exists */
    
    if (tr->req)
    {
        tr->req->value = cpu_to_le16 (tr->req->value);
        tr->req->index = cpu_to_le16 (tr->req->index);
        tr->req->length = cpu_to_le16 (tr->req->length);
    }

    tr->status = e_usb_tr_status_setup;
    return 0;
}


    /* Move from the SETUP, INSERTED, EXECUTING, ABORTED, DONE or
     * ERROR states to the FILLED state.
     *
     *  This function is used to get a transaction out of one of the
     * given states to the FILLED state, cleanly releasing all the
     * resources adquired (except if the state in which the
     * transaction is is below the FILLED state).
     *
     *  To accomplish this, we simply bring the transaction to the
     * INSERTED state and from there to FILLED.
     *
     * Parameters:  
     *
     *  ts_usb_tr*      Pointer to transaction structure.
     *
     * Returns:
     *
     *  int             0 if ok, errno code < 0 on error.
     *
     * Global:
     *
     *  int usb_errno   Unmodified if ok, USB errno code < 0 on
     *                  error. 
     */

int usb_tr_setdn (ts_usb_tr* tr)
{
    int errcode;
    
    gdlf_fntrace ("usb_tr_setdn (0x%p)", tr);

    gdlf_aassert (tr != NULL, return -EINVAL);
    gdlf_aassert (tr->open != NULL, return -EINVAL);
    gdlf_aassert (tr->open->host != NULL, return -EINVAL);
    gdlf_aassert (tr->open->host->tr_setdn != NULL, return -ENOSYS);

        /* Do some housekeeping */

    switch (tr->status)
    {
      case e_usb_tr_status_clean:   /* CLEAN -> CLEAN */
        gdlf_debug ("Tried to set down transaction 0x%p, "
                    "which is clean\n", tr);
        return 0;

      case e_usb_tr_status_filled:  /* FILLED -> FILLED */
        gdlf_debug ("Tried to set down transaction 0x%p, "
                    "which is already set down (filled)\n", tr);
        return 0;
        
      case e_usb_tr_status_inserted:/* INSERTED -> SETUP */
      case e_usb_tr_status_queued:  /* QUEUED -> INSERTED -> SETUP */
      case e_usb_tr_status_executing:/* EXECUTING -> ABORTED -> INSERTED -> SETUP */
      case e_usb_tr_status_ok:      /* DONE -> INSERTED -> SETUP */
      case e_usb_tr_status_aborted: /* ABORTED -> INSERTED -> SETUP */
      case e_usb_tr_status_error:   /* ERROR -> INSERTED -> SETUP */
        errcode = usb_tr_remove (tr); /* remove the transaction */
        if (errcode < 0)
            gdlf_err ("Error removing transaction 0x%p to set "
                      "it down: errno/usb %d/%d. "
                      "Continuing anyway\n", 
                      tr, errcode, usb_errno);

      case e_usb_tr_status_setup:    /* SETUP -> FILLED */
        break;

      default:
        gdlf_apassert (0, return -EINVAL,
                       "Unknown transaction 0x%p state %d\n",
                       tr, tr->status);
    }

        /*
         *  Ok, let's move from SETUP state to FILLED state; this
         * implies we have to free the tr->hci data, which is what the
         * HCI keeps. We just have to call the host-specific
         * destruction method for it.
         */

    if (tr->hci)                /* If there's HCI info */
    {                               /* destroy it */
#warning DEBUG: remove me
#if 1
        errcode = 0;
        gdlf_aassert (tr != NULL, errcode = -EINVAL);
        gdlf_aassert (tr->open != NULL, errcode = -EINVAL);
        gdlf_aassert (tr->open->host != NULL, errcode = -EINVAL);
        gdlf_aassert (tr->open->host->tr_setdn != NULL, errcode = -ENOSYS);
        errcode = errcode? : (*tr->open->host->tr_setdn) (tr);
#else
        errcode = (*tr->open->host->tr_setdn) (tr);
#endif
        if (errcode < 0)
            gdlf_err ("Error destroying HCI data for "
                      "transaction 0x%p: errno/usb %d/%d\n",
                      tr, errcode, usb_errno);
        usb_kfree (tr->hci);        /* free */
        tr->hci = NULL;             /* and clean */
    }
    else
        gdlf_debug ("Ein?? Transaction 0x%p in setup state and "
                    "no HCI data present\n", tr);

        /* Remove from the open descriptor's list */

    read_lock (&tr->open->tr.snode.lock);   /* set up 'finished' to ... */
    if (tr->open->finished == &tr->snode)   /* ... previous snode if last */
    {
        ts_snode *prev;         /* previous snode to tr in open->tr */
        prev = ts_slist_prev (&tr->open->list, &tr->snode);
        tr->open->finished = (prev == &tr->snode)? NULL : prev;
    }
    if (tr->open->running == &tr->snode)
    {
        ts_snode *prev;         /* previous snode to tr in open->tr */
        prev = ts_slist_prev (&tr->open->list, &tr->snode);
        tr->open->running = (prev == &tr->snode)? NULL : prev;
    }
    read_unlock (&tr->open->tr.snode.lock);
    
    errcode = ts_slist_del (&tr->open->list, &tr->snode);
    gdlf_passert (errcode == 0, 
                  "Couldn't remove transaction 0x%p from open "
                  "descriptors' 0x%p list!: errno %d\n",
                  tr, tr->open, errcode);

        /*
         * We are in the FILLED state. If this was a control xfer, we
         * got to leave the request data in host order, from little
         * endian order.
         */
    
    if (tr->req)
    {
        tr->req->value = le16_to_cpu (tr->req->value);
        tr->req->index = le16_to_cpu (tr->req->index);
    }

    tr->status = e_usb_tr_status_filled;
    return 0;
}


    /* Move the transaction from the SETUP state to the INSERTED state
     *              
     *  This implies that the HCD will load/insert the HCI data
     * (tr->tci) into it's transfer list, without activating it, to
     * have it ready to be fired.
     *
     * Parameters:  
     *
     *  ts_usb_tr*      Pointer to transaction structure.
     *
     * Returns:
     *
     *  int             0 if ok, errno code < 0 on error.
     *
     * Global:
     *
     *  int usb_errno   Unmodified if ok, USB errno code < 0 on
     *                  error. 
     */

#warning FIXME: add an schedulement time (absolute) for isochronous xfers

    /*
     *  The Audio Class Spec suggests this is neccessary to
     * synchronize different audio sources in phase [for effects]. See
     * Device Class Definition for Audio Devices v1.0, section
     * 3.4. Seems it will be neccessary to put it into ts_usb_tr
     */

int usb_tr_insert (ts_usb_tr* tr)
{
    int errcode;
    
    gdlf_fntrace ("usb_tr_insert (0x%p)", tr);

    gdlf_aassert (tr != NULL, return -EINVAL);
    gdlf_aassert (tr->open != NULL, return -EINVAL);
    gdlf_aassert (tr->open->host != NULL, return -EINVAL);

    if (tr->status != e_usb_tr_status_setup)
    {
        gdlf_err ("Cannot insert transaction 0x%p "
                  "as it is not in the SETUP state "
                  "(it is in state %d)\n",
                  tr, tr->status);
        return -EBADFD;
    }
    
        /* Ok, insert it */

#if defined (CONFIG_USB_UHCI) || defined (CONFIG_USB_UHCI_MODULE)
    errcode = tr->open->host->tr_insert? 
        (*tr->open->host->tr_insert) (tr) : 0;
    if (errcode < 0)
    {
        gdlf_debug ("Cannot insert transaction 0x%p: "
                    "errno/usb %d/%d\n", 
                    tr, errcode, usb_errno);
        return errcode;
    }
#endif
    tr->status = e_usb_tr_status_inserted;
    return 0;
}


    /* Move from the INSERTED, EXECUTING, ABORTED, DONE or ERROR
     * states to the SETUP state (remove from the transfer list).
     *
     *  If we are above the SETUP state, we move to the INSERTED
     * state, if not, we ignore the call. Then we move from INSERTED
     * to SETUP.
     *
     * Parameters:  
     *
     *  ts_usb_tr*      Pointer to transaction structure.
     *
     * Returns:
     *
     *  int             0 if ok, errno code < 0 on error.
     *
     * Global:
     *
     *  int usb_errno   Unmodified if ok, USB errno code < 0 on
     *                  error. 
     */

int usb_tr_remove (ts_usb_tr* tr)
{
    int errcode;

    gdlf_fntrace ("usb_tr_remove (0x%p)", tr);

    gdlf_aassert (tr != NULL, return -EINVAL);
    gdlf_aassert (tr->open != NULL, return -EINVAL);
    gdlf_aassert (tr->open->host != NULL, return -EINVAL);

        /* Do some housekeeping */

    switch (tr->status)
    {
      case e_usb_tr_status_clean:   /* CLEAN -> CLEAN */
        gdlf_debug ("Tried to remove transaction 0x%p, "
                    "which is clean\n", tr);
        return 0;

      case e_usb_tr_status_filled:  /* FILLED -> FILLED */
        gdlf_debug ("Tried to remove transaction 0x%p, "
                    "which is filled\n", tr);
        return 0;

      case e_usb_tr_status_setup:  /* SETUP -> SETUP */
        gdlf_debug ("Tried to remove transaction 0x%p, "
                    "which is already removed (setup)\n", tr);
        return 0;
        
      case e_usb_tr_status_queued:  /* QUEUED -> INSERTED */
      case e_usb_tr_status_executing:/* EXECUTING -> INSERTED */
      case e_usb_tr_status_ok:      /* DONE -> INSERTED */
      case e_usb_tr_status_aborted: /* ABORTED -> INSERTED */
      case e_usb_tr_status_error:   /* ERROR -> INSERTED */
        errcode = usb_tr_reset (tr); /* remove the transaction */
        if (errcode < 0)
            gdlf_err ("Error resetting transaction 0x%p to "
                      "remove it: errno/usb %d/%d. "
                      "Continuing anyway\n", 
                      tr, errcode, usb_errno);

      case e_usb_tr_status_inserted:/* INSERTED -> SETUP */
        break;

      default:
        gdlf_apassert (0, return -EINVAL,
                       "Unknown transaction 0x%p state %d\n",
                       tr, tr->status);
    }

        /* 
         * Move from the INSERTED to the SETUP status
         */

#if defined (CONFIG_USB_UHCI) || defined (CONFIG_USB_UHCI_MODULE)
    errcode = tr->open->host->tr_remove? 
        (*tr->open->host->tr_remove) (tr) : 0;
    if (errcode < 0)
        gdlf_debug ("Cannot remove transaction 0x%p: "
                    "errno/usb %d/%d\n", 
                    tr, errcode, usb_errno);
#endif
    tr->status = e_usb_tr_status_setup;
    return 0;
}


    /* Execute an INSERTED transaction
     *
     *  This will tell the USB host to execute the transaction,
     * doing all the loops, warning calls, etc
     *
     *  In case tr->done_warn is 1, this call will BLOCK if
     * tr->warn.wq == NULL (or tr->warn.callback == NULL, what's the
     * same) until the transaction goes out of the EXECUTING state (to
     * ABORTED, DONE or ERROR if following the normal flow, or to
     * CLEAN, FILLED, SETUP or INSERTED if externally made so).
     *
     *  If the call is non-blocking (there's a callback or wait queue
     * defined), then the function will return inmediately, and again,
     * on transaction's state changing from EXECUTING to any other,
     * and if tr->done_warn is 1, the callback or wait queue will be
     * launched. 
     *
     *  If there's set up a warning each tr->pkts packets (tr->pkts !=
     * 0), then, each time tr->pkts packets are correctly fed or
     * sucked to/from the wire, the callback or wait queue will be
     * launched (and you can check the call reason checking the tr
     * status and the number of transfered bytes (if it is an exact
     * multiple of the data buffer size and there's looping, it's
     * probably a loop - though in some circumstances it may
     * collide).
     *
     *  If there's set up a warning on each loop, each time the
     * transaction is looped (if it loops at all), then the same
     * mechanism is started. Refer to the above paragraph on how to
     * detect it.
     *
     *  Note you can only execute one transaction at the time. See the
     * docs for ts_usb_open at usb.h.
     *
     * Parameters:  
     *
     *  ts_usb_tr*      Pointer to transaction structure.
     *
     * Returns:
     *
     *  int             0 if ok, errno code < 0 on error. Note this
     *                  error code is the result of putting the
     *                  transaction to run, not the result of the
     *                  transaction itself. For that, use tr->errno,
     *                  tr->usb_errno and tr->status.
     *
     * Global:
     *
     *  int usb_errno   Unmodified if ok, USB errno code < 0 on
     *                  error.  
     */

int usb_tr_run (ts_usb_tr* tr)
{
    ts_snode *prev;
    int errcode;
    
    gdlf_fntrace ("usb_tr_run (0x%p)", tr);

    gdlf_aassert (tr != NULL, return -EINVAL);
    gdlf_aassert (tr->open != NULL, return -EINVAL);
    gdlf_aassert (tr->open->host != NULL, return -EINVAL);
    gdlf_aassert (tr->open->host->tr_run != NULL, return -ENOSYS);

    if (tr->status != e_usb_tr_status_inserted)
    {
        gdlf_err ("Cannot run transaction 0x%p "
                  "as it is not in the INSERTED state "
                  "(it is in state %d)\n",
                  tr, tr->status);
        return -EBADFD;
    }
    
        /* move to tail of running zone of list */

    prev = ts_slist_prev (&tr->open->list, &tr->snode);
    gdlf_aassert (prev != NULL, return -EBADFD);
    ts_slist_move (&tr->open->list, &tr->snode, tr->open->running);
    tr->open->running = &tr->snode;
    
        /* init timer */
    
    tr->timer.expires = jiffies + (tr->timeout*HZ)/USB_TIMEOUT_HZ;
    add_timer (&tr->timer);
    
        /* tell the HCD to execute it */    
    
    errcode = (*tr->open->host->tr_run) (tr);
    if (errcode < 0)
    {
        gdlf_debug ("Cannot run transaction 0x%p: "
                    "errno/usb %d/%d\n", 
                    tr, errcode, usb_errno);
        if (prev != &tr->snode)     /* restore position in list */
            ts_slist_move (&tr->open->list, &tr->snode, prev);
        return errcode;
    }
    return 0;
}


    /* Abort an EXECUTING transaction
     *
     *  This will abort a transaction, leaving it inserted, but
     * aborted.
     *
     *  HCI has to move to the finished part of the open descriptor
     * list! 
     *
     * Parameters:  
     *
     *  ts_usb_tr*      Pointer to transaction structure.
     *
     * Returns:
     *
     *  int             0 if ok, errno code < 0 on error.
     *
     * Global:
     *
     *  int usb_errno   Unmodified if ok, USB errno code < 0 on
     *                  error.  
     */

int usb_tr_abort (ts_usb_tr* tr)
{
    int errcode;    

    gdlf_fntrace ("usb_tr_abort (0x%p)", tr);

    gdlf_aassert (tr != NULL, return -EINVAL);
    gdlf_aassert (tr->open != NULL, return -EINVAL);
    gdlf_aassert (tr->open->host != NULL, return -EINVAL);
    gdlf_aassert (tr->open->host->tr_abort != NULL, return -ENOSYS);

        /* Do some housekeeping */

    if (tr->status != e_usb_tr_status_executing
        && tr->status != e_usb_tr_status_queued)
    {
        gdlf_err ("Cannot abort transaction 0x%p "
                  "as it is not in any of the running states "
                  "(it is in state %d)\n",
                  tr, tr->status);
        return -EBADFD;
    }

        /*
         * Move from the EXECUTING or QUEUED to the ABORTED
         * state 
         */

    errcode = (*tr->open->host->tr_abort) (tr);
    if (errcode < 0)
        gdlf_debug ("Cannot abort transaction 0x%p: "
                    "errno/usb %d/%d\n", 
                    tr, errcode, usb_errno);
    tr->status = e_usb_tr_status_aborted;
    return 0;
}


    /* Move a transaction from the EXECUTING, ABORTED, DONE or ERROR
     * states to the INSERTED state.
     *
     *  This function is used to 're-use' transactions. When you
     * execute or schedule one transaction and it finishes (ok, with
     * error or aborted), if you want to re-use it, you have first to
     * reset it to the INSERTED state. 
     *
     *  This will reset the status, usb_errno, errno, xfered, errors
     * and ratio fields to their default, sane, values. It will also
     * tell the HCD to prepare for execution the HCI data which is
     * hold within the transaction structure.
     *
     * Parameters:  
     *
     *  ts_usb_tr*      Pointer to transaction structure.
     *
     * Returns:
     *
     *  int             0 if ok, errno code < 0 on error.
     *
     * Global:
     *
     *  int usb_errno   Unmodified if ok, USB errno code < 0 on
     *                  error.  
     */

#warning FIXME: add a re-schedulement time (absolute) for isochronous xfers

    /*
     *  In UHCI, it could be done by interchanging all the values of
     * the TDs, except the link pointer dwords.
     */

int usb_tr_reset (ts_usb_tr* tr)
{
    int errcode;

    gdlf_fntrace ("usb_tr_reset (0x%p)", tr);

    gdlf_aassert (tr != NULL, return -EINVAL);
    gdlf_aassert (tr->open != NULL, return -EINVAL);
    gdlf_aassert (tr->open->host != NULL, return -EINVAL);
        /* We don't check host->tr_reset != NULL, it may be */
    
        /* Do some housekeeping */

    switch (tr->status)
    {
      case e_usb_tr_status_clean:   /* CLEAN -> CLEAN */
        gdlf_debug ("Tried to reset transaction 0x%p, "
                    "which is clean\n", tr);
        return 0;

      case e_usb_tr_status_filled:  /* FILLED -> FILLED */
        gdlf_debug ("Tried to reset transaction 0x%p, "
                    "which is filled\n", tr);
        return 0;

      case e_usb_tr_status_setup:  /* SETUP -> SETUP */
        gdlf_debug ("Tried to reset transaction 0x%p, "
                    "which is setup\n", tr);
        return 0;
        
      case e_usb_tr_status_inserted:/* INSERTED -> INSERTED */
        gdlf_debug ("Tried to reset transaction 0x%p, "
                    "which is already reset\n", tr);
        return 0;

      case e_usb_tr_status_queued:  /* QUEUED -> ABORTED */
      case e_usb_tr_status_executing:/* EXECUTING -> ABORTED */
        errcode = usb_tr_abort (tr);/* abort the transaction */
        if (errcode < 0)
            gdlf_err ("Error aborting transaction 0x%p to "
                      "reset it: errno/usb %d/%d. "
                      "Continuing anyway\n", 
                      tr, errcode, usb_errno);

      case e_usb_tr_status_ok:      /* DONE -> INSERTED */
      case e_usb_tr_status_aborted: /* INSERTED -> INSERTED */
      case e_usb_tr_status_error:   /* ERROR -> INSERTED */
        break;

      default:
        gdlf_apassert (0, return -EINVAL,
                       "Unknown transaction 0x%p state %d\n",
                       tr, tr->status);
    }

        /* 
         *  Move from the ABORTED, DONE or ERROR status to the
         * INSERTED status 
         */

    if (tr->open->host->tr_reset != NULL)
    {
        errcode = (*tr->open->host->tr_reset) (tr);
        if (errcode < 0)
            gdlf_debug ("Cannot reset transaction 0x%p: "
                        "errno/usb %d/%d\n", 
                        tr, errcode, usb_errno);
    }

    tr->status = e_usb_tr_status_inserted;
    tr->usb_errno = 0;
    tr->errno = 0;
    tr->xfered = 0;
    tr->errors = 0;

    return 0;
}


    /* Move from ANY state to the CLEAN state
     *
     *  We move from that ANY state to the FILLED state, and from
     * there, we clean, and we are done.
     *
     * Parameters:  
     *
     *  ts_usb_tr*      Pointer to transaction structure.
     *
     * Returns:
     *
     *  int             0 if ok, errno code < 0 on error.
     *
     * Global:
     *
     *  int usb_errno   Unmodified if ok, USB errno code < 0 on
     *                  error. 
     */

int usb_tr_destroy (ts_usb_tr* tr)
{
    int errcode;

    gdlf_fntrace ("usb_tr_destroy (0x%p)", tr);

    gdlf_aassert (tr != NULL, return -EINVAL);

        /* Do some housekeeping */

    ts_snode_ref (&tr->snode);
    switch (tr->status)
    {
      case e_usb_tr_status_clean:   /* CLEAN -> CLEAN */
        gdlf_debug ("Tried to clean transaction 0x%p, "
                    "which is already clean\n", tr);
        return 0;

      case e_usb_tr_status_setup:   /* SETUP -> FILLED */
      case e_usb_tr_status_inserted:/* INSERTED -> FILLED */
      case e_usb_tr_status_queued:  /* QUEUED -> FILLED */
      case e_usb_tr_status_executing:/* EXECUTING -> FILLED */
      case e_usb_tr_status_ok:      /* DONE -> FILLED */
      case e_usb_tr_status_aborted: /* ABORTED -> FILLED */
      case e_usb_tr_status_error:   /* ERROR -> FILLED */
        errcode = usb_tr_setdn (tr); /* set the transaction down */
        if (errcode < 0)
            gdlf_err ("Error setting down transaction 0x%p "
                      "to clean (destroy) it: errno/usb %d/%d. "
                      "Continuing anyway\n", 
                      tr, errcode, usb_errno);

      case e_usb_tr_status_filled:    /* SETUP -> FILLED */
        break;

      default:
        gdlf_apassert (0, return -EINVAL,
                       "Unknown transaction 0x%p state %d\n",
                       tr, tr->status);
    }

        /*
         *  Ok, let's move from FILLED state to CLEAN state; this
         * implies we have to clean all the data in the
         * structure. This is easy. ts_usb_tr_release() releases the
         * data, we don't clean it, so we can see the results after
         * destroying.
         */

    ts_snode_unref_nd (&tr->snode);
    ts_usb_tr_release (tr);
    return 0;
}


    /* Timer handler for transactions
     *
     *  When this function is called by the timer code, it means the
     * transaction it's got as argument has timed out and has to be
     * moved to error.
     *
     * Parameters:  
     *
     *  ts_uhci_tr_hci* Pointer to the transaction HCI specific data.
     */

void usb_tr_timer (ts_usb_tr* tr)
{
    gdlf_fntrace ("usb_tr_timer (0x%p)", tr);
    gdlf_aassert (tr != NULL, return);

#warning FIXME: lock? kernel? tr?
#warning FIXME: inc the host timeout error count, reset on too hi

    if (tr->status == e_usb_tr_status_executing)
    {
        gdlf_err ("Transaction 0x%p to device 0x%p (%d %d), "
                  "endpoint %d timed out\n",
                  tr, tr->open->dev,
                  tr->open->dev->hub->host->addr,
                  tr->open->dev->addr,
                  tr->open->endp->addr >> 1);
        tr->errno = -ETIMEDOUT;
        tr->usb_errno = -USBE_BUS_TIMEOUT;
        usb_tr_abort (tr);
    }
    return;
}


/* Log:
 *
 * $Log: usbd-tr.c,v $
 * Revision 1.19  1998/11/17 10:04:35  inaky
 * Fixed initial check at tr_setup() for correct open descriptor.
 *
 * Revision 1.18  1998/11/14 17:15:16  inaky
 * Moved the include files to uusbd/include/uusbd, to ease integration
 * into the Linux kernel.
 *
 * Revision 1.17  1998/11/09 21:31:30  inaky
 * Added optimizations when using only one kind of HCI [optimizing at
 * compile time].
 *
 * Revision 1.16  1998/10/14 21:43:06  inaky
 * Some esthetical changes.
 *
 * Revision 1.15  1998/09/24 15:57:47  inaky
 * Added comments.
 *
 * Revision 1.14  1998/09/14 19:47:11  inaky
 * Inserted the tr destructor in the tr.
 *
 * Revision 1.13  1998/08/30 13:51:33  inaky
 * Fixed endpoint number reporting.
 *
 * Revision 1.12  1998/08/27 22:17:46  inaky
 * Added timeouts to be controlled from USBD.
 *
 * Revision 1.11  1998/08/24 02:27:43  inaky
 * Changed name of all enumerations to be prefixed with the enumeration
 * type name.
 *
 * Added the check to actualize open->running when the tr moved from the
 * list was the first one on the queue.
 *
 * Revision 1.10  1998/08/03 13:35:15  inaky
 * Added the queued state where it was missing and moved ts_usb_tr.tr ->
 * ts_usb_tr.list.
 *
 * Revision 1.9  1998/07/30 22:58:30  inaky
 * Moved the unreferenciation of the tr being destroyed before the
 * it being freed.
 *
 * Revision 1.8  1998/07/29 23:34:53  inaky
 * Fixed the destructor to make a reference to the transaction, in order
 * to evitate it's premature destruction.
 *
 * The reset HCI method is now optional. Fixed also the check of the out
 * transfer on interrupt endpoints.
 *
 * Revision 1.7  1998/07/28 23:21:37  inaky
 * Added proper handling of the open descriptor transaction list to keep
 * the order of running, finished and unscheduled transfers.
 *
 * Revision 1.6  1998/07/27 19:26:11  inaky
 * Collected all error states into a single state.
 *
 * Revision 1.5  1998/07/24 15:36:41  inaky
 * Removed the PAUSED state, it was quite useless ...
 *
 * Revision 1.4  1998/07/24 15:32:27  inaky
 * Added more error codes.
 *
 * Revision 1.3  1998/07/23 14:36:06  inaky
 * Added docs for the new states (QUEUED and PAUSED).
 *
 * Renamed the ts_usb_trbuf struct to ts_usb_tr.
 *
 * Added a check to not allow doing OUTs on interrupt pointers.
 *
 * Made the _tr_run() function to really queue the transaction and warn
 * the HCD of it.
 *
 * Added the usb_tr_pause() function to pause a transaction, and the
 * usb_tr_cont() to continue it.
 *
 * Revision 1.2  1998/07/16 14:13:36  inaky
 * Fixed docs for usb_tr_run().
 *
 * Revision 1.1  1998/07/15 23:04:59  inaky
 * Coded all the new transaction-based schedulement code. Seems nicer.
 *
 */

