
/* Universal USB Driver Stack v$Revision: 1.47 $ 
 *
 * (C) 1997 by Iaky Prez Gonzlez <inaky@peloncho.fis.ucm.es>
 *
 * Distributed under the terms of the GNU Public License Version 2
 * See files COPYRIGHT and COPYING at the root distribution directory
 *
 * $Id: usbd.c,v 1.47 1998/11/16 00:00:09 inaky Exp $
 *
 * Changelog at the bottom 
 *
 *
 *  USB Driver
 *  ----------
 *
 * This is supossed to be the layer USBD, as explained in the 10.5
 * section of the USB Spec 1.0. So there are a set of operations that
 * must be supported, which are implemented here. However, some more
 * are still living in this file :).
 *
 */

#include <linux/module.h>
#include <linux/kernel.h>
#include <asm/current.h>

#include <uusbd/usbd.h>

#define GDLF_IDENT          "USBD"
#define GDLF_LOCAL          /* Define if want local settings */
#define GDLF_DEF_LOGLEVEL   9
#define GDLF_ASSERTIONS     /* want assertions (#undef if not) */
#undef GDLF_FNTRACE        /* want function traces */
#define GDLF_LOCATIONS      /* want locations in log messages */
#include <uusbd/gdlf.h>


    /* Check a ts_usb_open structure to have a correct full
     * destination (host, device and endpoint). No range check is
     * performed on the addresses. It is just a `they are set' check.
     *
     * NOTE: this function will be FULLY used on raw access via
     *       /dev/usb*, and is used by the usb_[c]{read,write}*
     *       functions.
     *
     * Parameters:  
     *
     *  ts_usb_open*    Pointer to the structure. NULL is considered
     *                  as an error.
     * Returns:
     *
     *  int             0 if ok, errno code < 0 on error.
     *
     * Global:
     *
     *  int usb_errno   On error, error code.  
     */

int usb_dst_check (const ts_usb_open *open)
{
    gdlf_fntrace ("usb_check_dst (0x%p)", open);

    if (open == NULL)                   /* check we've got data ... */
    {
        gdlf_err ("NULL pointer as ts_usb_open descriptor!\n");
        usb_errno = -USBE_OPEN_NULL;
        return -EINVAL;
    }

    if (open->host == NULL)             /* check we've got a destination host set */
    {
        gdlf_err ("No USB destination bus\n");
        usb_errno = -USBE_OPEN_NO_HOST;
        return -EINVAL;
    }

    if (open->dev == NULL)              /* check we've got a destination device set */
    {
        gdlf_err ("No USB destination device\n");
        usb_errno = -USBE_OPEN_NO_DEV;
        return -EINVAL;
    }

    if (open->endp == NULL)             /* check we've got a destination endpoint set */
    {
        gdlf_err ("No USB destination endpoint\n");
        usb_errno = -USBE_OPEN_NO_ENDP;
        return -EINVAL;
    }

    return 0;   /* If we are here, is 'cause the address seems to be ok */
}


    /* This set of functions are used to manage the destination of USB
     * transactions as described by the open descriptor. They will
     * manage to free any adquired resources.
     *
     * NOTE: When unsetting a destination endpoint, all the
     *       active transactions in that endpoint are aborted and
     *       retired. 
     *
     * Parameters:  
     *
     *  ts_usb_open*    Open descriptor structure
     *
     *  ts_usb_addr*    Destination (compact) address. If NULL all
     *                  destinatios will be released (usecounts will
     *                  be decremented etc...) and address fields will
     *                  be cleared.
     *
     *  t_usb_host      Destination host (bus number) to set to. This
     *                  is the order number of the host
     *                  (ts_usb_host.addr). If negative, the this and
     *                  all the device's address and endpoint will be
     *                  left resetted to an unset value.
     *
     *  t_usb_addr      Destination device's address in current host
     *                  to set to. If negative, the device and endp
     *                  address wil be left resetted to an unset
     *                  value.
     *
     *  t_usb_endp      Destination endpoint in interface. If
     *                  negative, the endpoint address will be left
     *                  resetted in an unset value. Note the address
     *                  you should give here is the one on THE
     *                  ENDPOINT STRUCTURE (ts_usb_endp.addr).
     *
     * Returns:
     *
     *  int             0 if ok, < 0 on error (errno code). In case
     *                  of error, the destination of the ts_usb_open
     *                  structure is kind of strange, as the error may have
     *                  happened on the host, device or endpoint stage
     *                  of the address setting. Then, until the error,
     *                  all the previous stages will be set up, but
     *                  the erroneous one will be left in initialized
     *                  state (not in the old one, to evitate confusions).
     *
     * Global:
     *
     *  int usb_errno   On error, error code.  
     */

int usb_dst_endp_unset (ts_usb_open *open)
{
    ts_usb_endp *endp;

    gdlf_fntrace ("usb_dst_endp_unset (0x%p)", open);
    gdlf_aassert (open != NULL, return -EINVAL);

    if (open->endp == NULL)
        return 0;
    
        /* 
         * no need to write lock open, nobody will add new xfers
         * as endp is unset. FIXME: posible problem if someone
         * tries to reopen `open' now.
         */

    endp = open->endp;
        
    if (endp->desc != NULL)     /* if there is an open descriptor twinkling */
    {
        ts_usb_tr *tr, *nx;

            /* 
             * We do no locking on the list, as we want
             * usb_tr_setdn() to be able to remove the xfer
             * from the queue. Also it is not needed, as we keep
             * the pointers who allows us walk the list. No one is
             * going to append new xfers to the list as we have
             * set endp to NULL before starting.
             *
             * Also note there's a suboptimal removing of the
             * list, as we aways know the previous item and
             * there's no way for usb_tr_setdn() to know.
             * FIXME.
             */
        
        tr = (ts_usb_tr*)ts_slist_head (&open->list);
        while (tr)                      /* and while we have pending transactions */
        {
            nx = (ts_usb_tr*) ts_slist_next (&tr->snode);/* aborting removes from list! */
            usb_tr_setdn (tr);          /* abort, remove and set it dn */
            tr = nx;                    /* and go next */
        }
        if (open->dev->active && endp->addr != 0)
            usb_configuration_unref (open->dev->active);
        atomic_dec (&open->list.usecnt);
        endp->desc = NULL;              /* free the reference to the descriptor from the endpoint */
    }

        /*
         *  Note we don't clear open->endp until the end, as
         * usb_tr_setdn will need it. 
         */

    open->endp = NULL; 
#if defined (CONFIG_USB_OHCI) || defined (CONFIG_USB_OHCI_MODULE)
    if (open->host->endp_setdn)     /* Set down endpoint's HCI data */
        return (*open->host->endp_setdn) (open->dev, endp);
    else
        return 0;
#else
    return 0;
#endif
}


int usb_dst_endp_set (ts_usb_open* open, t_usb_endp endpn)
{
    t_usb_endp addr;
    ts_usb_endp *endp = NULL;
    
    gdlf_fntrace ("usb_dst_endp_set (0x%p, %d)", open, endpn);
    gdlf_aassert (open != NULL, return -EINVAL);
    gdlf_aassert (open->dev != NULL, return -EINVAL);
    gdlf_aassert (open->host != NULL, return -EINVAL);
    gdlf_aassert (endpn >= -1 && endpn < USB_ENDPS + 2, return -EINVAL);

    usb_dst_endp_unset (open);                  /* clean previous destination endpoint */

    if (endpn < 0)                              /* Leave it resetted? ... */ 
        return 0;                               /* ... ok, leave it */

    if (endpn == 0)                             /* default endpoint */
        endp = &open->dev->endp0;               /* here it is :) */
    else
    {
        gdlf_aassert (open->dev->active != NULL, return -EINVAL);
        addr = (endpn >> 1) - 1;
        if (addr >= open->dev->active->pendps)
            gdlf_err ("Endpoint addres out of range, maybe prev error\n");
        else                                    /* get the pointer to the endpoint structure */
            endp = open->dev->active->pendp[addr];
        if (endp == NULL)                       /* Huh? isn't it used */
        {
            gdlf_err ("Endpoint address %d unexistant "
                      "at device 0x%p (%d %d)\n", 
                      endpn, open->dev, open->host->addr, 
                      open->dev->addr);
            usb_errno = -USBE_ADDR_UNKNOWN;
            return -ENXIO;
        }
    }    

    if (endp->desc == NULL)                     /* If the endpoint is not used ... */
    {
#if defined (CONFIG_USB_OHCI) || defined (CONFIG_USB_OHCI_MODULE)
        if (open->host->endp_setup)             /* Setup endpoint's HCI data */
        {
            int errcode;
            
            errcode = (*open->host->endp_setup) (open->dev, endp);
            if (errcode < 0)
                return errcode;
        }
#endif
        open->endp = endp;
        open->endp->desc = open;                    /* indicate which ts_usb_open is using it */
        atomic_inc (&open->list.usecnt);
        if (open->dev->active && endpn != 0)
            usb_configuration_ref (open->dev->active);
        return 0;
    }
    else                                        /* ... but if it is used ... */
        return -EBUSY;                          /* ... it is busy */
}


int usb_dst_dev_unset (ts_usb_open *open)
{
    gdlf_fntrace ("usb_dst_dev_unset (0x%p)", open);
    gdlf_aassert (open != NULL, return -EINVAL);

    usb_dst_endp_unset (open);      /* release endpoint */
    if (open->dev != NULL)          /* release any device reference */
    {
        usb_device_unref (open->dev, NULL);
        open->dev = NULL;
    }
    return 0;
}


int usb_dst_dev_set (ts_usb_open* open, t_usb_addr addrn)
{
    gdlf_fntrace ("usb_dst_dev_set (0x%p, %u)", open, addrn);
    gdlf_aassert (open != NULL, return -EINVAL);
    gdlf_aassert (open->host != NULL, return -EINVAL);

    usb_dst_dev_unset (open);       /* clean previous destination endpoint and device */

    if (addrn < 0)                  /* Leave it resetted? ... */
        return 0;                   /* ... ok, leave it */
                                    
        /*
         * Get the pointer to the device's structure 
         */

    open->dev = usb_device_ptr_hp (open->host, addrn);
    if (open->dev == NULL)
    {
        gdlf_err ("Device address %i unexistant at bus %i\n", 
                  addrn, open->host->addr);
        usb_errno = -USBE_ADDR_UNKNOWN;
        return -ENXIO;
    }
    usb_device_ref (open->dev); /* increment the device use count */
    return 0;
}


int usb_dst_host_unset (ts_usb_open *open)
{
    gdlf_fntrace ("usb_dst_host_unset (0x%p)", open);
    gdlf_aassert (open != NULL, return -EINVAL);

    usb_dst_dev_unset (open);       /* release endpoint and device */
    if (open->host != NULL)         /* release any host reference */
    {
        usb_host_unref (open->host, NULL);
        open->host->reference (NULL, -1);  /* we don't use the host ... */
        MOD_DEC_USE_COUNT;
        open->host = NULL; 
    }
    return 0;
}    


int usb_dst_host_set (ts_usb_open* open, t_usb_host hostn)
{
    gdlf_fntrace ("usb_dst_host_set (0x%p, %u)", open, hostn);

    gdlf_apassert (open != NULL,
                   {
                       usb_errno = -USBE_OPEN_NULL;
                       return -EINVAL;
                   },
                   "Tried to operate on a NULL open descriptor\n");

    usb_dst_host_unset (open);      /* clean previous destination endpoint, device and host */

    if (hostn < 0)                  /* Leave it resetted? ... */
        return 0;                   /* ... ok */

        /*
         * WARNING! Any change in this procedure _MUST_ be also
         *          updated in usb_device_attach(), as it has to
         *          manually open the device->endp0d endpoint
         *          descriptor. 
         */

    open->host = usb_host_ptr (hostn);  /* set the host pointer */
    if (open->host == NULL)             /* Not found? ... */
        return -ENXIO;                  /* ... bad address */
    usb_host_ref (open->host);          /* increment the host use count */
    open->host->reference (NULL, 1);    /* we use the host ... */
    MOD_INC_USE_COUNT;
    return 0;
}


int usb_dst_unset (ts_usb_open* open)
{
    gdlf_fntrace ("usb_dst_unset (0x%p)", open);

    return usb_dst_host_unset (open);
}


int usb_dst_set (ts_usb_open* open, const ts_usb_addr* addr)
{
    int errcode;

    gdlf_fntrace ("usb_dst_set (0x%p, 0x%p)", open, addr);
    gdlf_aassert (open != NULL, return -EINVAL);
    gdlf_aassert (addr != NULL, return -EINVAL);

    errcode = usb_dst_host_set (open, addr->host);
    if (errcode != 0)
        return errcode;
    
    errcode = usb_dst_dev_set (open, addr->device);
    if (errcode != 0)
        return errcode;
    
    return usb_dst_endp_set (open, addr->endp);
}


    /* Open a descriptor for accessing the USB. If succesful, the
     * module use count will be incremented.
     *
     * Parameters:  
     *
     *  const ts_usb_addr*  
     *
     *                  Destination address to open. It may be
     *                  uninitialized, in which case, there will be no
     *                  real destination set. However, that stuff may
     *                  be set via the usb_set_dst* functions.
     *
     *  ts_usb_open*    Pointer to a ts_usb_open, clean, that will be
     *                  overwritten without previous free of
     *                  resources.
     *
     * Returns:
     *
     *  int             0 if ok, errno code < 0 on error.
     *
     * Global:
     *
     * int usb_errno    On error, error code.
     */

int usb_open (const ts_usb_addr* addr, ts_usb_open* open)
{
    int errcode;        /* result of address evaluation */

    gdlf_fntrace ("usb_open (0x%p, 0x%p)", addr, open);
    
    gdlf_apassert (addr != NULL,
                   { 
                       usb_errno = -USBE_ADDR_NULL; 
                       return -EINVAL; 
                   },
                   "Cannot open a NULL ts_usb_addr!\n");

    gdlf_apassert (open != NULL,
                   {
                       usb_errno = -USBE_OPEN_NULL;
                       return -EINVAL;
                   },
                   "Cannot open into a NULL ts_usb_open!\n");

    ts_usb_open_clean (open);
    errcode = usb_dst_set (open, addr); /* set the destination address */
    if (errcode < 0)                    /* If there's an error ... */ 
        return errcode;                     /* return error code */
    return 0;                           /* ok, done */
}


    /* Close a descriptor for using the USB. If successful, it will
     * decrement the module use count.
     *
     * Parameters:  
     *
     *  ts_usb_open*    Pointer to the descriptor data structure.
     *
     * Returns:
     *
     *  int             0 if ok, errno code < 0 on error.
     *
     * Global:
     *
     *  int usb_errno   On error, error code.  
     */

int usb_close (ts_usb_open* open)
{
    gdlf_fntrace ("usb_close (0x%p)", open);
    
    if (open == NULL)                       /* Bad pointer? */
    {
        gdlf_err ("Tried to close a NULL ts_usb_open!\n");
        usb_errno = -USBE_OPEN_NULL;
        return -EINVAL;
    }
    
    usb_dst_unset (open);   /* clear destination (will abort transfers) */
    return 0;
}


    /*  Become a master client.
     *
     *  This grants the privilege to set the SOF for the host the
     * device is connected to. In order to let other devices set the
     * SOF, you must drop the master privilege by using
     * usb_device_master_drop().
     *
     * NOTE: If your driver will use USB bus mastering, you should
     *       make available the posibility to work without master
     *       privilege, in advance for any other device also
     *       requesting it with higher priority. This should be
     *       selectable by user.
     *
     * Parameters:  
     *
     *  ts_usb_device*  Pointer to the device structure to make
     *                  master. 
     *
     * Returns:
     *
     *  int             0 if ok, < 0 if master privilege not granted
     *                  (-EBUSY if it is already used by another
     *                  device), errno code < 0 on error.
     *
     * Global:
     *
     *  int usb_errno   On error, error code.  
     */

int usb_master_become (ts_usb_device* device)
{
    ts_usb_host *host;
    
    gdlf_fntrace ("usb_master_become (0x%p)", device);

    gdlf_apassert (device != NULL,
                   { 
                       usb_errno = -USBE_DEVICE_NULL;
                       return -EINVAL;
                   },
                   "Cannot make master a NULL ts_usb_device\n");
    
        /* FIXME: public fn, make checks */

    host = device->hub->host;
    if (host->master == NULL)
    {
        write_lock (&host->lock);
        write_lock (&device->lock);
        host->master = device;
        usb_device_ref (device);
        write_unlock (&device->lock);
        write_unlock (&host->lock);
        return 0;
    }
    usb_errno = -USBE_DEVICE_MASTER_BUSY;
    gdlf_info ("Bus mastering denied to device 0x%p "
               "(master is 0x%p)\n", 
               device, host->master);
    return -EBUSY;
}


    /*  Drop master privileges.
     *
     * Parameters:  
     *
     *  ts_usb_device*  Pointer to the device structure to drop
     *                  privileges.
     *
     * Returns:
     *
     *  int             0 if ok, < 0 (-EPERM) if you're not master,
     *                  errno code < 0 on error.
     *
     * Global:
     *
     *  int usb_errno   On error, error code.  
     */


int usb_master_drop (ts_usb_device* device)
{
    ts_usb_host *host;

    gdlf_fntrace ("usb_master_drp (0x%p)", device);
    
    gdlf_aassert (device != NULL, return -EINVAL);
    
        /* FIXME: public fn, make checks */

    host = device->hub->host;
    
    if (host->master != device)
    {
            /* No need to set usb_errno */
        gdlf_debug ("Device 0x%p isn't master, cannot drop it "
                    "(master is 0x%p)\n", 
                    device, host->master);        
            /* no error, a it simply, is not */
        return ENOENT;
    }
    else
    {
        write_lock (&host->lock);
        write_lock (&device->lock);
        host->master = NULL;
        usb_device_unref (device, NULL);
        write_unlock (&device->lock);
        write_unlock (&host->lock);
    }
    return 0;
}


    /*  Set the Start of Frame for a given host.
     *
     *  This function will set the start of frame for a given
     * host. You need master privilege for doing this, as with it
     * you're changing the timings of the USB frames, and maybe some
     * other devices depend on them.
     *  
     * Parameters:  
     *
     *  ts_usb_device*  Pointer to the device that gives you right to
     *                  change the SOF. Also used to identify which
     *                  host's SOF you want to change.
     *
     * Returns:
     *
     *  int             0 if ok, errno code < 0 on error.
     *
     * Global:
     *
     *  int usb_errno   On error, error code.  
     */

int usb_sof_set (ts_usb_device* device, unsigned frame_length)
{
    int errnocode;
    ts_usb_host *host;

    gdlf_fntrace ("usb_sof_set (0x%p, %u)", device, frame_length);
    
    gdlf_apassert (device != NULL,
                   {
                       usb_errno = -USBE_DEVICE_NULL;
                       return -EINVAL;
                   },
                   "Tried to operate on a NULL device\n");
    
    host = device->hub->host;

    if (host->master != device)
    {
        usb_errno = -USBE_DEVICE_NOT_MASTER;
        gdlf_info ("Device 0x%p isn't master and cannot "
                   "set SOF (master is 0x%p)\n", 
                   device, host->master);
        return -EPERM;
    }
    
    write_lock (&host->lock);
    errnocode = host->setsof (host, frame_length);
    if (errnocode == 0)
        host->sof = frame_length;
    write_unlock (&host->lock);
    return errnocode;
}


/* Log:
 *
 * $Log: usbd.c,v $
 * Revision 1.47  1998/11/16 00:00:09  inaky
 * Added the errcode declaration, needed when compiling OHCI support.
 *
 * Revision 1.46  1998/11/14 17:15:06  inaky
 * Moved the include files to uusbd/include/uusbd, to ease integration
 * into the Linux kernel.
 *
 * Revision 1.45  1998/11/09 21:32:01  inaky
 * Changed the name of the configuration pointer array to pcfg. Moved all
 * device ref/unref to usb_device_[un]ref(). Moved all host ref/unref to
 * usb_host_[un]ref().
 *
 * Revision 1.44  1998/10/31 02:14:06  inaky
 * Added the endpoint setting up on usb_close() and usb_open(), if OHCI
 * is selected.
 *
 * Revision 1.43  1998/10/22 00:03:52  inaky
 * Fixed destruction of transactions on endpoint close. Moved
 * transactions to be set down when closing endpoints. This is done so
 * because though a transaction could get along with only open->host
 * being open to be set down, it almost makes no sense.
 *
 * Revision 1.42  1998/09/14 19:46:46  inaky
 * Moved to the new dynamic scheme for endpoint pointers.
 *
 * Revision 1.41  1998/08/30 13:52:30  inaky
 * Moved to abort all transfers when closing on usb_close(). Update the
 * cfg use counts. Severed assertions.
 *
 * Revision 1.40  1998/08/28 23:48:03  inaky
 * Removed traces.
 *
 * Revision 1.39  1998/08/03 13:35:41  inaky
 * Moved ts_usb_tr.tr -> ts_usb_tr.list. Severed some assertions to make
 * the code clearer.
 *
 * Revision 1.38  1998/07/30 17:54:15  inaky
 * Changed a function to use another, as we had the pointer the other
 * would otherwise look for.
 *
 * Revision 1.37  1998/07/23 14:36:22  inaky
 * Renamed the ts_usb_trbuf struct to ts_usb_tr.
 *
 * Revision 1.36  1998/07/16 14:13:47  inaky
 * Moved from the old read/write interface to the new, request-based
 * interface.
 *
 * Revision 1.35  1998/07/14 22:01:05  inaky
 * Fixed some bits of usb_dst_endp_unset() which were completely broken
 * and prone to a NULL derefence exception. Also some bits of the endp
 * set code and added BIG warnings at the open sequences so the one at
 * usb_device_attach() doesn't get unsynchronized if we change these
 * ones.
 *
 * Also moved the error given by usb_master_drop() if the calling device
 * is not a master. Instead of returning error (<0), it prints a debug
 * message and returns a positive (non error, just warning) ENOENT.
 *
 * Revision 1.34  1998/07/12 22:56:07  inaky
 * Fixed the endp unsetting functions to _clear_ the endp member _after_
 * cleaning :)
 *
 * Moved the reference count inc/dec'ing to when the host is successfully
 * referenced/dereferenced. Also added calls to say we reference the
 * host, so the HCI module get's also it's reference count update.
 *
 * Fixed the fn trace name on usb_close().
 *
 * Revision 1.33  1998/06/20 21:54:54  inaky
 * Massive changes to split the old monolithic hubd driver into a hub
 * manager (hub wrappers and daemon) and a dynamic driver, which can be
 * replaced and its different parts reused.
 *
 * Revision 1.32  1998/03/28 19:37:12  inaky
 * Removed inclusion of usb-methods.h; fixed to use the new transfer
 * lists and use counts for ts_usb_open structures.
 *
 * Revision 1.31  1998/02/12 16:35:11  inaky
 * Finished the device ops, and moved synch_frame getting to usbd-endp,
 * where it belongs. Fixed some stuff here and there, and starting to get
 * worried on locks ... I'm gonna have to debug *lots* of deadlocks.
 *
 * Revision 1.30  1998/02/12 14:26:12  inaky
 * usb_dst_check(): No need to do more checkings on pointer consistency,
 * as if that happens, it is screwed up any way and will appear sooner or
 * later. This function has to be just a fast check.
 *
 * Revision 1.29  1998/02/10 17:37:23  inaky
 * Moved the direct atomic use count manipulations to use
 * ts_[N]snode_[un]ref().
 *
 * Revision 1.28  1998/02/09 13:32:28  inaky
 * Moved all the use counts to atomic
 *
 * Revision 1.27  1998/01/11 15:44:20  inaky
 * Finished functions from the descriptor, inteface and endpoint modules,
 * fixed docs, pretty printed and removed craps.
 *
 * Revision 1.26  1998/01/08 23:41:48  inaky
 * Removed the compact address from the ts_usb_open descriptors, they
 * were quite useless. Also added the open descriptor to the device
 * structure.
 *
 * Revision 1.25  1998/01/08 12:19:13  inaky
 * Added copyright strings to all important files
 *
 * Revision 1.24  1998/01/03 00:12:52  inaky
 * Modified the usb_open function not to allocate descriptors itself.
 *
 * Revision 1.23  1998/01/02 20:15:53  inaky
 * Fixed some bits in the destination selection code, to allow using any
 * negative value as invalid and to make enpoint numbers 0 and 1 both map
 * to endpoint 0.
 *
 * Revision 1.22  1998/01/02 19:39:38  inaky
 * Added the definition for endpoint zero to the device structure and
 * removed it from the list of endpoints kept at each configuration. Also
 * updated it on places where needed and removed the concept of basic
 * configuration, as is no longer needed.
 *
 * Revision 1.21  1998/01/02 19:28:59  inaky
 * Changed the open endp function to use for endpoint 0 the endp0 member
 * in the device structure.
 *
 * Revision 1.20  1997/12/30 20:15:45  inaky
 * Moved everything to the new lists scheme. Untested yet.
 *
 * Revision 1.19  1997/12/29 00:25:41  inaky
 * Fixed unimplemented functions to return -ENOSYS and moving the khubdd
 * to use signals as commands.
 * Also added function traces on most files that had them disabled (donkey!)
 *
 * Revision 1.18  1997/12/27 17:25:36  inaky
 * Finished moving to GDLF. Deprecated usb_printk and __LOCATION__
 *
 * Revision 1.17  1997/12/27 16:44:25  inaky
 * Moving to GDLF and fixinf docs. Removing more cruft.
 *
 * Revision 1.16  1997/12/27 02:24:04  inaky
 * Moving everything to use the GDLF for logging and debug checkings.
 * Removed the global array of USB hosts and moved it to a linked
 * list. Added an address field to each USB host controller.
 * Removed lots of cruft that were accumulated and useless. Still some
 * missing, but may be useful someday.
 *
 * Revision 1.15  1997/12/22 02:44:33  inaky
 * Trying to move all the documentation to the source files from the
 * header files, as it was duplicated.
 *
 * Revision 1.14  1997/12/08 02:41:02  inaky
 * Fixed some stuff of the #includes, modified the makefile targets to be
 * more flexible, tried to finish the device configuration section.
 *
 * Revision 1.13  1997/12/06 00:19:42  inaky
 * Almost finished the usbd/host, usbd/cfg and a lot of meat from
 * usbd/device. Also added the driver definition. Got to polish it yet
 * ...
 *
 * Revision 1.12  1997/11/29 20:21:49  inaky
 * Documented and almost fixes the read/write/abort API, documented a
 * damn lot and implemented some stuff
 *
 * Revision 1.11  1997/11/29 00:46:31  inaky
 * Documented almost all the hub basic API, added spin locks to
 * usb_host_register() and some more stuff. Going for more.
 *
 * Revision 1.10  1997/10/02 22:29:37  inaky
 * Finally managed to understand the workings of the MODVERSIONS (ok, not
 * exactly the working, but how to make it work).
 *
 * Now I can go to sleep happy :) -- I know, I'm really sick --
 *
 * Revision 1.9  1997/10/02 21:29:04  inaky
 * Remade all the directory structure to make it cleaner when the USB
 * class drivers come.
 *
 * Revision 1.8  1997/10/01 21:28:13  inaky
 * Re-making a lot of stuff in the hadnling of host, hub and device
 * initialization and handling, following a more correct approach of
 * registering/unregistering. Also moved the usb_host array to be an
 * array of pointers, so dynamic reallocation of it is easier, and the
 * host structure needs not to be moved in memory. The same applies to
 * the device arrays for each host.
 *
 * Revision 1.7  1997/09/08 10:54:51  inaky
 * Modified the generic functions to gain some performance, also some other checks
 *
 * Revision 1.6  1997/09/01 19:14:44  inaky
 * Backup check in
 *
 * Revision 1.5  1997/07/03 20:00:06  inaky
 * Moved all the hw dependencies to the HCD module, modified the
 * controller detections stuff to use just PCI means (okay, other buses
 * to come, but when they come). General fixes, but no really new deep
 * enhancements.
 *
 * Revision 1.4  1997/06/22 19:54:32  inaky
 * Seems that the moving of all the hw dependants stuff to the HCI
 * modules is done. Got to test it yet, but at least it compiles cleanly.
 *
 * Revision 1.3  1997/06/19 20:38:18  inaky
 * Moving al hardware dependant stuff to the HCD modules
 *
 * Revision 1.2  1997/05/03 21:22:50  inaky
 * Adding a lot of code, with the only objetive of adding the first
 * transfer type: control :) Still there's a lot of stuff to do, mostly
 * concerning data structures handling.
 *
 * Revision 1.1  1997/04/14 16:56:02  inaky
 * *** empty log message ***
 *
 * Revision 1.7  1997/04/12 16:47:49  inaky
 * Added the code for buffering reads/writes to/from userland
 *
 * Revision 1.6  1997/04/10 21:30:48  inaky
 * Fixed the file code, so it opens/closes ok; also reads/writes ok ... ok, just theory
 *
 * Revision 1.5  1997/04/06 22:56:44  inaky
 * Added the functions to set destinations of an USB open descriptor
 * (ts_usb_open); however, I haven't tested it yet ... bugs flowing!
 *
 * Revision 1.4  1997/04/06 21:14:31  inaky
 * Still with the definition of the access interface. More or less it is
 * taking an stable form.
 *
 * Revision 1.3  1997/03/31 20:24:46  inaky
 * Pfff ... a lot. Mostly added descriptors stuff (HUB class still is
 * missing a lot) and I think I've got a definitive idea of how to
 * implement the storing of configurations and interfaces.
 *
 * Revision 1.2  1997/03/09 23:10:31  inaky
 * *** empty log message ***
 *
 * Revision 1.1  1997/02/22 22:25:44  inaky
 * *** empty log message ***
 *
 *
 */

